/*
 * File repeater.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef REPEATER_H_
#define REPEATER_H_

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#define RPTR_LCLCTX "LOCALHOST"
#define RPTR_LCLFEC "__LCLRPTRFEC__"
#define RPTR_LCLSRV "__LCLRPTRSRV__"
#define RPTR_DEFAULT_PORT 101
#define RPTR_MAX_NDEVICES 5000
#define RPTR_POLLRATE 1000
#define RPTR_EQM "RPTEQM"

/* some alarms : */
#define target_value_unavailable  512
#define repeater_not_synchronized 513

#define MAX_SCHEDULED_PROPERTIES 25
extern NAME64 schPrpLst[MAX_SCHEDULED_PROPERTIES];
extern int nSchPrps;
extern int rpPollingRate;
extern int rpPortOffset;
extern int gTruncateDeviceNames;
extern int hasExclusiveRead;
extern int ignoreListeningErrors;
extern int repeatHistories;
extern int isSyncRepeater;
extern int isLocalRepeater;
extern char thisFec[FEC_NAME_SIZE+4];
extern char thisContext[CONTEXT_NAME_SIZE+4];
extern char thisFailoverMaster[EXPORT_NAME_SIZE+4];
extern char thisSlaveMaster[EXPORT_NAME_SIZE+4];
int rpPreSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer);
int rpPostSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer);

extern int gRptStartTime;
extern int gTgtStartTime;
void cbStartTime(int id, int cc);
void monitorTarget(void);


#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* REPEATER_H_ */
