/*
 * alaeqm.c    REVISION HISTORY:
 * Generated by SRVWIZARD on Fri Mar 11 09:12:47 2005
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "alaeqm.h"
#if defined(UNIX)
#include "dirent.h"
# define DBPATH "../database/"
#elif defined(WIN32)
#include "direct.h"
# define DBPATH "..\\database\\"
#endif

#define STRING_CONTAINS 0
#define STRING_MATCHES 1
typedef struct
{
  char state[64];
  int  compareType;
} StateType;

UINT32 g_notreadycount;
UINT32 g_isready;
UINT32 g_numalarms;
char gCasServer[64];
int gNumAlarmSystems = 0;
int gNumAlarmGroups = 0;
NAME64DBLDBL gAlarmSystems[MAX_ALARM_SYSTEMS];
int gAlarmCounter[MAX_ALARM_SYSTEMS];
int gAlarmCounterRunning[MAX_ALARM_SYSTEMS];
int gNumberOfAlarms[MAX_ALARM_SYSTEMS];
LNGINT gNumberOfAlarmsX[MAX_ALARM_SYSTEMS];
ALARM_GRP gAlarmGroups[MAX_ALARM_SYSTEMS];
int gIsRunning = 0;
int gIsOfficiallyRunning = 0;
int gIsOfficiallyProblems = 0;
int nNonRunningStates = 0;
int nRunningStates = 0;
int nProblemStates = 0;
char nonRunningState[MAX_NONRUNNING_STATES][64];
StateType isRunningState[MAX_RUNNING_STATES];
StateType isProblemState[MAX_RUNNING_STATES];
char gDeclaredState[64] = {0};
char gDeclStateSrv[128];
char gDeclStatePrp[64];
MTBP mtbpInfoSet[MAX_ALARM_SYSTEMS];
TimeSpan gAlarmSystemTimeSpan[MAX_ALARM_SYSTEMS];
time_t gMTBPEpochBegin = 0;
time_t mtbp_threshold = 900; // 15 minutes


//Threshold.RecoveredFromSubsystemFailure (the same subsystem with 'ready' - available time)
//Threshold.RecoveredFromFailure (any other subsystem with available time)
//Threshold.Failure (time in failure state to be regarded as failure)
//Threshold.Available (time between failures) = > must be in available state at least this time
//Threshold.UnidentifiedFailure (time to regard unidentified problems as a failure)
//Threshold.Available.MTFB (time between failures concerning MTBF calc.) = > must be in available state at least this time
#define NTHRESHOLD_TYPES 6
#define THR_TYPE_FAILURE_INDEX 0
#define THR_TYPE_AVAILABLE_INDEX 1
#define THR_TYPE_RECOVEREDFROMFAILURE_INDEX 2
#define THR_TYPE_RECOVEREDFROMSUBSYSTEMFAILURE_INDEX 3
#define THR_TYPE_UNIDENTIFIEDFAILURE_INDEX 4
#define THR_TYPE_AVAILABLE_MTBF_INDEX 5
int gThresholds[NTHRESHOLD_TYPES] = {
  5, 1000, 10, 3000, 5, 120
};
NAME64 gThresholdTypes[NTHRESHOLD_TYPES] = {
  "Failure",
  "Available",
  "RecoveredFromFailure",
  "RecoveredFromSubsystemFailure",
  "UnidentifiedFailure",
  "Available.MTBF"
};

#define admin_rights_required 512

int logproblem(char *text,...);
time_t getTimeStampFromLogEntry(char *str); // use this routine from the TINE kernel

#define CORRECTIONSFILE "corrections.csv"
#define BLAMEFILE "blame.csv"
#define CORRLOGFILE "corrections.log"
#define MTBFLOGFILE "mtbf.log"
#define LNKLOGFILE "link-updates.log"

typedef struct StateCorrection
{
  double start;
  double stop;
  int counts;
  int to;  // 0 or 1 (from is !to); '1' => 'Problems' or 'Blame', '0' => 'Available'
  char system[64];
  char state[64];
  char startStr[64];
  char stopStr[64];
  char usr[16];
  char cmt[256];
} StateCorrection;
#define MAX_CORRECTION_PER_DAY 2048

StateCorrection stateCorrection[MAX_CORRECTION_PER_DAY];

CSVCOLUMN csvStateCorrectionsFileColumns[] =
{
  { "System","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,system),NULL,0 },
  { "Start","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,startStr),NULL,0 },
  { "Stop","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,stopStr),NULL,0 },
  { "To","",0,CF_TEXT,64,(BYTE *)OFFSETIN(StateCorrection,state),NULL,0 },
  { "User","",0,CF_TEXT,16,(BYTE *)OFFSETIN(StateCorrection,usr),NULL,0 },
  { "Comment","",0,CF_TEXT,256,(BYTE *)OFFSETIN(StateCorrection,cmt),NULL,0 },
};
CSVDB csvStateCorrectionsFileDb =
{
  csvStateCorrectionsFileColumns,
  sizeof(csvStateCorrectionsFileColumns)/sizeof(csvStateCorrectionsFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(StateCorrection),
  0
};
int getStateAssignment(char *state)
{
  if (state == NULL) return 0;
  switch (toupper(state[0]))
  {
    case 'P':   // "Problems"
    case 'B':   // "Blamed"
      return 1;
    default: 
      break;
  }
  return 0;
}
// the property to get the list of stored correction list over a selected time range
int readStateCorrections(char *filename, time_t start, time_t stop, StateCorrection *list, int *num)
{
  int n=0, i, cc=0, length = *num;
  char fn[128];
  StateCorrection *stcr = stateCorrection;
  time_t t;
  struct tm *lt;
  int dd, mm, yy, yde;
  
  lt = localtime(&stop); yde = lt->tm_yday;
  for (t = start; t < stop+86400; t += 86400)
  { // Loop over the days involved
    lt = localtime(&t);
    dd = lt->tm_mday; mm = lt->tm_mon; yy = lt->tm_year;
    if (yde < lt->tm_yday) break;
    sprintf(fn,"%s%04d/%02d/%02d/",DBPATH,1900+yy,mm+1,dd);
    csvStateCorrectionsFileDb.siz = MAX_CORRECTION_PER_DAY;
    if ((cc=csvReadFile(fn,filename,&csvStateCorrectionsFileDb,(void **)&stcr)) != 0) continue;
    for (i=0; i<csvStateCorrectionsFileDb.siz && n < length; i++)
    { // check > start && < stop
      stcr[i].stop = (double)getTimeStampFromLogEntry(stcr[i].stopStr);
      stcr[i].start = (double)getTimeStampFromLogEntry(stcr[i].startStr);
      stcr[i].counts = (int)(stcr[i].stop - stcr[i].start);
      if (stcr[i].stop < (double)start) continue;
      if (stcr[i].start > (double)stop) continue;
      stcr[i].to = getStateAssignment(stcr[i].state);
      list[n++] = stcr[i];
    }
    if (n >= length) break;
  }
  *num = n;
  return 0;
}
// the property to get the corrections over a selected time range
// overlap of different alm state corrections is allowed
// overlap of same alm state is not allowed
int getStateCorrectionCounts(char *filename,time_t start, time_t stop, int *corcnts, int length)
{ // length typically = ndeclstates
  int i, cc=0, counts, idx, sgn;
  char fn[128];
  StateCorrection *stcr = stateCorrection;
  time_t t;
  struct tm *lt;
  int dd, mm, yy, yde;
    
  memset(corcnts, 0, length*sizeof(int));
  lt = localtime(&stop); yde = lt->tm_yday;
  for (t = start; t < stop+86400; t += 86400)
  { // Loop over the days involved
    lt = localtime(&t);
    if (yde < lt->tm_yday) break;
    dd = lt->tm_mday; mm = lt->tm_mon; yy = lt->tm_year;
    sprintf(fn,"%s%04d/%02d/%02d/",DBPATH,1900+yy,mm+1,dd); // dbpath ?
    csvStateCorrectionsFileDb.siz = MAX_CORRECTION_PER_DAY;
    if ((cc=csvReadFile(fn,filename,&csvStateCorrectionsFileDb,(void **)&stcr)) != 0) continue;
    for (i=0; i<csvStateCorrectionsFileDb.siz; i++)
    { // reduce array to just corrections in given interval
      // fill in UTC values for start, stop
      stcr[i].stop = (double)getTimeStampFromLogEntry(stcr[i].stopStr);
      stcr[i].start = (double)getTimeStampFromLogEntry(stcr[i].startStr);
      stcr[i].to = getStateAssignment(stcr[i].state);
      sgn = stcr[i].to ? 1 : -1;
      stcr[i].counts = (int)(stcr[i].stop - stcr[i].start + 1) * sgn;
      // if 'this' start/top fully contained then : as is
      if (stcr[i].stop < start || stcr[i].start > stop) 
      { // if fully excluded then skip
        continue;
      }
      if (stcr[i].stop <= (double)stop && stcr[i].start >= (double)start) 
      { // if fully contained then use as is ...
        counts = stcr[i].counts;
      }
      else
      { // need to inperpolate ...
        if (stcr[i].start < (double)start) counts = (int)(stcr[i].stop - start);
        else counts = (int)(stop - stcr[i].start);
        counts++; // include the stop second
        counts *= sgn;
      }
      // find each state index from the 'fromState' and 'toState' and accumulate the counts
      idx = getAlarmIndexFromSystemNumber(stcr[i].system);
      if (idx < 0 || idx >= length) continue;
      corcnts[idx] += counts;
    }
  }
  return 0;
}
int getGroupCorrectionCounts(time_t start, time_t stop, int *corcnts, int length)
{ // length typically = gNumAlarmGroups
  int sttcnts[256]; // this should always be enough
  int i, k, cc = 0;
  memset(corcnts, 0, length *sizeof(int));
  if ((cc = getStateCorrectionCounts(CORRECTIONSFILE,start, stop, sttcnts, 256)) != 0) return cc;
  for (i=0; i<length && i<gNumAlarmGroups; i++) for (k=0; k<gAlarmGroups[i].nindices; k++)
  {
    corcnts[i] += sttcnts[k]; 
  }
  return 0;
}
int getAvailabilityInRange(char *system, time_t start, time_t stop, time_t *nexttime)
{
  DBLDBL problemCounts[100];
  int i, npts, cntB = -1, cntE = -1, hbt = 0;
  HstRngType rng;
  HstReqType req;

  req.eqm = ALAEQM_TAG;
  req.prp = "NOTREADYCOUNT";
  req.dev = system;
  req.index = req.fieldindex = req.sample = 0;
  if (stop < start + 2) hbt = 10; // requested range otherwise likely to see too few points
  rng.starttime = (double)(start-hbt);
  rng.stoptime = (double)(stop+hbt);
  rng.lwrLimit = rng.uprLimit = 0;
  rng.startstamp = rng.stopstamp = rng.userstamp = 0;
  npts = getLTSData(&req,rng,(BYTE *)problemCounts,CF_DBLDBL,NULL,100);
  if (npts <= 0) return 1; // no points found or error -> return available
  for (i=0; i<npts; i++)
  {
    if (cntB < 0 || problemCounts[i].d2val <= (double)start) cntB = (int)problemCounts[i].d1val;
    if (problemCounts[i].d2val > (double)start) cntE = (int)problemCounts[i].d1val;
    if (problemCounts[i].d2val >= (double)stop) break;
  }
  if (cntE < 0 && i < npts) cntE = (int)problemCounts[i].d1val;
  return (cntE == cntB) ? 1 : 0;
}
int isCorrectable(char *sys)
{
  int i;
  for (i=0; i<gNumAlarmSystems; i++)
  {
    if (strnicmp(sys,gAlarmSystems[i].name,64)) continue; // not a match
    if ((int)gAlarmSystems[i].d1val == 5000) return FALSE;
    if ((int)gAlarmSystems[i].d1val == 5001) return FALSE;
    return TRUE;
  }
  return TRUE;
}
#define corrhdr "Start, Stop, System, To, User, Comment"
#define corrstt(v,c) ((v) ? (c) ? "Problems" : "Blamed" : "Available") 
int writeStateCorrectionToFile(FILE *fp,StateCorrection *newCorrection,time_t start, time_t stop, int isCorrection)
{
  char startstr[64], stopstr[64];
  char *txt = isCorrection ? "correct" : "assign";
  strncpy(startstr,GetDataTimeString((double)start,FALSE),64);
  strncpy(stopstr,GetDataTimeString((double)stop,FALSE),64);
  if (newCorrection->to >= 0)
  { // if 'to' < 0 then I'm removing a prior correction (back to the way it was)
    char *cs = corrstt(newCorrection->to,isCorrection);
    fprintf(fp,"%.32s, %s, %.64s, %.64s, %.16s, %.256s\n",startstr,stopstr,
      newCorrection->system,cs,newCorrection->usr,newCorrection->cmt);
    srvlog(CORRLOGFILE,"writeStateCorrectionToFile","%s %.64s to %s from %.32s to %.32s by %.16s",
      txt,newCorrection->system,cs,startstr,stopstr,newCorrection->usr);
    return 1;
  }
  txt = isCorrection ? "correction" : "assigned blame";
  srvlog(CORRLOGFILE,"writeStateCorrectionToFile","remove %s to %.64s from %.32s to %.32s by %.16s",
    txt,newCorrection->system,startstr,stopstr,newCorrection->usr);
  return 0;
}
int addStateCorrections(char *filename, StateCorrection *newCorrection) 
{
  int i, n, cc=0, nitems, corrWritten=FALSE,isAvailable=FALSE,isSameSystem=FALSE;
  char fn[128],nfn[128],bfn[132],ofn[128];
  FILE *fp = NULL;
  StateCorrection *stcr = stateCorrection;
  time_t start = (time_t)newCorrection->start, stop = (time_t)newCorrection->stop, t;
  struct tm *lt;
  int dd, mm, yy;
  int dde, mme, yye, yde;
  time_t nexttime = 0;
  int isCorrection = (!strcmp(filename,CORRECTIONSFILE));

  if (newCorrection == NULL)
  {
    srvlog(CORRLOGFILE,"addStateCorrections","NULL correction!");
    return argument_list_error;
  }
  isAvailable = getAvailabilityInRange(newCorrection->system,start,stop,&nexttime);
  if (isCorrection && newCorrection->to && !isCorrectable(newCorrection->system))
  {
    srvlog(CORRLOGFILE,"addStateCorrections","mark systematic subsystem %.64s as unavailable by %.16s is not allowed",
      newCorrection->system,newCorrection->usr);
    return not_allowed;
  }
  if (!isCorrection)
  { // assign blame
    if (!newCorrection->to) newCorrection->to  = -1; // correct the assignment (unassign blame)
  }
  else if ((newCorrection->to && !isAvailable) || (!newCorrection->to && isAvailable))
  { // back to stored value: signal remove correction
    newCorrection->to  = -1;
  }
  if (nexttime > 0 && stop >= nexttime) stop = nexttime - 1;
  lt = localtime(&stop);
  dde = lt->tm_mday; mme = lt->tm_mon; yye = lt->tm_year; yde = lt->tm_yday;   
  for (t = start; t < stop+86400 && start < stop; t += 86400)
  { // Loop over the days involved
    lt = localtime(&t);
    if (lt->tm_yday > yde) break; /* past the end */
    dd = lt->tm_mday; mm = lt->tm_mon; yy = lt->tm_year;
    sprintf(nfn,"%s%04d/%02d/%02d/corrections.nsv",DBPATH,1900+yy,mm+1,dd); // dbpath ?
    if ((fp=createAndOpen(nfn,"w")) == NULL) ccerr(file_error);
    fprintf(fp,"%s\n",corrhdr);
    if (lt->tm_yday < yde)
    { // spans more than 1 day !
      struct tm it;
      it = *lt;
      it.tm_hour = 23; it.tm_min = 59; it.tm_sec = 59;
      stop = mktime(&it);
    }
    nitems = 0; n = 0;
    sprintf(fn,"%s%04d/%02d/%02d/",DBPATH,1900+yy,mm+1,dd); 
    csvStateCorrectionsFileDb.siz = MAX_CORRECTION_PER_DAY;
    if (csvReadFile(fn,filename,&csvStateCorrectionsFileDb,(void **)&stcr) == 0) 
      nitems = csvStateCorrectionsFileDb.siz;
    for (i=0; i<nitems; i++)
    { // fill in some missing parameters from the items read from an existing file ...
      stcr[i].stop = (double)getTimeStampFromLogEntry(stcr[i].stopStr);
      stcr[i].start = (double)getTimeStampFromLogEntry(stcr[i].startStr);
      stcr[i].to = getStateAssignment(stcr[i].state);
      if (start == (time_t)stcr[i].start && 
          stop == (time_t)stcr[i].stop && 
          newCorrection->to == stcr[i].to && 
          !strnicmp(stcr[i].system,newCorrection->system,64))
      { // correction is already in the DB !
        char *txt = isCorrection ? "correction" : "assignment blame";
        srvlog(CORRLOGFILE,"addStateCorrections","%s to %.64s from %.32s to %.32s by %.16s already exists",
          txt,newCorrection->system,stcr[i].startStr,stcr[i].stopStr,newCorrection->usr);
        ccerr(0);
      }
    }
    corrWritten = FALSE;
    for (i=0; i<nitems; i++)
    { // scan items read from an existing file ...
      if (stcr[i].stop < (double)start) 
      { // pre-dates current correction: write back to file
        fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
          stcr[i].startStr,stcr[i].stopStr,stcr[i].system,stcr[i].state,stcr[i].usr,stcr[i].cmt);
        n++;
        continue;
      }
      isSameSystem = strnicmp(stcr[i].system,newCorrection->system,64) == 0 ? TRUE : FALSE;
      if (start < (time_t)stcr[i].start) 
      { // the new region starts before the old region (possibly overlapping?)
        if (!isSameSystem) 
        { // different systems: write back to file
          fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
            stcr[i].startStr,stcr[i].stopStr,stcr[i].system,stcr[i].state,stcr[i].usr,stcr[i].cmt);
          n++;
          continue;
        }
        else 
        {
          if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop,isCorrection)) n++;
          corrWritten = TRUE;
          if (stop >= (time_t)stcr[i].start && stop < (time_t)stcr[i].stop)
          { // overlapping: something left from previous correction
            fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
              GetDataTimeString((double)(stop+1),FALSE),stcr[i].stopStr,stcr[i].system,stcr[i].state,stcr[i].usr,stcr[i].cmt);
            n++;
            continue;
          }
        }
      }
      if (start >= (time_t)stcr[i].start && start <= (time_t)stcr[i].stop) 
      { // overlapping regions(the new region starts after the old region)
        if (!isSameSystem) 
        { // different systems: write back to file
          fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
            stcr[i].startStr,stcr[i].stopStr,stcr[i].system,stcr[i].state,stcr[i].usr,stcr[i].cmt);
          n++;
          continue;
        }
        else if (start > (time_t)stcr[i].start)
        { // same system, correction starts later => write back the initial piece to file
          fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
            stcr[i].startStr,GetDataTimeString((double)(start-1),FALSE),stcr[i].system,stcr[i].state,stcr[i].usr,stcr[i].cmt);
          n++;
          if (corrWritten) continue;
        }
        if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop,isCorrection)) n++;
        corrWritten = TRUE;
        if (isSameSystem && stop < (time_t)stcr[i].stop)
        { // same system, correction ends earlier => write back final piece to file
          fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
            GetDataTimeString((double)(stop+1),FALSE),stcr[i].stopStr,stcr[i].system,stcr[i].state,stcr[i].usr,stcr[i].cmt);
          n++;
          continue;
        }
      }
      if (stcr[i].start > (double)stop) 
      { // post-dates current correction: finish adding new correction and write back to file
        if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop,isCorrection)) n++;
        corrWritten = TRUE;
        fprintf(fp,"%.32s, %.32s, %.64s, %.64s, %.16s, %.256s\n", 
          stcr[i].startStr,stcr[i].stopStr,stcr[i].system,stcr[i].state,stcr[i].usr,stcr[i].cmt);    
        n++;
        continue;
      }
    }  
    if (!corrWritten) if (writeStateCorrectionToFile(fp,newCorrection,start,stop,isCorrection)) n++;
    fclose(fp); fp = NULL;
    sprintf(ofn,"%.110s%s",fn,filename);
    if (n >= MAX_CORRECTION_PER_DAY)
    {
      char *txt = isCorrection ? "corrections" : "blame assignments";
      srvlog(CORRLOGFILE,"addStateCorrections","total number of %s for %d.%d.%d would exceed capacity %d",
        txt,dd,mm,yy,MAX_CORRECTION_PER_DAY);
      return at_limit;
    }
    else
    { // okay: update the corrections file ...
      if (nitems > 0)
      {
        strncpy(bfn,ofn,128);
        strcat(bfn,".bak");
        rename(fn,bfn);
      }
      rename(nfn,ofn);
    }
    start = stop + 1;
    stop = (time_t)newCorrection->stop;
  }
err:
  if(fp != NULL) fclose(fp);
  return 0;  
}

time_t timeDeclaredRunning = 0;
time_t timeDeclaredProblems = 0;
int windowDeclaredRunning = 60;
int windowDeclaredProblems = 60;
int countsDeclaredRunning = 0;
int countsDeclaredProblems = 0;
int syscntsDeclaredProblems[MAX_ALARM_SYSTEMS];

int readDeclaredCounters(void)
{
  FILE *fp = NULL;
  char str[256],scratch[64];
  int rdcl_col=-1,rcnt_col=-1,rwnd_col=-1;
  int pdcl_col=-1,pcnt_col=-1,pwnd_col=-1;
  int cc = 0, done = 0;

  if ((fp=fopen(DECL_CNTRS_FILE,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((rdcl_col=findcol(str,"RUNNING_DECLARED")) < 0) ccerr(no_such_line);
      if ((rcnt_col=findcol(str,"RUNNING_COUNTS")) < 0) ccerr(no_such_line);
      if ((rwnd_col=findcol(str,"WINDOW_RUNNING")) < 0) ccerr(no_such_line);
      if ((pdcl_col=findcol(str,"PROBLEMS_DECLARED")) < 0) ccerr(no_such_line);
      if ((pcnt_col=findcol(str,"PROBLEMS_COUNTS")) < 0) ccerr(no_such_line);
      if ((pwnd_col=findcol(str,"WINDOW_PROBLEMS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(rdcl_col,str),64);
    timeDeclaredRunning = (time_t)strtol(scratch,NULL,10);
    strncpy(scratch,colptr(rcnt_col,str),64);
    countsDeclaredRunning = atoi(scratch);
    strncpy(scratch,colptr(rwnd_col,str),64);
    windowDeclaredRunning = atoi(scratch);
    strncpy(scratch,colptr(pdcl_col,str),64);
    timeDeclaredProblems = (time_t)strtol(scratch,NULL,10);
    strncpy(scratch,colptr(pcnt_col,str),64);
    countsDeclaredProblems = atoi(scratch);
    strncpy(scratch,colptr(pwnd_col,str),64);
    windowDeclaredProblems = atoi(scratch);
    break;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
#define DECL_CNTRS_HDR "RunningDeclared, RunningCounts, WindowRunning, ProblemsDeclared, ProblemsCounts, WindowProblems\n"
int writeDeclaredCounters(void)
{
  FILE *fp = NULL;
  int cc = 0;

  if ((fp=fopen(DECL_CNTRS_FILE,"w")) == NULL) ccerr(file_error);
  fprintf(fp,DECL_CNTRS_HDR);

  fprintf(fp,"%ld, %d, %d, %ld, %d, %d\n",
    (long)timeDeclaredRunning, countsDeclaredRunning, windowDeclaredRunning,
    (long)timeDeclaredProblems, countsDeclaredProblems, windowDeclaredProblems);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

void checkCounters(void)
{
  static time_t last_t = 0;
  static int accumulatedCountsRunning = 0;
  static int needToIncrementCountsProblems = TRUE;
  static int  lIsOfficiallyRunning = FALSE;
  static int  lIsOfficiallyProblems = FALSE;
  time_t t = time(NULL);
  int delta_t = 0;
  if (last_t == 0)
  { // first pass
    last_t = t;
  }
  if (last_t == t) return;
  delta_t = (int)(t - last_t);
  if (gIsOfficiallyRunning)
  {
    if (!lIsOfficiallyRunning)
    {
      srvlog(MTBFLOGFILE,"checkCounters","mark as running!");
    }
    lIsOfficiallyRunning = gIsOfficiallyRunning;
    if (t < timeDeclaredRunning + windowDeclaredRunning)
    {
      if (accumulatedCountsRunning == 0)
      {
        srvlog(MTBFLOGFILE,"checkCounters","accumulate running counts for %d seconds!",windowDeclaredRunning);
      }
      accumulatedCountsRunning += delta_t;
    }
    else
    { // passed the window, we're good!
      if (accumulatedCountsRunning > 0)
      {
        srvlog(MTBFLOGFILE,"checkCounters","officially begin incrementing declared running counts at %d!",countsDeclaredRunning);
      }
      countsDeclaredRunning += delta_t;
      countsDeclaredRunning += accumulatedCountsRunning;
      accumulatedCountsRunning = 0;
    }
  }
  else
  {
    if (lIsOfficiallyRunning)
    {
      srvlog(MTBFLOGFILE,"checkCounters","no longer running (counts stopped at %d!",countsDeclaredRunning);
    }
    lIsOfficiallyRunning = FALSE;
    accumulatedCountsRunning = 0;
  }
  if (gIsOfficiallyProblems)
  {
    if (!lIsOfficiallyProblems)
    {
      srvlog(MTBFLOGFILE,"checkCounters","mark as problems!");
    }
    lIsOfficiallyProblems = gIsOfficiallyProblems;
    if (needToIncrementCountsProblems && t >= timeDeclaredProblems + windowDeclaredProblems)
    {
      countsDeclaredProblems++;
      needToIncrementCountsProblems = FALSE;
      srvlog(MTBFLOGFILE,"checkCounters","increment problems to %d",countsDeclaredProblems);
    }
  }
  else
  {
    if (lIsOfficiallyProblems)
    {
      srvlog(MTBFLOGFILE,"checkCounters","no longer in problems state -> set needToIncrement flag");
    }
    needToIncrementCountsProblems = TRUE;
    lIsOfficiallyProblems = FALSE;
  }
  
  last_t = t;

}
void checkstate(int id, int cc)
{
  static int last_cc = 0;
  static int sIsRunning = 0;
  static int sIsOfficiallyRunning = 0;
  static int sIsOfficiallyProblems = 0;
  static char sDeclaredState[64] = {0};

  if (cc != 0) ccerr(cc);

  gIsRunning = isRunning();
  gIsOfficiallyRunning = isOfficiallyRunning();
  gIsOfficiallyProblems = isOfficallyProblems();

  if (strncmp(gDeclaredState,sDeclaredState,64))
  {
    feclog("declared state changed from %.64s to %.64s",sDeclaredState,gDeclaredState);
    srvlog(MTBFLOGFILE,"checkstate","declared state changed from %.64s to %.64s",sDeclaredState,gDeclaredState);
    strncpy(sDeclaredState,gDeclaredState,32);
  }
  if (sIsRunning != gIsRunning)
  {
    feclog("running state changed to %s",gIsRunning ? "running" : "non running");
    srvlog(MTBFLOGFILE,"checkstate","running state changed to %s",gIsRunning ? "running" : "non running");
    sIsRunning = gIsRunning;
  }
  if (gIsOfficiallyRunning && sIsOfficiallyRunning != gIsOfficiallyRunning)
  {
    timeDeclaredRunning = time(NULL);
    sIsOfficiallyRunning = gIsOfficiallyRunning;
    srvlog(MTBFLOGFILE,"checkstate","state marked as officially running");
  }
  if (gIsOfficiallyProblems && sIsOfficiallyProblems != gIsOfficiallyProblems)
  {
    timeDeclaredProblems = time(NULL);
    sIsOfficiallyProblems = gIsOfficiallyProblems;
    srvlog(MTBFLOGFILE,"checkstate","state marked as officially problems");
  }
err:
  srvlog(LNKLOGFILE,"update","link to %.128s : link status now = %.32s",gDeclStateSrv,cc2str(cc));
  last_cc = cc;
  return;
}
int isRunning(void)
{
  int i;
  char lDeclaredState[32];
  strncpy(lDeclaredState,gDeclaredState,32);
  strlwr(lDeclaredState);
  for (i=0; i<nNonRunningStates; i++)
    if (strstr(lDeclaredState,nonRunningState[i]) != NULL) return FALSE;
  return TRUE;
}
int isOfficiallyRunning(void)
{
  int i;
  char lDeclaredState[32];
  strncpy(lDeclaredState,gDeclaredState,32);
  strlwr(lDeclaredState);
  for (i=0; i<nRunningStates; i++)
  {
    if (isRunningState[i].compareType == STRING_CONTAINS)
    {
      if (stristr(lDeclaredState,isRunningState[i].state) != NULL) return TRUE;
    }
    else
    {
      if (!strnicmp(gDeclaredState,isRunningState[i].state,32)) return TRUE;
    }
  }
  return FALSE;
}
int isOfficallyProblems(void)
{
# ifdef SIMPLY_PROBLEMS
# define PROBLEMS "problems"
  char lDeclaredState[32];
  strncpy(lDeclaredState,gDeclaredState,32);
  strlwr(lDeclaredState);
  return (strstr(lDeclaredState,PROBLEMS) != NULL) ? TRUE : FALSE;
# else  
  int i;
  char lDeclaredState[32];
  strncpy(lDeclaredState,gDeclaredState,32);
  strlwr(lDeclaredState);
  for (i=0; i<nProblemStates; i++)
  {
    if (isProblemState[i].compareType == STRING_CONTAINS)
    {
      if (stristr(lDeclaredState,isProblemState[i].state) != NULL) return TRUE;
    }
    else
    {
      if (!strnicmp(gDeclaredState,isProblemState[i].state,32)) return TRUE;
    }
  }
  return FALSE;
# endif
}
int getAlarmIndexFromSystemNumber(char *sys)
{
  int i;
  char *asys;
  for (i=0; i<gNumAlarmSystems; i++)
  {
    asys = gAlarmSystems[i].name[0] == '.' ? &gAlarmSystems[i].name[1] : gAlarmSystems[i].name;
    if (!strncmp(asys,sys,64)) return i;
  }
  return -1;
}
int getAlarmGroupIndexFromGroup(char *grp)
{
  int i;
  for (i=0; i<gNumAlarmGroups; i++)
    if (!strncmp(gAlarmGroups[i].grp,grp,64)) return i;
  return -1;
}
int getOfficiallyRunningStates(void)
{
  FILE *fp = NULL;
  char str[256], scratch[64];
  int sta_col=-1,flg_col,cc = 0;
  int i, done = 0;

  if ((fp=fopen(STATES_RUNNING_FILE,"r")) != NULL)
  {
    for (i=0; csvGetLine(str,255,fp) != NULL && i<MAX_RUNNING_STATES; )
    {
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        if ((sta_col=findcol(str,"STATE")) < 0) ccerr(no_such_line);
        flg_col = findcol(str,"MATCH");
        done = 1;
        continue;
      }
      strncpy(isRunningState[i].state,colptr(sta_col,str),32);
      feclog("flagging %s as an official running state",isRunningState[i].state);
      if (flg_col >= 0)
      {
        strncpy(scratch,colptr(flg_col,str),64);
        if (!stricmp(scratch,"contains"))
          isRunningState[i].compareType = STRING_CONTAINS;
        if (!stricmp(scratch,"exact"))
          isRunningState[i].compareType = STRING_MATCHES;
      }
      i++;
    }
    nRunningStates = i;
  }
  else
  {
    feclog("make running states file from default set");
    strncpy(isRunningState[0].state,"operating",32);
    nRunningStates = 1;
    if ((fp=fopen(STATES_RUNNING_FILE,"w")) == NULL) ccerr(file_error);
    fprintf(fp,"STATE, MATCH\n");
    for (i=0; i<nRunningStates; i++)
    {
      fprintf(fp,"%s, contains\n",isRunningState[i].state);
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int getOfficiallyProblemStates(void)
{
  FILE *fp = NULL;
  char str[256], scratch[64];
  int sta_col=-1,flg_col,cc = 0;
  int i, done = 0;

  if ((fp=fopen(STATES_PROBLEMS_FILE,"r")) != NULL)
  {
    for (i=0; csvGetLine(str,255,fp) != NULL && i<MAX_RUNNING_STATES; )
    {
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        if ((sta_col=findcol(str,"STATE")) < 0) ccerr(no_such_line);
        flg_col = findcol(str,"MATCH");
        done = 1;
        continue;
      }
      strncpy(isProblemState[i].state,colptr(sta_col,str),32);
      feclog("flagging %s as an official problems state",isProblemState[i].state);
      if (flg_col >= 0)
      {
        strncpy(scratch,colptr(flg_col,str),64);
        if (!stricmp(scratch,"contains"))
          isProblemState[i].compareType = STRING_CONTAINS;
        if (!stricmp(scratch,"exact"))
          isProblemState[i].compareType = STRING_MATCHES;
      }
      i++;
    }
    nProblemStates = i;
  }
  else
  {
    feclog("make problmes states file from default set");
    strncpy(isProblemState[0].state,"problems",32);
    nProblemStates = 1;
    if ((fp=fopen(STATES_PROBLEMS_FILE,"w")) == NULL) ccerr(file_error);
    fprintf(fp,"STATE, MATCH\n");
    for (i=0; i<nProblemStates; i++)
    {
      fprintf(fp,"%s, contains\n",isProblemState[i].state);
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int getNonRunningStates(void)
{
  FILE *fp = NULL;
  char str[256];
  int col=-1,cc = 0;
  int i, done = 0;

  if ((fp=fopen(STATES_FILE,"r")) != NULL)
  {
    for (i=0; csvGetLine(str,255,fp) != NULL && i<MAX_NONRUNNING_STATES; )
    {
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        if ((col=findcol(str,"STATE")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      strncpy(nonRunningState[i],colptr(col,str),32);
      feclog("flagging %s as a non-running state",nonRunningState[i]);
      i++;
    }
    nNonRunningStates = i;
  }
  else
  {
    feclog("make non-running states file from default set");
    strncpy(nonRunningState[0],"access",32);
    strncpy(nonRunningState[1],"off",32);
    strncpy(nonRunningState[2],"service",32);
    strncpy(nonRunningState[3],"studies",32);
    strncpy(nonRunningState[3],"unknown",32);
    nNonRunningStates = 5;
    if ((fp=fopen(STATES_FILE,"w")) == NULL) ccerr(file_error);
    fprintf(fp,"STATE\n");
    for (i=0; i<nNonRunningStates; i++)
    {
      fprintf(fp,"%s\n",nonRunningState[i]);
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int updateALLGroup(void)
{
  FILE *fp = NULL;
  char fn[48];
  int i, cc = 0;
  if (gNumAlarmGroups == 0) return 0;
  sprintf(fn,"ALL.csv");
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"SYSTEM,COUNTS,RUNCOUNTS\n");
  for (i=0; i<gNumAlarmSystems; i++)
  {
    if ((int)gAlarmSystems[i].d1val == 5000) continue; // systematic 'System' category
    if ((int)gAlarmSystems[i].d1val == 5001) continue; // systematic 'Hardware' category
    fprintf(fp,"%.64s,%d,%d\n",gAlarmSystems[i].name,gAlarmCounter[i],gAlarmCounterRunning[i]);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int makeALLGroup(void)
{
  FILE *fp = NULL;
  char fn[48];
  int i, cc = 0;
  if (gNumAlarmGroups == 0) return 0;
  sprintf(fn,"ALL.csv");
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"SYSTEM,COUNTS,RUNCOUNTS\n");
  for (i=0; i<gNumAlarmSystems; i++)
  {
    if ((int)(gAlarmSystems[i].d1val) == 5000) continue; // systematic 'System' category
    if ((int)(gAlarmSystems[i].d1val) == 5001) continue; // systematic 'Hardware' category
    fprintf(fp,"%.64s,0,0\n",gAlarmSystems[i].name);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int fillInGroupTable(int i)
{
  FILE *fp = NULL;
  char str[256],scratch[64],fn[64];
  int sys_col=-1,cc = 0;
  int k, n, done = 0;

  sprintf(fn,"%s.csv",gAlarmGroups[i].grp);
  if ((n=GetDataBaseSize(fn)) < 1) ccerr(no_such_file);
  if ((gAlarmGroups[i].indices=(int *)calloc(n,sizeof(int))) == NULL) 
    ccerr(out_of_server_memory);
  gAlarmGroups[i].nindices = n;
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  for (k=0; csvGetLine(str,255,fp) != NULL && k<gAlarmGroups[i].nindices; )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((sys_col=findcol(str,"SYSTEM")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(sys_col,str),48);
    n = getAlarmIndexFromSystemNumber(scratch);
    if (n < 0)
    {
      feclog("alarm system %s not found in table !",scratch);
      continue;
    }
    if (!strnicmp(scratch,gAlarmGroups[i].grp,48))
    {
      feclog("duplicate name: group %.48s and system %.48s",
        gAlarmGroups[i].grp,scratch);
    }
    gAlarmGroups[i].indices[k] = n;
    k++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int getAlarmGroups(void)
{
  FILE *fp = NULL;
  char str[256],scratch[48];
  int grp_col=-1,dsc_col=-1;
  int sys_col=-1,cnt_col=-1,run_col=-1,cc = 0;
  int i, done = 0;

  if ((fp=fopen(ALARM_GROUPS_FILE,"r")) == NULL) ccerr(no_such_file);
  for (i=0; csvGetLine(str,255,fp) != NULL && i<MAX_ALARM_SYSTEMS; )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((grp_col=findcol(str,"GROUP")) < 0) ccerr(no_such_line);
      if ((dsc_col=findcol(str,"DESCRIPTION")) < 0) ccerr(no_such_line);
      if ((cnt_col=findcol(str,"COUNTS")) < 0) ccerr(no_such_line);
      if ((run_col=findcol(str,"RUNCOUNTS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(gAlarmGroups[i].grp,colptr(grp_col,str),48);
    strncpy(gAlarmGroups[i].dsc,colptr(dsc_col,str),48);
    strncpy(scratch,colptr(cnt_col,str),48);
    gAlarmCounter[gNumAlarmSystems+i] = atoi(scratch);
    strncpy(scratch,colptr(run_col,str),48);
    gAlarmCounterRunning[gNumAlarmSystems+i] = atoi(scratch);
    fillInGroupTable(i);
    i++;
  }
  gNumAlarmGroups = i;
  if (i == 0) cc = end_of_file;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int readAlarmCountersFromFile(void)
{
  FILE *fp = NULL;
  char str[256],scratch[48];
  int sys_col=-1,cnt_col=-1,run_col=-1,prb_col=-1,cc = 0;
  int i, n, done = 0;

  if ((fp=fopen(ALARM_COUNTERS_FILE,"r")) == NULL) ccerr(no_such_file);
  for (i=0; csvGetLine(str,255,fp) != NULL && i<gNumAlarmSystems; )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((sys_col=findcol(str,"SYSTEM")) < 0) ccerr(no_such_line);
      if ((cnt_col=findcol(str,"COUNTS")) < 0) ccerr(no_such_line);
      if ((run_col=findcol(str,"RUNCOUNTS")) < 0) ccerr(no_such_line);
      prb_col = findcol(str,"PROBLEMS");
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(sys_col,str),48);
    n = getAlarmIndexFromSystemNumber(scratch);
    if (n < 0)
    {
      feclog("alarm system %s not found in table !",scratch);
      continue;
    }
    strncpy(scratch,colptr(cnt_col,str),48);
    gAlarmCounter[n] = atoi(scratch);
    strncpy(scratch,colptr(run_col,str),48);
    gAlarmCounterRunning[n] = atoi(scratch);
    if (prb_col >= 0)
    {
      strncpy(scratch,colptr(prb_col,str),48);
      syscntsDeclaredProblems[n] = atoi(scratch);
    }
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int writeAlarmGroupsToFile(void)
{
  FILE *fp = NULL;
  int i, cc = 0;
  if (gNumAlarmGroups == 0) return 0;
  if ((fp=fopen(ALARM_GROUPS_FILE,"w")) == NULL) ccerr(no_such_file);
  fprintf(fp,"GROUP,DESCRIPTION,COUNTS,RUNCOUNTS\n");
  for (i=0; i<gNumAlarmGroups; i++)
  {
    fprintf(fp,"%s,%s,%d,%d\n",gAlarmGroups[i].grp,gAlarmGroups[i].dsc,
      gAlarmCounter[gNumAlarmSystems+i],gAlarmCounterRunning[gNumAlarmSystems+i]);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int writeMTBPEpochFile()
{
  char tstr[64];
  FILE *fp;
  if ((fp=fopen(EPOCH_FILE,"w")) == NULL)
  {
    perror("fopen");
    feclog("cannot open MTBP epoch file for writing!");
    return file_error;
  }
  strncpy(tstr,GetDataTimeString((double)gMTBPEpochBegin,FALSE),64);
  fprintf(fp,"EPOCH_BEGIN\n");
  fprintf(fp,"%.64s\n",tstr);
  fclose(fp);
  return 0;
}
int resetMTBPEpoch(void)
{
  int i;
  gMTBPEpochBegin = time(NULL);
  for (i=0; i<gNumAlarmSystems+gNumAlarmGroups && i<MAX_ALARM_SYSTEMS; i++)
  {
    mtbpInfoSet[i].mtbp = 0;
    mtbpInfoSet[i].nproblems = 0;
  }
  return writeMTBPEpochFile();
}

/**
 * \internal if called with span > 0 then begin = 0 => end 
 *  of availability/beginning of problems.
 *  if called with span = 0, begin = time then mark the 
 *  beginning of a run
 */
int updateMTBPInfo(int span,UINT32 begin,int index)
{
  if (index < 0 || index >= MAX_ALARM_SYSTEMS) return invalid_index;
  if (span > 0)
  {
    int mtbp = mtbpInfoSet[index].mtbp;
    mtbp = (mtbp * mtbpInfoSet[index].nproblems + span) / (mtbpInfoSet[index].nproblems + 1);
    mtbpInfoSet[index].mtbp = mtbp;
    mtbpInfoSet[index].nproblems++;
    begin = 0;
  }
  mtbpInfoSet[index].runBegin = begin;
  return 0;
}
int writeMTBPFile(int writeGrpFile)
{
# define MTBF_HDR "SYSTEM, NPROBLEMS, MTBP, RUN_BEGIN"
  FILE *fp = NULL;
  int i, k, cc = 0;
  char tstr[64], *nam;
  char *fn = writeGrpFile ? MTBP_GROUP_FILE : MTBP_FILE;
  int n = writeGrpFile ? gNumAlarmGroups : gNumAlarmSystems;
  
  if ((fp=fopen(fn,"w")) == NULL) ccerr(no_such_file);
  fprintf(fp,"%s\n",MTBF_HDR);
  for (i=0; i<n; i++)
  {
    if (writeGrpFile)
    {
      k = gNumAlarmSystems + i;
      nam = gAlarmGroups[i].grp;
    }
    else
    {
      k = i;
      nam = gAlarmSystems[i].name;
    }
    if (k >= MAX_ALARM_SYSTEMS) continue;
    if (mtbpInfoSet[k].runBegin > 0) 
      strncpy(tstr,GetDataTimeString((double)mtbpInfoSet[k].runBegin,FALSE),64);
    else
      strcpy(tstr,"PROBLEMS");
    
    fprintf(fp,"%.64s, %d, %d, %.64s\n",nam,mtbpInfoSet[k].nproblems,mtbpInfoSet[k].mtbp,tstr);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int readMTBPFile(int readGrpFile)
{
  FILE *fp = NULL;
  char str[256],scratch[64];
  int sys_col=-1,prob_col=-1,mtbp_col=-1,run_col=-1,cc = 0;
  int i, n = 0, done = 0;
  char *fn = readGrpFile ? MTBP_GROUP_FILE : MTBP_FILE;

  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  for (i=0; csvGetLine(str,255,fp) != NULL && i<gNumAlarmSystems; i++)
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((sys_col=findcol(str,"SYSTEM")) < 0) ccerr(no_such_line);
      if ((prob_col=findcol(str,"NPROBLEMS")) < 0) ccerr(no_such_line);
      if ((mtbp_col=findcol(str,"MTBP")) < 0) ccerr(no_such_line);
      if ((run_col=findcol(str,"RUN_BEGIN")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(sys_col,str),64);
    if (readGrpFile)
    {
      n = getAlarmGroupIndexFromGroup(scratch);
      if (n < 0) ccerr(invalid_index);
      n += gNumAlarmSystems;
    }
    else
    {
      n = getAlarmIndexFromSystemNumber(scratch);
      if (n < 0) ccerr(invalid_index);
    }
    
    strncpy(scratch,colptr(prob_col,str),48);
    mtbpInfoSet[n].nproblems = atoi(scratch);
    
    strncpy(scratch,colptr(mtbp_col,str),48);
    mtbpInfoSet[n].mtbp = atoi(scratch);
    strncpy(scratch,colptr(run_col,str),48);
    if (!strcmp(scratch,"PROBLEMS")) mtbpInfoSet[n].runBegin = 0;
    else mtbpInfoSet[n].runBegin = (UINT32)getTimeStampFromLogEntry(scratch);
    gAlarmSystemTimeSpan[n].begin = mtbpInfoSet[n].runBegin;
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0 && cc != no_such_file)
  {
    feclog("file %s : line %d : %.32s",fn,i,cc2str(cc));
  }
  return cc;
}

int writeAlarmCountersToFile(void)
{
  FILE *fp = NULL;
  int i, cc = 0;
  if ((fp=fopen(ALARM_COUNTERS_FILE,"w")) == NULL) ccerr(no_such_file);
  fprintf(fp,"SYSTEM,COUNTS,RUNCOUNTS,PROBLEMS\n");
  for (i=0; i<gNumAlarmSystems; i++)
  {
    fprintf(fp,"%.64s,%d,%d,%d\n",gAlarmSystems[i].name,gAlarmCounter[i],gAlarmCounterRunning[i],syscntsDeclaredProblems[i]);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
void checkalarms(int id,int cc)
{
  int i = id-1;
  static int last_cc[MAX_ALARM_SYSTEMS] = { 0 };

  if (cc == 0)
  {
    int hasErrorsA, hasErrorsB = (gNumberOfAlarms[i] == 0) ? 0 : 1;
    gNumberOfAlarms[i] = gNumberOfAlarmsX[i].ival;
    hasErrorsA = (gNumberOfAlarms[i] == 0) ? 0 : 1;
    if (hasErrorsA != hasErrorsB) SystemScheduleProperty(ALAEQM_TAG, "ISREADY");
  }
  if (cc != last_cc[i]) srvlog(LNKLOGFILE,"update","link to %.64s : link status now = %s",gAlarmSystems[i].name,erlst[cc]);
  last_cc[i] = cc;
}
int gIncludeHiddenSystems = TRUE;
int getAlarmSystems(char *cassrv,NAME64DBLDBL *almsys)
{
  int retries, cc=link_not_open;
  DTYPE dout;
  char devnam[64], *prpnam;
  DTYPEZERO(dout);
  prpnam = gIncludeHiddenSystems ? "ALARMSYSTEMS.ALL" : "ALARMSYSTEMS";
  sprintf(devnam,"%s/#0",cassrv);
  dout.dFormat = CF_NAME64DBLDBL;
  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  dout.dTag[0] = 0;
  dout.data.vptr = almsys;
  for (retries=0; retries<3 && cc != 0; retries++)
    cc = ExecLinkEx(devnam,prpnam,&dout,NULL,CA_READ|CA_MUTABLE,800);
  if (cc) return -cc;
  return dout.dArrayLength;
}
void alaeqm_bkg(void)
{
  int i,k,idx,deadband,span,sched=FALSE,updMTBP=FALSE,updAlmCnts=FALSE;
  static time_t lasttime = 0;
  time_t t = time(NULL);

  deadband = (int)(lasttime > 0 ? (t - lasttime) : 1);
  lasttime = t;

  ClearAlarm(ALAEQM_TAG,-1);
  checkCounters();
  for (i=0; i<gNumAlarmSystems; i++)
  {
    if (gNumberOfAlarms[i] > 0) gAlarmCounter[i] += deadband;
    if (gIsOfficiallyProblems && gNumberOfAlarms[i] > 0) 
    {
      syscntsDeclaredProblems[i]++;
      updAlmCnts = TRUE;
    }
    if (gIsRunning && gNumberOfAlarms[i] > 0) 
    { // fatal alarms showing ...
      if (gAlarmSystemTimeSpan[i].type == SPAN_AVAILABLE)
      { // was available
        if (gAlarmSystemTimeSpan[i].last_prob == 0) gAlarmSystemTimeSpan[i].last_prob = (UINT32)t;
        if (t < (int)gAlarmSystemTimeSpan[i].last_prob + mtbp_threshold) continue;
        gAlarmSystemTimeSpan[i].last_prob = (UINT32)(t - mtbp_threshold);
        if (gAlarmSystemTimeSpan[i].end == 0)
        { // terminate the span :
          gAlarmSystemTimeSpan[i].end = gAlarmSystemTimeSpan[i].last_prob - 1;
          sched = TRUE;
          span = (int)(gAlarmSystemTimeSpan[i].end - gAlarmSystemTimeSpan[i].begin);
          updateMTBPInfo(span,0,i);
          updMTBP = TRUE;
        }
        else
        { // start the problems span (next pass) :
          gAlarmSystemTimeSpan[i].begin = gAlarmSystemTimeSpan[i].end + 1;
          gAlarmSystemTimeSpan[i].end = 0;
          gAlarmSystemTimeSpan[i].nalarms = gNumberOfAlarms[i];
          gAlarmSystemTimeSpan[i].type = SPAN_PROBLEMS;
          gAlarmSystemTimeSpan[i].last_prob = 0; // reset
        }
      }
      gAlarmCounterRunning[i] += deadband;
      updAlmCnts = TRUE;
    }
    else
    { // no more alarms or no longer running ...
      if (gAlarmSystemTimeSpan[i].type == SPAN_PROBLEMS)
      { // was showing problems ...
        if (gAlarmSystemTimeSpan[i].last_avail == 0) gAlarmSystemTimeSpan[i].last_avail = (UINT32)t;
        if (t < (int)gAlarmSystemTimeSpan[i].last_avail + mtbp_threshold) continue;
        gAlarmSystemTimeSpan[i].last_avail = (UINT32)(t - mtbp_threshold);
        if (gAlarmSystemTimeSpan[i].end == 0)
        { // had problems but is now available : terminate the span :
          gAlarmSystemTimeSpan[i].end = (UINT32)gAlarmSystemTimeSpan[i].last_avail - 1;
          sched = TRUE;
          updateMTBPInfo(0,(UINT32)t,i);
          updMTBP = TRUE;
        }
        else
        { // start the available span (next pass) :
          gAlarmSystemTimeSpan[i].begin = gIsRunning ? gAlarmSystemTimeSpan[i].end + 1 : 0;
          gAlarmSystemTimeSpan[i].end = 0;
          gAlarmSystemTimeSpan[i].nalarms = 0;
          gAlarmSystemTimeSpan[i].type = SPAN_AVAILABLE;
          gAlarmSystemTimeSpan[i].last_avail = 0; // reset
        }
      }
      if (gAlarmSystemTimeSpan[i].type == SPAN_AVAILABLE) 
      { // marked available
        if (gAlarmSystemTimeSpan[i].begin == 0 && gIsRunning)
        { // and we're running (again) and haven't noted the begin time
          gAlarmSystemTimeSpan[i].last_avail = (UINT32)t;
          gAlarmSystemTimeSpan[i].begin = (UINT32)t;
          updateMTBPInfo(0,(UINT32)t,i);
          updMTBP = TRUE;
        }
      }
    }
  }
  for (i=0; i<gNumAlarmGroups; i++)
  {
    for (k=0; k<gAlarmGroups[i].nindices; k++)
    {
      idx = gAlarmGroups[i].indices[k];
      if (gNumberOfAlarms[idx] > 0) 
      {
        gAlarmCounter[gNumAlarmSystems+i] += deadband;
        if (gIsRunning) gAlarmCounterRunning[gNumAlarmSystems+i] += deadband;
        writeAlarmGroupsToFile();
        break;
      }
    }
    gNumberOfAlarms[gNumAlarmSystems+i] = 0;
    for (k=0; k<gAlarmGroups[i].nindices; k++) 
    {
      idx = gAlarmGroups[i].indices[k];
      gNumberOfAlarms[gNumAlarmSystems+i] += gNumberOfAlarms[idx];
    }
    k = gNumAlarmSystems + i; // simplify notation ...
    if (gNumberOfAlarms[k] > 0)
    {
      if (gIsOfficiallyProblems && gNumberOfAlarms[k] > 0) syscntsDeclaredProblems[k]++;
      if (gIsRunning && gAlarmSystemTimeSpan[k].type == SPAN_AVAILABLE)
      {
        if (gAlarmSystemTimeSpan[k].last_prob == 0) gAlarmSystemTimeSpan[k].last_prob = (UINT32)t;
        if (t < (int)gAlarmSystemTimeSpan[k].last_prob + mtbp_threshold) continue;
        gAlarmSystemTimeSpan[k].last_prob = (UINT32)(t - mtbp_threshold);
        if (gAlarmSystemTimeSpan[k].end == 0)
        { // group was available  : terminate the span
          gAlarmSystemTimeSpan[k].end = (UINT32)gAlarmSystemTimeSpan[k].last_prob - 1;
          sched = TRUE;
          span = (int)(gAlarmSystemTimeSpan[k].end - gAlarmSystemTimeSpan[k].begin);
          updateMTBPInfo(span,0,k);
          updMTBP = TRUE;
        }
        else
        { // start the problems span :
          gAlarmSystemTimeSpan[k].begin = gAlarmSystemTimeSpan[k].end + 1;
          gAlarmSystemTimeSpan[k].end = 0;
          gAlarmSystemTimeSpan[k].nalarms = gNumberOfAlarms[k];
          gAlarmSystemTimeSpan[k].type = SPAN_PROBLEMS;
          gAlarmSystemTimeSpan[k].last_prob = 0; // reset
        }
      }
    }
    else
    { // no alarms OR no longer running
      if (gAlarmSystemTimeSpan[k].type == SPAN_PROBLEMS)
      {
        if (gAlarmSystemTimeSpan[k].last_avail == 0) gAlarmSystemTimeSpan[k].last_avail = (UINT32)t;
        if (t < (int)gAlarmSystemTimeSpan[k].last_avail + mtbp_threshold) continue;
        gAlarmSystemTimeSpan[k].last_avail = (UINT32)(t - mtbp_threshold);
        if (gAlarmSystemTimeSpan[k].end == 0)
        { // had problems but is now available : terminate the span
          gAlarmSystemTimeSpan[k].end = (UINT32)gAlarmSystemTimeSpan[k].last_avail - 1;
          sched = TRUE;
          updateMTBPInfo(0,(UINT32)t,k);
          updMTBP = TRUE;
        }
        else
        { // start the available span :
          gAlarmSystemTimeSpan[k].begin = gIsRunning ? gAlarmSystemTimeSpan[k].end + 1 : 0;
          gAlarmSystemTimeSpan[k].end = 0;
          gAlarmSystemTimeSpan[k].nalarms = 0;
          gAlarmSystemTimeSpan[k].type = SPAN_AVAILABLE;
          gAlarmSystemTimeSpan[k].last_avail = 0; // reset
        }
      }
      if (gAlarmSystemTimeSpan[k].type == SPAN_AVAILABLE) 
      { // marked available
        if (gAlarmSystemTimeSpan[k].begin == 0 && gIsRunning)
        { // and we're running (again) and haven't noted the begin time
          gAlarmSystemTimeSpan[k].last_avail = (UINT32)t;
          gAlarmSystemTimeSpan[k].begin = (UINT32)t;
          updateMTBPInfo(0,(UINT32)t,k);
          updMTBP = TRUE;
        }
      }
    }
  }
  if (sched) SystemScheduleProperty(ALAEQM_TAG,"TimeSpan");
  if (updMTBP)
  {
    writeMTBPFile(FALSE);
    writeMTBPFile(TRUE);
  }
  if (updAlmCnts) writeAlarmCountersToFile();
  if ((t%60) == 0) writeDeclaredCounters();
}

NAME64 isRunningName[1];
int RegisterAlaeqmProperties(void)
{
  int cc = 0;
  DTYPE dout, din;
  HistorySpecification hspc;

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_INT32;
  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"NOTREADYRUNNING",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1e+009 counts]Not Ready while running counter",PRP_NOTREADYRUNNING,NULL);
  if (cc < 0) ccerr(cc);
  cc = RegisterPropertyInformation(ALAEQM_TAG,"NOTREADYCOUNT",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1e+009 counts]Not Ready State counter",PRP_NOTREADYCOUNT,NULL);
  if (cc < 0) ccerr(cc);
  cc = RegisterPropertyInformation(ALAEQM_TAG,"ISREADY",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1]Is the designated subsystem free of alarms?",PRP_ISREADY,NULL);
  if (cc < 0) ccerr(cc);
  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"NUMALARMS",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:100 ]Number of Current Alarms",PRP_NUMALARMS,NULL);
  if (cc < 0) ccerr(cc);
  dout.dFormat = CF_NAME32;
  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"GROUPS",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"Subsystem Group Names",PRP_GROUPS,NULL);
  if (cc < 0) ccerr(cc);
  cc = RegisterPropertyInformation(ALAEQM_TAG,"MEMBERS",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"Subsystem Group Members",PRP_MEMBERS,NULL);
  if (cc < 0) ccerr(cc);
  cc = RegisterPropertyInformation(ALAEQM_TAG,"MEMBERS.NAM",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"Subsystem Group Names",PRP_MEMBERS_NAM,NULL);
  if (cc < 0) ccerr(cc);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"ISRUNNING",
           &dout,NULL,CA_READ,AT_SCALAR,0,"[0:1]Is the facility in a declared running state?",PRP_ISRUNNING,NULL);
  if (cc < 0) ccerr(cc);

  strncpy(isRunningName[0].name,GetExportedContext(ALAEQM_TAG),CONTEXT_NAME_SIZE);
  AssignDeviceListToProperty(ALAEQM_TAG,"ISRUNNING",isRunningName,1);

  dout.dFormat = CF_INTINTINTINT;
  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"TIMESPAN",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1]Availability/Problems time span information",PRP_TIMESPAN,NULL);
  if (cc < 0) ccerr(cc);

  dout.dFormat = CF_NAME16;
  dout.dArrayLength = 4;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"TIMESPAN.ENUM",
           &dout,NULL,CA_READ,AT_UNKNOWN,0,"field enumerations of the TIMESPAN quad variable",PRP_TIMESPAN_ENUM,NULL);
  if (cc < 0) ccerr(cc);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"MeanTimeBetweenProblems",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1000000 sec]mean time between problems",PRP_MTBP,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"NumberProblems",
           &dout,NULL,CA_READ,AT_CHANNEL,0,"[0:1000 sec]number of problems since epoch begin",PRP_NUMPROBLEMS,NULL);
  if (cc < 0) ccerr(cc);

  dout.dArrayLength = 1;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"ProblemsCountingEpoch",
           &dout,NULL,CA_READ|CA_WRITE,AT_SCALAR,0,"start of the problems counting epoch as UTC",PRP_MTBP_EPOCH,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"TIMESPAN.THRESHOLD",
           &dout,&dout,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_SCALAR,0,"[0:6000 sec]threshold for recognizing a problem/available span",PRP_MTBP_THRESHOLD,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredProblems.Window",
           &dout,&dout,CA_READ|CA_WRITE,AT_SCALAR,0,"[0:900 sec]window for officially declared problems state",PRP_DECLPROBLEMS_WINDOW,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredRunning.Window",
           &dout,&dout,CA_READ|CA_WRITE,AT_SCALAR,0,"[0:900 sec]window for officially declared running state",PRP_DECLRUNNING_WINDOW,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredProblems.Time",
           &dout,NULL,CA_READ,AT_SCALAR,0,"time of officially declared problems state",PRP_DECLPROBLEMS_TIME,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredRunning.Time",
           &dout,NULL,CA_READ,AT_SCALAR,0,"time of last officially declared running state",PRP_DECLRUNNING_TIME,NULL);
  if (cc < 0) ccerr(cc);

  dout.dArrayLength = MAX_ALARM_SYSTEMS;

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredProblems.Counts",
           &dout,NULL,CA_READ,AT_SCALAR,0,"counter for officially declared problems state",PRP_DECLPROBLEMS_COUNT,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredRunning.Counts",
           &dout,NULL,CA_READ,AT_SCALAR,0,"counter for officially declared running state",PRP_DECLRUNNING_COUNT,NULL);
  if (cc < 0) ccerr(cc);
  
  dout.dFormat = CF_TEXT;
  dout.dArrayLength = 64;
  cc = RegisterPropertyInformation(ALAEQM_TAG,"ProblemsCountingEpoch",
           &dout,NULL,CA_READ,AT_SCALAR,0,"start of the problems counting epoch as string",PRP_MTBP_EPOCH,NULL);
  if (cc < 0) ccerr(cc);
  AssignDeviceListToProperty(ALAEQM_TAG,"ProblemsCountingEpoch",isRunningName,1);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredProblems.Time",
           &dout,NULL,CA_READ,AT_SCALAR,0,"time of officially declared problems state",PRP_DECLPROBLEMS_TIME,NULL);
  if (cc < 0) ccerr(cc);

  cc = RegisterPropertyInformation(ALAEQM_TAG,"DeclaredRunning.Time",
           &dout,NULL,CA_READ,AT_SCALAR,0,"time of last officially declared running state",PRP_DECLRUNNING_TIME,NULL);
  if (cc < 0) ccerr(cc);

  DTYPEZERO(dout);
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"AvailCor");
  RegisterPropertyInformation(ALAEQM_TAG,"Corrections",
           &dout,&din,CA_WRITE,AT_UNKNOWN,0,"add new avail correction",PRP_CORRECTIONS,NULL);

  RegisterPropertyInformation(ALAEQM_TAG,"AssignedBlame",
           &dout,&din,CA_WRITE,AT_UNKNOWN,0,"add new avail correction",PRP_BLAME,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = MAX_CORRECTION_PER_DAY;
  strcpy(dout.dTag,"AvailCor");
  DTYPEZERO(din);
  din.dFormat = CF_INT32;
  din.dArrayLength = 2;
  RegisterPropertyInformation(ALAEQM_TAG,"Corrections",
           &dout,&din,CA_READ,AT_UNKNOWN,0,"get avail corrections over input range",PRP_CORRECTIONS,NULL);

  RegisterPropertyInformation(ALAEQM_TAG,"AssignedBlame",
           &dout,&din,CA_READ,AT_UNKNOWN,0,"get assigned blame over input range",PRP_BLAME,NULL);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  DTYPEZERO(din);
  din.dFormat = CF_INT32;
  din.dArrayLength = 2;
  RegisterPropertyInformation(ALAEQM_TAG,"CorrectionCounts.Systems",
           &dout,&din,CA_READ,AT_CHANNEL,0,"get system correction Counts over a range",PRP_SYS_CORRECTION_CNTS,NULL);

  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  RegisterPropertyInformation(ALAEQM_TAG,"CorrectionCounts.Groups",
           &dout,&din,CA_READ,AT_CHANNEL,0,"get group correction Counts over a range",PRP_GRP_CORRECTION_CNTS,NULL);

  dout.dArrayLength = MAX_ALARM_SYSTEMS;
  RegisterPropertyInformation(ALAEQM_TAG,"BlameCounts.Systems",
           &dout,&din,CA_READ,AT_CHANNEL,0,"get system Blame Counts over a range",PRP_SYS_BLAME_CNTS,NULL);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = NTHRESHOLD_TYPES;
  din.dFormat = CF_INT32;
  din.dArrayLength = 1;
  RegisterPropertyInformation(ALAEQM_TAG, "Thresholds",
    &dout, &din, CA_READ|CA_WRITE|CA_SAVERESTORE, AT_CHANNEL, 0, "[0:3600 sec]operation history statistics thresholds (seconds)", PRP_THRESHOLDS, NULL);

  AssignDeviceListToProperty(ALAEQM_TAG, "Thresholds", gThresholdTypes, NTHRESHOLD_TYPES);

  memset(&hspc,0,sizeof(hspc));
  hspc.pollingRate = 1000;
  hspc.archiveRate = 1000;
  hspc.heartbeat = 1200;
  hspc.depthShort = 10;
  hspc.depthLong = 3;
  AppendHistoryInformation(ALAEQM_TAG,"NOTREADYCOUNT","#0",MAX_ALARM_SYSTEMS,CF_INT32,1,&hspc);
  AppendHistoryInformation(ALAEQM_TAG,"TIMESPAN","#0",MAX_ALARM_SYSTEMS,CF_INTINTINTINT,2,&hspc);
  AppendHistoryInformation(ALAEQM_TAG,"MeanTimeBetweenProblems","#0",MAX_ALARM_SYSTEMS,CF_INT32,3,&hspc);

  AppendHistoryInformation(ALAEQM_TAG,"DeclaredRunning.Counts","#0",MAX_ALARM_SYSTEMS,CF_INT32,10,&hspc);
  AppendHistoryInformation(ALAEQM_TAG,"DeclaredProblems.Counts","#0",MAX_ALARM_SYSTEMS,CF_INT32,11,&hspc);
  AppendHistoryInformation(ALAEQM_TAG,"DeclaredRunning.Time","#0",1,CF_INT32,12,&hspc);
  AppendHistoryInformation(ALAEQM_TAG,"DeclaredProblems.Time","#0",1,CF_INT32,13,&hspc);

err:
  if (cc < 0)
  {
    dbglog("Can't register property : %s\n>",erlst[-cc]);
  }
  return cc;
}

#define quit(i) { feclog("%s: %s",ALAEQM_TAG,erlst[i]); exit(1); }
void RegisterStructs(void)
{
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,start),1,CF_DOUBLE,"startUtc")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,stop),1,CF_DOUBLE,"stopUtc")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,counts),1,CF_INT32,"counts")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,to),1,CF_INT32,"hasProblems")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,system),64,CF_TEXT,"system")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,state),64,CF_TEXT,"state")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,startStr),64,CF_TEXT,"startStr")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,stopStr),64,CF_TEXT,"stopStr")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,usr),16,CF_TEXT,"user")) quit(out_of_local_memory);
  if (addFieldToStruct("AvailCor",OFFSETIN(StateCorrection,cmt),256,CF_TEXT,"comment")) quit(out_of_local_memory);
  if (sealTaggedStruct("AvailCor",sizeof(StateCorrection),MAX_CORRECTION_PER_DAY)) quit(out_of_local_memory);
}

void alaeqm_ini(void)
{
  char devnam[256];
  int cc=0, i, id, almrq[3];
  DTYPE dout, din;
  FILE *fp = NULL;
  char str[256];
  int srv_col=-1,prp_col=-1;
  int done = 0;
  NAME64 devlst[128];

  remove("problems.log");

  if (strlen(gDeviceContext) == 0)
  {
    feclog("Cannot determine device context !");
    exit(1);
  }

  RegisterStructs();

  RegisterAlaeqmProperties();

  srvlog(LNKLOGFILE,"init","initializing alarm state server");

  if ((fp=fopen(EPOCH_FILE,"r")) != NULL)
  { // found an epoch file
    int ep_col=-1;
    char scratch[64];
    done = FALSE;
    while (csvGetLine(str,255,fp) != NULL)
    {
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        ep_col = findcol(str,"EPOCH_BEGIN");
        done = 1;
        continue;
      }
      if (ep_col >= 0)
      {
        strncpy(scratch,colptr(ep_col,str),64);
        gMTBPEpochBegin = getTimeStampFromLogEntry(scratch);
      }
      break;
    }
    fclose(fp);
  }
  else
  { // not there yet ...
    gMTBPEpochBegin = time(NULL);
    writeMTBPEpochFile();
  }

  /* default settings ... */
  sprintf(gCasServer,"/%s/CAS",gDeviceContext);
  sprintf(gDeclStateSrv,"/%s/STATE/#0",gDeviceContext);
  strncpy(gDeclStatePrp,"DECLSTATE",64);
  if ((fp=fopen(STATEINFO_FILE,"r")) != NULL)
  { /* found a local config file : use it ! */
    done = FALSE;
    while (csvGetLine(str,255,fp) != NULL)
    {
      if (strlen(str) == 0) continue;
      if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
      if (!done)
      {
        srv_col = findcol(str,"STATESERVER");
        prp_col = findcol(str,"STATEPROPERTY");
        done = 1;
        continue;
      }
      if (srv_col >= 0)
        sprintf(gDeclStateSrv,"/%s/%s/#0",gDeviceContext,colptr(srv_col,str));
      if (prp_col >= 0)
        strncpy(gDeclStatePrp,colptr(prp_col,str),64);
      break;
    }
    fclose(fp);
  }
  feclog("using CAS %s",gCasServer);
  feclog("using declared state server %s and property %s",gDeclStateSrv,gDeclStatePrp);

  for (i=0; (gNumAlarmSystems=getAlarmSystems(gCasServer,gAlarmSystems)) < 0; i++)
  {
    if (i < 5)
    { /* the CAS sometimes needs a bit of time to get going (in case of simultaneous starts) */
      millisleep(1000);
      continue;
    }
    feclog("Cannot determine alarm systems from %s : %s",gCasServer,erlst[-gNumAlarmSystems]);
    exit(1);
  }
  if (gNumAlarmSystems > MAX_ALARM_SYSTEMS) 
  {
    feclog("truncating number of alarm systems found from %d to %d",gNumAlarmSystems,MAX_ALARM_SYSTEMS);
    gNumAlarmSystems = MAX_ALARM_SYSTEMS;
  }
  if ((cc=readAlarmCountersFromFile()) != 0)
  { /* no ALARM_COUNTERS_FILE */
    if ((cc=writeAlarmCountersToFile()) != 0)
    {
      feclog("cannot make alarm system counters file : %s",erlst[cc]);
      logproblem("cannot make alarm system counters file : %s",erlst[cc]);
    }
    else
    {
      feclog("make new alarm system counters file");
    }
  }
  if ((cc=getAlarmGroups()) != 0)
  {
    gNumAlarmGroups = 1;
    strncpy(gAlarmGroups[0].grp,"ALL",48);
    strncpy(gAlarmGroups[0].dsc,gDeviceContext,48);
    if ((cc=writeAlarmGroupsToFile()) != 0)
    {
      feclog("cannot make new groups file : %s",erlst[cc]);
      logproblem("cannot make new groups file : %s",erlst[cc]);
    }
    else
    {
      feclog("make new alarm group counters file");
      if ((cc=makeALLGroup()) != 0)
      {
        feclog("cannot make new ALL group file : %s",erlst[cc]);
        logproblem("cannot make new ALL group file : %s",erlst[cc]);
      }
    }
  }
  else
  {
    if ((cc=updateALLGroup()) != 0)
    {
        feclog("cannot update ALL group file : %s",erlst[cc]);
        logproblem("cannot update ALL group file : %s",erlst[cc]);
    }
  }
  if ((cc=readMTBPFile(FALSE)) != 0)
  {
    if (cc == no_such_file)
    { // initial file ...
      writeMTBPFile(FALSE);
    }
    else
    {
      feclog("read meantime between problems (alarm systems) : %.32s",cc2str(cc));
      logproblem("read meantime between problems (alarm systems) : %.32s",cc2str(cc));
    }
  }
  if ((cc=readMTBPFile(TRUE)) != 0)
  {
    if (cc == no_such_file)
    {
      writeMTBPFile(TRUE);
    }
    else
    {
      feclog("read meantime between problems (alarm groups) : %.32s",cc2str(cc));
      logproblem("read meantime between problems (alarm systems) : %.32s",cc2str(cc));
    }
  }
  if ((cc=readDeclaredCounters()) != 0)
  {
    if (cc == no_such_file)
    {
      writeDeclaredCounters();
    }
    else
    {
      feclog("read declared counters file : %.32s",cc2str(cc));
      logproblem("read declared counters file : %.32s",cc2str(cc));
    }
  }
  /* get the last 21 minutes of alarms (default heartbeat + 1) */
  almrq[0] = -1260; almrq[1] = -1; almrq[2] = ALM_SEVERITY_ERR+1;
  din.dFormat = CF_LONG;
  din.dArrayLength = 3;
  din.data.lptr = (SINT32 *)almrq;
  din.dTag[0] = 0;
  dout.dFormat = CF_LONG;
  dout.dArrayLength = 2;
  dout.dTag[0] = 0;
  
  cc = 0;
  for (i=0; i<gNumAlarmSystems; i++)
  {
    if (gAlarmSystems[i].name[0] == '.')
    { // a hidden alarm system
      strncpy(devlst[i].name,&gAlarmSystems[i].name[1],63);
      devlst[i].name[63] = 0;
    }
    else
    {
      strncpy(devlst[i].name,gAlarmSystems[i].name,64);
    }
    RegisterDeviceName(ALAEQM_TAG,devlst[i].name,i);
    dout.data.lptr = (SINT32 *)&gNumberOfAlarmsX[i];
    sprintf(devnam,"%.64s/SYS#%d",gCasServer,(int)gAlarmSystems[i].d1val);
    id = AttachLinkEx(devnam,"NUMALARMS",&dout,&din,CA_READ,1000,checkalarms,CM_DATACHANGE,i+1);
    if (id < 0) 
    {
      cc = -id;
      feclog("cannot start link to %s : %.32s",devnam,erlst[cc]);
      logproblem("cannot start link to %s : %.32s",devnam,erlst[cc]);
      srvlog(LNKLOGFILE,"init","cannot start link to %s : %.32s",devnam,erlst[cc]);
    }
    else
    {
      srvlog(LNKLOGFILE,"init","start link to %s : link id %d",devnam,id);
    }
  }
  AssignDeviceListToProperty(ALAEQM_TAG,"CorrectionCounts.Systems",devlst,gNumAlarmSystems);

  for (i=0; i<gNumAlarmGroups; i++)
  {
    sprintf(devnam,"%.60s",gAlarmGroups[i].grp);
    RegisterDeviceName(ALAEQM_TAG,devnam,i+gNumAlarmSystems);
    strncpy(devlst[i].name,devnam,48);
  }
  AssignDeviceListToProperty(ALAEQM_TAG,"CorrectionCounts.Groups",devlst,gNumAlarmGroups);

  AssignNumModules(ALAEQM_TAG,gNumAlarmSystems+gNumAlarmGroups);

  getOfficiallyRunningStates();
  getOfficiallyProblemStates();
  getNonRunningStates();
  if (nNonRunningStates > 0 || nRunningStates > 0)
  {
    dout.dFormat = CF_NAME64;
    dout.dArrayLength = 1;
    dout.data.cptr = gDeclaredState;
    id = AttachLink(gDeclStateSrv,gDeclStatePrp,&dout,NULL,CA_READ,1000,checkstate,CM_DATACHANGE);
    if (id < 0) 
    {
      cc = -id;
      feclog("cannot start link to %s : %.32s",gDeclStateSrv,erlst[cc]);
      logproblem("cannot start link to %s : %.32s",gDeclStateSrv,erlst[cc]);
      srvlog(LNKLOGFILE,"init","cannot start link to %s : %.32s",gDeclStateSrv,erlst[cc]);
    }
    else
    {
      srvlog(LNKLOGFILE,"init","start link to %s : link id %d",gDeclStateSrv,id);
    }
  }
  if (cc)
  {
    feclog("cannot contact central services : %.32s",erlst[cc]);
    logproblem("cannot contact central services : %.32s",erlst[cc]);
    exit(1);
  }
}

void alaeqm_exi(void)
{
  writeDeclaredCounters();
}

int alaeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,i,k,idx,slen,cc,ntotal,ncallers=1;
  int ready[MAX_ALARM_SYSTEMS];
  ClnInfo clnInfo;

  devnr = GetDeviceNumberEx(ALAEQM_TAG,devName,devProperty);
  if (devnr < 0) return illegal_equipment_number;

  prpid = GetPropertyId(ALAEQM_TAG,devProperty);

  GetCallerInformation(ALAEQM_TAG,&clnInfo,&ncallers); 

  ntotal = gNumAlarmSystems + gNumAlarmGroups;
  switch (prpid)
  {
    case PRP_NOTREADYRUNNING:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > (UINT32)ntotal) dout->dArrayLength = ntotal;
          if ((cc=putValuesFromLong(dout,(SINT32 *)&gAlarmCounterRunning[devnr],MAX_ALARM_SYSTEMS-devnr)) != 0) return cc;
        }
        return 0;
    case PRP_NOTREADYCOUNT:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > (UINT32)ntotal) dout->dArrayLength = ntotal;
          if ((cc=putValuesFromLong(dout,(SINT32 *)&gAlarmCounter[devnr],MAX_ALARM_SYSTEMS-devnr)) != 0) return cc;
        }
        return 0;
    case PRP_ISRUNNING:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if ((cc=putValuesFromLong(dout,(SINT32 *)&gIsRunning,1)) != 0) return cc;
        }
        return 0;
    case PRP_TIMESPAN:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_INTINTINTINT) return illegal_format;
          if (dout->dArrayLength > (UINT32)ntotal) dout->dArrayLength = ntotal;
          if ((cc=PutValuesFromAny(dout,gAlarmSystemTimeSpan,CF_INTINTINTINT,1,sizeof(TimeSpan),MAX_ALARM_SYSTEMS,devnr)) != 0)
            return cc;
        }
        return 0;
    case PRP_TIMESPAN_ENUM:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          static NAME16 TimeSpanEnum[4] = { "type", "nalarms", "begin", "end" };
          if (dout->dArrayLength > 4) dout->dArrayLength = 4;
          if ((cc=PutValuesFromString(dout,TimeSpanEnum[0].name,CF_NAME16,4,0)) != 0) return cc;
        }
        return 0;
    case PRP_MTBP_THRESHOLD:
        if (access&CA_WRITE) 
        {
          int th = 0;
          if ((cc=GetValuesAsLong(din,(SINT32 *)&th,1)) != 0) return cc;
          mtbp_threshold = th;
        }
        if (dout->dArrayLength > 0)
        {
          if ((cc=PutValuesFromLong(dout,(SINT32 *)&mtbp_threshold,1)) != 0) return cc;
        }
        return 0;
    case PRP_MTBP:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > (UINT32)ntotal) dout->dArrayLength = ntotal;
          if ((cc=PutValuesFromAny(dout,&mtbpInfoSet[0].mtbp,CF_INT32,1,sizeof(MTBP),MAX_ALARM_SYSTEMS,devnr)) != 0)
            return cc;
        }
        return 0;
    case PRP_NUMPROBLEMS:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > (UINT32)ntotal) dout->dArrayLength = ntotal;
          if ((cc=PutValuesFromAny(dout,&mtbpInfoSet[0].nproblems,CF_INT32,1,sizeof(MTBP),MAX_ALARM_SYSTEMS,devnr)) != 0)
            return cc;
        }
        return 0;
    case PRP_MTBP_EPOCH:
        if (access&CA_WRITE) 
        {
          cc = resetMTBPEpoch();
          if (cc != 0) return cc;
        }
        if (dout->dArrayLength > 0)
        {
          if (isSimpleStringFormat(dout->dFormat))
          {
            char tstr[64];
            strncpy(tstr,ctime(&gMTBPEpochBegin),24); tstr[24] = 0;
            cc = PutValuesFromString(dout,tstr,CF_TEXT,64,0);
          }
          else
          {
            UINT32 lv = (UINT32)gMTBPEpochBegin;
            cc = PutValuesFromLong(dout,(UINT32 *)&lv,1);
          }
          return cc;
        }
        return 0;
    case PRP_DECLRUNNING_TIME:
        if (dout->dArrayLength > 0)
        { 
          if (isSimpleStringFormat(dout->dFormat))
          {
            char tstr[64];
            strncpy(tstr,ctime(&timeDeclaredRunning),24); tstr[24] = 0;
            return PutValuesFromString(dout,tstr,CF_TEXT,64,0);
          }
          else
          {
            UINT32 lv = (UINT32)timeDeclaredRunning;
            return PutValuesFromLong(dout,(SINT32 *)&lv,1);
          }
        }
        return 0;
    case PRP_DECLRUNNING_COUNT:
        if (dout->dArrayLength > 0)
        {
          return putValuesFromLong(dout,(SINT32 *)&countsDeclaredRunning,1);
        }
        return 0;
    case PRP_DECLRUNNING_WINDOW:
        if (access & CA_WRITE)
        {
          int w = 0;
          if (din->dArrayLength == 0) return dimension_error;
          if ((cc=GetValuesAsLong(din,(SINT32 *)&w,1)) != 0) return cc;
          if (w < 0 || w > 900) return out_of_range;
          windowDeclaredRunning = w;
          writeDeclaredCounters();
        }
        if (dout->dArrayLength > 0)
        {
          return putValuesFromLong(dout,(SINT32 *)&windowDeclaredRunning,1);
        }
        return 0;
    case PRP_DECLPROBLEMS_TIME:
        if (dout->dArrayLength > 0)
        { 
          if (isSimpleStringFormat(dout->dFormat))
          {
            char tstr[64];
            strncpy(tstr,ctime(&timeDeclaredProblems),24); tstr[24] = 0;
            return PutValuesFromString(dout,tstr,CF_TEXT,64,0);
          }
          else
          {
            UINT32 lv = (UINT32)timeDeclaredProblems;
            return PutValuesFromLong(dout,(SINT32 *)&lv,1);
          }
        }
        return 0;
    case PRP_DECLPROBLEMS_COUNT:
        if (dout->dArrayLength > 0)
        {
          if (stricmp(devName,"ALL") == 0)
          {
            return putValuesFromLong(dout,(SINT32 *)&countsDeclaredProblems,1);
          }
          else
          {
            return PutValuesFromLongEx(dout,(SINT32 *)&syscntsDeclaredProblems,MAX_ALARM_SYSTEMS,devnr);
          }
        }
        return 0;
    case PRP_DECLPROBLEMS_WINDOW:
        if (access & CA_WRITE)
        {
          int w = 0;
          if (din->dArrayLength == 0) return dimension_error;
          if ((cc=GetValuesAsLong(din,(SINT32 *)&w,1)) != 0) return cc;
          if (w < 0 || w > 900) return out_of_range;
          windowDeclaredProblems = w;
          writeDeclaredCounters();
        }
        if (dout->dArrayLength > 0)
        {
          return putValuesFromLong(dout,(SINT32 *)&windowDeclaredRunning,1);
        }
        return 0;
    case PRP_ISREADY:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          for (i=0; i<MAX_ALARM_SYSTEMS; i++) ready[i] = gNumberOfAlarms[i] ? 0 : 1;
          if (dout->dArrayLength > (UINT32)ntotal) dout->dArrayLength = ntotal;
          switch (dout->dFormat)
          {
            case CF_TEXT:
              if (dout->dArrayLength < 8) return dimension_error;
              strcpy(dout->data.cptr, ready[devnr] ? "YES" : "NO");
              break;
            case CF_NAME16:
              if (dout->dArrayLength > (UINT32)MAX_ALARM_SYSTEMS) return dimension_error;
              for (i=devnr; i<(SINT32)dout->dArrayLength-devnr && i<MAX_ALARM_SYSTEMS; i++)
                strcpy(((NAME16 *)dout->data.cptr)[i].name,ready[i]?"YES":"NO");
              break;
            case CF_NAME32:
              if (dout->dArrayLength > (UINT32)MAX_ALARM_SYSTEMS) return dimension_error;
              for (i=devnr; i<(SINT32)dout->dArrayLength-devnr && i<MAX_ALARM_SYSTEMS; i++)
                strcpy(((NAME32 *)dout->data.cptr)[i].name,ready[i]?"YES":"NO");
              break;
            default:
              if (dout->dArrayLength > (UINT32)MAX_ALARM_SYSTEMS) return dimension_error;
              if ((cc=putValuesFromLong(dout,(SINT32 *)&ready[devnr],MAX_ALARM_SYSTEMS-devnr)) != 0) return cc;
              break;
          }
        }
        return 0;
    case PRP_NUMALARMS:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > (UINT32) ntotal) dout->dArrayLength = ntotal;
          if ((cc=putValuesFromLong(dout,(SINT32 *)&gNumberOfAlarms[devnr],MAX_ALARM_SYSTEMS-devnr)) != 0) return cc;
        }
        return 0;
    case PRP_MEMBERS:
        if (access&CA_WRITE) return illegal_read_write;
        //if (devnr < gNumAlarmSystems) return illegal_equipment_number;
        idx = devnr; // - gNumAlarmSystems;
        if (dout->dArrayLength > 0)
        {
          switch (dout->dFormat)
          {
            case CF_NAME16: 
            case CF_NAME32: 
              slen = fmtsizeof(dout->dFormat);
              break;
            default:
              return illegal_format;
          }
          if (dout->dArrayLength > (UINT32)gAlarmGroups[idx].nindices)
            dout->dArrayLength = gAlarmGroups[idx].nindices;
          for (i=0; i<(SINT32)dout->dArrayLength; i++)
          {
            k = gAlarmGroups[idx].indices[i];
            strncpy(dout->data.cptr+i*slen,gAlarmSystems[k].name,slen);
          }
        }
        return 0;
    case PRP_MEMBERS_NAM:
    case PRP_GROUPS:
        if (access&CA_WRITE) return illegal_read_write;
        if (devnr < 0 || devnr >= gNumAlarmGroups) return illegal_equipment_number;
        if (dout->dArrayLength > 0)
        {
          int clen = 48;
          if (dout->dArrayLength > (UINT32)(gNumAlarmGroups-devnr)) dout->dArrayLength = (UINT32)(gNumAlarmGroups-devnr);
          switch (dout->dFormat)
          {
            case CF_NAME16: 
            case CF_NAME32: 
              clen = fmtsizeof(dout->dFormat);
            case CF_NAME48:
            case CF_NAME64:
              slen = fmtsizeof(dout->dFormat);
              break;
            default:
              return illegal_format;
          }
          for (i=0; i<(SINT32)dout->dArrayLength; i++)
          {
            strncpy(dout->data.cptr+i*slen,gAlarmGroups[i+devnr].grp,clen);
          }
        }
        return 0;
    case PRP_BLAME:
    case PRP_CORRECTIONS:
      {
        time_t stop = time(NULL);
        time_t start = stop - 86400;
        char *fn = prpid == PRP_CORRECTIONS ? CORRECTIONSFILE : BLAMEFILE;
        if (din->dArrayLength > 0)
        {
          if (access & CA_WRITE)
          {
            StateCorrection *sc = (StateCorrection *)din->data.vptr;
            if (din->dFormat != CF_STRUCT) return illegal_format;
            if (strcmp(din->dTag,"AvailCor")) return invalid_structure_tag;
            if (sc->start < stop - 86400*7)
            { // must be an admin to apply 'old' corrections
              SetEqmCompletion(ALAEQM_TAG,"admin rights required for corrections older than 1 week");
              if (!AssertIsAdministrator(clnInfo.userName)) return admin_rights_required;
            }
            return addStateCorrections(fn,sc); 
          }
          if (din->dArrayLength < 2) return dimension_error;
          switch (din->dFormat)
          {
            case CF_INT32:
              start = (time_t)din->data.lptr[0];
              stop = (time_t)din->data.lptr[1];
              break;
            case CF_DOUBLE:
              start = (time_t)din->data.dptr[0];
              stop = (time_t)din->data.dptr[1];
              break;
            default:
              return illegal_format;
          }
        }
        if (dout->dArrayLength > 0)
        {
          int num = dout->dArrayLength;
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strcmp(dout->dTag,"AvailCor")) return invalid_structure_tag;
          if ((cc=readStateCorrections(fn,start,stop,(StateCorrection *)dout->data.vptr,&num)) != 0) return cc;
          if (num < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)num;
        }
        return 0;
      }
    case PRP_SYS_BLAME_CNTS:
    case PRP_SYS_CORRECTION_CNTS:
    case PRP_GRP_CORRECTION_CNTS:
      {
        time_t stop = time(NULL);
        time_t start = stop - 86400; 
        int n=0, statescor[256];
        if (dout->dArrayLength < 1) return dimension_error;
        if (din->dArrayLength > 0)
        {
          if (access & CA_WRITE) return illegal_read_write;
          if (din->dArrayLength < 2) return dimension_error;
          switch (din->dFormat)
          {
            case CF_INT32:
              start = (time_t)din->data.lptr[0];
              stop = (time_t)din->data.lptr[1];
              break;
            case CF_DOUBLE:
              start = (time_t)din->data.dptr[0];
              stop = (time_t)din->data.dptr[1];
              break;
            default:
              return illegal_format;
          }
        }
        if (prpid == PRP_SYS_CORRECTION_CNTS)
        {
          if ((cc=getStateCorrectionCounts(CORRECTIONSFILE,start,stop, statescor, gNumAlarmSystems)) != 0) return cc;
          n = gNumAlarmSystems;
        }
        else if (prpid == PRP_SYS_BLAME_CNTS)
        {
          if ((cc=getStateCorrectionCounts(BLAMEFILE,start,stop, statescor, gNumAlarmSystems)) != 0) return cc;
          n = gNumAlarmSystems;
        }
        else
        {
          if ((cc=getGroupCorrectionCounts(start,stop, statescor, gNumAlarmGroups)) != 0) return cc;
          n = gNumAlarmGroups;
        }
        if (dout->dArrayLength > (UINT32)n) dout->dArrayLength = (UINT32)n;
        return PutValuesFromLongEx(dout,statescor,n,devnr);
      }
    case PRP_THRESHOLDS:
    {
      if (devnr < 0 || devnr >= NTHRESHOLD_TYPES) return illegal_device_number;
      if (din->dArrayLength > 0)
      {
        int val = 0;
        if (!(access & CA_WRITE)) return illegal_read_write;
        if ((cc=GetValuesAsLong(din, &val, 1)) != 0) return cc;
        gThresholds[devnr] = val;
      }
      if (dout->dArrayLength > (UINT32)NTHRESHOLD_TYPES) dout->dArrayLength = (UINT32)NTHRESHOLD_TYPES;
      return PutValuesFromLongEx(dout, gThresholds, NTHRESHOLD_TYPES, devnr);
    }
    default:
        return illegal_property;
  }
}
int logproblem(char *text,...)
{
  int cc=0;
  double ts = makeDataTimeStamp();
  char str[256];
  FILE *fp = NULL;
  va_list args;
  if (strlen(text) > 80) ccerr(string_too_long);
  va_start(args,text);
  sprintf(str,"%s : ",getDataTimeString(ts,0));
  vsprintf(&str[strlen(str)],text,args);
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  va_end(args);
  if ((fp=fopen("problems.log","a")) == NULL) ccerr(file_error);
  fwrite(str,strlen(str),1,fp);
err:
  if (fp) fclose(fp);
  return cc;
}
FILE *createAndOpen(char *fname,char *attr)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  static char strdel[2] = { FS_DELIMITER , 0};

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,strdel); c != NULL; c=strtok(NULL,strdel))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return fopen(fname,attr);
    }
    strcat(dirname,strdel);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH)) return NULL;
    }
#   elif defined(WIN32)
    sprintf(fn,"%.120s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return NULL;
    }
#   endif
  }
  return NULL;
}

