/*
 * File devsrv.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "app_version.h"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include "tine.h"
#include "archive.h"

void PreSystemInit(void);
void PostSystemInit(void);
extern int needsRestart;
void RestartServer(void);
extern char adbFileName[64];

extern char thisFecName[FEC_NAME_SIZE];
extern char thisContext[CONTEXT_NAME_SIZE];
extern char thisLocation[FEC_LOCATION_SIZE];
extern char thisServerName[EXPORT_NAME_SIZE];
extern char thisFailoverMasterName[EXPORT_NAME_SIZE];
extern char thisFailoverSlaveMasterName[EXPORT_NAME_SIZE];
extern UINT16 thisPortOffset;
extern int startupInputsGiven;

char *usageText = "usage: arcsrv [/f=<fecname> /c=<context> /s=<server> /p=<port offset> /fm=<failover master name> /fs=<failover slave master> /e=<failover error count> /l=<location> /n=<database file name> /d=<debug level>";
int main(int argc, char *argv[])
{
  int i,cc,failoverErrorCount=0;
  char *c, scratch[256];

  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'f':
      {
        char *dst = thisFecName;
        int len = FEC_NAME_SIZE;
        switch (tolower(c[1]))
        {
          case 'm': /* failover master specified */
            dst = thisFailoverMasterName;
            len = EXPORT_NAME_SIZE;
            break;
          case 's': /* failover slave master specified */
            dst = thisFailoverSlaveMasterName;
            len = EXPORT_NAME_SIZE;
            break;
          default:
            startupInputsGiven = TRUE;
            break;
        }
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(dst,c,len);
        }
        continue;
      }
      case 'e':
      {
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          failoverErrorCount = atoi(scratch);      
        }
        continue;      
      }
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisContext,c,CONTEXT_NAME_SIZE);
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisServerName,c,EXPORT_NAME_SIZE);
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisLocation,c,FEC_LOCATION_SIZE);
        }
        continue;
      case 'p':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          thisPortOffset = atoi(scratch);      
        }
        startupInputsGiven = TRUE;
        continue;
      case 'n':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(adbFileName,c,64);
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
        }
        continue;
      case '?':
      case 'h':
        printf("%s",usageText);
        exit(0);
      default:
        continue;
    }
  }

  AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);
  SetPreSystemInitFunction(PreSystemInit);
  SetPostSystemInitFunction(PostSystemInit);
  /* init server: */
  if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }
  /* call FEC specific initialization: */
  if( failoverErrorCount) 
  {
    SetFailoverThreshold(failoverErrorCount);
  } 
  for(;;)
  {
    SystemCycle(TRUE);
    if (needsRestart) RestartServer();
    slave_mode_diagnostics();
  }
  return 0;
}
