/*
 * File mvanno.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>

int copyfile(char *src,char *dst)
{
  int in_fd = open(src, O_RDONLY);
  int out_fd = open(dst, O_RDWR|O_CREAT|O_APPEND,S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
  int siz, sizt = 0, cc = 0;
  char buf[8192];

  if (in_fd < 0 || out_fd < 0) 
  { 
    cc = -1; 
    goto err; 
  }
  while (1) 
  {
    siz = read(in_fd, &buf[0], sizeof(buf));
    sizt += siz;
    if (siz <= 0) break;
    write(out_fd, &buf[0], siz);
  }
err:
  printf("copy %s to %s: %s (%d bytes)\n",src,dst,(cc?"error":"ok"),sizt);
  return 0;
}
int mkannodir(char *dirname)
{
  DIR *dp=opendir(dirname);
  if (dp != NULL) 
  {
    closedir(dp);
    return 0;
  }
  return mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH);
}
int mvfiles(int yy,int mm)
{
  DIR *dp, *adp;
  struct stat s;
  struct dirent *dirp, *adirp;
  char dirname[256],adirname[256],newdir[256];
  char src[512],dst[512];

  sprintf(dirname,"../DATA/%04d/%02d",yy,mm);
  if ((dp=opendir(dirname)) == NULL) return 0;
  while ((dirp=readdir(dp)) != NULL)
  {
    if (strncmp(dirp->d_name,"annotate.",9) != 0) continue;
    if (stat(newdir,&s) == 0 && (s.st_mode & S_IFDIR) != S_IFDIR) continue;
    sprintf(newdir,"../DATA/%s",dirp->d_name);  
    mkannodir(newdir);
    sprintf(adirname,"%s/%s",dirname,dirp->d_name);
    if ((adp=opendir(adirname)) == NULL) continue;
    while ((adirp=readdir(adp)) != NULL)
    { // copy the files inside the annotate directory
      if (strcmp(adirp->d_name,".") == 0) continue;
      if (strcmp(adirp->d_name,"..") == 0) continue;
      sprintf(src,"%s/%s/%s",dirname,dirp->d_name,adirp->d_name);
      sprintf(dst,"%s/%s",newdir,adirp->d_name);
      copyfile(src,dst);
    }
    closedir(adp);
  }
  closedir(dp);
  return 0;
}

int main(int argc, char *argv[])
{
  int i,m;

  for(i=2010;i<2017;i++) for (m=1; m<13; m++)
  {
    mvfiles(i,m);
  }
  return 0;
}
