/*
 * File whichrec.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include "tine.h"
#include "archive.h"

extern int gAllocDataBuffers;
int getEntryFromRecno(int rec,int nrecs);
int getNumLines(FILE *fp);
DAQRecord *daqTbl;
int totalNrRecords, nrecords;
int nadbfiles = 0;
int archdebug = 0,archPending = 0;
int recordsCapacity = 0;
int g_isTerminating = 0;
int AppendProblemsLog(char *s) { return 0;}
char dbFileName[256];

int getDatabaseInfo(char *dbf)
{
  FILE *fp=NULL;
  char str[256];
  int done = 0,cc = 0;
  int database_col;

  if ((fp=fopen("../bin/database.csv","rt")) == NULL && 
      (fp=fopen("../bin/startup.csv","rt")) == NULL)
  {
    strcpy(dbf,"datam.csv");
    return 0;
  }
  while (fgets(str,255,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dbf,colptr(database_col,str),170);
    break;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
char *usageText = "usage: whichrec <recno>";
int main(int argc, char *argv[])
{
  int i, n, base=10, recno, cc = 0;
  char *c, scratch[256], fn[256], cwd[256];
  struct stat sbuf;
  static char adbp[256];

  if (argc < 2) 
  {
    printf("%s\n",usageText);
    exit(0);
  }
  strncpy(scratch,argv[1],64);
  strupr(scratch);
  if (!strncmp(scratch,"0X",2) || !strncmp(scratch,"X",1))
  {
    base = 16;
    c = strchr(scratch,'X');
    if (c == NULL) { printf("this is mega-weird !!\n"); exit(1); }
    *c++ = 0;
  }
  else if ((c=strrchr(scratch,'.')) != NULL)
  {
    base = 16;
    *c++ = 0;
  }
  else
  {
    if (strpbrk(scratch,"ABCDEF") != NULL) base = 16;
    c = scratch;
  }
  recno = strtol(c,NULL,base);
  gAllocDataBuffers = FALSE;
  if ((cc=getDatabaseInfo(dbFileName)) != 0) ccerr(cc);
  if ((c=getenv("PWD")) != NULL) strncpy(cwd,c,256); // want to gloss over symbolic links !
  //getcwd(cwd,256);
  if ((c=strrchr(cwd,FS_DELIMITER)) == NULL) ccerr(file_error);
  *c = 0;
  sprintf(fn,"%.128s/database/%.32s",cwd,dbFileName);
  if (stat(fn, &sbuf) != 0) ccerr(no_such_file);
  totalNrRecords = sbuf.st_size /50; // overestimate the # recs.
  recordsCapacity = totalNrRecords;
  sprintf(adbp,"%.128s/database/",cwd);
  ArchiveDBpath = adbp;
  if ((daqTbl=(DAQRecord *)calloc(totalNrRecords,sizeof(DAQRecord))) == NULL) ccerr(out_of_server_memory);
  if ((prpAlias=(PRP_ALIAS *)calloc(recordsCapacity,sizeof(PRP_ALIAS))) == NULL) ccerr(out_of_server_memory);
  if ((subSystems=(SUB_SYS *)calloc(recordsCapacity,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);
  n = loadDaqDb(dbFileName,0);
  loadDaqParamDb(n,0);
  i = getEntryFromRecno(recno,totalNrRecords);
  if (i < 0) 
  {
    printf("no record number %d found in database!\n",recno);
    ccerr(non_existent);
  }
  printf("record number: %d\n",daqTbl[i].rec);
  printf("source: /%.32s/%.32s/%.64s[%.64s]\n",daqTbl[i].ctx,daqTbl[i].srv,daqTbl[i].dev,daqTbl[i].prp);
  printf("keyword: %.64s\n",daqTbl[i].daqParam[0].param);

err:
  if (cc != 0)
  {
    printf("ERROR: %.32s\n",cc2str(cc));
  }
  return 0;
}
int getEntryFromRecno(int rec,int nrecs)
{
  int i;

  for (i=0; i<nrecs; i++)
    if (daqTbl[i].rec == rec) return i;
  return -1;
}
