/*
 * alaeqm.c    REVISION HISTORY:
 * Generated by SRVWIZARD on Fri Mar 11 09:12:47 2005
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "avaeqm.h"
#if defined(UNIX)
#include "dirent.h"
# define DBPATH "../database/"
#elif defined(WIN32)
#include "direct.h"
# define DBPATH "..\\database\\"
#endif

char thisFecName[FEC_NAME_SIZE + 4] = "";
char thisFecServer[32] = "Availability";
char thisFecContext[CONTEXT_NAME_SIZE] = "TEST";
char thisFecSubsystem[16] = "SER";
char thisFecDescription[FEC_DESC_SIZE] = "availability server";
char thisFecLocation[FEC_LOCATION_SIZE] = "";
char thisFecResponsible[32] = "fecadmin";
int thisFecPort = 0;
int avaeqm(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access);
long locateRecordFromLTS(char *fileName, int nidx, HstTblEntry *hst, double dt);
int writeCorrectionsFiles(time_t from, time_t to);

// from staeqm.c:
#define MAX_NUM_CONFIGS 500
#define MAX_STATES 256
#define MAX_PROCS 100
#define MAX_MACHINE_STATES 50

typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char type[16];
  char states[512];
  char color[16];
  int  metric;
  int  flags;
  int  reserved1;
  int  reserved2;
} OprProfileEx;

#define PRF_ALWAYS_VISIBLE 0x01
#define PRF_NON_OPERATIONAL 0x02
#define PRF_EXCLUDE_STATS 0x04
#define PRF_PROBLEMS 0x08
#define PRF_NOT_AVAILABLE 0x10
#define PRF_WARMUP 0x20

#define isAlwayVisible(f) (((f) & PRF_ALWAYS_VISIBLE) == PRF_ALWAYS_VISIBLE ? TRUE : FALSE)
#define isNonOperational(f) (((f) & PRF_NON_OPERATIONAL) == PRF_NON_OPERATIONAL ? TRUE : FALSE)
#define isExcludedInStatistics(f) (((f) & PRF_EXCLUDE_STATS) == PRF_EXCLUDE_STATS ? TRUE : FALSE)
#define isProblemStatistics(f) (((f) & PRF_PROBLEMS) == PRF_PROBLEMS ? TRUE : FALSE)
#define isNotAvailableStatistics(f) (((f) & PRF_NOT_AVAILABLE) == PRF_NOT_AVAILABLE ? TRUE : FALSE)
#define isWarmup(f) (((f) & PRF_WARMUP) == PRF_WARMUP ? TRUE : FALSE)

enum ProfileType
{
  PT_SLICE, PT_OPTIONAL, PT_ERROR, PT_BEAM,
};
int PrfTypeOf(char *type)
{
  if (!stricmp(type, "SLICE")) return PT_SLICE;
  if (!stricmp(type, "OPTIONAL")) return PT_OPTIONAL;
  if (!stricmp(type, "ERROR")) return PT_ERROR;
  if (!stricmp(type, "BEAM")) return PT_BEAM;
  return PT_OPTIONAL;
}
enum ProfileBucket {
  PB_TOTAL, PB_BEAM, PB_ERROR,
};

typedef struct ProfileBeanCounter
{
  char name[64];
  long sum;
  long correction;
  time_t start;
  time_t end;
} ProfileBeanCounter;

int gPrfCntsCurrent[100];
int gCorrCntsCurrent[100];
NAME64 procedureNames[100];
int numProcedures;
NAME64 gCurrentProcedure;

enum AvailabilityEpoch {
  AV_YEAR = 0, // 1st of January
  AV_MONTH, // 1st of current month
  AV_WEEK,  // this past Sunday
  AV_SHIFT, // 07:00, 15:00, or 23:00
  AV_USER,  // defined by WRITE call
};
#define NEPOCHS 5
NAME64 epochNames[5] = {
  "Beginning of YEAR",
  "Beginning of MONTH",
  "Beginning of WEEK",
  "Beginning of SHIFT",
  "User defined"
};
time_t tepoch[NEPOCHS] = {0};
time_t tsrvstart = 0;
NAME64 userEpochTag = {"User defined"};
int getEpochId(char *str)
{
  int eid = AV_YEAR;
  if (str != NULL)
  {
    if (!stricmp(str, "MONTH")) eid = AV_MONTH;
    else if (!stricmp(str, "WEEK")) eid = AV_WEEK;
    else if (!stricmp(str, "SHIFT")) eid = AV_SHIFT;
  }
  return eid;
}
time_t gEarliestCorrectionsTimestamp = 0;
time_t gCorrectionsAppliedTimestamp = 0;
int gLastCorrectionsTimestamp = 0;
int nDaysInMonth[2][12] = {
  {31,28,31,30,31,30,31,31,30,31,30,31},
  {31,29,31,30,31,30,31,31,30,31,30,31},
};
#define numDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonth[0][mm] : nDaysInMonth[1][mm])

int idCnts = -1;
int idLastCorr = -1;
int idDeclProc = -1;
void updProfileCounter(ProfileConfig *prf, char *proc)
{
  int i, k;
  for (i = 0; i < prf->nprofiles; i++)
  { // run thru all the profiles in a config
    for (k = 0; k < prf->profiles[i].nstates; k++)
    { // all the procedures in a profile
      if (!strncmp(prf->profiles[i].states[k].name, proc, 64))
      { // found it!
        strncpy(prf->currProfile, prf->profiles[i].name, 64);
        if (prf->profiles[i].exclude)
        {
          prf->currProfileCounter = 0;
        }
        else
        {
          prf->currProfileCounter =
            prf->profiles[i].subtract ? -1 : 1;
        }
        return; // done !
      }
    }
  }
}
void cbCounts(int id, int cc)
{
  if (cc != 0) return;

  if (id == idDeclProc)
  {
    int p;
    for (p = 0; p < numConfigs; p++)
    { // run thru all profile configs
      updProfileCounter(&profileConfigs[p], gCurrentProcedure.name);
    }
  }
}
int numtoks(const char *str, char tok)
{
  int count = 0;
  if (str == NULL || strlen(str) == 0) return -1;
  while (*str != '\0') count += *str++ == tok;
  return count;
}
int getProfileStates(char *parseableStateList, ProfileState **list)
{
  char stateList[512], *tok = ",", *c = NULL;
  int i, n = 0;
  ProfileState *states = NULL;
  if (parseableStateList == NULL || list == NULL) return 0;
  strncpy(stateList, parseableStateList, 512);
  n = numtoks(stateList, ',') + 1;
  if ((states=(ProfileState *)calloc(n, sizeof(ProfileState))) == NULL) return 0;
  for (c=strtok(stateList, tok), i=0; c != NULL && i<n; c=strtok(NULL, tok))
  {
    if (strlen(c) == 0) continue;
    strncpy(states[i++].name, c, 64);
  }
  *list = states;
  return n;
}
#define USR_FILENAME "userepoch.csv"
time_t getSavedUserEpoch(void)
{
  FILE *fp = NULL;
  char str[1024], scratch[64];
  int n = 0, done = 0, cc = 0;
  int tim_col = -1;

  if ((fp=fopen(USR_FILENAME, "rt")) == NULL) ccerr(no_such_file);
  while (fgets(str, 1023, fp) != NULL)
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n", str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((tim_col = findcol(str, "BEGIN")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    getColumnStr(tim_col, str, scratch, 64);
    tepoch[AV_USER] = (time_t)parseTineTimeString(scratch);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0 && cc != no_such_file)
  {
    feclog("error reading user epoch file %s : %.32s", USR_FILENAME, cc2str(cc));
  }
  return cc;
}
int saveUserEpoch(void)
{
  FILE *fp = NULL;
  int cc = 0;
  time_t et = tepoch[AV_USER];
  struct tm *etm = localtime(&et);
  if ((fp=fopen(USR_FILENAME, "wt")) == NULL) ccerr(no_such_file);
  fprintf(fp, "BEGIN\n");
  fprintf(fp, "%02d.%02d.%d %02d:%02d:%02d",
    etm->tm_mday, etm->tm_mon + 1, etm->tm_year + 1900,
    etm->tm_hour, etm->tm_min, etm->tm_sec);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
ProfileCounts *gCurrentProfileCounts = NULL;
ProfileCounts *getCurrentPrfCntsItem(char *name)
{
  int i;
  if (gCurrentProfileCounts == NULL)
  { // not yet allocated ...
    gCurrentProfileCounts = (ProfileCounts *)calloc(numProcedures, sizeof(ProfileCounts));
    if (gCurrentProfileCounts == NULL)
    {
      msglog(out_of_local_memory, "getCurrentPrfCntsItem: out of memory");
      return NULL;
    }
    for (i=0; i<numProcedures; i++)
    { 
      strncpy(gCurrentProfileCounts[i].name, procedureNames[i].name, DEVICE_NAME_SIZE);
    }
  }
  for (i=0; i<numProcedures; i++)
  { // find 'im
    if (!strncmp(gCurrentProfileCounts[i].name, name, DEVICE_NAME_SIZE))
    {
      gCurrentProfileCounts[i].counts = gPrfCntsCurrent[i];
      return &gCurrentProfileCounts[i];
    }
  }
  return NULL;
}
long getCurrentCounterDifference(ProfileState state, time_t start)
{
  ProfileCounts *prfCntsStart = getPrfCntsLstItem(state.name, start);
  ProfileCounts *prfCntsCurr = getCurrentPrfCntsItem(state.name);
  if (prfCntsCurr == NULL || prfCntsStart == NULL) return 0;
  return prfCntsCurr->counts - prfCntsStart->counts;
}
long getCounterDifference(ProfileState state,time_t start, time_t end)
{
  ProfileCounts *prfCntsStart = getPrfCntsLstItem(state.name, start);
  ProfileCounts *prfCntsEnd = getPrfCntsLstItem(state.name, end);
  if (prfCntsEnd == NULL || prfCntsStart == NULL) return 0;
  return prfCntsEnd->counts - prfCntsStart->counts;
}
void collectProfileData(ProfileBeanCounter *prfBeanCtr,ProfileConfig *prfConfig, time_t startTime, time_t endTime)
{
  int i, k;
  long cnt;

  for (i = 0; i < prfConfig->nprofiles; i++)
  {
    cnt = 0;
    for (k = 0; k < prfConfig->profiles[i].nstates; k++)
    {
      cnt += getCounterDifference(prfConfig->profiles[i].states[k], startTime, endTime);
    }
    prfBeanCtr[i].sum = cnt;
  }
}
void collectCurrentProfileData(ProfileBeanCounter *prfBeanCtr, ProfileConfig *prfConfig, time_t startTime)
{
  int i, k;
  long cnt;
  for (i = 0; i < prfConfig->nprofiles; i++)
  {
    cnt = 0;
    for (k = 0; k < prfConfig->profiles[i].nstates; k++)
    {
      cnt += getCurrentCounterDifference(prfConfig->profiles[i].states[k], startTime);
    }
    prfBeanCtr[i].sum = cnt;
  }
}
void loadStateCorrectionCountsForProfiles(ProfileBeanCounter *prfBeanCtr, ProfileConfig *prfConfig, time_t startTime, time_t endTime)
{
  int i, k, m, cc;
  static int ccnts[100] = {0}, range[2] = {0, 0};

  if (range[0] != (int)startTime || range[1] != (int)endTime)
  {
    DTYPE dout, din;
    char dev[128];
    char *prp = "CorrectionCounts.Procedures";
    range[0] = (int)startTime; range[1] = (int)endTime;
    DTYPEZERO(dout); DTYPEZERO(din);
    dout.dArrayLength = 100;
    dout.dFormat = CF_INT32;
    dout.data.lptr = ccnts;
    din.dArrayLength = 2;
    din.dFormat = CF_INT32;
    din.data.lptr = range;
    sprintf(dev, "/%.32s/STATE/#0", thisFecContext);
    cc = ExecLinkEx(dev, prp, &dout, &din, CA_READ | CA_MUTABLE, 800);
    if (cc != 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot get correction counts : %.32s", cc2str(cc));
      return;
    }
  }
  for (i=0; i< prfConfig->nprofiles; i++)
  { // a profile (pie slice) is made up of procedures ...
    prfBeanCtr[i].correction = 0;
    for (k = 0; k < numProcedures; k++)
    { // a procedure is made up of states ...
      if (ccnts[k] == 0) continue;
      for (m = 0; m < prfConfig->profiles[i].nstates; m++)
      {
        if (!strnicmp(prfConfig->profiles[i].states[m].name, procedureNames[k].name, DEVICE_NAME_SIZE))
        {
          prfBeanCtr[i].correction += ccnts[k];
          break;
        }
      }
    }
    prfBeanCtr[i].start = startTime;
    prfBeanCtr[i].end = endTime;
  }
  return;
}
typedef struct StateCorrection
{
  double start;
  double stop;
  int counts;
  int reserved;
  char startStr[64];
  char stopStr[64];
  char fromState[64];
  char toState[64];
  char usr[16];
  char cmt[256];
} StateCorrection;
#define MAX_CORRECTION_PER_DAY 2048
StateCorrection cprocs[1000];
int getProcedureCorrections(time_t startTime, time_t endTime, NAME64DBL *procCorrTbl, int *nitems)
{
  int i, k, cc, n=0;
  static int range[4] = {0, 0, 0, 1};
  static int last_nitems = 0;

  if (range[0] != (int)startTime || range[1] != (int)endTime)
  {
    DTYPE dout, din;
    int len = *nitems;
    char dev[128];
    char *prp = "Corrections";
    DTYPEZERO(dout); DTYPEZERO(din);
    range[0] = (int)startTime; range[1] = (int)endTime;
    dout.dArrayLength = 1000;
    dout.dFormat = CF_STRUCT;
    dout.data.vptr = cprocs;
    strcpy(dout.dTag, "StateCor");
    din.dArrayLength = 2;
    din.dFormat = CF_INT32;
    din.data.lptr = range;
    sprintf(dev, "/%.32s/STATE/#0", thisFecContext);
    cc = ExecLinkEx(dev, prp, &dout, &din, CA_READ | CA_MUTABLE, 800);
    if (cc != 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot get procedure corrections : %.32s", cc2str(cc));
      return cc;
    }
    n = dout.dArrayLength;
    if (n > len / 2) return buffer_too_small;
    last_nitems = 2 * n; // always a 'from' and a 'to'
  }
  // map cprocs to NAME64DBL ...
  for (i=0, k=0; i < n; i++)
  {
    procCorrTbl[k].dval = cprocs[i].start;
    strncpy(procCorrTbl[k].name, cprocs[i].toState, 64);
    k++;
    procCorrTbl[k].dval = cprocs[i].stop;
    strncpy(procCorrTbl[k].name, cprocs[i].toState, 64);
    k++;
  }
  *nitems = last_nitems;
  return 0;
}
int getArchivedProcedures(time_t startTime, time_t endTime, NAME64DBL *procDataTbl, int *nitems)
{
  int cc;
  static int range[4] = {0, 0, 0, 1};
  static int last_nitems = 0;

  if (range[0] != (int)startTime || range[1] != (int)endTime)
  {
    DTYPE dout, din;
    char dev[128];
    char *prp = "DeclaredProcedure";
    int len = *nitems;
    range[0] = (int)startTime; range[1] = (int)endTime;
    DTYPEZERO(dout); DTYPEZERO(din);
    dout.dArrayLength = 1000;
    dout.dFormat = CF_NAME64DBL;
    dout.data.vptr = procDataTbl;
    din.dArrayLength = 4;
    din.dFormat = CF_INT32;
    din.data.lptr = range;
    sprintf(dev, "/%.32s/HISTORY/#0", thisFecContext);
    cc = ExecLinkEx(dev, prp, &dout, &din, CA_READ | CA_MUTABLE, 800);
    if (cc != 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot get stored procedures : %.32s", cc2str(cc));
      return cc;
    }
    last_nitems = dout.dArrayLength;
  }
  *nitems = last_nitems;
  return 0;
}
double calculateAvailability(ProfileConfig *prfConfig, int applyCorrections, time_t startTime, time_t endTime)
{
  ProfileBeanCounter *prfBeanCntr = NULL;
  int cc = 0, i, n = 0;
  double availableTime = 0;
  double availableFullTime = 0;
  double value;
  if (prfConfig == NULL) return 0;
  n = prfConfig->nprofiles;
  if ((prfBeanCntr=(ProfileBeanCounter *)calloc(n, sizeof(ProfileBeanCounter))) == NULL)
    ccerr(out_of_local_memory);
  for (i=0; i<n; i++) 
    strncpy(prfBeanCntr[i].name, prfConfig->profiles[i].name, DEVICE_NAME_SIZE);
  collectProfileData(prfBeanCntr, prfConfig, startTime, endTime);
  loadStateCorrectionCountsForProfiles(prfBeanCntr, prfConfig, startTime, endTime);
  for (i = 0; i < n; i++)
  {
    value = (double)prfBeanCntr[i].sum;
    if (applyCorrections) value += prfBeanCntr[i].correction;
    if (!prfConfig->profiles[i].exclude)
    {
      availableFullTime += value;
      if (!prfConfig->profiles[i].subtract) availableTime += value;
    }
  }
err:
  if (prfBeanCntr) free(prfBeanCntr);
  if (availableFullTime == 0) return 100.0;
  return (availableTime / availableFullTime) * 100.;
}
ProfileBeanCounter currBeanCntr[100] = {0};
ProfileBeanCounter corrBeanCntrs[NEPOCHS][MAX_PROFILES][100];
double currentAvailability(ProfileConfig *prfConfig, int applyCorrections, int epoch)
{
  static int startup = TRUE;
  time_t startTime = tepoch[epoch];
  ProfileBeanCounter *prfBeanCntr = currBeanCntr;
  int cc = 0, i, p, n = 0;
  double availableTime = 0;
  double availableFullTime = 0;
  double value;
  if (prfConfig == NULL) return 0;
  if (startTime == 0) return 0;
  n = prfConfig->nprofiles;
//  for (i = 0; i < n; i++)
//    strncpy(prfBeanCntr[i].name, prfConfig->profiles[i].name, DEVICE_NAME_SIZE);
  collectCurrentProfileData(prfBeanCntr, prfConfig, startTime);
  for (p = 0; p < numConfigs && strncmp(profileConfigNames[p].name, prfConfig->name, 64); p++);
  for (i = 0; i < n; i++)
  {
    value = (double)prfBeanCntr[i].sum;
    if (applyCorrections) value += corrBeanCntrs[epoch][p][i].correction;
    if (!prfConfig->profiles[i].exclude)
    {
      availableFullTime += value;
      if (!prfConfig->profiles[i].subtract) availableTime += value;
    }
  }
  if (availableFullTime == 0) return 100.0;
  return (availableTime / availableFullTime) * 100.;
}
OprProfileEx profs[200];
int getOperationHistoryProfiles(char *context, char *configName, Profile **profilesList, int *nprofiles)
{
  DTYPE dout;
  char dev[128]; 
  char *prp = "PROFILE";
  char *tag = "PrfCfgEx";
  Profile *profiles = NULL;
  int cc, i, n = 0;
  if (context == NULL || configName == NULL) return argument_list_error;
  if (profilesList == NULL || nprofiles == NULL) return argument_list_error;
  DTYPEZERO(dout);
  dout.dArrayLength = 200;
  dout.dFormat = CF_STRUCT;
  strncpy(dout.dTag, tag, TAG_NAME_SIZE);
  dout.data.vptr = profs;
  sprintf(dev, "/%.32s/STATE/%.64s", context, configName);
  cc = ExecLinkEx(dev, prp, &dout, NULL, CA_READ|CA_MUTABLE, 800);
  if (cc != 0)
  { 
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot get profiles for config %.64s : %.32s", configName,cc2str(cc));
    return cc;
  }
  n = dout.dArrayLength;
  profiles = (Profile *)calloc(n, sizeof(Profile));
  for (i=0; i < n; i++)
  {
    strncpy(profiles[i].name, profs[i].name, 64);
    profiles[i].type = PrfTypeOf(profs[i].type);
    profiles[i].exclude = isExcludedInStatistics(profs[i].flags);
    profiles[i].subtract = isNotAvailableStatistics(profs[i].flags);
    profiles[i].nstates = getProfileStates(profs[i].states, &profiles[i].states);
  }
  *profilesList = profiles;
  *nprofiles = n;
  return cc;
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(thisFecServer, AVAEQM_TAG, AVAEQM_MAXDEVICES, avaeqm, avaeqm_ini, avaeqm_bkg, AVAEQM_BKG_RATE, avaeqm_exi);
}
void PreSystemInit(void)
{
  SetCatchConsoleBreak(TRUE);
  SetAllowOverloadedConcurrentLocalHistories(TRUE);
  if (strlen(thisFecContext) > 0)
  {
    int p;
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName, thisFecContext, 10); thisFecName[10] = 0;
      strcat(thisFecName, "AVAFEC");
    }
    if (thisFecPort == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisFecPort = p;
    }
    sprintf(thisFecDescription, "%.32s Alarm State server", thisFecContext);
    RegisterFecInformation(thisFecName, thisFecSubsystem, thisFecContext, thisFecDescription, thisFecLocation, "none", thisFecResponsible, thisFecPort);
    RegisterStructs();
  }
}
#define quit(t,i) { feclogEx(TINE_LOGLEVEL_FATAL,"%s: %.32s",t,cc2str(i)); exit(1); }
void RegisterStructs(void)
{
  static int done = FALSE;
  char *tag = "PrfCfgEx";
  int cc;
  if (done) return;
  done = TRUE;
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, name), DEVICE_NAME_SIZE, CF_TEXT, "name"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, type), 16, CF_TEXT, "type"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, states), 512, CF_TEXT, "states"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, color), 16, CF_TEXT, "color"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, metric), 1, CF_INT32, "metric"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, flags), 1, CF_INT32, "flags"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, reserved1), 1, CF_INT32, "reserved1"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, reserved2), 1, CF_INT32, "reserved2"))) quit(tag,cc);
  /* terminate the structure definition like this! */
  if ((cc=sealTaggedStruct(tag, sizeof(OprProfileEx), MAX_NUM_CONFIGS))) quit(tag,cc);

  tag = "StateCor";
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, start), 1, CF_DOUBLE, "startUtc"))) quit(tag,cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, stop), 1, CF_DOUBLE, "stopUtc"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, counts), 1, CF_INT32, "counts"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, reserved), 1, CF_INT32, "reserved"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, startStr), 64, CF_TEXT, "startStr"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, stopStr), 64, CF_TEXT, "stopStr"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, fromState), 64, CF_TEXT, "fromState"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, toState), 64, CF_TEXT, "toState"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, usr), 16, CF_TEXT, "user"))) quit(tag, cc);
  if ((cc = addFieldToStruct("StateCor", OFFSETIN(StateCorrection, cmt), 256, CF_TEXT, "comment"))) quit(tag, cc);
  if ((cc = sealTaggedStruct("StateCor", sizeof(StateCorrection), MAX_CORRECTION_PER_DAY))) quit(tag, cc);

}
int RegisterProperties(void)
{
  int id, cc = 0;
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = NEPOCHS;
  din.dFormat = CF_INT32;
  din.dArrayLength = 2;
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability",
    &dout, &din, CA_READ, AT_SINGLE, 0, "[0:100 %]Running availability (with corrections) since start of year", PRP_AVAILABILITY, NULL);
  if (id < 0) ccerr(-id);
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability.HIST",
    &dout, &din, CA_READ, AT_SINGLE, 0, "[0:100 %]History (overloaded) of running availability (with corrections) since start of year", PRP_AVAILABILITY_HIST, NULL);
  if (id < 0) ccerr(-id);
  SetCallPropertyInSeparateThread(AVAEQM_TAG, "Availability.HIST", TRUE);
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability.Raw.HIST",
    &dout, &din, CA_READ, AT_SINGLE, 0, "[0:100 %]History (overloaded) of running availability (without corrections) since start of year", PRP_AVAILABILITY_RAW_HIST, NULL);
  if (id < 0) ccerr(-id);
  SetCallPropertyInSeparateThread(AVAEQM_TAG, "Availability.Raw.HIST", TRUE);
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability.Raw",
    &dout, &din, CA_READ, AT_SINGLE, 0, "[0:100 %]Running availability (raw data: no corrections) since start of year", PRP_AVAILABILITY_RAW, NULL);
  if (id < 0) ccerr(-id);
  dout.dFormat = CF_INT32;
  dout.dArrayLength = NEPOCHS;
  din.dFormat = CF_INT32;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(AVAEQM_TAG, "Epoch",
    &dout, &din, CA_READ|CA_WRITE, AT_CHANNEL, 0, "Epoch begin as UTC (or string)", PRP_EPOCH, NULL);
  if (id < 0) ccerr(-id);
  AssignDeviceListToProperty(AVAEQM_TAG, "Epoch", epochNames, NEPOCHS);

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = 1;
  din.dFormat = CF_NAME64;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(AVAEQM_TAG, "Epoch.UserTag",
    &dout, &din, CA_READ|CA_WRITE|CA_SAVERESTORE, AT_SCALAR, 0, "User defined epoch decriptive tag", PRP_EPOCH_USERDEFINED_TAG, NULL);
  if (id < 0) ccerr(-id);
  AssignDeviceListToProperty(AVAEQM_TAG, "Epoch.UserTag", &epochNames[AV_USER], 1);

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = numConfigs;
  id = RegisterPropertyInformation(AVAEQM_TAG, "CurrentProfile",
    &dout, NULL, CA_READ, AT_CHANNEL, 0, "Current declared state profile (pie slice)", PRP_CURRENT_PROFILE, NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_NAME64;
  dout.dArrayLength = 1;
  id = RegisterPropertyInformation(AVAEQM_TAG, "CurrentProcedure",
    &dout, NULL, CA_READ, AT_SCALAR, 0, "Current declared state procedure", PRP_CURRENT_PROCEDURE, NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = numConfigs;
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability.CounterStep",
    &dout, NULL, CA_READ, AT_CHANNEL, 0, "[-1:1 steps] current availability counter increment", PRP_AVAILABILITY_COUNTER, NULL);
  if (id < 0) ccerr(-id);
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability.CounterStep.HIST",
    &dout, &din, CA_READ, AT_SINGLE, 0, "[0:100 %]History (overloaded) of counter increment (with corrections) since start of year", PRP_AVAILABILITY_COUNTER_HIST, NULL);
  if (id < 0) ccerr(-id);
  SetCallPropertyInSeparateThread(AVAEQM_TAG, "Availability.CounterStep.HIST", TRUE);
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability.CounterStep.Raw.HIST",
    &dout, &din, CA_READ, AT_SINGLE, 0, "[0:100 %]History (overloaded) of counter increment (with corrections) since start of year", PRP_AVAILABILITY_COUNTER_RAW_HIST, NULL);
  if (id < 0) ccerr(-id);
  SetCallPropertyInSeparateThread(AVAEQM_TAG, "Availability.CounterStep.Raw.HIST", TRUE);

err:
  if (cc != 0)
  {
    dbglog("Can't register property : %32s", cc2str(cc));
  }
  return cc;
}
NAME64 profileConfigNames[MAX_PROFILES];
NAME64 procedureNames[MAX_PROCEDURES];
int numConfigs = MAX_PROFILES;
int numProcedures = MAX_PROCEDURES;
ProfileConfig *profileConfigs = NULL;
int RegisterDevices(void)
{
  int i, cc = 0;
  char srv[256];
  sprintf(srv, "/%.32s/STATE/#0",thisFecContext);
  cc = GetSystemDevices(srv, "CONFIGURATION", profileConfigNames, &numConfigs);
  if (cc == 0)
  {
    if (numConfigs > MAX_PROFILES) numConfigs = MAX_PROFILES;
    for (i = 0; i < numConfigs; i++)
      RegisterDeviceName(AVAEQM_TAG, profileConfigNames[i].name, i);
  }
  return cc;
}
void removePrfCntsItems(time_t timestamp)
{
  int p, i, k;
  if (profileConfigs == NULL) return;
  if (timestamp == 0) return;
  for (p = 0; p < numConfigs; p++)
  { // all the profile configs ...
    for (i = 0; i < profileConfigs[p].nprofiles; i++)
    { // all the profiles (slices) in the config
      for (k = 0; k < profileConfigs[p].profiles[i].nstates; k++)
      { // all the state procedures in the profile
        rmvPrfCntsLstItem(profileConfigs[p].profiles[i].states[k].name, timestamp);
      }
    }
  }
}
/** 
 * return TRUE if an epoch has changed ...
 */
int resetEpochs(void)
{
  int stsChanged = FALSE;
  time_t t = time(NULL);
  time_t et = 0;
  struct tm ltm = {0}, *ltn = localtime(&t);

  ltm.tm_year = ltn->tm_year;
  ltm.tm_mday = 1;
  et = mktime(&ltm);
  if (et != tepoch[AV_YEAR])
  {
    removePrfCntsItems(tepoch[AV_YEAR]);
    stsChanged = TRUE;
  }
  tepoch[AV_YEAR] = et;

  ltm.tm_mon = ltn->tm_mon;
  ltm.tm_hour = 0;
  et = mktime(&ltm);
  if (ltm.tm_hour > 0) et -= 3600 * ltm.tm_hour; // weird DST behavior
  if (et != tepoch[AV_MONTH]) 
  {
    removePrfCntsItems(tepoch[AV_MONTH]);
    stsChanged = TRUE;
  }
  tepoch[AV_MONTH] = et;

  ltm.tm_hour = 0;
  ltm.tm_mday = ltn->tm_mday - (ltn->tm_wday-1); // in the BKR the week starts on Monday
  if (ltm.tm_mday < 0)
  {
    if (ltm.tm_mon > 0)
    {
      ltm.tm_mon--;
      ltm.tm_mday = numDaysInMonth(ltm.tm_year, ltm.tm_mon) + ltm.tm_mday;
    }
    else
    { // if beginning of January then start the week on Jan. 1st
      ltm.tm_mday = 1;
    }
  }
  et = mktime(&ltm);
  if (et != tepoch[AV_WEEK]) 
  {
    removePrfCntsItems(tepoch[AV_WEEK]);
    stsChanged = TRUE;
  }
  tepoch[AV_WEEK] = et;

  ltm.tm_mon = ltn->tm_mon;
  ltm.tm_mday = ltn->tm_mday;
  ltm.tm_hour = 8 * ((ltn->tm_hour + 1) / 8) - 1; // 7:00, 15:00, 23:00
  if (ltm.tm_hour < 0)
  { // = -1 if HH < 07
    ltm.tm_mday--;
    ltm.tm_hour = 23;
    if (ltm.tm_mday < 0)
    {
      if (ltm.tm_mon > 0)
      { // and we happened to have changed months during the late shift
        ltm.tm_mon--;
        ltm.tm_mday = numDaysInMonth(ltm.tm_year, ltm.tm_mon) + ltm.tm_mday;
      }
      else
      { // if beginning of January then start the week on Jan. 1st
        ltm.tm_mday = 1;
      }
    }
  }
  et = mktime(&ltm);
  if (et != tepoch[AV_SHIFT]) 
  {
    removePrfCntsItems(tepoch[AV_SHIFT]);
    stsChanged = TRUE;
  }
  tepoch[AV_SHIFT] = et;

  return stsChanged;
}
int cmdDumpCorrections(int depth,int arg1, int arg2, int arg3)
{
  time_t t1 = time(NULL);
  time_t t0 = t1 - depth;
  UNUSED(arg1); 
  UNUSED(arg2);
  UNUSED(arg3);
  return writeCorrectionsFiles(t0, t1);
}
#define NPROCS_BUFLEN 1000
NAME64DBL storedProcs[NPROCS_BUFLEN];
int nStoredProcs = NPROCS_BUFLEN;
NAME64DBL storedCorrs[NPROCS_BUFLEN];
int nStoredCorrs = NPROCS_BUFLEN;
#define NSLICES 500
int writeCorrectionsFiles(time_t from, time_t to)
{
  int i = 0, k, cc = 0, inc = (int)(to - from)/NSLICES;
  double ut;
  double av = 0.0, ac = 0.0;
  time_t t;
  FILE *fp = NULL;
  char fn[256], *proc;
  int itmp[2] = {0, 0};
  struct tm *stm = localtime(&gSystemStartupTime);
  ProfileConfig storedProfile;

  sprintf(fn, "%.128sAvailability-%02d-%02d-%02d-%02d-%02d-%02d",
    arcDbPath,
    stm->tm_year%100,stm->tm_mon+1,stm->tm_mday,
    stm->tm_hour,stm->tm_min,stm->tm_sec);
  if ((fp=fopen(fn, "w")) == NULL) ccerr(file_error);
  inc = MAX(300, (int)(to - from) / NSLICES);
  for (t=from; t < to; t += inc)
  {
    av = calculateAvailability(&profileConfigs[0], TRUE, tepoch[AV_YEAR], t);
    ut = (double)t;
    if (fwrite(&ut,sizeof(double),1,fp) < 0) ccerr(no_file_space);
    if (fwrite(itmp, sizeof(int), 2, fp) < 0) ccerr(no_file_space);
    if (fwrite(&av, sizeof(double), 1, fp) < 0) ccerr(no_file_space);
  }
  if (fp != NULL) fclose(fp);
  fp = NULL;
  sprintf(fn, "%.128sCounter.Step-%02d-%02d-%02d-%02d-%02d-%02d",
    arcDbPath,
    stm->tm_year % 100, stm->tm_mon + 1, stm->tm_mday,
    stm->tm_hour, stm->tm_min, stm->tm_sec);
  if ((fp = fopen(fn, "w")) == NULL) ccerr(file_error);
  storedProfile = profileConfigs[0];
  while (from < to)
  {
    nStoredProcs = NPROCS_BUFLEN;
    cc = getArchivedProcedures(from, to, storedProcs, &nStoredProcs);
    from = (nStoredProcs < NPROCS_BUFLEN) ? to : (time_t)storedProcs[NPROCS_BUFLEN - 1].dval + 1;
    if (cc != 0) ccerr(cc);
    nStoredCorrs = NPROCS_BUFLEN;
    cc = getProcedureCorrections(from, to, storedCorrs, &nStoredCorrs);
    if (cc != 0) ccerr(cc);
    k = i = 0;
    while (k < nStoredCorrs - 1)
    { // nStoredCorrs should be an even number ...
      if (i < nStoredProcs && storedProcs[i].dval < storedCorrs[k].dval)
      { // a corrections 'gap'
        proc = storedProcs[i].name;
        ut = storedProcs[i].dval;
        i++;
      }
      else
      { // a correction start time ...
        proc = storedCorrs[k].name;
        ut = storedCorrs[k].dval;
        k++;
      }
      updProfileCounter(&storedProfile, proc);
      ac = (double)storedProfile.currProfileCounter;
      if (fwrite(&ut, sizeof(double), 1, fp) < 0) ccerr(no_file_space);
      if (fwrite(itmp, sizeof(int), 2, fp) < 0) ccerr(no_file_space);
      if (fwrite(&ac, sizeof(double), 1, fp) < 0) ccerr(no_file_space);
      if ((k % 2) == 1)
      { // the correction stop time ...
        proc = storedCorrs[k].name;
        ut = storedCorrs[k].dval;
        updProfileCounter(&storedProfile, proc);
        ac = (double)storedProfile.currProfileCounter;
        if (fwrite(&ut, sizeof(double), 1, fp) < 0) ccerr(no_file_space);
        if (fwrite(itmp, sizeof(int), 2, fp) < 0) ccerr(no_file_space);
        if (fwrite(&ac, sizeof(double), 1, fp) < 0) ccerr(no_file_space);
        while (i < nStoredProcs && storedProcs[i].dval < storedCorrs[k].dval) i++;
        k++;
      }
    }
  }
err:
  if (fp != NULL) fclose(fp);
  return 0;
}
int GetCorrectedHistoryData(char *prp, char *dev, DTYPE *dout, DTYPE *din)
{
  int cc = 0, i, k, idx = 0, eidx = 0, n, yy, mm, rng[2];
  long pstart, pstop;
  int itmp[2];
  double timestamp = 0.0, dval = 0.0;
  HstTblEntry hst = {0};
  NAME64 flst[200];
  FILE *fp = NULL;
  DTYPE fdt;
  DBLDBL *hstdata = (DBLDBL *)dout->data.vptr;
  char pattern[64], fn[256], *tgt;
  time_t stop = time(NULL);
  time_t start = stop - 86400;
  struct tm *ltm = localtime(&stop);
  if (dout->dFormat != CF_DBLDBL && dout->dFormat != CF_DBLTIME)
  { // make life easy for us ...
    return illegal_format;
  }
  yy = ltm->tm_year % 100;
  mm = ltm->tm_mon;
  if (ltm->tm_mon == 0)
  { // last year, last month if we're at January
    mm = 11;
    yy--;
  }
  tgt = stristr(prp, ".CounterStep") == NULL ? "Availability" : "Counter.Step";
  sprintf(pattern, "%s-%02d-%02d*", tgt, yy, mm);
  n = getFilesOfType(arcDbPath,pattern,flst,100);
  yy = ltm->tm_year % 100;
  mm = ltm->tm_mon; // this month 
  sprintf(pattern, "%s-%02d-%02d*", tgt, yy, mm+1);
  n += getFilesOfType(arcDbPath,pattern, &flst[n], 200-n);
  DTYPEZERO(fdt);
  fdt.dArrayLength = n;
  fdt.data.vptr = flst;
  fdt.dFormat = CF_NAME64;
  sortDataValues(&fdt);
  if (din->dArrayLength > 0)
  {
    if ((cc=GetValuesAsLong(din, rng, 2)) != 0) return cc;
    start = (time_t)rng[0];
    if (din->dArrayLength > 1) stop = (time_t)rng[1];
  }
  hst.recordLength = sizeof(double);
  for (i = 0; i < n; i++)
  {
    sprintf(fn, "%.128s%.64s",arcDbPath,flst[i].name);
    if ((pstart=locateRecordFromLTS(fn, 0, &hst, (double)start)) < 0) continue;
    if ((pstop=locateRecordFromLTS(fn, 0, &hst, (double)stop)) < 0) continue;
    if ((fp=fopen(fn, "r")) == NULL) continue;
    idx = 0; // start again at the beginning
    fseek(fp, pstart, SEEK_SET);
    while (fread(&timestamp, 8, 1, fp) == 1)
    { // read through file 
      if (idx == 0)
      { // find the right latch point
        while (idx < (int)dout->dArrayLength && hstdata[idx].d2val < (double)timestamp) idx++;
      }
      if (timestamp > (double)stop) break;
      if (fread(itmp, 4, 2, fp) != 2) break;
      if (fread(&dval, 8, 1, fp) != 1) break;
      if (timestamp > hstdata[idx].d2val)
      {
        hstdata[idx].d2val = timestamp;
        hstdata[idx].d1val = dval;
        idx++;
      }
      if (idx >= (int)dout->dArrayLength - 1) break;
    }
    eidx = idx;
    while (idx < (int)dout->dArrayLength && hstdata[idx].d2val < (double)timestamp) idx++;
    for (k=0; k < (int)dout->dArrayLength-idx; k++)
    { // shore up the end
      hstdata[eidx+k] = hstdata[idx+k];
    }
    dout->dArrayLength -= (idx-eidx);
    if (fp != NULL) fclose(fp);
  }
  return cc;
}
HistorySpecification availHSpec =
{
  HRS_DEFAULT_POLL_RATE,HRS_DEFAULT_ARCHIVE_RATE,HRS_DEFAULT_DEPTHSHORT,HRS_DEFAULT_DEPTHLONG,
  HRS_DEFAULT_HEARTBEAT,0,0,NULL,NULL
};
void avaeqm_ini(void)
{
  int i, cc = 0;
  char dev[256];
  char *prp = "TIMECOUNTER.Procedure";
  DTYPE dout;
  tsrvstart = time(NULL);
  resetEpochs();
  getSavedUserEpoch();
  if ((cc=RegisterProperties()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"Could not register properties (%.32s) : Abort!",cc2str(cc));
    exit(1);
  }
  if ((cc=RegisterDevices()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "Could not register devices (%.32s) : Abort!", cc2str(cc));
    exit(1);
  }
  profileConfigs = (ProfileConfig *)calloc(numConfigs, sizeof(ProfileConfig));
  for (i = 0; i < numConfigs; i++)
  {
    getOperationHistoryProfiles(thisFecContext, profileConfigNames[i].name, &profileConfigs[i].profiles, &profileConfigs[i].nprofiles);
    strncpy(profileConfigs[i].name, profileConfigNames[i].name, DEVICE_NAME_SIZE);
  }
  sprintf(dev, "/%.32s/STATE/#0", thisFecContext);
  if ((cc=GetSystemDevices(dev, prp, procedureNames, &numProcedures)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot get procedure names : %.32s", cc2str(cc));
  }
  DTYPEZERO(dout);
  dout.dArrayLength = numProcedures;
  dout.dFormat = CF_INT32;
  dout.data.lptr = gPrfCntsCurrent;
  idCnts = AttachLink(dev,prp,&dout,NULL,CA_READ,1000,cbCounts,CM_TIMER);

  prp = "LastCorrectionTimestamp";
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &gLastCorrectionsTimestamp;
  idLastCorr = AttachLink(dev, prp, &dout, NULL, CA_READ, 1000, cbCounts, CM_TIMER);

  prp = "DeclaredProcedure";
  dout.dArrayLength = 1;
  dout.dFormat = CF_NAME64;
  dout.data.vptr = gCurrentProcedure.name;
  idDeclProc = AttachLink(dev, prp, &dout, NULL, CA_READ, 1000, cbCounts, CM_TIMER);

  for (i = 0; i < numConfigs; i++)
  {
    AppendHistoryInformationEx(AVAEQM_TAG, "Availability", profileConfigNames[i].name, NEPOCHS, CF_FLOAT, i, &availHSpec, 0);
    AppendHistoryInformationEx(AVAEQM_TAG, "Availability.CounterStep", profileConfigNames[i].name, numConfigs, CF_INT32, 100+i, &availHSpec, 0);
  }

  RegisterUserCommand("dumpCorrections", cmdDumpCorrections, NULL, NULL, CA_READ);

}
void avaeqm_exi(void)
{
  //writeDeclaredCounters();
}
int epochs_changed = FALSE;
int corrections_applied = FALSE;
void avaeqm_bkg(void)
{
  static time_t tlast_corrections = 0;
  time_t tnow = time(NULL);
  int i, eid, corrections_changed = FALSE, initial_application = TRUE;

  if ((tnow % 3600) == 0)
  { //on the hour: check for change in shift, week, month, etc ...
    epochs_changed = resetEpochs();
  }
  initial_application = gCorrectionsAppliedTimestamp == 0;
  if (gLastCorrectionsTimestamp != (int)tlast_corrections)
  { // there are new corrections !
    if (initial_application || gLastCorrectionsTimestamp < gEarliestCorrectionsTimestamp)
    { // 1st one or an ealier one ...
      gEarliestCorrectionsTimestamp = (time_t)gLastCorrectionsTimestamp;
    }
    gCorrectionsAppliedTimestamp = tnow;
    corrections_changed = TRUE;
  }
  tlast_corrections = (time_t)gLastCorrectionsTimestamp;
  if (!epochs_changed && !corrections_changed) return; // no reason to continue
  for (eid=0; eid<NEPOCHS; eid++)
    for (i = 0; i < numConfigs; i++)
    {
      if (tepoch[eid] == 0) continue;
      loadStateCorrectionCountsForProfiles(corrBeanCntrs[eid][i], &profileConfigs[i], tepoch[eid], tnow);
    }
  corrections_applied = TRUE; 
  if (corrections_changed && !initial_application)
  { // archived data corrections ...
    writeCorrectionsFiles(gEarliestCorrectionsTimestamp, gCorrectionsAppliedTimestamp);
  }
}
int avaeqm(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access)
{
  int cc, did, pid, eid = AV_YEAR;
  int applyCorrections = TRUE, useCurrentStats = FALSE;
  double av = 0;
  time_t t = time(NULL), tstop = t, tstart = 0;
  struct tm ltm = {0}, *lt = localtime(&t);
  did = GetDeviceNumberEx(AVAEQM_TAG, devName, devProperty);
  if (did < 0) return illegal_equipment_number;

  pid = GetPropertyId(AVAEQM_TAG, devProperty);

  //ntotal = gNumAlarmSystems + gNumAlarmGroups;
  switch (pid)
  {
    case PRP_AVAILABILITY_RAW:
      applyCorrections = FALSE;
    case PRP_AVAILABILITY:
      if (access & CA_WRITE) return illegal_read_write;
      if (!corrections_applied) return not_initialized;
      if (din->dArrayLength > 0)
      {
        if (IsIntegerFormat(din->dFormat))
        {
          int vals[2];
          if (din->dArrayLength == 1)
          {
            int yy = lt->tm_year + 1900;
            cc = GetValuesAsLong(din, vals, 1);
            if (cc != 0) return cc;
            if (vals[0] < 100) vals[0] += 2000;
            if (vals[0] < 2050)
            { // given as the desired year
              yy = vals[0];
              ltm.tm_year = yy - 1900;
              ltm.tm_mday = 1;
              tstart = mktime(&ltm);
              ltm.tm_year--;
              tstop = mktime(&ltm) - 1;
            }
            else
            { // given as a start time in UTC (stop is 'now')
              tstart = (time_t)vals[0];
              useCurrentStats = TRUE;
            }
          }
          else
          { // 2 values as start and stop in UTC 
            cc = GetValuesAsLong(din, vals, 2);
            if (cc != 0) return cc;
            tstart = (time_t)vals[0];
            tstop = (time_t)vals[1];
          }
        }
        else if (IsStringFormat(din->dFormat))
        {
          char epochStr[64];
          if ((cc=GetValuesAsStringEx(din, epochStr, CF_TEXT, 64, 0)) != 0)
            return cc;
          eid = getEpochId(epochStr);
          useCurrentStats = TRUE;
        }
      }
      else
      { // no input: start of the year up to 'now'
        tstart = tepoch[AV_YEAR];
        eid = AV_YEAR;
        useCurrentStats = TRUE;
      }
      if (dout->dArrayLength > 0)
      {
        if (access & CA_HIST) useCurrentStats = TRUE;
        if (useCurrentStats)
        {
          double avals[NEPOCHS];
          int i;
          for (i = 0; i < NEPOCHS; i++)
          {
            avals[i] = currentAvailability(&profileConfigs[did], applyCorrections, i);
          }
          if (dout->dArrayLength > NEPOCHS) dout->dArrayLength = NEPOCHS;
          return PutValuesFromDoubleEx(dout, avals, NEPOCHS, eid);
        }
        else
        {
          dout->dArrayLength = 1;
          av = calculateAvailability(&profileConfigs[did], applyCorrections, tstart, tstop);
          return PutValuesFromDoubleEx(dout, &av, 1, 0);
        }
      }
      return dimension_error;
    case PRP_AVAILABILITY_COUNTER_RAW_HIST:
      devProperty = "Availability.CounterStep";
    case PRP_AVAILABILITY_RAW_HIST:
      if (pid == PRP_AVAILABILITY_RAW_HIST) devProperty = "Availability";
      applyCorrections = FALSE;
    case PRP_AVAILABILITY_COUNTER_HIST:
    case PRP_AVAILABILITY_HIST:
      // get the 'normally stored' local history data
      cc = GetLocalHistoryData(AVAEQM_TAG, devProperty, devName, dout, din);
      if (!applyCorrections) return cc;
      if (cc == 0 && (dout->dFormat == CF_DBLTIME || dout->dFormat == CF_DBLDBL))
      { // apply any corrections ...
        DTYPE d = *din;
        int lvals[2];
        if (din->dArrayLength == 0)
        { // no input given ...
          lvals[0] = (int)((DBLDBL *)dout->data.vptr)[0].d2val;
          lvals[1] = (int)((DBLDBL *)dout->data.vptr)[dout->dArrayLength - 1].d2val;
          d.dArrayLength = 2;
          d.dFormat = CF_INT32;
          d.data.lptr = lvals;
        }
        GetCorrectedHistoryData(devProperty, devName, dout, &d);
      }
      return cc;
    case PRP_EPOCH:
      if (access & CA_WRITE)
      {
        int et = 0;
        if (din->dArrayLength == 0) return illegal_read_write;
        if (did != AV_USER) return not_allowed;
        if (IsNumberFormat(din->dFormat))
        {
          if ((cc=GetValuesAsLong(din, &et, 1)) != 0) return cc;
        }
        else if (IsStringFormat(din->dFormat))
        {
          char etstr[64];
          if ((cc=GetValuesAsStringEx(din, etstr, CF_TEXT, 64, 0)) != 0)
            return cc;
          et = (int)parseTineTimeString(etstr);
        }
        else
        {
          return illegal_format;
        }
        if (et < (int)tepoch[AV_YEAR] || et > (int)t) return out_of_range;
        if (tepoch[AV_USER] != (time_t)et)
        {
          if (tepoch[AV_USER] != 0) removePrfCntsItems(tepoch[AV_USER]);
          epochs_changed = TRUE;
        }
        tepoch[AV_USER] = (time_t)et;
        saveUserEpoch();
      }
      if (dout->dArrayLength > 0)
      {
        int i;
        if (IsStringFormat(dout->dFormat))
        {
          NAME64 estrvals[NEPOCHS] = {0};
          for (i = 0; i < NEPOCHS; i++) strncpy(estrvals[i].name, ctime(&tepoch[i]), 24);
          cc =PutValuesFromString(dout, estrvals[0].name, CF_NAME64, NEPOCHS, did);
        }
        else
        {
          int tvals[NEPOCHS];
          for (i = 0; i < NEPOCHS; i++) tvals[i] = (int)tepoch[i];
          cc = PutValuesFromLongEx(dout, tvals, NEPOCHS, did);
        }
        return cc;
      }
      return 0;
    case PRP_EPOCH_USERDEFINED_TAG:
      if (access & CA_WRITE)
      {
        char ettag[64];
        if (din->dArrayLength == 0) return dimension_error;
        if (!IsStringFormat(din->dFormat)) return illegal_format;
        if ((cc=GetValuesAsStringEx(din, ettag, CF_TEXT, 64, 0)) != 0)
          return cc;
        strncpy(userEpochTag.name, ettag, 64);
      }
      if (dout->dArrayLength > 0)
      {
        if (!IsStringFormat(dout->dFormat)) return illegal_format;
        return PutValuesFromString(dout, userEpochTag.name, CF_TEXT, 64, 0);
      }
      return 0;
    case PRP_CURRENT_PROCEDURE:
      if (dout->dArrayLength > 0)
      {
        if (!IsStringFormat(dout->dFormat)) return illegal_format;
        return PutValuesFromString(dout, gCurrentProcedure.name, CF_TEXT, 64, 0);
      }
      return 0;
    case PRP_CURRENT_PROFILE:
      if (dout->dArrayLength > 0)
      {
        NAME64 nams[100];
        int i;
        for (i = 0; i < numConfigs; i++) strncpy(nams[i].name, profileConfigs[i].currProfile, 64);
        return PutValuesFromString(dout, nams[0].name, CF_NAME64, numConfigs, did);
      }
      return 0;
    case PRP_AVAILABILITY_COUNTER:
      if (dout->dArrayLength > 0)
      {
        int vals[100];
        int i;
        for (i = 0; i < numConfigs; i++) vals[i] = profileConfigs[i].currProfileCounter;
        return PutValuesFromLongEx(dout, vals, numConfigs, did);
      }
      return 0;
    case PRP_SYNCHRONIZE:
      return 0;
  }
  return illegal_property;
}