/*
 * alaeqm.c    REVISION HISTORY:
 * Generated by SRVWIZARD on Fri Mar 11 09:12:47 2005
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "avaeqm.h"
#if defined(UNIX)
#include "dirent.h"
# define DBPATH "../database/"
#elif defined(WIN32)
#include "direct.h"
# define DBPATH "..\\database\\"
#endif

char thisFecName[FEC_NAME_SIZE + 4] = "";
char thisFecServer[32] = "Availability";
char thisFecContext[CONTEXT_NAME_SIZE] = "TEST";
char thisFecSubsystem[16] = "SER";
char thisFecDescription[FEC_DESC_SIZE] = "availability server";
char thisFecLocation[FEC_LOCATION_SIZE] = "";
char thisFecResponsible[32] = "fecadmin";
int thisFecPort = 0;
int avaeqm(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access);

// from staeqm.c:
#define MAX_NUM_CONFIGS 500
#define MAX_STATES 256
#define MAX_PROCS 100
#define MAX_MACHINE_STATES 50

typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char type[16];
  char states[512];
  char color[16];
  int  metric;
  int  flags;
  int  reserved1;
  int  reserved2;
} OprProfileEx;

#define PRF_ALWAYS_VISIBLE 0x01
#define PRF_NON_OPERATIONAL 0x02
#define PRF_EXCLUDE_STATS 0x04
#define PRF_PROBLEMS 0x08
#define PRF_NOT_AVAILABLE 0x10
#define PRF_WARMUP 0x20

#define isAlwayVisible(f) (((f) & PRF_ALWAYS_VISIBLE) == PRF_ALWAYS_VISIBLE ? TRUE : FALSE)
#define isNonOperational(f) (((f) & PRF_NON_OPERATIONAL) == PRF_NON_OPERATIONAL ? TRUE : FALSE)
#define isExcludedInStatistics(f) (((f) & PRF_EXCLUDE_STATS) == PRF_EXCLUDE_STATS ? TRUE : FALSE)
#define isProblemStatistics(f) (((f) & PRF_PROBLEMS) == PRF_PROBLEMS ? TRUE : FALSE)
#define isNotAvailableStatistics(f) (((f) & PRF_NOT_AVAILABLE) == PRF_NOT_AVAILABLE ? TRUE : FALSE)
#define isWarmup(f) (((f) & PRF_WARMUP) == PRF_WARMUP ? TRUE : FALSE)

enum ProfileType
{
  PT_SLICE, PT_OPTIONAL, PT_ERROR, PT_BEAM,
};
int PrfTypeOf(char *type)
{
  if (!stricmp(type, "SLICE")) return PT_SLICE;
  if (!stricmp(type, "OPTIONAL")) return PT_OPTIONAL;
  if (!stricmp(type, "ERROR")) return PT_ERROR;
  if (!stricmp(type, "BEAM")) return PT_BEAM;
  return PT_OPTIONAL;
}
enum ProfileBucket {
  PB_TOTAL, PB_BEAM, PB_ERROR,
};

typedef struct ProfileBeanCounter
{
  char name[64];
  long sum;
  long correction;
  time_t start;
  time_t end;
} ProfileBeanCounter;

int gPrfCntsCurrent[100];
int gCorrCntsCurrent[100];
NAME64 procedureNames[100];
int numProcedures;

enum AvailabilityEpoch {
  AV_YEAR = 0, // 1st of January
  AV_MONTH, // 1st of current month
  AV_WEEK,  // this past Sunday
  AV_SHIFT, // 07:00, 15:00, or 23:00
  AV_USER,  // defined by WRITE call
};
#define NEPOCHS 5
NAME64 epochNames[5] = {
  "Beginning of YEAR",
  "Beginning of MONTH",
  "Beginning of WEEK",
  "Beginning of SHIFT",
  "User defined"
};
time_t tepoch[NEPOCHS] = {0};
time_t tsrvstart = 0;
int getEpochId(char *str)
{
  int eid = AV_YEAR;
  if (str != NULL)
  {
    if (!stricmp(str, "MONTH")) eid = AV_MONTH;
    else if (!stricmp(str, "WEEK")) eid = AV_WEEK;
    else if (!stricmp(str, "SHIFT")) eid = AV_SHIFT;
  }
  return eid;
}
int gLastCorrectionsTimestamp = 0;
int nDaysInMonth[2][12] = {
  {31,28,31,30,31,30,31,31,30,31,30,31},
  {31,29,31,30,31,30,31,31,30,31,30,31},
};
#define numDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonth[0][mm] : nDaysInMonth[1][mm])

void cbCounts(int id, int cc)
{
  if (cc != 0) return;
  //??
}
int numtoks(const char *str, char tok)
{
  int count = 0;
  if (str == NULL || strlen(str) == 0) return -1;
  while (*str != '\0') count += *str++ == tok;
  return count;
}
int getProfileStates(char *parseableStateList, ProfileState **list)
{
  char stateList[512], *tok = ",", *c = NULL;
  int i, n = 0;
  ProfileState *states = NULL;
  if (parseableStateList == NULL || list == NULL) return 0;
  strncpy(stateList, parseableStateList, 512);
  n = numtoks(stateList, ',') + 1;
  if ((states=(ProfileState *)calloc(n, sizeof(ProfileState))) == NULL) return 0;
  for (c=strtok(stateList, tok), i=0; c != NULL && i<n; c=strtok(NULL, tok))
  {
    if (strlen(c) == 0) continue;
    strncpy(states[i++].name, c, 64);
  }
  *list = states;
  return n;
}
#define USR_FILENAME "userepoch.csv"
time_t getSavedUserEpoch(void)
{
  FILE *fp = NULL;
  char str[1024], scratch[64];
  int n = 0, done = 0, cc = 0;
  int tim_col = -1;

  if ((fp=fopen(USR_FILENAME, "rt")) == NULL) ccerr(no_such_file);
  while (fgets(str, 1023, fp) != NULL)
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n", str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((tim_col = findcol(str, "BEGIN")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    getColumnStr(tim_col, str, scratch, 64);
    tepoch[AV_USER] = (time_t)parseTineTimeString(scratch);
  }
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0 && cc != no_such_file)
  {
    feclog("error reading user epoch file %s : %.32s", USR_FILENAME, cc2str(cc));
  }
  return cc;
}
int saveUserEpoch(void)
{
  FILE *fp = NULL;
  int cc = 0;
  time_t et = tepoch[AV_USER];
  struct tm *etm = localtime(&et);
  if ((fp=fopen(USR_FILENAME, "wt")) == NULL) ccerr(no_such_file);
  fprintf(fp, "BEGIN\n");
  fprintf(fp, "%02d.%02d.%d %02d:%02d:%02d",
    etm->tm_mday, etm->tm_mon + 1, etm->tm_year + 1900,
    etm->tm_hour, etm->tm_min, etm->tm_sec);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
ProfileCounts *gCurrentProfileCounts = NULL;
ProfileCounts *getCurrentPrfCntsItem(char *name)
{
  int i;
  if (gCurrentProfileCounts == NULL)
  { // not yet allocated ...
    gCurrentProfileCounts = (ProfileCounts *)calloc(numProcedures, sizeof(ProfileCounts));
    if (gCurrentProfileCounts == NULL)
    {
      msglog(out_of_local_memory, "getCurrentPrfCntsItem: out of memory");
      return NULL;
    }
    for (i=0; i<numProcedures; i++)
    { 
      strncpy(gCurrentProfileCounts[i].name, procedureNames[i].name, DEVICE_NAME_SIZE);
    }
  }
  for (i=0; i<numProcedures; i++)
  { // find 'im
    if (!strncmp(gCurrentProfileCounts[i].name, name, DEVICE_NAME_SIZE))
    {
      gCurrentProfileCounts[i].counts = gPrfCntsCurrent[i];
      return &gCurrentProfileCounts[i];
    }
  }
  return NULL;
}
long getCurrentCounterDifference(ProfileState state, time_t start)
{
  ProfileCounts *prfCntsStart = getPrfCntsLstItem(state.name, start);
  ProfileCounts *prfCntsCurr = getCurrentPrfCntsItem(state.name);
  if (prfCntsCurr == NULL || prfCntsStart == NULL) return 0;
  return prfCntsCurr->counts - prfCntsStart->counts;
}
long getCounterDifference(ProfileState state,time_t start, time_t end)
{
  ProfileCounts *prfCntsStart = getPrfCntsLstItem(state.name, start);
  ProfileCounts *prfCntsEnd = getPrfCntsLstItem(state.name, end);
  if (prfCntsEnd == NULL || prfCntsStart == NULL) return 0;
  return prfCntsEnd->counts - prfCntsStart->counts;
}
void collectProfileData(ProfileBeanCounter *prfBeanCtr,ProfileConfig *prfConfig, time_t startTime, time_t endTime)
{
  int i, k;
  long cnt;

  for (i = 0; i < prfConfig->nprofiles; i++)
  {
    cnt = 0;
    for (k = 0; k < prfConfig->profiles[i].nstates; k++)
    {
      cnt += getCounterDifference(prfConfig->profiles[i].states[k], startTime, endTime);
    }
    prfBeanCtr[i].sum = cnt;
  }
}
void collectCurrentProfileData(ProfileBeanCounter *prfBeanCtr, ProfileConfig *prfConfig, time_t startTime)
{
  int i, k;
  long cnt;
  for (i = 0; i < prfConfig->nprofiles; i++)
  {
    cnt = 0;
    for (k = 0; k < prfConfig->profiles[i].nstates; k++)
    {
      cnt += getCurrentCounterDifference(prfConfig->profiles[i].states[k], startTime);
    }
    prfBeanCtr[i].sum = cnt;
  }
}
void loadStateCorrectionCountsForProfiles(ProfileBeanCounter *prfBeanCtr, ProfileConfig *prfConfig, time_t startTime, time_t endTime)
{
  int i, k, m, cc;
  static int ccnts[100] = {0}, range[2] = {0, 0};

  if (range[0] != (int)startTime || range[1] != (int)endTime)
  {
    DTYPE dout, din;
    char dev[128];
    char *prp = "CorrectionCounts.Procedures";
    range[0] = (int)startTime; range[1] = (int)endTime;
    DTYPEZERO(dout); DTYPEZERO(din);
    dout.dArrayLength = 100;
    dout.dFormat = CF_INT32;
    dout.data.lptr = ccnts;
    din.dArrayLength = 2;
    din.dFormat = CF_INT32;
    din.data.lptr = range;
    sprintf(dev, "/%.32s/STATE/#0", thisFecContext);
    cc = ExecLinkEx(dev, prp, &dout, &din, CA_READ | CA_MUTABLE, 800);
    if (cc != 0)
    {
      feclogEx(TINE_LOGLEVEL_ERROR, "cannot get correction counts : %.32s", cc2str(cc));
      return;
    }
  }
  for (i=0; i< prfConfig->nprofiles; i++)
  { // a profile (pie slice) is made up of procedures ...
    prfBeanCtr[i].correction = 0;
    for (k = 0; k < numProcedures; k++)
    { // a procedure is made up of states ...
      if (ccnts[k] == 0) continue;
      for (m = 0; m < prfConfig->profiles[i].nstates; m++)
      {
        if (!strnicmp(prfConfig->profiles[i].states[m].name, procedureNames[k].name, DEVICE_NAME_SIZE))
        {
          prfBeanCtr[i].correction += ccnts[k];
          break;
        }
      }
    }
    prfBeanCtr[i].start = startTime;
    prfBeanCtr[i].end = endTime;
  }
  return;
}
double calculateAvailability(ProfileConfig *prfConfig, int applyCorrections, time_t startTime, time_t endTime)
{
  ProfileBeanCounter *prfBeanCntr = NULL;
  int cc = 0, i, n = 0;
  double availableTime = 0;
  double availableFullTime = 0;
  double value;
  if (prfConfig == NULL) return 0;
  n = prfConfig->nprofiles;
  if ((prfBeanCntr=(ProfileBeanCounter *)calloc(n, sizeof(ProfileBeanCounter))) == NULL)
    ccerr(out_of_local_memory);
  for (i=0; i<n; i++) 
    strncpy(prfBeanCntr[i].name, prfConfig->profiles[i].name, DEVICE_NAME_SIZE);
  collectProfileData(prfBeanCntr, prfConfig, startTime, endTime);
  loadStateCorrectionCountsForProfiles(prfBeanCntr, prfConfig, startTime, endTime);
  for (i = 0; i < n; i++)
  {
    value = (double)prfBeanCntr[i].sum;
    if (applyCorrections) value += prfBeanCntr[i].correction;
    if (!prfConfig->profiles[i].exclude)
    {
      availableFullTime += value;
      if (!prfConfig->profiles[i].subtract) availableTime += value;
    }
  }
err:
  if (prfBeanCntr) free(prfBeanCntr);
  return (availableTime / availableFullTime) * 100.;
}
ProfileBeanCounter currBeanCntr[100] = {0};
ProfileBeanCounter corrBeanCntrs[NEPOCHS][MAX_PROFILES][100];
double currentAvailability(ProfileConfig *prfConfig, int applyCorrections, int epoch)
{
  static int startup = TRUE;
  time_t startTime = tepoch[epoch];
  ProfileBeanCounter *prfBeanCntr = currBeanCntr;
  int cc = 0, i, p, n = 0;
  double availableTime = 0;
  double availableFullTime = 0;
  double value;
  if (prfConfig == NULL) return 0;
  n = prfConfig->nprofiles;
//  for (i = 0; i < n; i++)
//    strncpy(prfBeanCntr[i].name, prfConfig->profiles[i].name, DEVICE_NAME_SIZE);
  collectCurrentProfileData(prfBeanCntr, prfConfig, startTime);
  for (p = 0; p < numConfigs && strncmp(profileConfigNames[p].name, prfConfig->name, 64); p++);
  for (i = 0; i < n; i++)
  {
    value = (double)prfBeanCntr[i].sum;
    if (applyCorrections) value += corrBeanCntrs[epoch][p][i].correction;
    if (!prfConfig->profiles[i].exclude)
    {
      availableFullTime += value;
      if (!prfConfig->profiles[i].subtract) availableTime += value;
    }
  }
  if (availableFullTime == 0) return 0.0;
  return (availableTime / availableFullTime) * 100.;
}
OprProfileEx profs[200];
int getOperationHistoryProfiles(char *context, char *configName, Profile **profilesList, int *nprofiles)
{
  DTYPE dout;
  char dev[128]; 
  char *prp = "PROFILE";
  char *tag = "PrfCfgEx";
  Profile *profiles = NULL;
  int cc, i, n = 0;
  if (context == NULL || configName == NULL) return argument_list_error;
  if (profilesList == NULL || nprofiles == NULL) return argument_list_error;
  DTYPEZERO(dout);
  dout.dArrayLength = 200;
  dout.dFormat = CF_STRUCT;
  strncpy(dout.dTag, tag, TAG_NAME_SIZE);
  dout.data.vptr = profs;
  sprintf(dev, "/%.32s/STATE/%.64s", context, configName);
  cc = ExecLinkEx(dev, prp, &dout, NULL, CA_READ|CA_MUTABLE, 800);
  if (cc != 0)
  { 
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot get profiles for config %.64s : %.32s", configName,cc2str(cc));
    return cc;
  }
  n = dout.dArrayLength;
  profiles = (Profile *)calloc(n, sizeof(Profile));
  for (i=0; i < n; i++)
  {
    strncpy(profiles[i].name, profs[i].name, 64);
    profiles[i].type = PrfTypeOf(profs[i].type);
    profiles[i].exclude = isExcludedInStatistics(profs[i].flags);
    profiles[i].subtract = isNotAvailableStatistics(profs[i].flags);
    profiles[i].nstates = getProfileStates(profs[i].states, &profiles[i].states);
  }
  *profilesList = profiles;
  *nprofiles = n;
  return cc;
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(thisFecServer, AVAEQM_TAG, AVAEQM_MAXDEVICES, avaeqm, avaeqm_ini, avaeqm_bkg, AVAEQM_BKG_RATE, avaeqm_exi);
}
void PreSystemInit(void)
{
  SetCatchConsoleBreak(TRUE);
  if (strlen(thisFecContext) > 0)
  {
    int p;
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName, thisFecContext, 10); thisFecName[10] = 0;
      strcat(thisFecName, "AVAFEC");
    }
    if (thisFecPort == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisFecPort = p;
    }
    sprintf(thisFecDescription, "%.32s Alarm State server", thisFecContext);
    RegisterFecInformation(thisFecName, thisFecSubsystem, thisFecContext, thisFecDescription, thisFecLocation, "none", thisFecResponsible, thisFecPort);
    RegisterStructs();
  }
}
#define quit(t,i) { feclogEx(TINE_LOGLEVEL_FATAL,"%s: %.32s",t,cc2str(i)); exit(1); }
void RegisterStructs(void)
{
  static int done = FALSE;
  char *tag = "PrfCfgEx";
  int cc;
  if (done) return;
  done = TRUE;
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, name), DEVICE_NAME_SIZE, CF_TEXT, "name"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, type), 16, CF_TEXT, "type"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, states), 512, CF_TEXT, "states"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, color), 16, CF_TEXT, "color"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, metric), 1, CF_INT32, "metric"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, flags), 1, CF_INT32, "flags"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, reserved1), 1, CF_INT32, "reserved1"))) quit(tag,cc);
  if ((cc=addFieldToStruct(tag, OFFSETIN(OprProfileEx, reserved2), 1, CF_INT32, "reserved2"))) quit(tag,cc);
  /* terminate the structure definition like this! */
  if ((cc=sealTaggedStruct(tag, sizeof(OprProfileEx), MAX_NUM_CONFIGS))) quit(tag,cc);
}
int RegisterProperties(void)
{
  int id, cc = 0;
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = NEPOCHS;
  din.dFormat = CF_INT32;
  din.dArrayLength = 2;
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability",
    &dout, &din, CA_READ|CA_HIST, AT_SINGLE, 0, "[0:100 %]Running availability (with corrections) since start of year", PRP_AVAILABILITY, NULL);
  if (id < 0) ccerr(-id);
  id = RegisterPropertyInformation(AVAEQM_TAG, "Availability.Raw",
    &dout, &din, CA_READ|CA_HIST, AT_SINGLE, 0, "[0:100 %]Running availability (raw data: no corrections) since start of year", PRP_AVAILABILITY_RAW, NULL);
  if (id < 0) ccerr(-id);
//  id = RegisterPropertyInformation(AVAEQM_TAG, "Synchronize",
//    NULL, NULL, CA_WRITE, AT_UNKNOWN, 0, "Synchronize the running availability since start of year", PRP_SYNCHRONIZE, NULL);
//  if (id < 0) ccerr(-id);
  dout.dFormat = CF_INT32;
  dout.dArrayLength = NEPOCHS;
  din.dFormat = CF_INT32;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(AVAEQM_TAG, "Epoch",
    &dout, &din, CA_READ|CA_WRITE|CA_SAVERESTORE, AT_CHANNEL, 0, "Epoch begin as UTC (or string)", PRP_EPOCH, NULL);
  if (id < 0) ccerr(-id);
  AssignDeviceListToProperty(AVAEQM_TAG, "Epoch", epochNames, NEPOCHS);

err:
  if (cc != 0)
  {
    dbglog("Can't register property : %32s", cc2str(cc));
  }
  return cc;
}
NAME64 profileConfigNames[MAX_PROFILES];
NAME64 procedureNames[MAX_PROCEDURES];
int numConfigs = MAX_PROFILES;
int numProcedures = MAX_PROCEDURES;
ProfileConfig *profileConfigs = NULL;
int RegisterDevices(void)
{
  int i, cc = 0;
  char srv[256];
  sprintf(srv, "/%.32s/STATE/#0",thisFecContext);
  cc = GetSystemDevices(srv, "CONFIGURATION", profileConfigNames, &numConfigs);
  if (cc == 0)
  {
    if (numConfigs > MAX_PROFILES) numConfigs = MAX_PROFILES;
    for (i = 0; i < numConfigs; i++)
      RegisterDeviceName(AVAEQM_TAG, profileConfigNames[i].name, i);
  }
  return cc;
}
void removePrfCntsItems(time_t timestamp)
{
  int p, i, k;
  if (profileConfigs == NULL) return;
  if (timestamp == 0) return;
  for (p = 0; p < numConfigs; p++)
  { // all the profile configs ...
    for (i = 0; i < profileConfigs[p].nprofiles; i++)
    { // all the profiles (slices) in the config
      for (k = 0; k < profileConfigs[p].profiles[i].nstates; k++)
      { // all the state procedures in the profile
        rmvPrfCntsLstItem(profileConfigs[p].profiles[i].states[k].name, timestamp);
      }
    }
  }
}
/** 
 * return TRUE if an epoch has changed ...
 */
int resetEpochs(void)
{
  int stsChanged = FALSE;
  time_t t = time(NULL);
  time_t et = 0;
  struct tm ltm = {0}, *ltn = localtime(&t);

  ltm.tm_year = ltn->tm_year;
  ltm.tm_mday = 1;
  et = mktime(&ltm);
  if (et != tepoch[AV_YEAR])
  {
    removePrfCntsItems(tepoch[AV_YEAR]);
    stsChanged = TRUE;
  }
  tepoch[AV_YEAR] = et;

  ltm.tm_mon = ltn->tm_mon;
  ltm.tm_hour = 0;
  et = mktime(&ltm);
  if (et != tepoch[AV_MONTH]) 
  {
    removePrfCntsItems(tepoch[AV_MONTH]);
    stsChanged = TRUE;
  }
  tepoch[AV_MONTH] = et;

  ltm.tm_hour = 0;
  ltm.tm_mday = ltn->tm_mday - ltn->tm_wday;
  if (ltm.tm_mday < 0)
  {
    if (ltm.tm_mon > 0)
    {
      ltm.tm_mon--;
      ltm.tm_mday = numDaysInMonth(ltm.tm_year, ltm.tm_mon) + ltm.tm_mday;
    }
    else
    { // if beginning of January then start the week on Jan. 1st
      ltm.tm_mday = 1;
    }
  }
  et = mktime(&ltm);
  if (et != tepoch[AV_WEEK]) 
  {
    removePrfCntsItems(tepoch[AV_WEEK]);
    stsChanged = TRUE;
  }
  tepoch[AV_WEEK] = et;

  ltm.tm_mday = ltn->tm_mday;
  ltm.tm_hour = 8 * ((ltn->tm_hour + 1) / 8) - 1; // 7:00, 15:00, 23:00
  if (ltm.tm_hour < 0)
  { // = -1 if HH < 07
    ltm.tm_mday--;
    ltm.tm_hour = 23;
  }
  et = mktime(&ltm);
  if (et != tepoch[AV_SHIFT]) 
  {
    removePrfCntsItems(tepoch[AV_SHIFT]);
    stsChanged = TRUE;
  }
  tepoch[AV_SHIFT] = et;

  return stsChanged;
}
void avaeqm_ini(void)
{
  int i, cc = 0;
  char dev[256];
  char *prp = "TIMECOUNTER.Procedure";
  DTYPE dout;
  tsrvstart = time(NULL);
  resetEpochs();
  getSavedUserEpoch();
  if ((cc=RegisterProperties()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR,"Could not register properties (%.32s) : Abort!",cc2str(cc));
    exit(1);
  }
  if ((cc=RegisterDevices()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "Could not register devices (%.32s) : Abort!", cc2str(cc));
    exit(1);
  }
  profileConfigs = (ProfileConfig *)calloc(numConfigs, sizeof(ProfileConfig));
  for (i = 0; i < numConfigs; i++)
  {
    getOperationHistoryProfiles(thisFecContext, profileConfigNames[i].name, &profileConfigs[i].profiles, &profileConfigs[i].nprofiles);
    strncpy(profileConfigs[i].name, profileConfigNames[i].name, DEVICE_NAME_SIZE);
  }
  sprintf(dev, "/%.32s/STATE/#0", thisFecContext);
  if ((cc=GetSystemDevices(dev, prp, procedureNames, &numProcedures)) != 0)
  {
    feclogEx(TINE_LOGLEVEL_ERROR, "cannot get procedure names : %.32s", cc2str(cc));
  }
  DTYPEZERO(dout);
  dout.dArrayLength = numProcedures;
  dout.dFormat = CF_INT32;
  dout.data.lptr = gPrfCntsCurrent;
  AttachLink(dev,prp,&dout,NULL,CA_READ,1000,cbCounts,CM_TIMER);

  prp = "LastCorrectionTimestamp";
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &gLastCorrectionsTimestamp;
  AttachLink(dev, prp, &dout, NULL, CA_READ, 1000, cbCounts, CM_TIMER);
}
void avaeqm_exi(void)
{
  //writeDeclaredCounters();
}
int epochs_changed = FALSE;
void avaeqm_bkg(void)
{
  static time_t tlast_corrections = 0;
  time_t tnow = time(NULL);
  int i, eid, corrections_changed = FALSE;

  if ((tnow % 3600) == 0)
  { //on the hour: check for change in shift, week, month, etc ...
    epochs_changed = resetEpochs();
  }
  if (gLastCorrectionsTimestamp != (int)tlast_corrections)
  { // there are new corrections !
    corrections_changed = TRUE;
  }
  tlast_corrections = (time_t)gLastCorrectionsTimestamp;
  if (!epochs_changed && !corrections_changed) return; // no reason to continue
  for (eid=0; eid<NEPOCHS; eid++)
    for (i = 0; i < numConfigs; i++)
    {
      if (tepoch[eid] == 0) continue;
      loadStateCorrectionCountsForProfiles(corrBeanCntrs[eid][i], &profileConfigs[i], tepoch[eid], tnow);
    }

}
int avaeqm(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access)
{
  int cc, did, pid, eid = AV_YEAR, applyCorrections = TRUE, useCurrentStats = FALSE;
  double av = 0;
  time_t t = time(NULL), tstop = t, tstart = 0;
  struct tm ltm = {0}, *lt = localtime(&t);
  did = GetDeviceNumberEx(AVAEQM_TAG, devName, devProperty);
  if (did < 0) return illegal_equipment_number;

  pid = GetPropertyId(AVAEQM_TAG, devProperty);

  //ntotal = gNumAlarmSystems + gNumAlarmGroups;
  switch (pid)
  {
    case PRP_AVAILABILITY_RAW:
      applyCorrections = FALSE;
    case PRP_AVAILABILITY:
      if (access & CA_WRITE) return illegal_read_write;
      if (din->dArrayLength > 0)
      {
        if (IsIntegerFormat(din->dFormat))
        {
          int vals[2];
          if (din->dArrayLength == 1)
          {
            int yy = lt->tm_year + 1900;
            cc = GetValuesAsLong(din, vals, 1);
            if (cc != 0) return cc;
            if (vals[0] < 100) vals[0] += 2000;
            if (vals[0] < 2050)
            { // given as the desired year
              yy = vals[0];
              ltm.tm_year = yy - 1900;
              ltm.tm_mday = 1;
              tstart = mktime(&ltm);
              ltm.tm_year--;
              tstop = mktime(&ltm) - 1;
            }
            else
            { // given as a start time in UTC (stop is 'now')
              tstart = (time_t)vals[0];
              useCurrentStats = TRUE;
            }
          }
          else
          { // 2 values as start and stop in UTC 
            cc = GetValuesAsLong(din, vals, 2);
            if (cc != 0) return cc;
            tstart = (time_t)vals[0];
            tstop = (time_t)vals[1];
          }
        }
        else if (IsStringFormat(din->dFormat))
        {
          char epochStr[64];
          if ((cc=GetValuesAsStringEx(din, epochStr, CF_TEXT, 64, 0)) != 0)
            return cc;
          eid = getEpochId(epochStr);
          useCurrentStats = TRUE;
        }
      }
      else
      { // no input: start of the year up to 'now'
        tstart = tepoch[AV_YEAR];
        eid = AV_YEAR;
        useCurrentStats = TRUE;
      }
      if (dout->dArrayLength > 0)
      {
        if (access & CA_HIST) useCurrentStats = TRUE;
        if (useCurrentStats)
        {
          double avals[NEPOCHS];
          int i;
          for (i = 0; i < NEPOCHS; i++)
          {
            avals[i] = currentAvailability(&profileConfigs[did], applyCorrections, i);
          }
          if (dout->dArrayLength > NEPOCHS) dout->dArrayLength = NEPOCHS;
          return PutValuesFromDoubleEx(dout, avals, NEPOCHS, eid);
        }
        else
        {
          dout->dArrayLength = 1;
          av = calculateAvailability(&profileConfigs[did], applyCorrections, tstart, tstop);
          return PutValuesFromDoubleEx(dout, &av, 1, 0);
        }
      }
      return dimension_error;
    case PRP_EPOCH:
      if (access & CA_WRITE)
      {
        int et = 0;
        if (din->dArrayLength == 0) return illegal_read_write;
        if (did != AV_USER) return not_allowed;
        if (IsNumberFormat(din->dFormat))
        {
          if ((cc=GetValuesAsLong(din, &et, 1)) != 0) return cc;
        }
        else if (IsStringFormat(din->dFormat))
        {
          char etstr[64];
          if ((cc=GetValuesAsStringEx(din, etstr, CF_TEXT, 64, 0)) != 0)
            return cc;
          et = (int)parseTineTimeString(etstr);
        }
        else
        {
          return illegal_format;
        }
        if (et < (int)tepoch[AV_YEAR] || et > (int)t) return out_of_range;
        if (tepoch[AV_USER] != (time_t)et)
        {
          if (tepoch[AV_USER] != 0) removePrfCntsItems(tepoch[AV_USER]);
          epochs_changed = TRUE;
        }
        tepoch[AV_USER] = (time_t)et;
        saveUserEpoch();
      }
      if (dout->dArrayLength > 0)
      {
        int i;
        if (IsStringFormat(dout->dFormat))
        {
          NAME64 estrvals[NEPOCHS] = {0};
          for (i = 0; i < NEPOCHS; i++) strncpy(estrvals[i].name, ctime(&tepoch[i]), 24);
          cc =PutValuesFromString(dout, estrvals[0].name, CF_NAME64, NEPOCHS, did);
        }
        else
        {
          int tvals[NEPOCHS];
          for (i = 0; i < NEPOCHS; i++) tvals[i] = (int)tepoch[i];
          cc = PutValuesFromLongEx(dout, tvals, NEPOCHS, did);
        }
        return cc;
      }
      return 0;
    case PRP_SYNCHRONIZE:
      return 0;
  }
  return illegal_property;
}