#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "avaeqm.h"

int GetArchivedCountsSnapshot(time_t target, int *counts, NAME64 *names, int size);

ProfileCounts *gPrfCntsTbl[LNKTBL_HASH_SIZE];
int gPrfCntsTblSize = 0;

ProfileCounts *getPrfCntsLstItemByKey(char *key)
{
  ProfileCounts *item;
  char itemkey[256];
  int idx;

  if (gPrfCntsTblSize == 0) return NULL; /* no need to look */
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);

  for (item=gPrfCntsTbl[idx]; item != NULL; item = item->nxt)
  {
    sprintf(itemkey,"/%.32s/STATE/%.64s:%ld",thisFecContext,item->name,item->timestamp);
    if (strnicmp(itemkey, key, 256)) continue;
    break;
  }
  return item;
}
int PrfCntxLstContains(char *key)
{
  ProfileCounts *item;
  char itemkey[256];
  int idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);
  if (gPrfCntsTblSize == 0) return FALSE; /* no need to look */
  for (item=gPrfCntsTbl[idx]; item != NULL; item = item->nxt)
  {
    sprintf(itemkey, "/%.32s/STATE/%.64s:%ld", thisFecContext, item->name, item->timestamp);
    if (strnicmp(itemkey, key, 256)) continue;
    return TRUE;
  }
  return FALSE;
}
void rmvPrfCntsLstItem(char *name, time_t timestamp)
{
  char key[256];
  int idx;
  ProfileCounts *item = NULL, *prv_item = NULL;

  sprintf(key, "/%.32s/STATE/%.64s:%ld", thisFecContext, name, timestamp);
  idx = (int)(ElfHash((unsigned char *)key) % LNKTBL_HASH_SIZE);

  for (item=gPrfCntsTbl[idx]; item != NULL; item = item->nxt)
  {
    if (!strnicmp(item->name, name, 64) && item->timestamp == timestamp)
    { // the item to remove !
      if (gPrfCntsTbl[idx] == item)
      { // right at the head ...
        gPrfCntsTbl[idx] = item->nxt;
      }
      if (prv_item != NULL)
      {
        prv_item->nxt = item->nxt;
      }
      free(item);
      srvlog(SRV_FILENAME, "remove %.192s from the counts list", key);
      return;
    }
    prv_item = item;   
  }
  return;
}
ProfileCounts *getPrfCntsLstItem(char *name,time_t timestamp)
{
  int idx, cc, i, j, k, m, n;
  int cnts[100];
  NAME64 nams[100];
  char itemkey[256];
  ProfileCounts *item = NULL;
  sprintf(itemkey, "/%.32s/STATE/%.64s:%ld", thisFecContext, name, timestamp);
  if ((item=getPrfCntsLstItemByKey(itemkey)) != NULL) return item; // what luck!
  n = 100;
  if ((cc=GetArchivedCountsSnapshot(timestamp, cnts, nams, 100)) == 0)
  {
    for (i = 0; i < 100; i++)
    { // scan over all possible archive entries ...
      if (strlen(nams[i].name) == 0) break; // break at the end of the line
      for (j = 0; j < numConfigs; j++)
      { // scan over all possible 'configs' 
        for (k = 0; k < profileConfigs[j].nprofiles; k++)
        { // and over the profiles in each config ...
          for (m = 0; m < profileConfigs[j].profiles[k].nstates; m++)
          { // and over the states in each profile ...
            if (!strnicmp(nams[i].name, profileConfigs[j].profiles[k].states[m].name, DEVICE_NAME_SIZE))
            { // got a match ! => make a new item
              sprintf(itemkey, "/%.32s/STATE/%.64s:%ld", thisFecContext, nams[i].name, timestamp);
              if (PrfCntxLstContains(itemkey)) break; // avoid multiple table entries ...
              idx = (int)(ElfHash((unsigned char *)itemkey) % LNKTBL_HASH_SIZE);
              if ((item=(ProfileCounts *)SystemCalloc(1, sizeof(ProfileCounts))) == NULL)
              {
                msglog(out_of_local_memory, "appendPrfCntsLst: out of memory");
                return NULL;
              }
              strncpy(item->name, nams[i].name, DEVICE_NAME_SIZE);
              item->counts = cnts[i];
              item->timestamp = timestamp;
              item->nxt = gPrfCntsTbl[idx];
              gPrfCntsTbl[idx] = item;
              gPrfCntsTblSize++;
              srvlog(SRV_FILENAME,"append %.192s to the counts list", itemkey);
              break;
            }
          }
        }
      }
    }
  }
  else
  {
    srvlog(SRV_FILENAME, "cannot get archived counts at %.24s : %.32s", ctime(&timestamp),cc2str(cc));
  }
  // should be there now : 
  sprintf(itemkey, "/%.32s/STATE/%.64s:%ld", thisFecContext, name, timestamp);
  return getPrfCntsLstItemByKey(itemkey);
}
int GetArchivedCountsSnapshot(time_t target, int *counts, NAME64 *names, int size)
{
  DTYPE din, dout;
  UINT32 range[3];
  char dev[STATUS_SIZE], prp[PROPERTY_NAME_SIZE];
  int cc;

  if (size <= 0) return buffer_too_small;
  if (target <= 0 || counts == NULL) return argument_list_error;
  DTYPEZERO(dout); DTYPEZERO(din);
  range[0] = (UINT32)(target);
  range[1] = (UINT32)(target);
  dout.dArrayLength = size;
  dout.dFormat = CF_NAME64;
  dout.data.vptr = names;
  din.dArrayLength = 1;
  din.dFormat = CF_INT32;
  din.data.ulptr = range;

  sprintf(dev, "/%.32s/HISTORY/#0", thisFecContext);
  cc = ExecLinkEx(dev, "STATECNT.NAM", &dout, &din, CA_READ | CA_NOCALLBACKS, 1000);
  if (cc != 0) return cc;

  dout.dArrayLength = size;
  dout.dFormat = CF_INT32;
  dout.data.lptr = counts;
  din.dArrayLength = 1;
  din.dFormat = CF_INT32;
  din.data.ulptr = range;

  sprintf(dev, "/%.32s/HISTORY/#0", thisFecContext);
  cc = ExecLinkEx(dev, "STATECNT", &dout, &din, CA_READ | CA_NOCALLBACKS, 1000);

  return cc;
}
