#include <stdio.h>
#include "tbufsrv.h"
#include "threader.h"

void textout(char *str);

#define DEVSVR_TARGET "BufferedSineServer"
#define DEVSVR_PROPERTY "Sine"
#define DEVWR_PROPERTY "Amplitude"
char tstLnkSrv[64] = DEVSVR_TARGET;
char tstLnkDev[256];
char tstLnkPrp[72] = DEVSVR_PROPERTY;
char tstWrPrp[72] = DEVWR_PROPERTY;

float fvals[1024] = {0};
int lid = -1;
int nerr = 0;
void tstLnkCb(int id, int cc)
{
  if (cc != 0) nerr++;
}
int checkLinkUpdate(void)
{
  int cc = 0;
  struct timeval t_now, t_start;
  DTYPE dout;
  DTYPEZERO(dout);
  dout.dArrayLength = 1024;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = fvals;
  gettimeofday(&t_start, NULL);
  sprintf(tstLnkDev, "/TEST/%.32s/SineGen0", tstLnkSrv);
  lid = AttachLink(tstLnkDev, tstLnkPrp, &dout, NULL, CA_READ, 1000, tstLnkCb, CM_TIMER|CM_WAIT);
  gettimeofday(&t_now, NULL);
  if (MSECS(t_now, t_start) > 200) cc = link_timeout;
  return cc;
}
int checkWriteValue(void)
{
  float v = 333.0;
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = 1;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = &v;
  din.dArrayLength = 1;
  din.dFormat = CF_FLOAT;
  din.data.fptr = &v;
  return ExecLinkEx(tstLnkDev, tstWrPrp, &dout, &din, CA_READ | CA_WRITE, 500);
}
int serverAsClient(void)
{
  int v, cc = 0, ne = 0;
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = &v;

  cc = ExecLinkEx(tstLnkDev, "Errors", &dout, NULL, CA_READ, 500);
  if (cc != 0) ne++; else ne = v;
  cc = ExecLinkEx(tstLnkDev, "Updates", &dout, NULL, CA_READ, 500);
  if (cc != 0) ne++;
  if (v < 1) ne++;
  return ne;
}
char *usageText = "usage: bufclient [/s=<server> /d=<debug level>";
int main(int argc, char *argv[])
{
  int i, n;
  char *c, scratch[256];
  time_t t0 = time(NULL);

  for (i = 1; i < argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 's':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(tstLnkSrv, c, EXPORT_NAME_SIZE);
        }
      continue;
      case 'd':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          tineDebug = atoi(scratch);
        }
      continue;     case '?':
      case 'h':
        printf("%s", usageText);
        exit(0);
      default:
        continue;
    }
  }

  SystemStartCycleTimer();
  n = checkLinkUpdate();
  printf("------ checkLinkUpdate with WAIT ----- : %s\n", (n == 0 ? "success" : "failed"));
  n = checkWriteValue();
  printf("------ checkWriteValue ----- : %s\n", (n == 0 ? "success" : "failed"));
  n = serverAsClient();
  printf("------ serverAsClient ----- : %d errors\n", n);
  textout("------ checkLinkUpdate callback ");
  while (time(NULL) < t0 + 5)
  {
    millisleep(1000);
    textout("-");
  }
  printf(" : %d errors\n", nerr);
  return 0;
}
void textout(char *str)
{ // even putchar_unlocked doesn't work on linux : you still have to wait !
  char *c = str;
  while (c != NULL && *c != 0) putchar(*(c++));
}