#include <stdio.h>
#include "tbufsrv.h"
#include "threader.h"

#define DEVSVR_TARGET "/TEST/SineServer/SineGen0"
#define DEVSVR_PROPERTY "Amplitude"
char tstLnkDev[256] = DEVSVR_TARGET;
char tstLnkPrp[72] = DEVSVR_PROPERTY;
int lidTstLnk = -1;
int tstLnkErrors = 0;
int tstLnkUpdates = 0;
float tstLnkValue;

void tstLnkCb(int id, int cc)
{
  if (cc != 0) tstLnkErrors++;
  tstLnkUpdates++;
  pushBufferedData("Errors", "#0", (BYTE *)&tstLnkErrors, 1, FALSE);
  pushBufferedData("Updates", "#0", (BYTE *)&tstLnkUpdates, 1, FALSE);
}
int startRepresentativeClient(void)
{
  DTYPE dout;
  DTYPEZERO(dout);
  dout.dArrayLength = 1;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = &tstLnkValue;
  lidTstLnk = AttachLink(tstLnkDev, tstLnkPrp, &dout, NULL, CA_READ, 1000, tstLnkCb, CM_TIMER);
  return lidTstLnk < 0 ? -lidTstLnk : 0;
}
#define FECNAME "BUFSINEFEC"
#define SRVNAME "BufferedSineServer"
#define PRP_SINE 1
#define PRP_AMPLITUDE 2
#define PRP_FREQUENCY 3
#define PRP_NOISE 4
#define PRP_PHASE 5
#define PRP_DAMPING 6


#define PRP_LINK_ERRORS 7
#define PRP_LINK_UPDATES 8

#define NUM_VALUES 1024
#define NUM_DEVICES 10

#define random(r) (r != 0 ? rand()%(r) : 0)
typedef struct
{
  float amplitude;
  float frequency;
  float noise;
  float phase;
  float damping;
  char  description[64];
} SineInfo;
SineInfo sineInfoTable[NUM_DEVICES];
float sinbuf[NUM_DEVICES][NUM_VALUES];
char *sinname[NUM_DEVICES] = {
  "SineGen0","SineGen1","SineGen2","SineGen3","SineGen4",
  "SineGen5","SineGen6","SineGen7","SineGen8","SineGen9"
};

void updateSine(int devnr)
{ /* used internally -> don't bother with bounds checking */
  int i, k;
  float *target = sinbuf[devnr];
  float f = sineInfoTable[devnr].frequency;
  int n = (int)sineInfoTable[devnr].noise;
  int p = (int)sineInfoTable[devnr].phase;
  float a = sineInfoTable[devnr].amplitude;
  float d = sineInfoTable[devnr].damping / (float)100.0;
  for (i = 0; i < NUM_VALUES; i++)
  {
    k = (i + p) % NUM_VALUES;
    target[k] = random(n) + (float)(sin(f * i * 6.2832 / (NUM_VALUES / 8)) * a);
    if (d > 0.0)
    {
      target[k] *= ((float)1.0 - (d * (float)i) / (float)NUM_VALUES);
    }
  }
  pushBufferedData("Sine", sinname[devnr], (BYTE *)target, NUM_VALUES, TRUE);
  return;
}
#define CHKPRP(p) if (hasInputChanged(p)) {\
    if ((cc=pullBufferedData(p, dev, (BYTE *)&val, 1)) != 0) return cc;\
    if ((cc=pushBufferedData(p, dev, (BYTE *)&val, 1, TRUE)) != 0) return cc;\
  }
int prpHndlr(void)
{
  int cc, val;
  int nr = GetInputDeviceNumber();
  char dev[DEVICE_NAME_SIZE];
  sprintf(dev, "#%d", nr);
  CHKPRP("Amplitude");
  CHKPRP("Frequency");
  CHKPRP("Noise");
  CHKPRP("Phase");
  CHKPRP("Damping");
  return 0;
}
void upd(void)
{
  int i;
  for (i = 0; i < NUM_DEVICES; i++) updateSine(i);
}
char thisFecName[FEC_NAME_SIZE + 4] = FECNAME;
char thisContext[CONTEXT_NAME_SIZE + 4] = "TEST";
char thisServerName[EXPORT_NAME_SIZE] = SRVNAME;
UINT16 thisPortOffset = 0;
char *usageText = "usage: bufserv [/f=<fecname> /c=<context> /s=<server> /p=<port offset> /l=<link target as TINE full name> /t=<TTL in seconds, 0 => forever> /d=<debug level>";
int main(int argc,char *argv[])
{
  char expstr[256];
  int i, cc, ttl = 0;
  char *c, scratch[256];
  time_t t0 = time(NULL);
 
  for (i = 1; i < argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'f':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(thisFecName, c, FEC_NAME_SIZE);
        }
        continue;
      case 'c':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(thisContext, c, CONTEXT_NAME_SIZE);
        }
        continue;
      case 's':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(thisServerName, c, EXPORT_NAME_SIZE);
        }
        continue;
      case 'p':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          thisPortOffset = atoi(scratch);
        }
        continue;
      case 't': // time to live
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          ttl = atoi(scratch);
          if (ttl < 0) ttl = 0;
          if (ttl > 0 && ttl < 30) ttl = 30;
        }
        continue;
      case 'l': // target test link
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 255);
          if (strchr(scratch, '[') != NULL)
          { // well formed
            strncpy(tstLnkDev, scratch, 255);
            if ((c=strchr(tstLnkDev, '[')) != NULL)
            {
              *c++ = 0;
              strncpy(tstLnkPrp, c, 64);
              if ((c=strchr(tstLnkPrp, ']')) != NULL) *c = 0;
            }
          }
        }
        continue;
      case 'd':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          tineDebug = atoi(scratch);
        }
      continue;     case '?':
      case 'h':
        printf("%s", usageText);
        exit(0);
      default:
        continue;
    }
  }

  for (i = 0; i < NUM_DEVICES; i++)
  {
    sineInfoTable[i].amplitude = 256.0;
    sineInfoTable[i].frequency = 1.0;
    sineInfoTable[i].phase = 0.0;
    sineInfoTable[i].noise = 30.0;
    sineInfoTable[i].damping = 0.0;
    sprintf(sineInfoTable[i].description, "buffered sine %d at your disposal", i);
  }
  if (thisPortOffset == 0) thisPortOffset = GetPortOffset(thisFecName);
  RegisterFecInformation(thisFecName, "TEST", thisContext, "A buffered sine server", "here", "none", "me", thisPortOffset);
  sprintf(expstr, "%.32s/INIT=SRV|KEEP", thisServerName);
  cc = AttachServerEx(expstr, "BUFEQM", NUM_DEVICES, upd, 500);

  RegisterBufferedPropertyEx("Sine", 0, CF_NULL, 1024, CF_FLOAT, 1000, -1000, "V", CA_READ, "Sine Curve", PRP_SINE);
  RegisterBufferedPropertyEx("Amplitude", 1, CF_FLOAT, 10, CF_FLOAT, 1000, -1000, "V", CA_READ|CA_WRITE, "Sine Curve amplitude", PRP_AMPLITUDE);
  RegisterBufferedPropertyEx("Frequency", 1, CF_FLOAT, 10, CF_FLOAT, 100, 1, "Hz", CA_READ | CA_WRITE, "Sine Curve frequency", PRP_FREQUENCY);
  RegisterBufferedPropertyEx("Noise", 1, CF_FLOAT, 10, CF_FLOAT, 100, 0, "V", CA_READ | CA_WRITE, "Sine Curve noise", PRP_NOISE);
  RegisterBufferedPropertyEx("Phase", 1, CF_FLOAT, 10, CF_FLOAT, 10, 0, "Hz", CA_READ | CA_WRITE, "Sine Curve phase", PRP_PHASE);
  RegisterBufferedPropertyEx("Damping", 1, CF_FLOAT, 10, CF_FLOAT, 100, 0, "V", CA_READ | CA_WRITE, "Sine Curve damping", PRP_DAMPING);
  RegisterBufferedPropertyEx("Errors", 0, CF_NULL, 1, CF_INT32, 1000, 0, "Cnts", CA_READ, "Link Errors", PRP_LINK_ERRORS);
  RegisterBufferedPropertyEx("Updates", 0, CF_NULL, 1, CF_INT32, 1000, 0, "Cnts", CA_READ, "Link Errors", PRP_LINK_UPDATES);
  pushBufferedData("Errors", sinname[0], (BYTE *)&tstLnkErrors, 1, FALSE);
  pushBufferedData("Updates", sinname[0], (BYTE *)&tstLnkUpdates, 1, FALSE);

  for (i = 0; i < NUM_DEVICES; i++)
  {
    RegisterDeviceName("BUFEQM", sinname[i], i);
    pushBufferedData("Amplitude", sinname[i], (BYTE *)&sineInfoTable[i].amplitude, 1, FALSE);
    pushBufferedData("Frequency", sinname[i], (BYTE *)&sineInfoTable[i].frequency, 1, FALSE);
    pushBufferedData("Noise", sinname[i], (BYTE *)&sineInfoTable[i].noise, 1, FALSE);
    pushBufferedData("Phase", sinname[i], (BYTE *)&sineInfoTable[i].phase, 1, FALSE);
    pushBufferedData("Damping", sinname[i], (BYTE *)&sineInfoTable[i].damping, 1, FALSE);
  }
  RegisterServerCallback("Amplitude", prpHndlr);
  RegisterServerCallback("Frequency", prpHndlr);
  RegisterServerCallback("Noise", prpHndlr);
  RegisterServerCallback("Phase", prpHndlr);
  RegisterServerCallback("Damping", prpHndlr);

  startRepresentativeClient();

  SystemStartCycleTimer();
  if (ttl == 0)
  {
    SystemWaitCycleTimer();
  }
  else
  {
    while (time(NULL) < t0 + ttl)
    {
      millisleep(100);
    }
  }
  return 0;
}

