#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "caseqm.h"
#include "cas2doocs.h"

DoocsAlmXref *doocsAlmXref = NULL;
int nDoocsAlmXref = 0;

/**
 * These are the constants to define the severities for the
 * third parameter of the set_error methods.
 * The following constants are defined :
 * - SEVERITY_NOERROR	The device has no error
 * - SEVERITY_ERROR	The device has an error. This is the default,if the severity parameter is not used.
 * - SEVERITY_OFF	The device is off
 * - SEVERITY_WARN	A warning a user should know
 * - SEVERITY_FATAL	This error stops machine operation
 * - SEVERITY_OFFLINE	The device is out of service
 * - SEVERITY_INFO	The device likes to send an information
 *
 * @author O.Hensler
**/
 
#define SEVERITY_NOERROR 0
#define SEVERITY_ERROR 1
#define SEVERITY_OFF 2
#define SEVERITY_WARN 3
#define SEVERITY_FATAL 4
#define SEVERITY_OFFLINE 6
#define SEVERITY_INFO 7

#ifndef UNIX
struct tm *localtime_r(const time_t *timer,struct tm *result)
{
  struct tm *t;
  if (timer == NULL || result == NULL) return NULL;
  t = localtime(timer);
  *result = *t;
  return result;
}
#endif
void create_isodate (char* isodate_str, time_t tm, int ms)
{
  struct tm t;
  char time_str [20];

  if (!isodate_str) return;

  localtime_r (&tm, &t);
  memset(time_str,0,20);
  sprintf (time_str,"%02d:%02d:%02d,%03d", t.tm_hour, t.tm_min, t.tm_sec, ms);
  sprintf (isodate_str, "%d-%02d-%02d", t.tm_year + 1900, t.tm_mon + 1, t.tm_mday);
  strncat (isodate_str, "T", 1);
  strncat (isodate_str, time_str, 20);
}
int tine2doocsSeverity(int tineSeverity)
{
  if (tineSeverity == 0) return SEVERITY_NOERROR;
  if (tineSeverity < 4) return SEVERITY_INFO;
  if (tineSeverity < 8) return SEVERITY_WARN;
  if (tineSeverity < 12) return SEVERITY_ERROR;
  return SEVERITY_FATAL;
}
char *severity_string(int sev)
{
  switch (sev) 
  {
    case SEVERITY_NOERROR: return "OK";
    case SEVERITY_INFO: return "INFO";
    case SEVERITY_OFF: return "OFF";
    case SEVERITY_WARN: return "WARNING";
    case SEVERITY_ERROR: return "ERROR";
    case SEVERITY_FATAL: return "FATAL";
    case SEVERITY_OFFLINE: return "OFFLINE";
    default: return "OK";
  }
  return "OK";
}
char *GetReplString(char target)
{
  static char starget[2] = {0,0};
  switch (target)
  {
    case '&': return "amp;";
    case '<': return "lt;";
    case '>': return "gt'";
    case '\'': return "apos;";
    case '\"': return "quot;";
    default: break;
  }
  starget[0] = target;
  return starget;
}
char *GetWellFormedMessage(char *msg)
{
  static char illchrs[] = "&<>\'\"";
  static char msgbuf[128];
  char *c = msg, *cbrk;
  int clen, len = 112; /* leave a bit of space */
  msgbuf[0] = 0; /* start off empty */
  if (msg == NULL) return msgbuf;
  while (*c != 0)
  {
    if ((cbrk=strpbrk(c,illchrs)) == NULL)
    { /* no illegal chars from here on out */
      strncat(msgbuf,c,len);
      break;
    }
    clen = cbrk - c;
    strncat(msgbuf,c,clen);
    strcat(msgbuf,GetReplString(*cbrk));
    len -= strlen(msgbuf);
    c = ++cbrk;
  }
  return msgbuf;
}


CSVCOLUMN csvDoocsAlmXref[] =
{
  { "CONTEXT","",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(DoocsAlmXref,tineContext),NULL, 0},
  { "SERVER","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(DoocsAlmXref,tineServer),NULL, 0},
  { "SUBSYSTEM","",0,CF_TEXT,SUBSYSTEM_NAME_SIZE,(BYTE *)OFFSETIN(DoocsAlmXref,tineSubsystem),NULL, 0},
  { "ALARM_SERVER","",0,CF_TEXT,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+8,(BYTE *)OFFSETIN(DoocsAlmXref,doocsAlarmServer),NULL, 0},
  { "FACILITY","",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(DoocsAlmXref,doocsFacility),NULL, 0},
  { "DEVICE","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(DoocsAlmXref,doocsDevice),NULL, 0},
};
CSVDB csvDoocsAlmXrefDb = 
{
  csvDoocsAlmXref,
  sizeof(csvDoocsAlmXref)/sizeof(csvDoocsAlmXref[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(DoocsAlmXref),
  0
};

/* type can be "run", "init", or "alive" */
char *GetAlarmAsXML(AMS *ams,DoocsAlmXref *xref,char *type)
{ /* TODO: incorporate the ads->alarmText into the fray ... */
  int len = 4096;
  char iso[64], ustr[256];
  static char xml[4096];
  static char *xmlhdr = "<?xml version=\"1.0\" encoding=\"ISO-8859-1\" ?>\n";

  xml[0] = 0;
  strncat(xml,xmlhdr,len); len -= strlen(xml);
  strncat(xml,"<list>\n    <isodate>",len); len -= strlen(xml);
  create_isodate(iso,(time_t)ams->timestamp,ams->timestampUSec/1000);
  strncat(xml,iso,len); len -= strlen(xml);
  sprintf(ustr,"</isodate>\n    <calltype>%s</calltype>\n    <fac name=\"",type);
  strncat(xml,ustr,len); len -= strlen(xml);
  strncat(xml,xref->doocsFacility,CONTEXT_NAME_SIZE); len -= strlen(xml);
  strncat(xml,"\">\n        <dev name=\"",len); len -= strlen(xml);
  strncat(xml,xref->doocsDevice,EXPORT_NAME_SIZE); len -= strlen(xml);
  strncat(xml,"\">\n            <loc name=\"",len); len -= strlen(xml);
  strncat(xml,ams->device,DEVICE_NAME_SIZE); len -= strlen(xml);
  strncat(xml,"\">\n                <prop name=\"ERROR\">\n                    <data type=\"USTR\">",len); len -= strlen(xml);
  sprintf(ustr,"%d %d 0 \"%s\" \"%s\"",
    ams->alarmCode,
    tine2doocsSeverity(ams->severity),
    iso,GetWellFormedMessage(ams->alarmTag));
  strncat(xml,ustr,len); len -= strlen(xml);
  sprintf(ustr,"</data>\n                    <info severity=\"%s\" mask=\"%d\" newerror=\"%d\" online=\"1\"/>\n",
    severity_string(tine2doocsSeverity(ams->severity)),
    ams->alarmMask,
    ams->descriptor & almNEWALARM ? 1 : 0);
  strncat(xml,ustr,len); len -= strlen(xml);
	strncat(xml,"                </prop>\n            </loc>\n        </dev>\n    </fac>\n</list>",len);
  return xml;
}

char *GetDeviceSubSystem(char *srv)
{
  static char subsys[SUBSYSTEM_NAME_SIZE];

  /* need to ask the ENS and cache this somewhere */
  return subsys;
}

int sndalm2doocs(int i1,int i2,int i3,int i4)
{
  AMS ams;
  int almdata[4] = {1,2,3,4};
  strncpy(ams.server,"TESTSVR",EXPORT_NAME_SIZE);
  strncpy(ams.device,"TESTDEV",DEVICE_NAME_SIZE);
  strncpy(ams.alarmTag,"Test TINE 2 DOOCS alarm",ALARM_TAG_SIZE);
  ams.alarmCode = 512;
  ams.timestamp = time(NULL);
  ams.timestampUSec = 0;
  ams.starttime = ams.timestamp;
  ams.starttimeUSec = ams.timestampUSec;
  ams.alarmMask = 0;
  memcpy(ams.alarmData,&almdata,ALARM_DATA_SIZE);
  ams.alarmDataFormat = BFMT(CF_INT32);
  ams.alarmDataArraySize = 4;
  ams.severity = 3;
  ams.descriptor = almNEWALARM|almTEST;
  ams.alarmSystem = 100;
  return SendAlarmToDoocs(&ams,"run");
}
int SendHeartbeatToDoocs(char *server,int system)
{
  time_t t = time(NULL);
  DoocsAlmXref *xref = GetDoocsXRef(server);
  char device[DEVICE_NAME_SIZE];
  if (xref == NULL) return un_allocated;
  if (t < xref->interim + DOOCS_HEARTBEAT) return 0;
  xref->interim = t;
  sprintf(device,"%.32s._SVR",server); /* TODO: find the host name ? */
  return SendInfoToDoocs("alive",server,device,system);
}
int SendStartupToDoocs(char *server,int system)
{
  time_t t = time(NULL);
  DoocsAlmXref *xref = GetDoocsXRef(server);
  int i;
  if (xref == NULL) return un_allocated;
  for (i=0; i<xref->ndevices; i++)
  {
    SendInfoToDoocs("init",server,xref->devices[i].name,system);
  }
  return 0;
}
int SendInfoToDoocs(char *type,char *server,char *device,int system)
{
  AMS ams;
  char *tag;
  if (!strcmp(type,"info")) tag = "server is initializing"; else tag = "server is alive";
  strncpy(ams.server,server,EXPORT_NAME_SIZE);
  strncpy(ams.device,device,DEVICE_NAME_SIZE);
  strncpy(ams.alarmTag,tag,ALARM_TAG_SIZE);
  ams.alarmCode = 0;
  ams.timestamp = time(NULL);
  ams.timestampUSec = 0;
  ams.starttime = ams.timestamp;
  ams.starttimeUSec = ams.timestampUSec;
  ams.alarmMask = 0;
  memset(ams.alarmData,0,ALARM_DATA_SIZE);
  ams.alarmDataFormat = BFMT(CF_NULL);
  ams.alarmDataArraySize = 0;
  ams.severity = 0;
  ams.descriptor = almNEWALARM;
  ams.alarmSystem = system;
  return SendAlarmToDoocs(&ams,type);
}
int getDoocsXRefList(char *fn)
{
  int cc;
  if ((cc=csvReadFile(fecDbPath,fn,&csvDoocsAlmXrefDb,(void **)&doocsAlmXref)) != 0) ccerr(cc);
  nDoocsAlmXref = csvDoocsAlmXrefDb.siz;
err:
  return cc;
}
int dumpxref(int i1,int i2,int i3,int i4)
{
  int i;
  dbglog("monitoring %d servers for DOOCS",nDoocsAlmXref);
  for (i=0; i<nDoocsAlmXref; i++)
  {
    dbglog("/%.32s/%.32s [%.16s] -> %.32s/%.32s (%.32s): %d devices, %d total alarms sent (last heartbeat: %.24s",
      doocsAlmXref[i].tineContext,doocsAlmXref[i].tineServer,doocsAlmXref[i].tineSubsystem,doocsAlmXref[i].doocsAlarmServer,
      doocsAlmXref[i].doocsFacility,doocsAlmXref[i].doocsDevice,doocsAlmXref[i].ndevices,doocsAlmXref[i].nsent,
      ctime(&doocsAlmXref[i].interim));
  }
  return 0;
}
NAME64 devbuffer[1024];
int initCas2Doocs(void)
{
  int i, k, n, cc=0;
  char dev[256];
  if ((cc=getDoocsXRefList(CAS2DOOCS_FILE)) != 0)
  {
    feclog("could not read %s : %s",CAS2DOOCS_FILE,erlst[cc]);
    goto err;
  }
  for (i=0; i<nDoocsAlmXref; i++)
  {
    n = 1024;
    sprintf(dev,"/%.32s/%.32s",doocsAlmXref[i].tineContext,doocsAlmXref[i].tineServer);
    if ((cc=GetSystemDevices(dev,NULL,devbuffer,&n)) == 0 || cc & CE_SENDDATA)
    {
      if ((doocsAlmXref[i].devices=(ADevInfo *)calloc(n,sizeof(ADevInfo))) != NULL)
      {
        doocsAlmXref[i].ndevices = n;
        for (k=0; k<n; k++) memcpy(doocsAlmXref[i].devices[k].name,devbuffer[k].name,64);
        feclog("acquired %d devices for server %s",n,dev);
      }
    }
    else
    {
      feclog("acquired to acquire devices for server %s : %s",dev,erlst[cc&0xff]);
    }
    if (strlen(doocsAlmXref[i].doocsAlarmServer) == 0)
    { /* not given: make a good guess ... */
      sprintf(doocsAlmXref[i].doocsAlarmServer,"/%.32s/ALARM_SERVER",doocsAlmXref[i].doocsFacility);
    }
    else if (doocsAlmXref[i].doocsAlarmServer[0] != '/')
    { /* make sure this begins with the leading slash */
      strncpy(dev,doocsAlmXref[i].doocsAlarmServer,64); dev[64] = 0;
      sprintf(doocsAlmXref[i].doocsAlarmServer,"/%.64s",dev);
    }
  }
  
  RegisterUserCommand("doocsstats",dumpxref,NULL,NULL,CA_READ);
err:
  return cc;
}
int getAlarmDevice(DoocsAlmXref *xref,AMS *ams)
{ /* TODO: hash this */
  int i;
  for (i=0; i<xref->ndevices; i++)
  {
    if (strnicmp(xref->devices[i].name,ams->device,DEVICE_NAME_SIZE)) continue;
    return i;
  }
  return -1;
}
DoocsAlmXref *GetDoocsXRef(char *server)
{
  int i;
  if (server == NULL) return NULL;
  for (i=0; i<nDoocsAlmXref; i++)
  {
    if (strnicmp(server,doocsAlmXref[i].tineServer,EXPORT_NAME_SIZE)) continue;
    return &doocsAlmXref[i];
  }
  return NULL;
}
void cas2doocsCallback(int id,int cc,void *ref)
{
  DoocsAlmXref *xref = (DoocsAlmXref *)ref;
  char dev[128];
  char errstr[256],str[512];
  static char *prp = "ERRORXML";
  if (IsServerRunning()) SetSystemUserName(GetExportedFecName());
  if (xref == NULL) return;
  sprintf(dev,"%s/%s",xref->doocsAlarmServer,xref->doocsFacility);
  if (cc != 0)
  {
    sprintf(str,"Cas2Doocs Link to %s %s: %s",dev,prp,GetLastLinkError((short)cc,errstr));
    dbglog(str);
    logproblem(str);
    return;
  }
  if (tineDebug) dbglog("Cas2Doocs to %s %s: success",dev,prp);
}
int SendAlarmToDoocs(AMS *ams,char *type)
{
  int lnkID = -1, d=-1, suppress=FALSE;
  DTYPE din;
  char dev[128],str[256];
  char *xml;
  static char *prp = "ERRORXML";
  DoocsAlmXref *xref = GetDoocsXRef(ams->server);

  if (xref == NULL) return un_allocated;
  if ((d=getAlarmDevice(xref,ams)) < 0) return illegal_device;
  if (ams->descriptor & almNEWALARM)
  {
    suppress = FALSE;
  }
  else if (xref->devices[d].code == 0)
  { /* mayber a CAS re-start ? */
    suppress = FALSE;
  }
  else
  {
    suppress = TRUE;
  }
  if (suppress) return already_assigned;
  xref->devices[d].code = ams->alarmCode; /* the last assigned alarm code */
  xref->devices[d].descriptor = ams->descriptor; /* the last assigned alarm descriptor */

  SetSystemUserName("DOOCSADM");
  sprintf(dev,"%s/%s",xref->doocsAlarmServer,xref->doocsFacility);

  xml = GetAlarmAsXML(ams,xref,type);
  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_XML;
  din.dArrayLength = strlen(xml);
  din.data.cptr = xml;
 
  xref->status = -1;
  lnkID = AttachLinkEx2(dev,prp,NULL,&din,CA_WRITE,100,cas2doocsCallback,CM_SINGLE,(int)xref,xref);
  if (lnkID < 0)
  {
    xref->status = -lnkID;
    sprintf(str,"Cas2Doocs Link to %s %s : %s",dev,prp,erlst[-lnkID]);
    feclog(str);
    logproblem(str);
  }
  if (IsServerRunning()) SetSystemUserName(GetExportedFecName());
  return lnkID < 0 ? -lnkID : 0;
}
