/*
 * CASEQM.h    REVISION HISTORY:
 * Generated by SRVWIZARD on Fri Jan 26 18:01:41 2001
*/

#ifndef CASEQM_H_
#define CASEQM_H_

#ifndef ccerr
# define ccerr(i) { cc = i; goto err; }
#endif

#define NOT_RESPONDING 999
#define TAG_NOT_RESPONDING "Not Responding"
#define TEXT_NOT_RESPONDING "The Front End Controller is not accessible"

#define CYCLE_NUMBER_STALE 998
#define TAG_CYCLE_NUMBER_STALE "Cycle Number not incrementing"
#define TEXT_CYCLE_NUMBER_STALE "The Cycler for this context is not sending cycle numbers"

#define INVALID_EPOCH 997
#define TAG_INVALID_EPOCH "Host Time/Date is invalid"
#define TEXT_INVALID_EPOCH "Host computer is showing a time and date from an invalid epoch"

#define IsCASGenerated(c) ((c) == NOT_RESPONDING || (c) == CYCLE_NUMBER_STALE || (c) == INVALID_EPOCH)

#define SIXMONTHS (86400 * 183)

#if defined(MSDOS) || defined(WIN32)
# define DEFAULT_ARCHIVE_PATH "..\\CACHE"
# define SEPARATER "\\"
#else
# define SEPARATER "/"
# define DEFAULT_ARCHIVE_PATH "../CACHE"
#endif

#ifndef INTEL
# define PrepOutgoingAMRB(rb,buf) {\
   SwapBytes((char *)rb,4,CF_LONG,LTOR);\
   memcpy(buf,rb,AMRB_SIZE);\
  }
# define PrepIncomingAMS(amsstruct,buf,n)  {\
   _StructSwapEx(AMSStructStruct,(unsigned char *)buf,(unsigned char *)amsstruct,RTOL,n,sizeof(AMS));\
  }
#else
# define PrepOutgoingAMRB(rb,buf) memcpy(buf,rb,AMRB_SIZE)
# define PrepIncomingAMS(amsstruct,buf,n) {if ((void *)buf != (void *)amsstruct) memcpy(amsstruct,buf,n*AMS_SIZE); }
#endif

#define CASEQM_TAG "CASEQM"

#define CAS_STS_REMOVE -1
#define CAS_STS_IGNORE_NEVER 0
#define CAS_STS_IGNORE_ALWAYS 1
#define CAS_STS_IGNORE_NOTRUNNING 2

#define CHK_TGT  0
#define CHK_HIGH 1
#define CHK_LOW  2

typedef struct
{
  char dev[256];
  char key[PROPERTY_NAME_SIZE];
  BYTE dat[64];
  BYTE cmp[64];
  int fmt;
  int siz;
  int chk;
  int lnkId;
} RunningCriteria;
extern int nRunningCriteria;
extern RunningCriteria *runningCriteria;

#define RUNNING_CRITERIA_FILE "runningCriteria.csv"

#define CAS_ACQ_RATE 500
#define CASEQM_RATE 1000
#define TIMEOUT_THRESHOLD 10
#define MAX_ALARMSNAPSHOT_SIZE 256

#define PRP_SYSALARMS 2
#define PRP_SYSALARMS_SIZE 1000
#define PRP_NUMALARMS 1
#define PRP_NUMALARMS_SIZE 1
#define PRP_SUBSYSTEMS 3
#define PRP_SERVERS 4
#define PRP_SERVERINFO 5
#define PRP_SERVERINFO_SIZE (16*4 + 4*5)
#define PRP_ACKTIME 6
#define PRP_ACKTIME_SIZE 1
#define PRP_ALARMSYSTEMS 7
#define PRP_ALARMSYSTEMS_SIZE 1000
#define PRP_REMOVEALARMS 8
#define PRP_RAWSYSALARMS 9
#define PRP_RAWNUMALARMS 10
#define PRP_SYSALMDEFS 11
#define PRP_SYSALARMSEXT 12
#define PRP_SYSTEMLIST 13
#define PRP_SERVERLIST 14
#define PRP_ISONLINE 15
#define PRP_DISABLE 16
#define PRP_ENABLE 17
#define PRP_SERVERDB 18
#define PRP_ALMSYSDB 19
#define PRP_ACTIONDB 20
#define PRP_RESTART 21
#define PRP_ALARMSYSTEMS_ALL 22
#define PRP_NUMALARMS_ALL 23
#define PRP_RAWNUMALARMS_ALL 24
#define PRP_SYSALARMS_ALL 25
#define PRP_RAWSYSALARMS_ALL 26
#define PRP_DISABLECMT 27
#define PRP_DISABLE_ACTIVE 28
#define PRP_DISABLE_ISACTIVE 29
#define PRP_BASETIME 30
#define PRP_NUMALMDEFS 31
#define PRP_NREGIONRULES 32
#define PRP_REGIONRULES 33
#define PRP_REGIONSDB 34
#define PRP_REGIONSXREF 35
#define PRP_CONTEXTS 36

#define ALMSYSTEM_INFRASTRUCTURE 5000
#define ALMSYSTEM_INFO   3
#define ALMSYSTEM_WARN   7
#define ALMSYSTEM_FATAL 12

#define XTN_NAME_SIZE 16
#define FILE_NAME_SIZE 32

typedef struct CASActionStruct
{
  char email[128];
  char actionSignalString[128];
  char devName[128];
  char devProperty[PROPERTY_NAME_SIZE];
  DTYPE din;
  DTYPE dout;
  int access;
  int actionSignalFormat;
  int alarmSystemForward;
  int sendAnnotation;
  UINT32 alarmCode;
  AMS *ams;
  int status;
  int casTblIdx;
  int retry;
  struct CASActionStruct *nxt;
} CasActLstEntry;

typedef struct
{
  char email[128];
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  char actionSignalString[128];
  int actionSignalFormat;
  int alarmCode;
} CasActionItem;
#define CASACTIONITEM_TAG "CasAct"
#define _ANNOTATE_ "ANNOTATE"

#define ALM_DISABLE_PERMANENT 0
#define ALM_DISABLE_CURRENT_ACTIVE 1
typedef struct CASDisabledStruct
{
  UINT32 alarmCode;
  int    disableType;
  time_t timeDisabled;
  char   device[DEVICE_NAME_SIZE];
  char   disabler[USERNAME_SIZE];
  struct CASDisabledStruct *nxt;
} CasDisableLstEntry;

typedef struct CASAlarmMsgStruct
{
  time_t starttime;
  time_t stoptime;
  int nalms;
  AMS *ams;
  struct CASAlarmMsgStruct *nxt;
} CAMS;

typedef struct
{
  UINT32 nalm;  /* number of alarms */
  UINT32 mrt;   /* most recent timestamp */
  UINT32 hsv;   /* highest severity */
  UINT32 nmrt;  /* number at most recent timestamp */
  UINT32 nhsv;  /* number at higest severity */
  UINT32 ndef;  /* number of alarm definitions */
} ALM_SNAPSHOT;

typedef struct
{
  char devContext[CONTEXT_NAME_SIZE];
  char devServer[EXPORT_NAME_SIZE];
  char subSystem[SUBSYSTEM_NAME_SIZE];
  char arcExt[XTN_NAME_SIZE];
  int SeverityLevel;
  int ArchiveLevel;
  int Retention;
  int AlarmLevel;
  int AlarmSystem;
  CasActLstEntry *actionList;
  CasDisableLstEntry *disabledList;
  int sizDisabledList;
  int linkId;
  int linkStatus;
  int linkTimeoutCounter;
  int linkRetryCounter;
  int markedAsDown;
  int nAlarmsLastUpdate;
  ALM_SNAPSHOT thisSnapshot;
  ALM_SNAPSHOT prevSnapshot;
  ALM_SNAPSHOT acqSnapshot;
  CAMS *almList;
  AMS *almBuffer;
  ADS *adsList;
  int *almSystemList;
  int nAlmList;
  int nAdsList;
  int nAlmSystemList;
  int adsListAcquired;
  int adsReacquire;
  int adsKeepStaticInfo;
  time_t timestamp;
  time_t timestampDisabled;
  int pending;
  int offline;
  int legacy;
  int epochInvalid;
  int cycleStale;
  int hasAcquiredInitialAlarmSet;
} CasTblEntry;

#define IsOffLine(cte) ((cte)->offline && (cte)->offline < CAS_STS_IGNORE_NOTRUNNING)

extern CasTblEntry *casTbl;
extern int casTblSiz;
extern int casTblCapacity;
#define CAS_CAPACITY_BUFFER 100

typedef struct
{
  char devContext[CONTEXT_NAME_SHORTSIZE];
  char devServer[EXPORT_NAME_SHORTSIZE];
  char subSystem[SUBSYSTEM_NAME_SIZE];
  char arcExt[XTN_NAME_SIZE];
  int SeverityLevel;
  int ArchiveLevel;
  int Retention;
  int AlarmLevel;
  int AlarmSystem;
} CASQS; /* CAS QUERY STRUCTURE (legacy) */

typedef struct
{
  char devContext[CONTEXT_NAME_SIZE];
  char devServer[EXPORT_NAME_SIZE];
  char subSystem[SUBSYSTEM_NAME_SIZE];
  char arcExt[XTN_NAME_SIZE];
  int SeverityLevel;
  int ArchiveLevel;
  int Retention;
  int AlarmLevel;
  int AlarmSystem;
} CQS; /* CAS QUERY STRUCTURE (modern) */

typedef struct
{
  char devContext[CONTEXT_NAME_SIZE];
  char devServer[EXPORT_NAME_SIZE];
  char subSystem[SUBSYSTEM_NAME_SIZE];
  char arcExt[XTN_NAME_SIZE];
  int SeverityLevel;
  int ArchiveLevel;
  int Retention;
  int AlarmLevel;
  int AlarmSystem;
  int offline;
} CasDbItem; /* CAS Database Item */
#define CASDBITEM_TAG "CasDb"

typedef struct
{
  int number;
  int column;
  int row;
  char almsystem[ALARM_TAG_SIZE];
  char subsystem[SUBSYSTEM_NAME_SIZE];
} CasAlmSysTblEntry;
#define CASALMSYSITEM_TAG "CasAlSys"
extern int nAlarmSystems;
extern CasAlmSysTblEntry *AlarmSystems;

typedef struct XAlarmMsgStruct5
{
  char server[EXPORT_NAME_SHORTSIZE];
  char device[DEVICE_NAME_SHORTSIZE];
  char alarmText[40];
  UINT32 alarmCode;
  UINT32 timestamp;
  UINT32 alarmMask;
  UINT32 alarmSystem;
  BYTE alarmData[8];
  BYTE alarmDataFormat;
  BYTE alarmDataArraySize;
  BYTE severity;
  BYTE descriptor;
} XAMS5; /* legacy eXtended AMS sent out upon request */
#define XAMS5_SIZE  100

typedef struct XAlarmMsgStruct
{
  char server[EXPORT_NAME_SIZE];
  char device[DEVICE_NAME_SIZE];
  char alarmText[ALARM_TEXT_SIZE];
  UINT32 alarmCode;
  UINT32 timestamp;
  UINT32 timestampUSec;
  UINT32 starttime;
  UINT32 starttimeUSec;
  UINT32 alarmMask;
  UINT32 alarmSystem;
  BYTE alarmData[ALARM_DATA_SIZE];
  BYTE alarmDataFormat;
  BYTE alarmDataArraySize;
  BYTE severity;
  BYTE descriptor;
} XAMS; /* eXtended AMS sent out upon request : uses alarmText rather than alarmTag */
#define XAMS_SIZE (EXPORT_NAME_SIZE + DEVICE_NAME_SIZE + ALARM_TEXT_SIZE + 7*4 + ALARM_DATA_SIZE + 4)  /* 256 */

#define ALARM_SYSTEM_FILE "AlarmCodes.csv"
#define CAS_LIST_FILE "ServerList.csv"

extern int nAlarmSubSystems;
extern NAME16 *AlarmSubSystems;

extern AMS g_sysalarmsBuffer[PRP_SYSALARMS_SIZE];
extern short g_numalarms;

int removeDisabledList(CasTblEntry *cl);
void TerminateDanglingAlarms(int i);
void TakeAction(CasActLstEntry *ca);
int caseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
void caseqm_bkg(void);
void caseqm_ini(void);
void caseqm_exi(void);
int logproblem(char *text,...);

void sendAlarmAsMail(char *addr,char *subj,char *body);
void GenerateAndForwardAlarm(int i,AMS *ams,int almsys);

THREAD_TYPE mtSendAlarmAsMail(THREAD_PARAM lpvThreadParam);

#endif

