/*
 * clseqm.c    REVISION HISTORY:
 * Generated by SRVWIZARD on Fri Sep 14 13:43:20 2007
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#if defined(UNIX)
# include "dirent.h"
# define CLSPATH "../LOGS"
# define FSTOK "/"
#elif defined(WIN32)
# include "direct.h"
# define CLSPATH "..\\LOGS"
# define FSTOK "\\"
#endif
#include "clseqm.h"

#if defined(DOS)
# define O_createmode O_RDWR|O_CREAT|O_APPEND|O_BINARY
# define O_writemode O_RDWR|O_BINARY
# define O_readmode O_RDONLY|O_BINARY
# define O_mode O_RDONLY|O_BINARY
# define FO_mode "rt"
#elif defined(WIN32)
# define O_createmode O_RDWR|O_CREAT|O_APPEND|O_BINARY,_S_IREAD|_S_IWRITE
# define O_writemode O_RDWR|O_BINARY,_S_IREAD|_S_IWRITE
# define O_readmode O_RDONLY|O_BINARY
# define O_mode O_RDONLY|O_BINARY
# define FO_mode "rt"
#elif defined(VMS)
# define O_createmode O_RDWR|O_CREAT|O_APPEND
# define O_writemode O_RDWR
# define O_readmode O_RDONLY
# define O_mode O_RDONLY
# define FO_mode "r"
#else
# define O_createmode O_RDWR|O_CREAT|O_APPEND,S_USERWRITE
# define O_writemode O_RDWR,S_USERWRITE
# define O_readmode O_RDONLY,S_USERWRITE
# define O_mode O_RDONLY
# define FO_mode "r"
#endif

/* from hstlib.h */
extern int nDaysInMonthTable[2][12];
#define nDaysInMonth(yy,mm) (((yy) % 4) ? nDaysInMonthTable[0][mm] : nDaysInMonthTable[1][mm])

#define CLOGSIZE sizeof(CLOG)
#define MakeLogFileName(nm,yy,mm,dd) (\
  sprintf((nm),"cls%02d%02d%02d.log",((yy)%100),(mm),(dd))\
)

int makeDirPath(char *fname)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,FSTOK); c != NULL; c=strtok(NULL,FSTOK))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return 0;
    }
    strcat(dirname,FSTOK);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return 0;
    }
#   elif defined(WIN32)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
#   endif
  }
  return -1;
}
int MakeLogFullName(char *fname,struct tm *lt)
{
  char fileName[64];
  MakeLogFileName(fileName,lt->tm_year,lt->tm_mon+1,lt->tm_mday);
  sprintf(fname,"%s%s%04d%s%02d%s%s",
    CLSPATH,FSTOK,lt->tm_year+1900,FSTOK,lt->tm_mon+1,FSTOK,fileName);
  return makeDirPath(fname);
}
int fltrcmp(char *tgt,char *ftr)
{ /* return 0 if matched */
  char beginToken[64], endToken[64], midToken[64];
  int blen = 0, elen = 0, mlen = 0;
  if (ftr == NULL || strlen(ftr) == 0) return 0;

  if (strchr(ftr,'*') != NULL)
  {
    getWildcardTokens(ftr,beginToken,endToken,midToken);
    blen = (int)strlen(beginToken);
    mlen = (int)strlen(midToken);
    elen = (int)strlen(endToken);
    return wildcardMatches(tgt,beginToken,blen,endToken,elen,midToken,mlen) ? 0 : 1;
  }
  return strnicmp(tgt,ftr,32);
}
int getLogEntries(CLOG *clsBuf,int size,CLOGF *filters, int nfilters)
{
  int hFile, i, n=0, pos = 0, cc = 0, fsize, passed, logend = FALSE, singlepass = FALSE;
  char fileName[128];
  CLOG clog;
  CLOGF lclfltr, *fltr;
  time_t tnow = time(NULL), t = tnow;
  struct tm *lt, it;

  if (clsBuf == NULL || size < 1) ccerr(argument_list_error);
  if (filters == NULL || nfilters < 1)
  { /* no filter => return the last 'size' entries from today */
    lt = localtime(&t);
    lt->tm_hour = 0;
    lt->tm_min = 0;
    lt->tm_sec = 0;
    memset(&lclfltr,0,sizeof(CLOGF));
    lclfltr.timeStart = (UINT32)mktime(lt);
    lclfltr.timeStop = (UINT32)t;
    fltr = &lclfltr;
    nfilters = 1;
    singlepass = TRUE;
  }
  else
  {
    fltr = filters;
  }
  t = (time_t)fltr->timeStart;
  lt = localtime(&t);
  if (MakeLogFullName(fileName,lt)) ccerr(file_error);
  if ((hFile=open(fileName,O_readmode)) != -1)
  { /* start time from filter could land somewhere where there are no files */
    if (filters == NULL)
    { /* no filter => return the last 'size' entries from today */
      fsize = lseek(hFile,0,SEEK_END);
      n = fsize/CLOGSIZE;
      if (n > size) pos = (n-size)*CLOGSIZE;
      n = 0;
    }
    lseek(hFile,pos,SEEK_SET);
  }
  while (!logend)
  { /* could span multiple files */
    if (singlepass) logend = TRUE;
    while(hFile != -1 && read(hFile,(char *)&clog,CLOGSIZE) == CLOGSIZE)
    { /* read through file */
      for (i=0, passed = FALSE; i<nfilters && !passed; i++)
      { /* run thru all filters */
        if (clog.timeLogged >= tnow) logend = TRUE;
        if (clog.timeLogged < (double)fltr[i].timeStart) continue;
        if (clog.timeLogged > (double)fltr[i].timeStop) continue;
        if (strlen(fltr[i].context) && strnicmp(clog.context,fltr[i].context,32)) continue;
        if (fltrcmp(clog.tag,fltr[i].tag)) continue;
        if (fltrcmp(clog.logger,fltr[i].logger)) continue;
        if (fltrcmp(clog.text,fltr[i].text)) continue;
        if (clog.priority < fltr[i].priority) continue;
        if (clog.status < fltr[i].status) continue;
        passed = TRUE;
      }
      if (!passed) continue;
      memcpy(&clsBuf[n],&clog,CLOGSIZE);
      if (++n >= size) break;
    }
    if (logend) ccerr(0);
    /* keep scanning: next file */
    it = *lt;
    it.tm_mday = (it.tm_mday+1)%(nDaysInMonth(it.tm_year,it.tm_mon)+1);
    if (it.tm_mday == 0)
    {
      it.tm_mday++;
      it.tm_mon = (it.tm_mon+1)%12;
      if (it.tm_mon == 0) it.tm_year++;
    }
    it.tm_hour = 0; it.tm_min = 0; it.tm_sec = 0;
    t = mktime(&it);
    lt = localtime((time_t *)&t);
    if (hFile != -1) close(hFile);
    if (t >= tnow) logend = TRUE;
    if (MakeLogFullName(fileName,lt)) ccerr(file_error);
    if ((hFile=open(fileName,O_readmode)) == -1) continue;
  }

err:
  if (hFile != -1) close(hFile);
  if (cc)
  {
    feclog("getLogEntries : %s",erlst[cc]);
    return -cc;
  }
  return n;
}
void dumpEntry(CLOG *entry)
{
  dbglog("context : %s\n",entry->context);
  dbglog("tag : %s\n",entry->tag);
  dbglog("logger : %s\n",entry->logger);
  dbglog("text : %s\n",entry->text);
  dbglog("priority : %d\n",entry->priority);
  dbglog("status : %d\n",entry->status);
  dbglog("time sent : %s\n",getDataTimeString(entry->timeSent,0));
  dbglog("time logged : %s\n",getDataTimeString(entry->timeLogged,0));
}
int logEntry(CLOG *entry)
{
  int hFile;
  static char fileName[128];
  time_t t = time(NULL);
  struct tm *lt=localtime(&t);
  static int mday = -1;

  if (entry == NULL) return argument_list_error;
  if (mday != lt->tm_mday)
  {
    MakeLogFullName(fileName,lt);
    mday = lt->tm_mday;
  }
  if ((hFile=open(fileName,O_createmode)) == -1) return file_error;
  lseek(hFile,0L,SEEK_END);
  write(hFile,(char *)entry,sizeof(CLOG));
  close(hFile);
  if (tineDebug) dumpEntry(entry);
  return 0;
}
CLOG g_entriesBuffer[PRP_ENTRIES_SIZE];

CLOG g_newentry;

int RegisterClogFilterStructure(void)
{
  int cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,context),32,CF_TEXT,"context")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,tag),32,CF_TEXT,"tag")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,logger),32,CF_TEXT,"logger")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,text),256,CF_TEXT,"text")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,priority),1,CF_LONG,"priority")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,status),1,CF_LONG,"status")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,timeStart),1,CF_LONG,"timeStart")) != 0) return cc;
  if ((cc=addFieldToStruct("CLOGF",OFFSETIN(CLOGF,timeStop),1,CF_LONG,"timeStop")) != 0) return cc;
  if ((cc=sealTaggedStruct("CLOGF",sizeof(CLOG),1)) != 0) return cc;
  return 0;
}
int RegisterClseqmProperties(void)
{
  int cc = 0;
  DTYPE dout,din;

  dout.dArrayLength = PRP_ENTRIES_SIZE;
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"CLOG");
  din.dArrayLength = 0;
  din.dFormat = CF_NULL;
  if ((cc=RegisterPropertyInformation(CLSEQM_TAG,"Entries",&dout,&din,CA_READ,AT_UNKNOWN,0,"[0:1 ]returns the last (datasize) Entries",PRP_ENTRIES,"")) < 0) goto err;

  dout.dArrayLength = PRP_NEWENTRY_SIZE;
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"CLOG");
  din.dArrayLength = PRP_NEWENTRY_SIZE;
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"CLOG");
  if ((cc=RegisterPropertyInformation(CLSEQM_TAG,"NewEntry",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"[0:1 ]New Log Entry",PRP_NEWENTRY,"")) < 0) goto err;

err:
  if (cc < 0)
  {
    printf("Can't register property : %s\n>",erlst[-cc]);
  }
  return cc;
}

void clseqm_bkg(void)
{
}

void clseqm_ini(void)
{
  RegisterClseqmProperties();
  RegisterDeviceName(CLSEQM_TAG,"Logger",0);
  RegisterClogFilterStructure();
}

void clseqm_exi(void)
{
  /* TODO: put your shutdown routines here */
}

int clseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,n;
  CLOG l_newentry;
  CLOGF *clsFilter = NULL;

  devnr = GetDeviceNumberEx(CLSEQM_TAG,devName,devProperty);
  prpid = GetPropertyId(CLSEQM_TAG,devProperty);

  switch (prpid)
  {
    case PRP_ENTRIES:
        if (access&CA_WRITE) return illegal_read_write;
        if ((n=din->dArrayLength) > 0)
        {
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strcmp(din->dTag,"CLOGF")) return invalid_structure_tag;
          clsFilter = (CLOGF *)din->data.vptr;
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_ENTRIES_SIZE) return dimension_error;
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strcmp(dout->dTag,"CLOG")) return invalid_structure_tag;
          n = getLogEntries((CLOG *)dout->data.vptr,dout->dArrayLength,clsFilter,n);
          if (n < 0) return -n;
          dout->dArrayLength = n;
        }
        return 0;
    case PRP_NEWENTRY:
        if (access&CA_WRITE)
        {
          if (din->dArrayLength > 0)
          {
            if (din->dArrayLength > PRP_NEWENTRY_SIZE) return dimension_error;
            if (din->dFormat != CF_STRUCT) return illegal_format;
            if (strcmp(din->dTag,"CLOG")) return invalid_structure_tag;
            memcpy(&l_newentry,din->data.vptr,din->dArrayLength*sizeof(CLOG));
            l_newentry.timeLogged = makeDataTimeStamp();
            logEntry(&l_newentry);
            g_newentry = l_newentry;
          }
        }
        if (dout->dArrayLength > 0)
        { /* return last entry */
          if (dout->dArrayLength > PRP_NEWENTRY_SIZE) return dimension_error;
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strcmp(dout->dTag,"CLOG")) return invalid_structure_tag;
          memcpy(dout->data.vptr,&g_newentry,sizeof(CLOG));
        }
        return 0;
    default:
        return illegal_property;
  }
}
