/* ============   EVNEQM.C    ============================================


   An archived file contains the following information according the
	 following protocol:

       DeviceServerName - 16 bytes   char string
       Property         - 32 bytes   char string
       DeviceName       - 16 bytes   char string
       size of call     - 2  bytes   short
       data format      - 2  bytes   short
       scale            - 4  bytes   float
       shift            - 4  bytes   float
       data buffer - N bytes according to format and size


  The above is repeated N times ....

  ========================================================================  */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "tine.h"
#include "arctools.h"
#include "apceqm.h"
#include "arceqm.h"

#define EVNEQM_TAG "EVNEQM"
#define EVNEQM_EXPORT "EVENTMGR"

#define APCEQM_TAG "APCEQM"
#define APCEQM_EXPORT "EVENT.HISTORY"

extern int numArchiveTypes;
extern int arbTblSiz;
int makeFullName(char *fname,char *ext);
int getContributingArchiveServers(int event,NAME16 *svrs,int *len);
int getContributingDeviceServers(int event,NAME16 *svrs,int *len);
int evneqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void evninit(void);
char *gExt = "EVENT";

#define dumpStructInfo(tag) {\
  if ((s=getStructFromTag(tag)) != NULL) {\
    printf("%s: num %d size %d rawlen %d\n",tag,s->num,s->size,s->rawlen);\
  } else printf("%s: not found\n",tag);\
}
int writeCommentToFile(char *fn,char *comment)
{
  FILE *fp = NULL;
  char *c,fullName[128],ext[16];

  if (comment == NULL) return -1;
  strncpy(fullName,fn,128);
  sprintf(ext,".%s",gExt);
  if ((c=strstr(fullName,ext)) == NULL) return -1;
  *c = 0;
  strcat(fullName,".CMT");
  if ((fp=fopen(fullName,"w")) == NULL) return -1;
  if (fwrite(comment,strlen(comment),1,fp) != 1) return -1;
  return 0;
}
#define pmHeaderSize 76
int writeEventHeaderToFile(FILE *fp,int id,int eventCode)
{
  BYTE s[512];
  short siz;
  short fmt;
  int dsiz;
  float shift = 0.0, scale = 1.0;
  if (id < 0 || id > numArchiveTypes) return -1;
  siz = 16; fmt = CF_TEXT; dsiz = siz*fmtsizeof(fmt);
  memset(s,0,128);
  memcpy(s,"EventServer",16);
  memcpy(&s[16],"TriggerSource",32);
  memcpy(&s[48],"#0",16);
  memcpy(&s[64],&siz,2);
  memcpy(&s[66],&fmt,2);
  memcpy(&s[68],&scale,4);
  memcpy(&s[72],&shift,4);
  memcpy(&s[76],ArchRecord[id].trg,dsiz);
  if (fwrite(s,pmHeaderSize+dsiz,1,fp) != 1) return -1;
  siz = 1; fmt = CF_LONG; dsiz = siz*fmtsizeof(fmt);
  memset(&s[16],0,32); 
  memcpy(&s[16],"EventCode",32);
  memcpy(&s[64],&siz,2);
  memcpy(&s[66],&fmt,2);
  memcpy(&s[76],&eventCode,dsiz);
  if (fwrite(s,pmHeaderSize+dsiz,1,fp) != 1) return -1;
  return 0;
}
int appendEventToFile(FILE *fp,char *strArchiveServer,char *strDeviceServer)
{
  BYTE s[128];
  short siz;
  short fmt;
  int dsiz;
  float shift = 0.0, scale = 1.0;
  siz = 16; fmt = CF_TEXT; dsiz = siz*fmtsizeof(fmt);
  memset(s,0,128);
  memcpy(s,strArchiveServer,16);
  memcpy(&s[16],"EventRecord",32);
  memcpy(&s[48],"#0",16);
  memcpy(&s[64],&siz,2);
  memcpy(&s[66],&fmt,2);
  memcpy(&s[68],&scale,4);
  memcpy(&s[72],&shift,4);
  strncpy((char *)&s[76],strDeviceServer,siz);
  if (fwrite(s,pmHeaderSize+dsiz,1,fp) != 1) return -1;
  return 0;
}

int eventExists(int event)
{
  FILE *fp = NULL;
  char fullName[128];

  sprintf(fullName,"%x.%s",event,gExt);
  if (makeFullName(fullName,gExt) == 0) 
  {
    if ((fp=fopen(fullName,"rb")) == NULL) return 0;
    fclose(fp);
    return -1;
  }
  return 0;
}
int ArchiveEventSet(int idSource, int eventCode, char *comment)
{
  char fullName[128];
  time_t interimEvent = time(NULL);
  static time_t lastEvent = 0;
  FILE *fp = NULL;
  int cc=0;

  if (interimEvent < lastEvent + 5) interimEvent = lastEvent;
  lastEvent = interimEvent;

  if (!eventExists(interimEvent))
  {
    sprintf(fullName,"%x.%s",(unsigned int)interimEvent,gExt);
    if (makeFullName(fullName,gExt) != 0) ccerr(file_error);
    if ((fp=fopen(fullName,"wb")) == NULL && 
        (fp=createAndOpen(fullName,"w")) == NULL) ccerr(file_error);
    // put the first enry into the file ...
    writeEventHeaderToFile(fp,idSource,eventCode);
    writeCommentToFile(fullName,comment);
    fclose(fp);
  }
err:
  if (cc)  feclog("cannot open event file for event %d",interimEvent);
  return interimEvent;
}
char thisFecName[FEC_NAME_SIZE] = "";
char thisContext[CONTEXT_NAME_SIZE] = "SITE";
char *thisSubSystem = "SER";
char thisFecDescription[FEC_DESC_SIZE] = "SITE Event Server";
char thisLocation[FEC_LOCATION_SIZE] = "";
UINT16 thisPortOffset = 0;
void PreSystemInit(void)
{
  int p;
  if (strlen(thisFecName) == 0)
  {
    strncpy(thisFecName,"EVENTS",FEC_NAME_SIZE);
  }
  if (thisPortOffset == 0)
  {
    p = GetPortOffset(thisFecName);
    if (p >= 0) thisPortOffset = p;
  }
  RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  return;
}
void PostSystemInit(void)
{
  registerStructs();
  RegisterEquipmentModule(EVNEQM_EXPORT,EVNEQM_TAG,0,evneqm,evninit,NULL,100,NULL);
  RegisterEquipmentModule(APCEQM_EXPORT,APCEQM_TAG,0,apceqm,apcinit,apctask,100,NULL);
}
int SetGlobalEvent(ulong eventCode,char *cmt)
{
  // just a dummy routine so that everything links ...
  return 0;
}
int RecordEvent(ulong eventNumber,ulong eventCode,char *devName)
{
  // just a dummy routine so that everything links ...
  return 0;
}
int idxSelectedARB = 0;
int GetARBIndex(char *srvName)
{
  int i,j;
  if (srvName == NULL || strlen(srvName) == 0) return -1;
  // try srvName = archive context first:
  i = getSetNumber(srvName);
  if (i >= 0 && i < arbTblSiz) return i;

  // try the selected index
  for (j=0; j<arbTbl[idxSelectedARB].nsteps; j++)
  {
    if (!strncmp(srvName,arbTbl[idxSelectedARB].das[j].srvArc,16))
      return idxSelectedARB;
  }
  // didn't match up, so try everything else
  for (i=0; i<arbTblSiz; i++) for (j=0; j<arbTbl[i].nsteps; j++)
  {
    if (!strncmp(srvName,arbTbl[i].das[j].srvArc,16) && arbTbl[i].das[j].acc&READ)
    {
      return i;
    }
  }
  // wasn't there
  return -1;
}

int ArchiveEventRecord(long eventNumber,char *strArchiveServer,char *strDeviceServer)
{
  int cc = 0;
  char fullName[128];
  FILE *fp = NULL;

  feclog("received Record for event %d from device %s",eventNumber,strDeviceServer);

  if (!eventExists(eventNumber)) 
  {
    feclog("Attemp to record %s to non-existent event by %s",
      strDeviceServer,strArchiveServer);
    return -1;
  }
  sprintf(fullName,"%x.%s",(unsigned int)eventNumber,gExt);
  if (makeFullName(fullName,gExt) == 0 && (fp=fopen(fullName,"ab")) != NULL) 
  {
    // append the event to the event file ...
    appendEventToFile(fp,strArchiveServer,strDeviceServer);
    fclose(fp);
    return 0;
  }
  return file_error;
}
void evninit(void)
{
  DTYPE dout, din;

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"AESNOT");
  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(EVNEQM_TAG,"SET",&dout,&din,CA_WRITE,AT_SCALAR,0,"Set event trigger",-1,NULL);

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"AESCON");
  RegisterPropertyInformation(EVNEQM_TAG,"RECORD",NULL,&din,CA_WRITE,AT_SCALAR,0,"record event",-1,NULL);
  RegisterPropertyInformation(EVNEQM_TAG,"REMOVE",NULL,&din,CA_WRITE,AT_SCALAR,0,"remove event",-1,NULL);

  RegisterProperty(EVNEQM_TAG,"NEVENTS",1,CF_LONG,CA_READ,"query number of events");
  RegisterProperty(EVNEQM_TAG,"EVENTS",100,CF_LONG,CA_READ,"query events");
  RegisterProperty(EVNEQM_TAG,"NASERVERS",1,CF_LONG,CA_READ,"query number of archive servers");
  RegisterProperty(EVNEQM_TAG,"ASERVERS",100,CF_NAME16,CA_READ,"query archive servers");
  RegisterProperty(EVNEQM_TAG,"NDSERVERS",1,CF_LONG,CA_READ,"query number of device servers");
  RegisterProperty(EVNEQM_TAG,"DSERVERS",100,CF_NAME16,CA_READ,"query device servers");

  RegisterProperty(EVNEQM_TAG,"SETCOMMENT",256,CF_TEXT,CA_WRITE,"write comment file");

}
int evneqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int i, siz, cc = 0;
  time_t starttime = 0,stoptime = time(NULL);
  char remSrvName[17];
  AESnotify  *aesNotify;
  AEScontext *aesContext;
  UINT32 evnnum;

  if (!strcmp(devProperty,"SET"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if ( din->dFormat != CF_STRUCT ||
         strncmp(din->dTag,"AESNOT",8) ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    aesNotify = (AESnotify *)din->data.vptr;
    strncpy(remSrvName,GetCaller(EVNEQM_TAG),16);
    feclog("%s: received SET request from %s",EVNEQM_TAG,remSrvName);
    evnnum = ArchiveEventSet(0,aesNotify->eventCode,aesNotify->comment[0]);
    if (dout->dArrayLength > 0) putValuesFromLong(dout,(SINT32 *)&evnnum,1);
    return 0;
  }
  if (!strcmp(devProperty,"RECORD"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESCON") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    aesContext = (AEScontext *)din->data.vptr;
    strncpy(remSrvName,GetCaller(EVNEQM_TAG),16);
    feclog("%s: received RECORD request from %s",EVNEQM_TAG,remSrvName);
    cc = ArchiveEventRecord(aesContext->eventNumber,remSrvName,aesContext->deviceName);
    return cc;
  }
  if (!strcmp(devProperty,"SETCOMMENT")) 
  {
    return apceqm(devName,"ANNOTATE",dout,din,access);
  }
  if (!strcmp(devProperty,"REMOVE")) 
  {
  }
  if (!strcmp(devProperty,"ASERVERS"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME16: siz = 16; break;
      case CF_NAME32: siz = 32; break;
      case CF_NAME48: siz = 48; break;
      case CF_NAME64: siz = 64; break;
      default: return illegal_format;
    }
    if (din->dArrayLength == 0)
    {
      for (i=0; i<numArchiveTypes && i<(int)dout->dArrayLength; i++)
      {
        strncpy(dout->data.cptr+i*siz,ArchRecord[i].trg,16);
      }
      dout->dArrayLength = i;
      return 0;
    }
    if ((cc=getValuesAsLong(din,(SINT32 *)&evnnum,1)) != 0) return cc;
    cc = getContributingArchiveServers(evnnum,(NAME16 *)dout->data.vptr,(SINT32 *)&dout->dArrayLength);
    return cc;
  }
  if (!strcmp(devProperty,"NASERVERS"))
  {
    return putValuesFromLong(dout,(SINT32 *)&numArchiveTypes,1);
  }
  if (!strcmp(devProperty,"NEVENTS"))
  {
    return apceqm(devName,"QUERY.NFILES",dout,din,access);
  }
  if (!strcmp(devProperty,"EVENTS"))
  {
    return apceqm(devName,"QUERY.FILES",dout,din,access);
  }
  if (!strcmp(devProperty,"NDSERVERS"))
  {
  }
  if (!strcmp(devProperty,"DSERVERS"))
  {
    switch (dout->dFormat)
    {
      case CF_NAME16: siz = 16; break;
      case CF_NAME32: siz = 32; break;
      case CF_NAME48: siz = 48; break;
      case CF_NAME64: siz = 64; break;
      default: return illegal_format;
    }
    if (din->dArrayLength == 0) return illegal_read_write;
    if ((cc=getValuesAsLong(din,(SINT32 *)&evnnum,1)) != 0) return cc;
    cc = getContributingDeviceServers(evnnum,(NAME16 *)dout->data.vptr,(SINT32 *)&dout->dArrayLength);
    return cc;
  }
  return illegal_property;
}
int getEventHandle(int event)
{
  int cc=0;
  UINT32 levent = event;
  short sid = -1;
  DTYPE dout,din;

  dout.dArrayLength = 1;
  dout.dFormat = CF_SHORT;
  dout.data.vptr = &sid;
  dout.dTag[0] = 0;
  din.dArrayLength = 1;
  din.dFormat = CF_LONG;
  din.data.lptr = (SINT32 *)&levent;
  din.dTag[0] = 0;
  if ((cc=apceqm("EVENT","OPEN",&dout,&din,CA_READ)) != 0) return -cc;
  return sid;
}
int closeEventHandle(short id)
{
  int cc=0;
  short sid = id;
  DTYPE dout,din;

  dout.dArrayLength = 0;
  dout.dFormat = CF_NULL;
  din.dArrayLength = 1;
  din.dFormat = CF_SHORT;
  din.data.sptr = &sid;
  din.dTag[0] = 0;
  return apceqm("EVENT","CLOSE",&dout,&din,CA_READ);
}
int getContributingArchiveServers(int event,NAME16 *svrs,int *len)
{
  int cc=0,i,n;
  short sid = getEventHandle(event);
  NAME16 names[100];
  DTYPE dout,din;

  if (sid < 0) return -sid;
  dout.dArrayLength = 100;
  dout.dFormat = CF_NAME16;
  dout.data.vptr = names;
  dout.dTag[0] = 0;
  din.dArrayLength = 1;
  din.dFormat = CF_SHORT;
  din.data.sptr = &sid;
  din.dTag[0] = 0;
  cc = apceqm("EVENT","QUERY.EQP",&dout,&din,CA_READ);
  for (i=0,n=0; cc == 0 && i<(int)dout.dArrayLength && n < *len; i++)
  {
    if (!strncmp(names[i].name,"EventServer",16)) continue;;
    strncpy(svrs[n++].name,names[i].name,16);
  }
  *len = n;
  closeEventHandle(sid);
  return cc;
}
int getContributingDeviceServers(int event,NAME16 *svrs,int *len)
{
  int cc=0,i,n,siz;
  short sid = getEventHandle(event);
  NAME16 names[100];
  char name[32];
  DTYPE dout,din;
  APCBLK apc;

  if (sid < 0) return -sid;
  dout.dArrayLength = 100;
  dout.dFormat = CF_NAME16;
  dout.data.vptr = names;
  dout.dTag[0] = 0;
  din.dArrayLength = 1;
  din.dFormat = CF_SHORT;
  din.data.sptr = &sid;
  din.dTag[0] = 0;
  if ((cc=apceqm("EVENT","QUERY.EQP",&dout,&din,CA_READ)) != 0) ccerr(cc);
  siz = dout.dArrayLength;
  dout.dFormat = CF_TEXT;
  dout.dArrayLength = 16;
  dout.data.cptr = name;
  din.dArrayLength = 1;
  din.dFormat = CF_STRUCT;
  din.data.vptr = &apc;
  strncpy(din.dTag,"APCBLK",8);
  apc.sessionID = sid;
  for (i=0,n=0; cc == 0 && i<siz && n < *len; i++)
  { 
    if (!strncmp(names[i].name,"EventServer",16)) continue;;
    strncpy(apc.expName,names[i].name,16);
    strncpy(apc.devProperty,"EventRecord",32);
    strncpy(apc.devName,"#0",16);
    if ((cc=apceqm("EVENT","READDATA",&dout,&din,CA_READ)) != 0) ccerr(cc);
    strncpy(svrs[n++].name,name,16);
  }
  *len = n;
err:
  closeEventHandle(sid);
  return cc;
}
