/* gens.c - Group Equipment Name Server */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include "tine.h"
#if defined(UNIX)
# include "dirent.h"
#elif defined(WIN32)
# include "direct.h"
#endif

int g_isTerminating = FALSE;
void term_handler(int signum)
{                  
  g_isTerminating = TRUE;
  feclog("caught SIGTERM");          
}
#define STD_QUERY_TIMEOUT 2000
extern TINE_EXPORT int gIgnoreAddressCacheOnUnknownAddress;
void nslog(char *text,...);
int scanTables(void);
int findcol(char *str,char *tag);
char *colptr(int col,char *str);
int isSystemContextDecoration(char *dec);
int sndRegExp(ExportListStruct *el);
int scanFlag = 0;
void grpini(void);
void grptsk(void);
int grpeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);
int asyncCmdForward(char *context,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int access,BYTE *badr);
void chkForwardsQueue(int idx);
int getForwardsList(char *fn);
int dumpFwdQueues(int i1,int i2, int i3, int i4);
void cbCheckDeviceNames(int id, int cc);
void chkForwardsList(int idx);
BYTE *getFwdIpAddr(char *tgt);
#define GENSFWD_FILE "forward.csv"
extern int nforwards;
typedef struct GrpDataXRefStruct
{
  char target[64];
  char devPrefix[8];
  char devPostfix[8];
  int offset;
  int index;
  int size;
  int lastAckSize;
  int disabled;
  int errcnt;
  int acquired;
  int isLegacy;
  int lastQrySize;
  int devlid; /* link id for async devnames acquisition */
  NAME64 *devlist;
} GRPXREF;
typedef struct GrpDataStruct
{
  char context[CONTEXT_NAME_SIZE];
  char subsystem[SUBSYSTEM_NAME_SIZE];
  char name[EXPORT_NAME_SIZE];
  char module[EQM_NAME_SIZE];
  int size;
  int npending;
  int nscans;
  int disabled;
  int deviceCapacity;
  GRPXREF *xref; /* array xref of members */
} GRPDATA;
GRPDATA *GRPData=NULL;
int nGroupServers = 0;
int GrpDataCapacity = 0;
static int debug=0;
#define GRP_FILE_NAME "grpdbase.csv"
#define GRP_RESERVED_SPACE 100
#define XREF_RESERVED_SPACE 100
#define MIN_GROUP_SPACE 256

void gensCleanup(void)
{
  if (GRPData != NULL) 
  {
    int i, k;
    for (i=0; i<nGroupServers; i++)
    {
      if (GRPData[i].xref != NULL) 
      {
        for (k=0; k<GRPData[i].size; k++)
          if (GRPData[i].xref[k].devlist != NULL) free(GRPData[i].xref[k].devlist);
        free(GRPData[i].xref);
      }
    }
    free(GRPData); GRPData = NULL;
  }
}

typedef struct memInfoStruct
{
  char name[EXPORT_NAME_SIZE];
  char devPrefix[8];
  char devPostfix[8];
  int index;
} MemInfo;
#define MemInfoTag "GrpMbr"
#define MAX_MEMBERS 256

#if defined(WIN32) || defined(NEED_UNSETENV)
int unsetenv(char *env)
{
  char envstr[128];
  strncpy(envstr,env,120);
  strcat(envstr,"=");
# ifdef WIN32
  return _putenv(envstr);
# else
  return putenv(envstr);
# endif
}
#endif

#define quit(i) { printf("Register struct: error\n"); exit(i); }
void registerGensStructs(void)
{
  static int done=FALSE;
  if (done) return;
  done = TRUE;
  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct(MemInfoTag,OFFSETIN(MemInfo,name),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(1);
  if (addFieldToStruct(MemInfoTag,OFFSETIN(MemInfo,devPrefix),8,CF_TEXT,"devPrefix")) quit(1);
  if (addFieldToStruct(MemInfoTag,OFFSETIN(MemInfo,devPostfix),8,CF_TEXT,"devPostfix")) quit(1);
  if (addFieldToStruct(MemInfoTag,OFFSETIN(MemInfo,index),1,CF_INT32,"index")) quit(1);
  if (sealTaggedStruct(MemInfoTag,sizeof(MemInfo),MAX_MEMBERS)) quit(1);
}
char thisLocation[FEC_LOCATION_SIZE] = "central location";
void PreSystemInit(void)
{
  int cc;
  char *c=NULL;
  int keepEnvSettings=FALSE;
  char ipaddr[ADDR_SIZE_IPV6];

  if ((c=getenv("USE_TINE_HOME")) != NULL && stricmp(c,"TRUE") == 0)
     keepEnvSettings = TRUE;
  if (!keepEnvSettings)
  {
    unsetenv("TINE_HOME");
    unsetenv("FEC_HOME");
    /* legacy env variable */
    unsetenv("CONTROLDB");
    unsetenv("FECDB");
  }
  inet_ntop(gAfInet,getmyipaddr(),ipaddr,ADDR_SIZE_IPV6);
  SetNameServerAddress(ipaddr);
  feclog("assign ENS address : %.16s",ipaddr);
  isGroupNameServer(TRUE);
  gUseLoopback = TRUE;
  MaxNumContracts = 1000;
  gIgnoreAddressCacheOnUnknownAddress = TRUE;
  FeclogDepth = 10000;
  gIgnoreRegionRules = TRUE;
  gIgnoreFollowers = TRUE;

  cc = RegisterFecInformation(GENS_FEC,"SER","SITE","ENS companion process",thisLocation,"none","control system admin",101);
  if (cc) feclog("RegisterFecInformation : %.32s",erlst[cc]);

  /* give this a bit of transport space in case of loooooooong devices lists ... */
  SetWorkAreaSize(0x40000); 
  /* a lot of simultaneous information at startup */
  SystemAssignBufferSpace(256000,64000);
  // signal(SIGTERM,term_handler);
  SetSystemCleanupFunction(gensCleanup);
  // don't allow dependent links (which would indicate alias servers within the same group)
  SetAllowDependentLinks(FALSE);
}
void PostSystemInit(void)
{
  int cc;
  DTYPE dout;
  memset(&dout,0,sizeof(DTYPE));
  cc = RegisterEquipmentModule(GENS_FEC,GENS_TAG,100,grpeqm,grpini,grptsk,1000,NULL);
  if (cc) 
  {
    feclog("RegisterEquipmentModule : %.32s",erlst[cc]);
    return;
  }
  cc = RegisterProperty(GENS_TAG,"GROUPS",100,CF_NAME32,CA_READ,"device group list");
  if (cc < 0) feclog("RegisterProperty GROUPS: %.32s",erlst[-cc]);
  dout.dFormat = CF_STRUCT; dout.dArrayLength = MAX_MEMBERS; strncpy(dout.dTag,MemInfoTag,TAG_NAME_SIZE);
  cc = RegisterPropertyInformation(GENS_TAG,"MEMBERS",&dout,NULL,CA_READ,AT_UNKNOWN,0,"full group member information list",-1,NULL);
  if (cc < 0) feclog("RegisterProperty MEMBERS: %.32s",erlst[-cc]);
  cc = RegisterProperty(GENS_TAG,"MEMBERS",100,CF_NAME32,CA_READ,"device group member list");
  if (cc < 0) feclog("RegisterProperty MEMBERS: %.32s",erlst[-cc]);
  cc = RegisterProperty(GENS_TAG,"JOIN",1,CF_NAME32,CA_WRITE,"join device group member list");
  if (cc < 0) feclog("RegisterProperty JOIN: %.32s",erlst[-cc]);
  cc = RegisterProperty(GENS_TAG,"SEPARATE",1,CF_NAME32,CA_WRITE,"join device group member list");
  if (cc < 0) feclog("RegisterProperty SEPARATE: %.32s",erlst[-cc]);
  cc = RegisterProperty(GENS_TAG,"REFRESH",0,CF_NULL,CA_WRITE,"refresh device list for the given group");
  if (cc < 0) feclog("RegisterProperty REFRESH: %.32s",erlst[-cc]);
  cc = RegisterProperty(GENS_TAG,"REEXPORT",0,CF_NULL,CA_WRITE,"re-export the group server to the ENS");
  if (cc < 0) feclog("RegisterProperty REEXPORT: %.32s",erlst[-cc]);
}
int makeDirPath(char *fname)
{
# if defined(UNIX)
  DIR *dp;
  struct dirent *dirp;
# elif defined(WIN32)
  char fn[128];
  struct _finddata_t fp;
# endif
  char *c,tmpname[128],dirname[128];
  int len;

  strncpy(tmpname,fname,128); dirname[0] = 0;
  for (c=strtok(tmpname,"/\\"); c != NULL; c=strtok(NULL,"/\\"))
  {
    strcat(dirname,c);
    if (!strncmp(dirname,fname,128)) /* okay the directory should be there */
    {
      return 0;
    }
    len = (int)strlen(dirname);
    dirname[len] = FS_DELIMITER; dirname[len+1] = 0;
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL) 
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return 0;
    }
#   elif defined(WIN32)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
#   endif
  }
  return -1;
}
#define isENS(nam) (!strncmp(nam,"ENS",3))
int isFec(char *nam)
{
  char hostname[64];
  if (GetHostFromFecName(nam,hostname) == 0) return TRUE;
  return FALSE;
}
#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,USERNAME_SIZE); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),16);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int GetDeviceNamesFromCache(char *srv,NAME64 *devbuffer,int *size)
{
  FILE *fp=NULL;
  char *c,fn[132],srvpath[64];
  int i,k,cc=0,done=0,dotToUnderscore=FALSE;
  int n, p;
  int dev_col=-1;
  char str[256],hdr[256],dev[72];

  if (srv == NULL || devbuffer == NULL || size == NULL) ccerr(argument_list_error);
  p = srv[0] == '/' ? 1 : 0;
  strncpy(srvpath,&srv[p],64);
  if ((c=strstr(srvpath,"/#")) != NULL) *c = 0;
retryGetDataBaseSize:
  sprintf(fn,"GROUPS/%.64s/devices.csv",srvpath);
  if ((n=GetDataBaseSize(fn)) < 0 && !dotToUnderscore) 
  {
    if ((c=strchr(srvpath,'.')) != NULL)
    {
      if (isSystemContextDecoration(&c[1]))
      { /* try legacy directory names ? */
        *c = '_';
        dotToUnderscore = TRUE;
        goto retryGetDataBaseSize;
      }
    }
    ccerr(-n);
  }
  if (*size < n) 
  {
    feclog("truncating size of device list for %.128s (%d to %d)!",srv,*size,n);
    nslog("truncating size of device list for %.128s (%d to %d)!",srv,*size,n);
    n = *size;
  }
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  for (done=0,i=0,k=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((dev_col=findcol(hdr,"DEVICES")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dev,colptr(dev_col,str),DEVICE_NAME_SIZE); dev[DEVICE_NAME_SIZE] = 0;
    strncpy(devbuffer[i].name,strtrm(dev),DEVICE_NAME_SIZE);
    i++;
  }
  *size = n;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int WriteDeviceNamesToCache(char *srv,NAME64 *devbuffer,int size)
{
  FILE *fp=NULL;
  char *c,fn[132],srvpath[64];
  int i,p,cc=0;

  if (srv == NULL || devbuffer == NULL || size < 1) ccerr(argument_list_error);
  p = srv[0] == '/' ? 1 : 0;
  strncpy(srvpath,&srv[p],64);
  if ((c=strstr(srvpath,"/#")) != NULL) *c = 0;
  sprintf(fn,"GROUPS/%.64s/devices.csv",srvpath);
  makeDirPath(fn);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Devices\n");
  for (i=0; i<size;i++)
  {
    if (strlen(devbuffer[i].name) == 0) continue;
    fprintf(fp,"%s\n",devbuffer[i].name);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

void checkContextCase(char *context,int list_size)
{
  int i;
  if (context == NULL || list_size < 0) return;
  for (i=0; i<list_size; i++)
  { /* find a case-consistent match */
    if (strnicmp(context,GRPData[i].context,CONTEXT_NAME_SIZE)) continue;
    if (strncmp(context,GRPData[i].context,CONTEXT_NAME_SIZE))
    {
      nslog("change context case from %.32s to %.32s for server %.32s",context,GRPData[i].context,GRPData[i].name);
      strncpy(context,GRPData[i].context,CONTEXT_NAME_SIZE);
    }
    return;
  }
  return;
}
int findServerGroup(char *group,char *context)
{
  int i;
  char *cntxt = context;
  if (context == NULL || strlen(context) == 0) cntxt = "DEFAULT";
  for (i=0; i<nGroupServers; i++)
  {
    if (strnicmp(cntxt,GRPData[i].context,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(group,GRPData[i].name,EXPORT_NAME_SIZE)) continue;
    break;
  }
  return i < nGroupServers ? i : -1;
}
#define DEV_BUFFER_SIZE 4000
NAME64 devbuffer[DEV_BUFFER_SIZE];
int writeMembersFile(char *filename,GRPXREF *xref,int listsize)
{
  int i,cc=0;
  FILE *fp=NULL;

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);
  if (listsize <= 0) 
  {
    nslog("!!!! writeMembersFile %s : group has no members !",filename);
    xref->disabled = TRUE;
    ccerr(0);
  }
  if (strlen(xref[0].target) == 0) ccerr(invalid_name);
  makeDirPath(filename);
  if ((fp=fopen(filename,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Members, Index, DevicePrefix, DevicePostfix\n"); /* write the header */
  for (i=0; i<listsize; i++)
  {
    if (xref[i].disabled) continue;
    fprintf(fp,"%.64s, %d, %.8s, %.8s\n",xref[i].target,xref[i].index,xref[i].devPrefix,xref[i].devPostfix);
  }
err:
  if (cc) nslog("!!!! writeMembersFile %s : %s",filename == NULL ? "[empty]" : filename,erlst[cc]);
  if (fp != NULL) fclose(fp);
  return cc;
}
void organizeMembers(GRPXREF *xref,int listsize)
{
  int i,k;
  GRPXREF tref;

  if (listsize <= 0) return;
  /* these will usually be ordered */
  /* the worst case (exactly backwards) can only happen once, so just do it this way */
  for (i=0; i<listsize; i++)
  {
    for (k=0; k<i; k++)
    {
      if (xref[i].index < xref[k].index)
      { /* then swap them */
        tref = xref[i];
        xref[i] = xref[k];
        xref[k] = tref;
      }
    }
  }
}
GRPXREF *loadMembersFile(char *filename,int *listsize)
{
  int i,idx,n=0,cc=0,done=0,dotToUnderscore=FALSE;
  int mem_col=-1,idx_col=-1,pre_col=-1,post_col=-1;
  char *c,*ce,str[256],hdr[256],scratch[128];
  GRPXREF *xref=NULL;
  FILE *fp=NULL;

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);

retryGetDataBaseSize:
  if ((n=GetDataBaseSize(filename)) < 0)
  {
    if ((c=strchr(filename,'.')) != NULL)
    {
      strncpy(scratch,&c[1],16);
      if ((ce=strchr(scratch,'/')) != NULL) *ce = 0;
      if (isSystemContextDecoration(scratch))
      { /* try legacy directory names ? */
        *c = '_';
        dotToUnderscore = TRUE;
        goto retryGetDataBaseSize;
      }
    }
    ccerr(-n);
  }
  if ((xref=(GRPXREF *)calloc(n+XREF_RESERVED_SPACE,sizeof(GRPXREF))) == NULL) ccerr(out_of_server_memory);
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<n;)   
  { /* scan file */
    strtrm(str);
    if (strlen(str) == 0) continue; /* empty line */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((mem_col=findcol(hdr,"MEMBERS")) < 0) ccerr(no_such_line);
      idx_col = findcol(hdr,"INDEX");
      pre_col = findcol(hdr,"DEVICE_PREFIX");
      post_col = findcol(hdr,"DEVICE_POSTFIX");
      done = 1;
      continue;
    }
    idx = i;
    if (idx_col >= 0) 
    {
      strncpy(scratch,colptr(idx_col,str),16);
      if ((idx=atoi(scratch)) < 0) continue;
    }
    xref[i].index = idx;
    if (pre_col >= 0)
    {
      strncpy(scratch,colptr(pre_col,str),8); scratch[8] = 0;
      if ((c=strchr(scratch,' ')) != NULL) *c = 0; // be strict with blanks !
      if (strlen(scratch) > 0) strncpy(xref[i].devPrefix,scratch,8);
    }
    if (post_col >= 0)
    {
      strncpy(scratch,colptr(post_col,str),8); scratch[8] = 0;
      if ((c=strchr(scratch,' ')) != NULL) *c = 0; // be strict with blanks !
      if (strlen(scratch) > 0) strncpy(xref[i].devPostfix,scratch,8);
    }
    strncpy(scratch,colptr(mem_col,str),EXPORT_NAME_SIZE);
    if (strlen(scratch) == 0) continue;
    strncpy(xref[i].target,scratch,EXPORT_NAME_SIZE);
    i++;
  }
  if (i != n)
  {
    nslog("scaned fewer members than database size (missing target)");
    n = i;
  }
  organizeMembers(xref,n);
err:
  if (n == 0) 
  {
    if (cc == 0) cc = string_expected;
    nslog("!!!! members file %s exists but has no members !!!!",filename);
  }
  if (listsize != NULL) *listsize = n;
  if (cc) 
  {
    if (xref != NULL) free(xref); xref = NULL;
    nslog("loadMembersFile %s : %s",filename == NULL ? "[empty]" : filename,erlst[cc]);
  }
  if (fp != NULL) fclose(fp);
  return xref;
}
int writeDatabaseFile(char *filename)
{
  FILE *fp = NULL;
  int i, cc=0;

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);
  if ((fp=fopen(filename,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Group, GroupModule, Context, SubSystem\n"); /* write the header */
  for (i=0; i<nGroupServers; i++)
  {
    if (GRPData[i].size <= 0) continue;
    fprintf(fp,"%s, %s, %s, %s\n",
      GRPData[i].name,GRPData[i].module,GRPData[i].context,GRPData[i].subsystem);
  }

err: 
  if (fp != NULL) fclose(fp);
  if (cc) nslog("writeDatabaseFile %s : %s",filename == NULL ? "[empty]" : filename, erlst[cc]);
  return cc;
}
int dumpDatabase(void)
{
  int i, len, rc = 0, cc = 0;
  char fn[128];

  for (i=0; i<nGroupServers; i++)
  {
    strcpy(fn,"GROUPS"); 
    len = (int)strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].context);
    len = (int)strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].name); strcat(fn,".csv");
    cc = writeMembersFile(fn,GRPData[i].xref,GRPData[i].size);
    if (cc != 0) rc = cc;
  }
  cc = writeDatabaseFile(GRP_FILE_NAME);
  if (cc != 0) rc = cc;
  /* can't re-read the files immediately ! */
  /* the call to get device names will fail ! */
  /* scanTables(); */
  return rc;
}
#define MAKEID(g,m) ((((m)&0xff)<<16)|((g)&0xff))
#define GROUPID(n) ((n)&0xff)
#define MEMBERID(n) (((n)>>16)&0xff)

int fixDeviceList(int i,int k,int cc)
{
  int n,len,size = -1, cbid, fid, resubmit = FALSE;
  char srv[128];
  GRPXREF *xref;
  DTYPE dout;
  xref = &GRPData[i].xref[k];

  if (xref->size > 0) return FALSE; /* already acquired */
  if (strlen(xref->target) == 0)
  {
    nslog("member %d of group /%.32s/%.32s has an empty target!",k,GRPData[i].context,GRPData[i].name);
    xref->disabled = TRUE;
    return FALSE;
  }
  /* fill in dout with default settings */
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_NAME64;
  dout.dArrayLength = xref->lastQrySize;
  sprintf(srv,"/%.32s/%.64s/#0",GRPData[i].context,xref->target);
  if ((cc&0xfff) == has_query_function)
  {
    nslog("server %s has property query precedence -> not a candidate for a server group !",srv);
    xref->disabled = TRUE;
    ccerr(invalid_data);
  }
  switch (cc)
  {
    case link_blacklisted:
      xref->acquired = TRUE;
      return FALSE;
    case illegal_format:
      if (xref->isLegacy) break; /* still illegal_format ? */
      if (xref->lastQrySize == 0) xref->lastQrySize = DEV_BUFFER_SIZE;
      xref->isLegacy = TRUE;
      dout.dFormat = CF_NAME16;
      dout.dArrayLength = xref->lastQrySize;
      resubmit = TRUE;
      break;
    case invalid_transport_size:
      if (xref->lastQrySize == 0) xref->lastQrySize = DEV_BUFFER_SIZE;
      xref->lastQrySize /= 2;
      dout.dFormat = (short)(xref->isLegacy ? CF_NAME16 : CF_NAME64);
      dout.dArrayLength = xref->lastQrySize;
      resubmit = TRUE;
      break;
    case non_existent_elem:
      nslog("acquire device names for %s non_existent -> re-submit request",srv);
      fid = conTbl[xref->devlid]->fecIdx;
      if (fid > 0 && fid < numFecTblEntries)
      { /* force address re-acquisition */
        FecTbl[fid].inetProtocol = 0; 
      }
      if (xref->errcnt++ < 1) resubmit = TRUE;
      break;
    case link_timeout:
      nslog("acquire device names for %s timed out -> re-submit request",srv);
      if (xref->errcnt++ < 2) 
      {
        if (ResubmitDataLink(xref->devlid) == 0) return TRUE;
      }
    default:
      break;
  }
  if (resubmit)
  { /* due to illegal_format or invalid_transport_size */
    nslog("re-submit device names acquisition for %.128s due to %.32s",srv,erlst[cc]);
    memset(dout.dTag,0,TAG_NAME_SIZE);
    dout.data.vptr = devbuffer;
    cbid = MAKEID(i,k);
    xref->devlid = AttachLinkEx(srv,"DEVICES",&dout,NULL,CA_READ|CA_RETRY,STD_QUERY_TIMEOUT,cbCheckDeviceNames,CM_SINGLE|CM_SYNCNOTIFY,cbid);
    if (xref->devlid < 0)
    { /* the callback will also be fired with whatever error this is */
      nslog("could not acquire device names for %s : %s",srv,erlst[-xref->devlid]);
      xref->disabled = TRUE;
    }
    return TRUE;
  }
  nslog("call to device list for %s returned: %d",srv,cc);
test_completion:
  if (cc != 0)
  { /* read from file */
    xref->errcnt++;
    xref->acquired = FALSE;
    if (cc != address_unknown)
    { /* try the local cache */
      nslog("acquire device names for %.128s : error <%d>, get stored names from file cache",srv,cc);
      size = DEV_BUFFER_SIZE;
      cc = GetDeviceNamesFromCache(srv,devbuffer,&size);
      if (cc == 0 && size > 0) xref->acquired = TRUE;
      nslog("get last cached device list returned %d",cc);
      if (size <= 0) cc = device_error;
    }
    else
    {
      nslog("acquire device names for %s : address unknown (disable this member!)",srv);
    }
  }
  else
  { /* successful acquisition: write to file */
    size = GetCompletionDataSize(xref->devlid);
    if (size == DEV_BUFFER_SIZE)
    {
      nslog("!!!!! server %s returned %d devices : increase query device buffer size !",srv,DEV_BUFFER_SIZE);
      xref->acquired = FALSE;
      xref->errcnt++;
      return FALSE;
    }
    if (xref->isLegacy)
    { /* a NAME16 list instead of a NAME64 list */
      for (n=size; n>=0; n--)
      { /* in situ reformatting */
        strncpy(devbuffer[n].name,((NAME16 *)devbuffer)[n].name,DEVICE_NAME_SHORTSIZE);
      }
    }
    for (n=0; n<size; n++)
    { /* check for blank names (break at the first offender) */
      if (strlen(devbuffer[n].name) == 0) break;
    }
    if (n < size) 
    {
      nslog("break device list for %s at empty name at element %d of %d",srv,n,size);
      if (n == 0)
      {
        cc = not_initialized;
        goto test_completion;
      }
      size = n;
    }
    xref->acquired = TRUE;
    xref->errcnt = 0;
    for (n=0; n<size; n++)
    { /* have a look at the supplied device names first */
      if (devbuffer[n].name[0] == '#')
      { /* no real device names given -> fake them : re-use the server name */
        strncpy(devbuffer[n].name,xref->target,DEVICE_NAME_SIZE);
        if (size > 1) 
        {
          devbuffer[n].name[12] = 0; /* 1st 12 chars from the srv name */
          len = (int)strlen(devbuffer[n].name);
          sprintf(&devbuffer[n].name[len],"#%3d",n);
        }
      }
    }
    if (size == 0)
    {
      nslog("!!!! fixDeviceList acquired no devices for member %s of /%.32s/%.32s",
        xref->target,GRPData[i].context,GRPData[i].name);
      xref->disabled = TRUE;
      ccerr(invalid_data);
    }
    else
    {
      nslog("write acquired device list to file (%d devices)",size);
      if (cc == 0)
      { /* only update if success */
        WriteDeviceNamesToCache(srv,devbuffer,size);
      }
    }
  }
  if (cc) 
  {
    if (xref->errcnt > 1)
    {
      nslog("!!!! do not have a device list for %.128s !",srv);
      nslog("!!!! cannot acquire devices for %.128s : %s",srv,erlst[cc%0xff]);
      size = 1;
      strcpy(devbuffer[0].name,"device0");
      xref->errcnt = 0;
      xref->disabled = TRUE;
    }
    else
    {
      return FALSE;
    }
  }
  if (size > 0)
  { /* got something */
    if (xref->lastAckSize < size || xref->devlist == NULL)
    {
      nslog("number of devices for %.128s : %d to %d",srv,xref->lastAckSize,size);
      if (xref->devlist != NULL) free(xref->devlist);
      if ((xref->devlist=(NAME64 *)calloc(size,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
    }
    xref->lastAckSize = size;
    xref->size = size;
    memcpy(xref->devlist,devbuffer,size*sizeof(NAME64));
  }
err:
  if (cc != 0) nslog("!!!! fixDeviceList: %.32s",erlst[cc]);
  return FALSE;
}
int reExportGroup(int idx)
{
  ExportListStruct *el;
  if (idx < 0 || idx >= nGroupServers) return invalid_index;
  if ((el=getExportListItem(GRPData[idx].module)) == NULL) 
  {
    feclogEx(TINE_LOGLEVEL_WARN,"could not find module %.6s for re-export",GRPData[idx].module);
    return non_existent_elem;
  }
  el->registered = 0; /* need to clear this ... */
  return sndRegExp(el);
}
void cbCheckDeviceNames(int id, int cc)
{
  int i,k,n,nd,grpsize,grpspace,rc;
  char rdrstr[256],devstr[DEVICE_NAME_SIZE+24];
  GRPXREF *xref;
  ExportListStruct *el;

  i = GROUPID(id);
  k = MEMBERID(id);

  if (i < 0 || i >= nGroupServers)
  {
    nslog("!!!! invalid group id (%d) in callback (ngroups = %d)",i,nGroupServers);
    return;
  }
  if (k < 0 || k >= GRPData[i].size)
  {
    nslog("!!!! invalid member id (%d) in callback (group /%.32s/%.32s (%d), nmembers = %d)",
      k,GRPData[i].context,GRPData[i].name,i,GRPData[i].size);
    GRPData[i].npending--;
    return;
  }

  /* fix the device list for this member of this group */
  if (fixDeviceList(i,k,cc)) return; /* non-zero -> was resubmitted */

  GRPData[i].npending--;
  /* don't continue until all in */
  if (GRPData[i].npending > 0) return; 
  /* now they're all in : determine total size and device list offset */
  grpsize = 0;
  for (k=0; k<GRPData[i].size; k++) 
  { /* loop over members of this group */
    xref = &GRPData[i].xref[k];
    if (xref->disabled) continue;
    xref->offset = grpsize;
    grpsize += xref->size;
  }
  if (grpsize == 0) 
  { /* database problems ? */
    nslog("!!!! group (%d) /%.32s/%.32s has no members !",i,GRPData[i].context,GRPData[i].name);
    GRPData[i].size = 0;
    GRPData[i].disabled = TRUE;
    return; 
  }
  if (strlen(GRPData[i].module) == 0)
  {
    nslog("group in context %s has no equipment module name!",GRPData[i].context);
    return;
  }
  if (!strcmp(GRPData[i].module,GENS_TAG))
  {
    nslog("group /%s/%s has equipment module name %s !",GRPData[i].context,GRPData[i].name,GENS_TAG);
    return;
  }
  if (grpsize > GRPData[i].deviceCapacity && GRPData[i].deviceCapacity > 0)
  { /* suddenly there are more ! */
    nslog("sudden increase in device space for group /%.32s/%.32s (module %.6s) to %d devices from %d",
      GRPData[i].context,GRPData[i].name,GRPData[i].module,grpsize,GRPData[i].deviceCapacity);
  }
  /* register equipment module with no equipment function, leave room for additional devices */
  if ((el=getExportListItem(GRPData[i].module)) == NULL) 
  { /* if not already registered */
    strncpy(gDeviceContext,GRPData[i].context,CONTEXT_NAME_SIZE);
    strncpy(gDeviceSubSystem,GRPData[i].subsystem,SUBSYSTEM_NAME_SIZE);
    /* leave some room for others to come: (if the total number exceeds this -> restart required!) */
    grpspace = MAX(grpsize*4,MIN_GROUP_SPACE);
    GRPData[i].deviceCapacity = grpspace;
    nslog("register /%.32s/%.32s (module %.6s) with space for %d devices",GRPData[i].context,GRPData[i].name,GRPData[i].module,grpspace);
    if ((rc=RegisterEquipmentModule(GRPData[i].name,GRPData[i].module,grpspace,NULL,NULL,NULL,0,NULL)) != 0)
      nslog("register equipment module /%.32s/%.32s (module %.6s) failed : %.32s",GRPData[i].context,GRPData[i].name,GRPData[i].module,erlst[rc]);
    /* register a nominal property */
    RegisterProperty(GRPData[i].module,"EMPTY",1,CF_SHORT,CA_READ,"empty property");
  }
  else
  { 
    if (el->EqmDevLstCapacity < grpsize)
    { /* now there are more than the original capacity ! */
      grpspace = MAX(grpsize*4,MIN_GROUP_SPACE);
      nslog("reassign device space for group /%.32s/%.32s (module %.6s) to %d devices",
        GRPData[i].context,GRPData[i].name,GRPData[i].module,grpspace);
      if ((rc=SetSizeDeviceCapacity(GRPData[i].module,grpspace)) != 0)
      {
        nslog("!!!!! unable to reassign device capacity : %s",cc2str(rc));
      }
      else
      { /* force next if-block */
        GRPData[i].deviceCapacity = 0;
      }
    }
    if (GRPData[i].deviceCapacity == 0) 
    { /* might be necessary after a re-load of the database */
      GRPData[i].deviceCapacity = el->EqmDevLstCapacity;
    }
  }
  nslog("assign number devices for /%.32s/%.32s (%.6s) to %d",GRPData[i].context,GRPData[i].name,GRPData[i].module,grpsize);
  if ((rc=AssignNumDevices(GRPData[i].module,grpsize)) != 0)
  {
    nslog("!!!! assign number devices for /%.32s/%.32s failed : %s",GRPData[i].context,GRPData[i].name,erlst[rc]);
  }

  /* register the device names */
  /* redirect the calls for each device */
  nd = 0;
  for (k=0; k<GRPData[i].size; k++)
  {
    xref = &GRPData[i].xref[k];
    for (n=0; n<xref->size; n++)
    {
      if (strlen(xref->devlist[n].name) == 0) continue;
      if (n + xref->offset >= GRPData[i].deviceCapacity)
      {
        nslog("!!!! register device %.64s (%d) for /%.32s/%.32s would exceed the current registered capacity %d",
          xref->devlist[n].name,n + xref->offset,GRPData[i].context,GRPData[i].name,GRPData[i].deviceCapacity);
        break;
      }
      memset(devstr,0,DEVICE_NAME_SIZE+24);
      if (strlen(xref->devPrefix) > 0) strncat(devstr,xref->devPrefix,8);
      strncat(devstr,xref->devlist[n].name,DEVICE_NAME_SIZE);
      if (strlen(xref->devPostfix) > 0) strncat(devstr,xref->devPostfix,8);
      strtrm(devstr);
      if ((rc=RegisterDeviceName(GRPData[i].module,devstr,n + xref->offset)) != 0)
      {
        nslog("!!!! register device (%.6s) %.80s (%d) failed : %.32s",
          GRPData[i].module,devstr,n+xref->offset,erlst[rc]);
      }
      else
      { /* record the number of successfully registered devices */
        nd++;
      }
      sprintf(rdrstr,"/%.32s/%.32s/%.64s",GRPData[i].context,xref->target,xref->devlist[n].name);
      if ((rc=RedirectDeviceName(GRPData[i].module,devstr,rdrstr)) != 0)
      {
        nslog("!!!! redirect (%.6s) %.80s (%d) to %.128s failed : %.32s",
          GRPData[i].module,devstr,n + xref->offset,rdrstr,erlst[rc]);
      }
    }
  }
  nslog("/%.32s/%.32s successfully registered %d devices",GRPData[i].context,GRPData[i].name,nd);
}
#define CHK_NON_ACQUIRED_DEADBAND 900
int gChkNonAckDeadband = CHK_NON_ACQUIRED_DEADBAND;
int gChkNotAck = FALSE;
void checkDeviceNamesAsync(void)
{
  int i, k, cbid, do_dump=FALSE;
  char srv[128];
  GRPXREF *xref;
  DTYPE dout;
  static time_t interim = 0;
  time_t tnow = time(NULL);

  if (gChkNonAckDeadband > 0 && tnow > interim + gChkNonAckDeadband)
  {
    if (interim != 0) gChkNotAck = TRUE;
    interim = tnow;
  }

  for (i=0; i<nGroupServers; i++)
  {
    if (GRPData[i].disabled == TRUE || GRPData[i].size == 0)
    {
      nslog("!!!! group /%.32s/%.32s has no members : remove it !",GRPData[i].context,GRPData[i].name);
      RemoveEquipmentModule(GRPData[i].module);
      do_dump = TRUE;
      continue;
    }
    if (GRPData[i].npending > 0) continue;
    GRPData[i].npending = GRPData[i].size;
    GRPData[i].nscans = 0;
    for (k=0; k<GRPData[i].size; k++)
    {
      xref = &GRPData[i].xref[k];
      if (xref->disabled) 
      {
        GRPData[i].npending--;
        continue;
      }
      if (gChkNotAck && xref->acquired == FALSE) 
      {
        nslog("attempt to acquire device names for /%.32s/%.32s in background",GRPData[i].context,xref->target);
        xref->size = 0;
      }
      if (xref->size > 0) 
      { /* already acquired */
        GRPData[i].npending--;
        continue; 
      }
      if (xref->lastQrySize == 0) xref->lastQrySize = DEV_BUFFER_SIZE;
      sprintf(srv,"/%.32s/%.32s/#0",GRPData[i].context,xref->target);      
      dout.dFormat = CF_NAME64;
      dout.dArrayLength = xref->lastQrySize;
      memset(dout.dTag,0,TAG_NAME_SIZE);
      dout.data.vptr = devbuffer;
      cbid = MAKEID(i,k);
      xref->devlid = AttachLinkEx(srv,"DEVICES",&dout,NULL,CA_READ|CA_RETRY,STD_QUERY_TIMEOUT,cbCheckDeviceNames,CM_SINGLE|CM_SYNCNOTIFY,cbid);
      if (xref->devlid < 0)
      { /* the callback will also be fired with whatever error this is (except link_exists) */
        nslog("could not acquire device names for %.128s : %.32s",srv,erlst[-xref->devlid]);
        xref->disabled = TRUE;
        if (xref->devlid == -link_exists) 
        {
          int elid = dout.xferReason;
          char esrv[128] = "another server";
          if (elid > 0) sprintf(esrv,"/%.32s/%.32s/#0",conTbl[elid]->ctxName,conTbl[elid]->expName);
          nslog("%.128s appears to be identical to %.128s",srv,esrv);
          GRPData[i].npending--;
        }
      }
    }
  }
  if (gChkNotAck) 
  {
    if (debug) dbglog("finished check device names cycle");
    gChkNotAck = FALSE;
  }
  if (do_dump) dumpDatabase();
  return;
}

int loadDatabaseFile(char *filename)
{
  FILE *fp = NULL;
  int i,len,nmembers,cc=0,done=0;
  int con_col=-1,sub_col=-1;
  int grp_col=-1,grpmod_col=-1;
  char *c,str[256],hdr[256],fn[128];

  if (filename == NULL || strlen(filename) == 0) ccerr(un_allocated);

  nGroupServers = 0;
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      con_col=findcol(hdr,"CONTEXT");
      sub_col=findcol(hdr,"SUBSYSTEM");
      if ((grp_col=findcol(hdr,"GROUP")) < 0) ccerr(no_such_line);
      if ((grpmod_col=findcol(hdr,"GROUPMODULE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(GRPData[i].name,colptr(grp_col,str),EXPORT_NAME_SIZE);
    if (strlen(GRPData[i].name) == 0) continue;
    strncpy(GRPData[i].module,colptr(grpmod_col,str),EQM_NAME_SIZE);
    if (strlen(GRPData[i].module) == 0) continue;
    strncpy(GRPData[i].context,con_col >= 0 ? colptr(con_col,str) : "DEFAULT",CONTEXT_NAME_SIZE);
    if ((c=strchr(GRPData[i].context,'_')) != NULL)
    { /* could this be "_TEST" or "_SIM" ? */
      if (isSystemContextDecoration(&c[1]))
      { /* seems to be! replace with a '.' */
        *c = '.';
      }
    }
    checkContextCase(GRPData[i].context,i-1);
    if (sub_col >= 0)
    {
      strncpy(GRPData[i].subsystem,colptr(sub_col,str),SUBSYSTEM_NAME_SIZE);
    }
    /* read the members list for this group */
    strcpy(fn,"GROUPS"); 
    len = (int)strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].context);
    len = (int)strlen(fn); fn[len] = FS_DELIMITER; fn[len+1] = 0;
    strcat(fn,GRPData[i].name); strcat(fn,".csv");
    if (GRPData[i].xref != NULL) free(GRPData[i].xref);
    if ((GRPData[i].xref=loadMembersFile(fn,&nmembers)) == NULL) continue;
    GRPData[i].size = nmembers;
    i++;
  }
  fclose(fp); fp = NULL;
  nGroupServers = i;
  nslog("read database: loaded %d groups",nGroupServers);
  checkDeviceNamesAsync();

err: 
  if (fp != NULL) fclose(fp);
  if (cc) nslog("!!!! loadDatabaseFile %s : %.32s",filename == NULL ? "[empty]" : filename, erlst[cc]);
  return cc;
}
int scanTables(void)
{
  int i, k, n, cc = 0;

  for (i=0; i<nGroupServers; i++)
  {
    if (GRPData[i].npending > 0) 
    {
      nslog("!!!! defer scanning database table as device acquistion is still ongoing for /%.32s/%.32s",
        GRPData[i].context,GRPData[i].name);
      if (GRPData[i].nscans++ > 10)
      {
        nslog("!!!! abort device acquisition attempts for /%.32s/%.32s",
          GRPData[i].context,GRPData[i].name);
        GRPData[i].npending--;
      }
      return not_ready;
    }
  }

  if (GRPData != NULL) 
  {
    nslog("re-scan the group database (free existing tables)");
    for (i=0; i<nGroupServers; i++)
    {
      if (GRPData[i].xref != NULL) 
      {
        for (k=0; k<GRPData[i].size; k++)
          if (GRPData[i].xref[k].devlist != NULL) free(GRPData[i].xref[k].devlist);
        free(GRPData[i].xref);
      }
    }
    free(GRPData); GRPData = NULL;
  }

  if ((n=GetDataBaseSize(GRP_FILE_NAME)) < 0) ccerr(-n);
  nslog("allocating space for %d groups",n+GRP_RESERVED_SPACE);
  if ((GRPData=(GRPDATA *)calloc(n+GRP_RESERVED_SPACE,sizeof(GRPDATA))) == NULL)
    ccerr(out_of_server_memory);
  GrpDataCapacity = n + GRP_RESERVED_SPACE;
  cc = loadDatabaseFile(GRP_FILE_NAME);
err:
  nslog("scan of group database : %.32s",erlst[cc]);
  return cc;
}
void grpini(void)
{
  registerGensStructs();
  SetAutoLinkErrorAlarms(FALSE);
  getAdministrators(ADMINS_FILE);
  getForwardsList(GENSFWD_FILE);
  RegisterUserCommand("grpdebug",NULL,&debug,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("forwards",dumpFwdQueues,NULL,NULL,CA_READ);
}
#ifdef WIN32
# define lstat(f,s) _stat(f,s)
#endif
void grptsk(void)
{
  int i, cc;
  static time_t lastfiletime = 0;
# ifdef WIN32
  struct _stat fnstat;
# else
  struct stat fnstat;
# endif

  if (lstat(GRP_FILE_NAME,&fnstat) < 0)
  {
    nslog("did not find gqp database : create new");
    if ((cc=writeDatabaseFile(GRP_FILE_NAME)) != 0)
    {
      feclog("panic: cannot create gqp database !");
      nslog("panic: cannot create gqp database !");
      return;
    }
  }
  if (lastfiletime != fnstat.st_mtime)
  {
    if ((cc=scanTables()) == not_ready) return;
    feclog("scanTables: %.32s",erlst[cc]);
    nslog("scanTables: %.32s",erlst[cc]);
    lastfiletime = fnstat.st_mtime;
  }

  checkDeviceNamesAsync();
  for (i=0; i<nforwards; i++) chkForwardsQueue(i);

  return;
}
char *makeModuleName(char *grpName)
{
  int i, k, len, idx = -1;
  char mod[8];
  static char modName[32];

  if (grpName == NULL || strlen(grpName) == 0) return NULL;
  if (strlen(grpName) < 7) 
  {
    strncpy(modName,grpName,6);
    for (i=0; i<nGroupServers && strncmp(modName,GRPData[i].module,6); i++);
    if (i == nGroupServers) return modName; /* simple name is available */
  }
  for (i=0; i<nGroupServers; i++)
  {
    if (GRPData[i].module[0] == 'G')
    { /* possibly assigned */
      strncpy(mod,&GRPData[i].module[1],5); mod[5] = 0;
      len = (int)strlen(mod);
      for (k=0; k<len && isdigit(mod[k]); k++);
      if (k == len) /* all digits */
      { /* definitely assigned! (find the highest group index) */
        k = atoi(mod);
        if (idx < k) idx = k;
      }
    }
  }
  idx++;
  sprintf(modName,"G%05d",idx);
  return modName;
}
int getNextIndexFromGroup(int n)
{
  int i, idx=0;
  GRPXREF *xref;
  for (i=0; i<GRPData[n].size; i++)
  {
    if ((xref = &GRPData[n].xref[i]) == NULL) continue;
    if (xref->index >= idx) idx = xref->index + 1;
  }
  return idx;
}
int splitDeviceName(char *devName,char *ctx,char *grp)
{
  char *c, *gptr=devName, *cptr="DEFAULT";
  if (devName == NULL || ctx == NULL || grp == NULL) return invalid_parameter;
  if (devName[0] == '/')
  {
    cptr = &devName[1];
    if ((gptr=strchr(cptr,'/')) == NULL) return invalid_parameter;
    *gptr++ = 0;
  }
  else
  {
    if ((c=strchr(devName,'/')) != NULL)
    {
      cptr = devName;
      *c++ = 0;
      gptr = c;
    }
  }
  strncpy(ctx,cptr,CONTEXT_NAME_SIZE);
  strncpy(grp,gptr,EXPORT_NAME_SIZE);
  return 0;
}
int grpeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int i,n,len,index,cc,dumpdb=FALSE,ncallers=1;
  char *c,*cs,str[256],cntName[CONTEXT_NAME_SIZE+8],grpName[EXPORT_NAME_SIZE+8];
  char srvName[EXPORT_NAME_SIZE+8],subName[SUBSYSTEM_NAME_SIZE+8],usr[USERNAME_SIZE+8];
  char prefix[16],postfix[16];
  GRPXREF *xref=NULL;
  UINT32 lIP[4];
  NAME16 caller;

  GetCallerInfo(GENS_TAG,&caller,NULL,lIP,NULL,&ncallers);
  strncpy(usr,caller.name,USERNAME_SIZE); usr[USERNAME_SIZE] = 0; strupr(usr);
  if (!stricmp(devProperty,"GROUPS"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    for (i=0; i<nGroupServers && i<(int)dout->dArrayLength; i++)
    {
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = fmtsizeof(dout->dFormat);
          sprintf(str,"/%.32s/%.32s",GRPData[i].context,GRPData[i].name);
          strncpy(dout->data.cptr+i*len,str,len);
          break;
        default:
          return illegal_format;
      }
    }
    dout->dArrayLength = (UINT32)i;
    return 0;
  }
  if (!stricmp(devProperty,"REEXPORT"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dArrayLength > 0)
    {
      memset(cntName,0,CONTEXT_NAME_SIZE);
      switch (din->dFormat)
      {
        case CF_TEXT:
          len = MIN(din->dArrayLength,16);
          break;
        case CF_NAME16:
          len = 16;
          break;
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = 32;
          break;
        default:
          return illegal_format;
      }
      strncpy(cntName,din->data.cptr,len);
      strncpy(grpName,devName,EXPORT_NAME_SIZE);
      cs = cntName;
    }
    else
    {
      if ((cc=splitDeviceName(devName,cntName,grpName)) != 0) return cc;
      cs = cntName;
    }
    if ((c=strrchr(cntName,'.')) != NULL && !isSystemContextDecoration(c))
    { /* context decorated with subsystem name */
      *c = 0;
    }
    if ((c=strchr(cntName,'_')) != NULL)
    { /* could this be "_TEST" or "_SIM" ? */
      if (isSystemContextDecoration(&c[1]))
      { /* seems to be! replace with a '.' */
        *c = '.';
      }
    }
    if ((n=findServerGroup(grpName,cs)) < 0) 
    {
      feclogEx(TINE_LOGLEVEL_WARN,"could not find group %.32s in context %.32s for re-export",grpName,cs);
      return non_existent_elem;
    }
    return reExportGroup(n);
  }
  if (!stricmp(devProperty,"MEMBERS"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dout->dArrayLength == 0) return dimension_error;
    cs = NULL;
    if (din->dArrayLength > 0)
    {
      memset(cntName,0,CONTEXT_NAME_SIZE);
      switch (din->dFormat)
      {
        case CF_TEXT:
          len = MIN(din->dArrayLength,16);
          break;
        case CF_NAME16:
          len = 16;
          break;
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = 32;
          break;
        default:
          return illegal_format;
      }
      strncpy(cntName,din->data.cptr,len);
      strncpy(grpName,devName,EXPORT_NAME_SIZE);
      cs = cntName;
    }
    else
    {
      if ((cc=splitDeviceName(devName,cntName,grpName)) != 0) return cc;
      cs = cntName;
    }
    if ((c=strrchr(cntName,'.')) != NULL && !isSystemContextDecoration(c))
    { /* context decorated with subsystem name */
      *c = 0;
    }
    if ((c=strchr(cntName,'_')) != NULL)
    { /* could this be "_TEST" or "_SIM" ? */
      if (isSystemContextDecoration(&c[1]))
      { /* seems to be! replace with a '.' */
        *c = '.';
      }
    }
    if ((n=findServerGroup(grpName,cs)) < 0) return non_existent_elem;
    for (i=0; i<GRPData[n].size; i++)
    {
      if ((xref = &GRPData[n].xref[i]) == NULL) continue;
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = fmtsizeof(dout->dFormat);
          strncpy(dout->data.cptr+i*len,xref->target,len);
          break;
        case CF_STRUCT:
          if (strncmp(dout->dTag,MemInfoTag,TAG_NAME_SIZE)) return invalid_structure_tag;
          strncpy(((MemInfo *)dout->data.vptr)[i].name,xref->target,EXPORT_NAME_SIZE);
          strncpy(((MemInfo *)dout->data.vptr)[i].devPrefix,xref->devPrefix,8);
          strncpy(((MemInfo *)dout->data.vptr)[i].devPostfix,xref->devPostfix,8);
          ((MemInfo *)dout->data.vptr)[i].index = xref->index;
          break;
        default:
          return illegal_format;
      }
    }
    dout->dArrayLength = (UINT32)i;
    return 0;
  }
  if (!stricmp(devProperty,"REFRESH"))
  {
    if ((access & CA_WRITE) == 0) return illegal_read_write;
    cs = NULL;
    if (din->dArrayLength > 0)
    {
      memset(cntName,0,CONTEXT_NAME_SIZE);
      switch (din->dFormat)
      {
        case CF_TEXT:
          len = MIN(din->dArrayLength,16);
          break;
        case CF_NAME16:
          len = 16;
          break;
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          len = 32;
          break;
        default:
          return illegal_format;
      }
      strncpy(cntName,din->data.cptr,len);
      strncpy(grpName,devName,EXPORT_NAME_SIZE);
      cs = cntName;
    }
    else
    {
      if ((cc=splitDeviceName(devName,cntName,grpName)) != 0) return cc;
      cs = cntName;
    }
    if ((c=strrchr(cntName,'.')) != NULL && !isSystemContextDecoration(c))
    { /* context decorated with subsystem name */
      *c = 0;
    }
    if ((c=strchr(cntName,'_')) != NULL)
    { /* could this be "_TEST" or "_SIM" ? */
      if (isSystemContextDecoration(&c[1]))
      { /* seems to be! replace with a '.' */
        *c = '.';
      }
    }
    if ((n=findServerGroup(grpName,cs)) < 0) return non_existent_elem;
    for (i=0; i<GRPData[n].size; i++)
    {
      if ((xref = &GRPData[n].xref[i]) == NULL) continue;
      nslog("refresh device names for %.32s",xref->target);
      gChkNotAck = TRUE;
      xref->acquired = FALSE;
      xref->disabled = FALSE;
    }
    return reExportGroup(n);
  }  
  if (!stricmp(devProperty,"JOIN"))
  { /* be strict as to the format and order of the input */
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isFec(usr) && !isAdministrator(usr) && !isENS(usr)) return access_denied;
    c = NULL;
    switch (din->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
        len = fmtsizeof(din->dFormat);
        break;
      default:
        nslog("JOIN refused due to illegal format %d\n",din->dFormat);
        return illegal_format;
    }
    memset(prefix,0,16); memset(postfix,0,16);
    memset(subName,0,SUBSYSTEM_NAME_SIZE+4);
    memset(srvName,0,EXPORT_NAME_SIZE+4);
    memset(grpName,0,EXPORT_NAME_SIZE+4);
    memset(cntName,0,CONTEXT_NAME_SIZE+4);
    index = 0;
    switch (din->dArrayLength)
    {
      case 7:
        strncpy(postfix,din->data.cptr+6*len,8);
      case 6:
        strncpy(prefix,din->data.cptr+5*len,8);
      case 5:
        strncpy(subName,din->data.cptr+4*len,SUBSYSTEM_NAME_SIZE);
      case 4:
        strncpy(str,din->data.cptr+3*len,16);
        index = atoi(str);
        if (index < 0) 
        {
          nslog("JOIN refused due to index = %d\n",index);
          return invalid_index;
        }
      case 3:
        strncpy(cntName,din->data.cptr+2*len,len);
        if (strlen(subName) == 0 && (cs=strrchr(cntName,'.')) != NULL && 
            !isSystemContextDecoration(cs))
        { /* context decorated with subsystem name */
          *cs++ = 0;
          strncpy(subName,cs,SUBSYSTEM_NAME_SIZE);
        }
        if ((c=strchr(cntName,'_')) != NULL)
        { /* could this be "_TEST" or "_SIM" ? */
          if (isSystemContextDecoration(&c[1]))
          { /* seems to be! replace with a '.' */
            *c = '.';
          }
        }
        checkContextCase(cntName,nGroupServers);
        c = cntName;
      case 2:
        strncpy(grpName,din->data.cptr+len,len); 
        strncpy(srvName,din->data.cptr,len);
        break;
      default:
        nslog("JOIN refused due to dimension error (%d)\n",din->dArrayLength);
        return dimension_error;
    }
    if (strlen(grpName) == 0 || strlen(srvName) == 0) return argument_list_error;
    nslog("JOIN group %.32s called by server %.32s (context %.32s, index %d)",grpName,srvName,cntName,index);
    if ((n=findServerGroup(srvName,c)) >= 0) 
    { /* a group server is trying to JOIN another group ? */
      feclog("group server %.32s tried to join group %.32s in context %.32s",srvName,grpName,cntName);
      nslog("group server %.32s tried to join group %.32s in context %.32s",srvName,grpName,cntName);
      return not_allowed;
    }
    /* now join the group */
    if ((n=findServerGroup(grpName,c)) < 0) 
    { /* new group ! */
      nslog("group %.32s appears to be NEW (not found among %d registered groups)!",grpName,nGroupServers);
      if (nGroupServers >= GrpDataCapacity)
      {
        nslog("number of groups would exceed capacity %d",GrpDataCapacity);
        return resources_exhausted;
      }
      n = nGroupServers; /* there will be enough reserved space ! */
      strncpy(GRPData[n].name,grpName,EXPORT_NAME_SIZE);
      strncpy(GRPData[n].module,makeModuleName(grpName),EQM_NAME_SIZE); /* grpName definitely contains a name ! */
      strncpy(GRPData[n].context,strlen(cntName) > 0 ? cntName : "DEFAULT",CONTEXT_NAME_SIZE);
      strncpy(GRPData[n].subsystem,subName,SUBSYSTEM_NAME_SIZE);
      if ((GRPData[n].xref=(GRPXREF *)calloc(XREF_RESERVED_SPACE,sizeof(GRPXREF))) == NULL) 
        return out_of_server_memory;
      nslog("prepare new group (%d) /%.32s/%.32s (subsystem %.16s) with equipment module name %.6s",
        n,GRPData[n].context,GRPData[n].name,GRPData[n].subsystem,GRPData[n].module);
      nGroupServers++;
    }
    for (i=0; i<GRPData[n].size; i++)
    {
      if ((xref = &GRPData[n].xref[i]) == NULL) continue;
      if (strnicmp(xref->target,srvName,EXPORT_NAME_SIZE)) continue;
      nslog("target %.32s found among %d members in group %.32s",srvName,GRPData[n].size,grpName);
      break;
    }
    /* so far so good, forward it ... */
    asyncCmdForward(cntName,devName,devProperty,dout,din,access,(BYTE *)lIP);
    if (GRPData[n].size > 0 && i < GRPData[n].size) 
    { /* remote server is simply restarting : re-acquire device names */
      feclog("server %.32s is restarting : re-acquire device names",srvName);
      nslog("server %.32s is restarting : re-acquire device names",srvName);
      gChkNotAck = TRUE;
      GRPData[n].xref[i].acquired = FALSE;
      GRPData[n].xref[i].disabled = FALSE;
      if (strncmp(GRPData[n].xref[i].devPrefix,prefix,8)) 
      {
        strncpy(GRPData[n].xref[i].devPrefix,prefix,8);
        dumpdb = TRUE;
      }
      if (strncmp(GRPData[n].xref[i].devPostfix,postfix,8)) 
      {
        strncpy(GRPData[n].xref[i].devPostfix,postfix,8);
        dumpdb = TRUE;
      }
      if (dumpdb) dumpDatabase();
      /* checkDeviceNamesAsync(); <- let the background task (1 Hz) pick this up */
      return 0;
    }
    /* add member to the group */
    i = GRPData[n].size;
    xref = &GRPData[n].xref[i];
    strncpy(xref->target,srvName,EXPORT_NAME_SIZE);
    if (i > 0 && index == 0) index = getNextIndexFromGroup(n);
    xref->index = index;
    if (strlen(prefix) > 0) strncpy(xref->devPrefix,prefix,8);
    if (strlen(postfix) > 0) strncpy(xref->devPostfix,postfix,8);
    GRPData[n].size++;
    nslog("add new member %.32s to group %.32s at index %d",srvName,grpName,index);
    /* now dump it out (causes immediate readback) */
    dumpDatabase();
    return 0;
  }
  if (!stricmp(devProperty,"SEPARATE"))
  { /* be strict as to the format and order of the input */
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isFec(usr) && !isAdministrator(usr) && !isENS(usr)) return access_denied;
    switch (din->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
        len = fmtsizeof(din->dFormat);
        break;
      default:
        nslog("SEPARATE refused due to illegal format %d\n",din->dFormat);
        return illegal_format;
    }
    c = NULL; cntName[0] = 0;
    switch (din->dArrayLength)
    {
      case 3:
        strncpy(cntName,din->data.cptr+2*len,len);
        if ((c=strrchr(cntName,'.')) != NULL && !isSystemContextDecoration(c))
        { /* context decorated with subsystem name */
          *c = 0;
        }
        if ((c=strchr(cntName,'_')) != NULL)
        { /* could this be "_TEST" or "_SIM" ? */
          if (isSystemContextDecoration(&c[1]))
          { /* seems to be! replace with a '.' */
            *c = '.';
          }
        }
        c = cntName;
      case 2:
        strncpy(grpName,din->data.cptr+len,len);
        strncpy(srvName,din->data.cptr,len);
        break;
      default: 
        nslog("SEPARATE refused due to dimension error (%d)\n",din->dArrayLength);
        return dimension_error;
    }
    if ((n=findServerGroup(grpName,c)) < 0) return non_existent;
    feclog("SEPARATE %.32s server %.32s from group %.32s",c,srvName,grpName);
    nslog("SEPARATE %.32s server %.32s from group %.32s",c,srvName,grpName);
    for (i=0; i<GRPData[n].size; i++)
    {
      if ((xref = &GRPData[n].xref[i]) == NULL) continue;
      if (strnicmp(xref->target,srvName,EXPORT_NAME_SIZE)) continue;
      nslog("member %.32s was found in group -> will be removed",srvName);
      break;
    }
    if (i == GRPData[n].size) 
    {
      feclog("member %.32s was not found in group",srvName);
      return non_existent_fec;
    }
    if (xref != NULL) xref->disabled = TRUE;
    /* so far so good, forward it ... */
    asyncCmdForward(cntName,devName,devProperty,dout,din,access,(BYTE *)lIP);
    dumpDatabase();
    return 0;
  }
  return not_implemented;
}

typedef struct FwdQ
{
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  DTYPE din;
  BYTE payload[512]; /* covers everything ... */
  int access;
  int ntries;
  int pending;
  int sts;
  struct FwdQ *nxt;
  struct FwdQ *prv;
} FWDQ;
typedef struct
{
  char ctx[32];
  char fwd[64];
  FWDQ *fwdq;
  INADR inadr;
  int qlen;
} FWD;
int nforwards = 0;
FWD *fwdLst = NULL;

int retryForwardOnError(int cc)
{
  switch (cc)
  {
    case link_timeout:
    case connection_timeout:
    case operation_busy:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
void chkForwardsList(int idx)
{
  DTYPE dout;
  FWDQ *q;
  int cc;
  char devsrv[128];

  memset(&dout,0,sizeof(DTYPE)); dout.dFormat = CF_NULL;
  while (fwdLst[idx].fwdq != NULL)
  {
    q = fwdLst[idx].fwdq;
    sprintf(devsrv,"%.64s/%.64s",fwdLst[idx].fwd,q->dev);
    cc = ExecLinkEx(devsrv,q->prp,&dout,&q->din,(short)q->access,200);
    if (retryForwardOnError(cc) && q->ntries++ < 60) break;
    if (cc != 0 && q->ntries >= 60)
    {
      feclog("abandon forward %.64s to %.64s : %.32s",q->prp,fwdLst[idx].fwd,cc2str(cc));
    }
    fwdLst[idx].fwdq = q->nxt;
    free(q);
    fwdLst[idx].qlen--;
  }
}
void addtoFwdQueueOld(int idx,char *dev,char *prp,DTYPE *din,int access)
{
  FWDQ *q, **qt;
  int sizeInBytes;
  if (din == NULL) return;
  sizeInBytes = din->dArrayLength * fmtsizeof(din->dFormat);
  if (sizeInBytes == 0) return;
  if (sizeInBytes > 512)
  { /* tilt !!!  */
    feclog("unexpected payload size (%d) in forward property %.64s to %.64s",sizeInBytes,prp,fwdLst[idx].fwd);
    return;
  }
  if (fwdLst[idx].qlen > 25)
  {
    feclog("forward queue %.64s at capacity !",fwdLst[idx].fwd);
    return;
  }
  if ((q=(FWDQ *)calloc(1,sizeof(FWDQ))) == NULL) return;
  strncpy(q->dev,dev,DEVICE_NAME_SIZE);
  strncpy(q->prp,prp,PROPERTY_NAME_SIZE);
  q->access = access;
  q->din.dFormat = din->dFormat;
  q->din.dArrayLength = din->dArrayLength;
  strncpy(q->din.dTag,din->dTag,TAG_NAME_SIZE);
  q->din.data.bptr = q->payload;
  memcpy(q->din.data.bptr,din->data.bptr,sizeInBytes);
  /* find the tail (this needs to be a FIFO queue */
  qt = &fwdLst[idx].fwdq;
  while (*qt != NULL) qt = &(*qt)->nxt;
  q->prv = *qt;
  *qt = q;
  fwdLst[idx].qlen++;
}
void addtoFwdQueue(int idx,char *dev,char *prp,DTYPE *din,int access)
{
  FWDQ *q, *qp, *qt;
  int sizeInBytes;
  if (din == NULL) return;
  sizeInBytes = din->dArrayLength * fmtsizeof(din->dFormat);
  if (din->dFormat == CF_STRUCT && din->dArrayLength == 1)
  {
    if (strlen(din->dTag) == 0) return;
    sizeInBytes *= GetStructSize(din->dTag);
  }
  if (sizeInBytes <= 0) return;
  if (sizeInBytes > 512)
  { /* tilt !!!  */
    feclog("unexpected payload size (%d) in forward property %.64s to %.64s",sizeInBytes,prp,fwdLst[idx].fwd);
    return;
  }
  if (fwdLst[idx].qlen > 25)
  {
    feclog("forward queue %.64s at capacity !",fwdLst[idx].fwd);
    return;
  }
  for (qt=fwdLst[idx].fwdq; qt != NULL; qt=qt->nxt)
  {
    if (strnicmp(qt->dev,dev,DEVICE_NAME_SIZE)) continue;
    if (strnicmp(qt->prp,prp,PROPERTY_NAME_SIZE)) continue;
    if (qt->din.dFormat != din->dFormat) continue;
    if (qt->din.dArrayLength != din->dArrayLength) continue;
    if (strncmp(qt->din.dTag,din->dTag,TAG_NAME_SIZE)) continue;
    if (memcmp(qt->payload,din->data.bptr,sizeInBytes)) continue;
    return; /* already in queue ! */
  }
  if ((q=(FWDQ *)calloc(1,sizeof(FWDQ))) == NULL) return;
  strncpy(q->dev,dev,DEVICE_NAME_SIZE);
  strncpy(q->prp,prp,PROPERTY_NAME_SIZE);
  q->access = access;
  q->din.dFormat = din->dFormat;
  q->din.dArrayLength = din->dArrayLength;
  strncpy(q->din.dTag,din->dTag,TAG_NAME_SIZE);
  q->din.data.bptr = q->payload;
  memcpy(q->din.data.bptr,din->data.bptr,sizeInBytes);
  /* find the tail (this needs to be a FIFO queue */
  qp = qt = fwdLst[idx].fwdq;
  while (qt != NULL) { qp = qt; qt = qt->nxt; }
  q->prv = qp;
  if (qp != NULL) qp->nxt = q; else fwdLst[idx].fwdq = q;
  fwdLst[idx].qlen++;
}
int getForwardsList(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int lst_col=-1,con_col=-1;
  char str[256],hdr[256];
  BYTE *badr;

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (fwdLst != NULL) 
  {
    free(fwdLst);
    fwdLst = NULL;
    nforwards = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((fwdLst=(FWD *)calloc(n,sizeof(FWD))) == NULL) ccerr(out_of_server_memory);
  nforwards = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nforwards;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((lst_col=findcol(hdr,"GENS_LIST")) < 0) ccerr(no_such_line);
      con_col = findcol(hdr,"CONTEXT");
      done = 1;
      continue;
    }
    strncpy(fwdLst[i].fwd,colptr(lst_col,str),64);
    if (con_col >= 0) strncpy(fwdLst[i].ctx,colptr(con_col,str),64);
    if (fwdLst[i].ctx[0] == '*') fwdLst[i].ctx[0] = 0;
    if ((badr=getFwdIpAddr(fwdLst[i].fwd)) == NULL) continue;
    memcpy(&fwdLst[i].inadr,badr,gInAdrLen);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
} 
BYTE *getFwdIpAddr(char *tgt)
{
  char fwdTgt[128],*ctx=NULL,*srv=NULL;
  FecAddrStruct *fec;
  strncpy(fwdTgt,tgt,64);
  if (fwdTgt[0] != '/') return 0;
  ctx = &fwdTgt[1];
  if ((srv=strchr(ctx,'/')) == NULL) return 0;
  *srv++ = 0;
  fec = GetFecAddrFromCache(ctx,srv);
  if (fec == NULL) return 0;
  return (BYTE *)&fec->netAdr;
}
void cbFwd(int id,int cc,void *ref)
{
  FWDQ *q = (FWDQ *)ref;
  char msg[128];
  if (id < 0 || id >= nforwards) return; /* shouldn't be ! */
  if (q == NULL || fwdLst[id].qlen == 0) return; /* no queue ? */
  q->pending = FALSE;
  q->sts = cc;
  if (retryForwardOnError(cc) && q->ntries++ < 10) return; /* keep trying */
  sprintf(msg,"%.32s",cc2str(cc));
  if (q->ntries >= 10) strcat(msg," (abandoned)");
  if (q->prv == NULL) 
  { /* then I'm at the head */
    fwdLst[id].fwdq = q->nxt; 
  }
  else 
  { /* splice */
    q->prv->nxt = q->nxt;
    if (q->nxt != NULL) q->nxt->prv = q->prv;
  }
  if (fwdLst[id].fwdq != NULL) fwdLst[id].fwdq->prv = NULL; /* ensure this */
  fwdLst[id].qlen--;
  feclog("forward %.64s to %.64s : %s -> queue length now %d",
    q->prp,fwdLst[id].fwd,msg,fwdLst[id].qlen);
  if (fwdLst[id].fwdq == NULL && fwdLst[id].qlen > 0)
  {
    fwdLst[id].qlen = 0;
    feclog("adjust forward %.64s queue length to 0!",fwdLst[id].fwd);
  }
  if (fwdLst[id].qlen == 0 && fwdLst[id].fwdq != NULL)
  {
    fwdLst[id].fwdq = NULL;
    feclog("force forward %.64s queue to NULL!",fwdLst[id].fwd);
  }
  free(q);
}
void chkForwardsQueue(int idx)
{
  DTYPE dout;
  FWDQ *q;
  int id,cc;
  char devsrv[256];

  DTYPEZERO(dout);
  for (q=fwdLst[idx].fwdq; q != NULL; q=q->nxt)
  {
    if (q->pending) continue; /* still waiting for an answer ... */
    sprintf(devsrv,"%.64s/%.64s",fwdLst[idx].fwd,q->dev);
    id = AttachLinkEx2(devsrv,q->prp,&dout,&q->din,(short)q->access,500,cbFwd,CM_SINGLE,idx,q);
    if (id < 0) 
    {
      cc = -id;
      feclog("Forward %.64s to %.128s : %.32s",q->prp,devsrv,cc2str(cc));
      q->ntries = 100;
    }
    q->pending = TRUE;
  }
}
int asyncCmdForward(char *context,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int access,BYTE *badr)
{
  int i;
  dout = dout;
  for (i=0; i<nforwards; i++)
  {
    if (!memcmp(&fwdLst[i].inadr,badr,gInAdrLen)) continue;
    if (strlen(fwdLst[i].ctx) > 0 && strstr(fwdLst[i].ctx,context) == NULL) continue;
    addtoFwdQueue(i,devName,devProperty,din,access);
  }
  return 0;
}
int dumpFwdQueues(int i1,int i2, int i3, int i4)
{
  int i;
  FWDQ *q;
  i1 = i1; i2 = i2; i3 = i3; i4 = i4;
  if (nforwards == 0)
  {
    ttyoutput("there is no forwarding information registered");
    return 0;
  }
  ttyoutput("forward queue information:");
  for (i=0; i<nforwards; i++)
  {
    ttyoutput("%.64s : current queue length %d",fwdLst[i].fwd,fwdLst[i].qlen);
    for (q=fwdLst[i].fwdq; q != NULL; q=q->nxt)
    {
      ttyoutput("  %.64s[%.64s] : %.32s, %d tries so far (pending: %s)",
        q->dev,q->prp,cc2str(q->sts),q->ntries,(q->pending?"TRUE":"FALSE"));
    }
  }
  return 0;
}
#define ENSLOGDEPTH 5000
void nslog(char *text,...)
{
  static char fn[64] = "";
  static char bfn[64] = "";
  FILE *fp=NULL;
  time_t t = time(NULL);
  char str[512];
  long flen;
  int cc=0;
  va_list args;

  if (strlen(fn) == 0) 
  {
    sprintf(fn,"%.32s.log",gFecName);
    sprintf(bfn,"%.32s.bak",gFecName);
    strlwr(fn); strlwr(bfn);
  }
  sprintf(str,"%s[%.16s] ",getDataTimeString((double)t,0),gFecName);
# if defined(UNIX)
  va_start(args,text);
  vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# elif defined(WIN32)
  va_start(args,text);
  _vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# else
  if (strlen(text) > 128) 
  {
    if (strchr(text,'%') == NULL)
    { /* then truncate */
      strncpy(&str[strlen(str)],text,256); str[256] = 0;
    }
    ccerr(string_too_long);
  }
  else
  {
    va_start(args,text);
    vsprintf(&str[strlen(str)],text,args);
    va_end(args);
  }
# endif
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  fseek(fp, 0L, SEEK_END); flen = ftell(fp);
  if (flen > ENSLOGDEPTH*80)
  {
    fclose(fp); remove(bfn); rename(fn,bfn);
    if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  }
  fwrite(str,strlen(str),1,fp);
  if (debug) dbglog(str);
err:
  if (cc) dbglog("nslog: %.32s",erlst[cc]);
  if (fp != NULL) fclose(fp);
  return;
}
