/*
 * File devsrv.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "app_version.h"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include "tine.h"

void PostSystemInit(void);
void PreSystemInit(void);

extern int g_isTerminating;
extern UINT32 gMininumFreeDiskSpace;
extern int globalsWatchdog;
extern int requireNetworkAccess;
extern char globalsMcastAddr[32];
extern char adbFileName[64];

extern char thisFecName[FEC_NAME_SIZE];
extern char thisContext[CONTEXT_NAME_SIZE];
extern char thisLocation[FEC_LOCATION_SIZE];
extern UINT16 thisPortOffset;

char *usageText = "usage: glbsrv [/f=<fecname> /p=<port offset> /l=<location> /n=<database file name> /k=<minimumFreeDiskSpaceInKbytes> /g=\"WATCHDOG\" /a=<globals mcast address> /r=\"TRUE\"(legacy mcast rules) /e=\"TRUE\"(require network access) /d=<debug level>";
int main(int argc, char *argv[])
{
  int i,cc;
  char *c, scratch[256];

  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'f':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisFecName,c,FEC_NAME_SIZE);
        }
        continue;
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisContext,c,CONTEXT_NAME_SIZE);
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisLocation,c,FEC_LOCATION_SIZE);
        }
        continue;
      case 'p':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          thisPortOffset = atoi(scratch);      
        }
        continue;
      case 'n':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(adbFileName,c,64);
        }
        continue;
      case 'k':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          gMininumFreeDiskSpace = strtol(scratch,NULL,0);      
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
        }
        continue;
      case 'g':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          if (stricmp(scratch,"WATCHDOG") == 0)
          {
            globalsWatchdog = TRUE;
          }
        }
        continue;
      case 'e':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          if (stricmp(scratch,"TRUE") == 0)
          {
            requireNetworkAccess = TRUE;
          }
        }
        continue;
      case 'a':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(globalsMcastAddr,c,32);
        }
        continue;
      case 'r':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          if (stricmp(scratch,"TRUE") == 0)
          {
            SetUseMCastLegacyRules(TRUE);
          }
        }
        continue;
      case '?':
      case 'h':
        printf("%s",usageText);
        exit(0);
      default:
        continue;
    }
  }

  AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);
  SetPreSystemInitFunction(PreSystemInit);
  SetPostSystemInitFunction(PostSystemInit);
  SetSendLegacyGlobals(TRUE); /* for the time being ... */
  if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }
  printf("Server running (debug level %d)\n",tineDebug);
  for(;;)
  {
    if (g_isTerminating) ttyquit(NULL,NULL);
    SystemCycle(TRUE);
  }
  return 0;
}
