// =============================================================
// ==
// ==   Database consistency check routine: checkdb.c
// ==
// =============================================================
#include "stdlib.h"
#include "math.h"
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <ctype.h>
#include <signal.h>
#include <tine.h>
#include "archive.h"

DAQRecord *daqTbl;
int  nrecords, nparam;
int archPending = 0;
int getNrActiveFecs(char *fname);
void getFecInfo(void);
int wrtFlg,archdebug=0;
char cbuff[NETMEX_BUF_SIZE];
int totalNrRecords = 0;
int g_isTerminating = 0;

#define WRITE     -1
#define READ       1
#define MAXFILES   100

char dbfile[] = "broadcst.csv";
//char ArchiveDBpath[] = "../database/";
//char *ArchiveDatapath = "../archdata/";
extern char ArchiveDBpath[];
extern char *ArchiveDatapath;

#define logerr(txt) { perror("open"); strcpy(s,txt); goto error; }

int AppendProblemsLog(char *s) { return 0;}

//============= General PKTR archiving globals ===========================

FILE *wfp, *rfp;
int Hwfp;

char sbuf[20];
char timebuf[9], datebuf[9];

int fileNo = 0;
char *struprXXX(char *s)
{
  char *c;
  for (c=s; c && *c; c++) if (*c >= 'a' && *c <= 'z') *c -= 32;
  return s;
}
char *strlwrXXX(char *s)
{
  char *c;
  for (c=s; c && *c; c++) if (*c >= 'A' && *c <= 'Z') *c += 32;
  return s;
}
char *strrevXXX(char *s)
{
  char c;
  register int i,len = strlen(s);
  for (i=0; i<len/2; i++) { c = s[i]; s[i] = s[len-1-i]; s[len-1-i] = c; }
  return s;
}
int stricmpXXX(char *s1,char *s2)
{
  char *c1,*c2;
  for (c1=s1,c2=s2; *c1 != 0 && *c2 != 0; c1++,c2++)
  if (toupper(*c1) != toupper(*c2)) break;
  return *c2 - *c1;
}
char *_strdateXXX(char *buf)
{
  time_t t;
  struct tm *tm;
  time(&t); tm=localtime(&t);
  sprintf(buf,"%02d/%02d/%02d",(tm->tm_mon)+1,tm->tm_mday,tm->tm_year);
  return buf;
}
char *_strtimeXXX(char *buf)
{
  time_t t;
  struct tm *tm;
  time(&t); tm=localtime(&t);
  sprintf(buf,"%02d:%02d:%02d",tm->tm_hour,tm->tm_min,tm->tm_sec);
  return buf;
}
int findcolXXX(char *str,char *tag)
{
  char *c;
  int i;

  for (i=0,c=str; strlen(c); i++,c++)
  {
    while (*c == '\"') { c = strchr(++c,'\"'); if (c++ == NULL) return -1; }
    strupr(c);
    if (!strncmp(c,tag,MAX(strcspn(c,",\n"),strlen(tag)))) return i;
    if ((c=strpbrk(c,",\n")) == NULL) return -1;
  }
  return -1;
}
char *colptrXXX(int col,char *str)
{
  static char buf[128];
  char *c;
  int i,sslen;

  for (i=0,c=str; i < col && strlen(c); i++,c++)
  {
    while (*c == '\"')
    {
      c = strchr(++c,'\"');
      if (c++ == NULL)
      {
        printf("%s : missing quote in column %d\n",str,col);
        exit(1);
      }
    }
    if ((c=strpbrk(c,",")) == NULL)
    {
      printf("%s : missing column %d\n",str,col);
      exit(1);
    }
  }
  if (*c == '\"') sslen = strcspn(++c,"\"");
  else sslen = strcspn(c,",\n");
  if (sslen > 126)
  {
    printf("sub string \"%s\" too long\n",c);
    exit(1);
  }
  strncpy(buf,c,i=MIN(sizeof(buf)-1,sslen));
  buf[i] = 0;
  return buf;
}
void AppendArchiveLog(char *str)
{
  int hfp;
  time_t t = time(NULL);
  char s[80];
  FILE *fp = NULL;

  s[0] = 0;
  if ((fp=fopen("archive.log","a")) == NULL) logerr("LOG: open");
  strncpy(s,str,78);
  if (s[strlen(s)-1] == '\n') s[strlen(s)-1] = 0;
  fprintf(fp,"%s : %s",s,ctime(&t));

error: if (fp) fclose(fp);
  return;
}
void dumpFecs(void)
{
  int i,j;
  
  for (i=0; i<nrecords; i++)
  {
    printf("%d: %.32s %.64s %.64s %d %d\n",
           i,daqTbl[i].srv,daqTbl[i].prp,daqTbl[i].dev,
           daqTbl[i].len,daqTbl[i].reclen);
    for (j=0; j<daqTbl[i].nparam; j++)
      printf("\t%.64s %d %d\n",
        daqTbl[i].daqParam[j].param,daqTbl[i].daqParam[j].fmt,daqTbl[i].daqParam[j].size);
  }
}
int dbInit(char *dbfile)
{
  int i, cc=0;

  totalNrRecords = getNrActiveFecs(dbfile);

  if ((daqTbl=(DAQRecord *)calloc(totalNrRecords,sizeof(DAQRecord))) == NULL) ccerr(out_of_server_memory);

if ((prpAlias=(PRP_ALIAS *)calloc(totalNrRecords,sizeof(PRP_ALIAS))) ==
NULL) ccerr(out_of_server_memory);

  if ((subSystems=(SUB_SYS *)calloc(totalNrRecords,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);

  nrecords=archLoad(dbfile,0);    // number of fecs being archived
  nparam=paramLoad(nrecords,0);    // number of parameters archived
  dumpFecs();
err:
  if (cc) printf("Init error: %s",erlst[cc]);
  return cc;
}
int main(int argc,char *argv[])
{
  switch ( argc )
  {
    case 2: dbInit(argv[1]); break;
    default: printf("Usage: checkdb <database file>\n"); break;
  }
  return 0;
}

