/*
 * mstlib.c (motor steering library)
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "p3meqm.h"
#include "cdi.h"
#include "mstlib.h"


int getMotorStatus(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  short sval;

  sprintf(devname,"/localhost/cdi/Motor%d.Status",motorNr+1);
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = &sval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) sval = -1;

  return sval;
}
float getMotorVelocityLowerLimit(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  float fval;

  sprintf(devname,"/localhost/cdi/Motor%d.LwrLmtV",motorNr+1);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &fval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) fval = 0;

  return fval;
}

float getMotorVelocityUpperLimit(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  float fval;

  sprintf(devname,"/localhost/cdi/Motor%d.UprLmtV",motorNr+1);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &fval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) fval = 0;

  return fval;
}

float getMotorAccelerationUpperLimit(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  float fval;

  sprintf(devname,"/localhost/cdi/Motor%d.UprLmtA",motorNr+1);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &fval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) fval = 0;

  return fval;
}
double getMotorPosition(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  float fval;

  sprintf(devname,"/localhost/cdi/Motor%d.rCurPos",motorNr+1);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &fval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) fval = 0;

  return (double)fval;
}
double getMotorTargetPosition(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  float fval;

  sprintf(devname,"/localhost/cdi/Motor%d.rTgtPos",motorNr+1);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &fval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) fval = 0;

  return (double)fval;
}
int setMotorTargetPosition(int motorNr,double newPosition)
{
  DTYPE dout,din;
  int cc = 0;
  char devname[64];
  float farr[6];

  farr[0] = 2; // Soll Position register
  farr[1] = 1;  // execute flag
  farr[2] = (float)newPosition;
  farr[4] = -1;

  sprintf(devname,"/localhost/cdi/Motor%d.fltCMD",motorNr+1);
  din.dFormat = CF_FLOAT;
  din.dArrayLength = 4;
  din.data.fptr = farr;

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 6;
  dout.data.fptr = farr;
  cc = ExecLinkEx(devname,"SEND.RECV.ATOM",&dout,&din,CA_WRITE | CA_READ,1000);
  if (cc != 0) return cc;

  if (farr[4] != 0 || farr[5] != 0) cc = hardware_error;
  return cc;
}
double getMotorVelocity(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  float fval;

  sprintf(devname,"/localhost/cdi/Motor%d.maxVel",motorNr+1);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &fval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) fval = 0;

  return (double)fval;
}
int setMotorVelocity(int motorNr,double newVelocity)
{
  DTYPE dout,din;
  int cc = 0;
  char devname[64];
  float farr[6];

  farr[0] = 39; // MAX_VELOCITY register
  farr[1] = 1;  // execute flag
  farr[2] = (float)newVelocity;
  farr[4] = -1;

  sprintf(devname,"/localhost/cdi/Motor%d.fltCMD",motorNr+1);
  din.dFormat = CF_FLOAT;
  din.dArrayLength = 4;
  din.data.fptr = farr;

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 6;
  dout.data.fptr = farr;

  cc = ExecLinkEx(devname,"SEND.RECV.ATOM",&dout,&din,CA_WRITE | CA_READ,1000);
  if (cc != 0) return cc;

  if (farr[4] != 0 || farr[5] != 0) cc = hardware_error;
  return cc;
}
double getMotorAcceleration(int motorNr)
{
  DTYPE dout;
  int cc = 0;
  char devname[64];
  float fval;

  sprintf(devname,"/localhost/cdi/Motor%d.maxAcc",motorNr+1);
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &fval;

  cc = ExecLinkEx(devname,"RECV",&dout,NULL,CA_READ,100);

  if (cc != 0) fval = 0;

  return (double)fval;
}
int setMotorAcceleration(int motorNr,double newAcceleration)
{
  DTYPE dout,din;
  int cc = 0;
  char devname[64];
  float farr[6];

  farr[0] = 40; // MAX_ACCELERATION register
  farr[1] = 1;  // execute flag
  farr[2] = (float)newAcceleration;
  farr[4] = -1;

  sprintf(devname,"/localhost/cdi/Motor%d.fltCMD",motorNr+1);
  din.dFormat = CF_FLOAT;
  din.dArrayLength = 4;
  din.data.fptr = farr;

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 6;
  dout.data.fptr = farr;
  cc = ExecLinkEx(devname,"SEND.RECV.ATOM",&dout,&din,CA_WRITE | CA_READ,1000);
  if (cc != 0) return cc;

  if (farr[4] != 0 || farr[5] != 0) cc = hardware_error;
  return cc;
}
