/*
 * mstlib.c (motor steering library)
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "mstlib.h"

#define quit(i) { feclog("MST: %s",cc2str(i)); exit(1); }
void registerStructs(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;

  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,motName),64,CF_TEXT,"name")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,stsBits),1,CF_INT32,"statusWord")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,stsAxis),1,CF_INT32,"statusAxis")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,stsCoupled),1,CF_INT32,"statusCoupl")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,stsClbr),1,CF_INT32,"statusCalibr")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,idxCntlLoop),1,CF_INT32,"indexCntrLoop")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,stateClbr),1,CF_INT32,"stateCalibr")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,iSpare7),1,CF_INT32,"intResv7")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,iSpare8),1,CF_INT32,"intResv8")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,pos),1,CF_FLOAT,"position")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,tgt),1,CF_FLOAT,"target")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,lag),1,CF_FLOAT,"lag")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,vel),1,CF_FLOAT,"velocity")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,load),1,CF_FLOAT,"load")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,ovrRide),1,CF_FLOAT,"overRide")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,fSpare7),1,CF_FLOAT,"fltResv7")) quit(out_of_local_memory);
  if (addFieldToStruct("MotSts",OFFSETIN(MotorStatus,fSpare8),1,CF_FLOAT,"fltResv8")) quit(out_of_local_memory);
  /* terminate the structure definition like this! */
  if (sealTaggedStruct("MotSts",sizeof(MotorStatus),NUM_MOTORS_MAX)) quit(out_of_local_memory);
}

int getifcnNotImpl(int i,int *v)
{
  i = i; v = v;
  return not_implemented;
}
double getdfcnNotImpl(int i,double *v)
{
  i = i; v = v;
  return not_implemented;
}
int setdfcnNoImpl(int i,double v)
{
  i = i; v = v;
  return not_implemented;
}
int mstNumMotors = 0;
NAME64 *mstMotorNames = NULL;

int sizMstStsList = 0;
NAME32 *mstStsList = NULL;

GETGRPFCNP getGroupMembersFcn = NULL;
MST_EXPORT void SetGetGroupMembersDispatch(GETGRPFCNP fcn)
{
  getGroupMembersFcn = fcn;
}
SETGRPFCNP setGroupMembersFcn = NULL;
MST_EXPORT void SetSetGroupMembersDispatch(SETGRPFCNP fcn)
{
  setGroupMembersFcn = fcn;
}

GETPFCNP getIncrementMoveStatusFcn = NULL;
MST_EXPORT void SetGetIncrementMoveStatusDispatch(GETPFCNP fcn)
{
  getIncrementMoveStatusFcn = fcn;
}
GETPFCNP getIncrementGroupMoveStatusFcn = NULL;
MST_EXPORT void SetGetIncrementGroupMoveStatusDispatch(GETPFCNP fcn)
{
  getIncrementGroupMoveStatusFcn = fcn;
}
GETPFCNP getIncrementGroupMoveProgressFcn = NULL;
MST_EXPORT void SetGetIncrementGroupMoveProgressDispatch(GETPFCNP fcn)
{
  getIncrementGroupMoveProgressFcn = fcn;
}
GETPFCNP getControlGroupMoveStatusFcn = NULL;
MST_EXPORT void SetGetControlGroupMoveStatusDispatch(GETPFCNP fcn)
{
  getControlGroupMoveStatusFcn = fcn;
}
GETPFCNP getControlGroupMoveProgressFcn = NULL;
MST_EXPORT void SetGetControlGroupMoveProgressDispatch(GETPFCNP fcn)
{
  getControlGroupMoveProgressFcn = fcn;
}
GETPFCNP getGroupMoveStatusFcn = NULL;
MST_EXPORT void SetGetGroupMoveStatusDispatch(GETPFCNP fcn)
{
  getGroupMoveStatusFcn = fcn;
}
GETPFCNP getGroupMoveProgressFcn = NULL;
MST_EXPORT void SetGetGroupMoveProgressDispatch(GETPFCNP fcn)
{
  getGroupMoveProgressFcn = fcn;
}
GETPFCNP getMoveToStatusFcn = NULL;
MST_EXPORT void SetGetMoveToStatusDispatch(GETPFCNP fcn)
{
  getMoveToStatusFcn = fcn;
}
GETPFCNP getMoveToProgressFcn = NULL;
MST_EXPORT void SetGetMoveToProgressDispatch(GETPFCNP fcn)
{
  getMoveToProgressFcn = fcn;
}
GETPFCNP getStepToStatusFcn = NULL;
MST_EXPORT void SetGetStepToStatusDispatch(GETPFCNP fcn)
{
  getStepToStatusFcn = fcn;
}
GETPFCNP getStepToProgressFcn = NULL;
MST_EXPORT void SetGetStepToProgressDispatch(GETPFCNP fcn)
{
  getStepToProgressFcn = fcn;
}

GETSPFCNP getScanMoveToStatusFcn = NULL;
MST_EXPORT void SetGetScanMoveToStatusDispatch(GETSPFCNP fcn)
{
  getScanMoveToStatusFcn = fcn;
}
SETSPFCNP setScanMoveToStatusFcn = NULL;
MST_EXPORT void SetSetScanMoveToStatusDispatch(SETSPFCNP fcn)
{
  setScanMoveToStatusFcn = fcn;
}

GETSPFCNP getScanGroupMoveToStatusFcn = NULL;
MST_EXPORT void SetGetScanGroupMoveToStatusDispatch(GETSPFCNP fcn)
{
  getScanGroupMoveToStatusFcn = fcn;
}
SETSPFCNP setScanGroupMoveToStatusFcn = NULL;
MST_EXPORT void SetSetScanGroupMoveToStatusDispatch(SETSPFCNP fcn)
{
  setScanGroupMoveToStatusFcn = fcn;
}
GETPFCNP getScanMoveToProgressFcn = NULL;
MST_EXPORT void SetGetScanMoveToProgressDispatch(GETPFCNP fcn)
{
  getScanMoveToProgressFcn = fcn;
}

SETPFCNP setIncrementMoveStatusFcn = NULL;
MST_EXPORT void SetSetIncrementMoveStatusDispatch(SETPFCNP fcn)
{
  setIncrementMoveStatusFcn = fcn;
}
SETPFCNP setIncrementGroupMoveStatusFcn = NULL;
MST_EXPORT void SetSetIncrementGroupMoveStatusDispatch(SETPFCNP fcn)
{
  setIncrementGroupMoveStatusFcn = fcn;
}
SETPFCNP setControlGroupMoveStatusFcn = NULL;
MST_EXPORT void SetSetControlGroupMoveStatusDispatch(SETPFCNP fcn)
{
  setControlGroupMoveStatusFcn = fcn;
}
SETPFCNP setGroupMoveStatusFcn = NULL;
MST_EXPORT void SetSetGroupMoveStatusDispatch(SETPFCNP fcn)
{
  setGroupMoveStatusFcn = fcn;
}
SETPFCNP setMoveToStatusFcn = NULL;
MST_EXPORT void SetSetMoveToStatusDispatch(SETPFCNP fcn)
{
  setMoveToStatusFcn = fcn;
}
SETPFCNP setStepToStatusFcn = NULL;
MST_EXPORT void SetSetStepToStatusDispatch(SETPFCNP fcn)
{
  setStepToStatusFcn = fcn;
}
SETPFCNP setFreeMoveToStatusFcn = NULL;
MST_EXPORT void SetSetFreeMoveToStatusDispatch(SETPFCNP fcn)
{
  setFreeMoveToStatusFcn = fcn;
}
SETPFCNP setMicroMoveStatusFcn = NULL;
MST_EXPORT void SetSetMicroMoveStatusDispatch(SETPFCNP fcn)
{
  setMicroMoveStatusFcn = fcn;
}
MST_EXPORT int SetMotorList(NAME64 *names,int num)
{
  int i, cc = 0;
  if (num < 1) return out_of_range;
  if ((mstMotorNames=(NAME64 *)calloc(num,sizeof(NAME64))) == NULL) return out_of_local_memory;
  mstNumMotors = num;
  for (i=0; i<num; i++) strncpy(mstMotorNames[i].name,names[i].name,64);
  return 0;
}
MST_EXPORT int SetStatusList(NAME32 *lst,int siz)
{
  int i, cc = 0;
  if (siz < 1) return out_of_range;
  if ((mstStsList=(NAME32 *)calloc(siz,sizeof(NAME32))) == NULL) return out_of_local_memory;
  sizMstStsList = siz;
  for (i=0; i<siz; i++) strncpy(mstStsList[i].name,lst[i].name,32);
  return 0;
}
CMDFCNP initMotorHdwFcn = NULL;
MST_EXPORT void SetInitMotorHardwareDispatch(CMDFCNP fcn)
{
  initMotorHdwFcn = fcn;
}

GETIFCNP getMotorOnlineFcn = NULL;
MST_EXPORT void SetGetMotorOnlineDispatch(GETIFCNP fcn)
{
  getMotorOnlineFcn = fcn;
}
SETIFCNP setMotorOnlineFcn = NULL;
MST_EXPORT void SetSetMotorOnlineDispatch(SETIFCNP fcn)
{
  setMotorOnlineFcn = fcn;
}

GETIFCNP resetMotorFcn = NULL;
MST_EXPORT void SetResetMotorDispatch(GETIFCNP fcn)
{
  resetMotorFcn = fcn;
}
GETIFCNP calibrateMotorFcn = NULL;
MST_EXPORT void SetCalibrateMotorDispatch(GETIFCNP fcn)
{
  calibrateMotorFcn = fcn;
}
SETDFCNP calibrateToOffsetFcn = NULL;
MST_EXPORT void SetCalibrateToOffsetDispatch(SETDFCNP fcn)
{
  calibrateToOffsetFcn = fcn;
}
GETDFCNP getMotorStepsPerMUFcn = NULL;
MST_EXPORT void SetGetMotorStepsPerMUDispatch(GETDFCNP fcn)
{
  getMotorStepsPerMUFcn = fcn;
}
GETDFCNP getMotorStepsPerRadianFcn = NULL;
MST_EXPORT void SetGetMotorStepsPerRadianDispatch(GETDFCNP fcn)
{
  getMotorStepsPerRadianFcn = fcn;
}
GETIFCNP getMotorMicroStepsPerStepFcn = NULL;
MST_EXPORT void SetGetMotorMicroStepsPerStepDispatch(GETIFCNP fcn)
{
  getMotorMicroStepsPerStepFcn = fcn;
}
GETIFCNP getMotorStatusFcn = NULL;
MST_EXPORT void SetGetMotorStatusDispatch(GETIFCNP fcn)
{
  getMotorStatusFcn = fcn;
}
GETIFCNP getGroupMotorStatusFcn = NULL;
MST_EXPORT void SetGetGroupMotorStatusDispatch(GETIFCNP fcn)
{
  getGroupMotorStatusFcn = fcn;
}
GETSTSFCNP getMotorStatusBlockFcn = NULL;
MST_EXPORT void SetGetMotorStatusBlockDispatch(GETSTSFCNP fcn)
{
  getMotorStatusBlockFcn = fcn;
}

GETDFCNP getMotorPositionMinimumFcn = NULL;
MST_EXPORT void SetGetMotorPositionMinimumDispatch(GETDFCNP fcn)
{
  getMotorPositionMinimumFcn = fcn;
}
GETDFCNP getMotorPositionMaximumFcn = NULL;
MST_EXPORT void SetGetMotorPositionMaximumDispatch(GETDFCNP fcn)
{
  getMotorPositionMaximumFcn = fcn;
}
GETDFCNP getMotorPositionOffsetFcn = NULL;
MST_EXPORT void SetGetMotorPositionOffsetDispatch(GETDFCNP fcn)
{
  getMotorPositionOffsetFcn = fcn;
}
GETDFCNP getMotorPositionScaleFcn = NULL;
MST_EXPORT void SetGetMotorPositionScaleDispatch(GETDFCNP fcn)
{
  getMotorPositionScaleFcn = fcn;
}
GETDFCNP getMotorVelocityLowerLimitFcn = NULL;
MST_EXPORT void SetGetMotorVelocityLowerLimitDispatch(GETDFCNP fcn)
{
  getMotorVelocityLowerLimitFcn = fcn;
}
GETDFCNP getMotorVelocityUpperLimitFcn = NULL;
MST_EXPORT void SetGetMotorVelocityUpperLimitDispatch(GETDFCNP fcn)
{
  getMotorVelocityUpperLimitFcn = fcn;
}
GETDFCNP getMotorVelocityFcn = NULL;
MST_EXPORT void SetGetMotorVelocityDispatch(GETDFCNP fcn)
{
  getMotorVelocityFcn = fcn;
}
GETDFCNP getMotorAccelerationFcn = NULL;
MST_EXPORT void SetGetMotorAccelerationDispatch(GETDFCNP fcn)
{
  getMotorAccelerationFcn = fcn;
}
GETDFCNP getMotorPositionFcn = NULL;
MST_EXPORT void SetGetMotorPositionDispatch(GETDFCNP fcn)
{
  getMotorPositionFcn = fcn;
}
GETDFCNP getGroupMotorPositionFcn = NULL;
MST_EXPORT void SetGetGroupMotorPositionDispatch(GETDFCNP fcn)
{
  getGroupMotorPositionFcn = fcn;
}
SETDFCNP setMotorPositionFcn = NULL;
MST_EXPORT void SetSetMotorPositionDispatch(SETDFCNP fcn)
{
  setMotorPositionFcn = fcn;
}
GETDFCNP getMotorTargetPositionFcn = NULL;
MST_EXPORT void SetGetMotorTargetPositionDispatch(GETDFCNP fcn)
{
  getMotorTargetPositionFcn = fcn;
}
GETDFCNP getMotorTargetStepsFcn = NULL;
MST_EXPORT void SetGetMotorTargetStepsDispatch(GETDFCNP fcn)
{
  getMotorTargetStepsFcn = fcn;
}
GETDFCNP getMotorTargetStepSizeFcn = NULL;
MST_EXPORT void SetGetMotorTargetStepSizeDispatch(GETDFCNP fcn)
{
  getMotorTargetStepSizeFcn = fcn;
}
GETIFCNP getMotorTargetMicroPositionFcn = NULL;
MST_EXPORT void SetGetMotorTargetMicroPositionDispatch(GETIFCNP fcn)
{
  getMotorTargetMicroPositionFcn = fcn;
}
GETDFCNP getMotorAccelerationUpperLimitFcn = NULL;
MST_EXPORT void SetGetMotorAccelerationUpperLimitDispatch(GETDFCNP fcn)
{
  getMotorAccelerationUpperLimitFcn = fcn;
}
GETDFCNP getMotorCurrentFcn = NULL;
MST_EXPORT void SetGetMotorCurrentDispatch(GETDFCNP fcn)
{
  getMotorCurrentFcn = fcn;
}
GETDFCNP getGroupMotorTargetPositionFcn = NULL;
MST_EXPORT void SetGetGroupMotorTargetPositionDispatch(GETDFCNP fcn)
{
  getGroupMotorTargetPositionFcn = fcn;
}

SETDFCNP setMotorVelocityFcn = NULL;
MST_EXPORT void SetSetMotorVelocityDispatch(SETDFCNP fcn)
{
  setMotorVelocityFcn = fcn;
}
SETDFCNP setMotorAccelerationFcn = NULL;
MST_EXPORT void SetSetMotorAccelerationDispatch(SETDFCNP fcn)
{
  setMotorAccelerationFcn = fcn;
}
SETDFCNP setMotorPositionOffsetFcn = NULL;
MST_EXPORT void SetSetMotorPositionOffsetDispatch(SETDFCNP fcn)
{
  setMotorPositionOffsetFcn = fcn;
}
SETDFCNP setMotorPositionScaleFcn = NULL;
MST_EXPORT void SetSetMotorPositionScaleDispatch(SETDFCNP fcn)
{
  setMotorPositionScaleFcn = fcn;
}
SETDFCNP setMotorTargetPositionFcn = NULL;
MST_EXPORT void SetSetMotorTargetPositionDispatch(SETDFCNP fcn)
{
  setMotorTargetPositionFcn = fcn;
}
SETDFCNP setMotorTargetStepsFcn = NULL;
MST_EXPORT void SetSetMotorTargetStepsDispatch(SETDFCNP fcn)
{
  setMotorTargetStepsFcn = fcn;
}
SETDFCNP setMotorTargetStepSizeFcn = NULL;
MST_EXPORT void SetSetMotorTargetStepSizeDispatch(SETDFCNP fcn)
{
  setMotorTargetStepSizeFcn = fcn;
}
SETDFCNP setGroupMotorTargetPositionFcn = NULL;
MST_EXPORT void SetSetGroupMotorTargetPositionDispatch(SETDFCNP fcn)
{
  setGroupMotorTargetPositionFcn = fcn;
}

SETIFCNP setMotorTargetMicroPositionFcn = NULL;
MST_EXPORT void SetSetMotorTargetMicroPositionDispatch(SETIFCNP fcn)
{
  setMotorTargetMicroPositionFcn = fcn;
}
SETDFCNP setMotorCurrentFcn = NULL;
MST_EXPORT void SetSetMotorCurrentDispatch(SETDFCNP fcn)
{
  setMotorCurrentFcn = fcn;
}
GETSFCNP getMotorUnitsFcn = NULL;
MST_EXPORT void SetGetMotorUnitsDispatch(GETSFCNP fcn)
{
  getMotorUnitsFcn = fcn;
}
GETDFCNP getGroupGearRatioFcn = NULL;
MST_EXPORT void SetGetGroupGearRatioDispatch(GETDFCNP fcn)
{
  getGroupGearRatioFcn = fcn;
}
SETDFCNP setGroupGearRatioFcn = NULL;
MST_EXPORT void SetSetGroupGearRatioDispatch(SETDFCNP fcn)
{
  setGroupGearRatioFcn = fcn;
}
GETERRFCNP getLastHdwErrFcn = NULL;
MST_EXPORT void SetGetLastHardwareErrorDispatch(GETERRFCNP fcn)
{
  getLastHdwErrFcn = fcn;
}

MotorGroup mstGroups[NUM_MOTORGROUPS_MAX];
int mstNumGroups = 0;

int gMstSetAlarms = TRUE;
MST_EXPORT void SetAllowMstAlarms(int value)
{
  gMstSetAlarms = value;
}
XMA *xmaLst = NULL;
int xmaLstSize = 0;
EQMFCNP xmaDispatch = NULL;
MST_EXPORT int SetXMAList(XMA *list,int listSize,EQMFCNP fcn)
{
  int cc, i, num;
  short acc, atyp = AT_CHANNEL;
  char desc[256];
  DTYPE dout,din;

  xmaLst = list;
  xmaLstSize = listSize;

  num = gNumMotors;
  if (num == 0) 
  { /* not yet known ? */
    num = 1; 
    atyp = AT_SCALAR;
  }
  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  dout.dArrayLength = num;
  dout.dFormat = CF_DOUBLE;

  for (i=0; i<xmaLstSize; i++)
  {
    if (xmaLst[i].access & CA_WRITE)
    {
      din.dArrayLength = 1;
      din.dFormat = CF_DOUBLE;
    }
    else
    {
      din.dArrayLength = 0;
      din.dFormat = CF_NULL;
    }
    acc = (short)xmaLst[i].access;
    sprintf(desc,"[%g:%g %.16s]%.64s",xmaLst[i].min,xmaLst[i].max,xmaLst[i].units,xmaLst[i].desc);
    if ((cc=RegisterPropertyInformation(MSTEQM_TAG,xmaLst[i].name,&dout,&din,acc,atyp,0,desc,PPR_XMA_STARTBASE+i,"")) < 0)
    {
      feclog("could not register extended motor attribute %.64s : %.32s",xmaLst[i].name,erlst[cc]);
    }
  }
  xmaDispatch = fcn;
  return 0;
}
int initializeMotorHardware(void)
{
  int cc = 0;

  if (initMotorHdwFcn == NULL) ccerr(not_implemented);
  cc = initMotorHdwFcn();
err:
  if (cc) feclog("initialize motor hardware: %s",cc2str(cc));
  return cc;
}
void getMotorList(NAME64 **motorNames,int *numMotors)
{
  if (motorNames != NULL) *motorNames = mstMotorNames;
  if (numMotors != NULL) *numMotors = mstNumMotors;
  return;
}
double getStepsPerMU(int mot)
{
  int cc = 0;
  double spm=1;
  if (getMotorStepsPerMUFcn == NULL) return spm;
  cc = getMotorStepsPerMUFcn(mot,&spm);
  if (cc) 
  { 
    feclog("get stepsPerMU %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return spm;
}
double getStepsPerRadian(int mot)
{
  int cc = 0;
  double spr=1;
  if (getMotorStepsPerRadianFcn == NULL) return spr;
  cc = getMotorStepsPerRadianFcn(mot,&spr);
  if (cc) 
  { 
    feclog("get stepsPerRadian %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return spr;
}
int getMicroStepsPerStep(int mot)
{
  int cc = 0;
  int msps=1;
  if (getMotorMicroStepsPerStepFcn == NULL) return msps;
  cc = getMotorMicroStepsPerStepFcn(mot,&msps);
  if (cc) 
  { 
    feclog("get getMicroStepsPerStep %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return msps;
}

int resetMotor(int mot)
{
  int cc = 0;

  if (resetMotorFcn == NULL) ccerr(not_implemented);
  cc = resetMotorFcn(mot,&cc);
err:
  if (cc) 
  { 
    feclog("resetMotor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int calibrateMotor(int mot)
{
  int cc = 0;
  int status = -1;

  if (calibrateMotorFcn == NULL) ccerr(not_implemented);
  cc = calibrateMotorFcn(mot,&status);
err:
  if (cc) 
  { 
    feclog("calibrateMotor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int calibrateToOffset(int mot,double val)
{
  int cc = 0;
  int status = -1;

  if (calibrateToOffsetFcn == NULL) ccerr(not_implemented);
  cc = calibrateToOffsetFcn(mot,val);
err:
  if (cc) 
  { 
    feclog("calibrateToOffset %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int getMotorStatus(int mot)
{
  int cc = 0;
  int status = -1;

  if (getMotorStatusFcn == NULL) ccerr(not_implemented);
  cc = getMotorStatusFcn(mot,&status);
err:
  if (cc) 
  {
    feclog("getMotorStatus for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return status;
}
int getGroupMotorStatus(int grpId)
{
  int cc = 0;
  int status = -1;

  if (getGroupMotorStatusFcn == NULL) ccerr(not_implemented);
  cc = getGroupMotorStatusFcn(grpId,&status);
err:
  if (cc) 
  {
    feclog("getGroupMotorStatus for group id %d : %s",grpId,cc2str(cc));
  }
  return status;
}
double getMotorVelocityLowerLimit(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorVelocityLowerLimitFcn == NULL) ccerr(not_implemented);
  cc = getMotorVelocityLowerLimitFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorVelocityLowerLimit for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getMotorVelocityUpperLimit(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorVelocityUpperLimitFcn == NULL) ccerr(not_implemented);
  cc = getMotorVelocityUpperLimitFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorVelocityUpperLimit for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getMotorAccelerationUpperLimit(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorAccelerationUpperLimitFcn == NULL) ccerr(not_implemented);
  cc = getMotorAccelerationUpperLimitFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorAccelerationUpperLimit for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getMotorPosition(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorPositionFcn == NULL) ccerr(not_implemented);
  cc = getMotorPositionFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorPosition for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getGroupMotorPosition(int grpId)
{
  int cc = 0;
  double dval = 0;

  if (getGroupMotorPositionFcn == NULL) ccerr(not_implemented);
  cc = getGroupMotorPositionFcn(grpId,&dval);
err:
  if (cc) 
  {
    feclog("getGroupMotorPosition for group id %d : %s",grpId,cc2str(cc));
  }
  return dval;
}
double getMotorPositionMinimum(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorPositionMinimumFcn == NULL) ccerr(not_implemented);
  cc = getMotorPositionMinimumFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorPositionMinimum for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getMotorPositionMaximum(int mot)
{
  int cc = 0;
  double dval = 1000;

  if (getMotorPositionMaximumFcn == NULL) ccerr(not_implemented);
  cc = getMotorPositionMaximumFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorPositionMaximum for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getMotorPositionOffset(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorPositionOffsetFcn == NULL) ccerr(not_implemented);
  cc = getMotorPositionOffsetFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorPositionOffset for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getMotorPositionScale(int mot)
{
  int cc = 0;
  double dval = 1;

  if (getMotorPositionScaleFcn == NULL) ccerr(not_implemented);
  cc = getMotorPositionScaleFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorPositionScale for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
double getMotorTargetPosition(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorTargetPositionFcn == NULL) ccerr(not_implemented);
  cc = getMotorTargetPositionFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorTargetPosition for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
int getMotorTargetMicroPosition(int mot)
{
  int cc = 0;
  int ival = 0;

  if (getMotorTargetMicroPositionFcn == NULL) ccerr(not_implemented);
  cc = getMotorTargetMicroPositionFcn(mot,&ival);
err:
  if (cc) 
  {
    feclog("getMotorTargetMicroPosition for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return ival;
}
double getGroupMotorTargetPosition(int grpid)
{
  int gid,m,cc = 0;
  double dval = 0;

  if (getGroupMotorTargetPositionFcn != NULL)
  {
    cc = getGroupMotorTargetPositionFcn(grpid,&dval);
  }
  else
  {
    if ((gid=mstGetMotorGroupTableIndex(grpid)) < 0) ccerr(invalid_index);
    m = mstGroups[gid].grphead - 1;
    dval = getMotorTargetPosition(m);
  }
err:
  if (cc) 
  {
    feclog("getGroupMotorTargetPosition for group motor %d : %s",gid,cc2str(cc));
  }
  return dval;
}
int setMotorPositionOffset(int mot,double newOffset)
{
  int cc = 0;

  if (setMotorPositionOffsetFcn == NULL) ccerr(not_implemented);
  cc = setMotorPositionOffsetFcn(mot,newOffset);
err:
  if (cc) 
  {
    feclog("setMotorPositionOffsetFcn for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int setMotorPositionScale(int mot,double newOffset)
{
  int cc = 0;

  if (setMotorPositionScaleFcn == NULL) ccerr(not_implemented);
  cc = setMotorPositionScaleFcn(mot,newOffset);
err:
  if (cc) 
  {
    feclog("setMotorPositionScaleFcn for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int setMotorPosition(int mot,double newPosition)
{
  int cc = 0;

  if (setMotorPositionFcn == NULL) ccerr(not_implemented);
  cc = setMotorPositionFcn(mot,newPosition);
err:
  if (cc) 
  {
    feclog("setMotorPosition for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int setMotorTargetPosition(int mot,double newPosition)
{
  int cc = 0;

  if (setMotorTargetPositionFcn == NULL) ccerr(not_implemented);
  cc = setMotorTargetPositionFcn(mot,newPosition);
err:
  if (cc) 
  {
    feclog("setMotorTargetPosition for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int setMotorTargetSteps(int mot,int numSteps)
{
  int cc = 0;

  if (setMotorTargetStepsFcn == NULL) ccerr(not_implemented);
  cc = setMotorTargetStepsFcn(mot,numSteps);
err:
  if (cc) 
  {
    feclog("setMotorTargetSteps for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int setGroupMotorTargetPosition(int grpId,double newPosition)
{
  int cc = 0, gid, m;
  double dval = newPosition;

  if (setGroupMotorTargetPositionFcn != NULL)
  {
    cc = setGroupMotorTargetPositionFcn(grpId,newPosition);
  }
  else
  {
    if ((gid=mstGetMotorGroupTableIndex(grpId)) < 0) ccerr(invalid_index);
    m = mstGroups[gid].grphead - 1;
    if ((cc=setMotorTargetPosition(m,dval)) != 0) ccerr(cc);
    mstTbl[m].tgtPosition = dval;
  }
err:
  if (cc) 
  {
    feclog("setGroupMotorTargetPosition for group %d : %s",gid,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)gid,cc,NULL);
  }
  return cc;
}
int setGroupGearRatio(int grpId,double newRatio)
{
  int cc = 0;

  if (grpId < MST_DYNAGRP_OFFST) ccerr(0);
  if (setGroupGearRatioFcn == NULL) ccerr(not_implemented);
  cc = setGroupGearRatioFcn(grpId,newRatio);
err:
  if (cc) 
  {
    feclog("setGroupGearRatioFcn for group %d : %s",grpId,cc2str(cc));
  }
  return cc;
}
double getGroupGearRatio(int grpId)
{
  int cc = 0;
  double dval = 0;

  if (getGroupGearRatioFcn == NULL) ccerr(not_implemented);
  cc = getGroupGearRatioFcn(grpId,&dval);
err:
  if (cc) 
  {
    feclog("getGroupGearRatio for group %d : %s",grpId,cc2str(cc));
  }
  return dval;
}
int setMotorTargetMicroPosition(int mot,int newPosition)
{
  int cc = 0;

  if (setMotorTargetMicroPositionFcn == NULL) ccerr(not_implemented);
  cc = setMotorTargetMicroPositionFcn(mot,newPosition);
err:
  if (cc) 
  {
    feclog("setMotorTargetMicroPosition for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
double getMotorVelocity(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorVelocityFcn == NULL) ccerr(not_implemented);
  cc = getMotorVelocityFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorVelocity for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
int setMotorVelocity(int mot,double newVelocity)
{
  int cc = 0;

  if (setMotorVelocityFcn == NULL) ccerr(not_implemented);
  cc = setMotorVelocityFcn(mot,newVelocity);
err:
  if (cc) 
  {
    feclog("setMotorVelocity for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
double getMotorAcceleration(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorAccelerationFcn == NULL) ccerr(not_implemented);
  cc = getMotorAccelerationFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorAcceleration for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
int setMotorAcceleration(int mot,double newAcceleration)
{
  int cc = 0;

  if (setMotorAccelerationFcn == NULL) ccerr(not_implemented);
  setMotorAccelerationFcn(mot,newAcceleration);
err:
  if (cc) 
  {
    feclog("setMotorAcceleration for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
double getMotorCurrent(int mot)
{
  int cc = 0;
  double dval = 0;

  if (getMotorCurrentFcn == NULL) ccerr(not_implemented);
  cc = getMotorCurrentFcn(mot,&dval);
err:
  if (cc) 
  {
    feclog("getMotorCurrent for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return dval;
}
int setMotorCurrent(int mot,double newCurrent)
{
  int cc = 0;

  if (setMotorCurrentFcn == NULL) ccerr(not_implemented);
  cc = setMotorCurrentFcn(mot,newCurrent);
err:
  if (cc) 
  {
    feclog("setMotorCurrent for motor %d : %s",mot,cc2str(cc));
    if (gMstSetAlarms) SetAlarm(MSTEQM_TAG,(short)mot,cc,NULL);
  }
  return cc;
}
int getMotorUnits(int mot,NAME16 *units)
{
  int cc = 0;
  if (units == NULL) return argument_list_error;
  if (getMotorUnitsFcn != NULL)
  {
    memset(units->name,0,16);
    /* leave last 6 chars for possible concatenation with "/sec", etc. */
    cc = getMotorUnitsFcn(mot,units->name,10);
  }
  else
  {
    strcpy(units->name,"mm");
  }
  return cc;
}
int getLastHardwareError(int mot,NAME64 *errstr,int *errcode)
{
  int cc = 0;
  if (errstr == NULL && errcode == NULL) return argument_list_error;
  if (getLastHdwErrFcn != NULL)
  {
    cc = getLastHdwErrFcn(mot,errstr,errcode);
  }
  else
  {
    if (errstr != NULL) strcpy(errstr->name,"not available");
    if (errcode != NULL) *errcode = 0;
  }
  return cc;
}

MonitorParams *g_monitorParamsList = NULL;

MonitorParams *getMonitorParams(const char *dev,const char *prp,int fmt,int len,int tmr,int dly,int inc)
{
  MonitorParams *mp = NULL;

  if (dev == NULL || prp == NULL) return NULL;
  if (tmr < 100) tmr = 100;
  if (!isPrimitiveFormat(fmt)) return NULL;
  if (len < 1) len = 1;
  for (mp=g_monitorParamsList; mp != NULL; mp = mp->nxt)
  {
    if (strnicmp(dev,mp->dev,132)) continue;
    if (strnicmp(prp,mp->prp,64)) continue;
    if (fmt != mp->fmt) continue;
    if (len != mp->len) continue;
    if (tmr != mp->tmr) continue;
    return mp;
  }
  if ((mp=(MonitorParams *)calloc(1,sizeof(MonitorParams))) != NULL)
  {
    strncpy(mp->dev,dev,132);
    strncpy(mp->prp,prp,64);
    mp->fmt = fmt;
    mp->len = len;
    mp->tmr = tmr;
    mp->nxt = g_monitorParamsList;
    g_monitorParamsList = mp;
  }
  return mp;
}

int mstGetFreeMotorGroup(void)
{
  int i;
  for (i=0; i<mstNumGroups; i++)
  {
    if (mstGroups[i].nmotors == 0) break;
  }
  if (i == NUM_MOTORGROUPS_MAX) return -resources_exhausted;
  mstNumGroups++;
  return i;
}
/**
 * Gives the automatic group table index (not configured)
 *
 * \note largely used in mstlib.c
 */
int mstGetMotorGroupTableIndex(int grpId)
{
  int i;
  for (i=0; i<mstNumGroups; i++)
  {
    if (mstGroups[i].grpid == grpId) return i;
  }
  return -1;
}
/**
 * gives the configured group table index (as learned from motorgroups.csv)
 *
 * \note largely used in msteqm.c
 */
int mstGetMotorGroupConfigurationIndex(int grpId)
{
  int i;
  for (i=0; i<mstGroupTblSize; i++)
  {
    if (mstGroupTbl[i].groupId == grpId) return i;
  }
  return -1;
}
int mstFreeMotorGroup(int groupIndex)
{
  if (groupIndex < 0 || groupIndex >= mstNumGroups) return invalid_index;
  mstGroups[groupIndex].grpid = 0;
  mstGroups[groupIndex].nmotors = 0;
  memset(mstGroups[groupIndex].fd_motors,0,NUM_MOTOR_FDSETS);
  return 0;
}
char *mstGetMotorGroupList(int groupIndex)
{ /* this is effectively thread-safe as it will only be used to set up a group call (from the eqm) */
  static char mstGroupStr[NUM_MOTORS_MAX*(DEVICE_NAME_SIZE+2)];
  int i, n=0, m, b;
  mstGroupStr[0] = 0;
  if (mstMotorNames == NULL) return mstGroupStr;
  if (groupIndex < 0 || groupIndex >= mstNumGroups) return mstGroupStr;
  for (i=0; i<NUM_MOTOR_FDSETS; i++)
  {
    for (b=0; b<8; b++)
    {
      if (mstGroups[groupIndex].fd_motors[i] & ((0x01)<<b))
      {
        m = i*8 + b;
        strncat(mstGroupStr,mstMotorNames[m].name,DEVICE_NAME_SIZE);
        strcat(mstGroupStr,",");
        n++;
      }
      if (n == mstGroups[groupIndex].nmotors) 
      { /* no need to continue */
        return mstGroupStr;
      }
    }
  }
  if (n != mstGroups[groupIndex].nmotors)
  { /* we didn't find as many as expected ? */
    feclog("motor group %d registration mis-match (%d vs %d motors)",groupIndex,n,mstGroups[groupIndex].nmotors);
  }
  return mstGroupStr;
}
int mstSplitGroupFromList(const char *motorList,int *num,int *motors,char *tag)
{
  int i, len, m, n=0, cc=0;
  char *g=NULL,*c=NULL,*cp=NULL;
  char tmpString[1048];

  if (motorList == NULL || num == NULL || motors == NULL) return argument_list_error;
  if (*num < 1) return argument_list_error;
  strncpy(tmpString,motorList,1024);
  for (i=0, g=strtok_r(tmpString,MST_SEPERATOR_TOKEN,&cp); g != NULL; i++, g=strtok_r(NULL,MST_SEPERATOR_TOKEN,&cp))
  { /* split group into the individual devices ... */
    while (*g == ' ') g++; /* leading blanks */
    for (len=(int)strlen(g); len > 0 && g[len-1] == ' '; g[len-1] = 0); /* trailing blanks */
    if ((m=GetDeviceNumber(MSTEQM_TAG,g)) < 0) 
    {
	    return illegal_device_number;
    }
    if (m >= mstNumMotors) return illegal_device_number;
    motors[n] = m;
    n++;
    if (n >= mstNumMotors || n >= *num) break;
  }
  *num = n;
  len = (int)strlen(tag);
  if (len == 0 || len > 64) 
  { /* invent a tag for an on-the-fly motor group */
    if (n > 0) sprintf(tag,"%d",motors[0]);
    for (i=1; i<n; i++)
    {
      sprintf(tmpString,"-%d",motors[i]);
      strcat(tag,tmpString);
      if (strlen(tag) > 120) break;
    }
  }
  return 0;
}
int mstGetMotorGroupFromTag(const char *tag)
{
  int i;
  for (i=0; i<mstNumGroups; i++)
  {
    if (mstGroups[i].nmotors == 0) continue;
    if (strnicmp(mstGroups[i].grpname,tag,128)) continue;
    return i;
  }
  return -1;
}
int mstGetMotorGroupFromList(const char *motorList,const char *groupName)
{
  int num=NUM_MOTORS_MAX, cc=0;
  int motors[NUM_MOTORS_MAX];
  char tag[128];
  memset(tag,0,128);
  if (groupName != NULL) strncpy(tag,groupName,64);
  if ((cc=mstSplitGroupFromList(motorList,&num,motors,tag)) != 0) return -cc;
  return mstGetMotorGroup(tag,num,motors);
}
int mstMakeGroupId(void)
{
  int i=0, id=MST_DYNAGRP_OFFST;
  while (i<mstNumGroups)
  {
    if (mstGroups[i++].grpid == id) 
    {
      i = 0; id++;
    }
  }
  return id;
}
int mstGetMotorGroup(const char *tag,int nmotors,int *motors)
{
  int i;
  BYTE fd_motors[NUM_MOTOR_FDSETS];
  memset(fd_motors,0,NUM_MOTOR_FDSETS);
  for (i=0; i<nmotors; i++)
  {
    fd_motors[motors[i]/8] |= (0x01)<<(motors[i]%8);
  }
  for (i=0; i<mstNumGroups; i++)
  {
    if (mstGroups[i].nmotors == 0) continue;
    if (memcmp(mstGroups[i].fd_motors,fd_motors,NUM_MOTOR_FDSETS)) continue;
    if (strnicmp(mstGroups[i].grpname,tag,128)) continue;
    return mstGroups[i].grpid;
  }
  if ((i=mstGetFreeMotorGroup()) < 0) return i;
  mstGroups[i].nmotors = nmotors;
  memcpy(mstGroups[i].fd_motors,fd_motors,NUM_MOTOR_FDSETS);
  strncpy(mstGroups[i].grpname,tag,128);
  mstGroups[i].grpid = mstMakeGroupId();
  if (setGroupMembersFcn != NULL) setGroupMembersFcn(mstGroups[i].grpid,&mstGroups[i]);
  return mstGroups[i].grpid;
}
int mstAddPseudoAxisMotorGroup(const char *tag,int grpid)
{
  int i;
  for (i=0; i<mstNumGroups; i++)
  {
    if (strnicmp(mstGroups[i].grpname,tag,128)) continue;
    if (mstGroups[i].grpid != grpid) continue;
    return grpid;
  }
  if ((i=mstGetFreeMotorGroup()) < 0) return i;
  strncpy(mstGroups[i].grpname,tag,128);
  mstGroups[i].grpid = grpid;
  if (setGroupMembersFcn != NULL) setGroupMembersFcn(grpid,&mstGroups[i]);
  return grpid;

}
//n.b. motorNr counts from 1, 'mot' = 'device number' counts from 0
int mstSetGroupHead(int grpId,int motorNr)
{
  int i;
  for (i=0; i<mstNumGroups; i++)
  {
    if (mstGroups[i].grpid == grpId) break;
  }
  if (i == mstNumGroups) return invalid_index;
  mstGroups[i].grphead = motorNr;
  return 0;
}
int getGroupMemberTableIndex(int grpId,int mot)
{
  int i;
  if (mot >= NUM_MSTEQM_DEVICES) return -1;
  for (i=0; i<mstNumGroups; i++)
  {
    if (mstGroups[i].grpid == grpId) break;
  }
  if (i == mstNumGroups) return -1;
  if (!IsMotorInGroup(mot,&mstGroups[i])) return -1;
  if ((i=mstGetMotorGroupConfigurationIndex(grpId)) < 0) return -1;
  return i;
}
float getMemberGearRatio(int grpId,int mot)
{
  int i = getGroupMemberTableIndex(grpId,mot);
  if (i < 0) return 0;
  return mstGroupTbl[i].members[mot]->gearRatio;
}
MST_EXPORT float GetGroupMotorGearRatio(int grpId,int mot)
{
  return getMemberGearRatio(grpId,mot);
}
MST_EXPORT float GetGroupMotorScale(int grpId,int mot)
{
  int i = getGroupMemberTableIndex(grpId,mot);
  if (i < 0) return 0;
  return mstGroupTbl[i].members[mot]->scale;
}
MST_EXPORT float GetGroupMotorWeight(int grpId,int mot)
{
  int i = getGroupMemberTableIndex(grpId,mot);
  if (i < 0) return 0;
  return mstGroupTbl[i].members[mot]->weight;
}
MST_EXPORT float GetGroupMotorOffset(int grpId,int mot)
{
  int i = getGroupMemberTableIndex(grpId,mot);
  if (i < 0) return 0;
  return mstGroupTbl[i].members[mot]->offset;
}

GETDSMPFCNP getSmarPodPositionFcn = NULL;
GETDSMPFCNP getSmarPodTargetPositionFcn = NULL;
GETDSMPFCNP getSmarPodPositionMinimumFcn = NULL;
GETDSMPFCNP getSmarPodPositionMaximumFcn = NULL;
GETDSMPFCNP getSmarPodPositionOffsetFcn = NULL;
GETDSMPFCNP getSmarPodPositionScaleFcn = NULL;

SETDSMPFCNP setSmarPodPositionFcn = NULL;
SETDSMPFCNP setSmarPodTargetPositionFcn = NULL;
SETDSMPFCNP setSmarPodPositionOffsetFcn = NULL;
SETDSMPFCNP setSmarPodPositionScaleFcn = NULL;

MST_EXPORT void SetGetSmarPodPositionDispatch(GETDSMPFCNP fcn)
{
  getSmarPodPositionFcn = fcn;
}
MST_EXPORT void SetGetSmarPodTargetPositionDispatch(GETDSMPFCNP fcn)
{
  getSmarPodTargetPositionFcn = fcn;
}
MST_EXPORT void SetGetSmarPodPositionMinimumDispatch(GETDSMPFCNP fcn)
{
  getSmarPodPositionMinimumFcn = fcn;
}
MST_EXPORT void SetGetSmarPodPositionMaximumDispatch(GETDSMPFCNP fcn)
{
  getSmarPodPositionMaximumFcn = fcn;
}
MST_EXPORT void SetGetSmarPodPositionOffsetDispatch(GETDSMPFCNP fcn)
{
  getSmarPodPositionOffsetFcn = fcn;
}
MST_EXPORT void SetGetSmarPodPositionScaleDispatch(GETDSMPFCNP fcn)
{
  getSmarPodPositionScaleFcn = fcn;
}
MST_EXPORT void SetSetSmarPodTargetPositionDispatch(GETDSMPFCNP fcn)
{
  setSmarPodTargetPositionFcn = fcn;
}
MST_EXPORT void SetSetSmarPodPositionDispatch(SETDSMPFCNP fcn)
{
  setSmarPodPositionFcn = fcn;
}
MST_EXPORT void SetSetSmarPodPositionOffsetDispatch(SETDSMPFCNP fcn)
{
  setSmarPodPositionOffsetFcn = fcn;
}
MST_EXPORT void SetSetSmarPodPositionScaleDispatch(SETDSMPFCNP fcn)
{
  setSmarPodPositionScaleFcn = fcn;
}
