/*
   This following code demonstrates the CONSOLE - ARCHIVE - Server
   communication API.  This is only an example.  Feel free to modify
   anything you see.  This is also a c-wrapper for the CPP API interface.
   Thus you can convert this to legitimate CPP code if you are so
   inclined.

   This project is a client-side project only ! No server functionality
*/
#include "cprolog.h"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include "errors.h"
#include "rpctypes.h"
#include "servdef.h"
#include "arceqm.h"
#include "arctypes.h"

UINT32 targetevn = 0;
// to satisfy server-like calls from arceqm.cpp
char *getClientName(void) { return ""; }
char *getClientTagName(void) {return ""; }
int RegisterEquipmentModule(
    char *tagname,
    char *eqpname,
    int NumModules,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short),
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
){ return 0;}

void initArchiveClient(void);
void doinit(void)
{
  int cc;
  initArchiveClient();
}

// use CommandLevel to dynamically issue query commands:

extern int CommandLevel;
char *commandText[] = {
  "No command","SetGlobalEvent","GetNumberOfEvents",
  "GetEvents","GetNumberOfArchiveServers","GetArchiveServers",
  "GetNumberOfDeviceServers","GetDeviceServers", "GetRemoteArchiveData",
  "GetArchiveInfo","GetPropertyInfo","GetNumArchiveServers",
  "GetNumArchiveDeviceNames","GetArchivedEvents", "GetRemoteArchiveHeader"
 };

int numCommandLevels = sizeof(commandText)/sizeof(commandText[0]);

#define ASRVBUFFERSIZE 10
#define DSRVBUFFERSIZE 50
#define AESBUFFERSIZE 50
#define rpcerror(i) {\
  printf("%s : %s\n>",commandText[CommandLevel],erlst[i]);\
}
#define commandFinished {\
  CommandLevel = 0;  break;\
}

#define OldcommandFinished {\
  for (i=0; i < numCommandLevels; i++)\
    printf("Set CommandLevel = %d: %s\n",i,commandText[i]);\
  CommandLevel = 0;  break;\
}

// just an example routine: (see CommandLevel 8)
#define SHORTS
int queryRemoteData(void)
{
  float fdata[100];
  short sdata[100];
  DEVSPEC devspec;
  int i,cc;
  strcpy(devspec.deviceServerName,"BLM");
  strcpy(devspec.deviceProperty,"SLOSS");
  strcpy(devspec.deviceName,"#1");
  devspec.dataDimension = 128;
#ifdef SHORTS
  devspec.dataFormat = CF_SHORT;
  if ((cc=GetRemoteArchiveData("MARCHSRV",(BYTE *)sdata,targetevn,0,0,&devspec)) == 0)
  {
    printf("Received data :\n");
    for (i=0; i<100; i++) printf((i+1)%8 ? "%d " : "%d\n",sdata[i]);
    printf("\n");
  }
#else
  devspec.dataFormat = CF_FLOAT;
  if ((cc=GetRemoteArchiveData("MARCHSRV",(BYTE *)fdata,0,0,0,&devspec)) == 0)
  {
    printf("Received data :\n");
    for (i=0; i<100; i++) printf((i+1)%8 ? "%g " : "%g\n",fdata[i]);
    printf("\n");
  }
#endif  
  return cc;
}
int queryRemoteHeader(void)
{
  DHS dhs;
  DEVSPEC devspec;
  int i,cc;
  strcpy(devspec.deviceServerName,"BLM");
  strcpy(devspec.deviceProperty,"SLOSS");
  strcpy(devspec.deviceName,"#1");
  devspec.dataDimension = 128;
  devspec.dataFormat = CF_FLOAT;
  if ((cc=GetRemoteArchiveHeader("MARCHSRV",&dhs,0,0,0,&devspec)) == 0)
  {
    printf("Received header :\n");
    printf("channelID : %s\n",dhs.channelID);
    printf("arraysize : %d\n",dhs.arraysize);
    printf("native format : %d\n",dhs.datatype);
    printf("\n");
  }
  return cc;
}

/*
make some commands to exercise these function calls:

int GetArchiveInfo(char *archiveServer,char *deviceServer,ulong evnnum,AIB *aib);
int GetPropertyInfo(char *archiveServer,char *deviceServer,char *deviceProperty,ulong evnnum,PIB *pib);
int GetNumArchiveServerEvents(char *archiveServer,ulong starttime,ulong stoptime,ulong eventNumber);
int GetNumArchiveDeviceServers(char *archiveServer);
int GetNumArchiveDeviceNames(char *archiveServer,NAME16 *names,int num);
int GetArchivedEvents(char *arcsrvTag,time_t starttime,time_t stoptime,ulong eventCode,AES *aes,int *num);
*/
void queryCommandLevel(void)
{
  time_t interim=0,t=time(NULL);
  int i,cc,num;
  AIB aib;
  PIB pib;
  AES aes[AESBUFFERSIZE];
  NAME16 srv[ASRVBUFFERSIZE],dev[DSRVBUFFERSIZE];

    switch (CommandLevel)
    {
      case 0: break;
      case 1: 
              commandFinished;
      case 2: if ((cc=GetNumberOfEvents(0,time(NULL))) >= 0)
                printf("There are %d recorded events\n>",cc);
              else  rpcerror(-cc);
              commandFinished;
      case 3: num = AESBUFFERSIZE;
              t = time(NULL);
              if ((cc=GetEvents(t-3600,t,0,aes,&num)) == 0)
              {
                if (num > AESBUFFERSIZE)
                {
                  printf("More events found in the specified time range ");
                  printf("than can be buffered.\nIncrease buffer size or ");
                  printf("partition the time range !\n");
                }
                printf("\n Found %d events\n",num);
                for (i=0; i<AESBUFFERSIZE && i<num; i++)
                  printf(" %d %d %s\n",
                         aes[i].eventNumber,
                         aes[i].eventCode,
                         aes[i].comment[0]);
                targetevn = aes[0].eventNumber;
              }
              else rpcerror(cc);
              commandFinished;
      case 4: if ((cc=GetNumberOfArchiveServers()) >= 0)
                printf("There are %d Archive Servers\n>",cc);
              else rpcerror(-cc);
              commandFinished;
      case 5: num = ASRVBUFFERSIZE;
              if ((cc=GetArchiveServers(-1,srv,&num)) == 0)
              {
                printf("archive servers:\n");
                for (i=0; i<ASRVBUFFERSIZE; i++)
                  if (strlen(srv[i].name))printf("%s\n",srv[i].name);
              }
              else rpcerror(cc);
              commandFinished;
      case 6: if ((cc=GetNumberOfDeviceServers("MARCHSRV")) >= 0)
                printf("There are %d Device Servers\n>",cc);
              else rpcerror(-cc);
              commandFinished;
      case 7: num = DSRVBUFFERSIZE;
              if ((cc=GetDeviceServers("MARCHSRV",-1,dev,&num)) == 0)
              {
                printf("device servers:\n");
                for (i=0; i<DSRVBUFFERSIZE; i++)
                  if (strlen(dev[i].name)) printf("%s\n",dev[i].name);
              }
              else rpcerror(cc);
              commandFinished;
      case 8: 
              num = AESBUFFERSIZE;
              t = time(NULL);
              if ((cc=GetEvents(t-3600,t,0,aes,&num)) == 0)
              {
                if (num > AESBUFFERSIZE)
                {
                  printf("More events found in the specified time range ");
                  printf("than can be buffered.\nIncrease buffer size or ");
                  printf("partition the time range !\n");
                }
                printf("\n Found %d events\n",num);
                for (i=0; i<AESBUFFERSIZE && i<num; i++)
                  printf(" %d %d %s\n",
                         aes[i].eventNumber,
                         aes[i].eventCode,
                         aes[i].comment[0]);
                targetevn = aes[0].eventNumber;
              }
              else rpcerror(cc);
              if ((cc=queryRemoteData()) != 0)
              {
                rpcerror(cc);
              }
              commandFinished;
      case 9: GetArchiveInfo("MARCHSRV","BLM",0,&aib);
              printf("num properties: %d\n",aib.numProperties);
              printf("num devices:    %d\n",aib.numDevices);
              printf("event code:     %d\n",aib.eventCode);
              printf("comment: %s\n",aib.comment);
              commandFinished;
      case 10:cc = GetPropertyInfo("MARCHSRV","BLM","SLOSS",0,&pib);
              if (cc) printf("error: <%d> %s\n",cc,erlst[cc]);
              commandFinished;
      case 11:GetNumArchiveServerEvents("MARCHSRV",0,time(NULL),0);
              commandFinished;
      case 12:i = GetNumArchiveDeviceServers("MARCHSRV");
              printf("Received info: %d device servers\n",i);
              commandFinished;
      case 13:num = DSRVBUFFERSIZE;
              if ((cc=GetArchiveDeviceServerNames("MARCHSRV",1234,time(NULL),dev,&num)) >= 0)
              {
                printf("archive servers:\n");
                for (i=0; i<DSRVBUFFERSIZE; i++)
                  if (strlen(dev[i].name))printf("%s\n",dev[i].name);
              }
              else rpcerror(-cc);
              commandFinished;
      case 14:num = AESBUFFERSIZE;
              GetArchivedEvents("MARCHSRV",0,time(NULL),0,aes,&num);
              commandFinished; 
      case 15:i = GetNumArchiveDevices("MARCHSRV","BLM");
              printf("Received info: %d device servers\n",i);
              commandFinished;
      case 16:num = DSRVBUFFERSIZE;
              if ((cc=GetArchiveDeviceNames("MARCHSRV","BLM",dev,&num)) >= 0)
              {
                printf("archive servers:\n");
                for (i=0; i<DSRVBUFFERSIZE; i++)
                  if (strlen(dev[i].name))printf("%s\n",dev[i].name);
              }
              else rpcerror(-cc);
              commandFinished;
      case 17:num = AESBUFFERSIZE;
              if ((cc=GetDeviceArchivedEvents("MARCHSRV","BLM","ALL",t-3600*24,t,aes,&num)) >= 0)
              {
                printf("pktr events :\n");
                for (i=0; i<AESBUFFERSIZE; i++)
                  printf("%d\n",aes[i].eventNumber);
              }
              else rpcerror(-cc);
              commandFinished; 
      case 18:
              if ((cc=queryRemoteHeader()) != 0)
              {
                rpcerror(cc);
              }
              commandFinished;
      default:
              commandFinished;
    }
}
int main(int argc, char *argv[])
{
  int cc;

  switch (argc)                           /* initial debug level */
  {                  /* on UNIX stations: set debug = 1 (2,or 3) */
    case 2:              /* will dynamically set the debug level */
      CommandLevel = atoi(argv[1]);
//      NGdebug = atoi(argv[1]);
      break;
    default:
      CommandLevel = 0;                              /* default level */
//       NGdebug = 0;                              /* default level */
  }

  if ((cc=SystemInit(FALSE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }

  doinit();

//  cc = GetNumberOfEvents(0,time(NULL));
//  if (cc > 0) printf("There are %d recorded events\n>",cc);
//  for(;;)
//  {
//    SystemCycle(FALSE);
    if (CommandLevel) queryCommandLevel();
//  }
  return 0;
}
