#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#if defined(UNIX)
#include "dirent.h"
#elif defined(WIN32)
#include "direct.h"
#endif
#include "arceqm.h"
#include "arctools.h"

void registerStructs(void);
int isAdministrator(char *nam);
extern int useConnectedLinks;
extern char thisExtension[4];
extern int gAppendProperyAliasNamesToQueries;

#define EVENTMGRTAG "/SITE/EVENTMGR"
#define EVENTMGRFEC "EVENTS"
#define DEF_DEV_NAME "#0"

CSVCOLUMN csvTrgChannelColumns[] =
{
  { "CONTEXT","",0,CF_TEXT,32,(BYTE *)OFFSETIN(TrgChannel,ctx),NULL,0 },
  { "SERVER","",0,CF_TEXT,32,(BYTE *)OFFSETIN(TrgChannel,srv),NULL,0 },
  { "DEVICE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgChannel,dev),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgChannel,prp),NULL,0 },
};
CSVDB csvTrgChannelDb =
{
  csvTrgChannelColumns,
  sizeof(csvTrgChannelColumns)/sizeof(csvTrgChannelColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TrgChannel),
  0
};

CSVCOLUMN csvTrgComponentColumns[] =
{
  { "FILE_NAME","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgComponent,fileName),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgComponent,desc),NULL,0 },
};
CSVDB csvTrgComponentDb =
{
  csvTrgComponentColumns,
  sizeof(csvTrgComponentColumns)/sizeof(csvTrgComponentColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TrgComponent),
  0
};

TrgViewFile *trgViewFiles = NULL;
int gNumTrgViewFiles = 0;

CSVCOLUMN csvTrgViewFileColumns[] =
{
  { "CONTEXT","",0,CF_TEXT,32,(BYTE *)OFFSETIN(TrgViewFile,ctx),NULL,0 },
  { "TRIGGER","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgViewFile,trg),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgViewFile,desc),NULL,0 },
  { "STRUCTURE_FILE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgViewFile,file),NULL,0 },
};
CSVDB csvTrgViewFileDb =
{
  csvTrgViewFileColumns,
  sizeof(csvTrgViewFileColumns)/sizeof(csvTrgViewFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TrgViewFile),
  0
};

char cnfgPath[128] = CONFIGS_PATH;
char *getTrcCfgPath(void)
{
  static char path[132] = "";

  if (strlen(path) == 0)
  {
#   if defined(UNIX)
    DIR *dp = NULL;
    if ((dp=opendir(cnfgPath)) == NULL) 
    {
      char tmp[128];
      sprintf(tmp,"%.96s/CONFIGS",ArchiveDirectoryPath);
      if ((dp=opendir(cnfgPath)) != NULL)
      { // legacy directory structure
        strncpy(cnfgPath,tmp,128);
      }
    }
    if (dp != NULL) closedir(dp);
#   endif
    strncpy(path,cnfgPath,128); strcat(path,"/TRC/");
    feclog("configs path: %.132s",path);
  }
  return path;
}
int readTriggerComponentFiles(TrgViewFile *tvf)
{
  int i, nchn, cc=0;
  char fn[64], *path;
  TrgComponent *trg=NULL;
  TrgChannel *chn=NULL;

  path = getTrcCfgPath();
  sprintf(fn,"%.48s",tvf->file);
  if (strstr(fn,".csv") == NULL) strcat(fn,".csv");
  if ((cc=csvReadFile(path,fn,&csvTrgComponentDb,(void **)&trg)) != 0) return cc;
  tvf->components = trg;
  tvf->nComponents = csvTrgComponentDb.siz;
  for (i=0; i<tvf->nComponents; i++)
  { /* go thru the list for the view channels */
    chn = NULL; nchn = 0;
    sprintf(fn,"%.48s",tvf->components[i].fileName);
    if (strstr(fn,".csv") == NULL) strcat(fn,".csv");
    nchn =  (csvReadFile(path,fn,&csvTrgChannelDb,(void **)&chn) == 0) ? csvTrgChannelDb.siz : 0;
    tvf->components[i].channels = chn;
    tvf->components[i].nChannels = nchn;
  }
  return 0;
}

int readViewerFiles(void)
{
  int i, cc=0;
  char *path = getTrcCfgPath();
  /* first read the triggers configuration file */
  if (gNumTrgViewFiles > 0) 
  {
    gNumTrgViewFiles = 0;
    if (trgViewFiles != NULL) free(trgViewFiles);
    trgViewFiles = NULL;
  }
  feclog("looking for TRC viewer configurations in %s",path);
  if ((cc=csvReadFile(path,TRCVIEW_FILE,&csvTrgViewFileDb,(void **)&trgViewFiles)) != 0) ccerr(cc);
  gNumTrgViewFiles = csvTrgViewFileDb.siz;
  feclog("read %d TRC viewer configurations",gNumTrgViewFiles);
  for (i=0; i<gNumTrgViewFiles; i++)
  { /* go thru the list for the view components (ignore file errors here */
    readTriggerComponentFiles(&trgViewFiles[i]);
  }
err:
  return cc;
}
int isViewFileInList(int idx,TrgViewFile *trg)
{
  int i;
  for (i=0; i<idx; i++)
  {
    if (!strncmp(trg[i].file,trg[idx].file,64)) return TRUE;
    if (!strncmp(trg[i].desc,trg[idx].desc,64)) return TRUE;
  }
  return FALSE;
}
int writeTriggerComponentFile(char *config,TrgViewFile *trg,int size)
{
  int i,n,cc = 0;
  char fn[256],tfn[72];
  FILE *fp=NULL;
  TrgComponent *trgcmp=NULL;
  char *path = getTrcCfgPath();

  for (i=0; i<gNumTrgViewFiles; i++) if (!strnicmp(trgViewFiles[i].desc,config,64)) break;
  if (i == gNumTrgViewFiles) return illegal_device;

  strncpy(tfn,trgViewFiles[i].file,64); tfn[64] = 0;
  if (strstr(tfn,".csv") == NULL) strcat(tfn,".csv");
  sprintf(fn,"%.132s%.64s",path,tfn);
  if ((fp=fopen(fn,"wt")) == NULL &&
      (fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"FileName, Description\n");
  for (n=0; n<size; n++)
  {
    if (isViewFileInList(n,trg)) continue;
    fprintf(fp,"%.64s, %.64s\n",trg[n].file,trg[n].desc);
  }
  fclose(fp); fp = NULL;
  if ((cc=csvReadFile(path,tfn,&csvTrgComponentDb,(void **)&trgcmp)) != 0) ccerr(cc);
  trgViewFiles[i].components = trgcmp;
  trgViewFiles[i].nComponents = csvTrgComponentDb.siz;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int writeTriggerChannelFile(char *desc,TrgChannel *trg,int size)
{
  int i,k,n,nchn=0,cc = 0,found = FALSE;
  char fn[256], cfn[72];
  FILE *fp=NULL;
  TrgChannel *chn=NULL;
  char *path = getTrcCfgPath();

  for (i=0; i<gNumTrgViewFiles; i++)
  {
    for (k=0; k<trgViewFiles[i].nComponents; k++) if (!strnicmp(trgViewFiles[i].components[k].desc,desc,64))
    { /* found it ! */
      strncpy(cfn,trgViewFiles[i].components[k].fileName,64); cfn[64] = 0;
      found = TRUE;
      break;
    }
    if (found) break;
  }
  if (!found) 
  { /* new channel configuration ? */
    return illegal_device;
  }
  if (strstr(cfn,".csv") == NULL) strcat(cfn,".csv");
  sprintf(fn,"%.128s%.64s",path,cfn);
  if ((fp=fopen(fn,"wt")) == NULL &&
      (fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"Context, Server, Property, Device\n");
  for (n=0; n<size; n++)
  {
    fprintf(fp,"%.32s, %.32s, %.64s, %.64s\n",
      trg[n].ctx,trg[n].srv,trg[n].prp,trg[n].dev);
  }
  fclose(fp); fp = NULL;
  /* now read back what we just wrote ... */
  if (strstr(cfn,".csv") == NULL) strcat(cfn,".csv");
  if ((cc=csvReadFile(path,cfn,&csvTrgChannelDb,(void **)&chn)) == 0) nchn = csvTrgChannelDb.siz;
  trgViewFiles[i].components[k].channels = chn;
  trgViewFiles[i].components[k].nChannels = nchn;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int writeViewerFiles(TrgViewFile *viewFiles,int nViewFiles)
{
  int cc=0, i;
  char fn[256];
  FILE *fp=NULL;
  char *path = getTrcCfgPath();

  sprintf(fn,"%.132s%s",path,TRCVIEW_FILE);
  
  if ((fp=fopen(fn,"wt")) == NULL &&
      (fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);

  fprintf(fp,"Context, Trigger, Description, StructureFile\n");
  for (i=0; i<nViewFiles; i++)
  {
    fprintf(fp,"%.32s, %.64s, %.64s, %.64s\n",
      viewFiles[i].ctx,viewFiles[i].trg,viewFiles[i].desc,viewFiles[i].file);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int GetTriggerComponents(char *config,TrgChannel *cmp,int cmpBufSize)
{
  int i,n;

  for (i=0; i<gNumTrgViewFiles; i++) if (!strnicmp(trgViewFiles[i].desc,config,64)) break;
  if (i == gNumTrgViewFiles) return -illegal_device;
  for (n=0; n<trgViewFiles[i].nComponents && n<cmpBufSize; n++)
  {
    strncpy(cmp[n].ctx,trgViewFiles[i].ctx,32);
    strncpy(cmp[n].srv,"EVENTS",32);
    strncpy(cmp[n].dev,trgViewFiles[i].components[n].desc,64);
    strncpy(cmp[n].prp,trgViewFiles[i].trg,64);
  }
  return n;
}

int GetTriggerChannels(char *desc,TrgChannel *chn,int chnBufSize)
{
  int i,k,n;

  for (i=0; i<gNumTrgViewFiles; i++) for (k=0; k<trgViewFiles[i].nComponents; k++) 
    if (!strnicmp(trgViewFiles[i].components[k].desc,desc,64))
    { /* found it ! */
      for (n=0; n<chnBufSize && n<trgViewFiles[i].components[k].nChannels; n++)
      {
        strncpy(chn[n].ctx,trgViewFiles[i].components[k].channels[n].ctx,32);
        strncpy(chn[n].srv,trgViewFiles[i].components[k].channels[n].srv,32);
        strncpy(chn[n].dev,trgViewFiles[i].components[k].channels[n].dev,64);
        strncpy(chn[n].prp,trgViewFiles[i].components[k].channels[n].prp,64);
      }
      return n;
    }
  return -illegal_device;
}

AESnotify  aesNotify;
AEScontext aesContext;
ADRB *g_adrb; /* modern archive data request block */
ADRB5 *g_adrb5; /* ancient archive data request block */
ADRBX *g_adrbx; /* old archive data request block */
UINT32 g_eventNumber;

void arcini(void)
{
  static int initialized = 0;

  readViewerFiles();
  if (initialized) return;
  registerStructs();
  initialized = 1;
  return;
}
int GotTriggerEvent = 0;

/* asynchronous callback related routines: */

/* many async requests can be outstanding at the same time
   so we keep a callback table
 */
struct
{
  int id;
  BYTE *buffer;
} callbackTable[100];

int getCallbackSlot(void)
{
  int i;
  for (i=0; i<100; i++) if (callbackTable[i].id == 0) return i;
  return -1;
}
int findCallbackSlot(int id)
{
  int i;
  for (i=0; i<100; i++) if (callbackTable[i].id == id) return i;
  return -1;
}
void RcdEventCallback(int id,int cc)
{
  int i = findCallbackSlot(id);
  if (i >= 0)
  {
    callbackTable[i].id = 0; 
  }
  checkRecordEvent(cc);
}
void SetGlobalEventCallback(int id,int cc)
{
  int i = findCallbackSlot(id);
  if (i >= 0)
  {
    callbackTable[i].id = 0; 
  }
  checkSetEvent(g_eventNumber,cc);
}

/* for uniqueness, we need to specify a local device id and
   use it as the equipment number.  Although it is not used
   at the server end, it guarantees a unique contract request
   (necessary for asynchronous notification)
*/

#define MAXDEVICES 64 /* this should be plenty */
int ndevices = 0;
char devNameTags[MAXDEVICES][16];

int getDeviceIdFromName(char *devName)
{
  int i;
  for (i=0; i<ndevices && strnicmp(devName,devNameTags[i],16); i++);
  if (i == ndevices)
  {
    if (ndevices == MAXDEVICES) 
    {
      feclog("getDeviceIdFromName: devName %s not found",devName);
      return -illegal_device;
    }
    strncpy(devNameTags[i],devName,16);
    ndevices++;
  }
  return i;
}

PM_EXPORT int RecordEvent(UINT32 eventNumber,UINT32 eventCode,char *devName)
{
  int i,id;
  char devname[48];
  DTYPE din;

  aesContext.version = ARCEQM_VERSION;
  aesContext.eventNumber = eventNumber;
  aesContext.eventCode   = eventCode;
  strncpy(aesContext.deviceName,devName,16);

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESCON");
  din.dArrayLength = 1;
  din.data.vptr = &aesContext;

  if ((i=getDeviceIdFromName(devName)) < 0) return -i;
  sprintf(devname,"%s/#%d",EVENTMGRTAG,i);
  id = AttachLink(devname,"RECORD",NULL,&din,CA_WRITE,0,RcdEventCallback,CM_SINGLE);
  if (id >= 0 && ((i=getCallbackSlot()) >= 0))
  {
    callbackTable[i].id = id;
  }
  return id < 0 ? -id : 0;
}

int defaultTriggerCallback(UINT32 eventNumber,UINT32 eventCode)
{
  int cc = 0;
  feclog("reacting to event %d code %d",eventNumber,eventCode);
  cc = RecordEvent(eventNumber,eventCode,"NULL");
  return cc;
}
int queryArchiveInfo(AEScontext *aescon,AIB *aib)
{
  if (aib == NULL) return un_allocated;
  return GetLocalArchiveInfo(aib,aescon->eventNumber,aescon->deviceName);
}

int (*queryPropertyInfoFunction)(PIB *pib,UINT32 eventNumber,char *deviceServer,char *devProperty) = NULL;
PM_EXPORT void setQueryPropertyInfoFunction(int (*fcn)(PIB *pib,UINT32 eventNumber,char *deviceServer,char *devProperty))
{
  queryPropertyInfoFunction = fcn;
}
int queryPropertyInfo(AEScontext *aescon,char *property,PIB *pib)
{
  int cc = not_implemented;
  if (pib == NULL) return un_allocated;
  if (queryPropertyInfoFunction) 
    cc = (*queryPropertyInfoFunction)(pib,aescon->eventNumber,aescon->deviceName,property);
  return cc;
}
int queryPropertyInfoEx(char *fullName,UINT32 evnnum,PrpQueryStruct *pqs)
{
  if (pqs == NULL || fullName == NULL) return un_allocated;
  return GetLocalPropertyInfo(fullName,evnnum,pqs);
}
void arctsk(void)
{
  if (GotTriggerEvent)
  {
    respondToTrigger(aesNotify.eventNumber,aesNotify.eventCode);
    GotTriggerEvent = 0;
  }
}
PM_EXPORT int SetGlobalEventComment(UINT32 eventCode,char *cmt,char *mode)
{
  AESnotify aesn;
  char devname[48];
  short modeNo = strchr(mode,'a') ? 1 : 0;
  DTYPE din;

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESNOT");
  din.dArrayLength = 1;
  din.data.vptr = &aesn;
  aesn.version = ARCEQM_VERSION;
  aesn.eventCode = eventCode;
  strncpy(aesn.comment[0],cmt,sizeof(aesn.comment));
  sprintf(devname,"%s/#%d",EVENTMGRTAG,modeNo);
  return ExecLink(devname,"SETCOMMENT",NULL,&din,CA_WRITE);
}
PM_EXPORT int SetGlobalEvent(UINT32 eventCode,char *cmt)
{
  int i,id;
  static AESnotify aesn;
  DTYPE din,dout;

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESNOT");
  din.dArrayLength = 1;
  din.data.vptr = &aesn;

  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  g_eventNumber = (UINT32)time(NULL); /* set it to now in case the call fails! */
  dout.data.ulptr = &g_eventNumber;

  aesn.version = ARCEQM_VERSION;
  aesn.eventCode = eventCode;  
  strncpy(aesn.comment[0],cmt,64);
  aesn.comment[1][0] = 0; aesn.comment[2][0] = 0; aesn.comment[3][0] = 0;

  id = AttachLink(EVENTMGRTAG,"SET",&dout,&din,CA_WRITE|CA_READ,0,SetGlobalEventCallback,CM_SINGLE);
  if (id >= 0 && ((i=getCallbackSlot()) >= 0))
  {
    callbackTable[i].id = id;
  }
  return id < 0 ? -id : 0;
}
/* ancient style: */
int _GetArchiveData(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec)
{
   if (tineDebug) dbglog("GetData : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return GetLocalArchiveData(d,evnNumber,starttime,stoptime,evnContext,devspec);
}
/* old style: */
int _GetArchiveDataEx(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec)
{
   if (tineDebug) dbglog("GetData : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return GetLocalArchiveDataEx(d,evnNumber,starttime,stoptime,evnContext,devspec);
}
/* new style: */
int _GetData(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec)
{
   if (tineDebug) dbglog("GetData : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return getArchiveData(d,evnNumber,starttime,stoptime,evnContext,devspec);
}
/* ancient style: */
int _GetArchiveHeader(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec)
{
   if (tineDebug) dbglog("GetHeader : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return GetLocalArchiveHeader((DHS *)dhs,evnNumber,starttime,stoptime,evnContext,devspec);
}
/* old style: */
int _GetArchiveHeaderEx(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec)
{
   if (tineDebug) dbglog("GetHeader : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return GetLocalArchiveHeaderEx((DHS *)dhs,evnNumber,starttime,stoptime,evnContext,devspec);
}
/* new style: (but old output header) */
int _GetHeader(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec)
{
   if (tineDebug) dbglog("GetHeader : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return getArchiveHeader((DHS *)dhs,evnNumber,starttime,stoptime,evnContext,devspec);
}
int closeArcFile(UINT32 evnNumber,char *evnTrigger)
{
   if (tineDebug) dbglog("CloseArchiveFile : event %d trigger %s",evnNumber,evnTrigger);
   return closeArchiveFile(evnNumber,evnTrigger);
}

UINT32 archiveState = 0;
UINT32 archiveWindow = 30;
enum {
  TriggerNormal = 0,
  TriggerNext,
  TriggerWindow,
  TriggerRefuse
} archiveStates;
int isTriggerAllowed(void)
{
  static time_t lasttrigger = 0;
  time_t t = time(NULL);
  
  switch (archiveState)
  {
    case TriggerNormal: 
      lasttrigger = t;
      return TRUE;
    case TriggerNext:
      lasttrigger = t;
      archiveState = TriggerRefuse;
      return TRUE;
    case TriggerWindow:
      if (t-lasttrigger > (time_t)archiveWindow)
      {
        lasttrigger = t;
        return TRUE;
      }
      return FALSE;
    case TriggerRefuse:
      return FALSE;
  }
  return FALSE;
}
#define assertIsStruct(dt,tg) \
if ((dt)->dFormat != CF_STRUCT) return illegal_format;\
if (strcmp((dt)->dTag, tg)) return invalid_structure_tag;\
if ((dt)->dArrayLength < 1) return dimension_error

#define assertIsSingleStruct(dt,tg) \
if ((dt)->dFormat != CF_STRUCT) return illegal_format;\
if (strcmp((dt)->dTag, tg)) return invalid_structure_tag;\
if ((dt)->dArrayLength != 1) return dimension_error

PM_EXPORT int arceqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char caller[USERNAME_SIZE+8];
  time_t eventTime;
  int pid,i,k,p,n,nevn,nlen,cc=0,reverseSort=FALSE;
  static AES aes[20];
  NAME32 *n32;
  NAME64 *n64;
  UINT32 starttime,stoptime,evnnum = 0;
  UINT32 ivals[4];
  char devServerName[128];
  char device[DEVICE_NAME_SIZE];
  DEVSPEC devsp;
  char *c,*ca,*ctx,*srv,*dev,*prp,*trg;

  memset(devServerName,0,128);
  memset(device,0,DEVICE_NAME_SIZE);
  strncpy(caller,GetCaller(ARCEQM_TAG),USERNAME_SIZE);

  pid = GetPropertyId(ARCEQM_TAG, devProperty);
  switch (pid)
  {
    case PRP_REMOVE:
    {
      UINT32 evns[2]; // start and stop events
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (din->dArrayLength < 1) return dimension_error;
      if ((cc=GetValuesAsUnsignedLong(din, evns, 2)) != 0) return cc;
      if (din->dArrayLength == 1) evns[1] = evns[0];
      cc = RemoveLocalArchiveEvents(evns[0], evns[1], devName);
      if (cc == not_implemented) cc = RemoveLocalArchiveEvent(evns[0], devName);
      return cc;
    }
    case PRP_ANNOTATE:
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (din->dFormat != CF_TEXT) return illegal_format;
      if (din->dArrayLength == 0) return dimension_error;
      if ((cc=parseAnnotationAndEvent(din->data.cptr, &c, &eventTime)) != 0) return cc;
      evnnum = (UINT32)eventTime;
      return AnnotateLocalArchiveEvent(evnnum, devName, c);
    case PRP_ARCHSTATE:
      if (access & CA_WRITE)
      {
        if ((cc=getValuesAsLong(din, (SINT32 *)&archiveState, 1)) != 0) return cc;
      }
      if (dout->dArrayLength > 0) cc = putValuesFromUnsignedLong(dout, &archiveState, 1);
      return cc;
    case PRP_ARCHWINDOW:
      if (access & CA_WRITE)
      {
        if ((cc=getValuesAsLong(din, (SINT32 *)&archiveWindow, 1)) != 0) return cc;
      }
      if (dout->dArrayLength > 0) cc = putValuesFromUnsignedLong(dout, &archiveWindow, 1);
      return cc;
    case PRP_ARCHINFO: // does not appear to used or to return anything !
      if (access & CA_WRITE) return illegal_read_write;
      assertIsSingleStruct(din, "AESCON");
      assertIsSingleStruct(dout, "AIB");
      return queryArchiveInfo((AEScontext *)din->data.vptr, (AIB *)dout->data.vptr);
    case PRP_PRPINFO:
      if (access & CA_WRITE) return illegal_read_write;
      if (din->dArrayLength > 0)
      {
        if ((cc=GetValuesAsLong(din, (SINT32 *)&evnnum, 1)) != 0) return cc;
      }
      assertIsSingleStruct(dout, "PRPQSr4");
      return queryPropertyInfoEx(devName, evnnum, (PrpQueryStruct *)dout->data.vptr);
    case PRP_TRIGGER: 
      // this ends up starting event recording for everyon in the DB!
      if (strncmp(caller,EVENTMGRFEC,USERNAME_SIZE) != 0) return not_supported;
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isTriggerAllowed()) return not_accepted;
      assertIsSingleStruct(din, "AESNOT");
      memcpy(&aesNotify, din->data.vptr, sizeof(AESnotify));
      if (aesNotify.eventCode == EVENT_IGNORE) return 0;
      if (tineDebug) dbglog("EVNEQM: received TRIGGER %d from %.16s", aesNotify.eventNumber, caller);
      /* set trigger flag so that the call returns quickly */
      GotTriggerEvent = 1;
      /* handle the trigger asynchronously */
      return 0;
    case PRP_CLOSEARCHIVE:
      if (access & CA_WRITE) return illegal_read_write;
      if ((cc=GetValuesAsUnsignedLong(din, &evnnum, 1)) != 0) return cc;
      return closeArcFile(evnnum, devName);
    case PRP_ARCHDATA:
      if (access & CA_WRITE) return illegal_read_write;
      if (din->dArrayLength < 1) return dimension_error;
      if (dout->dArrayLength < 1) return dimension_error;
      switch (din->dFormat)
      { // a lot of legacy stuff still being supported here ...
        case CF_INT32:
          stoptime = (UINT32)time(NULL);
          starttime = stoptime - 3600 * 24 * 3;
          switch (din->dArrayLength)
          {
            case 3: stoptime = din->data.lptr[2];
            case 2: starttime = din->data.lptr[1];
            case 1: evnnum = din->data.lptr[0];
              break;
            default:
              return dimension_error;
          }
          /* devName must be of the form: <trigger>/<context>/<server>/<device>[property] */
          if ((c=strchr(devName, '/')) == NULL) return argument_list_error;
          if ((ca=strchr(devName, ':')) != NULL)
          { /* of the form: <trigger>:/<context>/<server>/<device>[property] */
            *ca++ = 0; c = ca;
            if (*c != '/') return argument_list_error;
          }
          *c++ = 0; ctx = c;
          if ((c = strchr(ctx, '/')) == NULL) return argument_list_error;
          *c++ = 0; srv = c;
          if ((c = strchr(srv, '/')) == NULL) return argument_list_error;
          *c++ = 0; dev = c;
          if ((c = strchr(dev, '[')) == NULL) return argument_list_error;
          *c++ = 0; prp = c;
          if ((c = strchr(prp, ']')) == NULL) return argument_list_error;
          *c = 0;
          strncpy(devsp.deviceContext, ctx, CONTEXT_NAME_SIZE);
          strncpy(devsp.deviceServerName, srv, EXPORT_NAME_SIZE);
          strncpy(devsp.deviceProperty, prp, PROPERTY_NAME_SIZE);
          strncpy(devsp.deviceName, dev, DEVICE_NAME_SIZE);
          memset(devsp.dataTag, 0, TAG_NAME_SIZE);
          devsp.dataFormat = dout->dFormat;
          devsp.dataIndex = 0;
          devsp.dataDimension = dout->dArrayLength;
          n = _GetData(dout, evnnum, starttime, stoptime, devName, &devsp);
          if (n < 0) return -n;
          dout->dArrayLength = n;
          return 0;
        case CF_STRUCT:
          if (!strcmp(din->dTag, "ADRBr4"))
          {
            g_adrb = (ADRB *)din->data.vptr;
            g_adrb->devSpecifier.dataDimension = dout->dArrayLength;
            g_adrb->devSpecifier.dataFormat = dout->dFormat;
            strncpy(g_adrb->devSpecifier.dataTag, dout->dTag, TAG_NAME_SIZE);
            /* use devName as event context */
            n = _GetData(dout, g_adrb->eventNumber, g_adrb->starttime, g_adrb->stoptime, devName, &g_adrb->devSpecifier);
            if (n < 0) return -n;
            dout->dArrayLength = n;
            return 0;
          }
          if (!strcmp(din->dTag, "ADRB"))
          {
            g_adrb5 = (ADRB5 *)din->data.vptr;
            g_adrb5->devSpecifier.dataDimension = dout->dArrayLength;
            g_adrb5->devSpecifier.dataFormat = dout->dFormat;
            /* use devName as event context */
            n = _GetArchiveData(dout, g_adrb5->eventNumber, g_adrb5->starttime, g_adrb5->stoptime, devName, &g_adrb5->devSpecifier);
            if (n < 0) return -n;
            dout->dArrayLength = n;
            return 0;
          }
          if (!strcmp(din->dTag, "ADRBX"))
          {
            g_adrbx = (ADRBX *)din->data.vptr;
            g_adrbx->devSpecifier.dataDimension = dout->dArrayLength;
            g_adrbx->devSpecifier.dataFormat = dout->dFormat;
            /* use devName as event context */
            n = _GetArchiveDataEx(dout, g_adrbx->eventNumber, g_adrbx->starttime, g_adrbx->stoptime, devName, &g_adrbx->devSpecifier);
            if (n < 0) return -n;
            dout->dArrayLength = n;
            return 0;
          }
          return invalid_structure_tag;
        default:
          return illegal_format;
      }
    case PRP_ARCHHEADER:
      /* n.b. 'starttime' and 'stoptime' are fairly superfluous below: the event Nr says it all */
      if (access & CA_WRITE)
      { /* can update static header information only */
        assertIsStruct(din, "PHSr4");
        return dumpPHS((PHS *)din->data.vptr, din->dArrayLength);
      }
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (dout->dArrayLength < 1) return dimension_error;
      if (din->dArrayLength < 1) return dimension_error;
      switch (din->dFormat)
      {
        case CF_INT32:
          stoptime = (UINT32)time(NULL);
          starttime = stoptime - 3600 * 24 * 3;
          switch (din->dArrayLength)
          {
            case 3: stoptime = din->data.lptr[2];
            case 2: starttime = din->data.lptr[1];
            case 1: evnnum = din->data.lptr[0];
              break;
            default:
              return dimension_error;
          }
          /* devName must be of the form: <event trigger>/<context>/<server>/<device>[property] */
          if ((c=strchr(devName, '/')) == NULL) return argument_list_error;
          *c++ = 0; ctx = c;
          if ((c=strchr(ctx, '/')) == NULL) return argument_list_error;
          *c++ = 0; srv = c;
          if ((c=strchr(srv, '/')) == NULL) return argument_list_error;
          *c++ = 0; dev = c;
          if ((c=strchr(dev, '[')) == NULL) return argument_list_error;
          *c++ = 0; prp = c;
          if ((c=strchr(prp, ']')) == NULL) return argument_list_error;
          *c = 0;
          strncpy(devsp.deviceContext, ctx, CONTEXT_NAME_SIZE);
          strncpy(devsp.deviceServerName, srv, EXPORT_NAME_SIZE);
          strncpy(devsp.deviceProperty, prp, PROPERTY_NAME_SIZE);
          strncpy(devsp.deviceName, dev, DEVICE_NAME_SIZE);
          memset(devsp.dataTag, 0, TAG_NAME_SIZE);
          devsp.dataFormat = dout->dFormat;
          devsp.dataIndex = 0;
          devsp.dataDimension = dout->dArrayLength;
          /* devName now gives the event trigger */
          break;
        case CF_STRUCT:
          if (!strcmp(din->dTag, "ADRBr4"))
          {
            g_adrb = (ADRB *)din->data.vptr;
            memcpy(&devsp, &g_adrb->devSpecifier, sizeof(DEVSPEC));
            evnnum = g_adrb->eventNumber;
            break;
          }
          if (!strcmp(din->dTag, "ADRB"))
          {
            g_adrb5 = (ADRB5 *)din->data.vptr;
            strncpy(devsp.deviceContext, gDeviceContext, CONTEXT_NAME_SIZE);
            strncpy(devsp.deviceServerName, g_adrb5->devSpecifier.deviceServerName, 16);
            strncpy(devsp.deviceProperty, g_adrb5->devSpecifier.deviceProperty, 16);
            strncpy(devsp.deviceName, g_adrb5->devSpecifier.deviceName, 16);
            memset(devsp.dataTag, 0, TAG_NAME_SIZE);
            devsp.dataFormat = dout->dFormat;
            devsp.dataIndex = 0;
            devsp.dataDimension = dout->dArrayLength;
            evnnum = g_adrb5->eventNumber;
            break;
          }
          if (!strcmp(din->dTag, "ADRBX"))
          {
            g_adrbx = (ADRBX *)din->data.vptr;
            strncpy(devsp.deviceContext, g_adrbx->devSpecifier.deviceContext, CONTEXT_NAME_SHORTSIZE);
            strncpy(devsp.deviceServerName, g_adrbx->devSpecifier.deviceServerName, EXPORT_NAME_SHORTSIZE);
            strncpy(devsp.deviceProperty, g_adrbx->devSpecifier.deviceProperty, PROPERTY_NAME_SHORTSIZE);
            strncpy(devsp.deviceName, g_adrbx->devSpecifier.deviceName, 32);
            memset(devsp.dataTag, 0, TAG_NAME_SIZE);
            devsp.dataFormat = dout->dFormat;
            devsp.dataIndex = 0;
            devsp.dataDimension = dout->dArrayLength;
            evnnum = g_adrbx->eventNumber;
            break;
          }
        default:
          return illegal_format;
      }
      return getArcHdr(dout, evnnum, devName, &devsp);
    case PRP_NUMEVENTS:
      if (access & CA_WRITE) return illegal_read_write;
      if ((cc = GetValuesAsLong(din, ivals, 3)) != 0) return cc;
      evnnum = 0;
      switch (din->dArrayLength)
      {
        case 3: evnnum = ivals[2];
        case 2: starttime = ivals[0];
          stoptime = ivals[1];
          break;
        default:
          return dimension_error;
      }
      n = GetNumLocalArchivedEvents(devName, starttime, stoptime, evnnum);
      return PutValuesFromLongEx(dout, &n, 1, 0);
    case PRP_EVENTS:
      if (access & CA_WRITE) return illegal_read_write;
      stoptime = (UINT32)time(NULL);
      starttime = stoptime - 3600 * 24 * 3;
      if (din->dArrayLength > 0)
      {
        if ((cc=GetValuesAsLong(din, ivals, 2)) != 0) return cc;
        starttime = ivals[0];
        if (din->dArrayLength > 1) stoptime = ivals[1];
      }
      if (dout->dFormat != CF_INT32 && dout->dFormat != CF_UINT32) return illegal_format;
      /* devName either trigger (1st choice) or server name */
      n = GetLocalArchivedEvents(devName, starttime, stoptime, dout->data.ulptr, dout->dArrayLength,&cc);
      if (n < 0) return -n;
      if (cc != 0) cc = CE_SENDDATA + cc;
      dout->dArrayLength = n;
      return cc;
    case PRP_NDSERVERS:
      if (access & CA_WRITE) return illegal_read_write;
      if (dout->dArrayLength != 1) return dimension_error;
      n = GetLocalnumDeviceServers();
      return PutValuesFromLongEx(dout, &n, 1, 0);
    case PRP_DSERVERS:
      if (access & CA_WRITE) return illegal_read_write;
      starttime = 0; stoptime = (UINT32)time(NULL);
      if (din->dArrayLength > 0)
      {
        if (din->dArrayLength == 1 && dout->dFormat == CF_NAME32)
        {
          n = dout->dArrayLength;
          n32 = (NAME32 *)dout->data.vptr;
          cc = getArchiveServers(din->data.lptr[0], devName, NULL, n32, &n);
          dout->dArrayLength = n;
          return cc;
        }
        if (din->dArrayLength != 2) return dimension_error;
        if (din->dFormat != CF_INT32) return illegal_format;
        starttime = din->data.lptr[0];
        stoptime = din->data.lptr[1];
      }
      switch (dout->dFormat)
      {
        case CF_NAME64:
          n = GetLocalDeviceServersEx(devName, (NAME64 *)dout->data.vptr, starttime, stoptime, 0, (int)dout->dArrayLength);
          if (n >= 0) dout->dArrayLength = n;
          return 0;
        case CF_NAME32:
          n32 = (NAME32 *)dout->data.vptr;
          break;
        case CF_NAME16:
          n32 = (NAME32 *)qryMemPool;
          break;
        default:
          return illegal_format;
      }
      n = GetLocalDeviceServers(devName, n32, starttime, stoptime, 0, (int)dout->dArrayLength);
      if (dout->dFormat == CF_NAME16 && n > 0)
        N1toN2((NAME16 *)dout->data.vptr, n32, dout->dArrayLength, 16);
      if (n >= 0) dout->dArrayLength = n;
      return 0;
    case PRP_NADEVICES:
    {
      char trigger[CONTEXT_NAME_SIZE];
      trigger[0] = 0;
      if (access & CA_WRITE) return illegal_read_write;
      if (strlen(devName) == 0) return illegal_device_number;
      strncpy(devServerName, devName, DEVICE_NAME_SIZE);
      if (din->dArrayLength > 0)
      {
        if (din->dArrayLength != 1) return dimension_error;
        switch (din->dFormat)
        {
          case CF_INT32:
            evnnum = din->data.lptr[0];
            break;
          case CF_NAME32:
          case CF_NAME16:
            nlen = fmtsizeof(din->dFormat);
            strncpy(trigger, din->data.cptr, nlen);
            break;
          case CF_NAME32I:
            strncpy(trigger, ((NAME32I *)din->data.vptr)[0].name, 32);
            evnnum = ((NAME32I *)din->data.vptr)[0].ival;
            break;
          case CF_NAME16I:
            strncpy(trigger, ((NAME16I *)din->data.vptr)[0].name, 16);
            evnnum = ((NAME16I *)din->data.vptr)[0].ival;
            break;
          default: 
            return illegal_format;
        }
      }
      n = GetLocalnumDevices(devServerName, trigger, evnnum);
      return PutValuesFromLongEx(dout, &n, 1, 0);
    }
    case PRP_ADEVICES:
    { // StoredDevices
      char trigger[64];
      trigger[0] = 0;
      if (access & CA_WRITE) return illegal_read_write;
      if (strlen(devName) == 0) return illegal_device_number;
      strncpy(devServerName, devName, 128);
      if (devServerName[0] != '/' && strchr(devServerName, ':') != NULL)
      {
        trg = devServerName;
        if ((ctx = strchr(trg, ':')) == NULL) return argument_list_error;
        *ctx++ = 0;
        if (ctx[0] != '/')
        {
          srv = ctx;
          ctx = gDeviceContext;
        }
        else
        {
          ctx++;
          if ((srv = strchr(ctx, '/')) == NULL) return argument_list_error;
          *srv++ = 0;
        }
        if ((prp = strchr(srv, '[')) != NULL)
        {
          *prp++ = 0;
          if ((c = strchr(prp, ']')) != NULL) *c = 0;
        }
      }
      else
      {
        trg = NULL;
        srv = devServerName;
      }
      if (din->dArrayLength > 0)
      {
        if (din->dArrayLength != 1) return dimension_error;
        switch (din->dFormat)
        {
          case CF_INT32:
            evnnum = din->data.lptr[0];
            if (trg != NULL)
            { // devName in form of trigger:/context/server[property]
              n = dout->dArrayLength;
              n64 = (NAME64 *)dout->data.vptr;
              cc = getArchiveDevices(evnnum, trg, ctx, srv, prp, n64, &n);
              dout->dArrayLength = n;
              return cc;
            }
            break;
          case CF_NAME32:
          case CF_NAME16:
            nlen = fmtsizeof(din->dFormat);
            strncpy(trigger, din->data.cptr, nlen);
            break;
          case CF_NAME32I:
            strncpy(trigger, ((NAME32I *)din->data.vptr)[0].name, 32);
            evnnum = ((NAME32I *)din->data.vptr)[0].ival;
            break;
          case CF_NAME16I:
            strncpy(trigger, ((NAME16I *)din->data.vptr)[0].name, 16);
            evnnum = ((NAME16I *)din->data.vptr)[0].ival;
            break;
          default: return illegal_format;
        }
      }
      switch (dout->dFormat)
      {
        case CF_NAME64:
          n64 = (NAME64 *)dout->data.vptr;
          break;
        case CF_NAME32:
        case CF_NAME16:
          n64 = (NAME64 *)qryMemPool;
          break;
        default:
          return illegal_format;
      }
      n = GetLocalDevices(n64, srv, trigger, evnnum, dout->dArrayLength);
      if (n < 0) return -n;
      switch (dout->dFormat)
      {
        case CF_NAME32:
          N1toN2((NAME32 *)dout->data.vptr, n64, dout->dArrayLength, 32);
          break;
        case CF_NAME16:
          N1toN2((NAME16 *)dout->data.vptr, n64, dout->dArrayLength, 16);
          break;
      }
      dout->dArrayLength = n;
      return 0;
    }
    case PRP_NAPROPS:
    {
      char trgstr[DEVICE_NAME_SIZE] = {""};
      if (access & CA_WRITE) return illegal_read_write;
      if (strlen(devName) == 0) return illegal_device_number;
      strncpy(devServerName, devName, DEVICE_NAME_SIZE);
      if (din->dArrayLength > 0)
      {
        if (din->dArrayLength != 1) return dimension_error;
        switch (din->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            nlen = fmtsizeof(din->dFormat);
            break;
          default:
            return illegal_format;
        }
        strncpy(trgstr, din->data.cptr, nlen);
      }
      n = GetLocalnumDeviceProperties(devServerName, trgstr);
      return PutValuesFromLongEx(dout, &n, 1, 0);
    }
    case PRP_NDEVEVENTS:
    {
      UINT32 start = 0, stop = (UINT32)time(NULL);
      if (access & CA_WRITE) return illegal_read_write;
      if (strlen(devName) == 0) return illegal_device_number;
      strncpy(devServerName, devName, DEVICE_NAME_SIZE);
      strncpy(device, DEF_DEV_NAME, 16);
      if (din->dArrayLength > 0)
      {
        switch (din->dFormat)
        {
        case CF_INT32:
          if (din->dArrayLength != 2) return dimension_error;
          start = din->data.lptr[0];
          stop = din->data.lptr[1];
          break;
        case CF_NAME16I:
          if (din->dArrayLength != 2) return dimension_error;
          start = ((NAME16I *)din->data.vptr)[0].ival;
          stop = ((NAME16I *)din->data.vptr)[1].ival;
          strncpy(device, ((NAME16I *)din->data.vptr)[0].name, 16);
          break;
        }
      }
      nevn = GetNumLocalDeviceArchivedEvents(devServerName, device, start, stop);
      if (nevn < 0) return -nevn;
      return PutValuesFromLongEx(dout, &nevn, 1, 0);
    }
    case PRP_APROPS:
    {
      char devn[DEVICE_NAME_SIZE] = {""};
      if (access & CA_WRITE) return illegal_read_write;
      if (strlen(devName) == 0) return illegal_device_number;
      strncpy(devServerName, devName, 128);
      if (din->dArrayLength > 0)
      {
        if (din->dArrayLength != 1) return dimension_error;
        switch (din->dFormat)
        {
          case CF_INT32:
            // devName in form of trigger:/context/server/device
            trg = devServerName;
            if (trg[0] == '/') return argument_list_error;
            if ((ctx = strchr(trg, ':')) == NULL) return argument_list_error;
            *ctx++ = 0;
            if (ctx[0] != '/')
            {
              srv = ctx; ctx = NULL;
            }
            else
            {
              ctx++;
              if ((srv = strchr(ctx, '/')) == NULL) return argument_list_error;
              *srv++ = 0;
            }
            if ((dev = strchr(srv, '/')) != NULL) *dev++ = 0;
            n = dout->dArrayLength;
            n64 = (NAME64 *)dout->data.vptr;
            cc = getArchiveProperties(din->data.lptr[0], trg, ctx, srv, dev, n64, &n);
            dout->dArrayLength = n;
            return cc;
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            nlen = fmtsizeof(din->dFormat);
            break;
          default:
            return illegal_format;
        }
        strncpy(devn, din->data.cptr, nlen);
      }
      switch (dout->dFormat)
      {
        case CF_NAME64:
          n64 = (NAME64 *)dout->data.vptr;
          break;
        case CF_NAME32:
        case CF_NAME16:
          n64 = (NAME64 *)qryMemPool;
          break;
        default:
          return illegal_format;
      }
      n = GetLocalDeviceProperties(n64, devServerName, devn, dout->dArrayLength);
      if (n < 0) return -n;
      switch (dout->dFormat)
      {
        case CF_NAME32:
          N1toN2((NAME32 *)dout->data.vptr, n64, dout->dArrayLength, 32);
          break;
        case CF_NAME16:
          N1toN2((NAME16 *)dout->data.vptr, n64, dout->dArrayLength, 16);
          break;
      }
      dout->dArrayLength = n;
      return 0;
    }
    case PRP_DEVEVENTS:
    {
      UINT32 stop = (UINT32)time(NULL), start = stop - 86400 * 3; /* defaults: 3 days ago -> now */
      if (access & CA_WRITE) return illegal_read_write;
      /* devName must be of the form: <trigger>/<context>/<server>/<device>[property] */
      ctx = srv = dev = NULL;
      if ((c = strchr(devName, '/')) != NULL)
      {
        if ((c = strchr(devName, ':')) != NULL)
        { /* or of the form: <trigger>:/<context>/<server>/<device>[property] */
          *c++ = 0;
          if (*c != '/') return argument_list_error;
        }
        *c++ = 0; ctx = c;
        if ((c = strchr(ctx, '/')) != NULL)
        {
          *c++ = 0; srv = c;
          if ((c = strchr(srv, '/')) != NULL)
          {
            *c++ = 0; dev = c;
            if ((c = strchr(dev, '[')) != NULL) *c++ = 0;
          }
        }
      }
      if (strlen(devName) == 0) return illegal_device_number;
      if (din->dArrayLength > 0)
      {
        switch (din->dFormat)
        {
          case CF_INT32:
            if (din->dArrayLength < 2) return dimension_error;
            if (din->dArrayLength > 2) reverseSort = din->data.lptr[2];
            start = din->data.lptr[0];
            stop = din->data.lptr[1];
            break;
          case CF_NAME16I:
            if (din->dArrayLength < 2) return dimension_error;
            start = ((NAME16I *)din->data.vptr)[0].ival;
            stop = ((NAME16I *)din->data.vptr)[1].ival;
            strncpy(device, ((NAME16I *)din->data.vptr)[0].name, 16);
            break;
          default:
            return illegal_format;
        }
      }
      if (dout->dFormat == CF_STRUCT && !strcmp(dout->dTag, "AES"))
      {
        n = GetLocalDeviceArchivedEvents(devName, srv, dev, start, stop, (AES *)dout->data.vptr, dout->dArrayLength, reverseSort, &cc);
      }
      else if (dout->dFormat == CF_INT32)
      {
        n = GetLocalDeviceArchivedEventList(devName, srv, dev, start, stop, dout->data.ulptr, dout->dArrayLength, reverseSort, &cc);
      }
      else
      {
        return illegal_format;
      }
      if (n < 0) return -n;
      if (cc != 0) cc = CE_SENDDATA + cc;
      if (n < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)n;
      return cc;
    }
    case PRP_TRIGGERS:
    {
      char prp[DEVICE_NAME_SIZE] = {""};
      if (strlen(devName) > 0 && devName[0] != '#') strncpy(devServerName, devName, DEVICE_NAME_SIZE);
      if (din->dArrayLength > 0)
      {
        if (din->dArrayLength != 1) return illegal_read_write;
        switch (din->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            nlen = fmtsizeof(din->dFormat);
            break;
          default:
            return illegal_format;
        }
        strncpy(prp, din->data.cptr, nlen);
      }
      switch (dout->dFormat)
      {
        case CF_NAME64:
          n64 = (NAME64 *)dout->data.vptr;
          break;
        case CF_NAME32:
        case CF_NAME16:
          n64 = (NAME64 *)qryMemPool;
          break;
        default:
          return illegal_format;
      }
      n = GetLocalTriggers(n64, devServerName, prp, dout->dArrayLength);
      if (n < 0) return -n;
      switch (dout->dFormat)
      {
        case CF_NAME32:
          N1toN2((NAME32 *)dout->data.vptr, n64, dout->dArrayLength, 32);
          break;
        case CF_NAME16:
          N1toN2((NAME16 *)dout->data.vptr, n64, dout->dArrayLength, 16);
          break;
      }
      dout->dArrayLength = n;
      return 0;
    }
    case PRP_TRC_COMPONENTS_NAM:
      if (din->dArrayLength > 0)
      {
        if (!(access & CA_WRITE)) return illegal_read_write;
        if (!isAdministrator(caller)) return access_denied;
        if (din->dFormat != CF_STRUCT) return illegal_format;
        if (strcmp(din->dTag, "TrgView")) return invalid_structure_tag;
        cc = writeViewerFiles((TrgViewFile *)din->data.vptr, din->dArrayLength);
        if (cc != 0) return cc;
        if ((cc = readViewerFiles()) != 0) return cc;
        resetPropertyDeviceList(ARCEQM_TAG, "TRC.COMPONENTS");
      }
      if (dout->dArrayLength == 0) return 0;
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          nlen = fmtsizeof(dout->dFormat);
          break;
        case CF_STRUCT:
          if (strcmp(dout->dTag, "TrgView")) return invalid_structure_tag;
          nlen = 0;
          break;
        default:
          return illegal_format;
      }
      if (gNumTrgViewFiles == 0) readViewerFiles();
      for (i = 0; i < gNumTrgViewFiles && i < (int)dout->dArrayLength; i++)
      {
        if (nlen > 0)
        {
          strncpy(dout->data.cptr + i * nlen, trgViewFiles[i].desc, nlen);
        }
        else
        {
          ((TrgViewFile *)dout->data.vptr)[i] = trgViewFiles[i];
        }
      }
      dout->dArrayLength = i;
      return 0;
    case PRP_TRC_COMPONENTS:
      if ((access & CA_WRITE)) return illegal_read_write;
      if (dout->dArrayLength == 0) return 0;
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(dout->dTag, "TrgCmp") != 0) return invalid_structure_tag;
      n = GetTriggerComponents(devName, (TrgChannel *)dout->data.vptr, dout->dArrayLength);
      if (n < 0) return -n;
      dout->dArrayLength = n;
      return 0;
    case PRP_TRC_CHANNELS_NAM:
      if (din->dArrayLength > 0)
      { /* use the WRITE to update the components (filename vs. component name) for a config (devName) */
        if (!(access & CA_WRITE)) return illegal_read_write;
        if (!isAdministrator(caller)) return access_denied;
        if (din->dFormat != CF_STRUCT) return illegal_format;
        if (strcmp(din->dTag, "TrgView")) return invalid_structure_tag;
        cc = writeTriggerComponentFile(devName, (TrgViewFile *)din->data.vptr, din->dArrayLength);
        if (cc != 0) return cc;
        for (i = 0; i < (int)din->dArrayLength; i++) readTriggerComponentFiles(&((TrgViewFile *)din->data.vptr)[i]);
        resetPropertyDeviceList(ARCEQM_TAG, "TRC.CHANNELS");
      }
      if (dout->dArrayLength == 0) return 0;
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          nlen = fmtsizeof(dout->dFormat);
          break;
        case CF_STRUCT:
          if (strcmp(dout->dTag, "TrgView")) return invalid_structure_tag;
          nlen = 0;
          break;
        default:
          return illegal_format;
      }
      if (gNumTrgViewFiles == 0) readViewerFiles();
      for (n = 0, i = 0; i < gNumTrgViewFiles && n < (int)dout->dArrayLength; i++)
        for (k = 0; k < trgViewFiles[i].nComponents && n < (int)dout->dArrayLength; k++)
      {
        if (strlen(trgViewFiles[i].components[k].desc) == 0) continue;
        if (nlen > 0)
        { /* a name list requested (no duplicates!) */
          for (p = 0; p < n; p++)
            if (!strnicmp(dout->data.cptr + p * nlen, trgViewFiles[i].components[k].desc, nlen)) break;
          if (p < n) continue;
          strncpy(dout->data.cptr + n * nlen, trgViewFiles[i].components[k].desc, nlen);
        }
        else
        { /* a list of TrgView structs requested */
          if (stricmp(devName, "ALL") && strnicmp(devName, trgViewFiles[i].desc, 64))
          { /* config != requested */
            continue;
          }
          for (p = 0; p < n; p++)
            if (!strnicmp(((TrgViewFile *)dout->data.vptr)[p].desc, trgViewFiles[i].components[k].desc, 64)) break;
          if (p < n) continue;
          strncpy(((TrgViewFile *)dout->data.vptr)[n].desc, trgViewFiles[i].components[k].desc, 64);
          strncpy(((TrgViewFile *)dout->data.vptr)[n].file, trgViewFiles[i].components[k].fileName, 64);
        }
        n++;
      }
      dout->dArrayLength = n;
      return 0;
    case PRP_TRC_CHANNELS:
      if (din->dArrayLength > 0)
      {
        if (!(access & CA_WRITE)) return illegal_read_write;
        if (!isAdministrator(caller)) return access_denied;
        if (din->dFormat != CF_STRUCT) return illegal_format;
        if (strcmp(din->dTag, "TrgChn")) return invalid_structure_tag;
        cc = writeTriggerChannelFile(devName, (TrgChannel *)din->data.vptr, din->dArrayLength);
        if (cc != 0) return cc;
      }
      if (dout->dArrayLength == 0) return 0;
      if (dout->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(dout->dTag, "TrgChn") != 0) return invalid_structure_tag;
      n = GetTriggerChannels(devName, (TrgChannel *)dout->data.vptr, dout->dArrayLength);
      if (n < 0) return -n;
      dout->dArrayLength = n;
      return 0;
    default:
      break;
  } // end property ID case switch

  return illegal_property;
}
PM_EXPORT int initArchiveServer(void)
{
  int i, n;
  NAME32 n32[256];
  char listOfOneDevice[64] = "target";
  char exp[32] = "";
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  gAppendProperyAliasNamesToQueries = FALSE;

  strncpy(exp,ARCEQM_EXPORT,28); 
  if (strlen(thisExtension) > 0) strncat(exp,thisExtension,4);
  RegisterEquipmentModule(exp,ARCEQM_TAG,256,arceqm,arcini,arctsk,100,NULL);

  din.dArrayLength = 1;
  din.dFormat = CF_INT32;
  RegisterPropertyInformation(ARCEQM_TAG, "REMOVE", NULL, &din, CA_WRITE, AT_SCALAR,0,"remove event",PRP_REMOVE,NULL);
  AssignDeviceListToProperty(ARCEQM_TAG, "REMOVE", arbTrgList, arbTblSiz);

  din.dArrayLength = 1024;
  din.dFormat = CF_TEXT;
  RegisterPropertyInformation(ARCEQM_TAG, "ANNOTATE", NULL, &din, CA_WRITE, AT_UNKNOWN, 0, "annotate event", PRP_ANNOTATE, NULL);
  AssignDeviceListToProperty(ARCEQM_TAG, "ANNOTATE", arbTrgList, arbTblSiz);

  din.dArrayLength = 1;
  din.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  RegisterPropertyInformation(ARCEQM_TAG, "ArchiveTriggerStatus", &dout, &din, CA_WRITE | CA_READ, AT_SCALAR, 0, "read/set archive trigger status", PRP_ARCHSTATE, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "ArchiveTriggerStatus", "ARCHSTATE");
  RegisterPropertyInformation(ARCEQM_TAG, "ArchiveTriggerWindow", &dout, &din, CA_WRITE | CA_READ, AT_SCALAR, 0, "read/set archive trigger window", PRP_ARCHWINDOW, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "ArchiveTriggerWindow", "ARCHWINDOW");

  RegisterPropertyInformation(ARCEQM_TAG, "CloseArchive", NULL, &din, CA_READ, AT_SCALAR, 0, "closed the targeted archive session", PRP_CLOSEARCHIVE, NULL);
  AssignDeviceListToProperty(ARCEQM_TAG, "CloseArchive", arbTrgList, arbTblSiz);

  RegisterPropertyInformation(ARCEQM_TAG, "NumEvents", &dout, NULL, CA_READ, AT_SCALAR, 0, "number of events in a time interval for given event trigger", PRP_NUMEVENTS, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "NumEvents", "NEvents");
  AssignDeviceListToProperty(ARCEQM_TAG, "NumEvents", arbTrgList, arbTblSiz);

  RegisterPropertyInformation(ARCEQM_TAG, "NumStoredServers", &dout, NULL, CA_READ, AT_SCALAR, 0, "number of device servers", PRP_NDSERVERS, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "NumStoredServers", "NDSERVERS");

  RegisterPropertyInformation(ARCEQM_TAG, "NumStoredDevices", &dout, NULL, CA_READ, AT_SCALAR, 0, "number of device servers", PRP_NADEVICES, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "NumStoredDevices", "NADEVICES");

  RegisterPropertyInformation(ARCEQM_TAG, "NumStoredProperties", &dout, NULL, CA_READ, AT_SCALAR, 0, "number of device properties", PRP_NAPROPS, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "NumStoredProperties", "NAPROPS");

  RegisterPropertyInformation(ARCEQM_TAG, "NumDeviceEvents", &dout, NULL, CA_READ, AT_SCALAR, 0, "number of device events", PRP_NDEVEVENTS, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "NumDeviceEvents", "NDEVEVENTS");

  dout.dArrayLength = 1000;
  RegisterPropertyInformation(ARCEQM_TAG, "Events", &dout, NULL, CA_READ, AT_COLLECTION, 0, "events in a time interval for given event trigger", PRP_EVENTS, NULL);
  AssignDeviceListToProperty(ARCEQM_TAG, "Events", arbTrgList, arbTblSiz);

  dout.dArrayLength = 100;
  RegisterPropertyInformation(ARCEQM_TAG, "DeviceEvents", &dout, NULL, CA_READ, AT_COLLECTION, 0, "device specific event list", PRP_DEVEVENTS, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "DeviceEvents", "DEVEVENTS");

  dout.dFormat = CF_NAME32;
  RegisterPropertyInformation(ARCEQM_TAG, "StoredServers", &dout, NULL, CA_READ, AT_COLLECTION, 0, "stored device server names for event trigger given",PRP_DSERVERS,NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "StoredServers", "DSERVERS");
  AssignDeviceListToProperty(ARCEQM_TAG, "StoredServers", arbTrgList, arbTblSiz);

  dout.dFormat = CF_NAME64;
  RegisterPropertyInformation(ARCEQM_TAG, "StoredProperties", &dout, NULL, CA_READ, AT_COLLECTION, 0, "stored property list for event and server given", PRP_APROPS, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "StoredProperties", "APROPS");

  dout.dArrayLength = 1024;
  dout.dFormat = CF_NAME64;
  RegisterPropertyInformation(ARCEQM_TAG, "StoredDevices", &dout, &din, CA_READ, AT_COLLECTION, 0, "stored device list for event and server given", PRP_ADEVICES, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "StoredDevices", "ADEVICES");

  RegisterPropertyInformation(ARCEQM_TAG, "Triggers", &dout, &din, CA_READ, AT_COLLECTION, 0, "stored event trigger list for server and property given", PRP_TRIGGERS, NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "Triggers", "StoredContexts");

  // properties with structure input or output :
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag, "AESNOT");
  RegisterPropertyInformation(ARCEQM_TAG, "TRIGGER", NULL, &din, CA_WRITE, AT_SCALAR, 0, "receive event trigger", PRP_TRIGGER, NULL);
  AssignDeviceListToProperty(ARCEQM_TAG, "TRIGGER", (NAME64 *)listOfOneDevice, 1);

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"AESCON");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"AIB");
  RegisterPropertyInformation(ARCEQM_TAG,"ArchiveInformation",&dout,&din,CA_READ|CA_WRITE,AT_SCALAR,0,"num properties/devices, event code",PRP_ARCHINFO,NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "ArchiveInformation", "ARCHINFO");

  din.dFormat = CF_INT32;
  din.dArrayLength = 1;
  din.dTag[0] = 0;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"PRPQSr4");
  RegisterPropertyInformation(ARCEQM_TAG,"PropertyInformation",&dout,&din,CA_READ,AT_SCALAR,0,"property information header",PRP_PRPINFO,NULL);
  AssignDeviceListToProperty(ARCEQM_TAG, "PropertyInformation", (NAME64 *)listOfOneDevice, 1);

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"ADRB");
  dout.dFormat = CF_BYTE;
  dout.dArrayLength = 4096;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(ARCEQM_TAG,"StoredData",&dout,&din,CA_READ,AT_SCALAR,0,"stored data for given archive data request",PRP_ARCHDATA,NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "StoredData", "ARCHDATA");
  AssignDeviceListToProperty(ARCEQM_TAG, "StoredData", arbTrgList, arbTblSiz);

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"ADRB");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"DHS");
  RegisterPropertyInformation(ARCEQM_TAG,"DataHeader",&dout,&din,CA_READ,AT_SCALAR,0,"data header for given archive data request",PRP_ARCHHEADER,NULL);
  strcpy(dout.dTag,"EVNDHDR"); // overload ...
  RegisterPropertyInformation(ARCEQM_TAG,"DataHeader",&dout,&din,CA_READ,AT_SCALAR,0,"data header for given archive data request",PRP_ARCHHEADER,NULL);
  RegisterPropertyAlias(ARCEQM_TAG, "DataHeader", "ARCHHEADER");
  AssignDeviceListToProperty(ARCEQM_TAG, "DataHeader", arbTrgList, arbTblSiz);

  // database config properties ...
  DTYPEZERO(dout);
  dout.dArrayLength = 64;
  dout.dFormat = CF_NAME64;
  RegisterPropertyInformation(ARCEQM_TAG, "TRC.COMPONENTS.NAM", &dout, NULL, CA_READ, AT_COLLECTION, 0, "component name list", PRP_TRC_COMPONENTS_NAM, NULL);
  RegisterPropertyInformation(ARCEQM_TAG, "TRC.CHANNELS.NAM", &dout, NULL, CA_READ, AT_CHANNEL, 0, "channel name list", PRP_TRC_CHANNELS_NAM, NULL);
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag,"TrgCmp");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.COMPONENTS",&dout,NULL,CA_READ,AT_CHANNEL,0,"Transient Recorder viewer components",PRP_TRC_COMPONENTS,NULL);
  strcpy(dout.dTag,"TrgView");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.COMPONENTS.NAM",&dout,&dout,CA_READ|CA_WRITE,AT_CHANNEL,0,"component name list",PRP_TRC_COMPONENTS_NAM,NULL);
  strcpy(dout.dTag,"TrgChn");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.CHANNELS",&dout,&dout,CA_READ|CA_WRITE,AT_SCALAR,0,"Transient Recorder viewer channels",PRP_TRC_CHANNELS,NULL);
  strcpy(dout.dTag,"TrgView");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.CHANNELS.NAM",&dout,&dout,CA_READ|CA_WRITE,AT_CHANNEL,0,"channel name list", PRP_TRC_CHANNELS_NAM,NULL);

  n = GetLocalDeviceServers(NULL,n32,0,(int)time(NULL),-1,256);
  for (i=0; i<n; i++)
  {
    RegisterDeviceName(ARCEQM_TAG,n32[i].name,i);
  }
  if (n > 0) AssignNumDevices(ARCEQM_TAG,n);

  if (useConnectedLinks)
  {
    feclog("use connected transport communication for large payloads");
  }

  return 0;
}
PM_EXPORT int initArchiveClient(void)
{
  arcini();
  return 0;
}
