#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include "tine.h"
#include "arctools.h"
#include "arceqm.h"
#include "arctypes.h"
#include "apceqm.h"

#define STREQM_TAG "STREQM"
#define STREQM_EXPORT "EVENTSTORE"
#define APCEQM_TAG "APCEQM"
#define APCEQM_EXPORT "EVENTAPC"

#define MAX_NUM_EVENTS 10000

#define unknown_trigger 512
int gMaxNumEvents = MAX_NUM_EVENTS;
char myErrLst[][32] = {
  "unknown trigger"
};
int numMyErr = sizeof(myErrLst)/sizeof(myErrLst[0]);

int parseFullDeviceName(char *fullname,char *dev,char *prp,int isHistoryCall);
void streqm_exit(void);
extern int gUseLoopback;
extern int RemoveEvent(UINT32 evnt, char *ext);
extern int RemoveEvents(UINT32 starttime, UINT32 stoptime, char *ext);
extern int EventNotificationPending;
extern UINT32 lastEventNumber;
void setGetHeaderFunction(int (*fcn)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec));
int makeFullName(char *fname,char *ext);
int writeCMT(char *s,char *fname,int size);
int readCMT(char *s,char *fname,int size);
char *makefilename(char *fname,char *src,char *ext,time_t t);
int getFiles(NAME8 *filebuf,char *fileext,UINT32 starttime,UINT32 stoptime,int bufsiz);
void freeSession(int id);
int GetStoredDeviceNumber(NAME64 *devname,char *tagname,int evnId);
char *GetStoredDeviceNameFromAlias(char *alias,char *tagname,int evnId);
int readComment(int setNr,UINT32 evnt,char *txt,int txtsiz)
{
  char fn[128];
  makefilename(fn,stdArchiveDirectory,arbTbl[setNr].ext,evnt);
  return readCMT(txt,fn,txtsiz);
}
int writeComment(int setNr,UINT32 evnt,char *txt)
{ // txt is less than 1024 chars and NULL terminated ...
  char fn[128];
  makefilename(fn,stdArchiveDirectory,arbTbl[setNr].ext,evnt);
  return writeCMT(txt,fn,(int)strlen(txt));
}

void initArchiveTask(void);
void bkgArchiveTask(void);
int streqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void apctask(void);
void apcinit(void);
int apceqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
int idxSelectedARB = 0;

#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  if (nadmins == 0) return TRUE;
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,USERNAME_SIZE); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),USERNAME_SIZE);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

int respondToTrigger(UINT32 eventNumber,UINT32 eventCode)
{
  int i;
  char str[256];

  sprintf(str,"Received trigger from event %d",eventNumber);
  AppendHistoryLog(str);
  clslog(gDeviceContext,"Services",STREQM_EXPORT,CLOG_STATUS_WARN,CLOG_PRIORITY_IMPORTANT,str);
  if (EventNotificationPending) return 0;  /* nothing to do */
  if (eventNumber == lastEventNumber) return 0;  /* nothing to do */
  if (!(eventCode & MARCHSRV_EVENTCODE)) return 0; /* until KRYK has finished testing */
  for (i=0; i<arbTblSiz; i++) 
    if (arbTbl[i].recordEvents && arbTbl[i].step == 0) arbTbl[i].step = -1;
  return 0;
}
/* at the moment: this is just a long linked list */
/* does it make senes to hash this ? (expect < 100 items) */
LHS *lhsLst = NULL;
LHS *fillLHS(FILE *fp)
{
  int cc = 0, done = 0;
  int prp_col,dev_col,srv_col,dsc_col,siz_col,fmt_col,rat_col,
      egu_col,min_col,max_col,xegu_col,xmin_col,xmax_col,trg_col;
  char str[256],hdr[256],scratch[64];
  LHS *lhs,*list = NULL;
  float fval = 0;
  
  while (csvGetLine(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      str[255] = 0; strcpy(hdr,str); 
      if ((prp_col=findcol(hdr,"PROPERTY")) < 0) ccerr(no_such_line);
      dev_col = findcol(hdr,"DEVICE");
      srv_col = findcol(hdr,"SERVER");
      dsc_col = findcol(hdr,"DESCRIPTION");
      if ((siz_col=findcol(hdr,"ARRAYSIZE")) < 0) ccerr(no_such_line);
      if ((fmt_col=findcol(hdr,"DATATYPE")) < 0) ccerr(no_such_line);
      rat_col = findcol(hdr,"SAMPLINGRATE");
      if ((egu_col=findcol(hdr,"UNITS")) < 0 &&
          (egu_col=findcol(hdr,"UNITS_Y")) < 0 &&
          (egu_col=findcol(hdr,"EGU")) < 0) ccerr(no_such_line);
      xegu_col = findcol(hdr,"UNITS_X");
      if ((min_col=findcol(hdr,"MIN")) < 0 &&
          (min_col=findcol(hdr,"MIN_Y")) < 0 &&
          (min_col=findcol(hdr,"LOWEGU")) < 0) ccerr(no_such_line);
      if ((max_col=findcol(hdr,"MAX")) < 0 &&
          (max_col=findcol(hdr,"MAX_Y")) < 0 &&
          (max_col=findcol(hdr,"HIGHEGU")) < 0) ccerr(no_such_line);
      xmin_col = findcol(hdr,"MIN_X");
      xmax_col = findcol(hdr,"MAX_X");
      trg_col = findcol(hdr,"PRETRIGGER");
      done = 1;                              /* read header info */
      continue;
    }
    if ((lhs=(LHS *)calloc(1,sizeof(LHS))) == NULL) ccerr(out_of_server_memory);
    strncpy(lhs->hdr.prp,colptr(prp_col,str),PROPERTY_NAME_SIZE);
    if (srv_col >= 0) strncpy(lhs->hdr.srv,colptr(srv_col,str),EXPORT_NAME_SIZE);
    if (dev_col >= 0) strncpy(lhs->hdr.dev,colptr(dev_col,str),DEVICE_NAME_SIZE);
    strncpy(scratch,colptr(fmt_col,str),64);
    lhs->hdr.format = getFormatType(scratch);
    lhs->hdr.arrayType = getArrayType(scratch);
    strncpy(scratch,colptr(siz_col,str),16);
    lhs->hdr.arraySize = atoi(scratch);
    if (trg_col >= 0)
    {
      strncpy(scratch,colptr(trg_col,str),16);
      lhs->hdr.preTrigger = atoi(scratch);
    }
    strncpy(lhs->hdr.unitsY,colptr(egu_col,str),16);
    if (xegu_col >= 0) strncpy(lhs->hdr.unitsX,colptr(xegu_col,str),16);
    strncpy(scratch,colptr(min_col,str),16);
    lhs->hdr.scaleMinY = (float)strtod(scratch,NULL);
    strncpy(scratch,colptr(max_col,str),16);
    lhs->hdr.scaleMaxY = (float)strtod(scratch,NULL);
    if (rat_col >= 0) 
    {
      strncpy(scratch,colptr(rat_col,str),16);
      fval = (float)atof(scratch);
      if (fval > 0) lhs->hdr.rangeMaxX = lhs->hdr.arraySize / fval;
    }
    if (xmin_col >= 0)
    {
      strncpy(scratch,colptr(xmin_col,str),16);
      lhs->hdr.rangeMinX = (float)strtod(scratch,NULL);
    }
    if (xmax_col >= 0)
    {
      strncpy(scratch,colptr(xmax_col,str),16);
      lhs->hdr.rangeMaxX = (float)strtod(scratch,NULL);
    }
    lhs->next = list;
    list = lhs;
  }
err:
  return cc ? NULL : list;  
}
LHS *getLHS(char *srv,char *prp,char *dev)
{
  FILE *fp=NULL;
  static int initialized = 0;
  char s[256];
  LHS *lhs=NULL;
  
  for (lhs=lhsLst; lhs != NULL; lhs=lhs->next)
  {
    if (strnicmp(srv,lhs->hdr.srv,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(prp,lhs->hdr.prp,PROPERTY_NAME_SIZE)) continue;
    if (!strcmp(lhs->hdr.dev,DEVICE_IGNORE)) return lhs;
    if (strlen(lhs->hdr.dev) > 0 && strlen(dev) > 0 &&
        strnicmp(dev,lhs->hdr.dev,DEVICE_NAME_SIZE)) continue;
    return lhs;
  }
  sprintf(s,"%s%c%s%c%s%c%s%c%s",
    stdArchiveDirectory,FS_DELIMITER,SERVER_FILES_SUBDIR,FS_DELIMITER,
    srv,FS_DELIMITER,prp,FS_DELIMITER,HEADER_FILE_NAME);
  if ((fp=fopen(s,"r")) == NULL) return NULL;
  lhs = fillLHS(fp);
  fclose(fp);
  if (lhs != NULL)
  {
    if (strlen(lhs->hdr.srv) == 0) strncpy(lhs->hdr.srv,srv,EXPORT_NAME_SIZE);
    lhs->next = lhsLst;
    lhsLst = lhs;
  }
  return lhs;
}
int dumpPHS(PHS *phs,int nitems)
{
  FILE *fp=NULL;
  char s[256],fmtstr[64];
  char *prp, *srv;
  int i, cc = 0;
  if (phs == NULL || nitems == 0) return argument_list_error;
  /* typically a single item is passed with dev = ""
     (srv and prp are redundant in subsequent items) */
  srv = phs[0].srv; prp = phs[0].prp;
  sprintf(s,"%s%c%s%c%s%c%s%c%s",
    stdArchiveDirectory,FS_DELIMITER,SERVER_FILES_SUBDIR,FS_DELIMITER,
    srv,FS_DELIMITER,prp,FS_DELIMITER,HEADER_FILE_NAME);
  if ((fp=fopen(s,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Property, Device, Description, ArraySize, PreTrigger, DataType,");
  fprintf(fp,"Min, Max, Units, MinX, MaxX, UnitsX\n");
  for (i=0; i<nitems; i++)
  { /* could be different information for different device names */
    if (phs[i].arrayType != AT_UNKNOWN) 
      sprintf(fmtstr,"%s.%s",GetFormatAsString(phs[i].format),GetArrayTypeAsString(phs[i].format));
    else
      strcpy(fmtstr,GetFormatAsString(phs[i].format));
    fprintf(fp,"%.64s, %.64s, %.64s, %d, %d, %64s, %g, %g, %.16s, %g, %g, %.16s\n",
      phs[i].prp,phs[i].dev,phs[i].dsc,phs[i].arraySize,phs[i].preTrigger,fmtstr,
      phs[i].scaleMinY,phs[i].scaleMaxY,phs[i].unitsY,
      phs[i].rangeMinX,phs[i].rangeMaxX,phs[i].unitsX);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
DAS *GetARBDas(int idx, char *tag, char *prp)
{
  int i;
  char *c, *ctx=NULL, *srv=tag, tgt[68];
  if (tag == NULL || strlen(tag) == 0) return NULL;
  if (prp == NULL || strlen(prp) == 0) return NULL;
  if (idx < 0 || idx >= arbTblSiz) return NULL;
  strncpy(tgt,tag,64); tgt[64] = 0;
  if (tgt[0] == '/')
  { /* context decorated */
    ctx = &tgt[1];
    if ((c=strchr(ctx,'/')) != NULL)
    { /* looks well-formed */
      *c++ = 0;
      srv = c;
    }
    else
    { /* not well-formed: back to plan A */
      ctx = NULL;
    }
  }
  for (i=0; i<arbTbl[idx].nsteps; i++)
  {
    if (ctx != NULL && strnicmp(ctx,arbTbl[idx].das[i].ctx,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(srv,arbTbl[idx].das[i].srvArc,EXPORT_NAME_SIZE)) continue;
    if (!strnicmp(prp,arbTbl[idx].das[i].prpArc,EXPORT_NAME_SIZE))
      return &arbTbl[idx].das[i];
  }
  return NULL;
}
int isIgnoreDeviceForProperty(int arbTblIdx,char *property)
{
  int i;
  for (i=0; i<arbTbl[arbTblIdx].nsteps; i++) if (isReadable(arbTblIdx,i))
  {
    if (!strnicmp(arbTbl[arbTblIdx].das[i].prpArc,property,PROPERTY_NAME_SIZE))
    {
      if (!strcmp(arbTbl[arbTblIdx].das[i].dev,DEVICE_IGNORE)) return TRUE;
      break;
    }
  }
  return FALSE;
}
int GetARBIndex(char *tag)
{
  int i,j;
  char *c, *ctx=NULL, *srv=tag, tgt[68];
  if (tag == NULL || strlen(tag) == 0) return -1;
  if (arbTbl == NULL || arbTblSiz == 0) return -1;
  /* try tag = event trigger first: */
  i = getSetNumber(tag);
  if (i >= 0 && i < arbTblSiz) return i;

  /* try the selected index  */
  strncpy(tgt,tag,64); tgt[64] = 0;
  if (tgt[0] == '/')
  { /* context decorated */
    ctx = &tgt[1];
    if ((c=strchr(ctx,'/')) != NULL)
    { /* looks well-formed */
      *c++ = 0;
      srv = c;
    }
    else
    { /* not well-formed: back to plan A */
      ctx = NULL;
    }
  }
  for (j=0; j<arbTbl[idxSelectedARB].nsteps; j++)
  {
    if (ctx != NULL && strnicmp(ctx,arbTbl[idxSelectedARB].das[j].ctx,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(srv,arbTbl[idxSelectedARB].das[j].srvArc,EXPORT_NAME_SIZE))
      return idxSelectedARB;
  }
  /* didn't match up, so try everything else */
  for (i=0; i<arbTblSiz; i++) for (j=0; j<arbTbl[i].nsteps; j++)
  {
    if (ctx != NULL && strnicmp(ctx,arbTbl[i].das[j].ctx,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(srv,arbTbl[i].das[j].srvArc,EXPORT_NAME_SIZE) && isReadable(i,j))
    {
      return i;
    }
  }
  /* wasn't there */
  return -1;
}
void phs2dhs(PHS *phs,DHS *dhs,DEVSPEC *devsp)
{
  float range;
  if (phs == NULL || dhs == NULL || devsp == NULL) return;
  if ((range=(phs->rangeMaxX-phs->rangeMinX)) == 0) range = 1.0;
  memcpy(dhs->channelID,devsp->deviceName,32);
  strncpy(dhs->egu,phs->unitsY,16);
  dhs->arraysize = (UINT16)phs->arraySize;
  dhs->datalength = phs->arraySize;
  dhs->datatype = phs->format;
  dhs->eguhigh = phs->scaleMaxY;
  dhs->egulow = phs->scaleMinY;
  dhs->eventcode = 0;
  dhs->highcount = (UINT32)phs->scaleMaxY;
  dhs->lowcount = (UINT32)phs->scaleMinY;
  dhs->pretrigger = phs->preTrigger;
  if (phs->arraySize > 0) dhs->samplingrate = phs->arraySize/range;
  dhs->timeoffset = 0;
}
void phs2edh(PHS *phs,EVNDHDR *edh,UINT32 evn)
{
  if (phs == NULL || edh == NULL) return;
  edh->evnNr = evn;
  strncpy(edh->unitsY,phs->unitsY,16);
  edh->datSiz = phs->arraySize;
  edh->datFmt= phs->format;
  edh->scaleMaxY = phs->scaleMaxY;
  edh->scaleMinY = phs->scaleMinY;
  edh->rangeMaxX = phs->rangeMaxX;
  edh->rangeMinX = phs->rangeMinX;
  edh->evnCode = 0;
  edh->aryTyp = phs->arrayType;
  edh->preTrg = phs->preTrigger;
  edh->shiftX = 0;
  edh->scaleX = 1;
  edh->shiftY = 0;
  edh->scaleY = 1;
}
void edh2dhs(EVNDHDR *edh,DHS *dhs,char *dev)
{
  float scale = 1, range;
  if (dhs == NULL || edh == NULL) return;
  if ((range=(edh->rangeMaxX-edh->rangeMinX)) == 0) range = 1.0;
  memcpy(dhs->channelID,dev,32);
  strncpy(dhs->egu,edh->unitsY,16);
  dhs->arraysize = (UINT16)edh->datSiz;
  dhs->datalength = edh->datSiz;
  dhs->datatype = edh->datFmt;
  memcpy(dhs->desc,edh->desc,256);
  dhs->eguhigh = edh->scaleMaxY;
  dhs->egulow = edh->scaleMinY;
  dhs->eventcode = edh->evnCode;
  if (edh->scaleY != 0) scale = edh->scaleY;
  dhs->highcount = (UINT32)(edh->scaleMaxY/scale);
  dhs->lowcount = (UINT32)(edh->scaleMinY/scale);
  dhs->pretrigger = edh->preTrg;
  dhs->samplingrate = edh->datSiz/range;
  dhs->timeoffset = (UINT32)edh->shiftX;
}
void dhs2edh(DHS *dhs,EVNDHDR *edh,UINT32 evnNr)
{
  float rate = 1;
  if (dhs == NULL || edh == NULL) return;
  if (dhs->samplingrate != 0) rate = dhs->samplingrate;
  edh->aryTyp = AT_SINGLE|AT_SPECTRUM;
  edh->datFmt = dhs->datatype;
  edh->datSiz = dhs->datalength > 0 ? dhs->datalength : dhs->arraysize;
  memcpy(edh->desc,dhs->desc,256);
  edh->evnCode = dhs->eventcode;
  edh->evnNr = evnNr;
  strncpy(edh->unitsY,dhs->egu,16);
  edh->scaleMinY = dhs->egulow;
  edh->scaleMaxY = dhs->eguhigh;
  edh->preTrg = dhs->pretrigger;
  edh->scaleY = dhs->highcount != 0 ? dhs->eguhigh/dhs->highcount : 1;
  edh->shiftY = 0;
  edh->shiftX = 0;
  edh->scaleX = 1;
  edh->rangeMinX = (float)dhs->timeoffset;
  edh->rangeMaxX = edh->datSiz / rate;
  strcpy(edh->unitsX,"ms");
}
int loadArcHdr(DHS *dhs,EVNDHDR *edh,UINT32 evnNr,char *evnTrg,DEVSPEC *devsp)
{
  int cc=0,id=-1,i=-1,devnr=-1,dsize,dlen=0,fmtsize,n,free_session=FALSE,nPreTrg=0,ignore_device=FALSE;
  BYTE *sptr;
  float fdat;
  short sdat;
  UINT32 ldat;
  double ddat;
  ApcRecord *apc;
  LHS *lhs;
  DHS *dhsPtr;
  EVNDHDR *edhPtr;
  DAS *das;
  PrpQueryStruct pqs, *ppqs=&pqs;
  PHS phs;
  char *c,dev[FULL_NAME_MAXSIZE],lclprop[PROPERTY_NAME_SIZE];

  fmtsize = fmtsizeof(devsp->dataFormat);
  dsize = devsp->dataDimension*fmtsize;
  sptr = srcptr(devsp->dataFormat);
  if (evnNr == 0 || evnTrg == NULL || strlen(evnTrg) == 0) goto tryStatic;
  if (!strcmp(devsp->deviceName,DEVICE_IGNORE)) 
  {
    ignore_device = TRUE;
  }
  else
  {
    if ((c=GetStoredDeviceNameFromAliasEx(evnTrg,devsp->deviceName,devsp->deviceServerName,(int)evnNr)) != NULL)
    {
      strncpy(dev,c,DEVICE_NAME_SIZE);
    }
    else
    {
      devnr = GetStoredDeviceNumberEx(evnTrg,(NAME64 *)devsp->deviceName,devsp->deviceServerName,(int)evnNr);
      if (devnr < 0) devnr = 0;
      sprintf(dev,"#%d",devnr);
    }
  }
  if ((i=GetARBIndex(evnTrg)) < 0  &&
      (i=GetARBIndex(devsp->deviceServerName)) < 0) ccerr(un_allocated);
  if (isIgnoreDeviceForProperty(i,devsp->deviceProperty)) ignore_device = TRUE;
  if ((id=GetLocalSessionID(evnNr,arbTbl[i].ext)) < 0) 
  {
    goto tryStatic;
  }
  setDataTimeStamp((double)session[id].est);
  if (session[id].active)
  { /* try stored header from archive file: */
    session[id].active = time(NULL);  /* keep session alive */
    for (apc=session[id].apc; apc; apc=apc->Next)
    {
      if (strlen(devsp->deviceContext) > 0 &&
          strnicmp(apc->ctx,devsp->deviceContext,CONTEXT_NAME_SIZE)) continue;
      if (strnicmp(apc->srv,devsp->deviceServerName,EXPORT_NAME_SIZE)) continue;
      strncpy(lclprop,devsp->deviceProperty,PROPERTY_NAME_SIZE-4); lclprop[PROPERTY_NAME_SIZE-4] = 0;
      if (strnicmp(lclprop,apc->prp,PROPERTY_NAME_SIZE) == 0) dlen = apc->siz;
      if (strnicmp(apc->prp,"TRIGGEROFFSET",PROPERTY_NAME_SIZE) == 0)
      { /* stop looking, there won't be a header */
        if (apc->fmt != CF_INT32) ccerr(configuration_error);
        if ((dsize=apc->siz) < 1) ccerr(configuration_error);
        if (lseek(session[id].hFile,apc->ptr,SEEK_SET) == -1) ccerr(file_error);
        if (read(session[id].hFile,archiveBuffer,dsize*sizeof(SINT32)) == -1) ccerr(file_error);
        if (dsize == 1)
        { /* good for all */
          nPreTrg = (int)(*(SINT32 *)archiveBuffer);
        }
        else
        {
          if (devnr < 0 || devnr >= (int)apc->siz) ccerr(illegal_device);
          nPreTrg = (int)((SINT32 *)archiveBuffer)[devnr];
        }
        if (dlen == 0)
        { /* target record length still unknown */
          for (apc=apc->Next; apc != NULL; apc=apc->Next)
          {
            if (strlen(devsp->deviceContext) > 0 &&
                strnicmp(apc->ctx,devsp->deviceContext,CONTEXT_NAME_SIZE)) continue;
            if (strnicmp(apc->srv,devsp->deviceServerName,EXPORT_NAME_SIZE)) continue;
            if (strnicmp(apc->prp,devsp->deviceProperty,PROPERTY_NAME_SIZE)) continue;
            dlen = apc->siz;
            break;
          }
        }
        nPreTrg = dlen - nPreTrg; /* num pre-trigger values = total size - trigger offset */
        break;
      }
      strcat(lclprop,".HDR");
      if (strnicmp(lclprop,apc->prp,PROPERTY_NAME_SIZE)) continue;
      if (!ignore_device && strlen(apc->dev))
      {
        if (strnicmp(apc->dev,devsp->deviceName,DEVICE_NAME_SIZE) &&
            strnicmp(apc->dev,dev,DEVICE_NAME_SIZE)) continue;
      }
      if (!strcmp(apc->tag,"DHS") || !strcmp(apc->tag,"HEADER"))
      { /* stored 'old' */
        if ((dsize=sizeof(DHS)) > archiveBufferSize) ccerr(buffer_too_small);
        if (apc->siz < (UINT32)dsize) ccerr(dimension_error);
        if (lseek(session[id].hFile,apc->ptr,SEEK_SET) == -1) ccerr(file_error);
        if (read(session[id].hFile,archiveBuffer,dsize) == -1) ccerr(file_error);
        dhsPtr = (DHS *)archiveBuffer;
        if (dhsPtr->datatype < CF_DOUBLE) dhsPtr->datatype += CF_DOUBLE;
        if (dhs != NULL) memcpy(dhs,dhsPtr,dsize);
        else dhs2edh(dhsPtr,edh,evnNr);
      }
      if (!strcmp(apc->tag,"EVNDHDR"))
      { /* stored 'new' */
        if ((dsize=sizeof(EVNDHDR)) > archiveBufferSize) ccerr(buffer_too_small);
        if (apc->siz < (UINT32)dsize) ccerr(dimension_error);
        if (lseek(session[id].hFile,apc->ptr,SEEK_SET) == -1) ccerr(file_error);
        if (read(session[id].hFile,archiveBuffer,dsize) == -1) ccerr(file_error);
        edhPtr = (EVNDHDR *)archiveBuffer;
        if (edhPtr->datFmt < CF_DOUBLE) edhPtr->datFmt += CF_DOUBLE;
        if (edh != NULL) memcpy(edh,edhPtr,dsize);
        else edh2dhs(edhPtr,dhs,devsp->deviceName);
      }
      return 0; /* we're good ! */
    }
    free_session = TRUE;
  }
tryStatic:
  c = ignore_device ? DEVICE_IGNORE : devsp->deviceName;
  if ((lhs=getLHS(devsp->deviceServerName,devsp->deviceProperty,c)) != NULL) 
  { /* from config file: */
    if (nPreTrg > 0) lhs->hdr.preTrigger = nPreTrg;
    if (dhs != NULL) 
    {
      phs2dhs(&lhs->hdr,dhs,devsp);
      /* copy comment into dhs->desc; */
      readComment(i,evnNr,&dhs->desc[0][0],256);
    }
    else
    {
      phs2edh(&lhs->hdr,edh,evnNr);
    }
    return 0;
  }
  n = 1;
  sprintf(dev,"/%.32s/%.32s/%.64s",devsp->deviceContext,devsp->deviceServerName,devsp->deviceName);
  memset(&pqs,0,sizeof(PrpQueryStruct));
  das = GetARBDas(i, devsp->deviceServerName, devsp->deviceProperty);
  if (strstr(devsp->deviceProperty, ".HIST") != NULL)
  { /* head HIST calls off */
    strcpy(pqs.prpDescription, "trend history");
    strcpy(pqs.prpUnits, "none");
    strcpy(pqs.rngUnits, "sec");
    pqs.prpFormat = BFMT(CF_DBLDBL);
    pqs.prpSize = das == NULL ? 1000 : das->dout.dArrayLength;
    pqs.prpArrayType = AT_SPECTRUM;
    pqs.rngMinValue = 0;
    pqs.rngMaxValue = 0;
    pqs.prpMinValue = 1;
    pqs.prpMaxValue = 10000;
    goto last_chance;
  }
  if (GetSystemPropertyInformation(dev,devsp->deviceProperty,&ppqs,&n) == 0)
  { /* last chance: try a direct query (works if srv running and srv and prp not aliased) */
last_chance:
    strncpy(phs.srv,devsp->deviceServerName,EXPORT_NAME_SIZE);
    strncpy(phs.dev,devsp->deviceName,DEVICE_NAME_SIZE);
    strncpy(phs.prp,devsp->deviceProperty,PROPERTY_NAME_SIZE);
    strncpy(phs.dsc,pqs.prpDescription,PROPERTY_DESC_SIZE);
    strncpy(phs.unitsY,pqs.prpUnits,UNITS_SIZE);
    strncpy(phs.unitsX,pqs.rngUnits,UNITS_SIZE);
    phs.format = das == NULL ? LFMT(pqs.prpFormat) : das->dout.dFormat;
    phs.arraySize = das == NULL ? pqs.prpSize : das->dout.dArrayLength;
    phs.arrayType = pqs.prpArrayType;
    phs.rangeMinX = pqs.rngMinValue;
    phs.rangeMaxX = pqs.rngMaxValue;
    phs.scaleMinY = pqs.prpMinValue;
    phs.scaleMaxY = pqs.prpMaxValue;
    phs.preTrigger = nPreTrg;
    dumpPHS(&phs,1);
    if (dhs != NULL) 
    {
      phs2dhs(&phs,dhs,devsp);
    }
    else
    {
      phs2edh(&phs,edh,evnNr);
    }
    return 0;
  }
  if (strstr(devsp->deviceProperty,".NSEQ") != NULL)
  { /* now, really the last chance : a 'monitor' type ? */
    strcpy(pqs.prpDescription,"number in sequence");
    strcpy(pqs.prpUnits,"none");
    strcpy(pqs.rngUnits,"none");
    pqs.prpFormat = BFMT(CF_INT32);
    pqs.prpSize = 1;
    pqs.prpArrayType = AT_SCALAR;
    pqs.rngMinValue = 0;
    pqs.rngMaxValue = 0;
    pqs.prpMinValue = 1;
    pqs.prpMaxValue = 10000;
    goto last_chance;
  }
  if (das != NULL)
  { /* okay, (we really mean it now) the final, last-ditch effort: */
    strcpy(pqs.prpDescription,"ad-hoc header");
    strcpy(pqs.prpUnits,"none");
    strcpy(pqs.rngUnits,"none");
    pqs.prpFormat = BFMT(das->dout.dFormat);
    pqs.prpSize = das->dout.dArrayLength;
    pqs.prpArrayType = AT_UNKNOWN;
    pqs.rngMinValue = (float)0;
    pqs.rngMaxValue = (float)das->dout.dArrayLength;
    pqs.prpMinValue = -100.0;
    pqs.prpMaxValue =  100.0;
    goto last_chance;
  }
  ccerr(un_allocated);
err:
  if (cc && i >= 0 && arbTbl[i].step != 0) cc = operation_busy;
  if ((cc || free_session) && id >= 0) freeSession(id);
  return cc;
}
int getArchiveHeader(DHS *dhs,UINT32 evnNr,UINT32 start,UINT32 stop,char *evnTrg,DEVSPEC *devsp)
{
  start = start;
  stop = stop;
  return loadArcHdr(dhs,NULL,evnNr,evnTrg,devsp);
}
int GetLocalArchiveHeaderEx(DHS *dhs,UINT32 evnNr,UINT32 start,UINT32 stop,char *evnTrg,DEVSPECX *devspx)
{
  DEVSPEC devsp;
  memset(&devsp,0,sizeof(DEVSPEC));
  devsp.dataDimension = devspx->dataDimension;
  devsp.dataFormat = devspx->dataFormat;
  devsp.dataIndex = devspx->dataIndex;
  strncpy(devsp.deviceContext,devspx->deviceContext,CONTEXT_NAME_SHORTSIZE);
  strncpy(devsp.deviceServerName,devspx->deviceServerName,EXPORT_NAME_SHORTSIZE);
  strncpy(devsp.deviceProperty,devspx->deviceProperty,PROPERTY_NAME_SHORTSIZE);
  strncpy(devsp.deviceName,devspx->deviceName,32);
  return getArchiveHeader(dhs,evnNr,start,stop,evnTrg,&devsp);
}
int GetLocalArchiveHeader(DHS *dhs,UINT32 evnNr,UINT32 start,UINT32 stop,char *evnTrg,DEVSPEC5 *devsp5)
{
  DEVSPEC devsp;
  memset(&devsp,0,sizeof(DEVSPEC));
  devsp.dataDimension = devsp5->dataDimension;
  devsp.dataFormat = devsp5->dataFormat;
  devsp.dataIndex = 0;
  strncpy(devsp.deviceContext,gDeviceContext,CONTEXT_NAME_SIZE);
  strncpy(devsp.deviceServerName,devsp5->deviceServerName,EXPORT_NAME_SHORTSIZE);
  strncpy(devsp.deviceProperty,devsp5->deviceProperty,16);
  strncpy(devsp.deviceName,devsp5->deviceName,16);
  return getArchiveHeader(dhs,evnNr,start,stop,evnTrg,&devsp);
}
int getArcHdr(DTYPE *dout,UINT32 evnNr,char *evnTrg,DEVSPEC *devsp)
{
  if (!strcmp(dout->dTag,"DHS") || !strcmp(dout->dTag,"HEADER"))
  {
    return loadArcHdr((DHS *)dout->data.vptr,NULL,evnNr,evnTrg,devsp);
  }
  if (!strcmp(dout->dTag,"EVNDHDR"))
  {
    return loadArcHdr(NULL,(EVNDHDR *)dout->data.vptr,evnNr,evnTrg,devsp);
  }
  return invalid_structure_tag;
}
int closeArchiveFile(UINT32 evnNr,char *evnTrg)
{
  int i,id;
  if ((i=GetARBIndex(evnTrg)) < 0) return un_allocated;
  if ((id=GetLocalSessionID(evnNr,arbTbl[i].ext)) < 0) return -id;
  session[id].count = 0;
  freeSession(id);
  return 0;
}
int getArchiveProperties(int evnNr,char *evnTrg,char *ctx,char *srv,char *dev,NAME64 *prps,int *nprps)
{
  int i, k, id, cc = 0, n =0, mx;
  ApcRecord *apc;
  if ((i=GetARBIndex(evnTrg)) < 0) ccerr(un_allocated);
  if ((id=GetLocalSessionID((UINT32)evnNr,arbTbl[i].ext)) < 0) ccerr(-id);
  if (!session[id].active) ccerr(file_not_open);
  session[id].active = time(NULL);  /* keep session alive */
  mx = *nprps;
  for (apc=session[id].apc; apc; apc=apc->Next)
  {
    if (n >= mx) break;
    if (ctx != NULL && strlen(ctx) > 0 && strnicmp(apc->ctx,ctx,CONTEXT_NAME_SIZE)) continue;
    if (dev != NULL && strlen(dev) > 0 && strnicmp(apc->dev,dev,DEVICE_NAME_SIZE)) continue;
    if (strnicmp(apc->srv,srv,EXPORT_NAME_SIZE)) continue;
    for (k=0; k<n; k++) if (!strnicmp(apc->prp,prps[k].name,PROPERTY_NAME_SIZE)) break;
    if (k < n) continue;
    strncpy(prps[n].name,apc->prp,PROPERTY_NAME_SIZE);
    n++;
  }
  if (n == 0)
  {
    n = GetStoredProperties(prps, srv, evnTrg, mx);
  }
  *nprps = n;
err:
  return cc;
}
int getArchiveDevices(int evnNr,char *evnTrg,char *ctx,char *srv,char *prp,NAME64 *devs,int *ndevs)
{
  int i, k, id, cc = 0, n = 0, mx;
  ApcRecord *apc;
  if ((i=GetARBIndex(evnTrg)) < 0) ccerr(un_allocated);
  if (evnNr <= 0) goto get_stock_devices;
  if ((id=GetLocalSessionID((UINT32)evnNr,arbTbl[i].ext)) < 0) ccerr(-id);
  if (!session[id].active) ccerr(file_not_open);
  session[id].active = time(NULL);  /* keep session alive */
  mx = *ndevs;
  if (ctx == NULL) ctx = gDeviceContext;
  for (apc=session[id].apc; apc; apc=apc->Next)
  {
    if (n >= mx) break;
    if (strlen(ctx) > 0 && strnicmp(apc->ctx,ctx,CONTEXT_NAME_SIZE)) continue;
    if (prp != NULL && strlen(prp) > 0 && strnicmp(apc->prp,prp,PROPERTY_NAME_SIZE)) continue;
    if (strnicmp(apc->srv,srv,EXPORT_NAME_SIZE)) continue;
    for (k=0; k<n; k++) if (!strnicmp(apc->dev,devs[k].name,DEVICE_NAME_SIZE)) break;
    if (k < n) continue;
    if (apc->dev[0] == '#')
    { /* numbers stored instead of names: force a names query */
      n = 0;
      break;
    }
    strncpy(devs[n].name,apc->dev,DEVICE_NAME_SIZE);
    n++;
  }
get_stock_devices:
  if (n == 0)
  {
    n = GetStoredDevices(devs, srv, evnTrg, evnNr, mx);
  }
  *ndevs = n;
err:
  return cc;
}
int getArchiveServers(int evnNr,char *evnTrg,char *ctx,NAME32 *srvs,int *nsrvs)
{
  int i, k, id, cc = 0, n =0, mx;
  ApcRecord *apc;
  if ((i=GetARBIndex(evnTrg)) < 0) ccerr(un_allocated);
  if ((id=GetLocalSessionID((UINT32)evnNr,arbTbl[i].ext)) < 0) ccerr(-id);
  if (!session[id].active) ccerr(file_not_open);
  session[id].active = time(NULL);  /* keep session alive */
  mx = *nsrvs;
  for (apc=session[id].apc; apc; apc=apc->Next)
  {
    if (n >= mx) break;
    if (ctx != NULL && strlen(ctx) > 0 && strnicmp(apc->ctx,ctx,CONTEXT_NAME_SIZE)) continue;
    for (k=0; k<n; k++) if (!strnicmp(apc->srv,srvs[k].name,EXPORT_NAME_SIZE)) break;
    if (k < n) continue;
    strncpy(srvs[n].name,apc->srv,EXPORT_NAME_SIZE);
    n++;
  }
  *nsrvs = n;
err:
  return cc;
}
int getArchiveData(DTYPE *d,UINT32 evnNr,UINT32 start,UINT32 stop,char *evnTrg,DEVSPEC *devsp)
{
  int cc=0,id=-1,i=-1,dsize,ssize=1,dindex,fmtsize,nrecs=0,ddim,ignore_device=FALSE;
  BYTE *sptr;
  float fdat;
  short sdat;
  UINT32 ldat;
  double ddat;
  ApcRecord *apc;
  BYTE *data = d->data.bptr;
  char *c,tmpdev[DEVICE_NAME_SIZE+8];
  int devnr = -1;
  float scale = 1, shift = 0;
# ifdef SCALE_FROM_HEADER
  DHS dhs;
  if (getArchiveHeader(&dhs,evnNr,start,stop,evnTrg,devsp) == 0)
  {
    if (dhs.eguhigh && dhs.highcount) scale = dhs.eguhigh;
  }
# endif

  fmtsize = fmtsizeof(devsp->dataFormat);
  dsize = devsp->dataDimension*fmtsize;
  ddim = devsp->dataDimension;
  if (devsp->dataFormat == CF_STRUCT)
  {
    if ((ssize=GetStructSize(devsp->dataTag)) > 0) ddim *= ssize;
  }
  dindex = devsp->dataIndex;
  sptr = srcptr(devsp->dataFormat);
  if ((c=GetStoredDeviceNameFromAliasEx(evnTrg,devsp->deviceName,devsp->deviceServerName,(int)evnNr)) != NULL)
  {
    strncpy(tmpdev,c,DEVICE_NAME_SIZE);
  }
  else
  {
    devnr = GetStoredDeviceNumberEx(evnTrg,(NAME64 *)devsp->deviceName,devsp->deviceServerName,(int)evnNr);
    if (devnr < 0) devnr = 0;
    sprintf(tmpdev,"#%d",devnr);
  }
  if ((i=GetARBIndex(evnTrg)) < 0  &&
      (i=GetARBIndex(devsp->deviceServerName)) < 0) ccerr(un_allocated);
  if ((id=GetLocalSessionID(evnNr,arbTbl[i].ext)) < 0) ccerr(-id);
  if (isIgnoreDeviceForProperty(i,devsp->deviceProperty)) ignore_device = TRUE;
  setDataTimeStamp((double)session[id].est);
  if (!session[id].active) ccerr(file_not_open);
  session[id].active = time(NULL);  /* keep session alive */
  if (!strcmp(devsp->deviceName,DEVICE_IGNORE)) ignore_device = TRUE;
scanStoredCalls:
  for (apc=session[id].apc; apc; apc=apc->Next)
  {
    if (strlen(devsp->deviceContext) > 0 &&
        strnicmp(apc->ctx,devsp->deviceContext,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(apc->srv,devsp->deviceServerName,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(apc->prp,devsp->deviceProperty,PROPERTY_NAME_SIZE)) continue;
    if (!ignore_device && strlen(apc->dev))
    {
      if (strnicmp(apc->dev,devsp->deviceName,DEVICE_NAME_SIZE) &&
          strnicmp(apc->dev,tmpdev,DEVICE_NAME_SIZE)) continue;
    }
    dsize = fmtsizeof(apc->fmt)*apc->siz + getFormatHeaderSize(apc->fmt);
    /* if (apc->fmt == CF_STRUCT && (ssize=GetStructSize(apc->tag)) > 0) dsize *= ssize; */
    if (dsize > archiveBufferSize) ccerr(buffer_too_small);
    /*if (apc->siz < devsp->dataIndex+devsp->dataDimension) ccerr(dimension_error);*/
    nrecs = MIN((int)apc->siz,ddim);
    if (lseek(session[id].hFile,apc->ptr,SEEK_SET) == -1) ccerr(file_error);
    if (read(session[id].hFile,archiveBuffer,dsize) == -1) ccerr(file_error);
    if (scale == 1 && apc->scale != 1) scale = apc->scale;
    if (shift == 0 && apc->shift != 0) shift = apc->shift;
    if (strstr(devsp->deviceProperty,"RAW")) scale = shift = 0;
    cc = reformatArchiveDataEx((void *)data,devsp->dataFormat,
                              archiveBuffer,apc->fmt,
                              devsp->dataIndex,
                              nrecs,scale,shift);
    if (cc != 0) ccerr(cc);
    if (devsp->dataFormat == CF_STRUCT && ssize > 0) nrecs /= ssize;
    d->dTimeStamp = apc->timestamp;
    d->sysStamp = apc->sysstamp;
    d->dStamp = apc->usrstamp;
    return nrecs;
  }
  if (strstr(devsp->deviceProperty,".NSEQ") != NULL && arbTbl[i].isMonitoring)
  { /* now, really the last chance : a scalar treated as a 'monitor' type ? */
    int ns = 1;
    nrecs = 1;
    cc = reformatArchiveDataEx((void *)&ns,devsp->dataFormat,archiveBuffer,CF_INT32,0,1,1,0);
    if (cc != 0) ccerr(cc);
    d->dTimeStamp = (double)evnNr;
    d->sysStamp = 0;
    d->dStamp = 0;
    return 1;
  }
  if ((c=strstr(devsp->deviceProperty,".1")) != NULL)
  {
    *c = 0;
    goto scanStoredCalls;
  }
  ccerr(un_allocated);
err:
  if (cc && i >= 0 && arbTbl[i].step != 0) cc = operation_busy;
  if (cc && id >= 0) freeSession(id);
  return cc ? -cc : nrecs;
}
int GetLocalArchiveData(DTYPE *d,UINT32 evnNr,UINT32 start,UINT32 stop,char *evnTrg,DEVSPEC5 *devsp5)
{
  DEVSPEC devsp;
  memset(&devsp,0,sizeof(DEVSPEC));
  devsp.dataDimension = devsp5->dataDimension;
  devsp.dataFormat = devsp5->dataFormat;
  devsp.dataIndex = 0;
  strncpy(devsp.deviceContext,gDeviceContext,CONTEXT_NAME_SIZE);
  strncpy(devsp.deviceServerName,devsp5->deviceServerName,EXPORT_NAME_SHORTSIZE);
  strncpy(devsp.deviceProperty,devsp5->deviceProperty,16);
  strncpy(devsp.deviceName,devsp5->deviceName,16);
  d->dFormat = devsp.dataFormat;
  d->dArrayLength = devsp.dataDimension;
  return getArchiveData(d,evnNr,start,stop,evnTrg,&devsp);
}

int GetLocalArchiveDataEx(DTYPE *d,UINT32 evnNr,UINT32 start,UINT32 stop,char *evnTrg,DEVSPECX *devspx)
{
  DEVSPEC devsp;
  memset(&devsp,0,sizeof(DEVSPEC));
  devsp.dataDimension = devspx->dataDimension;
  devsp.dataFormat = devspx->dataFormat;
  devsp.dataIndex = devspx->dataIndex;
  strncpy(devsp.deviceContext,devspx->deviceContext,CONTEXT_NAME_SHORTSIZE);
  strncpy(devsp.deviceServerName,devspx->deviceServerName,EXPORT_NAME_SHORTSIZE);
  strncpy(devsp.deviceProperty,devspx->deviceProperty,PROPERTY_NAME_SHORTSIZE);
  strncpy(devsp.deviceName,devspx->deviceName,32);
  d->dFormat = devsp.dataFormat;
  d->dArrayLength = devsp.dataDimension;
  return getArchiveData(d,evnNr,start,stop,evnTrg,&devsp);
}

int AnnotateLocalArchiveEvent(UINT32 evntime,char *devName,char *cmt)
{
  int setNr;
 
  setNr = getSetNumber(devName);
  if (setNr < 0 || setNr >= arbTblSiz) return invalid_index;
  if (evntime == (UINT32)arbTbl[setNr].tot)
  {
    if (arbTbl[setNr].eventtime == 0) return operation_busy;
    return writeComment(setNr,(int)arbTbl[setNr].eventtime,cmt);
  }
  return writeComment(setNr,evntime,cmt);
}
int RemoveLocalArchiveEvents(UINT32 evnstart,UINT32 evnstop,char *dsrv)
{
  int cc,j,k=0,p = 0,found = 0;

  /* try the selected index  */
  for (j=0; j<arbTbl[idxSelectedARB].nsteps; j++)
  {
    if (!strnicmp(dsrv,arbTbl[idxSelectedARB].das[j].srvArc,EXPORT_NAME_SIZE))
    {
      found = -1;
      k = idxSelectedARB+1;
      break;
    }
  }
  /* if it didn't match up, try everything else */
  for (; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    if (!strnicmp(dsrv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
  if (!found) return no_such_file;

  /* this will have to match the first time or bail out ! */
  if ((cc=RemoveEvents(evnstart, evnstop, arbTbl[k-1].ext)) < 0) return -cc;
  return 0;
}
int RemoveLocalArchiveEvent(UINT32 evntime,char *dsrv)
{
  int cc,j,k=0,p = 0,found = 0;

  /* try the selected index  */
  for (j=0; j<arbTbl[idxSelectedARB].nsteps; j++)
  {
    if (!strnicmp(dsrv,arbTbl[idxSelectedARB].das[j].srvArc,EXPORT_NAME_SIZE))
    {
      found = -1;
      k = idxSelectedARB+1;
      break;
    }
  }
  /* if it didn't match up, try everything else */
scan_arbs:
  for (; k<arbTblSiz  && !found; k++) for (j=0; j<arbTbl[k].nsteps && !found; j++)
    if (!strnicmp(dsrv,arbTbl[k].das[j].srvArc,EXPORT_NAME_SIZE)) found = -1;
  if (!found) return no_such_file;

  if ((cc=RemoveEvent(evntime, arbTbl[k-1].ext)) != 0) 
  {
    found = 0;
    goto scan_arbs;
  }
  return cc;
}
int GetLocalnumDevices(char *devSrvName,char *trigger,UINT32 evnTime)
{
  int n;
  if (evnTime <= 0) evnTime = (UINT32)time(NULL);
  n = GetStoredDevices(NULL,devSrvName,trigger,evnTime,0);
  return n > 0 ? n : 0;
}
int GetLocalnumDeviceProperties(char *devSrvName,char *devName)
{
  int n = GetStoredProperties(NULL,devSrvName,devName,0);
  return n > 0 ? n : 0;
}
NAME32 *dsNames = NULL;
int numDsNames = 0;
int GetLocalnumDeviceServers(void)
{
  int num,i,j,k,m;
  static int n = 0;

  if (n == 0)
  {
    for (i=0; i<arbTblSiz; i++) for (j=0; j<arbTbl[i].nsteps; j++) if (isReadable(i,j))
    {
      for (m=0; m<=i; m++)
      {
        if (arbTbl[m].das == NULL) continue;
        for (k=0; k<j; k++) 
        {
          if (isReadable(m,k) && 
              !strnicmp(arbTbl[i].das[j].srvArc,arbTbl[m].das[k].srvArc,EXPORT_NAME_SIZE)) break;
        }
        if (k < j) break;
      }
      if (k == j && m > i) n++;
    }
    /* worst case n, now allocate and fill in. */
    if (dsNames != NULL) free(dsNames);
    if ((dsNames=(NAME32 *)calloc(n,sizeof(NAME32))) != NULL)
    {
      numDsNames = n;
      num = n; n = 0;
      for (i=0; i<arbTblSiz; i++) for (j=0; j<arbTbl[i].nsteps; j++) if (isReadable(i,j))
      {
        for (k=0; k<n; k++) if (!strnicmp(arbTbl[i].das[j].srvArc,dsNames[k].name,EXPORT_NAME_SIZE)) break;
        if (k == n)
        {
          if (n >= num) return 0;
          strncpy(dsNames[n++].name,arbTbl[i].das[j].srvArc,EXPORT_NAME_SIZE);
        }
      }      
    }
  }
  return n;
}
int GetLocalDeviceServers(char *trigger,NAME32 *names,UINT32 start,UINT32 stop,int event,int num)
{
  int i,j,k;
  int n = 0;  
 
  event = event;
  start = start;
  stop = stop;
  i = getSetNumber(trigger);
  if (i < 0 || i >= arbTblSiz) 
  { /* return them all */
    if (dsNames != NULL) 
    {
      if (num > numDsNames) n = numDsNames; else n = num;
      memcpy(names,dsNames,n*32);
      return n;
    }
    for (i=0; i<arbTblSiz; i++) for (j=0; j<arbTbl[i].nsteps; j++) if (isReadable(i,j))
    {
      if (strlen(arbTbl[i].das[j].srvArc) == 0) continue;
      for (k=0; k<n; k++) if (!strnicmp(arbTbl[i].das[j].srvArc,names[k].name,EXPORT_NAME_SIZE)) break;
      if (k == n)
      {
        if (n >= num) return n;
        strncpy(names[n++].name,arbTbl[i].das[j].srvArc,EXPORT_NAME_SIZE);
      }
    }
  }
  else
  { /* targeted trigger context */
    for (j=0; j<arbTbl[i].nsteps; j++) if (isReadable(i,j))
    {
      if (strlen(arbTbl[i].das[j].srvArc) == 0) continue;
      for (k=0; k<n; k++) if (!strnicmp(arbTbl[i].das[j].srvArc,names[k].name,EXPORT_NAME_SIZE)) break;
      if (k == n)
      {
        if (n >= num) return n;
        strncpy(names[n++].name,arbTbl[i].das[j].srvArc,EXPORT_NAME_SIZE);
      }
    }
  }
  return n;
}
int GetLocalDeviceServersEx(char *trigger,NAME64 *names,UINT32 start,UINT32 stop,int event,int num)
{
  int i,j,k;
  int n = 0;  
  char *ctx;
  char interim[64];

  event = event;
  start = start;
  stop = stop;
  i = getSetNumber(trigger);
  if (i < 0 || i >= arbTblSiz) 
  { /* return them all */
    for (i=0; i<arbTblSiz; i++) for (j=0; j<arbTbl[i].nsteps; j++) if (isReadable(i,j))
    {
      if (strlen(arbTbl[i].das[j].srvArc) == 0) continue;
      ctx = arbTbl[i].das[j].ctx;
      if (strlen(ctx) == 0) ctx = gDeviceContext;
      sprintf(interim,"/%.30s/%.32s",ctx,arbTbl[i].das[j].srvArc);
      for (k=0; k<n; k++) if (!strnicmp(interim,names[k].name,64)) break;
      if (k == n)
      {
        if (n >= num) return n;
        strncpy(names[n++].name,interim,64);
      }
    }
  }
  else
  { /* targeted trigger context */
    for (j=0; j<arbTbl[i].nsteps; j++) if (isReadable(i,j))
    {
      if (strlen(arbTbl[i].das[j].srvArc) == 0) continue;
      ctx = arbTbl[i].das[j].ctx;
      if (strlen(ctx) == 0) ctx = gDeviceContext;
      sprintf(interim,"/%.30s/%.32s",ctx,arbTbl[i].das[j].srvArc);
      for (k=0; k<n; k++) if (!strnicmp(interim,names[k].name,64)) break;
      if (k == n)
      {
        if (n >= num) return n;
        strncpy(names[n++].name,interim,64);
      }
    }
  }
  return n;
}
int GetLocalTriggers(NAME64 *names,char *srv,char *prp,int num)
{
  return GetStoredTriggers(names,srv,prp,num);
}
int GetLocalArchiveInfo(AIB *aib,UINT32 eventNumber,char *deviceServer)
{
  return 0;
}
int GetLocalPropertyInfo(char *fullName,UINT32 eventNumber,PrpQueryStruct *pqs)
{
  char devstr[STATUS_SIZE],prp[PROPERTY_NAME_SIZE];
  char *c, *ctx, *srv, *dev;
  int cc, i, j, plen, isseq=FALSE;

  if ((cc=parseFullDeviceName(fullName,devstr,prp,FALSE)) != 0) return cc;
  if (strlen(prp) == 0) return argument_list_error;
  ctx = (devstr[0] == '/') ? &devstr[1] : devstr;
  if ((srv=strchr(ctx,'/')) == NULL) return argument_list_error;
  *srv++ = 0;
  if ((dev=strchr(srv,'/')) != NULL) *dev++ = 0;
  if ((c=strrchr(prp,'.')) != NULL)
  {
    c++;
    if ((i=atoi(c)) > 0) isseq = TRUE;
    else if (strcmp(c,"NSEQ") == 0) isseq = TRUE;
  }

  for (i=0; i<arbTblSiz; i++) for (j=0; j<arbTbl[i].nsteps; j++) if (isReadable(i,j))
  {
    if (strnicmp(arbTbl[i].das[j].ctx,ctx,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(arbTbl[i].das[j].srvArc,srv,EXPORT_NAME_SIZE)) continue;
    plen = PROPERTY_NAME_SIZE;
    if (isseq && arbTbl[i].das[j].acc & MONITOR) plen = (int)MIN(strlen(arbTbl[i].das[j].prpArc),PROPERTY_NAME_SIZE);
    if (strnicmp(arbTbl[i].das[j].prpArc,prp,plen)) continue;
    strncpy(pqs->prpName,prp,PROPERTY_NAME_SIZE);
    memset(pqs->prpDescription,0,PROPERTY_DESC_SIZE);
    memset(pqs->prpRedirection,0,PROPERTY_REDIR_SIZE);
    strncpy(pqs->prpTag,arbTbl[i].das[j].dout.dTag,TAG_NAME_SIZE);
    strncpy(pqs->prpTagIn,arbTbl[i].das[j].din.dTag,TAG_NAME_SIZE);
    memset(pqs->prpUnits,0,UNITS_SIZE);
    pqs->prpMinValue = 0;
    pqs->prpMaxValue = 0;
    pqs->prpSize = arbTbl[i].das[j].dout.dArrayLength;
    pqs->prpSizeIn = arbTbl[i].das[j].din.dArrayLength;
    pqs->prpNumOverloads = 0;
    pqs->prpHistoryDepthShort = 0;
    pqs->prpHistoryDepthLong = 0;
    pqs->prpFormat = BFMT(arbTbl[i].das[j].dout.dFormat);
    pqs->prpFormatIn = BFMT(arbTbl[i].das[j].din.dFormat);
    pqs->prpAccess = CA_READ;
    pqs->prpGraphType = 0;
    memset(pqs->rngUnits,0,UNITS_SIZE);
    pqs->rngMinValue = 0;
    pqs->rngMaxValue = 0;
    pqs->numRows = 1;
    pqs->rowSize = arbTbl[i].das[j].dout.dArrayLength;;
    pqs->prpArrayType = AT_SPECTRUM;
    return 0;
  }
  return non_existent_property;
}
int GetNumLocalDeviceArchivedEvents(char *devServer,char *devName,UINT32 starttime,UINT32 stoptime)
{
  int i;

  /* get das from devServer : */
  if ((i=GetARBIndex(devServer)) < 0) return -non_existent_elem;
  /* find EXT from devServer */
  return getFiles(NULL,arbTbl[i].ext,starttime,stoptime, gMaxNumEvents);
}
int GetNumLocalArchivedEvents(char *trigger,UINT32 starttime,UINT32 stoptime,UINT32 eventNumber)
{
  int i;

  /* get das from devServer : */
  if ((i=GetARBIndex(trigger)) < 0) return -non_existent_elem;
  return getFiles(NULL,arbTbl[i].ext,starttime,stoptime, gMaxNumEvents);
}
NAME8 *evnbuf = NULL;
int GetLocalDeviceArchivedEvents(char *trigger,char *srv,char *dev,UINT32 starttime,UINT32 stoptime,AES *aes,UINT32 num,int reverseSort,int *sts)
{
  int cc=0,i,k,n,p = 0,found = 0;
  char s[16];

  if (trigger != NULL && strlen(trigger) > 0 && stricmp(trigger,"ANY") != 0)
  { /* get das from event context : */
    if ((k=GetARBIndex(trigger)) < 0) return -non_existent;
  }
  else
  { /* get das from devServer : */
    if ((k=GetARBIndex(srv)) < 0) return -non_existent_elem;
  }
  if (num > (UINT32)gMaxNumEvents)
  {
    NAME8 *tmp = (NAME8 *)realloc(evnbuf, (num + 1000) * sizeof(NAME8));
    if (tmp != NULL)
    {
      evnbuf = tmp;
      gMaxNumEvents = (int)(num + 1000);
    }
    else
    {
      num = (UINT32)gMaxNumEvents;
    }
  }
  if ((n = getFiles(evnbuf, arbTbl[k].ext, starttime, stoptime, num)) < 0)
  {
    *sts = -n;
    return n;
  }
  if (n > (int)num)
  {
    n = num; //return -buffer_too_small;
    *sts = buffer_too_small;
  }

  /* fill in aes with atoi(filebuf[]), local comment files, event mask */
  /* reverse order: put the newest first */
  for (i=0; i<n; i++) 
  {
    strncpy(s,evnbuf[i].name,8); s[8] = 0;
    if (reverseSort)
      aes[i].eventNumber = strtol(s,0,16);
    else
      aes[n-i-1].eventNumber = strtol(s,0,16);
  }
  return n;
}
int GetLocalDeviceArchivedEventList(char *trigger,char *srv,char *dev,UINT32 starttime,UINT32 stoptime,UINT32 *evnlst,UINT32 num,int reverseSort,int *sts)
{
  int i,k,n,p = 0,found = 0;
  char s[16];

  *sts = 0;
  if (trigger != NULL && strlen(trigger) > 0 && stricmp(trigger,"ANY") != 0)
  { /* get das from event context : */
    if ((k=GetARBIndex(trigger)) < 0)
    {
      *sts = non_existent;
      return -non_existent;
    }
  }
  else
  { /* get das from devServer : */
    if ((k=GetARBIndex(srv)) < 0)
    {
      *sts = non_existent_elem;
      return -non_existent_elem;
    }
  }
   /* ignore dev */
  if (num > (UINT32)gMaxNumEvents)
  {
    NAME8 *tmp = (NAME8 *)realloc(evnbuf, (num + 1000) * sizeof(NAME8));
    if (tmp != NULL)
    {
      evnbuf = tmp;
      gMaxNumEvents = (int)(num + 1000);
    }
    else
    {
      num = (UINT32)gMaxNumEvents;
    }
  }
  if ((n=getFiles(evnbuf,arbTbl[k].ext,starttime,stoptime,num)) < 0) return n;
  if (n > (int)num)
  {
    n = num; //return -buffer_too_small;
    *sts = buffer_too_small;
  }

  /* reverse order: put the newest first */
  for (i=0; i<n; i++) 
  {
    strncpy(s,evnbuf[i].name,8); s[8] = 0;
    if (reverseSort)
      evnlst[i] = strtol(s,0,16);
    else
      evnlst[n-i-1] = strtol(s,0,16);
  }

  return n;
}
int GetLocalArchivedEvents(char *trigger,UINT32 starttime,UINT32 stoptime,UINT32 *evnlst,UINT32 num,int *sts)
{ 
  int i,idx,n,p;
  char s[16];
  UINT32 evnt;

  /* get das from archive context : */
  if ((idx = GetARBIndex(trigger)) < 0)
  {
    *sts = non_existent;
    return -non_existent;
  }
  *sts = 0;
  /* use local filebuf array to hold NAME8 event numbers */
  if (num > (UINT32)gMaxNumEvents)
  {
    NAME8 *tmp = (NAME8 *)realloc(evnbuf, (num + 1000) * sizeof(NAME8));
    if (tmp != NULL)
    {
      evnbuf = tmp;
      gMaxNumEvents = (int)(num + 1000);
    }
    else
    {
      num = (UINT32)gMaxNumEvents;
    }
  }
  if ((n=getFiles(evnbuf,arbTbl[idx].ext,starttime,stoptime, gMaxNumEvents)) < 0) return n;
  if (n > gMaxNumEvents)
  {
    n = gMaxNumEvents;
    *sts = buffer_too_small;
  }
  /* reverse order: put the newest first */
  for (i=n-1,p=0; i >= 0 && p<(int)num; i--)
  {
    strncpy(s, evnbuf[i].name, 8); s[8] = 0;
    evnt = (UINT32)strtol(s, 0, 16);
    if (evnt == 0) continue;
    evnlst[p] = evnt;
    p++;
  }
  return p;
}
int GetLocalDevices(NAME64 *names,char *devSrvName,char *trigger,UINT32 id,int num)
{ /* react to id = event number */
  if (id <= 0) id = (int)time(NULL);
  return GetStoredDevices(names,devSrvName,trigger,id,num); 
}
int GetLocalDeviceProperties(NAME64 *props,char *devSrvName,char *trigger,int num)
{
  /* react to id = event number */
  return GetStoredProperties(props,devSrvName,trigger,num); 
}
char thisFecName[FEC_NAME_SIZE+4] = "";
char thisContext[CONTEXT_NAME_SIZE+4] = "";
char *thisSubSystem = "SER";
char thisFecDescription[FEC_DESC_SIZE] = "";
char thisLocation[FEC_LOCATION_SIZE] = "";
char thisServer[EXPORT_NAME_SIZE] = "";
char thisExtension[4] = "";
UINT16 thisPortOffset = 0;
int startupInputsGiven = FALSE;
void PreSystemInit(void)
{
  gRetardSingleContractRemoval = FALSE;
  gUseLoopback = TRUE;
  SystemPresetMemory = TRUE;
  fecLogFileDepth = 10000; // the 'FECNAME'.log file
  SetFeclogDepth(10000);   // all other systematic log files 
  archiveBufferSize = 16*1024*1024;
  SystemAssignBufferSpace(0x40000,0x8000);
  SetWorkAreaSize(archiveBufferSize);
  SetContractListCapacity(5000);
  SetPacketMTU(64000);
  SetAllowRemoteManagement(TRUE);

  if (startupInputsGiven)
  {
    int p;
    if (strlen(thisContext) == 0)
    {
      strcpy(thisContext,"TEST");
    }
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,10); thisFecName[10] = 0;
      strcat(thisFecName,"EVNFEC");
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    sprintf(thisFecDescription,"%.32s Event archive server",thisContext);
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }
  else
  {
    strncpy(thisContext,gDeviceContext,CONTEXT_NAME_SIZE);
    strncpy(thisServer,"EVENTSTORE",CONTEXT_NAME_SIZE);
  }
  return;
}


typedef struct 
{
  char trg[64];
  char rdr[192];
} TrgRedir;

TrgRedir *trgRedir = NULL;
int gNumTrgRedirs = 0;

CSVCOLUMN csvTrgRedirColumns[] =
{
  { "TRIGGER","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgRedir,trg),NULL,0 },
  { "REDIRECTION","",0,CF_TEXT,192,(BYTE *)OFFSETIN(TrgRedir,rdr),NULL,0 },
};
CSVDB csvTrgRedirDb =
{
  csvTrgRedirColumns,
  sizeof(csvTrgRedirColumns)/sizeof(csvTrgRedirColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TrgRedir),
  0
};
#define REDIR_FILE "redirections.csv"
int readTriggerRedirections(void)
{
  int cc;
  if ((cc=csvReadFile("",REDIR_FILE,&csvTrgRedirDb,(void **)&trgRedir)) != 0) return cc;
  gNumTrgRedirs = csvTrgRedirDb.siz;
  return 0;
}
int isSelfRedirection(char *dev)
{
  int i;
  char *c, *ctx, *srv, rdr[256] = {0};
  for (i=0; i<gNumTrgRedirs; i++) 
  { // anything to redirect ?
    if (strnicmp(dev,trgRedir[i].trg,DEVICE_NAME_SIZE)) continue;
    strncpy(rdr,trgRedir[i].rdr,192);
    if ((c=strchr(rdr,'/')) == NULL) return FALSE; // just context and it is logically different from my own
    ctx = &c[1];
    if ((c=strchr(ctx,'/')) == NULL) return FALSE; // no server given?
    *c++ = 0; srv = c;
    if ((c=strchr(srv,'/')) != NULL) *c = 0;
    if (strnicmp(ctx,thisContext,CONTEXT_NAME_SIZE)) return FALSE;
    if (strnicmp(srv,thisServer,EXPORT_NAME_SIZE)) return FALSE;
    return TRUE;
  }
  return FALSE;
}
void PostSystemInit()
{
  int cc, i, n, regdev=FALSE, isprimary=FALSE;
  DTYPE dout, din;
  char *c, rdr[256];
  
  strncpy(thisServer, STREQM_EXPORT, 28);
  if (strlen(thisExtension) > 0)
  {
    if (thisExtension[0] != '.') strcat(thisServer, ".");
    strncat(thisServer, thisExtension, 4);
  }
  registerStructs();
  RegisterEquipmentModule(thisServer, STREQM_TAG, 100, streqm, initArchiveTask, bkgArchiveTask, 100, streqm_exit);
  if ((cc=readArchiveSets(STREQM_TAG)) != 0) dbglog("initialization error: %s",erlst[cc]);
  initArchiveServer();
  if (evnbuf == NULL)
  {
    evnbuf = (NAME8 *)SystemCalloc(gMaxNumEvents, sizeof(NAME8));
    if (evnbuf == NULL)
    {
      dbglog("not enough memory to allocate event buffer to gather %d events",gMaxNumEvents);
      feclogEx(TINE_LOGLEVEL_FATAL, "could not allocate event buffer!");
      exit(1);
    }
  }
  RegisterProperty(STREQM_TAG,"TRIGGER",1,CF_INT32,CA_READ|CA_WRITE,"set: trigger event (initiate event archive); get: status");
  RegisterProperty(STREQM_TAG,"TRIGGER.STATUS",1,CF_LONG,CA_READ,"status of most recent trigger event (current process step)");
  RegisterProperty(STREQM_TAG,"TRIGGER.TIME",1,CF_LONG,CA_READ,"time of most recent trigger event");
  RegisterProperty(STREQM_TAG,"ANNOTATE",1024,CF_TEXT,CA_READ|CA_WRITE,"write comment to selected event");
  RegisterProperty(STREQM_TAG,"COMMENT",1024,CF_TEXT,CA_READ,"read comment from selected event");
  RegisterProperty(STREQM_TAG,"REMOVE",1,CF_LONG,CA_WRITE,"remove archive for selected event");
  RegisterProperty(STREQM_TAG,"SELECT",16,CF_TEXT,CA_READ,"set/get selected event context");
  RegisterProperty(STREQM_TAG,"STOP",0,CF_NULL,CA_WRITE,"abort current (on-going) archive process");
  RegisterProperty(STREQM_TAG, "TRIGGER.REFRESH_DEVICES", 0, CF_NULL, CA_WRITE, "refresh the device lists for servers associated with trigger");

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 100;
  strcpy(din.dTag,"ARBQS");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,"ARBQS");
  RegisterPropertyInformation(STREQM_TAG,"TRIGGER.TBL",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"event trigger table",-1,NULL);

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 100;
  strcpy(din.dTag,"DASQS");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,"DASQS");
  RegisterPropertyInformation(STREQM_TAG,"TRIGGER.DB",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"event trigger specific database",-1,NULL);

  din.dFormat = CF_NAME64;
  din.dArrayLength = 2;
  din.dTag[0] = 0;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"DASQS");
  RegisterPropertyInformation(STREQM_TAG,"TRIGGER.SOURCE",&dout,&din,CA_READ,AT_SCALAR,0,"event trigger specific server source",-1,NULL);

  RegisterProperty(STREQM_TAG,"TRIGGER.ITEM.DATA",100,CF_INT32,CA_READ|CA_WRITE,"trigger item input data");
  RegisterProperty(STREQM_TAG,"TRIGGER.ITEM.DATAFILES",100,CF_NAME16,CA_READ,"trigger item input data file list");
  RegisterProperty(STREQM_TAG,"TRIGGER.LIST",100,CF_NAME64,CA_READ,"get triggers involving the server input");

  readTriggerRedirections();
  isprimary = (strlen(thisExtension) == 0) ? TRUE : FALSE;
  for (i=0, n=0; i<arbTblSiz; i++) 
  { // a secondary should only register its 'own' devices
    regdev = isprimary;
    if (isSelfRedirection(arbTbl[i].trg)) regdev = TRUE;
    if (regdev) RegisterDeviceName(STREQM_TAG,arbTbl[i].trg,n++);
  }
  for (i=0; i<gNumTrgRedirs; i++) 
  { // anything to redirect ?
    if (!isprimary) break; // only the primary should redirect
    if (GetDeviceNumber(STREQM_TAG,trgRedir[i].trg) < 0)
      RegisterDeviceName(STREQM_TAG,trgRedir[i].trg,n++);
    if (strchr(trgRedir[i].rdr,'/') == NULL)
    { // just the context given ?
      sprintf(rdr,"/%.32s/EVENTSTORE/%.64s",trgRedir[i].rdr,trgRedir[i].trg);
      c = rdr;
    }
    else
    {
      c = trgRedir[i].rdr;
    }
    RedirectDeviceName(STREQM_TAG,trgRedir[i].trg,c);
  }

  AssignNumModules(STREQM_TAG,n);
  RegisterUserCommand("EventBufferSize", NULL, &gMaxNumEvents, NULL, CA_READ);

}
int forwardTrigger(int setNr,DTYPE *din)
{
  int cc = not_implemented;
  setNr = setNr; din = din;
# ifdef FIXED
  cc = ExecLinkEx(arbTbl[setNr].desc,"TRIGGER",NULL,din,CA_WRITE|CA_RETRY,800);
  feclog("%s -> forward trigger to %s : %s",
    arbTbl[setNr].trg,arbTbl[setNr].fwdsrv,cc2str(cc));
# endif
  return cc;

}
int logAndReturn(char *devName,char *devProperty,int returnCode,char *msg)
{
  char s[512], *c;

  if (returnCode >= 512 && returnCode < 512+numMyErr)
  {
    c = myErrLst[returnCode-512];
    SetEqmCompletion(STREQM_TAG,c);
  }
  else
  {
    c = cc2str(returnCode);
  }
  sprintf(s,"%.64s %.64s :: action rejected : %.32s",devProperty,devName,c);
  if (msg != NULL) 
  {
    strcat(s," -> ");
    strncat(s,msg,64);
  }
  AppendHistoryLog(s);
  /* clslog(gDeviceContext,"Services",STREQM_EXPORT,CLOG_STATUS_WARN,CLOG_PRIORITY_IMPORTANT,s); */
  return returnCode;
}
void streqm_exit(void)
{
  LHS *lhs=NULL, *nxt=NULL;
  
  for (lhs=lhsLst; lhs != NULL; lhs=nxt)
  {
    nxt = lhs->next;
    free(lhs);
  }
  lhsLst = NULL;

  freeSessions();
  freeArbTbl();
}
int addTriggerToQueue(int idx,int tot,int start,int stop,int max,int applyRange)
{
  EVNQ *e = calloc(1,sizeof(EVNQ));
  if (e == NULL) return out_of_server_memory;
  if (arbTbl[idx].nevnq >= arbTbl[idx].maxq) return at_limit;
  e->applyRange = applyRange;
  e->eventtime = tot;
  if (applyRange)
  {
    e->maxInRange = max;
    if (start > 0)
    { /* uses system stamp range specifications */
      if (stop <= 0) stop = gSystemDataStamp;
      if (stop < start)
      { /* meant as a increment to range start */
        if (stop > 1000) stop = 1000;
        stop += start;
      }
      e->tsStart = (double)(tot - RANGE_OFFSET);
      e->tsStop = (double)(time(NULL) + RANGE_OFFSET);
      e->sstStart = start;
      e->sstStop = stop;
    }
    else
    {
      e->tsStart = (double)(tot + start);
      e->tsStop = (double)(tot + stop);
      e->sstStart = 0;
      e->sstStop = 0;
    }
  }
  e->nxt = arbTbl[idx].evnq;
  arbTbl[idx].evnq = e;
  return 0;
}
#define EvntCanQueue(i) (arbTbl[i].maxq > 0)
int streqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  int n,len,cc,setNr,sts;
  int ivals[6], rngStart=0, rngStop=0, mxnum=0, hasRangeValues=FALSE, useusrstamp=FALSE;
  short firststep=1;
  char name[16],fullName[192],s[256],caller[USERNAME_SIZE];
  NAME64 n64[100];
  time_t tot = 0;
  char *c, *ctx, *srv, *prp;
  UINT32 evnt;
  static int lastSetNr = -1;
  static time_t lastTime = 0;

  if (!stricmp(devProperty,"TRIGGER"))
  {
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= arbTblSiz)
    {
      strncpy(caller,GetCaller(STREQM_TAG),USERNAME_SIZE);
      if (strstr(caller,"STATE") != NULL) return invalid_index;
      return logAndReturn(devName,devProperty,invalid_index,"the trigger entered was not found in the database");
    }
    if (arbTbl[setNr].maxq < 0) return logAndReturn(devName,devProperty,device_offline,"the trigger entered is offline");
    if (access&CA_WRITE && din->dArrayLength > 0)
    {
      char *caller = GetCaller(STREQM_TAG);
      sprintf(s,"%.16s : TRIGGER %.64s (set nr %d)",caller,devName,setNr);
      AppendHistoryLog(s);
      if (arbTbl[setNr].das == NULL) return logAndReturn(devName,devProperty,configuration_error,"the trigger entered has no contents");
      if ((cc=GetValuesAsLong(din,ivals,6)) != 0) return logAndReturn(devName,devProperty,cc,"the trigger input values could not be parsed");
      switch (din->dArrayLength)
      {
        case 6:
          if (ivals[5] != 0) useusrstamp = TRUE;
        case 5:
          mxnum = ivals[4];
        case 4:
          rngStop = ivals[3];
        case 3:
          rngStart = ivals[2];
          hasRangeValues = TRUE;
        case 2:
          tot = (time_t)ivals[1];
          AppendHistoryLog("time of trigger %d supplied as input",tot);
        case 1:
          firststep = ivals[0];
        default:
          break;
      }
      if (arbTbl[setNr].step || arbTbl[setNr].fp) 
      { /* busy handling a trigger already */
        if (EvntCanQueue(setNr) && tot > 0)
        {
          return addTriggerToQueue(setNr,(int)tot,rngStart,rngStop,mxnum,hasRangeValues);
        }
        return logAndReturn(devName,devProperty,operation_busy,"trigger currently active and the trigger queue is full");
      }
      if (tot < 1000) tot = time(NULL); /* official Time Of Trigger */
      arbTbl[setNr].tot = tot;
      if (firststep == 0) firststep = 1;
      if (lastSetNr == setNr && tot == lastTime) return logAndReturn(devName,devProperty,not_accepted,"the input trigger time is identical to the previous trigger time");
      clslog(gDeviceContext,"Services",STREQM_EXPORT,CLOG_STATUS_WARN,CLOG_PRIORITY_IMPORTANT,s);
      lastSetNr = setNr; 
      lastTime = tot;
      arbTbl[setNr].eventtime = tot; /* apply here: could be overriden if firststep < 0 */
      if (EventNotificationPending) /* I'm working on this event already */
      {
        arbTbl[setNr].resultPending = TRUE;
      }
      arbTbl[setNr].step = firststep;
      arbTbl[setNr].applyRange = APPLY_RANGE_NONE;
      AppendHistoryLog("trigger (%d steps) accepted",arbTbl[setNr].nsteps);
      if (hasRangeValues)
      {
        if (mxnum == 0) mxnum = 100;
        arbTbl[setNr].maxInRange = mxnum;
        if (rngStart > 0)
        { /* uses system/user stamp range specifications */
          arbTbl[setNr].applyRange = useusrstamp ? APPLY_USRSTAMP_RANGE : APPLY_SYSSTAMP_RANGE;
          if (rngStop <= 0) rngStop = gSystemDataStamp;
          if (rngStop < rngStart)
          { /* meant as a increment to range start */
            if (rngStop > 1000) rngStop = 1000;
            rngStop += rngStart;
          }
          arbTbl[setNr].tsStart = (double)(tot - RANGE_OFFSET);
          arbTbl[setNr].tsStop = (double)(time(NULL) + RANGE_OFFSET);
          arbTbl[setNr].sstStart = rngStart;
          arbTbl[setNr].sstStop = rngStop;
        }
        else
        { /* uses time stamp range specifications: rngStart is <= 0 (offset to tot) */
          arbTbl[setNr].applyRange = APPLY_TIMESTAMP_RANGE;
          arbTbl[setNr].tsStart = (double)(tot + rngStart);
          arbTbl[setNr].tsStop = (double)(tot + rngStop);
          arbTbl[setNr].sstStart = 0;
          arbTbl[setNr].sstStop = 0;
        }
      }
      else
      {
        arbTbl[setNr].tsStart = 0;
        arbTbl[setNr].tsStop = 0;
        arbTbl[setNr].sstStart = 0;
        arbTbl[setNr].sstStop = 0;
        arbTbl[setNr].maxInRange = 0;
      }
#     ifdef FIXED
      if (strlen(arbTbl[setNr].fwdsrv) > 0) forwardTrigger(setNr,din);
#     endif
      if (arbTbl[setNr].almCode > 0)
      {
        arbTbl[setNr].isAlarm = !strnicmp(arbTbl[setNr].almTrgUsr,caller,USERNAME_SIZE) ? TRUE : FALSE;
        AppendHistoryLog("set alarm (code %d) : %s", arbTbl[setNr].almCode, (arbTbl[setNr].isAlarm ? "TRUE":"FALSE"));
      }
    }
    if (dout->dArrayLength > 0)
    {
      return PutValuesFromLong(dout,(SINT32 *)&arbTbl[setNr].eventtime,1);
    }
    return 0;
  }
  if (!stricmp(devProperty,"TRIGGER.STATUS"))
  {
    if (strnicmp(devName,"ANY",DEVICE_NAME_SIZE) == 0)
    {
      sts = arbTblBusy;
    }
    else
    {
      setNr = getSetNumber(devName);
      if (setNr < 0 || setNr >= arbTblSiz) return logAndReturn(devName,devProperty,invalid_index,"the trigger entered was not found in the database");
      sts = arbTbl[setNr].step;
    }
    return PutValuesFromLong(dout,&sts,1);
  }
  if (!stricmp(devProperty,"TRIGGER.TIME"))
  {
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= arbTblSiz) return logAndReturn(devName,devProperty,invalid_index,"the trigger entered was not found in the database");
    return PutValuesFromLong(dout,(SINT32 *)&arbTbl[setNr].eventtime,1);
  }
  if (!stricmp(devProperty,"TRIGGER.REFRESH_DEVICES"))
  {
    if ((access & CA_WRITE) == 0) return illegal_read_write;
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= arbTblSiz) return logAndReturn(devName, devProperty, invalid_index, "the trigger entered was not found in the database");
    return RefreshDeviceLists(&arbTbl[setNr]);
  }
  if (!stricmp(devProperty,"SELECT"))  /* select a trigger context */
  {
    if (din->dArrayLength > 0)
    {
      setNr = getSetNumber(devName);
      if (setNr < 0 || setNr >= arbTblSiz) return invalid_index;
      idxSelectedARB = setNr;
    }
    if ((len=dout->dArrayLength) > 0)
    {
      switch (dout->dFormat)
      {
        case CF_TEXT: strncpy(dout->data.cptr,arbTbl[idxSelectedARB].trg,len);
                      break;
        case CF_NAME16: strncpy(dout->data.cptr,arbTbl[idxSelectedARB].trg,16);
                      break;
        default: return illegal_format;
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"COMMENT"))  /* read a comment */
  {
    int clen = 0;
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= arbTblSiz) return invalid_index;
    if (din->dArrayLength == 0) return illegal_read_write;
    if (dout->dFormat != CF_TEXT) return illegal_format;
    if (dout->dArrayLength < 1) return dimension_error;
    switch (din->dFormat)
    {
      case CF_LONG: evnt = din->data.lptr[0]; break;
      case CF_TEXT: evnt = atoi(din->data.cptr); break;
      default: return illegal_format;
    }
    len = (int)dout->dArrayLength;
    cc = readComment(setNr,evnt,dout->data.cptr,len);
    clen = (int)strlen(dout->data.cptr);
    if (clen < len) dout->dArrayLength = (UINT32)clen;
    return cc;
  }
  if (!stricmp(devProperty,"ANNOTATE"))  /* add a comment */
  {
    setNr = getSetNumber(devName);
    if (setNr < 0 || setNr >= arbTblSiz) return invalid_index;
    if (access&CA_WRITE && din->dArrayLength > 0)
    {
      if (din->dFormat != CF_TEXT) return illegal_format;
      if (din->dArrayLength == 0) return dimension_error;
      cc = parseAnnotationAndEvent(din->data.cptr,&c,&tot);
      if (cc != 0) return cc;
      feclog("ANNOTATE %.64s (%d) called (time of trigger: %10d)",devName,setNr,tot);
      if (tot == arbTbl[setNr].tot)
      {
        if (arbTbl[setNr].eventtime == 0) return operation_busy;
        feclog("EVENT %.64s (%d) @%.24s annotated",devName,setNr,ctime(&arbTbl[setNr].eventtime));
        return writeComment(setNr,(int)arbTbl[setNr].eventtime,c);
      }
      feclog("EVENT %.64s (%d) @%.26s annotated",devName,setNr,ctime(&tot));
      return writeComment(setNr,(UINT32)tot,c);
    }
    return illegal_read_write;
  }
  if (!stricmp(devProperty,"REMOVE"))
  {
    setNr = getSetNumber(devName);
    AppendHistoryLog("%s : REMOVE %s (set nr %d)",GetCaller(STREQM_TAG),devName,setNr);
    if (setNr < 0 || setNr >= arbTblSiz) return logAndReturn(devName,devProperty,invalid_index,"the trigger entered was not found in the database");
    if (!(access & CA_WRITE)) return logAndReturn(devName,devProperty,illegal_read_write,"WRITE access is required");
    if (din->dFormat != CF_INT32) return logAndReturn(devName,devProperty,illegal_format,"trigger time must be a32-bit integer");
    if (din->dArrayLength != 1) return logAndReturn(devName,devProperty,dimension_error,"input must be a single value");
    tot = din->data.lptr[0];
    if (tot == arbTbl[setNr].tot)
    { /* request removal on most recent archive for this set (on going?) */
      if (arbTbl[setNr].eventtime == 0 || arbTbl[setNr].step != 0)
      {
        arbTbl[setNr].step = arbTbl[setNr].nsteps+1;
        arbTbl[setNr].aborted = TRUE;
        return 0;
      }
      sprintf(name,"%x",(unsigned int)arbTbl[setNr].eventtime);
      sprintf(fullName,"%.8s.%.16s",name,arbTbl[setNr].ext);
      if (makeFullName(fullName,arbTbl[setNr].ext)) return logAndReturn(devName,devProperty,no_such_file,"unable to establish a trigger file name");
      if (arbTbl[setNr].step == 0) /* not actively archiving  */
      {
        if (archDebug) dbglog("STR: remove file %s",fullName);
        remove(fullName);
        sprintf(&fullName[strlen(fullName)-strlen(arbTbl[setNr].ext)],"CMT");
        if (archDebug) dbglog("STR: remove file %s",fullName);
        remove(fullName);
        return 0;
      }
    }
    else
    { /* removal of stored event */
      sprintf(fullName,"%x.%.16s",(unsigned int)tot,arbTbl[setNr].ext);
      if (makeFullName(fullName,arbTbl[setNr].ext)) return logAndReturn(devName,devProperty,no_such_file,"unable to establish a trigger file name");
      if (archDebug) dbglog("STR: remove file %s\n",fullName);
      remove(fullName);
      sprintf(&fullName[strlen(fullName)-strlen(arbTbl[setNr].ext)],"CMT");
      if (archDebug) dbglog("STR: remove file %s\n",fullName);
      remove(fullName);
      return 0;
    }
  }
  if (!stricmp(devProperty,"STOP"))
  {
    setNr = getSetNumber(devName);
    AppendHistoryLog("%s : STOP %s (set nr %d)",GetCaller(STREQM_TAG),devName,setNr);
    if (setNr < 0 || setNr >= arbTblSiz) return logAndReturn(devName,devProperty,invalid_index,"the trigger entered was not found in the database");
    if (!(access & CA_WRITE)) return logAndReturn(devName,devProperty,illegal_read_write,"WRITE access is required");
    if (arbTbl[setNr].fp == NULL) return logAndReturn(devName,devProperty,file_not_open,"trigger is not currently active");
    stopARB(&arbTbl[setNr]);
    return 0;
  }
  if (!stricmp(devProperty,"TRIGGER.TBL"))
  {
    if (access & CA_WRITE)
    {
      strncpy(caller,GetCaller(STREQM_TAG),USERNAME_SIZE);
      if (!isAdministrator(caller)) return access_denied;
      if (din->dArrayLength < 1) return dimension_error;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"ARBQS") != 0) return invalid_structure_tag;
      cc = AppendTriggerTable((ARBQS *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength == 0) return 0;
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (strcmp(dout->dTag,"ARBQS") != 0) return invalid_structure_tag;
    if ((setNr=getSetNumber(devName)) < 0) setNr = 0;
    n = GetTriggerTable(setNr,(ARBQS *)dout->data.vptr,dout->dArrayLength);
    if (n < 0) return -n;
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"TRIGGER.DB"))
  {
    setNr = getSetNumber(devName);
    if (access & CA_WRITE)
    {
      strncpy(caller,GetCaller(STREQM_TAG),USERNAME_SIZE);
      if (!isAdministrator(caller)) return access_denied;
      if (din->dArrayLength < 1) return dimension_error;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"DASQS") != 0) return invalid_structure_tag;
      cc = WriteDASTable(devName,(DASQS *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength == 0) return 0;
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (strcmp(dout->dTag,"DASQS") != 0) return invalid_structure_tag;
    n = GetDASTable(setNr,(DASQS *)dout->data.vptr,dout->dArrayLength);
    if (n < 0) return -n;
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"TRIGGER.SOURCE"))
  {
    DAS *das;
    setNr = getSetNumber(devName);
    if (access & CA_WRITE) return illegal_read_write;
    prp = NULL;
    switch (din->dArrayLength)
    {
      case 0: 
        return invalid_data;
      default:
        if (din->dFormat != CF_TEXT) return illegal_format;
        if (din->dArrayLength > 192) din->dArrayLength = 192;
        strncpy(fullName,din->data.cptr,din->dArrayLength);
        srv = fullName;
        if ((prp=strchr(fullName,'[')) != NULL)
        {
          *prp++ = 0; 
          if ((c=strchr(prp,']')) != NULL) *c = 0;
        }
        break;
      case 1:
        if (din->dFormat == CF_STRING)
        {
          srv = din->data.strptr[0];
        }
        else
        {
          if (!isSimpleStringFormat(din->dFormat)) return illegal_format;
          srv = din->data.cptr;
        }
        break;
      case 2:
        if (din->dFormat == CF_STRING)
        {
          srv = din->data.strptr[0];
          prp = din->data.strptr[1];
        }
        else
        {
          if (!isSimpleStringFormat(din->dFormat)) return illegal_format;
          srv = din->data.cptr;
          prp = din->data.cptr + fmtsizeof(din->dFormat);
        }
        break;
    }
    if (prp == NULL) return non_existent_property;
    if (dout->dArrayLength == 0) return 0;
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (strcmp(dout->dTag,"DASQS") != 0) return invalid_structure_tag;
    if ((das=GetARBDas(setNr,srv,prp)) == NULL) return non_existent;
    memcpy(dout->data.bptr,das,sizeof(DAS));
    return 0;
  }
  if (!stricmp(devProperty,"TRIGGER.ITEM.DATA"))
  {
    if (access & CA_WRITE)
    {
      strncpy(caller,GetCaller(STREQM_TAG),USERNAME_SIZE);
      if (!isAdministrator(caller)) return access_denied;
      cc = putDASInputData(devName,din);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength > 0)
    { /* type is known by db manager, size will be a 'large' guess */
      cc = getDASInputData(devName,dout);
    }
    return cc;
  }
  if (!stricmp(devProperty,"TRIGGER.ITEM.DATAFILES"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    return getDASInputFileNames(dout);
  }
  if (!stricmp(devProperty,"TRIGGER.LIST"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    ctx = srv = prp = NULL;
    if (dout->dArrayLength == 0) return dimension_error;
    if (din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsStringEx(din,n64[0].name,CF_NAME64,3,0)) != 0) return cc;
      switch (din->dArrayLength)
      {
        case 0: 
          break;
        case 1:
          strncpy(fullName,n64[0].name,64);
          if (fullName[0] == '/')
          {
            ctx = &fullName[1];
            if ((c=strchr(ctx,'/')) == NULL) return invalid_data;
            *c++ = 0;
            srv = c;
          }
          else
          {
            srv = fullName;
          }
          break;
        default:
        case 3:
          strncpy(s,n64[2].name,64); 
          prp = s; fullName[64] = 0;
        case 2:
          strncpy(fullName,n64[0].name,64);
          ctx = fullName; fullName[31] = 0;
          strncpy(&fullName[32],n64[1].name,64);
          srv = &fullName[32]; fullName[63] = 0;
          break;
      }
    }
    n = GetTriggersForServer(ctx,srv,prp,n64,100);
    if (n > 0)
    {
      if ((cc=PutValuesFromString(dout,n64[0].name,CF_NAME64,n,0)) != 0) return cc;
    }
    dout->dArrayLength = n;
    return 0;
  }
  return illegal_property;
}
