#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"

#define SETEQM_TAG "SETEQM"
#define PRP_SETTINGS_DB 5000
void PostSystemInit(void);
int seteqm(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access);
void seteqm_ini(void);
void seteqm_bkg(void);
void seteqm_exi(void);
int registerStructs(void);
#define MAX_CAPACITY 2048

char thisFecContext[32] = "TEST";
char thisFecServer[32] = "Settings";
char thisFecName[32];
char thisSubsystem[16] = "MISC";
char thisFecDescription[64];
char thisFecLocation[64] = "unknown";
char thisFecResponsible[32] = "";
int thisFecPort = 0;
int thisFecCapacity = MAX_CAPACITY; // this maximum number of servers + devices
extern char *gUsageString;

#define SETCFG_FILENAME "settings.csv"
char setCfgFile[64] = SETCFG_FILENAME;
int needsRestart = FALSE;
typedef struct PrpInfo
{
  char prp[64];
  NAME64 *devlst;
  int len; // num of devices (size of list)
  int fmt; // data format
  int siz; // num elems per device
  int pid;
  BYTE *values;
  char units[16];
  double tstamp;
} PrpInfo;
PrpInfo *prpInfo = NULL;

#define SETDB_TAG "StgDb"
typedef struct SPrpInfo
{
  char prp[64];
  char dev[64];
  char fmt[16];
  char val[64];
  char egu[16];
  int  siz; // num elems this device
  int  scanned;
  int  len; // total num devices
} SPrpInfo;
SPrpInfo *sprpinfo = NULL;
int nSettings = 0;
#define SPrpsFileHdr "Property, Device, Format, Size, Value, Units"
CSVCOLUMN csvSPrpsFileColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(SPrpInfo,prp),NULL,0 },
  { "DEVICE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(SPrpInfo,dev),NULL,0 },
  { "FORMAT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(SPrpInfo,fmt),NULL,0 },
  { "SIZE","",0,CF_INT32,1,(BYTE *)OFFSETIN(SPrpInfo,siz),NULL,0 },
  { "VALUE","0.0",0,CF_TEXT,64,(BYTE *)OFFSETIN(SPrpInfo,val),NULL,0 },
  { "UNITS","",0,CF_TEXT,16,(BYTE *)OFFSETIN(SPrpInfo,egu),NULL,0 },
};
CSVDB csvSPrpsFileDb =
{
  csvSPrpsFileColumns,
  sizeof(csvSPrpsFileColumns) / sizeof(csvSPrpsFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(SPrpInfo),
  0
};
// we should be able to retire this routine :
void RestartServer(void)
{
  static int restartCountdown = 5;
  if (restartCountdown-- > 0) return;
  FlushContractTable();
  SetServerCycleIdle(TRUE);
  FreeEQPModules();
  SystemFree(sprpinfo);
  SystemFree(prpInfo);
  nSettings = 0;
  ServerExitCondition = 0;
  PostSystemInit();
  SetServerCycleIdle(FALSE);
  needsRestart = FALSE;
  restartCountdown = 5;
}
char *getAbbreviatedFecName(char *srv, int maxlen)
{
  static char fec[FEC_NAME_SIZE + 8];
  int i, k, slen, sflen = 0;
  char *suf = NULL, tmp[EXPORT_NAME_SIZE + 8];
  strncpy(tmp, srv, EXPORT_NAME_SIZE); tmp[EXPORT_NAME_SIZE] = 0;
  slen = (int)strlen(tmp);
  for (i = slen - 1; i > 0 && isdigit(tmp[i]); i--);
  suf = &tmp[i + 1];
  if ((sflen = (int)strlen(suf)) > 5)
  { /* allow up to a 5-digit numerical name */
    suf = NULL;
    sflen = 0;
  }
  slen -= sflen;
  memset(fec, 0, FEC_NAME_SIZE + 8);
  for (i = 0, k = 0; i < slen && k < maxlen; i++)
  {
    if (tmp[i] == 0) break;
    if (tmp[i] == '*') continue;
    if (tmp[i] == '.') continue;
    if (tmp[i] == '_') continue;
    if (tmp[i] == '-') continue;
    fec[k++] = tmp[i];
  }
  if (sflen > 0)
  {
    if ((int)strlen(fec) > maxlen - sflen) fec[maxlen - sflen - 1] = 0;
    strncat(fec, suf, sflen);
  }
  return fec;
}
char *suggestFecName(char *ctx, char *srv, int maxlen)
{
  static char fec[FEC_NAME_SIZE + 8];
  int len;

  memset(fec, 0, FEC_NAME_SIZE + 8);
  if (srv == NULL) return fec;
  if (ctx != NULL)
  {
    char *c;
    if ((c = strstr(ctx, ".SIM")) != NULL) *c++ = 0;
    else if ((c = strstr(ctx, "_SIM")) != NULL) *c++ = 0;
    else if ((c = strstr(ctx, ".TEST")) != NULL) *c++ = 0;
    else if ((c = strstr(ctx, "_TEST")) != NULL) *c++ = 0;
    strncpy(fec, ctx, 2);
    if (c != NULL) strncat(fec, c, 1);
  }
  if (maxlen > FEC_NAME_SIZE) maxlen = FEC_NAME_SIZE;
  len = maxlen - (int)strlen(fec);
  strncat(fec, getAbbreviatedFecName(srv, len), len);
  strupr(fec);
  if ((int)strlen(fec) < (maxlen - 2) && strstr(fec, "FEC") == NULL) strcat(fec, "FEC");
  return fec;
}
void stgCleanup(void)
{
  SystemFree(sprpinfo);
  SystemFree(prpInfo);
  nSettings = 0;
  ServerExitCondition = 0;
}
void PreSystemInit(void)
{
  char *usr = getenv("USERNAME");
  if (strlen(thisFecServer) == 0) goto err;
  SetSystemCleanupFunction(stgCleanup);
  if (usr == NULL) usr = "UNKNOWN";
  if (strlen(thisFecName) == 0)
  {
    strncpy(thisFecName, suggestFecName(thisFecContext, thisFecServer, 16), 16);
  }
  if (strlen(thisFecResponsible) == 0) strncpy(thisFecResponsible, usr, 32);
  if (thisFecPort == 0) thisFecPort = GetPortOffset(thisFecName);
  FeclogDepth = 1000;
  RegisterFecInformation(thisFecName, thisSubsystem, thisFecContext, thisFecDescription, thisFecLocation, "none", thisFecResponsible, (UINT16)thisFecPort);
  return;
err:
  printf("%s\n", gUsageString);
  exit(0);
}
void PostSystemInit(void)
{
  int cc;
  if ((cc = registerStructs()) != 0)
  {
    feclogEx(TINE_LOGLEVEL_FATAL, "cannot register structures! : %.32s", cc2str(cc));
    exit(1);
  }
  // set 'capacity' = 1 (due to registration of 'trace' array properties
  // i.e. any database properties will need to assign device lists ...
  RegisterEquipmentModule(thisFecServer, SETEQM_TAG, 1, seteqm, seteqm_ini, seteqm_bkg, 200, seteqm_exi);
}
int getPrpInfoListIndex(char *prp, char *dev)
{
  int i;
  if (prp == NULL || strlen(prp) == 0) return -invalid_name;
  if (dev == NULL || strlen(dev) == 0) return -invalid_name;
  for (i = 0; i < nSettings; i++)
  {
    if (strnicmp(prp, sprpinfo[i].prp, 64) != 0) continue;
    if (strnicmp(dev, sprpinfo[i].dev, 64) != 0) continue;
    return i;
  }
  return -1;
}
int checkNewEntries(SPrpInfo *newprpinfo,int len)
{
  int i, idx, fmt, cc = 0, reinit=FALSE;
  char fn[512];
  FILE *fp = NULL;
  if (newprpinfo == NULL || len == 0) return argument_list_error;
  for (i = 0; i < len; i++)
  { // well formed ?
    if (strlen(newprpinfo[i].prp) == 0) return invalid_name;
    if (strlen(newprpinfo[i].dev) == 0) return invalid_name;
    fmt = GetFormatFromString(newprpinfo[i].fmt);
    if (GetFormatSize(fmt) == 0) return invalid_value;
    if (newprpinfo[i].siz <= 0) return illegal_data_size;
    idx = getPrpInfoListIndex(newprpinfo[i].prp, newprpinfo[i].dev);
    if (idx < 0)
    { // is new ...
      newprpinfo[i].scanned = 1;
      reinit = TRUE;
    }
    else
    { // is an edit ...
      if (newprpinfo[i].scanned < 0)
      { // signal to remove
        sprpinfo[idx].scanned = -1;
        reinit = TRUE;
        continue;
      }
      newprpinfo[i].scanned = 0;
      strncpy(sprpinfo[idx].egu, newprpinfo[i].egu, 16);
      strncpy(sprpinfo[idx].val, newprpinfo[i].val, 64);
      if (!strnicmp(sprpinfo[idx].fmt, newprpinfo[i].fmt, 16) &&
        sprpinfo[idx].siz == newprpinfo[i].siz)
      { // minor edit of egu, or default value
        sprpinfo[idx].scanned = 1;
      }
      else
      {
        strncpy(sprpinfo[idx].fmt, newprpinfo[i].fmt, 16);
        sprpinfo[idx].siz = newprpinfo[i].siz;
        sprpinfo[idx].scanned = 0;
        reinit = TRUE;
      }
    }
  }
  snprintf(fn, 128, "%s%c%s", SystemGetStartupDirectory(), FS_DELIMITER,setCfgFile);
  if ((fp = fopen(fn, "wt")) == NULL) return file_error;
  fprintf(fp,"%s\n", SPrpsFileHdr);
  //"Property, Device, Format, Size, Value, Units"
  for (i = 0; i < nSettings; i++)
  {
    if (sprpinfo[i].scanned < 0) continue;
    fprintf(fp,"%.64s, %.64s, %.16s, %d, %.64s, %.16s\n",
      sprpinfo[i].prp, sprpinfo[i].dev, sprpinfo[i].fmt,
      sprpinfo[i].siz, sprpinfo[i].val, sprpinfo[i].egu);
  }
  for (i = 0; i < len; i++)
  {
    if (newprpinfo[i].scanned <= 0) continue;
    fprintf(fp,"%.64s, %.64s, %.16s, %d, %.64s, %.16s\n",
      newprpinfo[i].prp, newprpinfo[i].dev, newprpinfo[i].fmt, 
      newprpinfo[i].siz, newprpinfo[i].val, newprpinfo[i].egu);
  }
  fclose(fp);
  if (reinit) cc = not_registered;
  return cc;
}
int dbhdlr(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access)
{
  int cc = 0;
  if (access & CA_WRITE)
  {
    SPrpInfo *inpt;
    if (din->dArrayLength < 1) return dimension_error;
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if (strncmp(din->dTag, SETDB_TAG, TAG_NAME_SIZE)) return invalid_structure_tag;
    inpt = (SPrpInfo *)din->data.vptr;
    // accept or reject the new entries / edit
    cc = checkNewEntries(inpt, din->dArrayLength);
    if (cc == not_registered)
    {
      SystemReInit(0);
      cc = 0;
    }
    return cc;
  }
  if (dout->dArrayLength > 0)
  {
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (strncmp(dout->dTag, SETDB_TAG, TAG_NAME_SIZE)) return invalid_structure_tag;
    if (dout->dArrayLength > (UINT32)csvSPrpsFileDb.siz)
      dout->dArrayLength = (UINT32)csvSPrpsFileDb.siz;
    memcpy(dout->data.vptr, sprpinfo, dout->dArrayLength * sizeof(SPrpInfo));
  }
  return 0;
}
int seteqm(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access)
{
  int did, pid, cc = 0;
  int osiz, len;
  BYTE *pvals;
  pid = GetPropertyId(SETEQM_TAG, devProperty);
  did = GetDeviceNumberEx(SETEQM_TAG, devName, devProperty);
  if (pid < 0) return illegal_property;
  if (did < 0) return illegal_device;
  osiz = GetFormatSize(prpInfo[pid].fmt) + GetFormatHeaderSize(prpInfo[pid].fmt);
  //osiz *= prpInfo[pid].siz;
  len = prpInfo[pid].siz; 
  if (prpInfo[pid].siz == 1) len *= prpInfo[pid].len;
  pvals = prpInfo[pid].values;
  if (prpInfo[pid].siz > 1)
  { // a DOUBLE array of something (i.e. not a CHANNEL)
    pvals += osiz * len * did;
    did = 0;
  }
  if (din->dArrayLength > 0)
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dArrayLength == 0 || din->dArrayLength > (UINT32)prpInfo[pid].siz) return dimension_error;
    if ((cc=GetValuesAsAny(din, pvals, prpInfo[pid].fmt, osiz, len, did)) != 0) return cc;
    prpInfo[pid].tstamp = MakeDataTimeStamp();
  }
  if (dout->dArrayLength > 0)
  {
    cc = PutValuesFromAny(dout, pvals, prpInfo[pid].fmt, TRUE, osiz, len, did);
    if ((int)dout->dArrayLength > len) dout->dArrayLength = (UINT32)len;
    dout->dTimeStamp = prpInfo[pid].tstamp;
  }
  return cc;
}


int registerStructs(void)
{
  static int done = FALSE;
  int cc;
  if (done) return 0;
  done = TRUE;  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, prp), 64, CF_TEXT, "property")) != 0) return cc;
  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, dev), 64, CF_TEXT, "device")) != 0) return cc;
  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, fmt), 16, CF_TEXT, "format")) != 0) return cc;
  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, val), 64, CF_TEXT, "default_value")) != 0) return cc;
  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, egu), 16, CF_TEXT, "units")) != 0) return cc;
  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, siz), 1, CF_INT32, "size")) != 0) return cc;
  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, scanned), 1, CF_INT32, "active")) != 0) return cc;
  if ((cc = AddFieldToStruct(SETDB_TAG, (int)OFFSETIN(SPrpInfo, len), 1, CF_INT32, "length")) != 0) return cc;
  if ((cc = sealTaggedStruct(SETDB_TAG, sizeof(SPrpInfo), 100)) != 0) return cc;
  return 0;
}

void seteqm_bkg(void)
{

}
void seteqm_exi(void)
{

}
HistorySpecification hspec =
{
  1000,1000,60,120,900,0,0,NULL,NULL
};
void seteqm_ini(void)
{
  char wd[128], prp[64], dsc[256], dev[64];
  int i, np, nd, id, idx = 0, k = -1;
  int cc = 0, len, dsiz = 0, inext = 0, atyp = AT_CHANNEL, acc;
  NAME64 *devs = NULL;
  BYTE *pval = NULL;
  DTYPE dout, din;
  double ts = MakeDataTimeStamp();
  snprintf(wd, 128, "%s%c", SystemGetStartupDirectory(), FS_DELIMITER);
  feclog("looking for settings file %s in %.128s", setCfgFile, wd);
  if ((cc=csvReadFile(wd, setCfgFile, &csvSPrpsFileDb, (void **)&sprpinfo)) != 0)
  {
    if (cc == no_such_file) goto fin;
    tineDebug = 1;
    feclogEx(TINE_LOGLEVEL_FATAL, "cannot read settings properties file %.64s! : %.32s", setCfgFile, cc2str(cc));
    exit(1);
  }
  nSettings = csvSPrpsFileDb.siz;
  feclog("read settings file from %.128s (%d entries)", wd, nSettings);
  devs = (NAME64 *)alloca(nSettings * sizeof(NAME64));
  if (devs == NULL)
  {
    feclogEx(TINE_LOGLEVEL_FATAL, "cannot allocate memory for device list files!");
    exit(1);
  }
  prp[0] = 0;
  for (i = 0, np = 0; i < nSettings; i++)
  {
    if (strnicmp(sprpinfo[i].prp, prp, 64) != 0 && !sprpinfo[i].scanned) 
    { // a new one ...
      np++; 
      strncpy(prp, sprpinfo[i].prp, 64);
      for (k = 0, nd = 0; k < nSettings; k++)
      { // find and flag any other references
        if (!strnicmp(sprpinfo[k].prp, prp, 64))
        {
          sprpinfo[k].scanned = 1;
          nd++;
        }
      }
      sprpinfo[i].len = nd;
    }
  }
  prpInfo = (PrpInfo *)calloc(np, sizeof(PrpInfo));
  if (prpInfo == NULL)
  {
    feclogEx(TINE_LOGLEVEL_FATAL, "cannot allocate memory for property list files!");
    exit(1);
  }
  for (i=0; i<nSettings; i++) sprpinfo[i].scanned = 0; // reset
  for (k=0; k<np; k++) 
  {
    if (inext < 0) break;
    nd = 0;
    prpInfo[k].fmt = CF_NULL;
    for (i=inext; i < nSettings; i++)
    {
      if (sprpinfo[i].scanned) continue;
      if (strlen(prpInfo[k].prp) == 0)
      { // new property in the list noted
        strncpy(prpInfo[k].prp, sprpinfo[i].prp, 64);
        sprpinfo[i].scanned = 1;
        inext = -1;
      }
      if (!strnicmp(sprpinfo[i].prp, prpInfo[k].prp, 64))
      { // another item (or the 1st one) in the list with same property name
        sprpinfo[i].scanned = 1;
        strncpy(devs[nd].name, sprpinfo[i].dev, 64);
        if (prpInfo[k].fmt == CF_NULL)
        {
          dsiz = 0;
          prpInfo[k].fmt = GetFormatFromString(sprpinfo[i].fmt);
          strncpy(prpInfo[k].units, sprpinfo[i].egu, 16);
          prpInfo[k].siz = sprpinfo[i].siz;
          if (sprpinfo[i].len > 0 && prpInfo[k].values == NULL)
          { // the first one in the set will have the length
            dsiz = prpInfo[k].siz * GetFormatSize(prpInfo[k].fmt) + GetFormatHeaderSize(prpInfo[k].fmt);
            prpInfo[k].values = (BYTE *)calloc(dsiz, sprpinfo[i].len);
            if (prpInfo[k].values == NULL)
            {
              feclogEx(TINE_LOGLEVEL_FATAL, "cannot allocate memory for values!");
              exit(1);
            }
          }
        }
        pval = prpInfo[k].values + nd * dsiz;
        csvAssignValue(pval, prpInfo[k].fmt, prpInfo[k].siz, sprpinfo[i].val);
        nd++;
      }
      else
      { // marked as the next property
        if (inext < 0) inext = i;
      }
    }
    prpInfo[k].tstamp = ts;
    prpInfo[k].len = nd;
    prpInfo[k].devlst = (NAME64 *)calloc(nd, sizeof(NAME64));
    if (prpInfo[k].devlst == NULL)
    {
      feclogEx(TINE_LOGLEVEL_FATAL, "cannot allocate memory for device list!");
      exit(1);
    }
    memcpy(prpInfo[k].devlst, devs, nd * sizeof(NAME64));
    atyp = prpInfo[k].siz == 1 ? AT_CHANNEL : AT_SPECTRUM;
    DTYPEZERO(dout);
    DTYPEZERO(din);
    len = prpInfo[k].len; // *prpInfo[k].siz;
    dout.dArrayLength = atyp == AT_CHANNEL ? len : prpInfo[k].siz;
    din.dArrayLength = prpInfo[k].siz;
    din.dFormat = dout.dFormat = prpInfo[k].fmt;
    acc = CA_READ | CA_WRITE;
    if (atyp == AT_CHANNEL || prpInfo[k].len == 1) acc |= CA_SAVERESTORE;
    sprintf(dsc, "%.64s settings", prpInfo[k].prp);
    prpInfo[k].pid = RegisterPropertyInformation(SETEQM_TAG, prpInfo[k].prp, &dout, &din, acc, atyp, prpInfo[k].siz, dsc, k, NULL);
    AssignDeviceListToProperty(SETEQM_TAG, prpInfo[k].prp, prpInfo[k].devlst, prpInfo[k].len);
    if (atyp == AT_CHANNEL) len = 1; // reset for the channel guys
    for (id = 0; id < len; id++)
    { // loop is mostly 1, except for the non-channel guys with multiple devices
      sprintf(dev, "#%d", id);
      idx = FindAvailableHistoryIndex(SETEQM_TAG, prpInfo[k].prp, dev, prpInfo[k].fmt, (int)dout.dArrayLength);
      AppendHistoryInformation(SETEQM_TAG, prpInfo[k].prp, dev, (int)dout.dArrayLength, prpInfo[k].fmt, idx, &hspec);
      if ((acc & CA_SAVERESTORE) == 0 && pval != NULL)
      {
        pval = prpInfo[k].values + id * dsiz;
        idx = id;
        GetLastStoredData(SETEQM_TAG, prpInfo[k].prp, dev, idx, pval, prpInfo[k].fmt, dout.dArrayLength);
      }
    }
  }
fin:
  dout.dArrayLength = din.dArrayLength = 256; // a nice number
  dout.dFormat = din.dFormat = CF_STRUCT;
  strncpy(dout.dTag, SETDB_TAG, TAG_NAME_SIZE);
  strncpy(din.dTag, SETDB_TAG, TAG_NAME_SIZE);
  RegisterPropertyAndHandler(SETEQM_TAG, dbhdlr, "Settings.DB", &dout, &din, CA_READ | CA_WRITE, AT_COLLECTION, 1, "settings database", NULL, NULL, PRP_SETTINGS_DB);
}