#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#ifdef UNIX
#include <dlfcn.h>
#endif
#include "tine.h"
#include "threader.h"

int gSystemInitialized=FALSE;
int cycle_busy = FALSE;
int cycleTerminated = FALSE;
unsigned int idCycleTimer = 0;
int userNameSet = FALSE;
int gSystemDelay = 20;

int gUseWindowsTimer = FALSE;
int gWindowsTimerInUse = FALSE;
int gThreadTimerInUse = TRUE;
int cycleStarted = FALSE;
int gIsClosingDynamic = FALSE;
int gTmrThreadDetached = FALSE;
THRHANDLE hTmrThread = (THRHANDLE)0;

#if defined(WIN32)
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
#elif defined(UNIX)
pthread_attr_t gCycleThreadAttr, *gPtrCycleThreadAttr = &gCycleThreadAttr;
pthread_mutex_t cycleStartMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t *hCyclerStartMutex = &cycleStartMutex;
pthread_mutex_t cycleStopMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t *hCyclerStopMutex = &cycleStopMutex;
#elif defined(VXWORKS)
#define MULTITHREADED
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
extern int gVxwDefaultTaskOptions;
extern int gVxwDefaultTaskPriority;
extern int gVxwDefaultTaskStack;
#endif

#ifndef VXWORKS
/* use for debugging purposes ... */
void cyclog(char *txt,...)
{
  static char fn[64] = "cycler.log";
  FILE *fp=NULL;
  time_t t = time(NULL);
  char str[1024];
  long flen;
  int cc=0;
  va_list args;

  va_start(args,txt);
  vsnprintf(str,512,txt,args); str[512] = 0;
  va_end(args);
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");

  if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  fseek(fp, 0L, SEEK_END); flen = ftell(fp);
  fwrite(str,strlen(str),1,fp);
err:
  if (fp != NULL) fclose(fp);
  return;
}
#endif
void InitCycleMutexSet(void);
void TineStopCycler(void)
{
  if (cycleTerminated) return;
  SystemKillCycleTimer();
  millisleep(20);
}
#if defined(UNIX) && !defined(SOLARIS)
#ifndef TINE_START_CYCLER
# define TINE_START_CYCLER FALSE
#endif
__attribute__((constructor)) void TineLoadCycler(void)
{
  int startCycler = TINE_START_CYCLER;
  char *ptr = getenv("TINE_START_CYCLER");
  gIsDynamicLoad = TRUE;
  if (ptr != NULL)
  {
    startCycler = stricmp(ptr,"TRUE") == 0 ? TRUE : FALSE;
  }
  if (startCycler)
  { /* start cycler immediately */
    SystemSetCycleTimer();
  }
  else
  { /* let the app manage the cycling */
    InitCycleMutexSet();
  }
}
__attribute__((destructor)) void TineStopCyclerEpilog(void)
{
  gIsClosingDynamic = TRUE;
  TineStopCycler();
  freeServerMemory();
}
#endif
void InitCycleMutexSet(void)
{
  static int done = FALSE;
  if (done) return;
  done = TRUE;
  InitMutex(&hCyclerStartMutex);
  InitMutex(&hCyclerStopMutex);
# if defined(UNIX)
  pthread_mutex_lock(hCyclerStartMutex);
  pthread_mutex_lock(hCyclerStopMutex);
# endif
  ReleaseSystemMutex(hCyclerStartMutex);
  ReleaseSystemMutex(hCyclerStopMutex);
}

void cycleTmrTask(void)
{
  if (!gSystemInitialized) 
  { /* don't spin in not initialized yet! */
    millisleep(gSystemDelay);
    return;
  }
  if (cycleTerminated) return;
  if (cycle_busy) return;
  cycle_busy = TRUE;
  SystemCycle(foregroundTTY);
  cycle_busy = FALSE;
  if (ServerExitCondition > 3) cycleTerminated = TRUE;
}

void cycleInitTask(int isServer)
{
  if (isServer) 
  { 
    SystemInit(FALSE);
  }
  else
  {
    initClient();
  }
  gSystemInitialized = TRUE;
  MaxPollingRate = MinPollingRate = 20;
}

THREAD_TYPE cycleTmrThread(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;

  SetThreadName("tcycler");
  if (WaitForSystemMutex(hCyclerStartMutex,-1) != 0) 
  {
    feclog("cycleTmrThread: start cycler mutex unavailable !");
    return 0;
  }
  UseMultiThreadedCycler(TRUE);
  while (!cycleTerminated)
  {
    cycleTmrTask();
  }
  cycleStarted = FALSE;
  CloseServers(TRUE);
  FreeEQPModules();
# ifdef UNIX
  pthread_detach(hTmrThread);
# endif
  hTmrThread = (THRHANDLE)0;
  gTmrThreadDetached = TRUE;
  ReleaseSystemMutex(hCyclerStartMutex);
  return 0;
}
#if defined(WIN32)
SECURITY_ATTRIBUTES tmrSA;
SECURITY_DESCRIPTOR tmrSD;
DWORD dwTmrThreadId;
THRHANDLE cycleCreateThread(void)
{
  THRHANDLE hThread = 0;
  int threadPriority = -1;
  char *ptr = getenv("TINE_CYCLE_THREAD_PRIORITY");

  InitializeSecurityDescriptor(&tmrSD,SECURITY_DESCRIPTOR_REVISION);
  SetSecurityDescriptorDacl(&tmrSD,TRUE,NULL,TRUE);
  tmrSA.nLength = sizeof(tmrSA);
  tmrSA.lpSecurityDescriptor = &tmrSD;
  tmrSA.bInheritHandle = TRUE;

  cycleTerminated = FALSE;
  hThread = (THRHANDLE)CreateThread(&tmrSA,0,cycleTmrThread,0,0,&dwTmrThreadId);
  if (ptr != NULL) threadPriority = atoi(ptr);
  if (threadPriority > 0) 
  {
    feclog("local cycle thread priority set to %d via environment",threadPriority);
    SetThreadPriority(hThread,threadPriority);
  }
  return hThread;
}
#elif defined(UNIX)
THRHANDLE cycleCreateThread(void)
{
  int cc;
  pthread_t tid;

  pthread_attr_init(gPtrCycleThreadAttr);
# ifdef FORCE_PTHREAD_SCOPE_SYSTEM
  if ((cc=pthread_attr_setscope(gPtrCycleThreadAttr,PTHREAD_SCOPE_SYSTEM)) != 0) 
  {
    feclog("Cannot set thread scope !\n");
  }
# endif
  cycleTerminated = FALSE;
  cc = pthread_create(&tid,gPtrCycleThreadAttr,cycleTmrThread,NULL);
  return cc ? 0 : (THRHANDLE)tid;
}
#elif defined(VXWORKS)
THRHANDLE cycleCreateThread(void)
{
  static int cnt=0;
  char tname[32];
  int tskid = taskIdSelf();
  int tskp = gVxwDefaultTaskPriority, tsko = gVxwDefaultTaskOptions;
  int tcks = sysClkRateGet()/100;
  cnt++;
  strncpy(tname,"tcycler",8); tname[7] = 0;
  taskPriorityGet(tskid,&tskp);
  taskOptionsGet(tskid,&tsko);
  if (tskp == 1) 
  { /* in case started from the shell prompt */
    tskp = gVxwDefaultTaskPriority;
    tsko = gVxwDefaultTaskOptions;
  }
  if (tcks < 1) tcks = 1;
  kernelTimeSlice(tcks);
  cycleTerminated = FALSE;
  return (THRHANDLE)taskSpawn(tname,tskp,tsko,gVxwDefaultTaskStack,(FUNCPTR)cycleTmrThread,0,0,0,0,0,0,0,0,0,0);
}
#endif

TINE_EXPORT void killCycleTimer(void)
{
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    feclog("SystemKillCycleTimer: stop cycler mutex unavailable !");
    return;
  }
  if ((!cycleStarted && !gIsClosingDynamic) || cycleTerminated)
  {
    feclog("attempt to kill timer refused : cycle %s",cycleStarted ? "already terminated" : "not started");
    ReleaseSystemMutex(hCyclerStopMutex);
    return;
  }
  CloseClients(TRUE);
  cycleTerminated = TRUE; /* causes the cycle thread to stop => issues a SystemReset() */
  CloseClientMutexSet();
  feclog("terminating local cycle");
  millisleep(gSystemDelay*2);
  CloseSystemMutexSet();
  ReleaseSystemMutex(hCyclerStopMutex);
  if (WaitForSystemMutex(hCyclerStartMutex,gSystemTick) != 0) 
  {
    return;
  }
  ReleaseSystemMutex(hCyclerStartMutex);
}
TINE_EXPORT void SystemWaitCycleTimer(void)
{
  if (gTmrThreadDetached) return;
  if (hTmrThread != (THRHANDLE)0) WaitForThreadTermination(hTmrThread);
  gTmrThreadDetached = TRUE;
}
TINE_EXPORT void SystemKillCycleTimer(void)
{
  killCycleTimer();
}
void TineStartCycler(void)
{
  SystemSetCycleTimer();
}
TINE_EXPORT void SystemStartCycleTimer(void)
{
  SystemSetCycleTimer();
}
void SystemSetCycleTimer(void)
{
  int cc = 0;
  InitCycleMutexSet();
  InitSystemMutexSet();
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    if (gIsRunningAsServer) feclog("SystemSetCycleTimer: stop cycler mutex unavailable !");
    return;
  }
  if (cycleStarted) ccerr(-1);
  cycleStarted = TRUE;
  cycleInitTask(FALSE);
  hTmrThread = cycleCreateThread();
  millisleep(20);
err:
  ReleaseSystemMutex(hCyclerStopMutex);
  return;
}
