#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "threader.h"
#include "unittest.h"

int echoTest(char *tgt,NAME64 *devs);
int tcpSyncTest(char *tgt,NAME64 *devs,int flags);
int structTest(char *tgt,NAME64 *devs,char *tag);
int stringTest(char *tgt,NAME64 *devs);
int spectrumTest(char *tgt,NAME64 *devs);
int spectrumArrayTest(char *tgt,NAME64 *devs);
int gspectrumTest(char *tgt, NAME64 *devs);
int mdaTest(char *tgt,NAME64 *devs);
int defaultTest(char *tgt,NAME64 *devs);
int imageTest(char *tgt,NAME64 *devs);
int imageArrayTest(char *tgt,NAME64 *devs);
int wildcardTest(char *tgt);
int mcaLinkTest(char *tgt,NAME64 *devs);
int bitfield16Test(char *tgt,NAME64 *devs);
int bitfield32Test(char *tgt,NAME64 *devs);
int stressTest(char *tgt,NAME64 *devs,int nloops,short flags);
int settingsTest(char *tgt,NAME64 *devs);
int commandTest(char *tgt);
int stkTest(char *tgt, NAME64 *devs);
int histTest(char *tgt,NAME64 *devs);
int largeArraytest(char *tgt, short flags);
int largeEchoTest(char *tgt, short flags);
int linkErrorTest(char *tgt);
int ensQueryTest(void);
#ifdef UNIX
void TineStopCycler(void);
#endif
char tgtContext[CONTEXT_NAME_SIZE] = "TEST";
char tgtServer[EXPORT_NAME_SIZE] = "UnitServer";
char *usageText = "usage: unitclient [/c=<context> /s=<server name> /v=<verbose level> /d=<debug level /k=<y/n (kill server after test, default is y)>";

#if defined(UNIX) || defined(WIN32)
# define random(r) (rand()%(r))
#else
# define random(r) ((i*(69069))%(r))
#endif

#define BLOB_SIZE 65536 * 8
BYTE blobIn[BLOB_SIZE]; 
BYTE blobOut[BLOB_SIZE]; 
char echoStr[1024];
int verbose = 0;
int linkactive = FALSE;
int stayactive = FALSE;
int nUnitDevices = 32;
void quit(int cc)
{
  writeTestResultQuit(cc);
  printf("access to /%.32s/%.32s failed : %.32s\n",tgtContext,tgtServer,cc2str(cc));
  exit(1);
}
int ntotal = 0;
void reportResult(char *txt,int n)
{
  if (n == 0) 
  { 
    printf("-------- %s test passed ! ---------\n",txt);
  }
  else
  {
    printf("!!!!!!!!! %d errors in %s test !!!!!!!!!\n",n,txt);
  }
  writeTestsuiteLogTestcase( "functiontest", txt, n);
  ntotal += n;
  ++testcaseCount;
}
int main(int argc, char *argv[])
{
  int i,cc,ec;
  char *c, ch, scratch[256];
  char tgt[128];
  NAME64 devs[32];

  writeTestsuiteLogStart();
  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(tgtContext,c,CONTEXT_NAME_SIZE);
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(tgtServer,c,EXPORT_NAME_SIZE);
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);
		      SetInterpretConsoleCommands(TRUE);
        }
        continue;
      case 'v':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          verbose = atoi(scratch);      
        }
        continue;
      case 'w':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          waitForServer = atoi(scratch);      
        }
        continue;
      case 'k':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          if((strchr(c,'y')) != NULL) {
            killserver = TRUE;
          }
          else if((strchr(c,'n')) != NULL) {
            killserver = FALSE;
          }
        }
        continue;
      case 'a':
        if ((c=strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          stayactive = atoi(scratch);
        }
        continue;
      case '?':
      case 'h':
        printf("%s",usageText);
        exit(0);
      default:
        continue;
    }
  }
  
  millisleep(waitForServer*1000);

  gConTblCapacity = 10;
  SystemStartCycleTimer();

  sprintf(tgt,"/%.32s/%.32s",tgtContext,tgtServer);

  nUnitDevices = 32;
  cc = GetSystemDevices(tgt,NULL,devs,&nUnitDevices);
  if (cc != 0) quit(cc);

  SetValuesAsStringSeparator(", ");  
  reportResult("ECHO",echoTest(tgt,devs));
  reportResult("UDP stress",stressTest(tgt,devs,1000,0));
  reportResult("TCP/IP",tcpSyncTest(tgt,devs,CA_CONNECT));
  reportResult("TCP/IP stress",stressTest(tgt,devs,1000,CA_CONNECT));
  reportResult("STREAM",tcpSyncTest(tgt,devs,CA_STREAM));
  reportResult("STREAM stress",stressTest(tgt,devs,1000,CA_STREAM));
  reportResult("large UDP array", largeArraytest(tgt, 0));
  reportResult("large TCP/IP array", largeArraytest(tgt, CA_CONNECT));
  reportResult("large STREAM array", largeArraytest(tgt, CA_STREAM));
  reportResult("large UDP echo", largeEchoTest(tgt, 0));
  reportResult("large TCP/IP echo", largeEchoTest(tgt, CA_CONNECT));
  reportResult("Simple STRUCT",structTest(tgt,devs,"Simple"));
  reportResult("Nested STRUCT",structTest(tgt,devs,"Nested"));
  reportResult("Complex STRUCT",structTest(tgt,devs,"Complex"));
  reportResult("16-bit BITFIELD",bitfield16Test(tgt,devs));
  reportResult("32-bit BITFIELD",bitfield32Test(tgt,devs));
  reportResult("STRING",stringTest(tgt,devs));
  reportResult("SPECTRUM",spectrumTest(tgt,devs));
  reportResult("SPECTRUM array",spectrumArrayTest(tgt,devs));
  reportResult("GSPECTRUM", gspectrumTest(tgt, devs));
  reportResult("IMAGE test",imageTest(tgt,devs));
  reportResult("IMAGE Array",imageArrayTest(tgt,devs));
  reportResult("MDA",mdaTest(tgt,devs));
  reportResult("DEFAULT",defaultTest(tgt,devs));
  reportResult("WILDCARD",wildcardTest(tgt));
  reportResult("STOCK Properties",stkTest(tgt,devs));
  reportResult("Settings",settingsTest(tgt,devs));
  reportResult("Command", commandTest(tgt));
  if( ! IsStandAlone()) {
    reportResult("ENS query", ensQueryTest());
  } else {
    printf("running StandAlone skipping ensQueryTests\n");
  }
  reportResult("Local History",histTest(tgt,devs));
  reportResult("Link Error Test",linkErrorTest(tgt));
mcatest:
  linkactive = TRUE;
  reportResult("MCA LINK",mcaLinkTest(tgt,devs));

  while (linkactive)
  {
    millisleep(1000);
  }
  ec = ntotal == 0 ? 0 : 2;
  if( killserver) 
  {
    shutdownServer(tgt,devs);
  }
  if (!stayactive) 
  {
#   ifdef UNIX
    TineStopCycler();
    _SystemFreeResources(FALSE);
#   endif
    writeTestsuiteLogEnd();
    exit(ec);
  }

  while ((ch=getchar()) != 'q')
  {
    putchar(ch);
    switch (ch)
    {
      case 'r': goto mcatest;
      default:
        break;
    }
  }
  //SystemWaitCycleTimer();
  writeTestsuiteLogEnd();
  return ec;
}

float mcaSingleFloat = 0;
int mcaLnkId = -1;
time_t mcaLnkStart = 0;
int flipflop = -1;
int linkerr = 0;
void cbMcaFloat(int id,int cc)
{
  time_t t = time(NULL);
  int dst = GetDataStamp(id);
  int sst = GetSystemDataStamp(id);
  double ts = GetCurrentDataTimeStamp(id);
  char *lnkstr = GetLinkName(id);
  if (verbose) printf("%s: %g @ %s (%d %d) <status: %d>\n",
                 lnkstr,mcaSingleFloat,getDataTimeString(ts,0),sst,dst,cc);
  if (cc != 0) linkerr++;
  if (t > mcaLnkStart + 5) 
  {
    if (verbose) printf("Close MCA Single element link %.128s\n",lnkstr);
    CloseLink(id);
    linkactive = FALSE;
    reportResult("MCA LINK callback",linkerr);
  }
  flipflop = 0;
}
int mcaLinkTest(char *tgt,NAME64 *devs)
{
  DTYPE dout;
  char devstr[128];

  int nerr=0,cc;
  static int devnr = 0;

  devnr = (devnr+1)%32;
  DTYPEZERO(dout);

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &mcaSingleFloat;

  sprintf(devstr,"%s/%.64s",tgt,devs[devnr].name);
  flipflop = -1; linkerr = 0;
  if (verbose) printf("Start MCA Single element link %.128s[MCAFloatValues]\n",devstr);
  mcaLnkStart = time(NULL);
  mcaLnkId = AttachLink(devstr,"MCAFloatValues",&dout,NULL,CA_READ,1000,cbMcaFloat,CM_TIMER|CM_WAIT);
  if (flipflop != 0)
  {
    printf("callback not called dispite CM_WAIT!\n");
  }
  flipflop = 1;
  if (mcaLnkId < 0) 
  {
    nerr = 1;
    cc = -mcaLnkId;
  }
  return nerr;
}
int wildcardTest(char *tgt)
{
  DTYPE dout,d;
  char devstr[128];
  NAME64DBLDBL ndd[100];
  DUSTRING ust[100];
  int i,nerr=0,cc;
  UINT32 n;
  static int devnr = 0;

  devnr = (devnr+1)%32;
  DTYPEZERO(dout);

  dout.dFormat = CF_NAME64DBLDBL;
  dout.dArrayLength = 100;
  dout.data.vptr = ndd;

  sprintf(devstr,"%s/*",tgt);
  cc = ExecLink(devstr,"MCAFloatValues",&dout,NULL,CA_READ|CA_MUTABLE);
  if (verbose) printf("READ Wildcard %.128s[Image] as NAME64DBLDBL: %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr = 1;
  else
  {
    for (i=0; verbose > 1 && i<(int)dout.dArrayLength; i++)
    {
      printf("%.64s : %g <%d>\n",ndd[i].name,ndd[i].d1val,(int)ndd[i].d2val);
    }
  }
  dout.dFormat = CF_USTRING;
  dout.dArrayLength = 100;
  dout.data.vptr = ust;

  sprintf(devstr,"%s/*",tgt);
  cc = ExecLink(devstr,"MCAFloatValues",&dout,NULL,CA_READ|CA_MUTABLE);
  if (verbose) printf("READ Wildcard %.128s[Image] as USTRING: %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr = 1;
  else
  {
    DTYPEZERO(d);
    d.dFormat = CF_USTRING;
    d.dArrayLength = 1;
    for (i=0; verbose > 1 && i<(int)dout.dArrayLength; i++)
    {
      d.data.vptr = &ust[i];
      n = 1024; GetValuesAsString(&d,echoStr,&n);
      printf("%s\n",echoStr);
    }
  }
  return nerr;
}

int imageTest(char *tgt,NAME64 *devs)
{
  DTYPE dout, din;
  DIMAGE img, img_in;
  char devstr[128];
  BYTE framebuffer[1000], framebuffer_in[1024];
  int nerr=0,cc,i;
  static int devnr = 0;

  devnr = (devnr+1)%32;
  DTYPEZERO(dout);

  img.frameBuffer = framebuffer;
  img.frameBufferSize = 1000;
  dout.dFormat = CF_IMAGE;
  dout.dArrayLength = 1000;
  dout.data.vptr = &img;

  sprintf(devstr,"%s/%.64s",tgt,devs[devnr].name);
  cc = ExecLink(devstr,"Image",&dout,NULL,CA_READ);
  if (verbose) printf("READ IMAGE %.128s[Image] : %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr = 1;
  else
  {
    if (verbose > 1) printf("base tag: %d, port id: %d, source format: %d, frame size: %d\n",
      img.sourceHeader.baseTag,
      img.sourceHeader.cameraPortId,
      img.frameHeader.sourceFormat,
      img.frameHeader.appendedFrameSize);
  }
  DTYPEZERO(din);
  memcpy(&img_in, &img, IMAGE_HDR_SIZE);
  img_in.frameBuffer = framebuffer_in;
  img_in.frameBufferSize = 1000;

  for (i = 0; i < (int)img_in.frameBufferSize; i++) img_in.frameBuffer[i] = 22;
  din.dFormat = CF_IMAGE;
  din.dArrayLength = 1000;
  din.data.vptr = &img_in;
  cc = ExecLink(devstr, "Image", &dout, &din, CA_READ);
  if (verbose) printf("ECHO %.128s[Image] : %.32s\n", devstr, cc2str(cc));
  if (cc != 0) nerr++;
  for (i = 0; i < 1000; i++)
  {
    if (framebuffer[i] != framebuffer_in[i])
    {
      if (verbose) printf("ECHO %.128s[Image] mis-match %x vs. %x\n", devstr, framebuffer[i], framebuffer_in[i]);
      nerr++;
    }
  }

  return nerr;
}
int imageArrayTest(char *tgt,NAME64 *devs)
{
  DTYPE dout;
  DIMAGE img[3];
  char devstr[128];
  int i,nerr=0,cc;
  static int devnr = 0;

  devnr = (devnr+1)%32;
  DTYPEZERO(dout); 

  dout.dFormat = CF_AIMAGE;
  dout.dArrayLength = 3;
  dout.data.vptr = &img;

  sprintf(devstr,"%s/%.64s",tgt,devs[devnr].name);
  cc = ExecLink(devstr,"ImageArray",&dout,NULL,CA_READ|CA_MUTABLE);
  if (verbose) printf("READ IMAGE Array %.128s[ImageArray] : %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr = 1;
  else
  {
    for (i=0; verbose > 1 && i<3 && i<(int)dout.dArrayLength; i++)
    {
      printf("base tag: %d, port id: %d, source format: %d, frame size: %d\n",
        img[i].sourceHeader.baseTag,
        img[i].sourceHeader.cameraPortId,
        img[i].frameHeader.sourceFormat,
        img[i].frameHeader.appendedFrameSize);
    }
  }
  return nerr;
}

int mdaTest(char *tgt,NAME64 *devs)
{
  DTYPE dout,din;
  DMDA mda;
  char devstr[128];
  float fa[10][10][10];
  int i,j,k,nerr=0,cc;
  static int devnr = 0;

  devnr = (devnr+1)%32;
  DTYPEZERO(dout); DTYPEZERO(din);

  dout.dFormat = CF_MDA;
  dout.dArrayLength = 1000;
  dout.data.vptr = &mda;
  mda.data.fptr = (float *)fa;
  strcpy(dout.dTag,"FLOAT");

  sprintf(devstr,"%s/%.64s",tgt,devs[devnr].name);
  cc = ExecLink(devstr,"3DFloatArray",&dout,NULL,CA_READ);
  if (verbose) printf("READ MDA %.128s[3DFloatArray] : %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr = 1;
  else if (verbose > 1)
  {
    printf("cmt: %.80s\n",mda.cmt);
    printf("axes: %d (%s - %s) by %d (%s - %s) by %d (%s - %s)\n",
      mda.axis[0].len,mda.axis[0].label,mda.axis[0].units,
      mda.axis[1].len,mda.axis[1].label,mda.axis[0].units,
      mda.axis[2].len,mda.axis[2].label,mda.axis[0].units);
    printf("values: 1st 2 by 1st 2 by 1st 2\n");
    for (i=0; i<2; i++) for (j=0; j<2; j++) for (k=0; k<2; k++)
    {
      printf("[%d,%d,%d] : %g\n",i,j,k,
        mda.data.fptr[(i*mda.axis[2].len*mda.axis[1].len)+(j*mda.axis[2].len)+k]);
    }
  }
  return nerr;
}

DSPECTRUM dspec;
int spectrumTest(char *tgt,NAME64 *devs)
{
  DTYPE dout, din;
  int i,nerr=0,cc;
  UINT32 n;
  char devstr[128];
  float infloats[1024],outfloats[1024];
  static int devnr = 0;

  devnr = (devnr+1)%32;
  DTYPEZERO(dout); DTYPEZERO(din);
  din.dFormat = CF_SPECTRUM;
  din.dArrayLength = 1024;
  dout.dFormat = CF_SPECTRUM;
  dout.dArrayLength = 1024;

  din.data.vptr = &dspec;
  dout.data.vptr = &dspec;
  sprintf(devstr,"%s/%.64s",tgt,devs[devnr].name);

  cc = ExecLink(devstr,"Spectrum",&dout,NULL,CA_READ);
  if (verbose) printf("READ %.128s[Spectrum] : %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr++;
  else 
  {
    if (verbose > 1)
    {
      n = 1024; GetValuesAsString(&dout,echoStr,&n);
      printf("%s start SPECTRUM:\n %s\n",devstr,echoStr);
    }
    memcpy(outfloats,dspec.d_spect_array,1024*sizeof(float));
  }
  for (i=0; i<1024; i++) infloats[i] = (float)1.2345;
  memcpy(dspec.d_spect_array,infloats,1024*sizeof(float));
  cc = ExecLink(devstr,"Spectrum",&dout,&din,CA_READ|CA_WRITE);
  if (verbose) printf("WRITE/READ %.128s[Spectrum] : %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr++;
  else
  {
    if (memcmp(infloats,dspec.d_spect_array,1024*sizeof(float)))
    {
      printf("returned spectrum differs from input!");
      n = 1024; GetValuesAsString(&din,echoStr,&n);
      printf("%s send SPECTRUM:\n %s\n",devstr,echoStr);
      n = 1024; GetValuesAsString(&dout,echoStr,&n);
      printf("%s recv SPECTRUM:\n %s\n",devstr,echoStr);
    }
    memcpy(dspec.d_spect_array,outfloats,1024*sizeof(float));
  }
  if (verbose) printf("RESTORE %.128s[Spectrum] : %.32s\n",devstr,cc2str(cc));
  cc = ExecLink(devstr,"Spectrum",&dout,&din,CA_READ|CA_WRITE);
  if (cc != 0) nerr++;
  else
  {
    if (verbose > 1)
    {
      n = 1024; GetValuesAsString(&din,echoStr,&n);
      printf("%s send initial SPECTRUM:\n %s\n",devstr,echoStr);
    }
    if (memcmp(outfloats,dspec.d_spect_array,1024*sizeof(float)))
    {
      printf("returned spectrum differs from input!");
    }
  }
  return nerr;
}
int spectrumArrayTest(char *tgt,NAME64 *devs)
{
  DTYPE dout;
  int nerr=0,cc;
  UINT32 n;
  char devstr[128];
  static int devnr = 0;
  ASPECTRUM aspec[4];

  devnr = (devnr+1)%32;
  DTYPEZERO(dout);
  dout.dFormat = CF_ASPECTRUM;
  dout.dArrayLength = 4;

  dout.data.vptr = aspec;
  sprintf(devstr,"%s/%.64s",tgt,devs[devnr].name);

  cc = ExecLink(devstr,"Spectra",&dout,NULL,CA_READ);
  if (verbose) printf("READ SPECTRUM Array %.128s[Spectra] : %.32s\n",devstr,cc2str(cc));
  if (cc != 0) nerr++;
  else if (verbose > 1)
  {
    n = 1024; GetValuesAsString(&dout,echoStr,&n);
    printf("%s read SPECTRUM arrays:\n %s\n",devstr,echoStr);
  }
  return nerr;
}
DGSPECTRUM gspec;
int gspectrumTest(char *tgt, NAME64 *devs)
{
  DTYPE dout, din;
  int i, nerr = 0, cc;
  UINT32 n;
  char devstr[128];
  float outfloats[1024], infloats[1024];
  DGSPECTRUM gspec_in;
  static int devnr = 0;

  devnr = (devnr + 1) % 32;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_GSPECTRUM;
  dout.dArrayLength = 1024;
  gspec.spectBuffer = outfloats;
  gspec.spectBufferSize = 1024;
  dout.data.vptr = &gspec;
  sprintf(devstr, "%s/%.64s", tgt, devs[devnr].name);

  cc = ExecLink(devstr, "G-Spectrum", &dout, NULL, CA_READ);
  if (verbose) printf("READ %.128s[G-Spectrum] : %.32s\n", devstr, cc2str(cc));
  if (cc != 0) nerr++;
  else
  {
    if (verbose > 1)
    {
      n = 1024; GetValuesAsString(&dout, echoStr, &n);
      printf("%s start GSPECTRUM:\n %s\n", devstr, echoStr);
    }
  }
  memcpy(&gspec_in, &gspec, GSPECTRUM_HDR_SIZE);
  gspec_in.spectBuffer = infloats;
  gspec_in.spectBufferSize = 1024;

  for (i = 0; i < (int)gspec_in.spectBufferSize; i++) gspec_in.spectBuffer[i] = 88;
  din.dFormat = CF_GSPECTRUM;
  din.dArrayLength = 1024;
  din.data.vptr = &gspec_in;
  cc = ExecLink(devstr, "G-Spectrum", &dout, &din, CA_READ);
  if (verbose) printf("ECHO %.128s[G-Spectrum] : %.32s\n", devstr, cc2str(cc));
  if (cc != 0) nerr++;
  for (i = 0; i < 1024; i++)
  {
    if (outfloats[i] != infloats[i])
    {
      if (verbose) printf("ECHO %.128s[G-Spectrum] mis-match %g vs. %g\n", devstr, outfloats[i],infloats[i]);
      nerr++;
    }
  }

  return nerr;
}

char *tstaddendum = " + a bunch of bytes to make the corresponding string really long, i.e. much longer than the standard 64 characters or 80 characters";
int stringTest(char *tgt,NAME64 *devs)
{
  DTYPE dout, din;
  int i,nerr=0,cc;
  time_t t;
  char devstr[128],tststr[256], *instr=tststr, *outstr = NULL, *rbstr[32];

  DTYPEZERO(dout); DTYPEZERO(din);
  din.dFormat = CF_STRING;
  din.dArrayLength = 1;
  dout.dFormat = CF_STRING;
  dout.dArrayLength = 1;

  for (i=0; i<32; i++)
  {
    t = time(NULL);
    sprintf(tststr,"unit test string from %.24s",ctime(&t));
    if (i%2) strcat(tststr,tstaddendum);
    din.data.strptr = &instr;
    dout.data.strptr = &outstr;
    sprintf(devstr,"%s/%.64s",tgt,devs[(i)%32].name);
    cc = ExecLink(devstr,"StringArray",&dout,&din,CA_READ|CA_WRITE);
    if (verbose) printf("set/get %s[StringArray] : %.32s\n",devstr,cc2str(cc));
    if (cc == 0)
    {
      if (strcmp(instr,outstr)) 
      {
        printf("sent:\n\t%s\n",instr);
        printf("returned string differs from input!:\n\t%s\n",outstr);
        nerr++;
      }
    }
  }

  dout.dArrayLength = 32;
  dout.data.strptr = rbstr;
  sprintf(devstr,"%s/%.64s",tgt,devs[0].name);
  cc = ExecLink(devstr,"StringArray",&dout,NULL,CA_READ);
  if (verbose) printf("get ALL strings %s[StringArray] : %.32s\n",devstr,cc2str(cc));

  return nerr;
}
int structTest(char *tgt,NAME64 *devs,char *tag)
{
  struct structStruct *ss;
  struct structFormat *sf;
  DTYPE dout, din;
  int i=0,nerr=0,cc,ntries=3;
  UINT32 n;
  char devstr[128],prpstr[64];

  DTYPEZERO(dout); DTYPEZERO(din);
  cc = AcquireAndRegisterStruct(tgt,tag,32);
  if (verbose) printf("acquire '%.16s' Struct: %.32s\n",tag,cc2str(cc));
  if ((ss=FindStruct(tag)) == NULL) return 1;

  sprintf(devstr,"%s/%.64s",tgt,devs[(i++)%32].name);
  sprintf(prpstr,"%sStruct",tag);
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strncpy(dout.dTag,tag,TAG_NAME_SIZE);
  dout.data.vptr = blobOut;
  cc = ExecLink(devstr,prpstr,&dout,NULL,CA_READ);
  if (cc != 0)
  {
    if (verbose) printf("%.128s[%.64s]: %.32s\n", devstr,prpstr,cc2str(cc));
    return 1;
  }
  cc = GetStructureAsString(tag,blobOut,echoStr,1024);
  if (cc == 0 && verbose > 1) printf("%s\n",echoStr);

  for (sf=ss->s; sf != NULL; sf=sf->nxt)
  {
    if (sf->fmt == CF_STRUCT) continue; /* skip these */
try_again:
    dout.dFormat = sf->fmt;
    dout.dArrayLength = sf->siz;
    sprintf(devstr,"%s/%.64s",tgt,devs[(i++)%32].name);
    sprintf(prpstr,"%sStruct.%.16s",tag,sf->field);
    dout.data.vptr = blobOut;
    cc = ExecLink(devstr,prpstr,&dout,NULL,CA_READ);
    if (cc != 0 && ntries-- > 0) 
    {
      goto try_again;
    }
    if (verbose) printf("acquire %s[%s] : %.32s\n",devstr,prpstr,cc2str(cc));
    if (cc == 0)
    {
      if (verbose > 1)
      {
        n = 1024; GetValuesAsString(&dout,echoStr,&n);
        printf("recv:\n\t%.1024s\n",echoStr);
      }
    }
    else
    {
      nerr++;
    }
  }
  return nerr;
}
int bitfield16Test(char *tgt,NAME64 *devs)
{
  DBITFIELD *bf;
  struct bitfieldFormat *bfs;
  DTYPE dout, din;
  int i=0,nerr=0,cc;
  UINT32 n;
  char devstr[128],prpstr[64],*tag;

  DTYPEZERO(dout); DTYPEZERO(din);
  tag = "bf16";
  cc = AcquireAndRegisterBitfield(tgt,tag);
  if (verbose) printf("acquire '%.16s' bitfield: %.32s\n",tag,cc2str(cc));
  if ((bf=findBitField(NULL,tag)) == NULL) return 1;

  sprintf(devstr,"%s/%.64s",tgt,devs[(i++)%32].name);
  dout.dFormat = bf->fmt;
  dout.dArrayLength = 1;
  strncpy(dout.dTag,tag,TAG_NAME_SIZE);
  dout.data.vptr = blobOut;
  cc = ExecLink(devstr,"Bitfield16Values",&dout,NULL,CA_READ);
  if (cc != 0)
  {
    if (verbose) printf("%.128s[Bitfield16Values]: %32s\n", devstr, cc2str(cc));
    return 1;
  }
  if (verbose) printf("received: 0x%x\n",dout.data.sptr[0]);
  cc = GetBitfieldAsString(NULL,tag,dout.data.sptr[0],echoStr,1024);
  if (cc != 0) nerr++; else if (verbose) printf("%s\n",echoStr);

  for (bfs=bf->bfs; bfs != NULL; bfs=bfs->nxt)
  {
    dout.dFormat = CF_INT16;
    dout.dArrayLength = 1;
    sprintf(devstr,"%s/%.64s",tgt,devs[(i++)%32].name);
    sprintf(prpstr,"Bitfield16Values.%.16s",bfs->field);
    dout.data.vptr = blobOut;
    cc = ExecLink(devstr,prpstr,&dout,NULL,CA_READ);
    if (verbose) printf("acquire %s[%s] : %.32s\n",devstr,prpstr,cc2str(cc));
    if (cc == 0)
    {
      if (verbose > 1)
      {
        n = 1024; GetValuesAsString(&dout,echoStr,&n);
        printf("recv:\n\t%.1024s\n",echoStr);
      }
    }
    else
    {
      nerr++;
    }
  }
  return nerr;
}
int bitfield32Test(char *tgt,NAME64 *devs)
{
  DBITFIELD *bf;
  struct bitfieldFormat *bfs;
  DTYPE dout, din;
  int i=0,nerr=0,cc;
  UINT32 n;
  char devstr[128],prpstr[64],*tag;

  DTYPEZERO(dout); DTYPEZERO(din);
  tag = "bf32";
  cc = AcquireAndRegisterBitfield(tgt,tag);
  if (verbose) printf("acquire '%.16s' bitfield: %.32s\n",tag,cc2str(cc));
  if ((bf=findBitField(NULL,tag)) == NULL) return 1;

  sprintf(devstr,"%s/%.64s",tgt,devs[(i++)%32].name);
  dout.dFormat = bf->fmt;
  dout.dArrayLength = 1;
  strncpy(dout.dTag,tag,TAG_NAME_SIZE);
  dout.data.vptr = blobOut;
  cc = ExecLink(devstr,"Bitfield32Values",&dout,NULL,CA_READ);
  if (cc != 0)
  {
    if (verbose) printf("%.128s[Bitfield32Values]: %32s\n", devstr,cc2str(cc));
    return 1;
  }
  if (verbose) printf("received: 0x%x\n", dout.data.lptr[0]);
  cc = GetBitfieldAsString(NULL,tag,dout.data.lptr[0],echoStr,1024);
  if (cc != 0) nerr++; else if (verbose) printf("%s\n",echoStr);

  for (bfs=bf->bfs; bfs != NULL; bfs=bfs->nxt)
  {
    dout.dFormat = CF_INT32;
    dout.dArrayLength = 1;
    sprintf(devstr,"%s/%.64s",tgt,devs[(i++)%32].name);
    sprintf(prpstr,"Bitfield32Values.%.16s",bfs->field);
    dout.data.vptr = blobOut;
    cc = ExecLink(devstr,prpstr,&dout,NULL,CA_READ);
    if (verbose) printf("acquire %s[%s] : %.32s\n",devstr,prpstr,cc2str(cc));
    if (cc == 0)
    {
      if (verbose > 1)
      {
        n = 1024; GetValuesAsString(&dout,echoStr,&n);
        printf("recv:\n\t%.1024s\n",echoStr);
      }
    }
    else
    {
      nerr++;
    }
  }
  return nerr;
}
char *randomString(int length)
{
  static char s[256];
  int i;
  if (length > 256) length = 256;
  for (i=0; i<length; i++) s[i] = (char)(65 + (int)random(57));

  return s;
}
int pushRandomInputData(DTYPE *d)
{
  int i;
  int siz = d->dArrayLength * fmtsizeof(d->dFormat);
  if (siz > BLOB_SIZE) return -buffer_too_small;
  for (i=0; i<(int)d->dArrayLength; i++)
  {
    switch (d->dFormat)
    {
      case CF_DOUBLE: d->data.dptr[i] = (double)random(1000); break;
      case CF_FLOAT: d->data.fptr[i] = (float)random(1000); break;
      case CF_INT16: d->data.sptr[i] = (short)random(1000); break;
      case CF_BOOLEAN:
      case CF_INT32: d->data.lptr[i] = (int)random(1000); break;
      case CF_INT64: d->data.llptr[i] = (SINT64)random(1000); break;
      case CF_BYTE: d->data.bptr[i] = (BYTE)random(256); break;
      case CF_TEXT: 
      case CF_XML:
        strncpy(d->data.cptr,randomString(d->dArrayLength),d->dArrayLength);
        return 0;
      case CF_NAME8:
        strncpy(((NAME8 *)d->data.vptr)[i].name,randomString(8),8);
        break;
      case CF_NAME16:
        strncpy(((NAME16 *)d->data.vptr)[i].name,randomString(16),16);
        break;
      case CF_NAME32:
        strncpy(((NAME32 *)d->data.vptr)[i].name,randomString(32),32);
        break;
      case CF_NAME48:
        strncpy(((NAME48 *)d->data.vptr)[i].name,randomString(48),48);
        break;
      case CF_NAME64:
        strncpy(((NAME64 *)d->data.vptr)[i].name,randomString(64),64);
        break;
      case CF_NAME8I:
        strncpy(((NAME8I *)d->data.vptr)[i].name,randomString(8),8);
        ((NAME8I *)d->data.vptr)[i].ival = (int)random(1000);
        break;
      case CF_NAME16I:
        strncpy(((NAME16I *)d->data.vptr)[i].name,randomString(16),16);
        ((NAME16I *)d->data.vptr)[i].ival = (int)random(1000);
        break;
      case CF_NAME32I:
        strncpy(((NAME32I *)d->data.vptr)[i].name,randomString(32),32);
        ((NAME32I *)d->data.vptr)[i].ival = (int)random(1000);
        break;
      case CF_NAME48I:
        strncpy(((NAME48I *)d->data.vptr)[i].name,randomString(48),48);
        ((NAME48I *)d->data.vptr)[i].ival = (int)random(1000);
        break;
      case CF_NAME64I:
        strncpy(((NAME64I *)d->data.vptr)[i].name,randomString(64),64);
        ((NAME64I *)d->data.vptr)[i].ival = (int)random(1000);
        break;
      case CF_NAME64TIME:
        strncpy(((NAME64DBL *)d->data.vptr)[i].name,randomString(64),64);
        ((NAME64DBL *)d->data.vptr)[i].dval = (double)random(1000);
        break;
      case CF_NAME64DBL:
        strncpy(((NAME64DBL *)d->data.vptr)[i].name,randomString(64),64);
        ((NAME64DBL *)d->data.vptr)[i].dval = (double)random(1000);
        break;
      case CF_FLTINT:
        ((FLTINT *)d->data.vptr)[i].ival = (int)random(1000);
        ((FLTINT *)d->data.vptr)[i].fval = (float)random(1000);
        break;
      case CF_INTINT:
        ((INTINT *)d->data.vptr)[i].i1val = (int)random(1000);
        ((INTINT *)d->data.vptr)[i].i2val = (int)random(1000);
        break;
      case CF_FLTFLT:
        ((FLTFLT *)d->data.vptr)[i].f1val = (float)random(1000);
        ((FLTFLT *)d->data.vptr)[i].f2val = (float)random(1000);
        break;
      case CF_DBLTIME:
      case CF_DBLDBL:
        ((DBLDBL *)d->data.vptr)[i].d1val = (double)random(1000);
        ((DBLDBL *)d->data.vptr)[i].d2val = (double)random(1000);
        break;
      case CF_NAME16DBLDBL:
        strncpy(((NAME16DBLDBL *)d->data.vptr)[i].name,randomString(16),16);
        ((NAME16DBLDBL *)d->data.vptr)[i].d1val = (double)random(1000);
        ((NAME16DBLDBL *)d->data.vptr)[i].d2val = (double)random(1000);
        break;
      case CF_NAME32DBLDBL:
        strncpy(((NAME32DBLDBL *)d->data.vptr)[i].name,randomString(32),32);
        ((NAME32DBLDBL *)d->data.vptr)[i].d1val = (double)random(1000);
        ((NAME32DBLDBL *)d->data.vptr)[i].d2val = (double)random(1000);
        break;
      case CF_NAME64DBLDBL:
        strncpy(((NAME64DBLDBL *)d->data.vptr)[i].name,randomString(64),64);
        ((NAME64DBLDBL *)d->data.vptr)[i].d1val = (double)random(1000);
        ((NAME64DBLDBL *)d->data.vptr)[i].d2val = (double)random(1000);
        break;
      case CF_NAME16FI:
        strncpy(((NAME16FI *)d->data.vptr)[i].name,randomString(16),16);
        ((NAME16FI *)d->data.vptr)[i].fval = (float)random(1000);
        ((NAME16FI *)d->data.vptr)[i].ival = (int)random(1000);
        break;
      case CF_NAME16II:
        strncpy(((NAME16II *)d->data.vptr)[i].name,randomString(16),16);
        ((NAME16II *)d->data.vptr)[i].i1val = (int)random(1000);
        ((NAME16II *)d->data.vptr)[i].i2val = (int)random(1000);
        break;
      case CF_INTFLTINT:
        ((INTFLTINT *)d->data.vptr)[i].i1val = (int)random(1000);
        ((INTFLTINT *)d->data.vptr)[i].fval = (float)random(1000);
        ((INTFLTINT *)d->data.vptr)[i].i2val = (int)random(1000);
        break;
      case CF_FLTFLTINT:
        ((FLTFLTINT *)d->data.vptr)[i].f1val = (float)random(1000);
        ((FLTFLTINT *)d->data.vptr)[i].f2val = (float)random(1000);
        ((FLTFLTINT *)d->data.vptr)[i].ival = (int)random(1000);
        break;
      case CF_FLTINTINT:
        ((FLTINTINT *)d->data.vptr)[i].fval = (float)random(1000);
        ((FLTINTINT *)d->data.vptr)[i].i1val = (int)random(1000);
        ((FLTINTINT *)d->data.vptr)[i].i2val = (int)random(1000);
        break;
      case CF_INTINTINT:
        ((INTINTINT *)d->data.vptr)[i].i1val = (int)random(1000);
        ((INTINTINT *)d->data.vptr)[i].i2val = (int)random(1000);
        ((INTINTINT *)d->data.vptr)[i].i3val = (int)random(1000);
        break;
      case CF_DBLDBLDBL:       
        ((DBLDBLDBL *)d->data.vptr)[i].d1val = (double)random(1000);
        ((DBLDBLDBL *)d->data.vptr)[i].d2val = (double)random(1000);
        ((DBLDBLDBL *)d->data.vptr)[i].d3val = (double)random(1000);
        break;
      case CF_INTFLTFLTFLT:
        ((DFILTER *)d->data.vptr)[i].ival = (int)random(1000);
        ((DFILTER *)d->data.vptr)[i].f1val = (float)random(1000);
        ((DFILTER *)d->data.vptr)[i].f2val = (float)random(1000);
        ((DFILTER *)d->data.vptr)[i].f3val = (float)random(1000);
        break;
      case CF_TTII:
      case CF_INTINTINTINT:
        ((DADDRESS *)d->data.vptr)[i].i1val = (int)random(1000);
        ((DADDRESS *)d->data.vptr)[i].i2val = (int)random(1000);
        ((DADDRESS *)d->data.vptr)[i].i3val = (int)random(1000);
        ((DADDRESS *)d->data.vptr)[i].i4val = (int)random(1000);
        break;
      case CF_INTINTFLTFLT:
        ((FWINDOW *)d->data.vptr)[i].i1val = (int)random(1000);
        ((FWINDOW *)d->data.vptr)[i].i2val = (int)random(1000);
        ((FWINDOW *)d->data.vptr)[i].f1val = (float)random(1000);
        ((FWINDOW *)d->data.vptr)[i].f2val = (float)random(1000);
        break;
      case CF_FLTINTFLTINT:
        ((FLTINTFLTINT *)d->data.vptr)[i].f1val = (float)random(1000);
        ((FLTINTFLTINT *)d->data.vptr)[i].i1val = (int)random(1000);
        ((FLTINTFLTINT *)d->data.vptr)[i].f2val = (float)random(1000);
        ((FLTINTFLTINT *)d->data.vptr)[i].i2val = (int)random(1000);
        break;
      case CF_NAME64DBLDBLDBL:
        strncpy(((NAME64DBLDBLDBL *)d->data.vptr)[i].name,randomString(64),64);
        ((NAME64DBLDBLDBL *)d->data.vptr)[i].d1val = (double)random(1000);
        ((NAME64DBLDBLDBL *)d->data.vptr)[i].d2val = (double)random(1000);
        ((NAME64DBLDBLDBL *)d->data.vptr)[i].d3val = (double)random(1000);
        break;
      case CF_INTFLTFLTFLTNAME16:
        ((DUNAME *)d->data.vptr)[i].ival = (int)random(1000);
        ((DUNAME *)d->data.vptr)[i].f1val = (float)random(1000);
        ((DUNAME *)d->data.vptr)[i].f2val = (float)random(1000);
        ((DUNAME *)d->data.vptr)[i].f3val = (float)random(1000);
        strncpy(((DUNAME *)d->data.vptr)[i].name,randomString(64),64);
        break;
      case CF_USTRING:
        ((DUSTRING *)d->data.vptr)[i].ival = (int)random(1000);
        ((DUSTRING *)d->data.vptr)[i].f1val = (float)random(1000);
        ((DUSTRING *)d->data.vptr)[i].f2val = (float)random(1000);
        ((DUSTRING *)d->data.vptr)[i].tm = (int)random(1000);
        strncpy(((DUSTRING *)d->data.vptr)[i].str,randomString(80),80);
        break;
        break;
    }
  }
  return siz;
}
int isDeprectated(short fmt)
{
  switch (fmt)
  {
    case CF_BIT:
    case CF_HITTS:
    case CF_DBLINT:
    case CF_SHTINT:
    case CF_INTDBL:
    case CF_FLTDBL:
    case CF_INTFLTSHTSHT:
      return TRUE;
  }
  return FALSE;
}
int stressTest(char *tgt,NAME64 *devs,int nloops,short flags)
{
  int cc=0, n, tt, tallowed, nerr=0, acc = CA_READ;
  char devstr[128];
  DTYPE dout, din;
  struct timeval tv1, tv2;
  double dval = 1.234;

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = din.dFormat = CF_DOUBLE;
  dout.dArrayLength = din.dArrayLength = 1;
  dout.data.dptr = &dval;
  din.data.dptr = &dval;
  sprintf(devstr,"%s/%.64s",tgt,devs[0].name);
  if (flags != 0) acc = CA_READ|flags;
  gettimeofday(&tv1,(struct timezone *)NULL);
  for (n=0; n<nloops; n++)
  {
    cc = ExecLinkEx(devstr,"ECHO",&dout,&din,(short)acc,500);
    if (cc == 0)
    {
      if (memcmp(blobOut,blobIn,sizeof(double))) 
      {
        printf("echoed data does not match\n");
        nerr++;
      }
    }
    else
    {
      printf("called failed: %.32s <%d>\n",cc2str(cc),cc);
      nerr++;
    }
  }
  gettimeofday(&tv2,(struct timezone *)NULL);
  tt = (tv2.tv_sec - tv1.tv_sec) * 1000 + (tv2.tv_usec - tv1.tv_usec) / 1000;
  if (verbose) 
  {
    char *txt = flags == CA_CONNECT ? "TCP" : flags == CA_STREAM ? "STREAM" : "UDP";
    printf("stress test (%s) %d calls total time %d msec\n",txt,nloops,tt);
  }
  tallowed = 2 * nloops;
  if (tt > tallowed) nerr++;
  return nerr;
}
int tcpSyncTest(char *tgt,NAME64 *devs,int flags)
{
  int cc=0, dsiz, nerr=0;
  UINT32 n = 1024;
  char devstr[128];
  DTYPE dout, din;

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = din.dFormat = CF_DOUBLE;
  dout.dArrayLength = din.dArrayLength = 1024;
  dout.data.vptr = blobOut;
  din.data.vptr = blobIn;
  if ((dsiz=pushRandomInputData(&din)) < 0) cc = -dsiz;
  sprintf(devstr,"%s/%.64s",tgt,devs[0].name);
  if (cc != 0) ccerr(cc);
  memset(blobOut,0,dsiz);
  cc = ExecLink(devstr,"ECHO",&dout,&din,(short)(CA_READ|flags));
  if (cc == 0)
  {
    if (memcmp(blobOut,blobIn,dsiz)) 
    {
      printf("%.128s: echo %d double values\n",devstr,n);
      n = 1024; GetValuesAsString(&din,echoStr,&n);
      printf("sent:\n\t%.1024s\n",echoStr);
      printf("input bytes not echoed correctly!\n");
      n = 1024; GetValuesAsString(&dout,echoStr,&n);
      printf("recv:\n\t%.1024s\n",echoStr);
      nerr++;
    }
  }
err:
  if (cc != 0)
  {
    printf("%.128s: echo 1024 double values (%d bytes) via TCP/IP: %.32s\n",devstr,dsiz,cc2str(cc));
    nerr++;
  }
  return nerr;
}
int histTest(char *tgt,NAME64 *devs)
{
  DTYPE din,dout;
  char devstr[128];
  UINT32 startstopArray[3];
  int nerr=0, n, cc;

  DTYPEZERO(dout); DTYPEZERO(din); 
  startstopArray[1] = (UINT32)time(NULL);
  startstopArray[0] = (UINT32)(startstopArray[1]-86400*30);
  dout.dArrayLength = 1000;
  dout.dFormat = CF_DBLDBL;
  dout.data.vptr = blobOut;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;
  n = random(32);
  sprintf(devstr,"%s/%.64s",tgt,devs[n%nUnitDevices].name);
  if ((cc=ExecLinkEx(devstr,"Setting.HIST",&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,200)) != 0) 
  {
    printf("Settings.HIST call to %.128s : %.32s\n",devstr,cc2str(cc));
    nerr++;
  }

  dout.dArrayLength = nUnitDevices;
  dout.dFormat = CF_INT32;
  din.dArrayLength = 1;
  sprintf(devstr,"%s/#0",tgt);
  if ((cc=ExecLinkEx(devstr,"Setting.HIST@",&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,200)) != 0) 
  {
    printf("Settings.HIST@ snapshot call to %.128s : %.32s\n",devstr,cc2str(cc));
    nerr++;
  }
  return nerr;
}
int ensQueryTest(void)
{
  UINT32 i;
  int cc;
  char tgt[132];
  NAME32 srvlst[1000];
  DTYPE dout;

  DTYPEZERO(dout);
  dout.dArrayLength = 1000;
  dout.dFormat = CF_NAME32;
  dout.data.vptr = srvlst;
  sprintf(tgt, "/SITE/ENS/%.32s", tgtContext);
  cc = ExecLink(tgt, "SERVERS", &dout, NULL, CA_READ);
  if (cc != 0)
  {
    printf("ENS query call to get SERVERS : %.32s\n", cc2str(cc));
    return 1;
  }
  for (i = 0; i < dout.dArrayLength; i++)
  {
    if (!strnicmp(srvlst[i].name, tgtServer, 32)) return 0;
  }
  printf("ENS SERVERS query call could not find unit server %.32s\n", tgtServer);
  return 1;
}
int commandTest(char *tgt)
{
  int cc;
  char devstr[128];
  sprintf(devstr, "%s/#0", tgt);
  cc = ExecLink(devstr, "Command", NULL, NULL, CA_WRITE);
  if (cc != 0)
  {
    printf("Command call to %.128s : %.32s\n", devstr, cc2str(cc));
    return 1;
  }
  return 0;
}
int settingsTest(char *tgt,NAME64 *devs)
{
  int i, cc=0, dsiz, nerr=0, v;
  char devstr[128];
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = din.dFormat = CF_INT32;
  dout.dArrayLength = nUnitDevices;
  din.dArrayLength = 1;
  for (i=0; i<nUnitDevices; i++)
  {
    dout.data.vptr = blobOut;
    din.data.vptr = &v;
    if ((dsiz=pushRandomInputData(&din)) < 0) cc = -dsiz;
    sprintf(devstr,"%s/%.64s",tgt,devs[i].name);
    if (cc != 0) ccerr(cc);
    cc = ExecLink(devstr,"Setting",&dout,&din,CA_READ|CA_WRITE);
err:
    if (cc != 0)
    {
      printf("Settings call to %.128s : %.32s\n",devstr,cc2str(cc));
      nerr++;
    }
  }
  return nerr;
}
int stkTest(char *tgt,NAME64 *devs)
{
  int i, cc=0, dsiz, nerr=0;
  char devstr[128];
  DTYPE dout, din;
  for (i=0; i<gNrStockProperties; i++)
  {
    DTYPEZERO(dout); DTYPEZERO(din);
    if (gStockProperty[i].prpAccessMode & CA_WRITE) continue; // skip these for now
    if (gStockProperty[i].prpSizeIn > 0) continue; // skip these for now
    if (gStockProperty[i].isFecProperty)
    { // device name is irrelevant
      sprintf(devstr,"%s",tgt);
    }
    else
    {
      sprintf(devstr,"%s/%.64s",tgt,devs[i%32].name);
    }
    dout.dFormat = gStockProperty[i].prpFormat;
    dout.dArrayLength = gStockProperty[i].prpSize;
    strncpy(dout.dTag,gStockProperty[i].prpStructTag,TAG_NAME_SIZE);  
    dout.data.bptr = blobOut;
    dsiz = dout.dArrayLength * getFormatSize(dout.dFormat) + getFormatHeaderSize(dout.dFormat);
    if (dout.dFormat == CF_STRUCT && strlen(gStockProperty[i].prpStructTag) > 0) 
    {
      dsiz = dout.dArrayLength * GetStructSize(gStockProperty[i].prpStructTag);
    }
    if (dsiz > BLOB_SIZE) dout.dArrayLength = 1; // just punt
    cc = ExecLink(devstr,gStockProperty[i].prpName,&dout,&din,CA_READ);
    if (cc != 0)
    {
      char fstr[32];
      if (cc == un_allocated || cc == ipx_not_supported || cc == no_such_file) continue;
      if (cc == file_error) continue; // returned by SRVLOGFILE when there's no file 
      if (cc == illegal_property && !strcmp(gStockProperty[i].prpName,"SRVNOTPOSTED")) continue; // not a WINDOWS server
      strncpy(fstr,GetFormatAsString(dout.dFormat),32);
      if (dout.dFormat == CF_STRUCT) 
      {
        strcat(fstr,"."); strncat(fstr,dout.dTag,TAG_NAME_SIZE);
      }
      printf("Stock property call %.128s[%.64s] %d %.32s values: %.32s\n",
        devstr,gStockProperty[i].prpName,dout.dArrayLength,fstr,cc2str(cc));
      nerr++;
    }   
  }
  return nerr;
}
int echoTest(char *tgt,NAME64 *devs)
{
  int cc=0, dsiz, nerr=0;
  UINT32 n;
  short i;
  char fstr[32],devstr[128];
  DTYPE dout, din;

  DTYPEZERO(dout); DTYPEZERO(din);
  for (i=CF_DOUBLE; i < CF_MDA; i++)
  {
    if (IsTaggableFormat(i)) continue; /* don't do these */
    if (getFormatHeaderSize(i) > 0) continue; /* or these */
    if (isDeprectated(i)) continue; /* or these */
    n = random(9) + 1;
    strncpy(fstr,GetFormatAsString(i),32);
    dout.dFormat = din.dFormat = i;
    dout.dArrayLength = din.dArrayLength = n;
    dout.data.vptr = blobOut;
    din.data.vptr = blobIn;
    if ((dsiz=pushRandomInputData(&din)) < 0) cc = -dsiz;
    sprintf(devstr,"%s/%.64s",tgt,devs[i%32].name);
    if (cc != 0) ccerr(cc);
    memset(blobOut,0,dsiz);
    cc = ExecLink(devstr,"ECHO",&dout,&din,CA_READ);
    if (cc == 0)
    {
      if (memcmp(blobOut,blobIn,dsiz)) 
      {
        printf("%.128s: echo %d %.16s values\n",devstr,n,fstr);
        n = 1024; GetValuesAsString(&din,echoStr,&n);
        printf("sent:\n\t%.1024s\n",echoStr);
        printf("input bytes not echoed correctly!\n");
        n = 1024; GetValuesAsString(&dout,echoStr,&n);
        printf("recv:\n\t%.1024s\n",echoStr);
        nerr++;
      }
    }
err:
    if (cc != 0)
    {
      printf("%.128s: echo %d %.16s values (%d bytes): %.32s\n",devstr,n,fstr,dsiz,cc2str(cc));
      nerr++;
    }
  }
  return nerr;
}

int getPropertyAsDefault(char *dev,char *prp)
{
  DTYPE dout, din;
  int cc=0;
  UINT32 n=0;
  DMDA mda;

  DTYPEZERO(dout); DTYPEZERO(din);

  dout.dFormat = CF_DEFAULT;
  dout.dArrayLength = 64;
  dout.data.vptr = blobOut;
  if (verbose) printf("get %.128s[%.64s] as DEFAULT type\n",dev,prp);
  memset(blobOut,0,64);
acqPrpData:
  cc = ExecLink(dev,prp,&dout,NULL,CA_READ|CA_MUTABLE);
  if (cc == 0)
  {
    if (verbose) printf("returned %d %.16s values\n",dout.dArrayLength,GetFormatAsString(dout.dFormat));
    if (verbose > 1)
    {
      n = 1024; GetValuesAsString(&dout,echoStr,&n);
      printf("got:\n\t%.1024s\n",echoStr);
    }
  }
  else
  {
    int fmt, siz, dsiz;
    switch (cc)
    {
      case buffer_too_small:
        cc = GetTargetPropertyInformation(dev,prp,&fmt,&siz,NULL);
        if (cc == 0)
        {
          if (verbose) printf("recast request to get %d %.16s values\n",siz,GetFormatAsString(fmt));
          dsiz = siz * fmtsizeof(fmt);
          if (dsiz > BLOB_SIZE)
          {
            printf("required data size is %d bytes !\n",dsiz);
            break;
          }
          dout.dFormat = fmt;
          dout.dArrayLength = siz;
          goto acqPrpData;
        }
        break;
      case dimension_error:
        if (n++ > 0) break;
        if (dout.dFormat == CF_MDA)
        { /* the CA_MUTABLE flag will allow this to have changed */
          if (verbose) printf("recast request to get MDA\n");
          mda.data.bptr = blobOut;
          dout.data.vptr = &mda;
          dout.dArrayLength = 4000;
          goto acqPrpData;        
        }
        break;
      default:
        printf("returned %.32s\n",cc2str(cc));
        break;
    }
  }
  return cc;
}

int defaultTest(char *tgt,NAME64 *devs)
{
  int cc=0, nerr=0;
  static int devnr = 0;
  char devstr[128];
  DTYPE dout, din;

  DTYPEZERO(dout); DTYPEZERO(din);
  devnr = (devnr+1)%32;
  sprintf(devstr,"%s/%.64s",tgt,devs[devnr].name);
  if ((cc=getPropertyAsDefault(devstr,"3DFloatArray")) != 0) nerr++;
  if ((cc=getPropertyAsDefault(devstr,"MCAFloatValues")) != 0) nerr++;
  /* if ((cc=getPropertyAsDefault(devstr,"SimpleStruct")) != 0) nerr++;  << only works under some circumstances */
  if ((cc=getPropertyAsDefault(devstr,"StringArray")) != 0) nerr++;
  if ((cc=getPropertyAsDefault(devstr,"Spectrum")) != 0) nerr++;
  if ((cc=getPropertyAsDefault(devstr,"TraceFloatValues")) != 0) nerr++;
  if ((cc=getPropertyAsDefault(devstr,"BooleanValues")) != 0) nerr++;
  if ((cc=getPropertyAsDefault(devstr,"KeyValuePairs")) != 0) nerr++;

  return nerr;
}
int largeArraytest(char *tgt, short flags)
{
  DTYPE din, dout;
  char devstr[128];
  int cc, acc = CA_READ|CA_MUTABLE;

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = 64000;
  dout.dFormat = CF_DOUBLE;
  dout.data.vptr = blobOut;
  sprintf(devstr, "%s/#0", tgt);
  if (flags != 0) acc = CA_READ | flags;
  if ((cc=ExecLinkEx(devstr, "DoubleArrayFloatValues", &dout, &din, acc, 2000)) != 0)
  {
    printf("largeArray call to %.128s : %.32s\n", devstr, cc2str(cc));
  }
  return cc ? 1 : 0;
}
int largeEchoTest(char *tgt, short flags)
{
  DTYPE din, dout;
  char devstr[128];
  int i, cc, acc = CA_READ | CA_MUTABLE;

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dArrayLength = 4000;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = (double *)blobOut;
  din.dArrayLength = 4000;
  din.dFormat = CF_DOUBLE;
  din.data.dptr = (double *)blobIn;
  for (i = 0; i < 4000; i++) din.data.dptr[i] = i;
  sprintf(devstr, "%s/#0", tgt);
  if (flags != 0) acc = CA_READ | flags;
  if ((cc = ExecLinkEx(devstr, "Echo", &dout, &din, acc, 1000)) != 0)
  {
    printf("Echo call to %.128s : %.32s\n", devstr, cc2str(cc));
    return 1;
  }
  for (i = 0; i < 4000; i++)
  {
    if (dout.data.dptr[i] != din.data.dptr[i])
    {
      printf("Echo call to %.128s failed to return element %d (%g vs. %g)\n",
        devstr, i, dout.data.dptr[i], din.data.dptr[i]);
      return 1;
    }
  }
  return 0;
}
int linkErrorTest(char *tgt)
{
  DTYPE dout;
  char devstr[128];
  int cc, acc = CA_READ | CA_MUTABLE;

  DTYPEZERO(dout); 
  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = (double *)blobOut;
  sprintf(devstr, "%s/#0", tgt);
  if ((cc=ExecLinkEx(devstr, "PropertyDoesNotExist", &dout, NULL, acc, 1000)) != illegal_property)
  {
    printf("PropertyDoesNotExist call to %.128s : '%.32s' instead of 'illegal property'\n", devstr, cc2str(cc));
    return 1;
  }
  return 0;
}
