
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "threader.h"
#include "unieqm.h"

#if defined(UNIX) || defined(WIN32)
# define random(r) (rand()%(r))
#else
# define random(r) ((i*(69069))%(r))
#endif

int needsRestart = FALSE;
int g_isTerminating = FALSE;
void term_handler(int signum)
{                  
  g_isTerminating = TRUE;
  feclog("caught SIGTERM");          
}
char thisFecName[FEC_NAME_SIZE+4] = "TUNITFEC";
char thisContext[CONTEXT_NAME_SIZE+4] = "TEST";
char thisServer[EXPORT_NAME_SIZE] = UNIEQM_EXPORT;
char *thisSubSystem = "SER";
char thisFecDescription[FEC_DESC_SIZE] = "TINE Unit Server";
char thisLocation[FEC_LOCATION_SIZE] = "";
UINT16 thisPortOffset = 50;
void PreSystemInit(void)
{
  signal(SIGTERM,term_handler);
  /* Some System Parameters and their default values    */
  /* are shown below.  You may change them as necessary */
  MaxNumContracts = 1000;
  MaxNumClients = 100;
  gConTblCapacity = 1000;
  FeclogDepth = 1000;
  gStartupDebug = TRUE;
  SetSystemRequireAcknowledgments(TRUE);
  MaxRPCTransportSize = 0x100000;
  gRetardSingleContractRemoval = FALSE;
  gIgnoreAddressCacheOnUnknownAddress = TRUE;
  SystemAssignBufferSpace(0x4000,0x2000);
  isCentralAlarmServer(TRUE);
  SetAMSCapacity(5000);

  if (strlen(thisContext) > 0)
  {
    int p;
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,9); thisFecName[9] = 0;
      strcat(thisFecName,"UNITFEC");
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }
}
void PostSystemInit(void)
{
  RegisterEquipmentModule(thisServer,UNIEQM_TAG,UNIEQM_NUM_DEVICES,unieqm,unieqm_ini,unieqm_bkg,UNIEQM_RATE,unieqm_exi);
}
void RestartServer(void)
{
  static int restartCountdown = 5;
  if (restartCountdown-- > 0) return;
  FlushContractTable();
  SetServerCycleIdle(TRUE);
  FreeEQPModules();
  ServerExitCondition = 0;
  PostSystemInit();
  SetServerCycleIdle(FALSE);
  needsRestart = FALSE;
  restartCountdown = 5;
}
#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
void registerMyStructs(void)
{
  static int done=FALSE;
  if (done) return;
  done = TRUE;
# ifdef CCCCCC
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,number),1,CF_INT32,"number")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,column),1,CF_INT32,"column")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,row),1,CF_INT32,"row")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,almsystem),ALARM_TAG_SIZE,CF_TEXT,"tag")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,subsystem),SUBSYSTEM_NAME_SIZE,CF_TEXT,"subsys")) quit(1);
  if (sealTaggedStruct(CASALMSYSITEM_TAG,sizeof(CasAlmSysTblEntry),100)) quit(1);
# endif
}
#define IMG_SIZE 20000
int RegisterUnieqmProperties(void)
{
  int id, cc = 0;
  DTYPE dout, din;
  PrpEgu egu, xegu;

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  din.dFormat = CF_INT32;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(UNIEQM_TAG,"ECHO",&dout,&din,CA_READ,AT_SCALAR,0,"echo the input data",PRP_ECHO,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 32;
  din.dFormat = CF_FLOAT;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(UNIEQM_TAG,"MCAFloatValues",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,0,"[0:1000 things]multi-channel array property",PRP_MCA_FLOAT,NULL);
  if (id < 0) ccerr(-id);
 
  dout.dFormat = CF_NAME64;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  din.dFormat = CF_NAME64;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(UNIEQM_TAG,"MCAName64Values",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,0,"multi-channel array property",PRP_MCA_NAME64,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  strcpy(dout.dTag,"Simple");
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"Simple");
  id = RegisterPropertyInformation(UNIEQM_TAG,"SimpleStruct",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"simple structure with primitive fields",PRP_STRUCT_SIMPLE,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  strcpy(dout.dTag,"Nested");
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"Nested");
  id = RegisterPropertyInformation(UNIEQM_TAG,"NestedStruct",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"nested structure with primitive and compound fields",PRP_STRUCT_NESTED,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  strcpy(dout.dTag,"Complex");
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"Complex");
  id = RegisterPropertyInformation(UNIEQM_TAG,"ComplexStruct",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"complex structure with adjustable length formats",PRP_STRUCT_COMPLEX,NULL);
  if (id < 0) ccerr(-id);

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_STRING;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  din.dFormat = CF_STRING;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(UNIEQM_TAG,"StringArray",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,0,"An array of free strings",PRP_STRING_ARRAY,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_SPECTRUM;
  dout.dArrayLength = SPECTRUM_SIZE;
  din.dFormat = CF_SPECTRUM;
  din.dArrayLength = SPECTRUM_SIZE;
  id = RegisterPropertyInformation(UNIEQM_TAG,"Spectrum",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"[-1000:1000 V]A doocs spectrum",PRP_SPECTRUM,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_GSPECTRUM;
  dout.dArrayLength = SPECTRUM_SIZE;
  id = RegisterPropertyInformation(UNIEQM_TAG, "G-Spectrum", &dout, NULL, CA_READ | CA_WRITE, AT_UNKNOWN, 0, "[-1000:1000 V]A doocs spectrum", PRP_GSPECTRUM, NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_ASPECTRUM;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  id = RegisterPropertyInformation(UNIEQM_TAG,"Spectra",&dout,NULL,CA_READ,AT_UNKNOWN,0,"An array of adjustable spectra",PRP_ASPECTRUM,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_IMAGE;
  dout.dArrayLength = IMG_SIZE;
  id = RegisterPropertyInformation(UNIEQM_TAG,"Image",&dout,NULL,CA_READ,AT_IMAGE,0,"An image type property",PRP_IMAGE,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_AIMAGE;
  dout.dArrayLength = 1000;
  id = RegisterPropertyInformation(UNIEQM_TAG,"ImageArray",&dout,NULL,CA_READ,AT_CHANNEL,0,"An array of images",PRP_AIMAGE,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_MDA;
  dout.dArrayLength = 1000;
  strcpy(dout.dTag,"FLOAT");
  din.dFormat = CF_MDA;
  din.dArrayLength = 1000;
  strcpy(din.dTag,"FLOAT");
  id = RegisterPropertyInformation(UNIEQM_TAG,"3DFloatArray",&dout,&din,CA_READ,AT_UNKNOWN,0,"A multi-dimensional array as data type",PRP_MDA,NULL);
  if (id < 0) ccerr(-id);

  DTYPEZERO(dout); DTYPEZERO(din);
  dout.dFormat = CF_KEYVALUE;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  din.dFormat = CF_KEYVALUE;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(UNIEQM_TAG,"KeyValuePairs",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"An XML string property",PRP_KEYVALUE,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = SPECTRUM_SIZE;
  id = RegisterPropertyInformation(UNIEQM_TAG,"TraceFloatValues",&dout,NULL,CA_READ,AT_SPECTRUM,0,"[-1000:1000 V][0:1000 mm]A Spectrum array type property",PRP_SPECTRUM_ARRAY,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = SPECTRUM_SIZE * UNIEQM_NUM_DEVICES;
  strcpy(egu.units,"V");
  egu.max = 1000;
  egu.min = -1000;
  strcpy(xegu.units,"mm");
  xegu.max = 100;
  xegu.min = 0;
  id = RegisterPropertyInformationEx(UNIEQM_TAG,"DoubleArrayFloatValues",&dout,NULL,
                  CA_READ,AT_SPECTRUM|AT_DOUBLE,SPECTRUM_SIZE,"A double float array",&egu,&xegu,PRP_DOUBLE_ARRAY,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_BOOLEAN;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  din.dFormat = CF_BOOLEAN;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(UNIEQM_TAG,"BooleanValues",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE,AT_CHANNEL,0,"A boolean array",PRP_BOOLEAN,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_BITFIELD16;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  strcpy(dout.dTag,"bf16");
  id = RegisterPropertyInformation(UNIEQM_TAG,"Bitfield16Values",&dout,NULL,CA_READ,AT_CHANNEL,0,"A 16-bit bitfield array",PRP_BITFIELD16,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_BITFIELD32;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  strcpy(dout.dTag,"bf32");
  id = RegisterPropertyInformation(UNIEQM_TAG,"Bitfield32Values",&dout,NULL,CA_READ,AT_CHANNEL,0,"A 32-bit bitfield array",PRP_BITFIELD32,NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = UNIEQM_NUM_DEVICES;
  din.dFormat = CF_INT32;
  din.dArrayLength = 1;
  id = RegisterPropertyInformation(UNIEQM_TAG,"Setting",&dout,&din,CA_READ|CA_WRITE|CA_SAVERESTORE|CA_HIST,AT_CHANNEL,0,"[-10000:10000 cts] data setting value",PRP_SETTING,NULL);
  if (id < 0) ccerr(-id);

  id = RegisterPropertyInformation(UNIEQM_TAG, "Command", NULL, NULL, CA_WRITE, AT_CHANNEL, 0, "empty command", PRP_CMD, NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 2048;
  dout.dTag[0] = 0;
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 2048;
  din.dTag[0] = 0;
  id = RegisterPropertyInformation(UNIEQM_TAG, "Bytes", &dout, &din, CA_READ | CA_WRITE, AT_UNKNOWN, 0, "just bytes", PRP_BYTES, NULL);
  if (id < 0) ccerr(-id);

  dout.dFormat = CF_DOUBLE;
  dout.dArrayLength = 1;
  dout.dTag[0] = 0;
  din.dFormat = CF_DOUBLE;
  din.dArrayLength = 1;
  din.dTag[0] = 0;
  id = RegisterPropertyInformation(UNIEQM_TAG, "Examine", &dout, &din, CA_READ | CA_WRITE, AT_UNKNOWN, 0, "display incoming data info (and echo if required)", PRP_EXAMINER, NULL);
  if (id < 0) ccerr(-id);

err:
  if (cc != 0)
  {
    feclog("Can't register property : %.32s",cc2str(cc));
  }
  return cc;
}
int RegisterDeviceNames(void)
{
  int i;
  char dev[DEVICE_NAME_SIZE];
  for (i=0; i<UNIEQM_NUM_DEVICES; i++)
  {
    sprintf(dev,"UnitDevice%d",i);
    RegisterDeviceName(UNIEQM_TAG,dev,i);
  }
  return 0;
}
char *kvStrings[UNIEQM_NUM_DEVICES];
typedef struct
{
  int   a;
  float b;
  char  t[16];
} StHdr;
typedef struct
{
  int c;
  float d;
  FLTINT e;
} StBod;
typedef struct
{
  StHdr hdr;
  StBod body[4];
} StCmp;
typedef struct
{
  float amplitude;
  float frequency;
  float noise;
  float phase;
  _Bool online;
  char  description[64];
} SineInfo;
typedef struct
{
  char cmt[80];
  float stt;
  float inc;
} SpecInfo;

typedef struct
{
  float fltfield;
  int intfield;
  double dblfield;
  char *strfields[4];
  DIMAGE imgfield;
} CmplxStruct;

CmplxStruct cmplxStruct;

void registerStructs(void)
{
  static int done = 0;
  if (done) return;
  done = TRUE;
  /* this must follow the order of the structure explicitly! */
  if (addFieldToStruct("Simple",OFFSETIN(SineInfo,amplitude),1,CF_FLOAT,"amplitude")) quit(1);
  if (addFieldToStruct("Simple",OFFSETIN(SineInfo,frequency),1,CF_FLOAT,"frequency")) quit(1);
  if (addFieldToStruct("Simple",OFFSETIN(SineInfo,noise),1,CF_FLOAT,"noise")) quit(1);
  if (addFieldToStruct("Simple",OFFSETIN(SineInfo,phase),1,CF_FLOAT,"phase")) quit(1);
  if (addFieldToStruct("Simple",OFFSETIN(SineInfo,online),1,CF_BOOLEAN,"online")) quit(1);
  if (addFieldToStruct("Simple",OFFSETIN(SineInfo,description),64,CF_TEXT,"description")) quit(1);
  if (sealTaggedStruct("Simple",sizeof(SineInfo),UNIEQM_NUM_DEVICES)) quit(1);

  /* below a status header struct */
  if (addFieldToStruct("StHdr",OFFSETIN(StHdr,a),1,CF_INT32,"a")) quit(1);
  if (addFieldToStruct("StHdr",OFFSETIN(StHdr,b),1,CF_FLOAT,"b")) quit(1);
  if (addFieldToStruct("StHdr",OFFSETIN(StHdr,t),16,CF_TEXT,"t")) quit(1);
  if (sealTaggedStruct("StHdr",sizeof(StHdr),UNIEQM_NUM_DEVICES)) quit(1);

  /* below a status body struct */
  if (addFieldToStruct("StBod",OFFSETIN(StBod,c),1,CF_INT32,"c")) quit(1);
  if (addFieldToStruct("StBod",OFFSETIN(StBod,d),1,CF_FLOAT,"d")) quit(1);
  if (addFieldToStruct("StBod",OFFSETIN(StBod,e),1,CF_FLTINT,"e")) quit(1);
  if (sealTaggedStruct("StBod",sizeof(StBod),UNIEQM_NUM_DEVICES)) quit(1);

  /* below a struct composed of the above header a 4 X the above body : */
  if (addFieldToStruct("Nested",OFFSETIN(StCmp,hdr),1,CF_STRUCT,"<StHdr>hdr")) quit(1);
  if (addFieldToStruct("Nested",OFFSETIN(StCmp,body),4,CF_STRUCT,"<StBod>body")) quit(1);
  if (sealTaggedStruct("Nested",sizeof(StCmp),UNIEQM_NUM_DEVICES)) quit(1);

  if (addFieldToStruct("Complex",OFFSETIN(CmplxStruct,fltfield),1,CF_FLOAT,"floatValue")) quit(1);
  if (addFieldToStruct("Complex",OFFSETIN(CmplxStruct,intfield),1,CF_INT32,"intValue")) quit(1);
  if (addFieldToStruct("Complex",OFFSETIN(CmplxStruct,dblfield),1,CF_DOUBLE,"doubleValue")) quit(1);
  if (addFieldToStruct("Complex",OFFSETIN(CmplxStruct,strfields),4,CF_STRING,"strings")) quit(1);
  if (addFieldToStruct("Complex",OFFSETIN(CmplxStruct,imgfield),1,CF_AIMAGE,"image")) quit(1);
  if (sealTaggedStruct("Complex",sizeof(CmplxStruct),10)) quit(1);

# define SRVTAG "UNITSERV"
# define BF16TAG "bf16"
  openBitField(SRVTAG,BF16TAG,CF_BITFIELD16);
  addFieldToBitField(SRVTAG,BF16TAG,0x01,"bit1");
  addFieldToBitField(SRVTAG,BF16TAG,0x02,"bit2");
  addFieldToBitField(SRVTAG,BF16TAG,0x04,"bit3");
  addFieldToBitField(SRVTAG,BF16TAG,0x08,"bit4");
  addFieldToBitField(SRVTAG,BF16TAG,0xf0,"field1");
  addFieldToBitField(SRVTAG,BF16TAG,0xf00,"field2");
  addFieldToBitField(SRVTAG,BF16TAG,0xf000,"field3");

# define BF32TAG "bf32"
  openBitField(SRVTAG,BF32TAG,CF_BITFIELD32);
  addFieldToBitField(SRVTAG,BF32TAG,0x01,"bit1");
  addFieldToBitField(SRVTAG,BF32TAG,0x02,"bit2");
  addFieldToBitField(SRVTAG,BF32TAG,0x04,"bit3");
  addFieldToBitField(SRVTAG,BF32TAG,0x08,"bit4");
  addFieldToBitField(SRVTAG,BF32TAG,0xf0,"field1");
  addFieldToBitField(SRVTAG,BF32TAG,0xf00,"field2");
  addFieldToBitField(SRVTAG,BF32TAG,0xf000,"field3");
  addFieldToBitField(SRVTAG,BF32TAG,0xff000,"field4");
  addFieldToBitField(SRVTAG,BF32TAG,0xf0f000,"field5");
  addFieldToBitField(SRVTAG,BF32TAG,0xf0f0f0f,"field6");

}

float sinbuf[UNIEQM_NUM_DEVICES][SPECTRUM_SIZE];

SineInfo sineInfoTable[UNIEQM_NUM_DEVICES];
SpecInfo specInfoTable[UNIEQM_NUM_DEVICES];

StCmp nsTable[UNIEQM_NUM_DEVICES];

char *staticUnitString[UNIEQM_NUM_DEVICES] = { 
  "string number 1     ",
  "string number 2",
  "string number 3",
  "string number 4",
  "really and truly, fantastically, incredibly, unimaginatively, gosh, let's see what this baby can do, as in give some bytes, you know, some longggggggg character strings that'll surely cause the sucker to choke, you-wouldn't-believe-it's-possible, long string 5",
  "really and truly, fantastically, incredibly, unimaginatively, gosh, let's see what this baby can do, as in give some bytes, you know, some longggggggg character strings that'll surely cause the sucker to choke, you-wouldn't-believe-it's-possible, long string 6",
  "really and truly, fantastically, incredibly, unimaginatively, you-wouldn't-believe-it's-possible, long string 7",
  "really and truly, fantastically, incredibly, unimaginatively, you-wouldn't-believe-it's-possible, long string 8",
  "str 9",
  "str 10"
};
char *unitString[UNIEQM_NUM_DEVICES];

NAME64 n64Table[UNIEQM_NUM_DEVICES];
int boolTable[UNIEQM_NUM_DEVICES];

int settings[UNIEQM_NUM_DEVICES];

DMDA mdaTable[UNIEQM_NUM_DEVICES];

void fillinMDA(int devnr)
{ /* make a 3-d float array */
  MDX xaxis = { 10, 0, "x", "mm", -100.0, 100.0 };
  MDX yaxis = { 10, 0, "y", "mm", -100.0, 100.0 };
  MDX zaxis = { 10, 0, "z", "m", 0, 1000.0 };
  char cmt[80];
  float *d=(float *)malloc(1000*sizeof(float));
  int i,j,k;
  if (d == NULL)
  {
    tineDebug = 1;
    feclog("no memory to allocate 3-D multi-dimensional float array nr. %d!",devnr);
    exit(1);
  }
  for (i=0; i<10; i++) for (j=0; j<10; j++) for (k=0; k<10; k++)
    d[i*100+j*10+k] = (float)((i+devnr)*1.5 + (j+devnr)*20.0 + (k+devnr)*100.0);

  sprintf(cmt,"How 'bout them apples,Mr. UnitDevice%d",devnr);
  MakeMDA(&mdaTable[devnr],cmt,CF_FLOAT,(void *)d,3,&xaxis,&yaxis,&zaxis);
}

void fillUnitString(int devnr)
{
  int slen;
  if (staticUnitString[devnr] != NULL)
  {
    slen = (int)strlen(staticUnitString[devnr]);
    unitString[devnr] = (char *)calloc(1,slen+1);
    strcpy(unitString[devnr],staticUnitString[devnr]);
  }
}

void fillSineBuffers(int devnr)
{ /* used internally -> don't bother with bounds checking */
  int i, k, n, p;
  float d, f, a;
  float *target = sinbuf[devnr];

  a = sineInfoTable[devnr].amplitude = 256.0;
  f = sineInfoTable[devnr].frequency = 1.0;
  p = (int)(sineInfoTable[devnr].phase = 0.0);
  n = (int)(sineInfoTable[devnr].noise = 30.0);
  sprintf(sineInfoTable[devnr].description,"Sine Generator %d at your disposal",devnr);
  sineInfoTable[devnr].online = TRUE;
  
  for (i=0; i<SPECTRUM_SIZE; i++)
  {
    k = (i + p) % SPECTRUM_SIZE;
    d = (float)(0.9 * (50.0 - devnr)/50.0);
    target[k] = random(n) + (float)(sin(f*i*6.2832/(SPECTRUM_SIZE/8))*a);
    target[k] *= ((float)1.0 - (d * (float)i)/(float)SPECTRUM_SIZE);
  }
  return;
}
void fillNestedStructTable(int devnr)
{
  memset(&nsTable[devnr],0,sizeof(StCmp));
  nsTable[devnr].hdr.a = devnr + 1;
  nsTable[devnr].hdr.b = (float)(devnr+2);
  sprintf(nsTable[devnr].hdr.t,"struct %d",devnr);
  nsTable[devnr].body[0].c = devnr+5;
  nsTable[devnr].body[0].d = (float)(devnr+6);
  nsTable[devnr].body[0].e.fval = (float)(devnr+7);
  nsTable[devnr].body[0].e.ival = devnr+17;
  nsTable[devnr].body[1].c = devnr+8;
  nsTable[devnr].body[1].d = (float)(devnr+9);
  nsTable[devnr].body[1].e.fval = (float)(devnr+10);
  nsTable[devnr].body[1].e.ival = devnr+20;
}
BYTE frame[IMG_SIZE];
void pushImageBytes(DIMAGE *img,int img_size)
{
  static int imgSeed = 1;
  int i;
  if (img_size > IMG_SIZE) img_size = IMG_SIZE;
  memset(&img->sourceHeader, 0, sizeof(img->sourceHeader));
  img->sourceHeader.baseTag = 55;
  img->sourceHeader.cameraPortId = 66;
  img->sourceHeader.totalLength = 188 + img_size;
  strcpy(img->sourceHeader.cameraPortName,"MY Port");
  img->frameHeader.sourceWidth = img_size/100;
  img->frameHeader.sourceHeight = 100;
  img->frameBufferSize = img_size;
  img->frameHeader.appendedFrameSize = img_size;
  for (i=0; i<img_size; i++) frame[i] = imgSeed; 
  img->frameBuffer = frame;
  imgSeed = (imgSeed + 1) % 155;
}
void dumpImageBytes(DIMAGE *img)
{
  int i;
  dbglog("baseTag: %d", img->sourceHeader.baseTag);
  dbglog("cameraPortId: %d", img->sourceHeader.cameraPortId);
  dbglog("cameraPortName: %s", img->sourceHeader.cameraPortName);
  dbglog("sourceWidth: %d", img->frameHeader.sourceWidth);
  dbglog("sourceHeight: %d", img->frameHeader.sourceHeight);
  dbglog("appendedFrameSize: %d", img->frameHeader.appendedFrameSize);
  dbglog("initial frame bytes:");
  for (i = 0; i < img->frameHeader.appendedFrameSize && i < 16; i++)
    dbglog("%x ", img->frameBuffer[i]);
}
void dumpGSpectrum(DGSPECTRUM *gsp)
{
  int i;
  dbglog("comment: %.80s", gsp->comment);
  dbglog("id: %d", gsp->id);
  dbglog("buffer size: %d", gsp->spectBufferSize);
  dbglog("initial spectrum values:");
  for (i = 0; i < (int)gsp->spectBufferSize && i < 10; i++)
    dbglog("%g ", gsp->spectBuffer[i]);
}
void unieqm_bkg(void)
{
  return;
}
void unieqm_ini(void)
{
  int i;
  char s[256];

  registerStructs();
  RegisterDeviceNames();
  RegisterUnieqmProperties();

  for (i=0; i<UNIEQM_NUM_DEVICES; i++)
  {
    sprintf(s,"key%d: value%d",i,i);
    kvStrings[i] = (char *)calloc(1,strlen(s)+1);
    strcpy(kvStrings[i],s);
    fillSineBuffers(i);
    fillNestedStructTable(i);
    fillUnitString(i);
    sprintf(specInfoTable[i].cmt,"this is spectrum %d",i);
    specInfoTable[i].inc = 1.0;
    specInfoTable[i].stt = 0;
    sprintf(n64Table[i].name,"ThisIsName %d",i);
    fillinMDA(i);
    boolTable[i] = i%2;
  }
}
void unieqm_exi(void)
{
}
int unieqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc,i,ival,devnr,prpid;
  int datasize;
  float mcarray[UNIEQM_NUM_DEVICES], fval;
  short smcarray[UNIEQM_NUM_DEVICES];
  int imcarray[UNIEQM_NUM_DEVICES];
  char *c;
  DSPECTRUM *sp;
  /* get device number from device name  */
  devnr = GetDeviceNumber(UNIEQM_TAG,devName); /* -1 => ALL ? */
  if (devnr < 0 || devnr >= UNIEQM_NUM_DEVICES) return illegal_device;
  prpid = GetPropertyId(UNIEQM_TAG,devProperty);

  switch (prpid)
  {
    case PRP_ECHO:
      if (dout->dArrayLength == 0) return dimension_error;
      if (din->dArrayLength != dout->dArrayLength) return data_size_mismatch;
      if (din->dFormat != dout->dFormat) return illegal_format;
      if (IsVariableLengthFormat(din->dFormat)) return not_supported;
      if (strncmp(dout->dTag,din->dTag,TAG_NAME_SIZE) != 0) return invalid_structure_tag;
      datasize = dout->dArrayLength * GetFormatSize(dout->dFormat);
      datasize += getFormatHeaderSize(dout->dFormat);
      if (dout->dFormat == CF_MDA) datasize = sizeof(DMDA);
      memcpy(dout->data.vptr,din->data.vptr,datasize);
      return 0;
    case PRP_MCA_FLOAT:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=GetValuesAsFloat(din,&fval,1)) != 0) return cc;
          if (fval < 1 || fval > 1000) return out_of_range;
          sineInfoTable[devnr].amplitude = fval;
          for (i=0; i<UNIEQM_NUM_DEVICES; i++) mcarray[i] = sineInfoTable[i].amplitude;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<UNIEQM_NUM_DEVICES; i++) mcarray[i] = sineInfoTable[i].amplitude;
          if ((cc=PutValuesFromFloatEx(dout,mcarray,UNIEQM_NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_BOOLEAN:
        if (din->dArrayLength > 0)
        {
          if (!(access & CA_WRITE)) return illegal_read_write;
          if ((cc=GetValuesAsLong(din,&ival,1)) != 0) return cc;
          if (ival != 0) ival = 1;
          boolTable[devnr] = ival;
        }
        if (dout->dArrayLength > 0)
        {
          if ((cc=PutValuesFromLongEx(dout,boolTable,UNIEQM_NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_BITFIELD16:
        if ((access & CA_WRITE)) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          for (i=0; i<UNIEQM_NUM_DEVICES; i++) smcarray[i] = random(48000);
          if (dout->dFormat != CF_BITFIELD16 && dout->dFormat != CF_INT16) return illegal_format;
          if ((cc=PutValuesFromShortEx(dout,smcarray,UNIEQM_NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_BITFIELD32:
        if (dout->dArrayLength > 0)
        {
          for (i=0; i<UNIEQM_NUM_DEVICES; i++) imcarray[i] = random(0x7fffffff);
          if (dout->dFormat != CF_BITFIELD32 && dout->dFormat != CF_INT32) return illegal_format;
          if ((cc=PutValuesFromLongEx(dout,imcarray,UNIEQM_NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_MCA_NAME64:
        if (din->dArrayLength > 0)
        { /* input data => require write access */
          if (!(access & CA_WRITE)) return illegal_read_write;
          if (din->dArrayLength != 1) return dimension_error;
          if ((cc=GetValuesAsStringEx(din,n64Table[0].name,CF_NAME64,UNIEQM_NUM_DEVICES,devnr)) != 0) return cc;
        }
        if (dout->dArrayLength > 0)
        { /* prepare multichannel array */
          for (i=0; i<UNIEQM_NUM_DEVICES; i++) mcarray[i] = sineInfoTable[i].amplitude;
          if ((cc=PutValuesFromString(dout,n64Table[0].name,CF_NAME64,UNIEQM_NUM_DEVICES,devnr)) != 0) return cc;
        }
        return 0;
    case PRP_STRUCT_SIMPLE:
      {
        SineInfo *sinf;
        if (din->dArrayLength)
        { /* allow an atomic set of all relevant values this way : */
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(din->dTag,"Simple",TAG_NAME_SIZE)) return invalid_structure_tag;
          if (din->dArrayLength != 1) return dimension_error;
          sinf = (SineInfo *)(din->data.vptr);
          if (sinf->amplitude < 1 || sinf->amplitude > 1000) return out_of_range;
          if (sinf->frequency < 1 || sinf->frequency > 100) return out_of_range;
          if (sinf->phase < 0 || sinf->phase > 512) return out_of_range;
          if (sinf->noise < 0 || sinf->noise > 100) return out_of_range;
          sineInfoTable[devnr] = *sinf;
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,"Simple",TAG_NAME_SIZE)) return invalid_structure_tag;
          //if (dout->dArrayLength > NUM_DEVICES) dout->dArrayLength = NUM_DEVICES;
          sinf = (SineInfo *)(dout->data.vptr);
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            sinf[i] = sineInfoTable[(i+devnr)%UNIEQM_NUM_DEVICES];
          }
        }
        return 0;
      }
    case PRP_STRUCT_NESTED:
      {
        StCmp *st;
        if (din->dArrayLength)
        { /* allow an atomic set of all relevant values this way : */
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(din->dTag,"Nested",TAG_NAME_SIZE)) return invalid_structure_tag;
          if (din->dArrayLength != 1) return dimension_error;
          st = (StCmp *)(din->data.vptr);
          nsTable[devnr] = *st;
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,"Nested",TAG_NAME_SIZE)) return invalid_structure_tag;
          if (dout->dArrayLength > (UINT32)UNIEQM_NUM_DEVICES) dout->dArrayLength = (UINT32)UNIEQM_NUM_DEVICES;
          st = (StCmp *)(dout->data.vptr);
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            st[i] = nsTable[(i+devnr)%UNIEQM_NUM_DEVICES];
          }
        }
        return 0;
      }
    case PRP_STRUCT_COMPLEX:
      {
        CmplxStruct *fs;
        if (din->dArrayLength)
        {
          if (din->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(din->dTag,"Complex",TAG_NAME_SIZE)) return invalid_structure_tag;
          if (din->dArrayLength != 1) return dimension_error;
          fs = (CmplxStruct *)(din->data.vptr);
          printf("floatValue: %g\n",fs->fltfield);
          printf("intValue: %d\n",fs->intfield);
          printf("doubleValue: %g\n",fs->dblfield);
          printf("%s\n",fs->strfields[0]);
          printf("%s\n",fs->strfields[1]);
          printf("%s\n",fs->strfields[2]);
          printf("%s\n",fs->strfields[3]);
          printf("cam port id %d\n",fs->imgfield.sourceHeader.cameraPortId);
          printf("frame byte 1 %d\n",fs->imgfield.frameBuffer[0]);
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,"Complex",TAG_NAME_SIZE)) return invalid_structure_tag;
          if (dout->dArrayLength > UNIEQM_NUM_DEVICES) dout->dArrayLength = UNIEQM_NUM_DEVICES;
          fs = (CmplxStruct *)(dout->data.vptr);
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            memset(&fs[i],0,sizeof(CmplxStruct));
            fs[i].fltfield = sineInfoTable[(i+devnr)%UNIEQM_NUM_DEVICES].amplitude;
            fs[i].intfield = (int)sineInfoTable[(i+devnr)%UNIEQM_NUM_DEVICES].frequency;
            fs[i].dblfield = sineInfoTable[(i+devnr)%UNIEQM_NUM_DEVICES].noise;
            fs[i].strfields[0] = sineInfoTable[(i+devnr)%UNIEQM_NUM_DEVICES].description;
            fs[i].strfields[1] = "and something in the data segment";
            fs[i].strfields[2] = unitString[(i+devnr)%10];
            fs[i].strfields[3] = "just a filler";
            pushImageBytes(&fs[i].imgfield,IMG_SIZE);
          }
        }
        return 0;
      }
    case PRP_STRING_ARRAY:
        if (din->dArrayLength > 0)
        {
          char *s = din->data.strptr[0];
          if (!(access & CA_WRITE)) return illegal_read_write;
          if (din->dFormat != CF_STRING) return illegal_format;
          if (din->dArrayLength != 1) return dimension_error;
          free(unitString[devnr]);
          unitString[devnr] = (char *)calloc(1,strlen(s)+1);
          if (unitString[devnr] == NULL) return out_of_server_memory;
          strcpy(unitString[devnr],s);
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_STRING) return illegal_format;
          if (dout->dArrayLength > (UINT32)UNIEQM_NUM_DEVICES) dout->dArrayLength = (UINT32)UNIEQM_NUM_DEVICES;
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            dout->data.strptr[i] = unitString[(i+devnr)%UNIEQM_NUM_DEVICES];
          }
        }
        return 0;
    case PRP_SPECTRUM:
      if (din->dArrayLength > 0)
      {
        if (!(access & CA_WRITE)) return illegal_read_write;
        if (din->dArrayLength > SPECTRUM_SIZE) return dimension_error;
        sp =(DSPECTRUM *)(din->data.vptr);
        strncpy(specInfoTable[devnr].cmt,sp->comment,80);
        specInfoTable[devnr].stt = sp->s_start;
        specInfoTable[devnr].inc = sp->s_inc;
        memset(sinbuf[devnr],0,SPECTRUM_SIZE);
        memcpy(sinbuf[devnr],sp->d_spect_array,sizeof(float)*din->dArrayLength);
      }
      if (dout->dArrayLength > 0)
      {
        sp =(DSPECTRUM *)(dout->data.vptr);
        if (dout->dArrayLength > SPECTRUM_SIZE) dout->dArrayLength = SPECTRUM_SIZE;
        strncpy(sp->comment,specInfoTable[devnr].cmt,80);
        sp->tm = (int)time(NULL);
        sp->s_start = specInfoTable[devnr].stt;
        sp->s_inc = specInfoTable[devnr].inc;
        sp->status = 0;
        memcpy(sp->d_spect_array,sinbuf[devnr],sizeof(float)*dout->dArrayLength);
      }
      return 0;
    case PRP_ASPECTRUM:
      {
        ASPECTRUM *asp = (ASPECTRUM *)(dout->data.vptr);
        if (dout->dFormat != CF_ASPECTRUM) return illegal_format;
        if (dout->dArrayLength > (UINT32)UNIEQM_NUM_DEVICES) dout->dArrayLength = (UINT32)UNIEQM_NUM_DEVICES;
        for (i=0; i<(int)dout->dArrayLength; i++)
        {
          sprintf(asp[i].comment,"spectrum number %d",i);
          asp[i].tm = (int)time(NULL);
          asp[i].s_start = 0;
          asp[i].s_inc = 1;
          asp[i].status = 0;
          asp[i].spectBufferSize = SPECTRUM_SIZE;
          asp[i].spectBuffer = sinbuf[i];
        }
        return 0;
      }
    case PRP_GSPECTRUM:
      if (din->dArrayLength > 0)
      {
        DGSPECTRUM *gspi = (DGSPECTRUM *)(din->data.vptr);
        if (din->dArrayLength > SPECTRUM_SIZE) return dimension_error;
        if (dout->dArrayLength >= din->dArrayLength)
        { // an echo request ...
          memcpy(dout->data.vptr, din->data.vptr, sizeof(DGSPECTRUM));
          return 0;
        }  
        dumpGSpectrum(gspi);
      }
      if (dout->dArrayLength > 0)
      {
        DGSPECTRUM *gsp = (DGSPECTRUM *)(dout->data.vptr);
        if (dout->dFormat != CF_GSPECTRUM) return illegal_format;
        if (dout->dArrayLength > (UINT32)SPECTRUM_SIZE) dout->dArrayLength = (UINT32)SPECTRUM_SIZE;
        sprintf(gsp->comment, "Gee, a spectrum number test");
        gsp->tm = (int)time(NULL);
        gsp->s_start = 0;
        gsp->s_inc = 1;
        gsp->status = 0;
        gsp->groups = 1;
        gsp->group_inc = 1;
        gsp->id = 42;
        for (i = 0; i < 24; i++) gsp->statistics[i] = (float)(i + 0.123);
        gsp->stat_mask = 7;
        for (i = 0; i < 8; i++) gsp->timing_param[i] = i;
        gsp->spectBufferSize = SPECTRUM_SIZE;
        gsp->spectBuffer = sinbuf[i];
      }
      return 0;
    case PRP_IMAGE:
        if (din->dArrayLength > 0)
        {
          DIMAGE *img = (DIMAGE *)(din->data.vptr);
          if (din->dFormat != CF_IMAGE) return illegal_format;
          if (dout->dArrayLength >= din->dArrayLength)
          { // an echo request ...
            memcpy(dout->data.vptr, din->data.vptr, sizeof(DIMAGE));
            return 0;
          }
          dumpImageBytes(img);
        }
        if (dout->dArrayLength)
        {
          if (dout->dFormat != CF_IMAGE) return illegal_format;
          ival = IMG_SIZE*9/10 + random(IMG_SIZE/10);
          if (dout->dArrayLength < (UINT32)ival) ival = dout->dArrayLength;
          if (dout->dArrayLength > (UINT32)ival) dout->dArrayLength = ival;
          pushImageBytes((DIMAGE *)(dout->data.vptr),ival);
        }
        return 0;
    case PRP_AIMAGE:
        if (dout->dArrayLength)
        {
          DIMAGE *img = (DIMAGE *)(dout->data.vptr);
          if (dout->dFormat != CF_AIMAGE) return illegal_format;
          if (dout->dArrayLength > (UINT32)UNIEQM_NUM_DEVICES) dout->dArrayLength = (UINT32)UNIEQM_NUM_DEVICES;
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            ival = IMG_SIZE*i/10 + random(IMG_SIZE/10);
            pushImageBytes(&img[i],ival);
          }
        }
        return 0;
    case PRP_MDA:
      {
        DMDA mda, *pmda;
        if (din->dArrayLength > 0)
        {
          if (din->dFormat == CF_MDA)
          {
            if (stricmp(din->dTag,"FLOAT")) return illegal_format;
            pmda = (DMDA *)din->data.vptr;
            if (pmda->ndim != 3) return invalid_data;
            if (din->dArrayLength != 1000) return dimension_error;
            memcpy(&mdaTable[devnr],pmda,MDA_HDR_SIZE);
            memcpy(mdaTable[devnr].data.fptr,pmda->data.fptr,din->dArrayLength*sizeof(float));
          }
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_MDA) return illegal_format;
          if (dout->dArrayLength < 100) return dimension_error;
          if (stricmp(dout->dTag,"FLOAT")) return illegal_format;
          memcpy(&mda,&mdaTable[devnr],sizeof(DMDA));
          mda.axis[2].len = (dout->dArrayLength/100);
          if (mda.axis[2].len > 10) mda.axis[2].len = 10;
          memcpy(dout->data.vptr,&mda,sizeof(DMDA));
        }
        return 0;
      }
    case PRP_KEYVALUE:
      if (din->dArrayLength > 0)
      {
        if (!(access & CA_WRITE)) return illegal_read_write;
        if (din->dArrayLength != 1) return dimension_error;
        if (din->dFormat != CF_KEYVALUE) return illegal_format;
        c = din->data.strptr[0];
        free(kvStrings[devnr]);
        if((kvStrings[devnr]=(char *)calloc(1,strlen(c)+1)) == NULL) return out_of_server_memory;
        strcpy(kvStrings[devnr],c);
      }
      if (dout->dArrayLength > 0)
      {
        if (dout->dFormat != CF_KEYVALUE) return illegal_format;
        for (i=0; i<(int)dout->dArrayLength && i<UNIEQM_NUM_DEVICES-devnr; i++)
        {
          dout->data.strptr[i] = kvStrings[i+devnr];
        }
        dout->dArrayLength = (UINT32)i;
      }
      return 0;
    case PRP_SPECTRUM_ARRAY:
      if (access&CA_WRITE) return illegal_read_write;
      return PutValuesFromFloat(dout,sinbuf[devnr],SPECTRUM_SIZE);
    case PRP_DOUBLE_ARRAY:
      if (access&CA_WRITE) return illegal_read_write;
      if ((cc=PutValuesFromFloat(dout,sinbuf[0],SPECTRUM_SIZE*UNIEQM_NUM_DEVICES)) != 0) return cc;
      return 0;
    case PRP_SETTING:      
      if (din->dArrayLength > 0)
      {
        if (!(access & CA_WRITE)) return illegal_read_write;
        if ((cc=GetValuesAsLong(din,&ival,1)) != 0) return cc;
        settings[devnr] = ival;
      }
      if (dout->dArrayLength > 0)
      {
        if ((cc=PutValuesFromLongEx(dout,settings,UNIEQM_NUM_DEVICES,devnr)) != 0) return cc;
      }
      return 0;
    case PRP_CMD:
      if (!(access & CA_WRITE)) return illegal_read_write;
      return 0;
    case PRP_BYTES:
      if (din->dArrayLength > 0)
      {
        int dsize = fmtsizeof(din->dFormat) * din->dArrayLength;
        for (i = 0; i<dsize; i++)
        {
          ttyoutput("0x%x ", din->data.bptr[i]);
        }
      }
      if (dout->dArrayLength > 0)
      {
        int dsize = fmtsizeof(dout->dFormat) * dout->dArrayLength;
        for (i = 0; i<dsize; i++)
        {
          dout->data.bptr[i] = (i % 256);
        }
      }
      return 0;
    case PRP_EXAMINER:
      if (din->dArrayLength > 0)
      {
        ttyoutput("Received: %d %s objects", din->dArrayLength, GetFormatAsString(din->dFormat));
        ttyoutput("\n");
        for (i = 0; i < (int)din->dArrayLength; i++)
        {
          //GetDataAsString(BYTE *data, int idx, short fmt, char *tag)
          ttyoutput(GetDataAsString(din->data.bptr, i, din->dFormat, din->dTag));
        }
      }
      if (dout->dArrayLength > 0)
      {
        if (dout->dFormat != din->dFormat) return illegal_format;
        if (dout->dArrayLength != din->dArrayLength) return dimension_error;
        memcpy(dout->data.bptr, din->data.bptr, dout->dArrayLength*fmtsizeof(dout->dFormat)+GetFormatHeaderSize(dout->dFormat));
      }
      return 0;
    default:
      break;
  }
  return illegal_property;
}
