/*
 * File apceqm.h
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#define APCEQM_TAG "APCEQM"

#ifndef ccerr
#define ccerr(i) { cc = i; goto err; }
#endif

#ifndef ulong
# define ulong UINT32
#endif

#define VERSION_LENGTH 16

void apcinit(void);
void apctask(void);
int apceqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void fswap(NAME8 *fbuf,int i, int j);
void fileSort(NAME8 *fbuf,int start,int num);
void freeSession(int id);

typedef struct
{
  UINT32 sessionID;
  UINT32 event;
  char ext[4];
  char expName[EXPORT_NAME_SHORTSIZE];
  char devName[DEVICE_NAME_SHORTSIZE];
  char devProperty[PROPERTY_NAME_SHORTSIZE];
  UINT16 dSize;
  UINT16 dFormat;
} APCBLK5;

typedef struct
{
  UINT32 sessionID;
  UINT32 event;
  char ext[16];
  char ctxName[CONTEXT_NAME_SIZE];
  char expName[EXPORT_NAME_SIZE];
  char devName[DEVICE_NAME_SIZE];
  char devProperty[PROPERTY_NAME_SIZE];
  char dTag[TAG_NAME_SIZE];
  UINT32 dSize;
  UINT16 dFormat;
  UINT16 reserved;
} APCBLK;

typedef struct
{
  char ext[16];
  char trg[16];
  char src[80];
  char dst[80];
} ARCHRECORD;

typedef struct APCStruct
{
  char   ctx[CONTEXT_NAME_SIZE];
  char   srv[EXPORT_NAME_SIZE]; 
  char   prp[PROPERTY_NAME_SIZE];
  char   dev[DEVICE_NAME_SIZE];
  char   tag[TAG_NAME_SIZE];
  char   sts[STATUS_SHORTSIZE];
  UINT32 siz;
  UINT16 fmt;
  UINT16 err;
  float  scale,
         shift;
  double timestamp;
  UINT32 sysstamp;
  UINT32 usrstamp;
  UINT32 ptr; /* location in file */
  struct APCStruct *Next;
} ApcRecord;

typedef struct
{
  time_t active,        /* last session request time */
         est;           /* session established */
  int count;
  int hFile;
  char fullName[128];
  char ext[16];
  ApcRecord *apc;       /* pointer to a chain */
} SESSION;
#define MAXSESSION  512
extern SESSION session[MAXSESSION];

#define SRCVAL(d,n,fmt) (\
   fmt == CF_SHORT  ? (double) ((short *)d)[n] :\
   fmt == CF_LONG   ? (double) ((long *)d)[n] :\
   fmt == CF_LNGINT ? (double) ((LNGINT *)d)[n].lval :\
   fmt == CF_FLOAT  ? (double) ((float *)d)[n] :\
   fmt == CF_FLTINT ? (double) ((FLTINT *)d)[n].fval :\
   fmt == CF_DOUBLE ? (double) ((double *)d)[n] :\
   fmt == CF_DBLDBL ? (double) ((DBLDBL *)d)[n].d1val :\
   fmt == CF_BYTE   ? (double) ((BYTE *)d)[n] :\
   fmt == CF_TEXT   ? (double) ((char *)d)[n] : (double)0\
)

#define srcptr(fmt) (\
  fmt == CF_SHORT ? (BYTE *)&sdat :\
  fmt == CF_LONG  ? (BYTE *)&ldat :\
  fmt == CF_FLOAT ? (BYTE *)&fdat :\
  fmt == CF_DOUBLE ? (BYTE *)&ddat: NULL\
)

extern BYTE *archiveBuffer;
extern int archiveBufferSize;
extern ARCHRECORD *ArchRecord;
extern void *LPCdata;
extern int numDataSets;
extern int archDebug;
extern int numArchiveTypes;

extern int (*filterArchiveData)(char *prp,char *nam,BYTE *buf,int fmt,UINT32 len,int EqNo);
int GetEventIdFromFileName(char *fname);
int GetLocalSessionID(ulong eventNumber,char *ext);
int reformatArchiveData(void *dst,short dstFmt,void *src,short srcFmt,int size,float scale,float shift);
int reformatArchiveDataEx(void *dst,short dstFmt,void *src,short srcFmt,int index,int size,float scale,float shift);
int apceqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void apcinit(void);
void apctask(void);
void apceqm_exit(void);
void dumpSessions(void);
void freeSessions(void);
int cmdDumpSessions(int arg0,int arg1,int arg2,int arg3);
int cmdFreeSessions(int arg0,int arg1,int arg2,int arg3);
int isArchiveOngoing(char *fn);
int loadStoredData(int year,char *fname,char *ext);
int makeFullName(char *fname,char *ext);
int getArchiveContracts(void);
