/*
 * File arceqm.h
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _ARCEQM_H_
#define _ARCEQM_H_

#ifndef PM_EXPORT
# define PM_EXPORT
#endif

#ifndef ulong
# define ulong UINT32
#endif

#define PRP_TRIGGER 1
#define PRP_REMOVE 2
#define PRP_ANNOTATE 3
#define PRP_CLOSEARCHIVE 4
#define PRP_ADEVICES 50
#define PRP_ARCHINFO 51
#define PRP_ARCHSTATE 52
#define PRP_ARCHWINDOW 53
#define PRP_PRPINFO 54
#define PRP_ARCHDATA 55
#define PRP_ARCHHEADER 56
#define PRP_NUMEVENTS 57
#define PRP_EVENTS 58
#define PRP_NDSERVERS 59
#define PRP_DSERVERS 60
#define PRP_NADEVICES 61
#define PRP_NAPROPS 62
#define PRP_NDEVEVENTS 63
#define PRP_APROPS 64
#define PRP_DEVEVENTS 65
#define PRP_TRIGGERS 66
#define PRP_TRC_COMPONENTS 67
#define PRP_TRC_COMPONENTS_NAM 68
#define PRP_TRC_CHANNELS 69
#define PRP_TRC_CHANNELS_NAM 70

extern char ArchiveDirectoryPath[128];
#define TRCVIEW_FILE "triggers.csv"
#ifdef WIN32
#define CONFIGS_PATH "..\\CONFIGS"
#else
#define CONFIGS_PATH "../CONFIGS"
#endif
typedef struct
{
  char ctx[32]; 
  char srv[32];
  char prp[64];
  char dev[64]; 
} TrgChannel;

typedef struct
{
  char fileName[64];
  char desc[64];
  TrgChannel *channels;
  int nChannels;
} TrgComponent;

typedef struct
{
  char ctx[32];
  char trg[64];
  char desc[64];
  char file[64];
  TrgComponent *components;
  int nComponents;
} TrgViewFile;

typedef struct
{
  short numProperties;
  short numDevices;
  long  eventCode;
  char  comment[64];
} AIB;  /* Archive Info Block */
#define AIBStructSize (2*2+4+64)
typedef struct
{
  float SamplingRate;
  long  numPreTriggerValues;
  long  TimeOffset;
  long  LowCount;
  long  HighCount;
  char  Units[16];
  double LowEGU;
  double HighEGU;
  short  numDevices;
  short  reserved;
} PIB;  /* Property Info Block */
#define PIBStructSize (5*4+16+2*8+2*2)

typedef struct
{
   char channelID[32]; /* Kanalidentifikator                         */
   UINT32 eventcode;   /* Ereignisbewerter                           */
   UINT16 arraysize;   /* Anzahl der Daten                           */
   UINT16 datatype;    /* Datentyp                                   */
   float samplingrate; /* Abtastfrequenz                             */
   UINT32 pretrigger;  /* Anzahl der Daten vor dem Triggerzeitpunkt  */
   UINT32 timeoffset;  /* Zeitoffset                                 */
   UINT32 lowcount;    /* kleinster Mewert in digits                */
   UINT32 highcount;   /* grter Mewert in digits                  */
   char egu[16];       /* physikalische Einheit                      */
   float egulow;       /* kleinster Mewert in physikalischer Einheit*/
   float eguhigh;      /* grter Mewert in physikalischer Einheit  */
   UINT32 datalength;  /* supercedes arraysize, when arraysize = -1  */
   char desc[4][64];   /* 4x Kommentarzeile                          */
} DHS;  /* Data Header Structure */
#define DHSStuctSize (32 + 4 + 2*2 + 4 + 4*4 + 16 + 4*2 + 4 + 4*64)

typedef struct
{
   UINT32 evnCode;
   UINT32 evnNr;
   UINT32 datSiz;
   UINT16 datFmt;
   UINT16 aryTyp;
   UINT32 preTrg;
   UINT32 reserved;
   float shiftX;
   float scaleX;
   float shiftY;
   float scaleY;
   float rangeMinX; 
   float rangeMaxX;
   float scaleMinY;
   float scaleMaxY;
   char unitsX[16];
   char unitsY[16];
   char desc[256];
   char fcnTag[64];
} EVNDHDR;  /* Event Data Header Structure (modern) */
#define EVNDHDR_SIZE (14*4 + 2*16 + 256 + 64) /* 408 bytes */

typedef struct
{
  char  deviceServerName[16];
  char  deviceProperty[16];
  char  deviceName[16];
  short dataFormat;
  short dataDimension;
} DEVSPEC5; /* Device Specifier */

typedef struct
{
  ulong   eventNumber;
  UINT32  starttime;
  UINT32  stoptime;
  DEVSPEC5 devSpecifier;
} ADRB5; /* Archived Data Request Block */

#define ADRBStructSize (3*4 + 3*16 + 2*2)

typedef struct
{
  char  deviceContext[CONTEXT_NAME_SHORTSIZE];
  char  deviceServerName[EXPORT_NAME_SHORTSIZE];
  char  deviceProperty[PROPERTY_NAME_SHORTSIZE];
  char  deviceName[32];
  short dataFormat;
  short reserved;
  ulong dataIndex;
  ulong dataDimension;
} DEVSPECX; /* Extended Device Specifier */

typedef struct
{
  ulong    eventNumber;
  UINT32   starttime;
  UINT32   stoptime;
  DEVSPECX devSpecifier;
} ADRBX; /* Extended Archived Data Request Block */

#define ADRBXStructSize (3*4 + 2*16 + 2*32 + 2*2 + 2*4)

typedef struct
{
  char  deviceContext[CONTEXT_NAME_SIZE];
  char  deviceServerName[EXPORT_NAME_SIZE];
  char  deviceProperty[PROPERTY_NAME_SIZE];
  char  deviceName[DEVICE_NAME_SIZE];
  char  dataTag[TAG_NAME_SIZE];
  short dataFormat;
  short reserved;
  UINT32 dataIndex;
  UINT32 dataDimension;
} DEVSPEC; /* Device Specifier Release 4 */

typedef struct
{
  UINT32  eventNumber;
  UINT32  starttime;
  UINT32  stoptime;
  DEVSPEC devSpecifier;
} ADRB; /* Extended Archived Data Request Block */

int getArcHdr(DTYPE *dout, UINT32 evnNr, char *evnTrg, DEVSPEC *devsp);
void checkSetEvent(UINT32 eventnumber, int cc);
void checkRecordEvent(int cc);
int GetLocalSessionID(UINT32 eventNumber, char *ext);
int GetStoredDevices(NAME64 *devnames, char *devsrv, char *trigger, UINT32 evnTime, int num);
int GetStoredProperties(NAME64 *props, char *devsrv, char *trg, int num);
int respondToTrigger(UINT32 eventNumber, UINT32 eventCode);
int getArchiveData(DTYPE *d, UINT32 evnNr, UINT32 start, UINT32 stop, char *evnTrg, DEVSPEC *devsp);
int GetLocalArchiveData(DTYPE *d, UINT32 evnNr, UINT32 start, UINT32 stop, char *evnTrg, DEVSPEC5 *devsp5);
int GetLocalArchiveDataEx(DTYPE *d, UINT32 evnNr, UINT32 start, UINT32 stop, char *evnTrg, DEVSPECX *devspx);
int closeArchiveFile(UINT32 evnNr, char *evnTrg);
int RemoveLocalArchiveEvent(UINT32 evntime, char *dsrv);
int RemoveLocalArchiveEvents(UINT32 evnstart, UINT32 evnstop, char *dsrv);
int AnnotateLocalArchiveEvent(UINT32 evntime, char *devName, char *cmt);
int GetLocalArchiveInfo(AIB *aib, UINT32 eventNumber, char *deviceServer);
int GetLocalPropertyInfo(char *fullName, UINT32 eventNumber, PrpQueryStruct *pqs);
int getArchiveHeader(DHS *dhs, UINT32 evnNr, UINT32 start, UINT32 stop, char *evnTrg, DEVSPEC *devsp);
int GetLocalArchiveHeader(DHS *dhs, UINT32 evnNr, UINT32 start, UINT32 stop, char *evnTrg, DEVSPEC5 *devsp5);
int GetLocalArchiveHeaderEx(DHS *dhs, UINT32 evnNr, UINT32 start, UINT32 stop, char *evnTrg, DEVSPECX *devspx);
int GetLocalnumDevices(char *devSrvName, char *trigger, UINT32 evnTime);
int GetLocalnumDeviceProperties(char *devSrvName, char *devName);

#define MAKECODE(id,cd) (((id)<<16)|(cd))
#define BITSinBYTE 8
#define EVENT_IGNORE 0xffffffff

#define ARCEQM_TAG "ARCEQM"
#define ARCEQM_EXPORT "EVENTS"
#define ARCEQM_VERSION 1

#endif /* _ARCEQM_H_ */

