/*
 * File arctypes.h
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

int initArchiveServer(void);
int initRPC(void);
int SetGlobalEvent(UINT32 eventCode,char *cmt);
int RecordEvent(UINT32 eventNumber,UINT32 eventCode,char *devName);

PM_EXPORT void setTriggerCallback(int (*fcn)(UINT32 eventNumber,UINT32 eventCode));
PM_EXPORT void setRecordEventCallback(void (*fcn)(int cc));
PM_EXPORT void setSetEventCallback(void (*fcn)(UINT32 evntnr,int cc));
PM_EXPORT void setRemoveArchiveEventFunction(int (*fcn)(UINT32 eventNumber,char *dsrv));
PM_EXPORT void setRemoveArchiveEventsFunction(int (*fcn)(UINT32 eventFirst,UINT32 eventLast,char *dsrv));
PM_EXPORT void setAnnotateArchiveEventFunction(int (*fcn)(UINT32 eventNumber,char *dsrv,char *cmt));
PM_EXPORT void setGetDataFunction(int (*fcn)(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devSpec));
PM_EXPORT void setGetHeaderFunction(int (*fcn)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec));
PM_EXPORT void setGetHeaderDispatch(int (*fcn)(DTYPE *dout,UINT32 evnNumber,char *evnContext,DEVSPEC *devspec));
PM_EXPORT void setGetArchiveDataFunction(int (*fcn)(DTYPE *d,UINT32 eventNumber,UINT32 starttime,UINT32 stoptime,char *eventContext,DEVSPEC5 *devSpecifier));
PM_EXPORT void setGetArchiveHeaderFunction(int (*fcn)(DHS *dhs,UINT32 eventNumber,UINT32 starttime,UINT32 stoptime,char *eventContext,DEVSPEC5 *devSpecifier));
PM_EXPORT void setQueryArchiveInfoFunction(int (*fcn)(AIB *aib,UINT32 eventNumber,char *deviceServer));
PM_EXPORT void setQueryPropertyInfoFunction(int (*fcn)(PIB *pib,UINT32 eventNumber,char *deviceServer,char *devProperty));
PM_EXPORT void setQueryPropertyInfoExFunction(int (*fcn)(char *fullName,UINT32 evncode,PrpQueryStruct *pqs));
PM_EXPORT void setQueryNumLocalArchivedEventsFunction(int (*fcn)(char *arcctx,UINT32 starttime,UINT32 stoptime,UINT32 eventNumber));
PM_EXPORT void setQueryLocalArchivedEventsFunction(int (*fcn)(char *arcctx,UINT32 starttime,UINT32 stoptime,UINT32 *evnlst,UINT32 num));
PM_EXPORT void setQueryNumLocalDeviceServersFunction(int (*fcn)(void));
PM_EXPORT void setQueryLocalTriggersFunction(int (*fcn)(NAME64 *ctxs,char *srvname,char *prpname,int num));
PM_EXPORT void setQueryNumLocalDevicesFunction(int (*fcn)(char *devServer,char *context,int id));
PM_EXPORT void setQueryNumLocalDevicePropertiesFunction(int (*fcn)(char *devServer,char *devName));
PM_EXPORT void setQueryLocalDeviceNamesFunction(int (*fcn)(NAME64 *names,char *devServer,char *context,int id,int num));
PM_EXPORT void setQueryLocalDeviceServerNamesFunction(int (*fcn)(char *arcctx,NAME32 *names,UINT32 start,UINT32 stop,int event,int num));
PM_EXPORT void setQueryLocalDeviceServerNamesExFunction(int (*fcn)(char *arcctx,NAME64 *names,UINT32 start,UINT32 stop,int event,int num));
PM_EXPORT void setQueryLocalDevicePropertiesFunction(int (*fcn)(NAME64 *properties,char *devSrvName,char *devName,int num));
PM_EXPORT void setQueryNumDeviceArchivedEventsFunction(int (*fcn)(char *devsrv,char *device,UINT32 starttime,UINT32 stoptime));
PM_EXPORT void setQueryDeviceArchivedEventsFunction(int (*fcn)(char *arcctx,char *devsrv,char *dev,UINT32 starttime,UINT32 stoptime,AES *aes,UINT32 num,int reverseSort));
PM_EXPORT void setQueryDeviceArchivedEventListFunction(int (*fcn)(char *arcctx,char *srv,char *dev,UINT32 starttime,UINT32 stoptime,UINT32 *evn,UINT32 num,int reverseSort));
PM_EXPORT void setCloseArchiveFileDispatch(int (*fcn)(UINT32 evnNumber,char *evnTrigger));

int GetNumberOfEvents(UINT32 starttime,UINT32 stoptime);
int GetEvents(UINT32 starttime,UINT32 stoptime,UINT32 eventCode,AES *aes,int *num);
int GetNumberOfArchiveServers(void);
int GetArchiveServers(int idSrv,NAME32 *asrv,int *num);
int GetNumberOfDeviceServers(char *archiveServer);
int GetDeviceServersX(char *archiveServer,int idDev,NAME32 *dsrv,int *num);
int GetRemoteArchiveData(char *archiveServer,BYTE *data,UINT32 eventNumber,UINT32 starttime,UINT32 stoptime,DEVSPEC5 *devSpecifier);
int GetRemoteArchiveHeader(char *archiveServer,DHS *dhs,UINT32 eventNumber,UINT32 starttime,UINT32 stoptime,DEVSPEC5 *devSpecifier);

int GetArchiveInfo(char *archiveServer,char *deviceServer,UINT32 evnnum,AIB *aib);
int GetPropertyInfo(char *archiveServer,char *deviceServer,char *deviceProperty,UINT32 evnnum,PIB *pib);
int GetNumArchiveServerEvents(char *archiveServer,UINT32 starttime,UINT32 stoptime,UINT32 eventNumber);
int GetNumArchiveDeviceServers(char *archiveServer);
int GetNumArchiveDevices(char *archiveServer,char *devServer);
int GetNumArchiveDeviceProperties(char *archiveServer,char *devServer);
int GetArchiveDeviceServerNames(char *archiveServer,unsigned long start,unsigned long stop,NAME32 *names,int *num);
#define GetArchiveDeviceServers(as,start,stop,nam,no) GetArchiveDeviceServerNames(as,start,stop,nam,no)
int GetArchiveDeviceNames(char *archiveServer,char *devServer,NAME64 *names,int *num);
#define GetDeviceNames(as,ds,nam,no) GetArchiveDeviceNames(as,ds,nam,no)
#define GetArchiveDevices(as,ds,nam,no) GetArchiveDeviceNames(as,ds,nam,no)
int GetArchiveDeviceProperties(char *archiveServer,char *devServer,char *devName,NAME64 *properties,int *num);
int GetArchivedEvents(char *archiveServer,UINT32 starttime,UINT32 stoptime,UINT32 eventCode,AES *aes,int *num);
int GetNumDeviceArchivedEvents(char *archiveServer,char *devServer,char *device,UINT32 starttime,UINT32 stoptime);
int GetDeviceArchivedEvents(char *archiveServer,char *devServer,char *device,UINT32 starttime,UINT32 stoptime,AES *aes,int *num);
int GetDeviceArchivedEventList(char *archiveServer,char *devServer,char *device,UINT32 starttime,UINT32 stoptime,UINT32 *evnlst,int *num);
