#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "tine.h"
#include "listener.h"

#define RPTREQM "RPTEQM"
#define RPTR_DEFAULT_PORT 101
#define RPTR_MAX_NDEVICES 5000
#define RPTR_POLLRATE 1000

int gTruncateDeviceNames = 0;
int rpPollingRate = RPTR_POLLRATE;
NAME64 *rpDeviceNames = NULL;
PropertyQueryStruct *rpProperties = NULL;
int rpteqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void rptini(void);
void rptbkg(void);
void rptexi(void);

char gTargetDeviceServer[256];
int GetNumberOf(char *srv,char *stkprp)
{
  DTYPE dout;
  int cc = 0;
  short n;
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = &n;
  if ((cc=ExecLinkEx(srv,stkprp,&dout,NULL,CA_READ,800)) != 0 && !(cc & CE_SENDDATA)) return -cc;
  return n;
}
int rpPreSystemInit(char *tgtContext,char *tgtServer) 
{
  int cc=0;
  char *ptr = NULL;
  char thisFec[FEC_NAME_SIZE+4], thisContext[CONTEXT_NAME_SIZE+4], thisServer[EXPORT_NAME_SIZE+4];
  char thisDescription[FEC_DESC_SIZE], thisHost[32], thisUser[32];
  UINT16 portOffset = RPTR_DEFAULT_PORT;

  if ((ptr=getenv("TINE_PORT")) != NULL) portOffset = (UINT16)atoi(ptr);
  if ((ptr=getenv("TINE_CONTEXT")) == NULL) 
  {
    strncpy(thisContext,tgtContext,CONTEXT_NAME_SIZE);
  }
  else
  {
    strncpy(thisContext,ptr,CONTEXT_NAME_SIZE); 
  }
  thisContext[CONTEXT_NAME_SIZE] = 0;
  if ((ptr=getenv("TINE_SERVER")) == NULL) 
  {
    strncpy(thisServer,tgtServer,12); thisServer[12] = 0;
    strcat(thisServer,".RPT");
  }
  else
  {
    strncpy(thisServer,ptr,EXPORT_NAME_SIZE);
  }
  thisServer[EXPORT_NAME_SIZE] = 0;
  strncpy(thisFec,thisServer,FEC_NAME_SIZE);
  if ((ptr=getenv("HOSTNAME")) == NULL && (ptr=getenv("COMPUTERNAME")) == NULL) ptr = "local.host";
  strncpy(thisHost,ptr,32);
  sprintf(thisDescription,"/%s/%s repeater",tgtContext,tgtServer);
  if ((ptr=getenv("USER")) == NULL && (ptr=getenv("USERNAME")) == NULL) ptr = "admin";
  strncpy(thisUser,ptr,16); 
  thisUser[16] = 0;

  if ((cc=RegisterFecInformation(thisFec,"RPT",thisContext,thisDescription,thisHost,"none",thisUser,portOffset)) != 0)
  {
    feclog("could not register repeater FEC name : %s",erlst[cc]);
  }
  return 0;
}
int rpPostSystemInit(char *tgtContext,char *tgtServer)
{
  int cc=0,i,n,ndevices,nproperties;
  char *ptr = NULL;
  char thisFec[FEC_NAME_SIZE+4], thisContext[CONTEXT_NAME_SIZE+4], thisServer[EXPORT_NAME_SIZE+4];
  char thisDescription[FEC_DESC_SIZE], thisHost[32], thisUser[32];
  char tgt[256], dname[DEVICE_NAME_SIZE+4];
  UINT16 portOffset = RPTR_DEFAULT_PORT;

  sprintf(tgt,"/%s/%s/#0",tgtContext,tgtServer);
  sprintf(gTargetDeviceServer,"/%s/%s/",tgtContext,tgtServer);
  if ((ndevices=GetNumberOf(tgt,"NDEVICES")) < 0)
  {
    dbglog("cannot get number of devices from target %s: %s",tgt,erlst[-ndevices]);
    exit(1);
  }
  if ((rpDeviceNames=(NAME64 *)calloc(ndevices,sizeof(NAME64))) == NULL)
  {
    dbglog("cannot allocate space for device names!");
    exit(1);
  }
  n = ndevices;
  if ((cc=GetSystemDevices(tgt,NULL,rpDeviceNames,&n)) != 0)
  {
    dbglog("cannot get device list from target %s: %s",tgt,erlst[cc]);
    exit(1);
  }
  if ((nproperties=GetNumberOf(tgt,"NPROPS")) < 0)
  {
    dbglog("cannot get number of properties from target %s: %s",tgt,erlst[-ndevices]);
    exit(1);
  }
  if ((rpProperties=(PropertyQueryStruct *)calloc(nproperties,sizeof(PropertyQueryStruct))) == NULL)
  {
    dbglog("cannot allocate space for property list!");
    exit(1);
  }
  n = nproperties;
  if ((cc=GetDevicePropertyInformation(tgt,rpProperties,&n)) != 0)
  {
    dbglog("cannot get property list from target %s: %s",tgt,erlst[cc]);
    exit(1);
  }

  if ((ptr=getenv("TINE_PORT")) != NULL) portOffset = (UINT16)atoi(ptr);
  if ((ptr=getenv("TINE_CONTEXT")) == NULL) ptr = tgtContext;
  strncpy(thisContext,ptr,CONTEXT_NAME_SIZE); 
  thisContext[CONTEXT_NAME_SIZE] = 0;
  if ((ptr=getenv("TINE_SERVER")) == NULL) 
  {
    strncpy(thisServer,tgtServer,EXPORT_NAME_SIZE-4); thisServer[EXPORT_NAME_SIZE-4] = 0;
    strcat(thisServer,".RPT");
  }
  else
  {
    strncpy(thisServer,ptr,EXPORT_NAME_SIZE);
  }
  thisServer[16] = 0;
  strncpy(thisFec,thisServer,8);
  thisFec[8] = 0;
  sprintf(&thisFec[strlen(thisFec)],".SRV.%d",(portOffset%1000));
  thisFec[16] = 0;
  if ((ptr=getenv("HOSTNAME")) == NULL && (ptr=getenv("COMPUTERNAME")) == NULL) ptr = "local.host";
  strncpy(thisHost,ptr,32);
  sprintf(thisDescription,"/%s/%s repeater",tgtContext,tgtServer);
  if ((ptr=getenv("USER")) == NULL && (ptr=getenv("USERNAME")) == NULL) ptr = "admin";
  strncpy(thisUser,ptr,16); 
  thisUser[16] = 0;

  if (gTruncateDeviceNames > 0) ndevices = gTruncateDeviceNames;
  n = ndevices;
  if ((cc=RegisterEquipmentModule(thisServer,RPTREQM,n,rpteqm,rptini,rptbkg,1000,rptexi)) != 0)
  {
    feclog("could not register repeater equipment module : %s",erlst[cc]);
  }

  dname[0] = 0;
  for (i=0,n=0; i<ndevices; i++)
  {
    strncat(dname,rpDeviceNames[i].name,DEVICE_NAME_SIZE);
    if (strlen(dname) == 0) continue;
    if ((cc=RegisterDeviceName(RPTREQM,dname,n++)) != 0)
    {
      dbglog("cannot register device name %s : %s",dname,erlst[cc]);
    }
    dname[0] = 0;
  }

  for (i=0; i<nproperties; i++)
  {
    if (strstr(rpProperties[i].prpName,".HIST") != NULL) continue;
    cc = RegisterProperty(RPTREQM,rpProperties[i].prpName,rpProperties[i].prpSize,
           rpProperties[i].prpFormat,rpProperties[i].prpAccess,rpProperties[i].prpDescription);
    if (cc < 0)
    {
      dbglog("cannot register property name %s : %s",rpProperties[i].prpName,erlst[-cc]);
    }
  }  
  return 0;
}
int rpteqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char lclTargetDeviceName[256];
  DTYPE *dinptr=NULL;
  int rate;
  strcpy(lclTargetDeviceName,gTargetDeviceServer);
  strcat(lclTargetDeviceName,devName);
  if (din != NULL && din->dArrayLength > 0) dinptr = din;
  rate = access & CA_LAST ? 800 : rpPollingRate;
  return alsnExecLink(lclTargetDeviceName,devProperty,dout,dinptr,access,rate);
}
void rptini(void)
{
}
void rptbkg(void)
{
}
void rptexi(void)
{
}
