/*
* console.c
*
* Provides helper routines for manipulating the application console.
*
*  Created on: Nov 28, 2015
*      Author: Jaka Bobnar
*/
#include "console.h"
#include "project.def"

#if defined(UNIX)

int DisableCloseWindow(void)
{
    return 1;
}

void EnableCloseWindow(void)
{
}

#elif defined(WIN32)

#include <Windows.h>
#include <shellapi.h>
#include <math.h>
#include "tine.h"
#include "Resource.h"
#include "util.h"
#include "server.h"

#define WM_TRAYICON 0x0401
#define WM_MENU_EXIT 0x0402
#define WM_MENU_SHOW 0x0403
#define WM_MENU_HIDE 0x0404

NOTIFYICONDATA tray;
HWND consoleWindow;
HINSTANCE hInstance;
HWND hwndWindow;
HMENU trayMenu;

int gMenuItemIndex = -1;
int gMenuItemOldState = 0;

HWND console;

void BreakHandler(int val)
{
    SystemKillCycleTimer();
    StopServer();
    //give the server time to terminate gracefully
    Wait(300);
    Shutdown();
}

void DestroyConsole(void)
{
    Shell_NotifyIcon(NIM_DELETE, &tray);
    DestroyWindow(hwndWindow);
    DestroyWindow(consoleWindow);
    PostQuitMessage(0);
}

void CheckMessages(void)
{
    MSG msg;
    while (PeekMessage(&msg, hwndWindow, 0, 0, PM_REMOVE))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
}

void HideConsole(void)
{
    ShowWindow(consoleWindow, SW_HIDE);
}

void ShowConsole(void)
{
    ShowWindow(consoleWindow, SW_SHOW);
    SetWindowPos(consoleWindow, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    //LONG_PTR a = SetWindowLongPtr(consoleWindow, GWLP_WNDPROC, (LONG_PTR)&WndProc);
    if (message == WM_TRAYICON)
    {
        if (lParam == WM_RBUTTONUP || lParam == WM_LBUTTONUP)
        {
            POINT curPoint;
            GetCursorPos(&curPoint);
            SetForegroundWindow(hwnd);
            UINT clicked = TrackPopupMenu(trayMenu, TPM_RETURNCMD, curPoint.x, curPoint.y, 0, hwnd, NULL);

            if (clicked == WM_MENU_EXIT)
            {
                BreakHandler(0);
            }
            else if (clicked == WM_MENU_SHOW)
            {
                ShowConsole();
            }
            else if (clicked == WM_MENU_HIDE)
            {
                HideConsole();
            }
        }
        else if (lParam == WM_LBUTTONDBLCLK)
        {
            ShowConsole();
        }
    }
    else if (message == WM_CREATE)
    {
        MENUITEMINFO separatorBtn = { 0 };
        separatorBtn.cbSize = sizeof(MENUITEMINFO);
        separatorBtn.fMask = MIIM_FTYPE;
        separatorBtn.fType = MFT_SEPARATOR;

        trayMenu = CreatePopupMenu();
        InsertMenu(trayMenu, -1, MF_BYPOSITION | MF_STRING, WM_MENU_SHOW, L"Show Console");
        InsertMenu(trayMenu, -1, MF_BYPOSITION | MF_STRING, WM_MENU_HIDE, L"Hide Console");
        InsertMenuItem(trayMenu, -1, FALSE, &separatorBtn);
        InsertMenu(trayMenu, -1, MF_BYPOSITION | MF_STRING, WM_MENU_EXIT, L"Exit");
    }
    else if (message == WM_SYSCOMMAND)
    {
        if ((wParam & SC_MINIMIZE) == SC_MINIMIZE)
        {
            HideConsole();
        }
    }

    return DefWindowProc(hwnd, message, wParam, lParam);
}

void IconToTray(void)
{
    size_t s;
    memset(&tray, 0, sizeof(NOTIFYICONDATA));
    consoleWindow = GetConsoleWindow();
    hInstance = GetModuleHandle(0);

    WNDCLASS wc = { 0 };
    wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wc.hCursor = LoadCursor(hInstance, IDC_ARROW);
    wc.hIcon = LoadIcon(hInstance, IDI_APPLICATION);
    wc.hInstance = hInstance;
    wc.lpfnWndProc = WndProc;
    wc.lpszClassName = TEXT("consoledaemon");
    wc.style = CS_HREDRAW | CS_VREDRAW;
    RegisterClass(&wc);

    hwndWindow = CreateWindow(L"consoledaemon", L"Console Daemon", 0, 0, 0, 30, 30, NULL, NULL, hInstance, NULL);

    tray.cbSize = sizeof(tray);
    tray.hIcon = LoadIcon(hInstance, (LPCTSTR)MAKEINTRESOURCE(IDI_ICON)); 
    tray.hWnd = hwndWindow;
    mbstowcs_s(&s, tray.szTip, 128, "Console Daemon", _TRUNCATE);
    mbstowcs_s(&s, tray.szInfo, 256, "Click the banana icon to show the console or shut down the daemon.", _TRUNCATE);
    mbstowcs_s(&s, tray.szInfoTitle, 64, "Console Daemon is hidden", _TRUNCATE);
    //mbstowcs_s(&s, tray.szInfo, 256, "", _TRUNCATE);
    tray.uCallbackMessage = WM_TRAYICON;
    tray.uFlags = NIF_INFO | NIF_ICON | NIF_TIP | NIF_MESSAGE;
    tray.uID = 147258369;
    Shell_NotifyIcon(NIM_ADD, &tray);

    ShowWindow(hwndWindow, SW_HIDE);
    HideConsole();
}

boolean DisableCloseWindow(void)
{
    HWND hwnd;
    HMENU hmenu;

    //RECT rect;
    CONSOLE_SCREEN_BUFFER_INFOEX cinfo = { 0 };
    CONSOLE_SCREEN_BUFFER_INFO ccinfo = { 0 };
    
    // disables ALT-F4, disables Close default window (system) menu entry, disables 'x' on title bar 
    hwnd = GetConsoleWindow();
    if (hwnd == NULL) return FALSE;

    hmenu = GetSystemMenu(hwnd, FALSE);
    if (hmenu == NULL) return FALSE;

    int nCnt = GetMenuItemCount(hmenu);

    for (int i = 0; i<nCnt; i++)
    {
        MENUITEMINFO mii;

        mii.fMask = MIIM_ID | MIIM_STATE;
        mii.cbSize = sizeof(MENUITEMINFO);

        BOOL retB = GetMenuItemInfo(hmenu, i, TRUE, &mii);

        if (retB == TRUE)
        {
            if (mii.wID == SC_CLOSE)
            {
                gMenuItemOldState = mii.fState;

                mii.fState = MFS_GRAYED;
                mii.wID = 0;

                retB = SetMenuItemInfo(hmenu, i, TRUE, &mii);
                if (retB == TRUE)
                {
                    gMenuItemIndex = i;
                    DrawMenuBar(hwnd);
                    return TRUE;
                }

            }
        }
    }
    return FALSE;
}

void EnableCloseWindow(void)
{
    // FIXME: if someone modified the system menu in the meantime, the old index will no longer be valid! SWEISSE Jul 27, 2015
    MENUITEMINFO mii;
    HWND hwnd;
    HMENU hmenu;

    if (gMenuItemIndex == -1) return;

    hwnd = GetConsoleWindow();
    if (hwnd == NULL) return;

    hmenu = GetSystemMenu(hwnd, FALSE);
    if (hmenu == NULL) return;

    mii.cbSize = sizeof(MENUITEMINFO);
    mii.fMask = MIIM_ID | MIIM_STATE;

    BOOL retB = GetMenuItemInfo(hmenu, gMenuItemIndex, TRUE, &mii);

    if (retB == FALSE) return;

    if (mii.wID != 0) return; // for safety

    mii.wID = SC_CLOSE;
    mii.fState = gMenuItemOldState;

    retB = SetMenuItemInfo(hmenu, gMenuItemIndex, TRUE, &mii);

    DrawMenuBar(hwnd);
}
#endif
