/*
* serverdef.h
*
* Provides constants and structures required by the watchdog server.
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#ifndef SERVERDEF_H_
#define SERVERDEF_H_

#define LOCAL_NAME "WDOG" //the local name of the server
#define DEVICE_PROPERTY_LIST_NAME "WDOG_DEVICE_PROPERTY_LIST" //the property list name for the devices
#define WDOG_PROPERTY_LIST_NAME "WDOG_PROPERTY_LIST" //the watchdog property list

//property IDs
#define PRP_INFO                    0
#define PRP_CPU                     1
#define PRP_MEMORY                  2
#define PRP_NUM_CPU_VIOLATIONS      3
#define PRP_NUM_MEMORY_VIOLATIONS   4
#define PRP_NOT_RUNNING_SINCE       5
#define PRP_LAST_RETRY              6
#define PRP_CURRENT_RETRY           7
#define PRP_STATUS                  8
#define PRP_START                   9
#define PRP_RESTART                 10
#define PRP_START_TINE              11
#define PRP_STOP_TINE               12
#define PRP_STOP                    13
#define PRP_LAST_TINE_RESPONSE      14
#define PRP_LAST_TINE_RETRY         15
#define PRP_TINE_STATUS             16
#define PRP_NUM_STOPPED_WD          17
#define PRP_NUM_RESTARTED           18
#define PRP_SUSPEND                 19
#define PRP_AVAILABLE_MEMORY        20
#define PRP_CPU_TOTAL               21
#define PRP_SHUTDOWN                22
#define PRP_REBOOT                  23
#define PRP_REMAINING_SUSPENDED     24
#define PRP_RELOAD                  25
#define PRP_NETWORK_NAMES           26
#define PRP_NETWORK_UTILIZATION     27
#define PRP_NETWORK_RELATIVE        28
#define PRP_NETWORK_FRIENDLY_NAMES  29
#define PRP_RESET_STATISTICS        30
#define PRP_LAST_RESTART_REASON     31
#define PRP_MEMORY_MB               32

//property names
#define P_N_INFO                    "Info"
#define P_N_CPU                     "CPU"
#define P_N_MEMORY                  "Memory"
#define P_N_MEMORY_MB               "MemoryMB"
#define P_N_NUM_CPU_VIOLATIONS      "NCPUViolations"
#define P_N_NUM_MEMORY_VIOLATIONS   "NMemoryViolations"
#define P_N_NOT_RUNNING_SINCE       "NotRunningSince"
#define P_N_LAST_RETRY              "LastRetry"
#define P_N_CURRENT_RETRY           "CurrentRetry"
#define P_N_STATUS                  "Status"
#define P_N_START                   "Start"
#define P_N_RESTART                 "Restart"
#define P_N_START_TINE              "StartTineMonitor"
#define P_N_STOP_TINE               "StopTineMonitor"
#define P_N_STOP                    "Stop"
#define P_N_LAST_TINE_RESPONSE      "LastTineResponse"
#define P_N_LAST_TINE_RETRY         "LastTineRetry"
#define P_N_TINE_STATUS             "TineStatus"
#define P_N_NUM_STOPPED_WD          "NStoppedByWD"
#define P_N_NUM_RESTARTED           "NStartedByWD"
#define P_N_SUSPEND                 "Suspend"
#define P_N_AVAILABLE_MEMORY        "AvailableMemory"
#define P_N_CPU_TOTAL               "CPUTotal"
#define P_N_SHUTDOWN                "Shutdown"
#define P_N_REBOOT                  "Reboot"
#define P_N_REMAINING_SUSPENDED     "RemainingSuspendedTime"
#define P_N_RELOAD                  "Reload"
#define P_N_NETWORK_NAMES           "NetworkInterfaceNames"
#define P_N_NETWORK_UTILIZATION     "NetworkLoad"
#define P_N_NETWORK_RELATIVE        "NetworkLoadRelative"
#define P_N_NETWORK_FRIENDLY_NAMES  "NetworkInterfaceFriendlyNames"
#define P_N_RESET_STATISTICS        "ResetStatistics"
#define P_N_LAST_RESTART_REASON     "LastRestartReason"

//user commands
#define C_WATCHDOG_DEBUG    "watchdogDebug"
#define C_RESTART           "restart"
#define C_STOP              "stop"
#define C_START             "start"
#define C_LINK_RESTART      "linkRestart"
#define C_LINK_STOP         "linkStop"
#define C_LINK_START        "linkStart"
#define C_SUSPEND           "suspend"
#define C_RESUME            "resume"
#define C_RELOAD            "reload"
#define C_PROC_LIST         "proclist"
#define C_PROC_INFO         "procinfo"
#define C_SHUTDOWN          "shutdown"
#define C_RESET_STATISTICS  "resetStat"

#define RESTARTED_BY_USER_MESSAGE "RESTARTED BY USER"
#define STOPPED_BY_USER_MESSAGE "STOPPED BY USER"
#define CPU_TOO_HIGH_MESSAGE "CPU TOO HIGH"
#define MEMORY_TOO_HIGH_MESSAGE "MEMORY TOO HIGH"
#define LINK_TIMEOUT_MESSAGE "LINK TIMEOUT"
#define UNKNOWN_MESSAGE "PROCESS NOT RUNNING"
#define NONE_MESSAGE "NOT STARTED BY WATCHDOG"

//alarm definitions
#define A_WARN_CPU_TOO_HIGH 5101
#define A_WARN_MEMORY_TOO_HIGH 5102
#define A_WARN_LINK_TIMEOUT 5103
#define A_ERR_CPU_TOO_HIGH 5104
#define A_ERR_MEMORY_TOO_HIGH 5105
#define A_ERR_LINK_TIMEOUT 5106
#define A_ERR_NOT_RUNNING 5107
#define A_ERR_DISABLED 5108

//status bits
#define RUNNING             0x01
#define TRYING_START        0x02
#define PROPERTY_CONNECTED  0x04
#define PROPERTY_TIMEOUT    0x08
#define PROPERTY_DISABLED   0x10
#define CPU_OVERLOAD        0x20
#define MEMORY_OVERLOAD     0x40
#define DISABLED            0x80

#define NUM_PROPERTIES 22 //number of process related properties
#define NUM_WD_PROPERTIES 14 //number of watchdog properties
#define WD_PROCESS_INFO "WDProcessInfo" //the tag for the process info structure
#define WD_STATUS "WDStsBits" //the tag for the status bits

#define CONFIRMATION_TIMEOUT 10 //the timeout in which to confirm the property execution after it has been armed

#define MAX_NUMBER_OF_DEVICES 500 //the maximum number of devices (processes) that one server can have

#define SIZE_TINE_DEVICE 132
#define SIZE_TINE_PROPERTY 64
// WDProcessInfo contains the settings of individual process. It is a read only container initialized at server startup
typedef struct
{
    char alias[64];
    char name[64];
    short is_service;
    char path[256];
    char parameters[512];
    short match_parameters;
    short match_path;
    char working_directory[256];
    char caption[256];
    char preceding_processes[512];
    char environment[512];
    short monitor_n_seconds_after_start;
    short pause_n_seconds_after_start;
    short number_of_retries;
    short retry_interval;
    short auto_start;
    short max_cpu_load;
    short number_of_allowed_cpu_load_violations;
    float max_memory;
    short number_of_allowed_memory_violations;
    char tine_device[SIZE_TINE_DEVICE]; //32 (context) + 32 (server) + 64 (device) + 4 (delimiters)
    char tine_property[SIZE_TINE_PROPERTY];
    short polling_rate;
    short timeout;
    int timeout_restart;

} WDProcessInfo;

#endif
