#if defined (WIN32) && !defined(TINE32_DLL) || defined(STATIC)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#if defined(UNIX)
#include <dlfcn.h>
#endif
#include "tine.h"
#include "threader.h"
#include "watchdog.h"
#include "util.h"

int gSystemInitialized = FALSE;
int cycle_busy = FALSE;
int cycleTerminated = FALSE;
unsigned int idCycleTimer = 0;
int userNameSet = FALSE;
int gSystemDelay = 20;

int gUseWindowsTimer = FALSE;
int gWindowsTimerInUse = FALSE;
int gThreadTimerInUse = TRUE;
int cycleStarted = FALSE;
THRHANDLE hTmrThread = (THRHANDLE)0;

#if defined(WIN32)
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
#elif defined(UNIX)
pthread_attr_t gCycleThreadAttr, *gPtrCycleThreadAttr = &gCycleThreadAttr;
pthread_mutex_t cycleStartMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t *hCyclerStartMutex = &cycleStartMutex;
pthread_mutex_t cycleStopMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t *hCyclerStopMutex = &cycleStopMutex;
#elif defined(VXWORKS)
#define MULTITHREADED
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
extern int gVxwDefaultTaskOptions;
extern int gVxwDefaultTaskPriority;
extern int gVxwDefaultTaskStack;
#endif

#ifndef VXWORKS
/* use for debugging purposes ... */
void cyclog(char *txt, ...)
{
    static char fn[64] = "cycler.log";
    FILE *fp = NULL;

    char str[1024];
    long flen;
    int cc = 0;
    va_list args;

    va_start(args, txt);
    vsnprintf_s(str, 1024, 512, txt, args); str[512] = 0;
    va_end(args);
    if (str[strlen(str) - 1] != '\n') strcat_s(str, 1024, "\n");
    fopen_s(&fp, fn, "a");
    if (fp == NULL) ccerr(file_error);
    fseek(fp, 0L, SEEK_END); flen = ftell(fp);
    fwrite(str, strlen(str), 1, fp);
err:
    if (fp != NULL) fclose(fp);
    return;
}
#endif
void InitCycleMutexSet(void);

#if defined(UNIX) && !defined(SOLARIS)
#ifndef TINE_START_CYCLER
# define TINE_START_CYCLER FALSE
#endif
__attribute__((constructor)) void TineLoadCycler(void)
{
    int startCycler = TINE_START_CYCLER;
    char *ptr = getenv("TINE_START_CYCLER");
    if (ptr != NULL)
    {
        startCycler = stricmp(ptr, "TRUE") == 0 ? TRUE : FALSE;
    }
    if (startCycler)
    { /* start cycler immediately */
        SystemSetCycleTimer();
    }
    else
    { /* let the app manage the cycling */
        InitCycleMutexSet();
    }
}
__attribute__((destructor)) void TineStopCycler(void)
{
    if (cycleTerminated) return;
    SystemKillCycleTimer();
    CloseClientMutexSet();
    millisleep(20);
}
#endif
void InitCycleMutexSet(void)
{
    static int done = FALSE;
    if (done) return;
    done = TRUE;
    InitMutex(&hCyclerStartMutex);
    InitMutex(&hCyclerStopMutex);
# if defined(UNIX)
    pthread_mutex_lock(hCyclerStartMutex);
    pthread_mutex_lock(hCyclerStopMutex);
# endif
    ReleaseSystemMutex(hCyclerStartMutex);
    ReleaseSystemMutex(hCyclerStopMutex);
}

void cycleTmrTask(void)
{
    if (!gSystemInitialized)
    { /* don't spin in not initialized yet! */
        millisleep(gSystemDelay);
        return;
    }
    if (cycleTerminated) return;
    if (cycle_busy) return;
    cycle_busy = TRUE;
    SystemCycle(FALSE);
    cycle_busy = FALSE;
}

void cycleInitTask(int isServer)
{
    if (isServer)
    {
        SystemInit(FALSE);
    }
    else
    {
        initClient();
    }
    gSystemInitialized = TRUE;
    MaxPollingRate = MinPollingRate = 20;
}

THREAD_TYPE cycleTmrThread(THREAD_PARAM lpvThreadParam)
{
    lpvThreadParam = lpvThreadParam;

    if (WaitForSystemMutex(hCyclerStartMutex, -1) != 0)
    {
        feclog("cycleTmrThread: start cycler mutex unavailable !");
        return 0;
    }
    UseMultiThreadedCycler(TRUE);
    while (!cycleTerminated)
    {
        cycleTmrTask();
    }
    cycleStarted = FALSE;
    CloseServers(TRUE);
    FreeEQPModules();
    ReleaseSystemMutex(hCyclerStartMutex);
    return 0;
}
#if defined(WIN32)
SECURITY_ATTRIBUTES tmrSA;
SECURITY_DESCRIPTOR tmrSD;
DWORD dwTmrThreadId;
THRHANDLE cycleCreateThread(void)
{
    THRHANDLE hThread = 0;
    int threadPriority = -1;
    char *ptr = 0;
    size_t ptr_size = 0;
    _dupenv_s(&ptr, &ptr_size, "TINE_CYCLE_THREAD_PRIORITY");
    //char *ptr = getenv("TINE_CYCLE_THREAD_PRIORITY");

    InitializeSecurityDescriptor(&tmrSD, SECURITY_DESCRIPTOR_REVISION);
    SetSecurityDescriptorDacl(&tmrSD, TRUE, NULL, TRUE);
    tmrSA.nLength = sizeof(tmrSA);
    tmrSA.lpSecurityDescriptor = &tmrSD;
    tmrSA.bInheritHandle = TRUE;

    cycleTerminated = FALSE;
    hThread = (THRHANDLE)CreateThread(&tmrSA, 0, cycleTmrThread, 0, 0, &dwTmrThreadId);
    if (hThread == NULL) OutOfMemoryExit(9);
    if (ptr != NULL) threadPriority = atoi(ptr);
    if (threadPriority > 0)
    {
        feclog("local cycle thread priority set to %d via environment", threadPriority);
        SetThreadPriority(hThread, threadPriority);
    }
    if (ptr != NULL) free(ptr);
    return hThread;
}
#elif defined(UNIX)
THRHANDLE cycleCreateThread(void)
{
    int cc;
    pthread_t tid;

    pthread_attr_init(gPtrCycleThreadAttr);
# ifdef FORCE_PTHREAD_SCOPE_SYSTEM
    if ((cc = pthread_attr_setscope(gPtrCycleThreadAttr, PTHREAD_SCOPE_SYSTEM)) != 0)
    {
        feclog("Cannot set thread scope !\n");
    }
# endif
    cc = pthread_create(&tid, gPtrCycleThreadAttr, cycleTmrThread, NULL);
    return cc ? 0 : (THRHANDLE)tid;
}
#elif defined(VXWORKS)
THRHANDLE cycleCreateThread(void)
{
    static int cnt = 0;
    char tname[32];
    int tskid = taskIdSelf();
    int tskp = gVxwDefaultTaskPriority, tsko = gVxwDefaultTaskOptions;
    int tcks = sysClkRateGet() / 100;
    cnt++;
    strncpy(tname, "tcycler", 8); tname[7] = 0;
    taskPriorityGet(tskid, &tskp);
    taskOptionsGet(tskid, &tsko);
    if (tskp == 1)
    { /* in case started from the shell prompt */
        tskp = gVxwDefaultTaskPriority;
        tsko = gVxwDefaultTaskOptions;
    }
    if (tcks < 1) tcks = 1;
    kernelTimeSlice(tcks);
    return (THRHANDLE)taskSpawn(tname, tskp, tsko, gVxwDefaultTaskStack, (FUNCPTR)cycleTmrThread, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
}
#endif

TINE_EXPORT void killCycleTimer(void)
{
    if (WaitForSystemMutex(hCyclerStopMutex, -1) != 0)
    {
        feclog("SystemKillCycleTimer: stop cycler mutex unavailable !");
        return;
    }
    if (!cycleStarted || cycleTerminated)
    {
        feclog("attempt to kill timer refused : cycle %s", cycleStarted ? "already terminated" : "not started");
        ReleaseSystemMutex(hCyclerStopMutex);
        return;
    }
    cycleTerminated = TRUE;
    CloseClients(TRUE);
    feclog("terminating local cycle");
    millisleep(gSystemDelay * 2);
    CloseSystemMutexSet();
    ReleaseSystemMutex(hCyclerStopMutex);
    if (WaitForSystemMutex(hCyclerStartMutex, gSystemTick) != 0)
    {
        return;
    }
    ReleaseSystemMutex(hCyclerStartMutex);
}
TINE_EXPORT void SystemWaitCycleTimer(void)
{
    if (hTmrThread != (THRHANDLE)0) WaitForThreadTermination(hTmrThread);
}
TINE_EXPORT void SystemKillCycleTimer(void)
{
    killCycleTimer();
}
void TineStartCycler(void)
{
    SystemSetCycleTimer();
}
TINE_EXPORT void SystemStartCycleTimer(void)
{
    SystemSetCycleTimer();
}
void SystemSetCycleTimer(void)
{
    int cc = 0;
    InitCycleMutexSet();
    InitSystemMutexSet();
    if (WaitForSystemMutex(hCyclerStopMutex, -1) != 0)
    {
        feclog("SystemSetCycleTimer: stop cycler mutex unavailable !");
        return;
    }
    if (cycleStarted) ccerr(-1);
    cycleStarted = TRUE;
    cycleInitTask(FALSE);
    hTmrThread = cycleCreateThread();
err:
    ReleaseSystemMutex(hCyclerStopMutex);
    millisleep(20);
    return;
}

#endif /* WIN32 && !TINE32_DLL || STATIC */
