/*
* tineserver.c
*
* Implements routines from server.h as a TINE server.
*
*  Created on: Aug 5, 2014
*      Author: Jaka Bobnar
*/
#include <math.h>
#include "tine.h"
#include "server.h"
#include "util.h"
#include "parser.h"
#include <limits.h>
#include <float.h>
#include "serverdef.h"

time_t shutdown_request; //time at which the shutdown property was armed. allow CONFIRMATION_TIMEOUT seconds to trigger shutdown
time_t reboot_request; //time at which the reboot property was armed. allow CONFIRMATION_TIMEOUT seconds to trigger reboot
char my_ip_address[16]; //the ip address of the watchdog host
boolean ip_set = FALSE;
unsigned int this_wait_period;
time_t server_reload_requested = 0; //a flag, which is set to true when the configuration is being regenerated. When true the server should not do anything to any of the processes.
int fmt[1];
int siz[1];
SERVER_INFO this_info;

WDProcessInfo *wd_process_info = NULL;

/* min() not availale in standard c */
#ifndef min
size_t min(size_t a, size_t b) {
    if (a > b)
        return b;
    else
        return a;
}
#endif


void createStructures(void)
{
    char *temp;
    if (wd_process_info != NULL)
    {
        free(wd_process_info);
        wd_process_info = NULL;
    }
    wd_process_info = (WDProcessInfo*)malloc(sizeof(WDProcessInfo)*num_processes);
    if (wd_process_info == NULL) OutOfMemoryExit(2);
    unsigned int i, j;
    for (i = 0; i < num_processes; i++)
    {
        _snprintf_s(wd_process_info[i].name, 64, 63, "%s", g_process[i].name);
        wd_process_info[i].is_service = g_process[i].is_service;
        _snprintf_s(wd_process_info[i].alias, 64, 63, "%s", g_process[i].alias);
        _snprintf_s(wd_process_info[i].path, 256, 255, "%s", g_process[i].path_to_executable);
        if (g_process[i].parameters == NULL)
            sprintf_s(wd_process_info[i].parameters, 512, "%s", "");
        else
            _snprintf_s(wd_process_info[i].parameters, 512, 511, "%s", g_process[i].parameters);
        wd_process_info[i].match_parameters = g_process[i].match_parameters;
        wd_process_info[i].match_path = g_process[i].match_path;
        if (g_process[i].working_directory == NULL)
            sprintf_s(wd_process_info[i].working_directory, 256, "%s", "");
        else
            _snprintf_s(wd_process_info[i].working_directory, 256, 255, "%s", g_process[i].working_directory);
        if (g_process[i].number_of_preceding_processes > 0) {
            size_t length = 1;
            for (j = 0; j < g_process[i].number_of_preceding_processes; j++)
            {
                length += strlen(g_process[i].preceding_process[j]) + 1;
            }
            temp = (char *)alloca(sizeof(char)*length);
            if (temp == NULL) OutOfMemoryExit(6);
            temp[length - 1] = '\0';
            size_t c = 0;
            for (j = 0; j < g_process[i].number_of_preceding_processes; j++)
            {
                length = strlen(g_process[i].preceding_process[j]);
                memcpy(&temp[c], g_process[i].preceding_process[j], length);
                c += length + 1;
                temp[c - 1] = ';';
            }
            _snprintf_s(wd_process_info[i].preceding_processes, 512, 511, "%s", temp);
        }
        else
        {
            sprintf_s(wd_process_info[i].preceding_processes, 512, "%s", "");
        }
        if (g_process[i].number_of_environment_variables > 0)
        {
            size_t length = 1;
            for (j = 0; j < g_process[i].number_of_environment_variables; j++)
            {
                length += strlen(g_process[i].environment_variables[j]) + 1;
            }
            temp = (char *)alloca(sizeof(char)*length);
            if (temp == NULL) OutOfMemoryExit(6);
            temp[length - 1] = '\0';
            size_t c = 0;
            for (j = 0; j < g_process[i].number_of_environment_variables; j++)
            {
                length = strlen(g_process[i].environment_variables[j]);
                memcpy(&temp[c], g_process[i].environment_variables[j], length);
                c += length + 1;
                temp[c - 1] = ';';
            }
            _snprintf_s(wd_process_info[i].environment, 512, 511, "%s", temp);
        }
        else
        {
            sprintf_s(wd_process_info[i].environment, 512, "%s", "");
        }
        if (g_process[i].caption == NULL)
            sprintf_s(wd_process_info[i].caption, 256, "%s", "");
        else
            _snprintf_s(wd_process_info[i].caption, 256, 255, "%s", g_process[i].caption);
        wd_process_info[i].monitor_n_seconds_after_start = (short)g_process[i].start_after_n_seconds;
        wd_process_info[i].pause_n_seconds_after_start = (short)g_process[i].pause_after_previous_start;
        wd_process_info[i].number_of_retries = (short)g_process[i].number_of_retries;
        wd_process_info[i].retry_interval = (short)g_process[i].retry_interval;
        wd_process_info[i].auto_start = (short)g_process[i].auto_start;
        wd_process_info[i].max_cpu_load = (short)g_process[i].max_cpu_load;
        wd_process_info[i].number_of_allowed_cpu_load_violations = (short)g_process[i].number_of_allowed_cpu_load_violations;
        wd_process_info[i].max_memory = g_process[i].max_memory / (1024.0f * 1024.0f);
        wd_process_info[i].number_of_allowed_memory_violations = (short)g_process[i].number_of_allowed_memory_violations;
        if (g_process[i].tine_device == NULL)
            sprintf_s(wd_process_info[i].tine_device, 132, "%s", "");
        else
            _snprintf_s(wd_process_info[i].tine_device, 132, 131, "%s", g_process[i].tine_device);
        if (g_process[i].tine_property == NULL)
            _snprintf_s(wd_process_info[i].tine_property, 64, 63, "%s", "");
        else
            _snprintf_s(wd_process_info[i].tine_property, 64, 63, "%s", g_process[i].tine_property);
        wd_process_info[i].polling_rate = g_process[i].polling_rate;
        wd_process_info[i].timeout = g_process[i].timeout;
        wd_process_info[i].timeout_restart = g_process[i].timeout_restart;
    }
}

void registerStructsAndBitfields(void)
{
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, alias), 64, CF_TEXT, "alias")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, name), 64, CF_TEXT, "name")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, is_service), 1, CF_SHORT, "isService")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, path), 256, CF_TEXT, "path")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, parameters), 512, CF_TEXT, "parameters")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, match_parameters), 1, CF_SHORT, "matchParameters")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, match_path), 1, CF_SHORT, "matchPath")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, working_directory), 256, CF_TEXT, "workingDir")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, caption), 256, CF_TEXT, "caption")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, preceding_processes), 512, CF_TEXT, "precedingProc")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, environment), 512, CF_TEXT, "envVariables")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, monitor_n_seconds_after_start), 1, CF_SHORT, "monitorAfter")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, pause_n_seconds_after_start), 1, CF_SHORT, "pauseAfter")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, number_of_retries), 1, CF_SHORT, "numOfRetries")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, retry_interval), 1, CF_SHORT, "retryInterval")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, auto_start), 1, CF_SHORT, "autoStart")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, max_cpu_load), 1, CF_SHORT, "maxCPULoad")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, number_of_allowed_cpu_load_violations), 1, CF_SHORT, "nCPUViolations")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, max_memory), 1, CF_FLOAT, "maxMemoryInMB")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, number_of_allowed_memory_violations), 1, CF_SHORT, "nMemViolations")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, tine_device), 132, CF_TEXT, "tineDevice")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, tine_property), 64, CF_TEXT, "tineProperty")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, polling_rate), 1, CF_SHORT, "pollingRate")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, timeout), 1, CF_SHORT, "timeout")) OutOfMemoryExit(2);
    if (addFieldToStruct(WD_PROCESS_INFO, OFFSETIN(WDProcessInfo, timeout_restart), 1, CF_INT32, "restartTimeout")) OutOfMemoryExit(2);
    if (sealTaggedStruct(WD_PROCESS_INFO, sizeof(WDProcessInfo), MAX_NUMBER_OF_DEVICES)) OutOfMemoryExit(2);
    if (OpenBitField(NULL, WD_STATUS, CF_BITFIELD8)) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, RUNNING, "Running")) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, TRYING_START, "Retrying")) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, PROPERTY_CONNECTED, "PropConnected")) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, PROPERTY_TIMEOUT, "PropTimeout")) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, PROPERTY_DISABLED, "PropDisabled")) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, CPU_OVERLOAD, "CPUOverload")) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, MEMORY_OVERLOAD, "MemoryOverload")) OutOfMemoryExit(2);
    if (AddFieldToBitField(NULL, WD_STATUS, DISABLED, "Stopped")) OutOfMemoryExit(2);
}

void registerPropertiesAndDevices()
{
    PrpEgu egu, xegu;
    DTYPE dout, din;
    HistorySpecification hspec;
    NAME64 dev_property_list[NUM_PROPERTIES] = {
        { P_N_INFO }, { P_N_CPU }, { P_N_MEMORY }, { P_N_MEMORY_MB }, { P_N_NUM_CPU_VIOLATIONS }, { P_N_NUM_MEMORY_VIOLATIONS },
        { P_N_NOT_RUNNING_SINCE }, { P_N_LAST_RETRY }, { P_N_CURRENT_RETRY }, { P_N_STATUS }, { P_N_START }, { P_N_RESTART },
        { P_N_START_TINE }, { P_N_STOP_TINE }, { P_N_STOP }, { P_N_LAST_TINE_RESPONSE }, { P_N_LAST_TINE_RETRY }, { P_N_TINE_STATUS },
        { P_N_NUM_RESTARTED }, { P_N_NUM_STOPPED_WD }, {P_N_RESET_STATISTICS}, { P_N_LAST_RESTART_REASON } };
    NAME64 wd_property_list[NUM_WD_PROPERTIES] = {
        { P_N_SUSPEND }, { P_N_CPU_TOTAL }, { P_N_AVAILABLE_MEMORY }, { P_N_SHUTDOWN }, { P_N_REBOOT }, { P_N_REMAINING_SUSPENDED },
        { P_N_RELOAD }, { P_N_CPU }, { P_N_MEMORY }, { P_N_MEMORY_MB }, { P_N_NETWORK_NAMES }, { P_N_NETWORK_UTILIZATION }, { P_N_NETWORK_RELATIVE },
        { P_N_NETWORK_FRIENDLY_NAMES } };


    dout.dArrayLength = 1;
    dout.dFormat = CF_SHORT;
    dout.dTag[0] = 0;
    din.dArrayLength = 1;
    din.dFormat = CF_SHORT;
    din.dTag[0] = 0;
    egu.graph = GT_LIN;
    egu.max = 100;
    egu.min = 0;
    strcpy_s(egu.units, UNITS_SIZE, "%");
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_CPU_TOTAL, &dout, NULL, CA_READ, AT_SINGLE, 0, "Total CPU Load (%)", &egu, NULL, PRP_CPU_TOTAL, NULL);
    dout.dArrayLength = num_processes + 1;
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_CPU, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "CPU usage", &egu, NULL, PRP_CPU, NULL);
    
    dout.dArrayLength = num_network_interfaces;
    dout.dFormat = CF_STRING;
    RegisterPropertyInformation(LOCAL_NAME, P_N_NETWORK_NAMES, &dout, NULL, CA_READ, AT_SINGLE, 0, "Network Interfaces Names",PRP_NETWORK_NAMES, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_NETWORK_FRIENDLY_NAMES, &dout, NULL, CA_READ, AT_SINGLE, 0, "Network Interfaces Friendly Names", PRP_NETWORK_FRIENDLY_NAMES, NULL);
    xegu.graph = GT_LIN;
    xegu.max = (float)(num_network_interfaces - 1);
    xegu.min = 0;
    dout.dFormat = CF_FLOAT;
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_NETWORK_RELATIVE, &dout, NULL, CA_READ, AT_SINGLE, 0, "Total Network Load (%)", &egu, &xegu, PRP_NETWORK_RELATIVE, NULL);
    egu.max = 1E6;
    egu.min = 0;
    strcpy_s(egu.units, UNITS_SIZE, "kB/s");
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_NETWORK_UTILIZATION, &dout, NULL, CA_READ, AT_SINGLE, 0, "Total Network Load (kB/s)", &egu, &xegu, PRP_NETWORK_UTILIZATION, NULL);

    dout.dFormat = CF_NAME64;
    dout.dArrayLength = num_processes;
    RegisterPropertyInformation(LOCAL_NAME, P_N_LAST_RESTART_REASON, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Reason for the last restart of the process", PRP_LAST_RESTART_REASON, NULL);

    dout.dFormat = CF_SHORT;
    dout.dArrayLength = num_processes;
    egu.max = 255;
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_TINE_STATUS, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "TINE property link status", &egu, NULL, PRP_TINE_STATUS, NULL);

    egu.max = (float)suspended_timeout;
    dout.dArrayLength = 1;
    strcpy_s(egu.units, UNITS_SIZE, "seconds");
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_REMAINING_SUSPENDED, &dout, NULL, CA_READ, AT_SINGLE, 0, "Time in seconds after which the watchdog will leave the suspended state", &egu, NULL, PRP_REMAINING_SUSPENDED, NULL);

    strcpy_s(egu.units, UNITS_SIZE, "counts");
    egu.max = (float)INT_MAX;
    egu.min = 0;
    dout.dArrayLength = num_processes;
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_NUM_CPU_VIOLATIONS, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Number of CPU limit violations", &egu, NULL, PRP_NUM_CPU_VIOLATIONS, NULL);
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_NUM_MEMORY_VIOLATIONS, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Number of memory limit violations", &egu, NULL, PRP_NUM_MEMORY_VIOLATIONS, NULL);
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_CURRENT_RETRY, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Number of attempted starts", &egu, NULL, PRP_CURRENT_RETRY, NULL);
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_NUM_RESTARTED, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Number of times WD started the process", &egu, NULL, PRP_NUM_RESTARTED, NULL);
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_NUM_STOPPED_WD, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Number of times WD stopped the process", &egu, NULL, PRP_NUM_STOPPED_WD, NULL);

    dout.dArrayLength = 1;
    RegisterPropertyInformation(LOCAL_NAME, P_N_SUSPEND, &dout, &din, CA_WRITE | CA_READ, AT_SINGLE, 0, "Suspend or resume the watchdog", PRP_SUSPEND, NULL);

    dout.dFormat = CF_NULL;
    dout.dArrayLength = 0;
    RegisterPropertyInformation(LOCAL_NAME, P_N_SHUTDOWN, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Shutdown watchdog (requires confirmation)", PRP_SHUTDOWN, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_REBOOT, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Reboot computer (requires confirmation)", PRP_REBOOT, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_RELOAD, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Reload configuration file", PRP_RELOAD, NULL);
    din.dFormat = CF_NULL;
    din.dArrayLength = 0;
    RegisterPropertyInformation(LOCAL_NAME, P_N_STOP, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Stop the process until manually restarted", PRP_STOP, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_START, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Enable and Start the process", PRP_START, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_RESTART, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Stop the process and let watchdog restart it", PRP_RESTART, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_START_TINE, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Enable and Start polling of TINE property", PRP_START_TINE, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_STOP_TINE, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Disable TINE property polling", PRP_STOP_TINE, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_RESET_STATISTICS, &dout, &din, CA_WRITE, AT_SINGLE, 0, "Reset statistics", PRP_RESET_STATISTICS, NULL);

    dout.dFormat = CF_INT64;
    dout.dArrayLength = 1;
    strcpy_s(egu.units, UNITS_SIZE, "bytes");
    egu.max = FLT_MAX;
    egu.graph = GT_LIN;
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_AVAILABLE_MEMORY, &dout, NULL, CA_READ, AT_SINGLE, 0, "Available Physical Memory", &egu, NULL, PRP_AVAILABLE_MEMORY, NULL);
    dout.dArrayLength = num_processes + 1;
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_MEMORY, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Memory usage", &egu, NULL, PRP_MEMORY, NULL);

    dout.dFormat = CF_DOUBLE;
    strcpy_s(egu.units, UNITS_SIZE, "MB");
    egu.max = 2048;
    RegisterPropertyInformationEx(LOCAL_NAME, P_N_MEMORY_MB, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Memory usage MB", &egu, NULL, PRP_MEMORY_MB, NULL);

    dout.dArrayLength = num_processes;
    RegisterPropertyInformation(LOCAL_NAME, P_N_NOT_RUNNING_SINCE, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Last timestamp when not running", PRP_NOT_RUNNING_SINCE, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_LAST_RETRY, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Timestamp of the last start retry", PRP_LAST_RETRY, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_LAST_TINE_RESPONSE, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Timestamp of the last successful update from TINE property", PRP_LAST_TINE_RESPONSE, NULL);
    RegisterPropertyInformation(LOCAL_NAME, P_N_LAST_TINE_RETRY, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Timestamp of the last retry to attach to TINE property", PRP_LAST_TINE_RETRY, NULL);

    dout.dFormat = CF_BITFIELD8;
    strcpy_s(dout.dTag, TAG_NAME_SIZE, WD_STATUS);
    RegisterPropertyInformation(LOCAL_NAME, P_N_STATUS, &dout, NULL, CA_READ, AT_CHANNEL | AT_SINGLE, 0, "Process status", PRP_STATUS, NULL);

    dout.dFormat = CF_STRUCT;
    strcpy_s(dout.dTag, TAG_NAME_SIZE, WD_PROCESS_INFO);
    din.dArrayLength = dout.dArrayLength;
    din.dFormat = CF_STRUCT;
    strcpy_s(din.dTag, TAG_NAME_SIZE, WD_PROCESS_INFO);
    RegisterPropertyInformation(LOCAL_NAME, P_N_INFO, &dout, &din, CA_READ | CA_WRITE, AT_CHANNEL | AT_SINGLE, 0, "Process configuration", PRP_INFO, NULL);

    //assign server properties to keyword
    RegisterDeviceName(LOCAL_NAME, this_info.server_name, 0);
    AssignPropertyList(LOCAL_NAME, this_info.server_name, WDOG_PROPERTY_LIST_NAME, NUM_WD_PROPERTIES, wd_property_list);
    
    //assign properties to processes devices
    unsigned int i;
    for (i = 0; i < num_processes; i++)
    {
        RegisterDeviceName(LOCAL_NAME, g_process[i].id, i + 1);
        AssignPropertyList(LOCAL_NAME, g_process[i].id, DEVICE_PROPERTY_LIST_NAME, NUM_PROPERTIES, dev_property_list);
        if (_stricmp(g_process[i].id, g_process[i].alias) != 0)
            SetSystemAlias(g_process[i].alias, g_process[i].id);
    }

    hspec.archiveRate = HRS_DEFAULT_ARCHIVE_RATE;
    hspec.aTolerance = HRS_DEFAULT_ABS_TOLERANCE;
    hspec.pTolerance = HRS_DEFAULT_REL_TOLERANCE;
    hspec.heartbeat = HRS_DEFAULT_HEARTBEAT;
    hspec.pollingRate = HRS_DEFAULT_POLL_RATE;
    hspec.rhsPropertyName = NULL;
    hspec.rhsServerName = NULL;
    hspec.depthLong = long_term_history_months;
    hspec.depthShort = short_term_history_seconds;
    
    AppendHistoryInformation(LOCAL_NAME, P_N_CPU_TOTAL, this_info.server_name, 1, CF_SHORT, 0, &hspec);
    AppendHistoryInformation(LOCAL_NAME, P_N_CPU, this_info.server_name, num_processes + 1, CF_SHORT, 1, &hspec);
    AppendHistoryInformation(LOCAL_NAME, P_N_AVAILABLE_MEMORY, this_info.server_name, 1, CF_INT64, 2, &hspec);
    AppendHistoryInformation(LOCAL_NAME, P_N_MEMORY, this_info.server_name, num_processes + 1, CF_INT64, 3, &hspec);
    AppendHistoryInformation(LOCAL_NAME, P_N_NETWORK_UTILIZATION, this_info.server_name, num_network_interfaces, CF_FLOAT, 4, &hspec);
    AppendHistoryInformation(LOCAL_NAME, P_N_MEMORY_MB, this_info.server_name, num_processes + 1, CF_DOUBLE, 5, &hspec);

    AssignNumDevices(LOCAL_NAME, num_processes + 1);
}

void clearStatistics(WATCHED_PROCESS *process)
{
    process->current_retry = 0;
    process->determined_not_running_timestamp = 0;
    process->last_retry_timestamp = 0;
    process->last_tine_retry = 0;
    process->memory_at_stop_time = 0;
    process->number_of_cpu_violations = 0;
    process->number_of_memory_violations = 0;
    process->number_of_times_started_by_wd = 0;
    process->number_of_times_stopped_by_wd = 0;
    process->num_of_cpu_violations_at_stop_time = 0;
    process->num_of_memory_violations_at_stop_time = 0;
    process->stopped_because_of_cpu = FALSE;
    process->stopped_because_of_link = FALSE;
    process->stopped_because_of_memory = FALSE;
    process->stopped_by_user = FALSE;
}

int cRestart(int argc, char **argv)
{
    unsigned int i;
    int ret;
    if (server_reload_requested > 0) return  device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    if (_stricmp("ALL", argv[0]) == 0)
    {
        for (i = 0; i < num_processes; i++)
        {
            g_process[i].disabled = FALSE;
            g_process[i].process_restart_requested = TRUE;
            g_process[i].last_restart_reason = STOPPED_BY_USER;
            ret = StopWatchedProcess(&g_process[i], FALSE, FALSE);
            if (ret > 0)
            {
                LockMutex(FALSE);
                return systemCodeToTineCode(ret);
            }
        }
        LockMutex(FALSE);
        return operation_success;
    }

    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            g_process[i].disabled = FALSE;
            g_process[i].process_restart_requested = TRUE;
            g_process[i].last_restart_reason = STOPPED_BY_USER;
            ret = StopWatchedProcess(&g_process[i], FALSE, FALSE);
            LockMutex(FALSE);
            if (ret == 0) return operation_success;
            else if (ret == -1) return device_busy;
            else return systemCodeToTineCode(ret);
        }
    }

    //device with that name not found, check if it is index
    int a = toNumber(argv[0]) - 1;
    if (errno == NOERROR)
    {
        if (a < (int)num_processes && a > -1)
        {
            g_process[a].disabled = FALSE;
            g_process[a].process_restart_requested = TRUE;
            g_process[a].last_restart_reason = STOPPED_BY_USER;
            ret = StopWatchedProcess(&g_process[a], FALSE, FALSE);
            LockMutex(FALSE);
            if (ret == 0) return operation_success;
            else if (ret == -1) return device_busy;
            else return systemCodeToTineCode(ret);
        }
    }

    LockMutex(FALSE);
    return illegal_device;
}

int cStop(int argc, char **argv)
{
    unsigned int i;
    int ret;
    if (server_reload_requested > 0) return  device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    if (_stricmp("ALL", argv[0]) == 0)
    {
        for (i = 0; i < num_processes; i++)
        {
            g_process[i].disabled = TRUE;
            g_process[i].stopped_by_user = TRUE;
            g_process[i].last_restart_reason = STOPPED_BY_USER;
            ret = StopWatchedProcess(&g_process[i], TRUE, FALSE);
            if (ret > 0)
            {
                LockMutex(FALSE);
                return systemCodeToTineCode(ret);
            }
        }
        LockMutex(FALSE);
        return operation_success;
    }
    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            g_process[i].disabled = TRUE;
            g_process[i].stopped_by_user = TRUE;
            g_process[i].last_restart_reason = STOPPED_BY_USER;
            ret = StopWatchedProcess(&g_process[i], TRUE, FALSE);
            LockMutex(FALSE);
            if (ret == 0) return operation_success;
            else if (ret == -1) return device_busy;
            else return systemCodeToTineCode(ret);
        }
    }

    //device with that name not found, check if it is index
    int a = toNumber(argv[0]) - 1;
    if (errno == NOERROR)
    {
        if (a < (int)num_processes && a > -1)
        {
            g_process[a].disabled = TRUE;
            g_process[a].stopped_by_user = TRUE;
            g_process[a].last_restart_reason = STOPPED_BY_USER;
            ret = StopWatchedProcess(&g_process[a], TRUE, FALSE);
            LockMutex(FALSE);
            if (ret == 0) return operation_success;
            else if (ret == -1) return device_busy;
            else return systemCodeToTineCode(ret);
        }
    }

    LockMutex(FALSE);
    return illegal_device;
}

int cStart(int argc, char **argv)
{
    unsigned int i;
    int ret;
    if (server_reload_requested > 0) return  device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    if (_stricmp("ALL", argv[0]) == 0) 
    {
        for (i = 0; i < num_processes; i++)
        {
            g_process[i].current_retry = 0;
            ret = StartWatchedProcess(&g_process[i]);
            if (ret != operation_success)
            {
                LockMutex(FALSE);
                return systemCodeToTineCode(ret);
            }
        }
        LockMutex(FALSE);
        return operation_success;
    }
    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            g_process[i].current_retry = 0;
            ret = StartWatchedProcess(&g_process[i]);
            LockMutex(FALSE);
            if (ret == operation_success) return operation_success;
            else return systemCodeToTineCode(ret);
        }
    }

    int a = toNumber(argv[0]) - 1;
    if (errno == NOERROR)
    {
        if (a < (int)num_processes && a > -1)
        {
            g_process[a].current_retry = 0;
            ret = StartWatchedProcess(&g_process[a]);
            LockMutex(FALSE);
            if (ret == operation_success) return operation_success;
            else return systemCodeToTineCode(ret);
        }
    }
    LockMutex(FALSE);
    return illegal_device;
}

int cLinkStop(int argc, char **argv)
{
    unsigned int i;
    int ret;
    if (server_reload_requested > 0) return device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    if (_stricmp("ALL", argv[0]) == 0)
    {
        for (i = 0; i < num_processes; i++)
        {
            g_process[i].property_disabled = TRUE;
            ret = CloseServerMonitor(&g_process[i]);
            if (ret != operation_success)
            {
                LockMutex(FALSE);
                return ret;
            }
        }
        LockMutex(FALSE);
        return operation_success;
    }
    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            g_process[i].property_disabled = TRUE;
            ret = CloseServerMonitor(&g_process[i]);
            LockMutex(FALSE);
            return ret;
        }
    }

    int a = toNumber(argv[0]) - 1;
    if (errno == NOERROR)
    {
        if (a < (int)num_processes && a > -1)
        {
            g_process[a].property_disabled = TRUE;
            ret = CloseServerMonitor(&g_process[a]);
            LockMutex(FALSE);
            return ret;
        }
    }
    LockMutex(FALSE);
    return illegal_device;
}

int cLinkRestart(int argc, char **argv)
{
    unsigned int i;
    int ret;
    if (server_reload_requested > 0) return  device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    if (_stricmp("ALL", argv[0]) == 0)
    {
        for (i = 0; i < num_processes; i++)
        {
            g_process[i].last_tine_retry = time(0);
            ret = CloseServerMonitor(&g_process[i]);
            if (ret != operation_success)
            {
                LockMutex(FALSE);
                return ret;
            }
        }
        LockMutex(FALSE);
        return operation_success;
    }
    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            g_process[i].last_tine_retry = time(0);
            ret = CloseServerMonitor(&g_process[i]);
            LockMutex(FALSE);
            return ret;
        }
    }
    int a = toNumber(argv[0]) - 1;
    if (errno == NOERROR)
    {
        if (a < (int)num_processes && a > -1)
        {
            g_process[a].last_tine_retry = time(0);
            ret = CloseServerMonitor(&g_process[a]);
            LockMutex(FALSE);
            return ret;
        }
    }
    LockMutex(FALSE);
    return illegal_device;
}

int cLinkStart(int argc, char **argv)
{
    unsigned int i;
    if (server_reload_requested > 0) return  device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    if (_stricmp("ALL", argv[0]) == 0)
    {
        for (i = 0; i < num_processes; i++)
        {
            g_process[i].last_tine_retry = time(0);
            g_process[i].property_disabled = FALSE;
        }
        LockMutex(FALSE);
        return operation_success;
    }
    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            g_process[i].last_tine_retry = time(0);
            g_process[i].property_disabled = FALSE;
            LockMutex(FALSE);
            return operation_success;
        }
    }
    int a = toNumber(argv[0]) - 1;
    if (errno == NOERROR)
    {
        if (a < (int)num_processes && a > -1)
        {
            g_process[a].last_tine_retry = time(0);
            g_process[a].property_disabled = FALSE;
            LockMutex(FALSE);
            return operation_success;
        }
    }
    LockMutex(FALSE);
    return illegal_device;
}

int cSuspend(int argc, char **argv)
{
    if (server_reload_requested > 0) return  device_busy;
    LockMutex(TRUE);
    suspended = TRUE;
    made_suspended = time(0);
    LockMutex(FALSE);
    return operation_success;
}

int cResume(int argc, char **argv)
{
    if (server_reload_requested > 0) return  device_busy;
    LockMutex(TRUE);
    suspended = FALSE;
    remaining_suspended_time = 0;
    LockMutex(FALSE);
    return operation_success;
}

int cReload(int argc, char **argv)
{
    if (server_reload_requested > 0) return  device_busy;
    LockMutex(TRUE);
    server_reload_requested = time(0);
    LockMutex(FALSE);
    return operation_success;
}

int cListProcesses(int argc, char **argv)
{
    if (server_reload_requested > 0) return  device_busy;
    LockMutex(TRUE);
    ttyoutput("\n#: ID (Alias) | AutoStart | Monitored | Running | Started | CPU [%%] | Memory [MB] | Current Restarts | # Stopped | # Started | Description \n");
    char time_buffer[26];
    unsigned int i, j;
    for (i = 0; i < num_processes; i++)
    {
    	static int sz = 26 * sizeof(char);
        ctime_s(time_buffer, sz, &g_process[i].last_retry_timestamp);
        for (j = 0; j < 26; j++)
        {
            if (time_buffer[j] == '\n') {
                time_buffer[j] = '\0';
                break;
            }
        }
        ttyoutput("%d: %s | %s | %s | %s | %s | %d | %3.2f | %d | %d | %d | %s\n", i+1, 
            g_process[i].alias, g_process[i].auto_start ? (g_process[i].auto_start == YES ? "YES" : "ONCE") : "NO", 
            g_process[i].disabled == 0 && g_process[i].auto_start != 2 ? "YES" : "NO",
            g_process[i].is_running ? "YES" : (g_process[i].auto_start == ONCE ? "--" : "NO"),
            g_process[i].is_running ? (g_process[i].last_retry_timestamp == 0 ? "Long long time ago" : time_buffer) : 
                (g_process[i].auto_start == ONCE ? "I don't care" : "Not Running"),
            g_process[i].current_cpu_load, (double)(g_process[i].current_memory_consumption/(1024.*1024.)),
            g_process[i].current_retry, g_process[i].number_of_times_stopped_by_wd,
            g_process[i].number_of_times_started_by_wd, 
            g_process[i].description == NULL ? "N/A" : g_process[i].description);
    }
    ttyoutput("\n");
    LockMutex(FALSE);
    return operation_success;
}

int cPrintProcessInfo(int argc, char ** argv)
{
    unsigned int i;
    if (server_reload_requested > 0) return  device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            PrintProcessInfo(&g_process[i], 1);
            LockMutex(FALSE);
            return operation_success;
        }
    }
    LockMutex(FALSE);
    return illegal_device;
}

int cShutDown(int argc, char ** argv)
{
    running = FALSE;
    return operation_success;
}

int cResetStatistics(int argc, char ** argv)
{
    unsigned int i;
    if (server_reload_requested > 0) return  device_busy;
    if (argc != 1) return invalid_parameter;
    LockMutex(TRUE);
    if (_stricmp("ALL", argv[0]) == 0)
    {
        for (i = 0; i < num_processes; i++)
        {
            clearStatistics(&g_process[i]);
        }
        LockMutex(FALSE);
        return operation_success;
    }
    for (i = 0; i < num_processes; i++)
    {
        if (_stricmp(g_process[i].alias, argv[0]) == 0 || _stricmp(g_process[i].id, argv[0]) == 0)
        {
            clearStatistics(&g_process[i]);
            LockMutex(FALSE);
            return operation_success;
        }
    }
    LockMutex(FALSE);
    return illegal_device;
}

void registerUserCommands()
{
    RegisterUserCommand(C_WATCHDOG_DEBUG, NULL, &debug, NULL, CA_READ | CA_WRITE);
    RegisterUserFunction(C_LINK_START, cLinkStart);
    RegisterUserFunction(C_LINK_STOP, cLinkStop);
    RegisterUserFunction(C_LINK_RESTART, cLinkRestart);
    RegisterUserFunction(C_START, cStart);
    RegisterUserFunction(C_STOP, cStop);
    RegisterUserFunction(C_RESTART, cRestart);
    RegisterUserFunction(C_RESUME, cResume);
    RegisterUserFunction(C_SUSPEND, cSuspend);
    RegisterUserFunction(C_RELOAD, cReload);
    RegisterUserFunction(C_PROC_LIST, cListProcesses);
    RegisterUserFunction(C_PROC_INFO, cPrintProcessInfo);
    RegisterUserFunction(C_SHUTDOWN, cShutDown);
    RegisterUserFunction(C_RESET_STATISTICS, cResetStatistics);
}

void registerAlarms()
{
    ADS ads;
    ads.alarmSystem = 0;
    ads.alarmDataArraySize = 4;
    ads.alarmOscillationWindow = ALM_OSCILLATION_WINDOW;
    strcpy_s(ads.deviceText, 64, "");
    strcpy_s(ads.url, 64, "");

    strcpy_s(ads.alarmTag, 32, "CPU too high");
    strcpy_s(ads.alarmText, 64, "CPU usage of the process is too high");
    strcpy_s(ads.dataText, 64, "Current CPU, Max CPU, Current Violation, Allowed Violations");
    ads.alarmDataFormat = (BYTE)CF_INT32;
    ads.alarmSeverity = 3;
    ads.alarmCode = A_WARN_CPU_TOO_HIGH;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);
    
    strcpy_s(ads.alarmTag, 32, "CPU too high: limit exceeded");
    strcpy_s(ads.alarmText, 64, "CPU too high: limit exceeded");
    ads.alarmCode = A_ERR_CPU_TOO_HIGH;
    ads.alarmSeverity = 9;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);

    strcpy_s(ads.alarmTag, 32, "Memory too high");
    strcpy_s(ads.alarmText, 64, "Memory consumption of the process is too high");
    strcpy_s(ads.dataText, 64, "Current Mem, Max Mem, Current Violation, Allowed Violations");
    ads.alarmDataFormat = (BYTE)CF_FLOAT;
    ads.alarmSeverity = 3;
    ads.alarmCode = A_WARN_MEMORY_TOO_HIGH;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);

    strcpy_s(ads.alarmTag, 32, "Memory too high: limit exceeded");
    strcpy_s(ads.alarmText, 64, "Memory too high: limit exceeded");
    ads.alarmCode = A_ERR_MEMORY_TOO_HIGH;
    ads.alarmSeverity = 9;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);

    strcpy_s(ads.alarmTag, 32, "Property link timeout");
    strcpy_s(ads.alarmText, 64, "TINE property link has not received updates for a while");
    strcpy_s(ads.dataText, 64, "");
    ads.alarmSeverity = 3;
    ads.alarmDataArraySize = 0;
    ads.alarmDataFormat = (BYTE)CF_NULL;
    ads.alarmCode = A_WARN_LINK_TIMEOUT;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);

    ads.alarmCode = A_ERR_LINK_TIMEOUT;
    ads.alarmSeverity = 9;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);

    strcpy_s(ads.alarmTag, 32, "Watched process not running");
    strcpy_s(ads.alarmText, 64, "Process monitored by watchdog is not running");
    strcpy_s(ads.dataText, 64, "");
    ads.alarmSeverity = 9;
    ads.alarmDataArraySize = 0;
    ads.alarmDataFormat = (BYTE)CF_NULL;
    ads.alarmCode = A_ERR_NOT_RUNNING;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);

    strcpy_s(ads.alarmTag, 32, "Watched process disabled");
    strcpy_s(ads.alarmText, 64, "Monitoring disabled due to configured retry threshold exceeded");
    strcpy_s(ads.dataText, 64, "");
    ads.alarmSeverity = 9;
    ads.alarmDataArraySize = 0;
    ads.alarmDataFormat = (BYTE)CF_NULL;
    ads.alarmCode = A_ERR_DISABLED;
    AppendAlarmInfoTable(LOCAL_NAME, &ads);

    if (minimum_free_disk_path)
    {
        // Only set the alarm if the path is given. If it is null ignore.
        SetFreeBlocksAlarmThreshold(minimum_free_disk_path, (UINT32)(minimum_free_disk_mbyte*1024));
    }
}

void wdinit(void)
{
    SetSizeDeviceCapacity(LOCAL_NAME,MAX_NUMBER_OF_DEVICES+1);
    createStructures();
    registerStructsAndBitfields();
    registerPropertiesAndDevices();
    registerUserCommands();
    registerAlarms();
}

int copyInfo(WDProcessInfo *dest, WDProcessInfo source, WATCHED_PROCESS *process)
{
    boolean property_changed;
    unsigned int k;
    char **temp;
    
    if (strlen(source.alias) != 0)
    {
        if (_stricmp(dest->alias, source.alias) != 0)
        {
            return invalid_parameter;
        }
    }
    if (strlen(source.name) != 0)
    {
        if (_stricmp(dest->name, source.name) != 0)
        {
            return invalid_parameter;
        }
    }
    if (strlen(source.path) != 0)
    {
        if (_stricmp(dest->path, source.path) != 0)
        {
            return invalid_parameter;
        }
    }
    if (strlen(source.parameters) != 0)
    {
        if (_stricmp(dest->parameters, source.parameters) != 0)
        {
            return invalid_parameter;
        }
    }
    if (strlen(source.environment) != 0)
    {
        if (_stricmp(dest->environment, source.environment) != 0)
        {
            return invalid_parameter;
        }
    }
    if (strlen(source.working_directory) != 0)
    {
        if (_stricmp(dest->working_directory, source.working_directory) != 0)
        {
            return invalid_parameter;
        }
    }
    if (strlen(source.preceding_processes) != 0)
    {
        if (_stricmp(dest->preceding_processes, source.preceding_processes) != 0)
        {
            return invalid_parameter;
        }
    }
    if (strlen(source.caption) != 0)
    {
        if (_stricmp(dest->caption, source.caption) != 0)
        {
            return invalid_parameter;
        }
    }
    if ((dest->is_service == 0 && source.is_service != 0) ||
        (dest->is_service != 0 && source.is_service == 0))
    {
        return invalid_parameter;
    }
    if ((dest->match_parameters == 0 && source.match_parameters != 0) ||
        (dest->match_parameters != 0 && source.match_parameters == 0))
    {
        return invalid_parameter;
    }
    if ((dest->match_path == 0 && source.match_path != 0) ||
        (dest->match_path != 0 && source.match_path == 0))
    {
        return invalid_parameter;
    }
     
    dest->auto_start = source.auto_start;
    dest->max_cpu_load = source.max_cpu_load;
    dest->max_memory = source.max_memory;
    dest->monitor_n_seconds_after_start = source.monitor_n_seconds_after_start;
    dest->number_of_allowed_cpu_load_violations = source.number_of_allowed_cpu_load_violations;
    dest->number_of_allowed_memory_violations = source.number_of_allowed_memory_violations;
    dest->number_of_retries = source.number_of_retries;
    dest->pause_n_seconds_after_start = source.pause_n_seconds_after_start;
    dest->polling_rate = source.polling_rate;
    dest->retry_interval = source.retry_interval;
    dest->timeout = source.timeout;
    dest->timeout_restart = source.timeout_restart;
    
    strcpy_s(dest->tine_device, 132, source.tine_device);
    strcpy_s(dest->tine_property, 64, source.tine_property);

    process->auto_start = source.auto_start;
    process->match_parameters = source.match_parameters;
    process->match_path = source.match_path;
    process->max_cpu_load = source.max_cpu_load;
    process->max_memory = (UINT64)(source.max_memory * 1024 * 1024);
    process->start_after_n_seconds = source.monitor_n_seconds_after_start;
    process->number_of_allowed_cpu_load_violations = source.number_of_allowed_cpu_load_violations;
    process->number_of_allowed_memory_violations = source.number_of_allowed_memory_violations;
    process->number_of_retries = source.number_of_retries;
    process->pause_after_previous_start = source.pause_n_seconds_after_start;
    process->retry_interval = source.retry_interval;
    process->timeout = source.timeout;
    process->timeout_restart = source.timeout_restart;
    
    property_changed = FALSE;
    if (process->polling_rate != source.polling_rate)
    {
        property_changed = TRUE;
        process->polling_rate = source.polling_rate;
    }
    
    if (process->tine_device != NULL) 
    {
        if (_stricmp(process->tine_device, source.tine_device) != 0)
        {
            property_changed = TRUE;
            free(process->tine_device);
            size_t length = min(SIZE_TINE_DEVICE,strlen(source.tine_device));
            process->tine_device = (char*)malloc(sizeof(char) * (length+1));
            if (process->tine_device == NULL) OutOfMemoryExit(6);
            strncpy_s(process->tine_device, length+1, source.tine_device,SIZE_TINE_DEVICE);
            process->tine_device[length] = '\0';
        }
    }
    else if (strlen(source.tine_device) > 0)
    {
        property_changed = TRUE;
        size_t length = min(SIZE_TINE_DEVICE,strlen(source.tine_device));
        process->tine_device = (char*)malloc(sizeof(char)* (length + 1));
        if (process->tine_device == NULL) OutOfMemoryExit(6);
        strncpy_s(process->tine_device, length + 1, source.tine_device, SIZE_TINE_DEVICE);
        process->tine_device[length] = '\0';
    }

    if (property_changed)
    {
        unsigned int siz;
        temp = split(process->tine_device, &siz, '/', FALSE, FALSE, FALSE);
        k = 0;
        if (siz >= 3)
        {
            if (process->context != NULL) free(process->context);
            if (process->server != NULL) free(process->server);
            //if the names starts with a / we get an empty string first
            if (strlen(temp[0]) == 0)
            {
                process->context = temp[1];
                process->server = temp[2];
                k = 3;
                free(temp[0]);
            }
            else
            {
                process->context = temp[0];
                process->server = temp[1];
                k = 2;
            }
        }
        //free everything that is no longer needed
        for (; k < siz; k++)
        {
            free(temp[k]);
        }
        free(temp);
    }
    
    if (process->tine_property != NULL)
    {
        if (_stricmp(process->tine_property, source.tine_property) != 0)
        {
            property_changed = 1;
            free(process->tine_property);
            size_t length = min(SIZE_TINE_PROPERTY,strlen(source.tine_property));
            process->tine_property = (char*)malloc(sizeof(char)* (length + 1));
            if (process->tine_property == NULL) OutOfMemoryExit(6);
            strncpy_s(process->tine_property, length+1, source.tine_property, SIZE_TINE_PROPERTY);
            process->tine_property[length] = '\0';
        }
    }
    else if (strlen(source.tine_property) > 0)
    {
        property_changed = 1;
        size_t length = min(SIZE_TINE_PROPERTY,strlen(source.tine_property));
        process->tine_property = (char*)malloc(sizeof(char)* (length + 1));
        if (process->tine_property == NULL) OutOfMemoryExit(6);
        strncpy_s(process->tine_property, length + 1, source.tine_property, SIZE_TINE_PROPERTY);
        process->tine_property[length] = '\0';
    }

    if (property_changed)
    {
        process->last_tine_retry = time(0);
        process->property_disabled = FALSE;
        int ret = CloseServerMonitor(process);
        if (ret != 0) return ret;
    }
    
    return operation_success;
}

//equipment module task to read or write data
int wdeqm(char *devName, char *devProperty, DTYPE *dout, DTYPE *din, short access)
{
    int prpid, devid, cc, all;
    unsigned int i;
    short val, *values;
    SINT64 *llval, lval;
    double *dval;
    WDProcessInfo *info;
    time_t now;
    if (!running) return not_running;
    if (server_reload_requested > 0) return  device_busy;
    prpid = GetPropertyId(LOCAL_NAME, devProperty);
    devid = GetDeviceNumber(LOCAL_NAME, devName);
    cc = operation_success;

    now = time(0);
    LockMutex(TRUE);
    if (devid == 0)
    {
        //device independent properties
        switch (prpid)
        {
        case PRP_SUSPEND:
            if (access&CA_WRITE)
            {
                cc = GetValuesAsShort(din, &val, 1);
                if (cc == 0)
                {
                    if (val == FALSE)
                    {
                        suspended = FALSE;
                        remaining_suspended_time = 0;
                    }
                    else if (val == TRUE)
                    {
                        suspended = TRUE;
                        made_suspended = time(0);
                    }
                    else
                    {
                        LockMutex(FALSE);
                        return illegal_command;
                    }
                }
                else
                {
                    LockMutex(FALSE);
                    return cc;
                }
            }
            if (access&CA_READ)
            {
                val = suspended;
                LockMutex(FALSE);
                if ((cc = PutValuesFromShort(dout, &val, 1)) != 0)
                {
                    return cc;
                }
                return operation_success;
            }
            LockMutex(FALSE);
            return illegal_command;
        case PRP_CPU_TOTAL:
            CHECK_ACCESS_READ(access);
            val = total_cpu_usage;
            LockMutex(FALSE);
            if ((cc = PutValuesFromShort(dout, &val, 1)) != 0)
            {
                return cc;
            }
            return operation_success;
        case PRP_NETWORK_NAMES:
            CHECK_ACCESS_READ(access);
            if (dout->dArrayLength > 0)
            {
                dout->data.strptr = network_interface_names;
            }
            LockMutex(FALSE);
            return operation_success;
        case PRP_NETWORK_FRIENDLY_NAMES:
            CHECK_ACCESS_READ(access);
            if (dout->dArrayLength > 0)
            {
                dout->data.strptr = network_interface_friendly_names;
            }
            LockMutex(FALSE);
            return operation_success;
        case PRP_NETWORK_UTILIZATION:
            CHECK_ACCESS_READ(access);
            if (dout->dArrayLength > 0)
            {
                if ((cc = PutValuesFromFloat(dout, network_utilization_absolute, num_network_interfaces)) != 0)
                {
                    LockMutex(FALSE);
                    return cc;
                }
            }
            LockMutex(FALSE);
            return operation_success;
        case PRP_NETWORK_RELATIVE:
            CHECK_ACCESS_READ(access);
            if (dout->dArrayLength > 0)
            {
                if ((cc = PutValuesFromFloat(dout, network_utilization_percentage, num_network_interfaces)) != 0)
                {
                    LockMutex(FALSE);
                    return cc;
                }
            }
            LockMutex(FALSE);
            return operation_success;
        case PRP_AVAILABLE_MEMORY:
            CHECK_ACCESS_READ(access);
            lval = (SINT64)available_memory;
            LockMutex(FALSE);
            cc = PutValuesFromDLong(dout, &lval, 1);
            return cc;
        case PRP_REMAINING_SUSPENDED:
            CHECK_ACCESS_READ(access);
            lval = (SINT64)remaining_suspended_time;
            LockMutex(FALSE);
            cc = PutValuesFromDLong(dout, &lval, 1);
            return cc;
        case PRP_RELOAD:
            CHECK_ACCESS_WRITE(access);
            server_reload_requested = time(0);
            LockMutex(FALSE);
            return operation_success;
        case PRP_SHUTDOWN:
            CHECK_ACCESS_WRITE(access);
            LockMutex(FALSE);
            cc = GetValuesAsShort(din, &val, 1);
            if (cc == 0 && (val == 1 || val == 2))
            {
                if (val == 1)
                {
                    shutdown_request = time(0);
                }
                else if (val == 2)
                {
                    if (now - shutdown_request < CONFIRMATION_TIMEOUT) running = FALSE;
                    else return illegal_command;
                    shutdown_request = 0;
                }
            }
            else if (val < 0 || val > 2) return invalid_parameter;
            return cc;
        case PRP_REBOOT:
            CHECK_ACCESS_WRITE(access);
            LockMutex(FALSE);
            cc = GetValuesAsShort(din, &val, 1);
            if (cc == 0 && (val == 1 || val == 2))
            {
                if (val == 1)
                {
                    reboot_request = time(0);
                }
                else if (val == 2)
                {
                    if (now - reboot_request < CONFIRMATION_TIMEOUT)
                    {
                        if (RebootComputer() > 0)
                        {
                            reboot_request = 0;
                            return unauthorised_action;
                        }

                    }
                    else return illegal_command;
                    reboot_request = 0;
                }
            }
            else if (val < 0 || val > 2) return invalid_parameter;
            return cc;
        case PRP_CPU:
            //this is for history storage purposes, which is always called with the device 0
            CHECK_ACCESS_READ(access);
            if (dout->dArrayLength > 0)
            {
                values = (short*)alloca(sizeof(short)*(num_processes+1));
                if (values == NULL) OutOfMemoryExit(6);
                values[0] = watchdog_cpu_usage;
                for (i = 0; i < num_processes; i++)
                {
                    values[i+1] = g_process[i].current_cpu_load;
                }                
                cc = PutValuesFromShortEx(dout, values, num_processes + 1, 0);
                LockMutex(FALSE);
                return cc;
            }
            else
            {
                LockMutex(FALSE);
                return illegal_data_size;
            }
        case PRP_MEMORY:
            //this is for history storage purposes, which is always called with the device 0
            CHECK_ACCESS_READ(access);
            if (dout->dArrayLength > 0)
            {
                llval = (SINT64*)alloca(sizeof(SINT64)*(num_processes+1));
                if (llval == NULL) OutOfMemoryExit(6);
                llval[0] = watchdog_memory_consumption;
                for (i = 0; i < num_processes; i++)
                {
                    llval[i+1] = (SINT64)g_process[i].current_memory_consumption;
                }
                cc = PutValuesFromDLongEx(dout, llval, num_processes + 1, 0);
                LockMutex(FALSE);
                return cc;
            }
            else
            {
                LockMutex(FALSE);
                return illegal_data_size;
            }
        case PRP_MEMORY_MB:
            //this is for history storage purposes, which is always called with the device 0
            CHECK_ACCESS_READ(access);
            if (dout->dArrayLength > 0)
            {
                dval = (double*)alloca(sizeof(double)*(num_processes + 1));
                if (dval == NULL) OutOfMemoryExit(6);
                dval[0] = ((UINT64)(watchdog_memory_consumption * 1000 / 1024. / 1024.)) / 1000.;
                for (i = 0; i < num_processes; i++)
                {
                    dval[i + 1] = ((UINT64)(g_process[i].current_memory_consumption * 1000 / 1024. / 1024.)) / 1000.;
                }
                cc = PutValuesFromDoubleEx(dout, dval, num_processes + 1, 0);
                LockMutex(FALSE);
                return cc;
            }
            else
            {
                LockMutex(FALSE);
                return illegal_data_size;
            }
        }
        LockMutex(FALSE);
        return illegal_equipment_number;
    }

    //check if the device name is all. In case that one of the commands was issued on device ALL, apply that command to all processes
    all = 0;
    if (devid < 0)
    {
        if (_stricmp("ALL", devName) == 0) 
        {
            all = 1;
            devid = 1;
        }
        else
        {
            LockMutex(FALSE);
            return illegal_equipment_name;
        }
    }
    else
    {
        if (devid < 1 || devid >(int)num_processes)
        {
            LockMutex(FALSE);
            return illegal_equipment_number;
        }
    }

    //reduce the device id for 1, becuase device 0 is "keyword"
    devid--;
    //device depended properties
    switch (prpid)
    {
    case PRP_INFO:
        if (!(access&(CA_READ | CA_WRITE)))
        {
            LockMutex(FALSE);
            return illegal_read_write;
        }
        if (access&CA_WRITE)
        {
            if (din->dFormat != CF_STRUCT) cc = illegal_format;
            else if (_stricmp(din->dTag, WD_PROCESS_INFO)) cc = invalid_structure_tag;
            else if ((int)din->dArrayLength > num_processes) cc = invalid_structure_size;
            if (cc != operation_success)
            {
                LockMutex(FALSE);
                return cc;
            }
            info = (WDProcessInfo *)(din->data.vptr);
            for (i = 0; i<(int)din->dArrayLength; i++)
            {
                cc = copyInfo(&wd_process_info[(i + devid) % num_processes], info[i],&g_process[(i+devid) % num_processes]);
                if (cc != 0)
                {
                    LockMutex(FALSE);
                    break;
                }
            }
            if (cc != 0)
            {
                return cc;
            }
            cc = RewriteConfiguration();
            LockMutex(FALSE);
            return cc;
        }
        if (access&CA_READ)
        {
            if (dout->dFormat != CF_STRUCT) cc = illegal_format;
            else if (_stricmp(dout->dTag, WD_PROCESS_INFO)) cc = invalid_structure_tag;
            if (cc != operation_success)
            {
                LockMutex(FALSE);
                return cc;
            }
            if ((int)dout->dArrayLength > num_processes) dout->dArrayLength = num_processes;
            info = (WDProcessInfo *)(dout->data.vptr);
            for (i = 0; i<(int)dout->dArrayLength; i++)
            {
                info[i] = wd_process_info[(i + devid) % num_processes];
            }
        }
        LockMutex(FALSE);
        return operation_success;
    case PRP_CPU:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            //for history purposes we include additional element
            values = (short*)alloca(sizeof(short)*(num_processes+1));
            if (values == NULL) OutOfMemoryExit(6);
            values[0] = watchdog_cpu_usage;
            for (i = 0; i < num_processes; i++)
            {
                values[i+1] = g_process[i].current_cpu_load;
            }
            cc = PutValuesFromShortEx(dout, values, num_processes + 1, devid + 1);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_NUM_CPU_VIOLATIONS:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            values = (short*)alloca(sizeof(short)*num_processes);
            if (values == NULL) OutOfMemoryExit(6);
            values[0] = total_cpu_usage;
            for (i = 0; i < num_processes; i++)
            {
                values[i] = g_process[i].number_of_cpu_violations;
            }
            cc = PutValuesFromShortEx(dout, values, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_MEMORY:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            llval = (SINT64*)alloca(sizeof(SINT64)*(num_processes+1));
            if (llval == NULL) OutOfMemoryExit(6);
            llval[0] = watchdog_memory_consumption;
            for (i = 0; i < num_processes; i++)
            {
                llval[i+1] = (SINT64)g_process[i].current_memory_consumption;
            }
            cc = PutValuesFromDLongEx(dout, llval, num_processes + 1, devid + 1);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_MEMORY_MB:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            dval = (double*)alloca(sizeof(double)*(num_processes + 1));
            if (dval == NULL) OutOfMemoryExit(6);
            dval[0] = ((UINT64)(watchdog_memory_consumption * 1000 / 1024. / 1024.)) / 1000.;
            for (i = 0; i < num_processes; i++)
            {
                dval[i + 1] = ((UINT64)(g_process[i].current_memory_consumption * 1000 / 1024. / 1024.)) / 1000.;
            }
            cc = PutValuesFromDoubleEx(dout, dval, num_processes + 1, devid + 1);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_NUM_MEMORY_VIOLATIONS:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            values = (short*)alloca(sizeof(short)*num_processes);
            if (values == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                values[i] = g_process[i].number_of_memory_violations;
            }
            cc = PutValuesFromShortEx(dout, values, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_NOT_RUNNING_SINCE:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            dval = (double*)alloca(sizeof(double)*num_processes);
            if (dval == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                dval[i] = difftime(g_process[i].determined_not_running_timestamp, 0);
            }
            cc = PutValuesFromDoubleEx(dout, dval, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_LAST_RETRY:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            dval = (double*)alloca(sizeof(double)*num_processes);
            if (dval == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                dval[i] = difftime(g_process[i].last_retry_timestamp, 0);
            }
            cc = PutValuesFromDoubleEx(dout, dval, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_CURRENT_RETRY:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            values = (short*)alloca(sizeof(short)*num_processes);
            if (values == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                values[i] = g_process[i].current_retry;
            }
            cc = PutValuesFromShortEx(dout, values, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_LAST_TINE_RESPONSE:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            dval = (double*)alloca(sizeof(double)*num_processes);
            if (dval == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                dval[i] = difftime(g_process[i].last_tine_response, 0);
            }
            cc = PutValuesFromDoubleEx(dout, dval, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_LAST_TINE_RETRY:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            dval = (double*)alloca(sizeof(double)*num_processes);
            if (dval == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                dval[i] = difftime(g_process[i].last_tine_retry, 0);
            }
            cc = PutValuesFromDoubleEx(dout, dval, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_TINE_STATUS:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            values = (short*)alloca(sizeof(short)*num_processes);
            if (values == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                values[i] = g_process[i].tine_status;
            }
            cc = PutValuesFromShortEx(dout, values, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_NUM_RESTARTED:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            values = (short*)alloca(sizeof(short)*num_processes);
            if (values == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                values[i] = g_process[i].number_of_times_started_by_wd;
            }
            cc = PutValuesFromShortEx(dout, values, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_NUM_STOPPED_WD:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            values = (short*)alloca(sizeof(short)*num_processes);
            if (values == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                values[i] = g_process[i].number_of_times_stopped_by_wd;
            }
            cc = PutValuesFromShortEx(dout, values, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_STATUS:
        CHECK_ACCESS_READ(access);
        if (dout->dArrayLength > 0)
        {
            values = (short*)alloca(sizeof(short)*num_processes);
            if (values == NULL) OutOfMemoryExit(6);
            for (i = 0; i < num_processes; i++)
            {
                values[i] = (
                    ((g_process[i].is_running) ? RUNNING : 0)
                    | ((g_process[i].current_retry > 0) ? TRYING_START : 0)
                    | ((g_process[i].tine_status == 0) ? PROPERTY_CONNECTED : 0)
                    | ((g_process[i].property_link_timeout > 0) ? PROPERTY_TIMEOUT : 0)
                    | ((g_process[i].property_disabled > 0) ? PROPERTY_DISABLED : 0)
                    | ((g_process[i].number_of_cpu_violations > 0) ? CPU_OVERLOAD : 0)
                    | ((g_process[i].number_of_memory_violations > 0) ? MEMORY_OVERLOAD : 0)
                    | (g_process[i].auto_start == ONCE ? DISABLED : (g_process[i].disabled ? DISABLED : 0)));
            }
            cc = PutValuesFromShortEx(dout, values, num_processes, devid);
        }
        LockMutex(FALSE);
        return cc;
    case PRP_START:
        CHECK_ACCESS_WRITE(access);
        //do not require any parameter
        if (all == 0)
        {
            g_process[devid].current_retry = 0;
            cc = StartWatchedProcess(&g_process[devid]);
            LockMutex(FALSE);
            if (cc == 0) return operation_success;
            else return systemCodeToTineCode(cc);
        }
        else
        {
            for (i = 0; i < num_processes; i++)
            {
                g_process[i].current_retry = 0;
                cc = StartWatchedProcess(&g_process[i]);
                if (cc != 0)
                {
                    LockMutex(FALSE);
                    return systemCodeToTineCode(cc);
                }
            }
            LockMutex(FALSE);
            return operation_success;
        }        
    case PRP_RESTART:
        CHECK_ACCESS_WRITE(access);
        if (all == 0)
        {
            g_process[devid].current_retry = 0;
            g_process[devid].process_restart_requested = TRUE;
            g_process[devid].disabled = FALSE;
            g_process[devid].last_restart_reason = STOPPED_BY_USER;
            cc = StopWatchedProcess(&g_process[devid], FALSE, FALSE);
            LockMutex(FALSE);
            if (cc == 0) return operation_success;
            else if (cc == -1) return device_busy;
            else return systemCodeToTineCode(cc);
        }
        else
        {
            for (i = 0; i < num_processes; i++)
            {
                g_process[i].current_retry = 0;
                g_process[i].process_restart_requested = TRUE;
                g_process[i].disabled = FALSE;
                g_process[i].last_restart_reason = STOPPED_BY_USER;
                cc = StopWatchedProcess(&g_process[i], FALSE, FALSE);
                if (cc > 0)
                {
                    LockMutex(FALSE);
                    return systemCodeToTineCode(cc);
                }
            }
            LockMutex(FALSE);
            return operation_success;
        }
        
    case PRP_START_TINE:
        CHECK_ACCESS_WRITE(access);
        //reset the state, it will be picked up by the watchdog
        if (all == 0)
        {
            g_process[devid].last_tine_retry = time(0);
            g_process[devid].property_disabled = FALSE;
        }
        else
        {
            for (i = 0; i < num_processes; i++)
            {
                g_process[i].last_tine_retry = time(0);
                g_process[i].property_disabled = FALSE;
            }
        }
        LockMutex(FALSE);
        return operation_success;
    case PRP_STOP_TINE:
        CHECK_ACCESS_WRITE(access);
        if (all == 0)
        {
            g_process[devid].property_disabled = TRUE;
            cc = CloseServerMonitor(&g_process[devid]);
            LockMutex(FALSE);
            return cc;
        }
        else
        {
            for (i = 0; i < num_processes; i++)
            {
                g_process[i].property_disabled = TRUE;
                cc = CloseServerMonitor(&g_process[i]);
                if (cc != 0)
                {
                    LockMutex(FALSE);
                    return cc;
                }
            }
            LockMutex(FALSE);
            return operation_success;
        }
    case PRP_STOP:
        CHECK_ACCESS_WRITE(access);
        if (all == 0)
        {
            g_process[devid].disabled = TRUE;
            g_process[devid].stopped_by_user = TRUE;
            g_process[devid].current_retry = 0;
            g_process[devid].last_restart_reason = STOPPED_BY_USER;
            cc = StopWatchedProcess(&g_process[devid], TRUE, FALSE);
            LockMutex(FALSE);
            if (cc == 0) return operation_success;
            else if (cc == -1) return device_busy;
            else return systemCodeToTineCode(cc);
        }
        else
        {
            for (i = 0; i < num_processes; i++)
            {
                g_process[i].disabled = TRUE;
                g_process[i].stopped_by_user = TRUE;
                g_process[i].current_retry = 0;
                g_process[i].last_restart_reason = STOPPED_BY_USER;
                cc = StopWatchedProcess(&g_process[i], TRUE, FALSE);
                if (cc > 0)
                {
                    LockMutex(FALSE);
                    return systemCodeToTineCode(cc);
                }
            }
            LockMutex(FALSE);
            return operation_success;
        }
    case PRP_RESET_STATISTICS:
        CHECK_ACCESS_WRITE(access);
        if (all == 0)
        {
            clearStatistics(&g_process[devid]);
        }
        else
        {
            for (i = 0; i < num_processes; i++)
            {
                clearStatistics(&g_process[i]);
            }
        }
        LockMutex(FALSE);
        return operation_success;
    case PRP_LAST_RESTART_REASON:
        CHECK_ACCESS_READ(access);
        if ((int)dout->dArrayLength > num_processes) dout->dArrayLength = num_processes;
        NAME64* reason = (NAME64 *)(dout->data.vptr);
        for (i = 0; i<(int)dout->dArrayLength; i++)
        {
        	static unsigned int nameSize = sizeof(NAME64);
            switch (g_process[(i + devid) % num_processes].last_restart_reason)
            {
            case RESTARTED_BY_USER: strcpy_s(reason[i].name, nameSize, RESTARTED_BY_USER_MESSAGE);
                break;
            case STOPPED_BY_USER: strcpy_s(reason[i].name, nameSize, STOPPED_BY_USER_MESSAGE);
                break;
            case CPU_TOO_HIGH: strcpy_s(reason[i].name, nameSize, CPU_TOO_HIGH_MESSAGE);
                break;
            case MEMORY_TOO_HIGH: strcpy_s(reason[i].name, nameSize, MEMORY_TOO_HIGH_MESSAGE);
                break;
            case LINK_TIMEOUT: strcpy_s(reason[i].name, nameSize, LINK_TIMEOUT_MESSAGE);
                break;
            case UNKNOWN: strcpy_s(reason[i].name, nameSize, UNKNOWN_MESSAGE);
                break;
            case NONE: strcpy_s(reason[i].name, nameSize, NONE_MESSAGE);
                break;
            }
        }
        LockMutex(FALSE);
        return operation_success;
    }
    LockMutex(FALSE);
    return illegal_property;
}

void intToChar(int val, BYTE *dest)
{
    dest[3] = (val >> 24) & 0xFF;
    dest[2] = (val >> 16) & 0xFF;
    dest[1] = (val >> 8) & 0xFF;
    dest[0] = val & 0xFF;
}

void floatToChar(float val, BYTE *dest)
{
    unsigned int asInt = *((int*)&val);
    dest[3] = (asInt >> 24) & 0xFF;
    dest[2] = (asInt >> 16) & 0xFF;
    dest[1] = (asInt >> 8) & 0xFF;
    dest[0] = asInt & 0xFF;
}

void wdbgk(void)
{
    unsigned int i;
    int flag;
    BYTE almData[16];
    if (!running) return;
    if (server_reload_requested > 0) return;
    for (i = 0; i < num_processes; i++)
    {
        ClearDeviceAlarm(LOCAL_NAME, g_process[i].id);
    }
    Watch();
    for (i = 0; i < num_processes; i++)
    {
        if (g_process[i].auto_start == ONCE) continue;
        if (g_process[i].auto_start == YES) flag = almNEWALARM;
        else flag = 0;
        if (!g_process[i].disabled && !g_process[i].stopped_by_user)
        {
            if (g_process[i].number_of_cpu_violations > 0 || g_process[i].stopped_because_of_cpu)
            {
                intToChar(g_process[i].max_cpu_load, &almData[4]);
                intToChar(g_process[i].number_of_allowed_cpu_load_violations, &almData[12]);
                if (g_process[i].stopped_because_of_cpu)
                {
                    intToChar(g_process[i].cpu_at_stop_time, almData);
                    intToChar(g_process[i].num_of_cpu_violations_at_stop_time, &almData[8]);
                    SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_ERR_CPU_TOO_HIGH, almData, almTRANSIENT);
                    g_process[i].last_restart_reason = CPU_TOO_HIGH;
                }
                if (g_process[i].number_of_cpu_violations > 0 && !g_process[i].is_starting)
                {
                    intToChar(g_process[i].current_cpu_load, almData);
                    intToChar(g_process[i].number_of_cpu_violations, &almData[8]);
                    SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_WARN_CPU_TOO_HIGH, almData, flag);
                }
            }
            if (g_process[i].number_of_memory_violations > 0 || g_process[i].stopped_because_of_memory)
            {
                floatToChar(g_process[i].max_memory / 1024.0f / 1024.0f, &almData[4]);
                floatToChar((float)g_process[i].number_of_allowed_memory_violations, &almData[12]);
                if (g_process[i].stopped_because_of_memory)
                {
                    floatToChar(g_process[i].memory_at_stop_time / 1024.0f / 1024.0f, almData);
                    floatToChar((float)g_process[i].num_of_memory_violations_at_stop_time, &almData[8]);
                    SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_ERR_MEMORY_TOO_HIGH, almData, almTRANSIENT);
                    g_process[i].last_restart_reason = MEMORY_TOO_HIGH;
                }
                if (g_process[i].number_of_allowed_memory_violations > 0 && !g_process[i].is_starting)
                {
                    floatToChar(g_process[i].current_memory_consumption / 1024.0f / 1024.0f, almData);
                    floatToChar((float)g_process[i].number_of_memory_violations, &almData[8]);
                    SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_WARN_MEMORY_TOO_HIGH, almData, flag);
                }
            }
            if (g_process[i].tine_status != -2 && (g_process[i].property_link_timeout || g_process[i].stopped_because_of_link))
            {
                if (g_process[i].stopped_because_of_link)
                {
                    SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_ERR_LINK_TIMEOUT, NULL, almTRANSIENT);
                    g_process[i].last_restart_reason = LINK_TIMEOUT;
                }
                if (g_process[i].property_link_timeout && !g_process[i].is_starting)
                {
                    SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_WARN_LINK_TIMEOUT, NULL, flag);
                }
            }
        }
        if (!g_process[i].is_running && !g_process[i].stopped_by_user)
        {   
            if (!g_process[i].is_starting && g_process[i].disabled)
            {
                SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_ERR_DISABLED, NULL, 0);
            }
            else if (!g_process[i].has_alarm && !g_process[i].is_starting)
            {
                SetDeviceAlarm(LOCAL_NAME, g_process[i].id, A_ERR_NOT_RUNNING, NULL, almTRANSIENT);
                g_process[i].last_restart_reason = UNKNOWN;
            }
            g_process[i].has_alarm = 1;
        }
    }
}

void preSystemInit(void)
{
    SetCycleDelay(0);
    if (this_info.context != NULL)
    {
        RegisterFecInformation(this_info.fec_name, this_info.subsystem, this_info.context, this_info.description, this_info.location, this_info.hardware, this_info.responsible, this_info.port_offset);
    }
}

void postSystemInit(void)
{
    RegisterEquipmentModule(this_info.server_name, LOCAL_NAME, num_processes + 1, wdeqm, wdinit, wdbgk, this_wait_period, Shutdown);
}

void initServer(unsigned int wait_period, unsigned int polling_rate, SERVER_INFO info)
{
    int cc;
    char errstr[128];
    SetExcludePointsOfInterest(TRUE);
    SetUseLoopback(TRUE);
    SetPreSystemInitFunction(preSystemInit);
    SetPostSystemInitFunction(postSystemInit);

    this_wait_period = wait_period;
    this_info = info;

    if ((cc = SystemInit(TRUE)) != 0)
    {
        wlog("Server initialization error: %s\n", "Server", GetLastLinkError(cc, errstr));
        feclog("Server initialization error: %s\n", GetLastLinkError(cc, errstr));
        exit(1);
    }
    MinPollingRate = polling_rate;
    server_reload_requested = 0;
}

void StartServer(unsigned int wait_period, unsigned int polling_rate, SERVER_INFO info)
{
    time_t now;
    AssignFecApplicationInfo(WAPPV_MAJOR, WAPPV_MINOR, WAPPV_REVISION);
    initServer(wait_period, polling_rate, info);
    while (running)
    {
        if (server_reload_requested > 0)
        {
            now = time(0);
            if ((now - server_reload_requested)*1000 > (4 * wait_period))
            {
                ReloadProcesses();
                createStructures();
                registerPropertiesAndDevices();
                server_reload_requested = 0;
            }
            Wait(1);
        }
        else 
        {
            SystemCycle(TRUE);
        }        
    }
    running = FALSE;
    FreeResources();
    if (wd_process_info != NULL) free(wd_process_info);
    wd_process_info = NULL;
}

void StopServer(void)
{
    running = FALSE;
}

int StartServerMonitor(WATCHED_PROCESS *process, void(*callback)(int, int))
{
    int cc, ans;
    char errstr[128];
    char ip[16];
    PrpQueryStruct st;
    PrpQueryStruct *ptrSt = &st;

    if (allow_monitoring_other_hosts == 0)
    {
        ExpDataStruct srvaddr;
        FecAddrStruct fecaddr;
        GetServerAddr(process->context, process->server, &srvaddr, &fecaddr);
        sprintf_s(ip, 16, "%d.%d.%d.%d", fecaddr.netAdr.ipv4[0], fecaddr.netAdr.ipv4[1], fecaddr.netAdr.ipv4[2], fecaddr.netAdr.ipv4[3]);

        if (!ip_set)
        {
            strcpy_s(my_ip_address, 16, GetMyIpAddress());
            ip_set = TRUE;
        }

        if (strncmp(ip, my_ip_address, 16) != 0)
        {
            wlog("The IP address (%s) of process [%s] is different than the IP of this watchdog (%s). The property will not be monitored.\n", "Link", ip, process->alias, my_ip_address);
            process->property_disabled = TRUE;
            process->tine_status = -2;
            return -1;
        }
    } 

    cc = GetTargetPropertyInformation(process->tine_device, process->tine_property, fmt, siz, NULL);
    if (cc == non_existent_property)
    {
        ans = 1;
        cc = GetSystemPropertyInformation(process->tine_device, process->tine_property, &ptrSt, &ans);
        if (cc == 0 && ans > 0)
        {
            fmt[0] = st.prpFormat;
            siz[0] = st.prpSize;
        }
    }
    
    process->dout_allocated_by_me = FALSE;
    if (cc == 0)
    {
        DTYPEZERO(process->dout);
        process->dout.dFormat = fmt[0];
        process->dout.dArrayLength = siz[0];
        process->dout.data.vptr = calloc(siz[0], GetFormatSize(fmt[0]));
        if (process->dout.data.vptr == NULL) OutOfMemoryExit(3);
        process->dout_allocated_by_me = TRUE;
        process->default_property_info_used = FALSE;
    }     
    else 
    {
        wlog("Failed to read property information for %s/%s for process [%s]. Error: %s.\n", "Link",
            process->tine_device, process->tine_property, process->alias, GetLastLinkError(cc, errstr));
        process->dout.dFormat = CF_DEFAULT;
        process->dout.dArrayLength = 1;
        process->dout.data.vptr = NULL;
        process->default_property_info_used = TRUE;
    }

    cc = AttachLink(process->tine_device, process->tine_property, &(process->dout), NULL, CA_READ, process->polling_rate, callback, CM_POLL);
    if (cc < 0)
    {
        wlog("Failed to establish a link to %s/%s for process [%s] with format %d. AttachLink Error: %s.\n", "Link",
            process->tine_device, process->tine_property, process->alias, *fmt, GetLastLinkError(-cc, errstr));
        process->tine_status = -cc;
        process->link_id = -1;
    }
    else
    {
        wlog("Created connection to %s/%s for process [%s]\n", "Link", process->tine_device, process->tine_property, process->alias);
        process->link_id = cc;
        process->last_tine_response = time(0);
    }

    process->last_tine_retry = time(0);
    process->property_disabled = FALSE;

    return cc;
}

int CloseServerMonitor(WATCHED_PROCESS *process)
{
    int cc;
    char errstr[128];

    if (process->link_id < 0) return 0;
    cc = CloseLink(process->link_id);
    if (cc == 0)
    {
        wlog("Closed connection to %s/%s for process [%s].\n", "Link", process->tine_device, process->tine_property, process->alias);
        process->link_id = -1;
        process->tine_status = -1;
        if (process->dout.data.vptr != NULL && process->dout_allocated_by_me) free(process->dout.data.vptr);
        process->dout.data.vptr = NULL;
        process->dout_allocated_by_me = FALSE;
        return 0;
    }
    else
    {
        wlog("Failed to close the link to %s/%s for process [%s]. CloseLink Error : %s.\n", "Link", process->tine_device, process->tine_property, process->alias, GetLastLinkError(-cc, errstr));
        return cc;
    }
}
