/*
 * File TBufferedSrv.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file TBufferedSrv.c
 * \brief TINE Buffered Server DLL documentation 
 *
 */
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "tine.h"

__declspec(dllimport) short PASCAL SetBufferedCompletionA(int compCode);
__declspec(dllimport) short PASCAL AttachServerA(char *srvExportName,char *srvEQPName,int ndevices);
__declspec(dllimport) short PASCAL AttachDatabaseA();
__declspec(dllimport) short PASCAL RegisterDeviceNameA(char *devName, short devNumber);
__declspec(dllimport) short PASCAL GetRegisteredPropertiesA(NAME32 *prps, short *nprps);
__declspec(dllimport) short PASCAL RegisterBufferedPropertyA
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc
);
__declspec(dllimport) short PASCAL pushBufferedDataA(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule);
__declspec(dllimport) short PASCAL pullBufferedDataA(char *prpName,char *devName,BYTE *prpData,long prpSiz);
__declspec(dllimport) short PASCAL hasInputChangedA(char *prpName);
__declspec(dllimport) short PASCAL GetInputDeviceNumberA(void);
__declspec(dllimport) short PASCAL RegisterServerCallbackA(char *prpName, int (*cb)(void));
__declspec(dllimport) short PASCAL RegisterServerNotifierA(char *prpName, void (*nf)(int));
__declspec(dllimport) short PASCAL RegisterServerNotifierExA(char *prpName, void (*nf)(int), int nid);


__declspec(dllexport) short PASCAL SetBufferedCompletion(int compCode)
{
  return SetBufferedCompletionA(compCode);
}
__declspec(dllexport) short PASCAL AttachServer(char *srvExportName,char *srvEQPName,int ndevices) 
{
  return AttachServerA(srvExportName,srvEQPName,ndevices);
}
__declspec(dllexport) short PASCAL AttachDatabase()
{
  return AttachDatabaseA();
}
__declspec(dllexport) short PASCAL RegisterDeviceNameX(char *devName, short devNumber) 
{
  return RegisterDeviceNameA(devName,devNumber);
}
__declspec(dllexport) short PASCAL GetRegisteredProperties(NAME32 *prps, short *nprps)
{
  return GetRegisteredPropertiesA(prps,nprps);
}
__declspec(dllexport) short PASCAL RegisterBufferedProperty
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc
) 
{
  return RegisterBufferedPropertyA(prpName,prpInSiz,prpInFmt,prpOutSiz,prpOutFmt,prpMax,prpMin,prpEgu,access,prpDsc);
}
__declspec(dllexport) short PASCAL pushBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule)
{
  return pushBufferedDataA(prpName,devName,prpData,prpSiz,prpSchedule);
}
__declspec(dllexport) short PASCAL pullBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz)
{
  return pullBufferedDataA(prpName,devName,prpData,prpSiz);
}
__declspec(dllexport) short PASCAL hasInputChanged(char *prpName)
{
  return hasInputChangedA(prpName);
}
__declspec(dllexport) short PASCAL GetInputDeviceNumber(void)
{
  return GetInputDeviceNumberA();
}
__declspec(dllexport) short PASCAL RegisterServerCallback(char *prpName, int (*cb)(void))
{
  return RegisterServerCallbackA(prpName,cb);
}
__declspec(dllexport) short PASCAL RegisterServerNotifier(char *prpName, void (*nf)(int))
{
  return RegisterServerNotifierA(prpName,nf);
}
__declspec(dllexport) short PASCAL RegisterServerNotifierEx(char *prpName, void (*nf)(int), int nid)
{
  return RegisterServerNotifierExA(prpName,nf,nid);
}

BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
	switch (fdwReason)
	{
	 case DLL_PROCESS_ATTACH:  // DLL being loaded
			break;
	 case DLL_THREAD_ATTACH:
			break;
	 case DLL_THREAD_DETACH:
			break;
	 case DLL_PROCESS_DETACH:  // DLL being unloaded
			break;
  }
  return (TRUE);
}
