/*
 * File alminfocheck.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
int almInfoCheck(char *cnt, char *cmpstr, char *field, int sev, int sys);
int ReportError(int cc);

char *usageTxt[] = {
  "\nScans the alarm definintions for the given context and outputs those meeting the comparison criteria\n",
  "alminfocheck /c=<context> /<field><comparison><value> [/d=debug level]",
  "\n\t <field> can be 'severity' or 'system'"
  "\n\te.g. alminfocheck /c=TEST /severity^>=8",
  "\n\t (note the escape '^' character needed when > or < are used.)"
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  int i, len = 0, n = 0, sev = 0, sys = 0;
  char *c, *ctx = NULL, *fld = "severity", *cmp = NULL;
  char *tok = "!><=", *cp=NULL;
  char scratch[32], context[64];
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 200;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  printf("\n");
  ctx = argv[1];
  if (argc < 2)
  {
  dumpUsage:
    for (i = 0; i < nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n", usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  for (i = 1; i < argc; i++)
  { /* options given */
    c = argv[i];
    if (*c == '/') c++;
    switch (tolower(*c))
    {
      case 'c':
        if ((c=strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(context, c, 32);
          ctx = context;
        }
        continue;
      case 's':
        if (!strncmp(c, "sy", 2))
        {
          sys = -1;
          fld = "system";
        }
        if ((cmp=strpbrk(c, tok)) != NULL)
        {
          for (c=cmp; strpbrk(c, tok) != NULL; c++);
          if (sys < 0)
            sys = atoi(c);
          else
            sev = atoi(c);
          *c = 0;
        }
        continue;
      case 'd':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          tineDebug = atoi(scratch);
          foregroundTTY = TRUE;
        }
        continue;
      case 'h':
      case '?':
        goto dumpUsage;
      default:
        continue;
    }
  }
  return ReportError(almInfoCheck(ctx,cmp,fld,sev,sys));
}
