/*
 * File errlist.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tinedefs.h"
#include "errors.h"
#include "srvcore.h"

TINE_EXPORT char erlst[][32] = {
  "success",
  "illegal line number",
  "illegal format specification",
  "illegal arithmetic expression",
  "ambiguous request",
  "illegal delimiter",
  "attempt to divide by zero",
  "working area full",
  "nonexistent name",
  "transport medium is invalid",
  "data size mismatch",
  "no data found in range",
  "not allocated",
  "nonexistent line addressed",
  "illegal data size",
  "i/o error",
  "illegal context",
  "runtime error",
  "system error",
  "hardware operation in progress",
  "parameter error",
  "file error",
  "resort to stream transport",
  "array dimension error",
  "square root of neg.number",
  "data buffer too small",
  "string too long",
  "ipx socket error",
  "net read error",
  "call not ready",
  "invalid transport size",
  "logarithm argument <= 0",
  "device not connected",
  "unauthorised action",
  "hardware error",
  "illegal device",
  "illegal property",
  "value out of range",
  "not implemented yet",
  "no such computer",
  "tagged struct is sealed",
  "syntax error",
  "no such file",
  "file already exists",
  "no file space",
  "link timeout",
  "remitted data lost",
  "end of file",
  "archive in progress",
  "server name in use",
  "no such column",
  "out of client memory",
  "network access required",
  "address unresolved",
  "invalid or missing property",
  "address unknown",
  "name unknown",
  "invalid keyword",
  "invalid link handle",
  "String expected",
  "out of local memory",
  "out of shared memory",
  "invalid structure tag",
  "invalid index",
  "illegal equipment name",
  "link error",
  "code failure",
  "nonexistent server",
  "function deprecated",
  "not supported",
  "parsing error",
  "illegal relational operator",
  "not allowed",
  "illegal read/write attempted",
  "out of server memory",
  "database not loaded",
  "illegal command",
  "resources exhausted",
  "header not found",
  "sedac error",
  "reacquire address",
  "semaphore error",
  "driver not installed",
  "hardware port not available",
  "asynchronous access required",
  "semaphore blocked",
  "nonexistent element",
  "nonexistent fec",
  "nonexistent client",
  "unable to lock the memory",
  "server not responding",
  "result not posted",
  "command not accepted",
  "operation timeout",
  "illegal network protocol",
  "gpib error",
  "rs232 error",
  "operation busy",
  "connection timeout",
  "illegal mode",
  "not owner",
  "not defined",
  "net write error",
  "invalid data",
  "software error",
  "access denied",
  "tcp not supported",
  "ipx not supported",
  "tcp host not resolved",
  "tcp connect error",
  "tcp socket error",
  "configuration error",
  "invalid return code",
  "link is blacklisted",
  "link exists",
  "alarms buffer full",
  "not exported",
  "is switching",
  "at limit",
  "get subscription id",
  "renegotiate contract",
  "server redirection",
  "value too high",
  "value too low",
  "warn too high",
  "warn too low",
  "wait for completion",
  "cannot select",
  "has query function",
  "not signalled",
  "call redirection",
  "udp socket error",
  "mutex error",
  "data not local",
  "name already exists",
  "already assigned",
  "invalid field",
  "multicast init error",
  "invalid multicast group",
  "nonexistent function",
  "use channel array",
  "invalid name",
  "invalid value",
  "device error",
  "device not ready",
  "device busy",
  "invalid time interval",
  "notification pending",
  "thread error",
  "process is terminating",
  "name server unknown",
  "access lock required",
  "not initialized",
  "invalid structure size",
  "canbus error",
  "profibus error",
  "data stale",
  "has bitfield tag",
  "not locked",
  "lock expired",
  "not registered",
  "nonexistent property",
  "memory overwrite",
  "server idle",
  "not applicable",
  "access locked",
  "invalid data reference",
  "has structure tag",
  "disk space is low",
  "disk space is very low",
  "information is static",
  "reset mca property",
  "record too long",
  "unix socket or pipe error",
  "shared memory error",
  "tcp connection closed",
  "service unavailable",
  "device offline",
  "out of sequence",
  "invalid data request",
  "use elevated channel array",
  "name not found",
  "out of tolerance",
  "file not open",
  "low main memory",
  "tcp access required"
};

int numErr = sizeof(erlst)/sizeof(erlst[0]);

TINE_EXPORT int GetErrorListSize(void) { return numErr; }

char logLevelList[][8] = {
  "INFO",
  "WARN",
  "ERROR",
  "FATAL"
};

int numLogLevels = sizeof(logLevelList)/sizeof(logLevelList[0]);

TINE_EXPORT int GetLogLevelListSize(void) { return numLogLevels; }

TINE_EXPORT char *errLongMsg[] = {
  "the call completed successfully",
  "the hardware bus line requested is invalid",
  "the data type format is invalid or not supported for this call",
  "an illegal arithmetic expression was encountered",
  "ambigous request or missing input information",
  "illegal delimiter in expression",
  "an operation encountered division by zero",
  "an internal working area buffer is at capacity",
  "requested target does not exist",
  "the requested transport medium is not valid",
  "incoming data payload larger than requested (internal protocol error)",

  "no data found in requested time range",
  "requested target is not allocated in the given context",
  "expected line was not found in database",
  "give data size is illegal or not valid",
  "bus or network io error",
  "the requested context is invalid or does not exist",
  "the request encountered a runtime error",
  "the request encountered a system internal error",
  "the requested operation could not be completed because the targeted hardware is busy",
  "an input argument or parameter is invalid",

  "the request needed to access a file and encountered an error",
  "datagram message size exceeds the allowable mtu (reserved)",
  "the requested array size does not match that of the target or is too large",
  "the operation encounterd the square root of a negative number",
  "the supplied buffer is too small to handle the request",
  "supplied string is too long to buffer and cannot be truncated",
  "the network encountered an ipx socket error (reserved)",
  "the request encountered a network read error",
  "call to equipment module dispatch is not ready and will be repeated (internal, reserved)",
  "total message size exceeds the configured allowable message size",

  "the operation encounterd the log of a negative or zero number",
  "requested device is not connected to the server",
  "requested operation is not authorized for the user making the request",
  "the operation encountered a hardware error",
  "requested device is not available or is not known to the server",
  "requested property is not exported by the server or service (request will be blacklisted)",
  "the given input value is out of range",
  "the requested operation has not been implemented",
  "requested host computer does not exist",
  "the tagged structure has already been sealed and cannot accept further field entries (reserved)",

  "expression syntax error (legacy NODAL error currently unused)",
  "the requested operation tried to open a file which does not exist",
  "the requested operation tried to create a file which already exists",
  "there is not enough capacity in pre-allocated random access file to complete the requested operation",
  "a persistent link update did not occur within the expected time frame",
  "the returned data payload contained only a portion of the requested data set (reserved)",
  "the request attempted to read past the end of a file (unused)",
  "an archive operation is already in progress",
  "an attempt was made to register a device server name which is already being used",
  "the expected column was not found in a csv database",

  "a client call could not allocate sufficient memory",
  "access requires CM_NETWORK which was not supplied! (used in contract coercion)",
  "the requested target address could not be resolved",
  "the requested property is not properly formed",
  "the requested target address could not be found (request will be blacklisted)",
  "the requested name could not be found",
  "the input keyword (e.g. globals keyword) is invalid",
  "the input link id specified is invalid",
  "a database file read encountered an end-of-line while expecting a string",
  "a local call to allocated memory was not successful",

  "a call to obtain a shared-memory buffer was not successful",
  "the specified structure tag is not in the structure registry",
  "the table index given does not reference a table entry",
  "the equipment module name specified is not valid",
  "internal error processing a data link (reserved)",
  "an internal failure occured due to unhandled exceptions, unavailable resource, OS failure, etc.",
  "the requested server does not exist (request will be blacklisted)",
  "the requested action or functionality has been deprecated",
  "the requested action or functionality is not supported",
  "the request encountered a error parsing input (illegal character?)",

  "the input arithmetic or logical operator is not allowed",
  "the requested operation is not allowed",
  "the requested access is not allowed (typically a WRITE was attempted on a read-only property)",
  "a server call could not allocate sufficient memory",
  "the requested operation requires a database which has not been loaded",
  "the requested command is illegal",
  "the requested operation would exceed the configured capacity of an allocation table (clients, contracts, connections, etc.)",
  "the expected csv file does not have a header",
  "a sedac bus error was encountered",
  "a signal to reacquire a server's address from ENS (reserved, internal)",

  "a semaphore or mutex could not be obtained or released",
  "the request requires a hardware driver which is not installed",
  "the request requires access to a port which is not available",
  "access to selected property should be asynchronous (contract coercion)",
  "a semahpre or mutex could not be obtained within the requested time",
  "the requested equipment modules does not exist in the registry (request will be blacklisted)",
  "the requested encountered a FEC address which does not exist",
  "the requested encountered a non-existent client (unused)",
  "the request was not able to obtain an access lock or lock resources",
  "the requested resource is not running",

  "a dispatch routine did not post a reply",
  "the requested operation was refused",
  "the requested operation did not complete within the time limit specified",
  "the incoming tine protocol is not recognized (reserved, internal)",
  "the request encountered a gpib bus error",
  "the request encountered a rs232 bus error",
  "the requested operation is busy and needs more time to complete",
  "a synchronous link did not return within the time limit specified",
  "the requested access mode is not valid",
  "the caller is not the owner of the requested resource",

  "the requested operation has not yet been defined",
  "the request was unable to send data on the network",
  "the input data are not valid",
  "the request encounterd a general software error",
  "call was refused by the security subsystem",
  "an attempt to use TCP/IP access was made and is not supported on this host (reserved)",
  "an attempt to use IPX access was made and is not supported on this host (reserved)",
  "the requested host could not be resolved (request will be blacklisted)",
  "unable to connect to a TCP/IP socket (reserved)",
  "general TCP/IP socket error (reserved)",

  "the request encountered a resource configuration error (e.g. registering redirection information)",
  "indicates a systematic error in processing the equipment dispatch handler (reserved)",
  "the requested link parameters cannot succeed (due to initial return of non_existent_elem, address_unknown, etc.)",
  "the requested link already exists in the link table indicating a dependent link (reserved,internal)",
  "the request attempted to add an alarm to the local alarm system whose capacity has been reached",
  "not exported (unused)",
  "switching transition is in progress (unused)",
  "setting/value is at a maximum or minimum or range exceeds storage (local history)",
  "code used by access mode CM_NETWORK (reserved, internal)",
  "code used to repackage output data when few bytes are returned than requested (reserved, internal)",

  "code used to supply redirection information (reserved, internal)",
  "the returned value is out of range: too high",
  "the returned value is out of range: too low ",
  "the returned value is approaching too high ",
  "the returned value is approaching too low ",
  "a dependent link is waiting for parent to complete",
  "select() or poll() operation failed (reserved)",
  "query call to PROPERTIES or DEVICES has completed but indicated a hierachical preference (reserved)",
  "contract is waiting to call the equipment module dispatch routine (reserved, internal)",
  "nominal alias for server_redirection",

  "general UDP socket error (reserved)",
  "a mutex could not be obtained or released",
  "indicates that a device within a wildcard call needs to span multiple servers (reserved, internal)",
  "a structure or bitfield field name already exists",
  "the requested resource has already been assigned",
  "a structure or bitfield field is invalid",
  "could not initialize a multicast socket (reserved)",
  "multicast group address is invalid",
  "an equipment module's dispatch routine is not registered",
  "requested property is a multi-channel array (contract coercion)",

  "the specified input name is invalid",
  "the specified input value is invalid",
  "the requested device is in an error state",
  "the requested device is not ready",
  "the requested device is busy",
  "the requested polling interval is invalid (contract coercion)",
  "a link is pending notification (reserved, internal)",
  "a thread operation failed",
  "the device server is terminating (reserved, internal)",
  "the equipment name server address is unknown (reserved, internal)",

  "an access lock is required to process the operation",
  "server is not yet initialized",
  "the input structure size does not match the size in the structure registry",
  "the request encountered a can bus error",
  "the request encountered a profibus error",
  "contract's data are stale (reserved)",
  "the reqested contract needs to request a bitfield (reserved, internal)",
  "the requested resource does not have an access lock (reserved)",
  "the requested access lock has expired (reserved)",
  "the requested property (e.g. SystemScheduleProperty()) has not been registered",

  "the requested property does not exist",
  "a heap error within a dispatch routine has been detected",
  "the server is in an idle state and is not handling requests",
  "the requested opertion is not applicable in the current context",
  "access temporarily suspended due to access lock",
  "an invalid data reference handle was encountered",
  "reqested contract needs to request a structure (reserved, internal)",
  "a warning that disk space is low (local alarm system)",
  "a signal that disk space is very low (local alarm system) ",

  "a signal that static information is being 'polled' (contract coercion)",
  "a signal to re-establish all mca single element links",
  "record length is too long to process within the tine kernel",
  "an error condition on a unix (pipe) socket was encountered",
  "an error occured accessing shared memory",
  "a tcp peer has closed the connection",
  "the requested service is unavailable",
  "the requested device is offline",
  "data value is out of sequence",
  "requested data length or type is invalid (contract coercion)",

  "the requested property is a multi-channel array with elevated format (contract coercion)",
  "named resource was not found in database",
  "the returned data are out of tolerance",
  "a read/write attempt was made on a file which is not open",
  "low available main memory detected",
  "access requires CM_CONNECT which was not supplied! (used in contract coercion)"
};

int numErrLongMsg = sizeof(errLongMsg) / sizeof(errLongMsg[0]);

int isCoerciveStatus(int code)
{
  switch (code)
  {
    case server_redirection:
    case has_structure_tag:
    case has_bitfield_tag:
    case mcast_access_required:
    case invalid_datarequest:
    case invalid_interval:
    case tcp_access_required:
    case property_is_mca:
    case reset_mca_property:
    case property_is_mca_elevated:
      return TRUE;
    default:
      return FALSE;
  }
}
TINE_EXPORT int isTimeout(int code)
{
  switch (code)
  {
    case link_timeout:
    case connection_timeout:
    case remitted_data_lost:
    case operation_busy:
    case operation_timeout:
      return TRUE;
    default:
      return FALSE;
  }
}
TINE_EXPORT char *cc2str(int code)
{
  static char str[STATUS_SIZE];
  code = BASE_CC(code);
  if (code < numErr) return erlst[(code)& 0xff];
  sprintf(str, "Error %d", code);
  return str;
}
TINE_EXPORT char *logLevel2str(int code)
{
  if (code < numLogLevels) return logLevelList[(code)& 0xff];
  return "UNKNOWN";
}
TINE_EXPORT char *cc2longString(int code)
{
  static char str[64];
  if (code < numErrLongMsg) return errLongMsg[(code)& 0xff];
  sprintf(str, "Error %d (not a system error code)", code);
  return str;
}
int numRegisteredErrors = 0;
ErrTblItem *gRegisteredErrors = NULL;
HashRefXRef **RegisteredErrorsXRef = NULL;
int addErrTableItemToHash(ErrTblItem *eti)
{
  int idx;
  char hashStr[32];
  HashRefXRef *xref = NULL;
  sprintf(hashStr, "%d", eti->code);
  idx = (int)(ElfHash((unsigned char*)hashStr) % ERRORTBL_HASH_SIZE);
  if ((xref = (HashRefXRef*)SystemCalloc(1, sizeof(HashRefXRef))) == NULL) return out_of_server_memory;
  xref->ref = eti;
  xref->next = RegisteredErrorsXRef[idx];
  RegisteredErrorsXRef[idx] = xref;
  return 0;
}
TINE_EXPORT int RegisterErrorCode(int code, char *msgShort, char *msgLong)
{
  int cc;
  ErrTblItem *eti = NULL;
  if (code < 512) return invalid_return_code;
  if ((eti=(ErrTblItem *)SystemCalloc(1, sizeof(ErrTblItem))) == NULL) return out_of_local_memory;
  eti->code = code;
  if (msgShort != NULL) strncpy(eti->msg, msgShort, 32);
  if (msgLong != NULL) strncpy(eti->msgLong, msgLong, 128);
  eti->nxt = gRegisteredErrors;
  gRegisteredErrors = eti;
  numRegisteredErrors++;
  if (numRegisteredErrors < ERRORTBL_HASH_PAYOFF) return 0;
  if (RegisteredErrorsXRef == NULL)
  { /* got enough to make a hash table */
    if ((RegisteredErrorsXRef=(HashRefXRef **)SystemCalloc(ERRORTBL_HASH_SIZE, sizeof(HashRefXRef*))) == NULL)
      return out_of_local_memory;
    for (eti = gRegisteredErrors; eti != NULL; eti = eti->nxt) if ((cc=addErrTableItemToHash(eti)) != 0) return cc;
    return 0;
  }
  return addErrTableItemToHash(eti);
}
void freeErrorTable()
{
  if (numRegisteredErrors == 0) return;
  if (RegisteredErrorsXRef != NULL)
  {
    ErrTblItem *eti = NULL, *neti = NULL;
    HashRefXRef *xref = NULL, *nref = NULL;
    int i;
    for (i = 0; i < ERRORTBL_HASH_SIZE; i++)
    { /* free the hash table if there */
      for (xref=RegisteredErrorsXRef[i]; xref != NULL; xref = nref)
      {
        nref = xref->next;
        SystemFree(xref);
      }
    }
    for (eti = gRegisteredErrors; eti != NULL; eti = neti)
    { /* free the registered errors list */
      neti = eti->nxt;
      SystemFree(eti);
    }
  }
  numRegisteredErrors = 0;
}
ErrTblItem *getErrTblItem(int code)
{
  int idx;
  char hashStr[32];
  ErrTblItem *eti = NULL;
  HashRefXRef *xref = NULL;
  if (numRegisteredErrors < ERRORTBL_HASH_PAYOFF)
  {
    for (eti = gRegisteredErrors; eti != NULL; eti = eti->nxt)
    {
      if (eti->code == code) return eti;
    }
    return NULL;
  }
  sprintf(hashStr, "%d", code);
  idx = (int)(ElfHash((unsigned char*)hashStr) % ERRORTBL_HASH_SIZE);
  for (xref=RegisteredErrorsXRef[idx]; xref != NULL; xref = xref->next)
  {
    if (xref->ref == NULL) continue; /* shouuldn't be possible */
    if (xref->ref->code == code) return xref->ref;
  }
  return NULL;
}
TINE_EXPORT char *GetErrorText(int code,char *errstr,int length)
{
  char *err = NULL;
  ErrTblItem *eti = NULL;
  int longtext = length > 32 ? TRUE : FALSE;
  code = code & 0x0fff;
  if (code < 0) err = "(invalid code)";
  else if (code < numErr) err = longtext ? cc2longString(code) : cc2str(code);
  else if ((eti=getErrTblItem(code)) != NULL) err = longtext ? eti->msgLong : eti->msg;
  if (errstr != NULL)
  {
    if (err == NULL) sprintf(errstr, "Error %d : unknown", code); else strncpy(errstr, err, length);
    return errstr;
  }
  return err;
}
