/*
 * File tbufkrnl.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#ifdef UNIX
#include <dlfcn.h>
#endif
#include "tine.h"
#include "threader.h"
#include "tbufsrv.h"

extern int gBufSrvInitialized;
int cycle_busy = FALSE;
unsigned int idCycleTimer = 0;
int srvTerminated = FALSE;
int userNameSet = FALSE;

int gUseWindowsTimer = FALSE;
int gWindowsTimerInUse = FALSE;
int gThreadTimerInUse = TRUE;

void krnlTmrTask(void)
{
  if (!gBufSrvInitialized) return;
  if (srvTerminated) return;
  if (cycle_busy) return;
  cycle_busy = TRUE;
  SystemCycle(FALSE);
  cycle_busy = FALSE;
}

void krnlInitTask(int isServer)
{
  if (isServer) SystemInit(FALSE);
  gBufSrvInitialized = TRUE;
  MaxPollingRate = MinPollingRate = 20;
}

THREAD_TYPE krnlTmrThread(THREAD_PARAM lpvThreadParam)
{
  lpvThreadParam = lpvThreadParam;
  UseMultiThreadedCycler(TRUE);
  while (!srvTerminated)
  {
    krnlTmrTask();
  }
  return 0;
}
#if defined(UNIX)
int WaitForMutex(pthread_mutex_t *mx,int to);

pthread_attr_t gKrnlThreadAttr, *gPtrKrnlThreadAttr = &gKrnlThreadAttr;
pthread_mutex_t krnlThreadMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t *hKrnlThreadMutex = &krnlThreadMutex;
int krnlCreateThread(void)
{
  int cc;
  pthread_t tid;

  if ((cc=pthread_mutex_init(hKrnlThreadMutex,NULL)) != 0)
  {
    feclog("Cannot create Tine Kernel Mutex: error %d",cc);
    return 0;
  }
  if ((cc=pthread_mutex_lock(hKrnlThreadMutex)) != 0)
  {
    feclog("Cannot lock Tine Kernel Mutex: error %d",cc);
    return 0;
  }
  pthread_attr_init(gPtrKrnlThreadAttr);
# ifdef FORCE_PTHREAD_SCOPE_SYSTEM
  if ((cc=pthread_attr_setscope(gPtrKrnlThreadAttr,PTHREAD_SCOPE_SYSTEM)) != 0) 
  {
    feclog("Cannot set thread scope !\n");
  }
# endif
  cc = pthread_create(&tid,gPtrKrnlThreadAttr,krnlTmrThread,NULL);
  feclog("Create Kernel Thread : %s",erlst[cc]);
  return cc ? 0 : (int)tid;
}

void TineStartCycler(void)
{
  static int hasStarted = FALSE;
  if (hasStarted) return;
  hasStarted = TRUE;
  krnlCreateThread();
  krnlInitTask(FALSE);
  return;
}
__attribute__((destructor)) void TineStopCycler(void)
{
  srvTerminated = TRUE;
  millisleep(20);
}
#endif /* unix */
