/*
 * File tbufsrv.h
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** \file tbufsrv.h
 * \brief TINE buffered server documentation 
 *
 */
#ifndef TBUFSRV_H_
#define TBUFSRV_H_

#include "tine.h"
#include "listener.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifdef TBUFSRVDLL_EXPORTS
# define EXPORT_B __declspec(dllexport) 
#endif 
#ifdef TBUFSRVDLL_IMPORTS
# define EXPORT_B __declspec(dllimport) 
#endif
#ifndef EXPORT_B
# define EXPORT_B 
#endif

#define  bufsrvMajor 5
#define  bufrvMinor 1
#define  bufsrvRevision 7

typedef void (*NTFRCBP)(int);
typedef int (*HNDLRCBP)(void);
typedef int (*XHNDLRCBP)(char *,char *,void *);

EXPORT_B int AttachServer(char *srvExportName,char *srvEQPName,int ndevices);
EXPORT_B int AttachServerEx(char *srvExportName,char *srvEQPName,int ndevices,void (*tmr)(void),int tmrInterval);
EXPORT_B int AttachServerFromFecXml(void);
EXPORT_B void runServer(void);
EXPORT_B int pushBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule);
#define PushBufferedData(pn,dn,pd,ps,s) pushBufferedData(pn,dn,pd,ps,s)
EXPORT_B int pushBufferedDataEx(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule,double timeStamp,int sysStamp,int usrStamp);
#define PushBufferedDataEx(pn,dn,pd,ps,s,ts,ss,us) pushBufferedDataEx(pn,dn,pd,ps,s,ts,ss,us)
EXPORT_B int pullBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz);
#define PullBufferedData(pn,dn,pd,ps) pullBufferedData(pn,dn,pd,ps)
EXPORT_B int pushImageData(char *prpName,char *devName,TImageSourceHeader *sourceHeader,TImageFrameHeader *frameHeader,BYTE *frameBytes,int frameSizeBytes,int prpSchedule);
EXPORT_B int hasInputChanged(char *prpName);
EXPORT_B int RegisterServerCallback(char *prpName, HNDLRCBP cb);
EXPORT_B int RegisterServerCallbackEx(char *prpName, XHNDLRCBP cb, void *ref);
EXPORT_B int GetInputDeviceNumber(void);
EXPORT_B int RegisterServerNotifierEx(char *prpName, NTFRCBP nf, int nid);
EXPORT_B int RegisterServerNotifier(char *prpName, NTFRCBP nf);
EXPORT_B int GetRegisteredProperties(NAME32 *prps, short *nprps);
EXPORT_B char *GetEquipmentModuleName(void);
EXPORT_B void TerminateServer();
EXPORT_B int SetBufferedCompletion(int compCode);
EXPORT_B int SetBufferedPropertyCompletion(char *prpName, int compCode);
EXPORT_B int SetBufferedPropertyStatus(char *prpName, int compCode);
EXPORT_B void SetUnpushedDataCompletionCode(int compCode);
EXPORT_B int GetUnpushedDataCompletionCode(void);
EXPORT_B int RegisterBufferedProperty
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc
);
EXPORT_B int RegisterBufferedPropertyEx
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc, int prpId
);
EXPORT_B int RegisterBufferedPropertyEx2
(
  char *prpName, long prpInSiz, short prpInFmt, char *prpInTag, long prpOutSiz, short prpOutFmt, char *prpOutTag,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc, int prpId, int arrayType, int rowLength
);
EXPORT_B int GetRegisteredPropertyInformation
(
  char *eqm, char *prp, /* input */
  float *max, float *min,char *units,
  int *sizeOut, int *sizeIn, char *fmtOut, char *fmtIn,
  char *description
);
EXPORT_B int RegisterBufferedDeviceName(char *devName,int devNr,char *devRdr,char *devDesc);
EXPORT_B int RegisterBufferedDeviceNameEx(char *devName,int devNr,int devMask,float zPos,char *devRdr,char *devDesc,char *devLocation);
EXPORT_B int getNotifiedProperty(char *prpName);
EXPORT_B int getNotifiedPropertyAndDevice(char *prpName,char *devName);
EXPORT_B void SetNotifierDeadband(int timeInMilliseconds);
EXPORT_B int RegisterCycleTriggerNotifier(void (*nf)(int), int nid);
EXPORT_B int GetTriggerCycle(int *cycleNumber,int *cycleStatus);
EXPORT_B int TriggerEvent(char *devname,char *cmt,short triggerLevel);
EXPORT_B int GetLastCaller(char *un,char *ip);
EXPORT_B int StartRuntimeDebugging(void);
EXPORT_B int SetBufferedDataSize(char *prpName,int dataSiz);
EXPORT_B int GetBufferedDataSize(char *prpName);
EXPORT_B int GetBufferIndex(char *prpName,char *devName);
EXPORT_B double GetTineTimeStampFromString(char *timestring);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* TBUFSRV_H_ */
