/*
 * File tbufsrv32.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "tine.h"
#include "tbufsrv.h"

TINE_EXPORT void SystemSetCycleTimer(void);
TINE_EXPORT void SystemKillCycleTimer(void);

extern int gBufSrvInitialized;
extern int gNumDevices;
extern char gEqmName[16];

__declspec(dllexport) short PASCAL SetBufferedCompletionA(int compCode)
{
  return (short)SetBufferedCompletion(compCode);
}
__declspec(dllexport) short PASCAL AttachServerA(char *srvExportName,char *srvEQPName,int ndevices) 
{
  return AttachServer(srvExportName,srvEQPName,ndevices);
}
__declspec(dllexport) short PASCAL AttachDatabaseA()
{
  return AttachServer(NULL,NULL,0);
}
// TODO:  shouldn't this pass the eqm param ?
__declspec(dllexport) short PASCAL RegisterDeviceNameA(char *devName, short devNumber) 
{
  return (short)RegisterDeviceName(gEqmName,devName,devNumber);
}
__declspec(dllexport) short PASCAL GetRegisteredPropertiesA(NAME32 *prps, short *nprps)
{
  return (short)GetRegisteredProperties(prps,nprps);
}
// TODO: move this to the general lib; parse the desc for atype; pass the eqm param?
__declspec(dllexport) short PASCAL RegisterBufferedPropertyA
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc
) 
{
  return (int)RegisterBufferedProperty(prpName,prpInSiz,prpInFmt,prpOutSiz,prpOutFmt,prpMax,prpMin,prpEgu,access,prpDsc);
}
__declspec(dllexport) short PASCAL pushBufferedDataA(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule)
{
  return (short)pushBufferedData(prpName,devName,prpData,prpSiz,prpSchedule);
}
__declspec(dllexport) short PASCAL pullBufferedDataA(char *prpName,char *devName,BYTE *prpData,long prpSiz)
{
  return (short)pullBufferedData(prpName,devName,prpData,prpSiz);
}
__declspec(dllexport) short PASCAL hasInputChangedA(char *prpName)
{
  return (short)hasInputChanged(prpName);
}
__declspec(dllexport) short PASCAL getNotifiedPropertyA(char *prpName)
{
  return (short) getNotifiedProperty(prpName);
}
__declspec(dllexport) short PASCAL getNotifiedPropertyAndDeviceA(char *prpName,char *devName)
{
  return (short) getNotifiedPropertyAndDevice(prpName,devName);
}
__declspec(dllexport) short PASCAL GetInputDeviceNumberA(void)
{
  return (short)GetInputDeviceNumber();
}
__declspec(dllexport) short PASCAL RegisterServerCallbackA(char *prpName, int (*cb)(void))
{
  return (short)RegisterServerCallback(prpName,cb);
}
__declspec(dllexport) short PASCAL RegisterServerNotifierA(char *prpName, void (*nf)(int))
{
  return (short)RegisterServerNotifier(prpName,nf);
}
__declspec(dllexport) short PASCAL RegisterServerNotifierExA(char *prpName, void (*nf)(int), int nid)
{
  return (short)RegisterServerNotifierEx(prpName,nf,nid);
}
__declspec(dllexport) void PASCAL SetNotifierDeadbandA(int timeInMilliseconds)
{
  SetNotifierDeadband(timeInMilliseconds);
}
__declspec(dllexport) short PASCAL GetLastCallerA(char *un,char *ip)
{
  return (short)GetLastCaller(un,ip);
}

void TineStopCycler(void)
{
  feclog("TineStopCycler has been called");
  // kill it if it's running
  SystemKillCycleTimer();
  gBufSrvInitialized = FALSE;
}
void TineStartCycler(void)
{
  feclog("TineStartCycler has been called");
  // start anew ...
  SystemSetCycleTimer();
  gBufSrvInitialized = TRUE;
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      SetConnectionTableCapacity(MexListenerTableCapacity);
      SetClnRecvQueueDepth(0);
      SystemSetCycleTimer();
      gBufSrvInitialized = TRUE;
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      SystemKillCycleTimer();
      freeMexListenerTable();
      break;
  }
  return (TRUE);
}
