/*
 * File tproplist.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int displayErrorMessages;
int tprops(char *dev);
int tproplist(char *dev);
int ReportError(int cc);
char *usageTxt[] = {
  "\nRetrieves a list of registered properties for the given device server\n",
  "tproplist /<context/<server>[/<device>] [/e=<TRUE,FALSE> /d=<debug level>]",
  "e.g. tproplist /HERA/HEPBLM",
  "  or tproplist /HERA/HEPBLM/WL105K1",
  "\nIf you specify a specific device name and it contains blanks use \"\"",
  "e.g. tproplist \"/HERA/HEPBPM/WL197 MX\"",
  "\nUse /e=TRUE or FALSE to display error messages when the call is not successfull (default=TRUE)",
  "\nSee also: 'tservers', 'tdevlist' for retrieving information",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  char *c, scratch[256];
  int i;
  if (argc < 2) 
  {
dumpUsage:
    for (i=0; i<nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  if (*(argv[1]) == '?' || !strncmp(argv[1],"/?",2)) goto dumpUsage;
  for (i=2; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'e':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          displayErrorMessages = strtobool(scratch);
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
          foregroundTTY = TRUE;
        }
        continue;
      case '?': goto dumpUsage;
      default:
        continue;
    }
  }  
  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  printf("\n");
  return ReportError(tproplist(argv[1]));
}
