/*
 * File tputhistory.c
 *
 * Authors Philip Duval (DESY), Steve Herb (DESY), Marina Nikolova (EMBL), 
 *    Karol Bartkiewicz (DESY), Marcus Walla (DESY), Jan Szczesny (DESY),
 *    Arthur Aghababyan (DESY), Stefan Weisse (DESY), Tobias Tempel (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * The TINE .c codeset made available by open source license ultimately
 * stems from the original ISOLDE control system ((c)CERN, released under
 * a permissive CERN licence akin to the later well known MIT licence)
 * in use at CERN in 1992.  Although none of the current software is in
 * anyway a redistribution of software from the ISOLDE system of that
 * era (apart from various error/status code definitions), it does in
 * fact inherit several conceptual ideas.  We therefore owe a substantial
 * debt to CERN for its inception, and would be remiss if we did not express
 * our gratitude in the TINE open source headers.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int defaultClientTimeout;
time_t parseTimeString(char *ts);
int ReportError(int cc);
char *usageTxt[] = {
  "\nPuts data and timestamps into the central archive system for 'SELF' entries",
  "\n        e.g. tputhistory PETRA MySelfEntry 1 2 3 4 5 /t=now /s=42 /u=1\n",
  "Usage : tputhistory <context> <keyword> [/t=<timestamp> /s=<systemstamp> /u=<userstamp>]"
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  int cc, i, fmt, nvals=0,sst=0, ust=0;
  char *c, scratch[256], tgt[128];
  char *context = NULL, *keyword = NULL;
  char timestamp[64] = "NOW";
  char **vals = NULL;
  BYTE *bvals = NULL;
  int *ivals = NULL;
  float *fvals = NULL;
  double ts = (double)time(NULL);
  if (argc < 3)
  {
    for (i = 0; i<nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n", usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  context = argv[1];
  keyword = argv[2];
  vals = (char **)calloc(argc - 3, sizeof(char *));
  for (i = 3; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c == '/') c++;
    switch (tolower(*c))
    {
      case 't':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(timestamp, c, 32);
        }
        ts = GetTimeStampFromString(timestamp);
        continue;
      case 's':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          sst = atoi(scratch);
          foregroundTTY = TRUE;
        }
        continue;
      case 'u':
        if ((c = strchr(c, '=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          ust = atoi(scratch);
        }
        continue;
      default:
        vals[nvals++] = c;
        continue;
    }
  }
  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  defaultClientTimeout = 50;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  sprintf(tgt, "/%.32s/ARCHIVER/#0[%.64s]", context,keyword);
  fmt = CF_INT32;
  for (i = 0; i < nvals && fmt == CF_INT32; i++) if (strchr(vals[i], '.')) fmt = CF_FLOAT;
  if (fmt == CF_INT32)
  {
    ivals = (int *)calloc(nvals, sizeof(int));
    for (i = 0; i < nvals; i++) ivals[i] = atoi(vals[i]);
    bvals = (BYTE *)ivals;
  }
  if (fmt == CF_FLOAT)
  {
    fvals = (float *)calloc(nvals, sizeof(int));
    for (i = 0; i < nvals; i++) fvals[i] = (float)strtod(vals[i],NULL);
    bvals = (BYTE *)fvals;
  }
  cc = PutArchivedData(tgt, bvals, fmt, nvals, ts, sst, ust);
  return cc;
}
