/*
 * CASEQM.c    REVISION HISTORY:
 * Generated by SRVWIZARD on Fri Jan 26 18:01:41 2001
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "threader.h"
#include "arctools.h"
#include "caseqm.h"
#include "apceqm.h"
#ifdef CAS2DOOCS
# include "cas2doocs.h"
#endif

#if defined(UNIX)
pthread_mutex_t gSendMailMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hSendMailMutex = &gSendMailMutex;
pthread_mutex_t gAmsTblMutex = PTHREAD_MUTEX_INITIALIZER;
MXHANDLE hAmsTblMutex = &gAmsTblMutex;
#elif defined(WIN32)
MXHANDLE hSendMailMutex;
MXHANDLE hAmsTblMutex;
#else
int hSendMailMutex = 0;
int hAmsTblMutex = 0;
#endif

/**
 * \internal
 * \note 
 * almDISABLED is used via local logic to exclude alarm messages from
 * queries.  
 * almSUPPRESS is a flag set externally to exclude alarm messages
 * from queries.
 *
 */
#define MIN_ARCHIVE_LEVEL 1

char casExportName[EXPORT_NAME_SIZE];
char AlarmArchivePath[80] = DEFAULT_ARCHIVE_PATH;
AMS g_sysalarmsBuffer[PRP_SYSALARMS_SIZE];
short g_numalarms;
CasTblEntry *casTbl = NULL;
int casTblSiz = 0;
int casTblCapacity = 0;
time_t g_AckThreshold = 0;
char *casListFile = CAS_LIST_FILE;
extern int g_isTerminating;
int logproblem(char *text,...);
AMS *findAlarmInList(CAMS *cams,UINT32 almCode,char *almTag);
extern ADSType almErrTblADS[];
ADS *getErrTblADS(short cc);
ADS *getAlarmDefinition(CasTblEntry *caslst,UINT32 almCode);

#define APCEQM_EXPORT "CAS.ARCHIVE"
#define CASEQM_EXPORT "CAS"
#define CASEQM_NUM_SYSTEMS 100

int casdebug = FALSE;
int g_isTerminating = FALSE;
int g_facilityNotRunning = FALSE;
int needsRestart = FALSE;
void term_handler(int signum)
{                  
  g_isTerminating = TRUE;
  feclog("caught SIGTERM");          
}
char thisFecName[FEC_NAME_SIZE] = "";
char thisContext[CONTEXT_NAME_SIZE] = "";
char *thisSubSystem = "SER";
char thisFecDescription[FEC_DESC_SIZE] = "";
char thisLocation[FEC_LOCATION_SIZE] = "";
UINT16 thisPortOffset = 0;
time_t gCasInitializedTs = 0;
void PreSystemInit(void)
{
  signal(SIGTERM,term_handler);
  /* Some System Parameters and their default values    */
  /* are shown below.  You may change them as necessary */
  MaxNumContracts = 1000;
  MaxNumClients = 100;
  MaxNumConnections = 1000;
  FeclogDepth = 1000;
  StartupDebug = TRUE;
  RequireAcknowledgments = TRUE;
  MaxRPCTransportSize = 0x100000;
  gRetardSingleContractRemoval = FALSE;
  gIgnoreAddressCacheOnUnknownAddress = TRUE;
  SystemAssignBufferSpace(0x4000,0x2000);
  isCentralAlarmServer(TRUE);
  SetRunServerCycleInSeparateThread(FALSE);
  SetAMSCapacity(5000);

  if (strlen(thisContext) > 0)
  {
    int p;
    if (strlen(thisFecName) == 0)
    {
      strncpy(thisFecName,thisContext,10); thisFecName[10] = 0;
      strncat(thisFecName,"CASFEC",6);
    }
    if (thisPortOffset == 0)
    {
      p = GetPortOffset(thisFecName);
      if (p >= 0) thisPortOffset = p;
    }
    sprintf(thisFecDescription,"%.32s Central Alarm Server",thisContext);
    RegisterFecInformation(thisFecName,thisSubSystem,thisContext,thisFecDescription,thisLocation,"none","administrator",thisPortOffset);
  }
}
void PostSystemInit(void)
{
  SetSuppressHeartbeatNotification(FALSE);
  archiveListFile = CAS_LIST_FILE; /* set this to the cas list file */
  RegisterEquipmentModule(APCEQM_EXPORT,APCEQM_TAG,CASEQM_NUM_SYSTEMS,apceqm,apcinit,apctask,100,NULL);
 
  RegisterEquipmentModule(CASEQM_EXPORT,CASEQM_TAG,CASEQM_NUM_SYSTEMS,caseqm,caseqm_ini,caseqm_bkg,CASEQM_RATE,caseqm_exi);
}
void RestartServer(void)
{
  static int restartCountdown = 5;
  if (restartCountdown-- > 0) return;
  FlushContractTable();
  SetServerCycleIdle(TRUE);
  FreeEQPModules();
  ServerExitCondition = 0;
  PostSystemInit();
  SetServerCycleIdle(FALSE);
  needsRestart = FALSE;
  restartCountdown = 5;
  gCasInitializedTs = 0;
}
int isDuplicatedAlarm(AMS *ams, CAMS *pcams);
void AppendGeneratedAlarm(int i,int almsys,int almsev,int almdsc,UINT32 starttime);
void AcquireAlarmInfo(int i);
int GetARBIndex(char *srvName) { return -1; } /* cover function */
int getAlarmSystemNumberFromTag(char *tag);
char *getAlarmSystemTagFromNumber(int nr);

typedef struct
{
  char srv[EXPORT_NAME_SIZE];
  char dev[DEVICE_NAME_SIZE];
  char reg[TAG_NAME_SIZE];
} RegionDbItem;

RegionDbItem *regionDbItems = NULL;
int nRegionDbItems = 0;
#define REGIONSDB_FILE "regionsdb.csv"

typedef struct
{
  char reg[TAG_NAME_SIZE];
  int  num;
} RegionDbXRef;

NAME16I *regionsXRef = NULL;
int nRegionsXRef = 0;
#define REGIONS_FILE "regions.csv"

CSVCOLUMN csvRegionDbColumns[] =
{
  { "SERVER","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(RegionDbItem,srv),NULL,0 },
  { "DEVICE","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(RegionDbItem,dev),NULL,0 },
  { "REGION","",0,CF_TEXT,TAG_NAME_SIZE,(BYTE *)OFFSETIN(RegionDbItem,reg),NULL,0 },
};
CSVDB csvRegionDb =
{
  csvRegionDbColumns,
  sizeof(csvRegionDbColumns)/sizeof(csvRegionDbColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(RegionDbItem),
  0
};

CSVCOLUMN csvRegionsXRefColumns[] =
{
  { "REGION","",0,CF_TEXT,TAG_NAME_SIZE,(BYTE *)OFFSETIN(NAME16I,name),NULL,0 },
  { "NUMBER","",0,CF_INT32,1,(BYTE *)OFFSETIN(NAME16I,ival),NULL,0 },
};
CSVDB csvRegions =
{
  csvRegionsXRefColumns,
  sizeof(csvRegionsXRefColumns)/sizeof(csvRegionsXRefColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME16I),
  0
};

#define DBTAG "CASRegDb"
int initRegionsDb(void)
{
  int cc;

  if ((cc=csvReadFile("",REGIONSDB_FILE,&csvRegionDb,(void **)&regionDbItems)) == 0)
  {
    nRegionDbItems = csvRegionDb.siz;
  }
  else
  {
    feclog("could not read regions database : %.32s",cc2str(cc));
  }

  if ((cc=csvReadFile("",REGIONS_FILE,&csvRegions,(void **)&regionsXRef)) == 0)
  {
    nRegionsXRef = csvRegions.siz;
  }
  else
  {
    feclog("could not read regions cross-references : %.32s",cc2str(cc));
  }

  if ((cc=addFieldToStruct(DBTAG,(int)OFFSETIN(RegionDbItem,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBTAG,(int)OFFSETIN(RegionDbItem,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBTAG,(int)OFFSETIN(RegionDbItem,reg),TAG_NAME_SIZE,CF_TEXT,"region")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(DBTAG,sizeof(RegionDbItem),100)) != 0) ccerr(cc);

err:
  return cc;
}
int getDbRegionCode(char *reg)
{
  int i;
  for (i=0; i<nRegionsXRef; i++)
  {
    if (!strnicmp(regionsXRef[i].name,reg,16)) return regionsXRef[i].ival;
  }
  return DEV_REGION_UNKNOWN;
}
int getRegionRulesForServer(char *srv,DeviceRegionItem *rules,int *size)
{
  int i, n=0;

  if (size == NULL) return argument_list_error;
  for (i=0; i<nRegionDbItems; i++)
  {
    if (wc_stricmp(regionDbItems[i].srv,srv)) continue;
    if (rules != NULL && n < *size)
    {
      strncpy(rules[n].dev,regionDbItems[i].dev,DEVICE_NAME_SIZE);
      strncpy(rules[n].reg,regionDbItems[i].reg,TAG_NAME_SIZE);
      rules[n].code = getDbRegionCode(regionDbItems[i].reg);
    }
    n++;
  }
  *size = n;
  return 0;
}
int getPrimaryRegionForServer(char *srv)
{
  int i, n=0;

  for (i=0; i<nRegionDbItems; i++)
  {
    if (wc_stricmp(regionDbItems[i].srv,srv)) continue;
    return getDbRegionCode(regionDbItems[i].reg);
  }
  return DEV_REGION_UNKNOWN;
}

#define SYSALIAS_FILE "system-alias.csv"
int nsysalias = 0;
typedef struct
{
  int system;
  int alias;
} SystemAlias;
SystemAlias *sysalias = NULL;

int getSystemAliases(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int sys_col=-1,alias_col=-1;
  char str[256],hdr[256],scratch[32];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (sysalias != NULL)
  {
    free(sysalias);
    nsysalias = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((sysalias=(SystemAlias *)calloc(n,sizeof(SystemAlias))) == NULL) ccerr(out_of_server_memory);
  nsysalias = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nsysalias;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((sys_col=findcol(hdr,"ALARM_SYSTEM")) < 0) ccerr(no_such_line);
      if ((alias_col=findcol(hdr,"SYSTEM_ALIAS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(sys_col,str),32);
    sysalias[i].system = atoi(scratch);
    strncpy(scratch,colptr(alias_col,str),32);
    sysalias[i].alias = atoi(scratch);
    if (sysalias[i].system == 0 || sysalias[i].alias == 0) continue;
    i++;
  }
  nsysalias = i;
err:
  if (cc != no_such_file) feclog("read system alias file: %.32s",cc2str(cc));
  if (fp != NULL) fclose(fp);
  return cc;
}
int dumpSysAlias(int i1,int i2,int i3,int i4)
{
  int i;
  if (nsysalias == 0)
  {
    dbglog("no alarm system aliases registered");
  }
  else
  {
    for (i=0; i<nsysalias; i++)
    {
      dbglog("alarms with system %d are routed to %d (%.32s)",
        sysalias[i].system,sysalias[i].alias,getAlarmSystemTagFromNumber(sysalias[i].alias));
    }
  }
  return 0;
}
int dumpCasTbl(int i1,int i2,int i3,int i4)
{
  int i;

  if (casTblSiz == 0)
  {
    dbglog("no alarm table entries !");
  }
  else
  {
    for (i=0; i<casTblSiz; i++) 
    {
      ttyoutput("/%.32s/%.32s : link %d, online %s, pending %s, has initial alarm set %s",
        casTbl[i].devContext,casTbl[i].devServer,casTbl[i].linkId,
        (casTbl[i].offline ? "FALSE" : "TRUE"),
        (casTbl[i].pending ? "TRUE" : "FALSE"),
        (casTbl[i].hasAcquiredInitialAlarmSet ? "TRUE" : "FALSE"));
    }
  }
  return 0;
}
#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  if (nadmins == 0) return TRUE;
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,USERNAME_SIZE); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; fgets(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),USERNAME_SIZE);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
int RecordEvent(ulong eventNumber,ulong eventCode,char *devName)
{
  return 0;
}
int SetGlobalEvent(ulong eventCode,char *name)
{
  return 0;
}
void registerCASStructs(void)
{
  static int done=FALSE;
  if (done) return;
  done = TRUE;
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,devContext),CONTEXT_NAME_SHORTSIZE,CF_TEXT,"devContext")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,devServer),EXPORT_NAME_SHORTSIZE,CF_TEXT,"devServer")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,subSystem),SUBSYSTEM_NAME_SIZE,CF_TEXT,"subSystem")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,arcExt),XTN_NAME_SIZE,CF_TEXT,"arcExt")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,SeverityLevel),1,CF_LONG,"SeverityLevel")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,ArchiveLevel),1,CF_LONG,"ArchiveLevel")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,Retention),1,CF_LONG,"Retention")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,AlarmLevel),1,CF_LONG,"AlarmLevel")) quit(1);
  if (addFieldToStruct("CASQS",OFFSETIN(CASQS,AlarmSystem),1,CF_LONG,"AlarmSystem")) quit(1);
  if (sealTaggedStruct("CASQS",sizeof(CASQS),1000)) quit(1);

  if (addFieldToStruct("CQS",OFFSETIN(CQS,devContext),CONTEXT_NAME_SIZE,CF_TEXT,"devContext")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,devServer),EXPORT_NAME_SIZE,CF_TEXT,"devServer")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,subSystem),SUBSYSTEM_NAME_SIZE,CF_TEXT,"subSystem")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,arcExt),XTN_NAME_SIZE,CF_TEXT,"arcExt")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,SeverityLevel),1,CF_LONG,"SeverityLevel")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,ArchiveLevel),1,CF_LONG,"ArchiveLevel")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,Retention),1,CF_LONG,"Retention")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,AlarmLevel),1,CF_LONG,"AlarmLevel")) quit(1);
  if (addFieldToStruct("CQS",OFFSETIN(CQS,AlarmSystem),1,CF_LONG,"AlarmSystem")) quit(1);
  if (sealTaggedStruct("CQS",sizeof(CQS),1000)) quit(1);

  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,server),EXPORT_NAME_SHORTSIZE,CF_TEXT,"server")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,device),DEVICE_NAME_SHORTSIZE,CF_TEXT,"device")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,alarmText),40,CF_TEXT,"alarmText")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,alarmCode),1,CF_LONG,"alarmCode")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,timestamp),1,CF_LONG,"timestamp")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,alarmMask),1,CF_LONG,"alarmMask")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,alarmSystem),1,CF_LONG,"alarmSystem")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,alarmData),8,CF_BYTE,"alarmData")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,alarmDataFormat),1,CF_BYTE,"alarmDataFormat")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,alarmDataArraySize),1,CF_BYTE,"alarmDataArraySize")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,severity),1,CF_BYTE,"severity")) quit(1);
  if (addFieldToStruct("XAMS",OFFSETIN(XAMS5,descriptor),1,CF_BYTE,"descriptor")) quit(1);
  if (sealTaggedStruct("XAMS",sizeof(XAMS5),1000)) quit(1);

  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,server),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,device),DEVICE_NAME_SIZE,CF_TEXT,"device")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,alarmText),ALARM_TEXT_SIZE,CF_TEXT,"alarmText")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,alarmCode),1,CF_LONG,"alarmCode")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,timestamp),1,CF_LONG,"time")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,timestampUSec),1,CF_LONG,"timeUsec")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,starttime),1,CF_LONG,"start")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,starttimeUSec),1,CF_LONG,"startUSec")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,alarmMask),1,CF_LONG,"alarmMask")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,alarmSystem),1,CF_LONG,"alarmSystem")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,alarmData),ALARM_DATA_SIZE,CF_BYTE,"alarmData")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,alarmDataFormat),1,CF_BYTE,"alarmDataFormat")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,alarmDataArraySize),1,CF_BYTE,"alarmDataArraySize")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,severity),1,CF_BYTE,"severity")) quit(1);
  if (addFieldToStruct("XAMSr4",OFFSETIN(XAMS,descriptor),1,CF_BYTE,"descriptor")) quit(1);
  if (sealTaggedStruct("XAMSr4",sizeof(XAMS),1000)) quit(1);

  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,email),128,CF_TEXT,"email")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,ctx),CONTEXT_NAME_SIZE,CF_TEXT,"context")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,srv),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,dev),DEVICE_NAME_SIZE,CF_TEXT,"device")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,prp),PROPERTY_NAME_SIZE,CF_TEXT,"property")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,actionSignalString),128,CF_TEXT,"actionData")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,actionSignalFormat),1,CF_INT32,"actionFormat")) quit(1);
  if (addFieldToStruct(CASACTIONITEM_TAG,OFFSETIN(CasActionItem,alarmCode),1,CF_INT32,"alarmCode")) quit(1);
  if (sealTaggedStruct(CASACTIONITEM_TAG,sizeof(CasActionItem),100)) quit(1);

  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,devContext),CONTEXT_NAME_SIZE,CF_TEXT,"context")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,devServer),EXPORT_NAME_SIZE,CF_TEXT,"server")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,subSystem),SUBSYSTEM_NAME_SIZE,CF_TEXT,"subsys")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,arcExt),XTN_NAME_SIZE,CF_TEXT,"ext")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,SeverityLevel),1,CF_INT32,"severityLevel")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,ArchiveLevel),1,CF_INT32,"archiveLevel")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,Retention),1,CF_INT32,"retention")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,AlarmLevel),1,CF_INT32,"alarmLevel")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,AlarmSystem),1,CF_INT32,"alarmSystem")) quit(1);
  if (addFieldToStruct(CASDBITEM_TAG,OFFSETIN(CasDbItem,offline),1,CF_INT32,"offline")) quit(1);
  if (sealTaggedStruct(CASDBITEM_TAG,sizeof(CasDbItem),500)) quit(1);

  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,number),1,CF_INT32,"number")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,column),1,CF_INT32,"column")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,row),1,CF_INT32,"row")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,almsystem),ALARM_TAG_SIZE,CF_TEXT,"tag")) quit(1);
  if (addFieldToStruct(CASALMSYSITEM_TAG,OFFSETIN(CasAlmSysTblEntry,subsystem),SUBSYSTEM_NAME_SIZE,CF_TEXT,"subsys")) quit(1);
  if (sealTaggedStruct(CASALMSYSITEM_TAG,sizeof(CasAlmSysTblEntry),100)) quit(1);
}
int makecasname(char *fname,char *src,char *ext,time_t t)
{
  struct tm *ts;
  int yr, mo;
  FILE *fp = NULL;
  time_t ht = (t / 3600) * 3600;

  ts = localtime(&t);
  yr = ts->tm_year + 1900;
  mo = ts->tm_mon + 1;

  sprintf(fname,"%s/%d/%02d/%s/%lx.%s",src,yr,mo,ext,ht,ext);
  if ((fp=fopen(fname,"r")) == NULL) 
  { /* does not yet exist => don't decorate the name */
    return 0;
  }
  fclose(fp);
  strcat(fname,"-add");
  return -1;
}
#define EVN_STORE_HDR_SIZE 256
void AppendAMS(char *filename)
{
  int n,num=0,cc=0,ival;
  short sval;
  float fval;
  FILE *fp = NULL, *afp = NULL, *nfp = NULL;
  char *c,vbuf[16],fname[128],nfname[128],buf[512],rbuf[512];

  strncpy(fname,filename,128);
  if ((c=strstr(fname,"-add")) == NULL) 
  { /* not the right file name ! */
    return;
  }
  *c = 0; /* strip off -> now 'fname' is the canonical file name ! */
  if ((fp=fopen(fname,"r")) == NULL) ccerr(file_error);
  if ((afp=fopen(filename,"r")) == NULL) ccerr(file_error);
  strncpy(nfname,fname,128); strcat(nfname,"-new");
  if ((nfp=fopen(nfname,"w")) == NULL) ccerr(file_error);
  if (fread(vbuf,VERSION_LENGTH,1,fp) != 1) ccerr(file_error);
  if (fread(buf,EVN_STORE_HDR_SIZE+4,1,fp) == 1)
  { /* read how many have been stored so far */
    memcpy(rbuf,buf,EVN_STORE_HDR_SIZE);
    memcpy(&n,&buf[EVN_STORE_HDR_SIZE],4);
    num += n;
  }
  if (fread(buf,EVN_STORE_HDR_SIZE+4,1,afp) == 1)
  { /* read how many should be appended and add them */
    memcpy(&n,&buf[EVN_STORE_HDR_SIZE],4);
    num += n;
  }
  /* write back out */
  fwrite(vbuf,VERSION_LENGTH,1,nfp);
  fwrite(rbuf,EVN_STORE_HDR_SIZE,1,nfp);
  memcpy(buf,&num,4);
  fwrite(buf,4,1,nfp);
  /* now read the alarms already stored (first update the header) */
  if (fread(buf,EVN_STORE_HDR_SIZE,1,fp) != 1) ccerr(file_error);
  fwrite(buf,EVN_STORE_HDR_SIZE-16,1,nfp);
  ival = num*AMS_SIZE; memcpy(&buf[0],&ival,4);
  sval = CF_STRUCT; memcpy(&buf[4],&sval,2);
  sval = 0; memcpy(&buf[6],&sval,2);
  fval = 1.0; memcpy(&buf[8],&fval,4);
  fval = 0.0; memcpy(&buf[12],&fval,4);
  fwrite(buf,16,1,nfp);
  /* now the alarms (original set) */
  while (fread(buf,AMS_SIZE,1,fp) == 1)
  {
    fwrite(buf,AMS_SIZE,1,nfp);
  }
  /* skip over the new alarms header */
  if (fread(buf,EVN_STORE_HDR_SIZE,1,afp) != 1) ccerr(file_error);
  /* now append the new alarms to the file */
  while (fread(buf,AMS_SIZE,1,afp) == 1)
  {
    fwrite(buf,AMS_SIZE,1,nfp);
  }
  /* close up shop */
  fclose(fp); fp = NULL; remove(fname);
  fclose(afp); afp = NULL; remove(filename);
  fclose(nfp); nfp = NULL;
  /* move the new file into the canonical file */
  rename(nfname,fname);
  return;
err:
  feclog("AppendAMS: %s -> %.32s",filename, cc2str(cc));
  if (fp != NULL) fclose(fp);
  if (afp != NULL) fclose(afp);
  if (nfp != NULL) fclose(nfp);
}
void ArchiveAMS(int nalms, AMS *ams,char *devsvr, char *ext,UINT32 stoptime)
{
  char fn[512];
  FILE *fp=NULL;
  int ival,cc=0,found=0;
  short sval;
  float fval;
  char s[512];

  if (g_isTerminating || tineDebug) feclog("archive alarms for %.32s",devsvr);

  if (stoptime < 100) 
  { /* invalid timestamp !! */
    feclog("attempted to archive %.32s with invalid timestamp %d",devsvr,stoptime);
    return;
  }
  /* use EVENT archive structure for nalms and then the AMS list */
  found = makecasname(fn,AlarmArchivePath,ext,stoptime);
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(no_such_file);
  if (!found) 
  { /* new file : supply the version tag */
    memset(s,0,VERSION_LENGTH); strncpy(s,"4.00",VERSION_LENGTH);
    if (fwrite(s,VERSION_LENGTH,1,fp) != 1) ccerr(file_error);
  }
  /* put number of alarms in archive: */
  memset(s,0,CONTEXT_NAME_SIZE); strncpy(s,gDeviceContext,CONTEXT_NAME_SIZE);
  if (fwrite(s,CONTEXT_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,EXPORT_NAME_SIZE); strncpy(s,casExportName,EXPORT_NAME_SIZE);
  if (fwrite(s,EXPORT_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,PROPERTY_NAME_SIZE); strncpy(s,"NUMALARMS",PROPERTY_NAME_SIZE);
  if (fwrite(s,PROPERTY_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,DEVICE_NAME_SIZE); 
  if (fwrite(s,DEVICE_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,TAG_NAME_SIZE); 
  if (fwrite(s,TAG_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,STATUS_SHORTSIZE); strncpy(s,GetCompletionStatus(0),STATUS_SHORTSIZE);
  if (fwrite(s,STATUS_SHORTSIZE,1,fp) != 1) ccerr(file_error);
  ival = 1; memcpy(&s[0],&ival,4);
  sval = CF_LONG; memcpy(&s[4],&sval,2);
  sval = 0; memcpy(&s[6],&sval,2);
  if (fwrite(s,8,1,fp) != 1) ccerr(file_error);
  fval = 1.0; memcpy(&s[0],&fval,4);
  fval = 0.0; memcpy(&s[4],&fval,4);
  if (fwrite(s,8,1,fp) != 1) ccerr(file_error);
  ival = nalms; memcpy(&s[0],&ival,4);
  if (fwrite(s,4,1,fp) != 1) ccerr(file_error);

  /* now put the alarms in archive */
  memset(s,0,CONTEXT_NAME_SIZE); strncpy(s,gDeviceContext,CONTEXT_NAME_SIZE);
  if (fwrite(s,CONTEXT_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,EXPORT_NAME_SIZE); strncpy(s,casExportName,EXPORT_NAME_SIZE);
  if (fwrite(s,EXPORT_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,PROPERTY_NAME_SIZE); strncpy(s,"SYSALARMS",PROPERTY_NAME_SIZE);
  if (fwrite(s,PROPERTY_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,DEVICE_NAME_SIZE); 
  if (fwrite(s,DEVICE_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,TAG_NAME_SIZE); strncpy(s,"AMSr4",TAG_NAME_SIZE);
  if (fwrite(s,TAG_NAME_SIZE,1,fp) != 1) ccerr(file_error);
  memset(s,0,STATUS_SHORTSIZE); strncpy(s,GetCompletionStatus(0),STATUS_SHORTSIZE);
  if (fwrite(s,STATUS_SHORTSIZE,1,fp) != 1) ccerr(file_error);
  ival = nalms*AMS_SIZE; memcpy(&s[0],&ival,4);
  sval = CF_STRUCT; memcpy(&s[4],&sval,2);
  sval = 0; memcpy(&s[6],&sval,2);
  if (fwrite(s,8,1,fp) != 1) ccerr(file_error);
  fval = 1.0; memcpy(&s[0],&fval,4);
  fval = 0.0; memcpy(&s[4],&fval,4);
  if (fwrite(s,8,1,fp) != 1) ccerr(file_error);
  if (fwrite(ams,AMS_SIZE,nalms,fp) != (UINT32)nalms) ccerr(file_error);

err:
  if (fp) fclose(fp);
  if (cc != 0) 
  {
    feclog("Alarm Archive %s : %s",fn,erlst[cc]);
    found = 0;
  }
  if (found) AppendAMS(fn);
  return;
}
int RegisterCaseqmProperties(void)
{
  int cc = 0;
  DTYPE dout, din;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PRP_SYSALARMS_SIZE;
  strcpy(dout.dTag,"AMSr4");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS",&dout,NULL,CA_READ,AT_SCALAR,0,"current system alarms",PRP_SYSALARMS,NULL);
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.ALL",&dout,NULL,CA_READ,AT_SCALAR,0,"current system alarms",PRP_SYSALARMS_ALL,NULL);
  strcpy(dout.dTag,"AMS");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS",&dout,NULL,CA_READ,AT_SCALAR,0,"current system alarms",PRP_SYSALARMS,NULL);
  strcpy(dout.dTag,"XAMSr4");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS",&dout,NULL,CA_READ,AT_SCALAR,0,"current system alarms",PRP_SYSALARMS,NULL);
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.ALL",&dout,NULL,CA_READ,AT_SCALAR,0,"current system alarms",PRP_SYSALARMS_ALL,NULL);
  strcpy(dout.dTag,"XAMS");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS",&dout,NULL,CA_READ,AT_SCALAR,0,"current system alarms",PRP_SYSALARMS,NULL);

  if ((cc=RegisterPropertyEx(CASEQM_TAG,"NUMALARMS",PRP_NUMALARMS_SIZE,CF_LONG,CA_READ,"Nr of current alarms",PRP_NUMALARMS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"NUMALARMS.ALL",PRP_NUMALARMS_SIZE,CF_LONG,CA_READ,"Nr of current alarms (+ hidden)",PRP_NUMALARMS_ALL)) < 0) goto err;

  if ((cc=RegisterPropertyEx(CASEQM_TAG,"SUBSYSTEMS",20,CF_NAME16,CA_READ,"available sub systems",PRP_SUBSYSTEMS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"SERVERS",100,CF_NAME16,CA_READ,"available device servers",PRP_SERVERS)) < 0) goto err;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"CASQS");
  RegisterPropertyInformation(CASEQM_TAG,"SERVERINFO",&dout,NULL,CA_READ,AT_SCALAR,0,"alarm configuration information",PRP_SERVERINFO,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"CQS");
  RegisterPropertyInformation(CASEQM_TAG,"SERVERINFO",&dout,NULL,CA_READ,AT_SCALAR,0,"alarm configuration information",PRP_SERVERINFO,NULL);

  if ((cc=RegisterPropertyEx(CASEQM_TAG,"ACKTIME",PRP_ACKTIME_SIZE,CF_LONG,CA_READ|CA_WRITE,"[0:1E11 sec]read/write alarm view threshold",PRP_ACKTIME)) < 0) goto err;

  if ((cc=RegisterPropertyEx(CASEQM_TAG,"ALARMSYSTEMS",100,CF_NAME16I,CA_READ,"registered alarm systems + codes",PRP_ALARMSYSTEMS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"ALARMSYSTEMS",100,CF_NAME16II,CA_READ,"registered alarm systems + codes + columns",PRP_ALARMSYSTEMS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"ALARMSYSTEMS.ALL",100,CF_NAME16I,CA_READ,"registered (and hidden) alarm systems + codes",PRP_ALARMSYSTEMS_ALL)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"ALARMSYSTEMS.ALL",100,CF_NAME16II,CA_READ,"registered (and hidden) alarm systems + codes + columns",PRP_ALARMSYSTEMS_ALL)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"REMOVEALARMS",0,CF_NULL,CA_WRITE,"remove alarms from table",PRP_REMOVEALARMS)) < 0) goto err;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PRP_SYSALARMS_SIZE;
  strcpy(dout.dTag,"AMS");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.RAW",&dout,NULL,CA_READ,AT_SCALAR,0,"all current system alarms",PRP_RAWSYSALARMS,NULL);
  strcpy(dout.dTag,"AMSr4");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.RAW",&dout,NULL,CA_READ,AT_SCALAR,0,"all current system alarms",PRP_RAWSYSALARMS,NULL);
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.RAW.ALL",&dout,NULL,CA_READ,AT_SCALAR,0,"all current system alarms",PRP_RAWSYSALARMS_ALL,NULL);
  strcpy(dout.dTag,"XAMSr4");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.RAW",&dout,NULL,CA_READ,AT_SCALAR,0,"all current system alarms",PRP_RAWSYSALARMS,NULL);
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.RAW.ALL",&dout,NULL,CA_READ,AT_SCALAR,0,"all current system alarms",PRP_RAWSYSALARMS_ALL,NULL);
  strcpy(dout.dTag,"XAMS");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALARMS.RAW",&dout,NULL,CA_READ,AT_SCALAR,0,"all current system alarms",PRP_RAWSYSALARMS,NULL);

  if ((cc=RegisterPropertyEx(CASEQM_TAG,"NUMALARMS.RAW",1,CF_LONG,CA_READ,"Nr of all current alarms",PRP_RAWNUMALARMS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"NUMALARMS.RAW.ALL",1,CF_LONG,CA_READ,"Nr of all current alarms (+ hidden)",PRP_RAWNUMALARMS_ALL)) < 0) goto err;

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 25;
  strcpy(dout.dTag,"ADS");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALMDEFS",&dout,NULL,CA_READ,AT_SCALAR,0,"Alarm definitions",PRP_SYSALMDEFS,NULL);
  strcpy(dout.dTag,"ADSr4");
  RegisterPropertyInformation(CASEQM_TAG,"SYSALMDEFS",&dout,NULL,CA_READ,AT_SCALAR,0,"Alarm definitions",PRP_SYSALMDEFS,NULL);
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"NUMALMDEFS",1,CF_LONG,CA_READ,"Nr of registered alarm definitions",PRP_NUMALMDEFS)) < 0) goto err;


  if ((cc=RegisterPropertyEx(CASEQM_TAG,"AlmSystem4Server",10,CF_LONG,CA_READ,"alarm systems managed by server",PRP_SYSTEMLIST)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CASEQM_TAG,"Server4AlmSystem",10,CF_NAME32,CA_READ,"servers which can manage given system",PRP_SERVERLIST)) < 0) goto err;

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 100;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"ServerIsOnline",&dout,NULL,CA_READ,AT_CHANNEL,0,"server on line flag",PRP_ISONLINE,NULL);

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 100;
  dout.dTag[0] = 0;
  din.dFormat = CF_LONG;
  din.dArrayLength = 100;
  din.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"DisableAlarm",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set disabled alarm(s)",PRP_DISABLE,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,"AMSr4");
  RegisterPropertyInformation(CASEQM_TAG,"DisableAlarm",&dout,NULL,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set disabled alarm(s)",PRP_DISABLE,NULL);

  dout.dFormat = CF_TEXT;
  dout.dArrayLength = 512;
  dout.dTag[0] = 0;
  din.dFormat = CF_TEXT;
  din.dArrayLength = 512;
  din.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"DisabledAlarmComment",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set disabled alarm comment",PRP_DISABLECMT,NULL);

  dout.dFormat = CF_BOOLEAN;
  dout.dArrayLength = 1;
  dout.dTag[0] = 0;
  din.dFormat = CF_LONG;
  din.dArrayLength = 1;
  din.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"DisabledAlarmIsActive",&dout,&din,CA_READ,AT_UNKNOWN,0,"is disabled alarm active",PRP_DISABLE_ISACTIVE,NULL);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"BaseTime",&dout,NULL,CA_READ,AT_SCALAR,0,"timestamp of base configuration",PRP_BASETIME,NULL);

  dout.dFormat = CF_NAME64I;
  dout.dArrayLength = 100;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"ActiveDisabledAlarms",&dout,NULL,CA_READ,AT_UNKNOWN,0,"list of disabled alarms which are active",PRP_DISABLE_ACTIVE,NULL);

  din.dFormat = CF_INT32;
  din.dArrayLength = 100;
  din.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"EnableAlarm",NULL,&din,CA_WRITE,AT_UNKNOWN,0,"ensable alarm(s)",PRP_ENABLE,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 500;
  strcpy(dout.dTag,CASDBITEM_TAG);
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 500;
  strcpy(din.dTag,CASDBITEM_TAG);
  RegisterPropertyInformation(CASEQM_TAG,"SERVER.DB",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set servers database",PRP_SERVERDB,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,CASALMSYSITEM_TAG);
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 100;
  strcpy(din.dTag,CASALMSYSITEM_TAG);
  RegisterPropertyInformation(CASEQM_TAG,"ALMSYSTEM.DB",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set alarm systems database",PRP_ALMSYSDB,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,CASACTIONITEM_TAG);
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 100;
  strcpy(din.dTag,CASACTIONITEM_TAG);
  RegisterPropertyInformation(CASEQM_TAG,"ACTION.DB",&dout,&din,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set alarm action database",PRP_ACTIONDB,NULL);

  RegisterPropertyInformation(CASEQM_TAG,"RESTART",NULL,NULL,CA_WRITE,AT_UNKNOWN,0,"restart the server's equipment modules",PRP_RESTART,NULL);

  dout.dFormat = CF_INT32;
  dout.dArrayLength = 1;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"NREGIONRULES",&dout,NULL,CA_READ,AT_SCALAR,0,"[0:100]number of region rules",PRP_NREGIONRULES,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,"RgnRule");
  RegisterPropertyInformation(CASEQM_TAG,"REGIONRULES",&dout,NULL,CA_READ,AT_SINGLE,0,"region rules for target server",PRP_REGIONRULES,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 100;
  strcpy(dout.dTag,DBTAG);
  RegisterPropertyInformation(CASEQM_TAG,"REGIONRULES.DB",&dout,&dout,CA_READ|CA_WRITE,AT_SINGLE,0,"region rules database",PRP_REGIONSDB,NULL);

  dout.dFormat = CF_NAME16I;
  dout.dArrayLength = 100;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(CASEQM_TAG,"REGIONXREF.DB",&dout,&dout,CA_READ|CA_WRITE,AT_SINGLE,0,"region codes cross-reference database",PRP_REGIONSXREF,NULL);

err:
  if (cc < 0)
  {
    printf("Can't register property : %s\n>",erlst[-cc]);
  }
  return cc;
}
int writeCASDisabledComment(int i,char *dev,int almcode,char *cmt,int len)
{
  char fn[256];
  FILE *fp = NULL;
  int cc=0;
  if (dev == NULL || strlen(dev) == 0) dev = "ALL";
  sprintf(fn,"%.80s/AlarmInfo/%.16s/%.32s-%.64s-%d.cmt",
    AlarmArchivePath,casTbl[i].arcExt,casTbl[i].devServer,dev,almcode);

  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  if (fwrite(cmt,len,1,fp) != 1) ccerr(file_error);

err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int readCASDisabledComment(int i,char *dev,int almcode,char *cmt,int *len)
{
  char fn[256];
  FILE *fp = NULL;
  int cc=0,rlen,blen;

  if (cmt == NULL || len == NULL) return invalid_parameter;
  if (dev == NULL || strlen(dev) == 0) dev = "ALL";
  sprintf(fn,"%.80s/AlarmInfo/%.16s/%.32s-%.64s-%d.cmt",
    AlarmArchivePath,casTbl[i].arcExt,casTbl[i].devServer,dev,almcode);

  if ((fp=fopen(fn,"r")) == NULL) 
  { /* not annotated */
    *len = 0;
    return 0;
  }
  blen = *len;
  if ((rlen=fread(cmt,1,blen,fp)) < blen) *len = rlen;
  if (fp != NULL) fclose(fp);
  return cc;
}

int writeCASDisabledList(int i)
{
  char fn[256];
  FILE *fp = NULL;
  int cc=0,n=0;
  CasDisableLstEntry *cd;

  sprintf(fn,"%.80s/AlarmInfo/%.16s/%.32s-disabled.csv",AlarmArchivePath,casTbl[i].arcExt,casTbl[i].devServer);
  if (casTbl[i].disabledList == NULL) 
  {
    unlink(fn);
    return 0;
  }
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"Alarms, Device, Time, Type, Caller\n");
  for (cd=casTbl[i].disabledList; cd != NULL; cd=cd->nxt)
  {
    fprintf(fp,"%d, %.64s, %.24s, %d, %.16s\n",
      cd->alarmCode,cd->device,ctime(&cd->timeDisabled),cd->disableType,cd->disabler);
    n++;
  }
  if (casTbl[i].sizDisabledList != n)
  {
    feclog("CAS correct size of disabled list from %d to %d",casTbl[i].sizDisabledList,n);
    casTbl[i].sizDisabledList = n;
  } 
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
CasDisableLstEntry *getCASDisabledList(int i)
{
  FILE *fp = NULL;
  int cc=0,done=0,n=0;
  int alm_col,dev_col,ts_col,usr_col,typ_col;
  char str[1024],hdr[256],scratch[128],fn[256];
  CasDisableLstEntry *cd,*cdList = NULL;

  sprintf(fn,"%.80s/AlarmInfo/%.16s/%.32s-disabled.csv",AlarmArchivePath,casTbl[i].arcExt,casTbl[i].devServer);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,1023,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((alm_col=findcol(hdr,"ALARMS")) < 0) ccerr(no_such_column);
      dev_col = findcol(hdr,"DEVICE");
      ts_col = findcol(hdr,"TIME");
      usr_col = findcol(hdr,"CALLER");
      typ_col = findcol(hdr,"TYPE");
      done = 1;                              /* read header info */
      continue;
    }
    if ((cd=(CasDisableLstEntry *)calloc(1,sizeof(CasDisableLstEntry))) == NULL) ccerr(out_of_server_memory);
    strncpy(scratch,colptr(alm_col,str),16);
    cd->alarmCode = atoi(scratch);
    if (typ_col >= 0)
    {
      strncpy(scratch,colptr(typ_col,str),16); scratch[16] = 0;
      cd->disableType = atoi(scratch);
    }
    if (dev_col >= 0) strncpy(cd->device,colptr(dev_col,str),DEVICE_NAME_SIZE);
    if (ts_col >= 0) 
    {
      strncpy(scratch,colptr(ts_col,str),24); scratch[24] = 0;
      cd->timeDisabled = parseCtimeStr(scratch);
    }
    if (usr_col >= 0) strncpy(cd->disabler,colptr(usr_col,str),USERNAME_SIZE);
    if (cd->timeDisabled == 0) cd->timeDisabled = time(NULL); /* adiabatically introduce this new feature */
    cd->nxt = cdList;
    cdList = cd;
    n++;
  }
err:
  casTbl[i].sizDisabledList = n;
  if (fp) fclose(fp);
  if (cc == 0) feclog("CAS Disabled file %s : found (%d disabled)",fn,n);
  return cdList;
}
int parseActionInput(DTYPE *din,char *str,int fmt)
{
  int i, nlen, dlen = 0, dsiz = 0, dfmt = fmt, cc = 0;
  char s[136],*tokens[16];
  char *c,*cp=NULL,*delim = " ,";

  strncpy(s,str,128); s[128] = 0;
  switch (fmt)
  {
    case CF_TEXT:
      dlen = strlen(s);
      break;
    case CF_INT16:
    case CF_INT32:
    case CF_FLOAT:
    case CF_DOUBLE:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      for (i=0, c=strtok_r(s,delim,&cp); i<16 && c != NULL; i++, c=strtok_r(NULL,delim,&cp))
      {
        tokens[i] = c;
      }
      dlen = i;
      break;
    default:
      dfmt = CF_NULL;
      break;
  }
  dsiz = dlen * fmtsizeof(dfmt);
  if ((din->data.vptr=calloc(1,dsiz)) == NULL) 
  {
    dlen = 0; dfmt = CF_NULL;
    ccerr(out_of_local_memory);
  }
  switch (fmt)
  {
    case CF_TEXT:
      strncpy(din->data.cptr,str,dlen);
      break;
    case CF_INT16:
      for (i=0; i<dlen; i++) din->data.sptr[i] = (SINT16)atoi(tokens[i]);
      break;
    case CF_INT32:
      for (i=0; i<dlen; i++) din->data.lptr[i] = (SINT32)atoi(tokens[i]);
      break;
    case CF_FLOAT:
      for (i=0; i<dlen; i++) din->data.fptr[i] = (float)strtod(tokens[i],NULL);
      break;
    case CF_DOUBLE:
      for (i=0; i<dlen; i++) din->data.dptr[i] = strtod(tokens[i],NULL);
      break;
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      nlen = fmtsizeof(fmt);
      for (i=0; i<dlen; i++) 
      {
        strncpy(din->data.cptr + nlen*i,tokens[i],nlen);
      }
      break;
  }
err:
  din->dFormat = dfmt;
  din->dArrayLength = dlen;
  return cc;
}
CasActLstEntry *getCASActionList(int i)
{
  FILE *fp = NULL;
  int cc=0,pcc,done=0;
  int con_col,srv_col,dev_col,prp_col,code_col,fmt_col,mto_col,str_col;
  char contxt[CONTEXT_NAME_SIZE+8],devsrv[EXPORT_NAME_SIZE+8],dev[DEVICE_NAME_SIZE+8];
  char str[1024],hdr[256],scratch[128],fn[256];
  CasActLstEntry *ca,*caList = NULL;

  sprintf(fn,"%.80s/AlarmInfo/%.16s/actions.csv",AlarmArchivePath,casTbl[i].arcExt);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,1023,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      mto_col = findcol(hdr,"MAILTO");
      con_col = findcol(hdr,"CONTEXT");
      srv_col=findcol(hdr,"SERVER");
      dev_col = findcol(hdr,"DEVICE");
      prp_col = findcol(hdr,"PROPERTY");
      code_col = findcol(hdr,"ALARMCODE");
      fmt_col = findcol(hdr,"ACTIONFORMAT"); 
      str_col = findcol(hdr,"ACTIONSTRING"); 
      done = 1;                              /* read header info */
      continue;
    }
    if ((ca=(CasActLstEntry *)calloc(1,sizeof(CasActLstEntry))) == NULL) ccerr(out_of_server_memory);
    ca->casTblIdx = i; /* cross reference back to the cas table */
    strncpy(scratch,colptr(code_col,str),16);
    ca->alarmCode = atoi(scratch);
    memset(devsrv,0,EXPORT_NAME_SIZE+8);
    if (srv_col >= 0) strncpy(devsrv,colptr(srv_col,str),EXPORT_NAME_SIZE);
    if (strlen(devsrv) > 0)
    { /* send command to a server */
      strncpy(contxt,gDeviceContext,CONTEXT_NAME_SIZE);
      if (con_col >= 0) strncpy(contxt,colptr(con_col,str),CONTEXT_NAME_SIZE); 
      contxt[CONTEXT_NAME_SIZE] = 0;
      sprintf(ca->devName,"/%s/",contxt);
      strncpy(devsrv,colptr(srv_col,str),EXPORT_NAME_SIZE); devsrv[EXPORT_NAME_SIZE] = 0;
      strcat(ca->devName,devsrv); strcat(ca->devName,"/");
      strncpy(dev,colptr(dev_col,str),DEVICE_NAME_SIZE); dev[DEVICE_NAME_SIZE] = 0;
      strcat(ca->devName,dev);
      strncpy(ca->devProperty,colptr(prp_col,str),PROPERTY_NAME_SIZE);
      pcc = 0;
      if (fmt_col >= 0 && str_col >= 0)
      {
        strncpy(scratch,colptr(fmt_col,str),16);
        ca->actionSignalFormat = ftoi(scratch);
        strncpy(ca->actionSignalString,colptr(str_col,str),128);
        pcc = parseActionInput(&ca->din,ca->actionSignalString,ca->actionSignalFormat);
      }
      ca->dout.dArrayLength = 0;
      ca->dout.dFormat = CF_NULL;
      ca->access = CA_WRITE|CA_READ;
      feclog("action: send command %.128s[%.64s] for alarm %d from server %.32s (parse input : <%d>)",
        ca->devName,ca->devProperty,ca->alarmCode,casTbl[i].devServer,pcc);
    }
    if (mto_col >= 0)
    { /* send email */
      strncpy(ca->email,colptr(mto_col,str),128);
      feclog("action: mailto %s for alarm %d from server %s",
        ca->email,ca->alarmCode,casTbl[i].devServer);
    }
    ca->nxt = caList;
    caList = ca;
  }
err:
  if (fp) fclose(fp);
  if (cc == 0) feclog("CAS ACTION file %.256s : found",fn);
  return caList;
}
void rmvCASActionList(int i)
{
  CasActLstEntry *ca, *can=NULL;
  for (ca=casTbl[i].actionList; ca != NULL; ca=can)
  {
    can = ca->nxt;
    free(ca);
  }
  casTbl[i].actionList = NULL;
}
CasAlmSysTblEntry defAlarmSystems[] = {
  { 100, 1, 1, "Magnets", "MAG" },
  { 110, 1, 2, "Ref. Magnets", "MAG" },
  { 150, 1, 3, "RF", "RF" },
  { 350, 1, 4, "Vacuum", "VAC" },
  { 400, 1, 5, "Kicker/Septum", "INJ" },
  { 410, 1, 6, "Transfer", "INJ" },
  { 450, 2, 1, "Beam Dump", "INJ" },
  { 500, 2, 2, "Feedback", "FB" },
  { 550, 2, 3, "Timing", "TIM" },
  { 600, 2, 4, "BPMs", "DIAG" },
  { 610, 2, 5, "BLMs", "DIAG" },
  { 800, 2, 6, "Collimators", "EXP" },
  {1200, 3, 6, "Infrastructure", "INFR" },
  {1600, 3, 3, "Services", "SER" },
  {1610, 3, 4, "Alarm Server", "SER" },
  {1620, 3, 5, "Archive Server", "HIST" },
  {5000, 3, 1, "System", "SYS" },
  {5001, 3, 2, "Hardware", "HDW" },
};
#define almSysTblHdr "NR, COLUMN, ROW, TAG, SUBSYSTEM"
int putAlarmSystemsList(char *fn,CasAlmSysTblEntry *items,int nitems)
{
  FILE *fp=NULL;
  int i, cc = 0;
  static time_t backup_done = 0;
  time_t t = time(NULL);

  if (t > backup_done + 86400)
  {
    remove("AlarmCodes.bak");
    rename("AlarmCodes.csv","AlarmCodes.bak");
    backup_done = t;
  }

  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"%s\n",almSysTblHdr);
  for (i=0; i<nitems; i++)
  {
    fprintf(fp,"%d, %d, %d, %.32s, %.16s\n",
      items[i].number,items[i].column,items[i].row,
      items[i].almsystem,items[i].subsystem);
  }
err:
  if (cc) feclog("putAlarmSystemsList : %.32s",cc2str(cc));
  if (fp != NULL) fclose(fp);
  return cc;
}

int nAlarmSystems = 0;    /* Each Alarm has an alarm system number */
CasAlmSysTblEntry *AlarmSystems = NULL;
CSVCOLUMN csvAlarmSystem[] =
{
  { "NR","",0,CF_LONG,1,(BYTE *)OFFSETIN(CasAlmSysTblEntry,number),NULL, 0},
  { "ROW","1",0,CF_LONG,1,(BYTE *)OFFSETIN(CasAlmSysTblEntry,row),NULL, 0},
  { "COLUMN","1",0,CF_LONG,1,(BYTE *)OFFSETIN(CasAlmSysTblEntry,column),NULL, 0},
  { "TAG","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE *)OFFSETIN(CasAlmSysTblEntry,almsystem),NULL, 0},
  { "SUBSYSTEM","",0,CF_TEXT,SUBSYSTEM_NAME_SIZE,(BYTE *)OFFSETIN(CasAlmSysTblEntry,subsystem),NULL, 0}
};
CSVDB csvAlarmSystemDb = 
{
  csvAlarmSystem,
  sizeof(csvAlarmSystem)/sizeof(csvAlarmSystem[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CasAlmSysTblEntry),
  0
};
int getAlarmSystemList(char *fn)
{
  int cc;
  if ((cc=csvReadFile(NULL,fn,&csvAlarmSystemDb,(void **)&AlarmSystems)) != 0) ccerr(cc);
  nAlarmSystems = csvAlarmSystemDb.siz;
err:
  return cc;
}
extern CSVDB csvAlarmInfoDb;
int needToAcquireAlarmDefs = FALSE;
int csvCASAlarmInfoGetFormat(char *fmtstr);
CSVCOLUMN csvCASAlarm5InfoColumns[] =
{
  { "ALARM_TAG","",0,CF_TEXT,ALARM_TAG_SHORTSIZE,(BYTE *)OFFSETIN(ADS5,alarmTag),NULL,0 },
  { "ALARM_CODE","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS5,alarmCode),NULL,0 },
  { "ALARM_MASK","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS5,alarmMask),NULL,0 },
  { "SEVERITY","1",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS5,severity),NULL,0 },
  { "DATA_FORMAT","",0,CF_NULL,1,NULL,csvCASAlarmInfoGetFormat,0 },
  { "DATA_ARRAYSIZE","0",0,CF_BYTE,1,(BYTE *)OFFSETIN(ADS5,alarmDataArraySize),NULL,0 },
  { "TIME_RESOLUTION","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS5,timeRes),NULL,0 },
  { "HITS_CHANNEL","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS5,hitsChannel),NULL,0 },
  { "HITS_MODULE","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS5,hitsModule),NULL,0 },
  { "ALARM_TEXT","not supplied",0,CF_TEXT,40,(BYTE *)OFFSETIN(ADS5,alarmText),NULL,0 },
  { "DEVICE_TEXT","not supplied",0,CF_TEXT,38,(BYTE *)OFFSETIN(ADS5,deviceText),NULL,0 },
  { "DATA_TEXT","not supplied",0,CF_TEXT,40,(BYTE *)OFFSETIN(ADS5,dataText),NULL,0 },
  { "URL","not supplied",0,CF_TEXT,40,(BYTE *)OFFSETIN(ADS5,url),NULL,0 },
  { "ALARM_SYSTEM","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS5,alarmSystem),NULL,0 }
};
CSVCOLUMN csvCASAlarmInfoColumns[] =
{
  { "ALARM_TAG","",0,CF_TEXT,ALARM_TAG_SIZE,(BYTE *)OFFSETIN(ADS,alarmTag),NULL,0 },
  { "ALARM_CODE","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS,alarmCode),NULL,0 },
  { "ALARM_MASK","0",0,CF_LONG,1,(BYTE *)OFFSETIN(ADS,alarmMask),NULL,0 },
  { "ALARM_SYSTEM","0",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS,alarmSystem),NULL,0 },
  { "SEVERITY","1",0,CF_SHORT,1,(BYTE *)OFFSETIN(ADS,alarmSeverity),NULL,0 },
  { "DATA_FORMAT","",0,CF_NULL,1,NULL,csvCASAlarmInfoGetFormat,0 },
  { "DATA_ARRAYSIZE","0",0,CF_BYTE,1,(BYTE *)OFFSETIN(ADS,alarmDataArraySize),NULL,0 },
  { "ALARM_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADS,alarmText),NULL,0 },
  { "DEVICE_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADS,deviceText),NULL,0 },
  { "DATA_TEXT","not supplied",0,CF_TEXT,ALARM_TEXT_SIZE,(BYTE *)OFFSETIN(ADS,dataText),NULL,0 },
  { "URL","not supplied",0,CF_TEXT,ALARM_TEXT_LONGSIZE,(BYTE *)OFFSETIN(ADS,url),NULL,0 },
};
CSVDB csvCASAlarmInfoDb =
{
  csvCASAlarmInfoColumns,
  sizeof(csvCASAlarmInfoColumns)/sizeof(csvCASAlarmInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ADS),
  0
};
int csvCASAlarmInfoGetFormat(char *fmtstr)
{
  if (fmtstr == NULL) return 0;
  ((ADS *)csvCASAlarmInfoDb.tgt)[csvCASAlarmInfoDb.cnt].alarmDataFormat = BFMT(ftoi(fmtstr));
  return 0;
}
int writeAlarmDefinitionsToFile(CasTblEntry *cl)
{
  int i, code, cc=0;
  FILE *fp=NULL;
  char fn[512],str[512];
  static char strdel[2] = { FS_DELIMITER , 0};

  sprintf(fn,"%.80s%sAlarmInfo%s%.16s%s%.32s-alarms.csv",
    AlarmArchivePath,strdel,strdel,cl->arcExt,strdel,cl->devServer);
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(no_such_file);
  strcpy(str,"ALARM_TAG,ALARM_CODE,ALARM_MASK");
  strcat(str,",SEVERITY,DATA_FORMAT,DATA_ARRAYSIZE");
  strcat(str,",ALARM_TEXT,DEVICE_TEXT,DATA_TEXT,URL,ALARM_SYSTEM");
  fprintf(fp,"%s\n",str); /* header */
  for (i=0; i<cl->nAdsList; i++)
  {
    if ((code=((ADS *)cl->adsList)[i].alarmCode) == 0) continue;
    str[ALARM_TAG_SIZE] = 0;
    strncpy(str,((ADS *)cl->adsList)[i].alarmTag,ALARM_TAG_SIZE);
    sprintf(&str[strlen(str)],",%d",code);
    sprintf(&str[strlen(str)],",%d",((ADS *)cl->adsList)[i].alarmMask);
    sprintf(&str[strlen(str)],",%d",((ADS *)cl->adsList)[i].alarmSeverity);
    sprintf(&str[strlen(str)],",%s",itof(((ADS *)cl->adsList)[i].alarmDataFormat));
    sprintf(&str[strlen(str)],",%d",((ADS *)cl->adsList)[i].alarmDataArraySize);
    strcat(str,","); str[strlen(str)+ALARM_TEXT_SIZE] = 0;
    strncpy(&str[strlen(str)],((ADS *)cl->adsList)[i].alarmText,ALARM_TEXT_SIZE);
    strcat(str,","); str[strlen(str)+ALARM_TEXT_SIZE] = 0;
    strncpy(&str[strlen(str)],((ADS *)cl->adsList)[i].deviceText,ALARM_TEXT_SIZE);
    strcat(str,","); str[strlen(str)+ALARM_TEXT_SIZE] = 0;
    strncpy(&str[strlen(str)],((ADS *)cl->adsList)[i].dataText,ALARM_TEXT_SIZE);
    strcat(str,","); str[strlen(str)+ALARM_TEXT_SIZE] = 0;
    strncpy(&str[strlen(str)],((ADS *)cl->adsList)[i].url,ALARM_TEXT_LONGSIZE);
    sprintf(&str[strlen(str)],",%d",((ADS *)cl->adsList)[i].alarmSystem);
    fprintf(fp,"%s\n",str);
  } 
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int getAlarmDefinitionsFromFile(CasTblEntry *cl)
{
  int cc=0;
  FILE *fp=NULL;
  char fn[512];
  static int done = 0;
  static char strdel[2] = { FS_DELIMITER , 0};

  sprintf(fn,"%.80s%sAlarmInfo%s%.16s%s%.32s-alarms.csv",
    AlarmArchivePath,strdel,strdel,cl->arcExt,strdel,cl->devServer);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  csvReadFile(NULL,fn,&csvCASAlarmInfoDb,(void **)&cl->adsList);
  cl->nAdsList = csvCASAlarmInfoDb.siz;
  cl->adsListAcquired = TRUE;
err:
  if (cc != 0 && cc != no_such_file) feclog("getAlarmDefinitionsFromFile: %.512s : %.32s",fn,erlst[cc]);
  return cc;
}
int ADStoADS5(ADS5 *ads5,ADS *ads)
{
  memset(ads5,0,ADS5_SIZE);
  strncpy(ads5->alarmTag,ads->alarmTag,ALARM_TAG_SHORTSIZE);
  ads5->alarmCode = ads->alarmCode;
  ads5->alarmMask = ads->alarmMask;
  ads5->alarmSystem = ads->alarmSystem;
  ads5->severity = ads->alarmSeverity;
  ads5->alarmDataFormat = ads->alarmDataFormat;
  ads5->alarmDataArraySize = ads->alarmDataArraySize;
  strncpy(ads5->alarmText,ads->alarmText,ALARM_TEXT_SHORTSIZE);
  strncpy(ads5->deviceText,ads->deviceText,ALARM_TEXT_SHORTSIZE-2);
  strncpy(ads5->dataText,ads->dataText,ALARM_TEXT_SHORTSIZE);
  strncpy(ads5->url,ads->url,ALARM_TEXT_SHORTSIZE);
  return 0;
}
int ADS5toADS(BYTE *adsstream,int length)
{
  ADS5 ads5;
  ADS *ads = (ADS *)adsstream;
  int i;

  for (i=length-1; i>=0; i--)
  { /* in-situ conversion */
    memcpy(&ads5,adsstream + i*ADS5_SIZE,sizeof(ADS5));
    memset(&ads[i],0,ADS_SIZE);
    strncpy(ads[i].alarmTag,ads5.alarmTag,ALARM_TAG_SHORTSIZE);
    ads[i].alarmCode = ads5.alarmCode;
    ads[i].alarmMask = ads5.alarmMask;
    ads[i].alarmSystem = ads5.alarmSystem;
    ads[i].alarmSeverity = ads5.severity;
    ads[i].alarmDataFormat = ads5.alarmDataFormat;
    ads[i].alarmDataArraySize = ads5.alarmDataArraySize;
    strncpy(ads[i].alarmText,ads5.alarmText,ALARM_TEXT_SHORTSIZE);
    strncpy(ads[i].deviceText,ads5.deviceText,ALARM_TEXT_SHORTSIZE-2);
    strncpy(ads[i].dataText,ads5.dataText,ALARM_TEXT_SHORTSIZE);
    strncpy(ads[i].url,ads5.url,ALARM_TEXT_SHORTSIZE);
  }
  return 0;
}
int getAlarmDefinitionsFromServer(CasTblEntry *cl)
{
  int i,k,cc=0,retrytag=FALSE;
  char devname[128],tag[TAG_NAME_SIZE],devprp[PROPERTY_NAME_SIZE];
  DTYPE dout;
  short nads, nads_last;
  ADS *ads=NULL, *ads_last=NULL;

  strcpy(devname,"(null)");
  if (cl == NULL) ccerr(argument_list_error);
  if (IsOffLine(cl)) ccerr(link_not_open);
  if (cl->adsKeepStaticInfo) ccerr(already_assigned);
  if (cl->linkStatus != 0) 
  {
    cl->adsListAcquired = TRUE;
    ccerr(cl->linkStatus);
  }
  nads_last = cl->nAdsList;
  sprintf(devname,"/%.32s/%.32s/*",cl->devContext,cl->devServer);
  strcpy(devprp,"NALMDEFS");
retryPrp:
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.dTag[0] = 0;
  dout.data.sptr = (SINT16 *)&nads;
  cc = ExecLinkEx(devname,devprp,&dout,NULL,CA_READ|CA_RETRY,500);
  if (cc == address_unknown || cc == non_existent_elem || cc == non_existent)
  {
    cl->offline = CAS_STS_REMOVE;
    cl->adsListAcquired = TRUE;
    feclog("error acquiring alarm definitions for %.128s : %.32s -> mark as off-line",devname,cc2str(cc));
    return cc;
  }
  if (cc == illegal_property && strcmp(devprp,"NALMDEFS") == 0)
  { /* work-around for a java server bug */
    strcpy(devprp,"NALARMDEFS");
    goto retryPrp;
  }
  if (cc != 0) ccerr(cc);
  if (nads == 0) ccerr(0);
  if (nads > 1000) nads = 1000; /* combat an old epics2tine problem */
  if ((ads=(ADS *)calloc(nads+nads_last,sizeof(ADS))) == NULL)
    ccerr(out_of_server_memory);
  strncpy(tag,"ADSr4",TAG_NAME_SIZE);
retryTag:
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = nads;
  strcpy(dout.dTag,tag);
  dout.data.vptr = ads;
  cc = ExecLinkEx(devname,&devprp[1],&dout,NULL,CA_READ|CA_RETRY,1000);
  if (cc) 
  { /* completion error */
    if (cc == semaphore_busy) goto retryTag;
    if (cc != connection_timeout && !retrytag)
    { /* seems to be a pre-release 4 server */
      retrytag = TRUE;
      strncpy(tag,"ADS",TAG_NAME_SIZE);
      goto retryTag;
    }
    ccerr(cc);
  }
  if (retrytag && !strcmp(tag,"ADS"))
  { /* a retrytag was successful => covert ADS */
    ADS5toADS((BYTE *)ads,nads);
  }
  if (cl->adsList != NULL) 
  { /* merge if necessary ! */
    ads_last = cl->adsList;
    for (i=0; i<nads_last; i++)
    {
      for (k=0; k<nads; k++) if (ads_last[i].alarmCode == ads[k].alarmCode) break;
      if (k == nads)
      { /* this 'last' ads was NOT among the new ones: keep it! (there's room) */
        ads[nads++] = ads_last[i];
      }
    }
    free(cl->adsList);
  }
  for (i=0,k=0; i<nads; i++)
  {
    if (ads[i].alarmCode == 0) continue; /* invalid information: skip */
    if (ads[i].alarmSystem == 0) ads[i].alarmSystem = cl->AlarmSystem;
    if (i > k) memcpy(&ads[k],&ads[i],sizeof(ADS));
    k++;
  }
  nads = k;
  cl->nAdsList = nads;
  cl->adsList = ads;
  writeAlarmDefinitionsToFile(cl); 
err:
  if (cc != 0) 
  {
    if (ads != NULL) free(ads);
    if (cl != NULL && !IsOffLine(cl)) feclog("error acquiring alarm definitions for %.128s : %.32s",devname,cc2str(cc));
  }
  if (cc == 0 && cl != NULL) 
  {
    feclog("acquired alarm definitions for %s",devname);
    cl->adsListAcquired = TRUE;
  }
  return cc;
}
int getAlarmDefinitions(CasTblEntry *cl)
{
  int cc1, cc2;
  cc1 = getAlarmDefinitionsFromFile(cl);
  cc2 = getAlarmDefinitionsFromServer(cl);
  if (cc1 == 0 || cc2 == 0) return 0;
  return cc2;
}
char *getAlarmMessageFromDefinition(int i,int almcode)
{
  int k;
  
  for (k=0; k<casTbl[i].nAdsList; k++)
  {
    if (casTbl[i].adsList[k].alarmCode == (UINT32)almcode) return casTbl[i].adsList[k].alarmTag;
  }
  return "unknown";
}
int putServerAlarmSystemList(CasTblEntry *cl,int *lst,int siz)
{
  FILE *fp = NULL;
  int cc=0,i;
  char fn[64];

  if (cl == NULL || strlen(cl->devServer) == 0) ccerr(argument_list_error);
  sprintf(fn,"%.80s/AlarmInfo/%.16s/alarmsystems.csv",AlarmArchivePath,cl->arcExt);
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"SYSTEM\n");
  for (i=0; i<siz; i++)
  {
    if (lst[i] == cl->AlarmSystem) continue;
    fprintf(fp,"%d\n",lst[i]);
  }
err:
  if (fp) fclose(fp);
  return cc;
}
int getServerAlarmSystemList(CasTblEntry *cl)
{
  FILE *fp = NULL;
  int cc=0,done=0,i=0,n,nr;
  int sys_col=-1;
  char fn[64],str[256],hdr[256],scratch[16];

  if (cl == NULL || strlen(cl->devServer) == 0) ccerr(argument_list_error);
  sprintf(fn,"%.80s/AlarmInfo/%.16s/alarmsystems.csv",AlarmArchivePath,cl->arcExt);
  /* sprintf(fn,"%s-alarmsystems.csv",cl->devServer); <- was */
  if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
  if (n == 0) ccerr(file_error);
  if ((cl->almSystemList=(int *)calloc(n,sizeof(int))) == NULL) ccerr(out_of_server_memory);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;    /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((sys_col=findcol(hdr,"SYSTEM")) < 0) ccerr(no_such_column);
      done = TRUE;                            /* read header info */
      continue;
    }
    strncpy(scratch,colptr(sys_col,str),16);
    nr = atoi(scratch);
    if (nr == 0) nr = getAlarmSystemNumberFromTag(scratch);
    cl->almSystemList[i] = nr;
    feclog("additional alarm system %d for %s registered (default %d)",nr,cl->devServer,cl->AlarmSystem);
    i++;
  }
  cl->nAlmSystemList = n;
err:
  if (fp) fclose(fp);
  return cc;
}
void rmvServerAlarmSystemList(CasTblEntry *cl)
{
  if (cl->almSystemList == NULL) return;
  free(cl->almSystemList); cl->nAlmSystemList = 0;
}
int nAlarmSubSystems = 0; /* Each FEC belongs to a sub-system */
NAME16 *AlarmSubSystems = NULL;
INTINT *AlarmSubRanges  = NULL;
int isInServerList(char *ctx, char *srv)
{
  int i;

  for (i=0; i<casTblSiz; i++)
  {
    if (strnicmp(casTbl[i].devServer,srv,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(casTbl[i].devContext,ctx,CONTEXT_NAME_SIZE)) continue;
    return TRUE;
  }
  return FALSE;
}
int getCasTbl(char *fn)
{
  FILE *fp = NULL;
  int cc=0,done=0,i=0,n;
  int con_col,srv_col,sev_col,arc_col,ret_col,alm_col,sub_col,sys_col,ext_col,off_col;
  char str[256],hdr[256],scratch[16];

  if ((casTblSiz=GetDataBaseSize(fn)) < 0) ccerr(-casTblSiz);
  if (casTblSiz == 0) ccerr(0);
  if (casTbl != NULL) free(casTbl);
  casTblCapacity = casTblSiz + CAS_CAPACITY_BUFFER;
  if ((casTbl=(CasTblEntry *)calloc(casTblCapacity,sizeof(CasTblEntry))) == NULL) ccerr(out_of_server_memory);
  /* it won't be any worse than this: */
  if ((AlarmSubSystems=(NAME16 *)calloc(casTblCapacity,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  if ((AlarmSubRanges=(INTINT *)calloc(casTblCapacity,sizeof(INTINT))) == NULL) ccerr(out_of_server_memory);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      con_col = findcol(hdr,"CONTEXT");
      if ((srv_col=findcol(hdr,"SERVER")) < 0) ccerr(no_such_column);
      sev_col = findcol(hdr,"SEVERITYLEVEL");
      arc_col = findcol(hdr,"ARCHIVELEVEL");
      ret_col = findcol(hdr,"RETENTION");
      alm_col = findcol(hdr,"ALARMLEVEL");
      sub_col = findcol(hdr,"SUBSYSTEM");
      sys_col = findcol(hdr,"ALARMSYSTEM");
      ext_col = findcol(hdr,"EXTENSION");
      off_col = findcol(hdr,"OFFLINE");
      done = 1;                              /* read header info */
      continue;
    }
    strncpy(casTbl[i].devServer,colptr(srv_col,str),EXPORT_NAME_SIZE);
    strncpy(casTbl[i].devContext,con_col >= 0 ? colptr(con_col,str) : "DEFAULT",CONTEXT_NAME_SIZE);
    strncpy(casTbl[i].arcExt,ext_col >= 0 ? colptr(ext_col,str) : casTbl[i].devServer,XTN_NAME_SIZE);

    if (off_col >= 0)
    {
      strncpy(scratch,colptr(off_col,str),16);
      casTbl[i].offline = atoi(scratch);
      if (casTbl[i].offline == 5555)
      { /* special insider job ... */
        casTbl[i].offline = 0;
        casTbl[i].adsKeepStaticInfo = TRUE;
      }
    }
    if (sev_col >= 0)
    {
      strncpy(scratch,colptr(sev_col,str),16);
      casTbl[i].SeverityLevel = atoi(scratch);
    }
    if (casTbl[i].SeverityLevel < 1) casTbl[i].SeverityLevel = 1;
    if (arc_col >= 0)
    {
      strncpy(scratch,colptr(arc_col,str),16);
      casTbl[i].ArchiveLevel = atoi(scratch);
    }
    if (ret_col >= 0)
    {
      strncpy(scratch,colptr(ret_col,str),16);
      casTbl[i].Retention = atoi(scratch);
    }
    if (casTbl[i].Retention < DEFAULT_ALARM_HEARTBEAT * 3)
      casTbl[i].Retention = DEFAULT_ALARM_HEARTBEAT * 3;
    if (alm_col >= 0)
    {
      strncpy(scratch,colptr(alm_col,str),16);
      casTbl[i].AlarmLevel = atoi(scratch);
      if (casTbl[i].AlarmLevel > ALMSYSTEM_FATAL) casTbl[i].AlarmLevel = ALMSYSTEM_FATAL;
    }
    if (casTbl[i].AlarmLevel < ALMSYSTEM_INFO) casTbl[i].AlarmLevel = ALMSYSTEM_INFO;

    if (sys_col >= 0)
    {
      strncpy(scratch,colptr(sys_col,str),16);
      casTbl[i].AlarmSystem = atoi(scratch);
    }
    if (sub_col >= 0)
    {
      strncpy(casTbl[i].subSystem,colptr(sub_col,str),SUBSYSTEM_NAME_SIZE);
      for (n=0; n<nAlarmSubSystems; n++) if (!strncmp(casTbl[i].subSystem,AlarmSubSystems[n].name,SUBSYSTEM_NAME_SIZE)) break;
      if (n == nAlarmSubSystems)
      {
        strncpy(AlarmSubSystems[n].name,casTbl[i].subSystem,SUBSYSTEM_NAME_SIZE);
        nAlarmSubSystems++;
      }
      if (casTbl[i].AlarmSystem <= (int)AlarmSubRanges[n].i1val || 
          AlarmSubRanges[n].i1val == 0)
        AlarmSubRanges[n].i1val = casTbl[i].AlarmSystem;
      if (casTbl[i].AlarmSystem >= (int)AlarmSubRanges[n].i2val)
        AlarmSubRanges[n].i2val = casTbl[i].AlarmSystem;
    }
    else strncpy(casTbl[i].subSystem,"TST",SUBSYSTEM_NAME_SIZE);

    casTbl[i].disabledList = getCASDisabledList(i);
    casTbl[i].actionList = getCASActionList(i);
    getServerAlarmSystemList(&casTbl[i]);

    i++;
  }
err:
  if (fp) fclose(fp);
  feclog("CAS file %s : %.32s",fn,cc ? erlst[cc] : "found");
  return cc;
}

#define casLstHdr "Context, Server, Extension, SeverityLevel, ArchiveLevel, Retention, AlarmLevel, SubSystem, AlarmSystem, Offline"
int dumpCASDb(void)
{
  FILE *fp=NULL;
  int i, cc = 0;
  static time_t backup_done = 0;
  time_t t = time(NULL);

  if (t > backup_done + 86400)
  {
    remove("ServerList.bak");
    rename("ServerList.csv","ServerList.bak");
    backup_done = t;
  }
  if ((fp=fopen(casListFile,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"%s\n",casLstHdr);
  for (i=0; i<casTblSiz; i++)
  {
    if (casTbl[i].offline < 0) continue; /* marked for removal */
    fprintf(fp,"%.32s, %.32s, %.16s, %d, %d, %d, %d, %.16s, %d, %d\n",
      casTbl[i].devContext,casTbl[i].devServer,casTbl[i].arcExt,
      casTbl[i].SeverityLevel,casTbl[i].ArchiveLevel,casTbl[i].Retention,casTbl[i].AlarmLevel,
      casTbl[i].subSystem,casTbl[i].AlarmSystem,casTbl[i].offline);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int appendCASDb(CasDbItem *dbItems,int nitems)
{
  int i, k, cc = 0;

  for (k=0; k<nitems; k++)
  {
    for (i=0; i<casTblSiz; i++)
    { /* is already in the table ? (just an edit) */
      if (strncmp(casTbl[i].devContext,dbItems[k].devContext,CONTEXT_NAME_SIZE)) continue;
      if (strncmp(casTbl[i].devServer,dbItems[k].devServer,EXPORT_NAME_SIZE)) continue;
      strncpy(casTbl[i].subSystem,dbItems[k].subSystem,SUBSYSTEM_NAME_SIZE);
      strncpy(casTbl[i].arcExt,dbItems[k].arcExt,XTN_NAME_SIZE);
      casTbl[i].SeverityLevel = dbItems[k].SeverityLevel;
      casTbl[i].ArchiveLevel = dbItems[k].ArchiveLevel;
      casTbl[i].Retention = dbItems[k].Retention;
      casTbl[i].AlarmLevel = dbItems[k].AlarmLevel;
      casTbl[i].AlarmSystem = dbItems[k].AlarmSystem;
      casTbl[i].offline = dbItems[k].offline;
      break;
    }
    if (i < casTblSiz) continue;
    /* is new ! */
    if (casTblSiz >= casTblCapacity) 
    {
      feclog("appendCASDb: number of records would exceed buffer capacity (%d)",casTblCapacity);
      return resources_exhausted;
    }
    if (strlen(dbItems[k].arcExt) == 0) strncpy(dbItems[k].arcExt,dbItems[k].devServer,XTN_NAME_SIZE);
    strncpy(casTbl[i].devContext,dbItems[k].devContext,CONTEXT_NAME_SIZE);
    strncpy(casTbl[i].devServer,dbItems[k].devServer,EXPORT_NAME_SIZE);
    strncpy(casTbl[i].subSystem,dbItems[k].subSystem,SUBSYSTEM_NAME_SIZE);
    strncpy(casTbl[i].arcExt,dbItems[k].arcExt,XTN_NAME_SIZE);
    casTbl[i].SeverityLevel = dbItems[k].SeverityLevel;
    casTbl[i].ArchiveLevel = dbItems[k].ArchiveLevel;
    casTbl[i].Retention = dbItems[k].Retention;
    casTbl[i].AlarmLevel = dbItems[k].AlarmLevel;
    casTbl[i].AlarmSystem = dbItems[k].AlarmSystem;
    casTbl[i].offline = dbItems[k].offline;
    casTblSiz++;
  }
  return dumpCASDb();
}
#define casActHdr "AlarmCode, MailTo, Context, Server, Device, Property, ActionFormat, ActionString"
int writeAlmActionLst(int devnr,CasActionItem *actItems,int nitems)
{
  int i, cc = 0;
  FILE *fp=NULL;
  char fn[256];

  sprintf(fn,"%.80s/AlarmInfo/%s/actions.csv",AlarmArchivePath,casTbl[devnr].arcExt);
  if (nitems < 1)
  {
    unlink(fn);
    feclog("remove actions file %.256s",fn);
    return 0; /* just return success (maybe file wasn't there to begin with) */
  }
  if ((fp=fopen(fn,"w")) == NULL &&
      (fp=createAndOpen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"%s\n",casActHdr);
  for (i=0; i < nitems; i++)
  {
    fprintf(fp,"%d, %.128s, %.32s, %.32s, %.64s, %.64s, %.16s, %.128s\n",
      actItems[i].alarmCode,actItems[i].email,actItems[i].ctx,actItems[i].srv,actItems[i].dev,actItems[i].prp,
      GetFormatAsString(actItems[i].actionSignalFormat),actItems[i].actionSignalString);
  }
err:
  if (fp != NULL) fclose(fp);
  feclog("write actions file %.256s : %.32s",fn,cc2str(cc));
  return cc;
}
int getAlmSubSys(char *subsys)
{
  int i;
  for (i=0; i<nAlarmSystems; i++)
  {
    if (!stricmp(subsys,AlarmSystems[i].subsystem))
      return AlarmSystems[i].number;
  }
  return 1600; /* services */
}
int appendCasTbl(int isNew)
{
  DTYPE dout,din;
  NAME16 inp[2];
  DUSTRING outp[200];
  char devName[128],srv[32],sub[16],*ptr;
  int cc=0, i, num, asys;
  FILE *fp=NULL;

  if (!isNew) 
  { /* if there's a list, don't append unless env says so ! */
    if ((ptr=getenv("APPEND_LIST")) == NULL) return 0;
    if (stricmp(ptr,"TRUE") != 0) return 0;
    remove("ServerList.bak");
    rename("ServerList.csv","ServerList.bak");
  }
  strcpy(inp[0].name,"ALL");
  strcpy(inp[1].name,"IMPORTANT");
  din.dArrayLength = 2;
  din.dFormat = CF_NAME16;
  din.data.vptr = inp;

  /* TODO: a USTRING will return both server name and subsystem, but will truncate the server name to 16 chars ! */
  dout.dArrayLength = 200;
  dout.dFormat = CF_USTRING;
  dout.data.vptr = outp;

  sprintf(devName,"/SITE/ENS/%s",gDeviceContext);
  cc = ExecLinkEx(devName,"TAGS",&dout,&din,CA_READ|CA_RETRY|CA_MUTABLE,500);

  if (cc != 0)
  {
    feclog("could not query ENS for important servers : %s",erlst[cc]);
    return cc;
  }

  if ((fp=fopen("ServerList.csv","w")) == NULL)
  {
    feclog("could not create file ServerList.csv !");
    return file_error;
  }
  fprintf(fp,"%s\n",casLstHdr);
  for (i=0; i<casTblSiz; i++)
  { /* any entries already read in ? */
    fprintf(fp,"%.32s, %.32s, %.16s, %d, %d, %d, %d, %.16s, %d, %d\n",
      casTbl[i].devContext,casTbl[i].devServer,casTbl[i].arcExt,
      casTbl[i].SeverityLevel,casTbl[i].ArchiveLevel,casTbl[i].Retention,casTbl[i].AlarmLevel,
      casTbl[i].subSystem,casTbl[i].AlarmSystem,casTbl[i].offline);
  }  
  num = dout.dArrayLength;
  for (i=0; i<num; i++)
  {
    sprintf(srv,"%.16s",outp[i].str); srv[16] = 0;
    sprintf(sub,"%.8s",&outp[i].str[24]); sub[8] = 0;
    asys = getAlmSubSys(sub);
    if (!isNew && isInServerList(gDeviceContext,srv)) continue;
    if (!strcmp(srv,"CAS")) continue; /* don't put myself in the list */
    fprintf(fp,"%.32s, %.32s, %.16s, 0, 7, 7200, 1, %.16s, %d, 0\n",
      gDeviceContext, srv, srv, sub, asys);
  }

  fclose(fp);

  return getCasTbl(casListFile);
}

void TakeActionCallback(int id,int cc,void *ref)
{
  CasActLstEntry *ca = (CasActLstEntry *)ref;
  char str[256],tmp[64];
  int i;
  DTYPE din;

  if (cc)
  {
    feclog("Action %s %s: %s",ca->devName,ca->devProperty,erlst[cc]);
    if (ca->retry-- > 0)
    {
      sprintf(str,"retry %s %s\n",ca->devName,ca->devProperty); feclog(str);
      TakeAction(ca);
    }
  }
  else
  {
    if (!strcmp(ca->devProperty,"TRIGGER"))
    { /* prepare annotation */
      sprintf(str,"%s (generated by %s %s) : ",
        ca->ams->alarmTag,ca->ams->server,ca->ams->device);
      for (i=0; i<ca->ams->alarmDataArraySize; i++)
      {
        switch (LFMT(ca->ams->alarmDataFormat))
        {
          case CF_SHORT:
            sprintf(tmp,"%d ",((short *)ca->ams->alarmData)[i]);
            strcat(str,tmp);
            break;
          case CF_LONG:
            sprintf(tmp,"%d ",((UINT32 *)ca->ams->alarmData)[i]);
            strcat(str,tmp);
            break;
          case CF_FLOAT:
            sprintf(tmp,"%g ",((float *)ca->ams->alarmData)[i]);
            strcat(str,tmp);
            break;
          case CF_DOUBLE:
            sprintf(tmp,"%g ",((double *)ca->ams->alarmData)[i]);
            strcat(str,tmp);
            break;
          case CF_TEXT:
            sprintf(tmp,"%s ",(char *)ca->ams->alarmData);
            strncat(str,tmp,64);
            i = ca->ams->alarmDataArraySize;
            break;
          default:
            sprintf(tmp,"%d ",ca->ams->alarmData[i]);
            strcat(str,tmp);
            break;
        }
      }
      sprintf(&str[strlen(str)],"@%d",(SINT32)GetDataTimeStamp());
      din.dFormat = CF_TEXT;
      din.data.cptr = str;
      din.dArrayLength = strlen(str);
      if ((i=AttachLink(ca->devName,"ANNOTATE",NULL,&din,CA_WRITE,100,NULL,CM_SINGLE)) < 0) cc = -i;
    }
  }
  ca->status = cc;
}
void TakeAction(CasActLstEntry *ca)
{
  int lnkID;
  char str[128];
  time_t ts;
# if defined(UNIX)
  pthread_t tid;
  pthread_attr_t gMailThreadAttr, *gPtrMailThreadAttr = &gMailThreadAttr;
# elif defined(WIN32)
  DWORD tid;
# endif

  if (!(ca->ams->descriptor & almNEWALARM)) return;
  feclog("take action for %s %s",ca->devName,ca->devProperty);
  if (ca->access > 0)
  { /* action involves sending a command */
    ca->status = -1;
    lnkID = AttachLinkEx2(ca->devName,ca->devProperty,&ca->dout,&ca->din,(short)ca->access,100,TakeActionCallback,CM_SINGLE,UNASSIGNED_CALLBACKID,ca);
    if (lnkID < 0)
    {
      ca->status = -lnkID;
      sprintf(str,"Action Link to %s %s : %s",ca->devName,ca->devProperty,erlst[-lnkID]);
      feclog(str);
      logproblem(str);
    }
  }
  /* send all action items to the central logger */
  ts = (time_t)ca->ams->timestamp;
  sprintf(str,"On %s server %s -> %s alarm %s occurred at %.24s",
     gDeviceContext,ca->ams->server,ca->ams->device,ca->ams->alarmTag,ctime(&ts));
  clslog(gDeviceContext,getAlarmSystemTagFromNumber(ca->ams->alarmSystem),"CAS",CLOG_STATUS_WARN,CLOG_PRIORITY_IMPORTANT,str);
  if (strlen(ca->email) > 0)
  { /* action involves sending an email to a recipient */
#   if defined(UNIX)
    pthread_attr_init(gPtrMailThreadAttr);
    pthread_create(&tid,gPtrMailThreadAttr,mtSendAlarmAsMail,(void *)ca);
#   elif defined(WIN32)
    CreateThread(NULL,0,mtSendAlarmAsMail,(LPVOID)ca,0,&tid);
#   else
    sendAlarmAsMail(ca->email,ca->ams->alarmTag,str);
#   endif
  }
}

THREAD_TYPE mtSendAlarmAsMail(THREAD_PARAM lpvThreadParam)
{
  CasActLstEntry *ca = (CasActLstEntry *)lpvThreadParam;
  char str[1024], dstr[512];
  time_t ts = (time_t)ca->ams->timestamp;
  ADS *ads = NULL;
  if (ca->ams->alarmDataFormat != CF_NULL && ca->ams->alarmDataArraySize > 0)
  {
    prepDataDump(dstr,(char *)ca->ams->alarmData,(int)ca->ams->alarmDataArraySize,(short)LFMT(ca->ams->alarmDataFormat),FALSE);    
  }
  else
  {
    dstr[0] = 0;
  }
  snprintf(str,512,"On %.32s server %.32s -> %.64s alarm %.32s occurred at %.24s",
     gDeviceContext,ca->ams->server,ca->ams->device,ca->ams->alarmTag,ctime(&ts));
  if (strlen(dstr) > 0) 
  {
    if ((ads=getAlarmDefinition(&casTbl[ca->casTblIdx],ca->ams->alarmCode)) != NULL)
    {
      strcat(str,"\n");
      strncat(str,ads->dataText,ALARM_TEXT_SIZE);
      strcat(str,": ");
    }
    else
    {
      strcat(str,"\nData: ");
    }
    strncat(str,dstr,64);
  }
  if (WaitForSystemMutex(hSendMailMutex,-1) != 0) 
  {
    feclog("mtSendAlarmAsMail: mailer mutex unavailable !");
    return 0;
  }
  sendAlarmAsMail(ca->email,ca->ams->alarmTag,str);
  ReleaseSystemMutex(hSendMailMutex);
  return 0;
}

void sendAlarmAsMail(char *addr,char *subj,char *body)
{
# if defined(UNIX)
  FILE *fp;
  char hdr[128];
  if (addr == NULL || body == NULL) return;
  if (subj == NULL) subj = "Alarm From Central Alarm Server";
  sprintf(hdr,"/usr/bin/mail -s '%s' %s",subj,addr);
  fp = popen(hdr, "w");
  fprintf(fp, "%s.\n",body);
  pclose(fp);
# elif defined(WIN32)
  dbglog("mail not implemented on windows yet");
# else
  dbglog("not implemented yet");
# endif
}
UINT32 getStartFromAlmSet(AMS *almbuf,int *num)
{
  int i, n;
  UINT32 t = (UINT32)time(NULL);
  if (almbuf == NULL || num == NULL) return 0;
  if ((n = *num) == 0) return 0;
  for (i=0; i<n; i++)
  {
    if (almbuf[i].timestamp == 0) break;
    if (almbuf[i].timestamp < t) t = almbuf[i].timestamp;
  }
  *num = i;
  return t;
}
int prepIncomingAMS(AMS *ams,BYTE *amsstream,int num,int isLegacy)
{
  AMS5 ams5;
# ifndef INTEL
  BYTE ams5buffer[AMS5_SIZE];
# endif
  int i;

  if (isLegacy)
  { /* old-style : convert from AMS5 to AMS */
    for (i=num-1; i>=0; i--)
    { /* work backwards for in-situ conversion */
      memset(&ams5,0,sizeof(AMS5));
#     ifndef INTEL
      memcpy(ams5buffer, amsstream + i*AMS5_SIZE, AMS5_SIZE);
      _StructSwapEx(AMS5StructStruct,ams5buffer,(BYTE *)&ams5,RTOL,1,sizeof(AMS5));
#     else
      memcpy(&ams5, amsstream + i*AMS5_SIZE, AMS5_SIZE);
#     endif
      memset(&ams[i],0,sizeof(AMS));
      strncpy(ams[i].server,ams5.server,EXPORT_NAME_SHORTSIZE);
      strncpy(ams[i].device,ams5.device,DEVICE_NAME_SHORTSIZE);
      strncpy(ams[i].alarmTag,ams5.alarmTag,ALARM_TAG_SHORTSIZE);
      ams[i].alarmCode = ams5.alarmCode;
      ams[i].timestamp = ams5.timestamp;
      ams[i].alarmMask = ams5.alarmMask;
      memcpy(ams[i].alarmData,ams5.alarmData,ALARM_DATA_SHORTSIZE);
      ams[i].alarmDataFormat = ams5.alarmDataFormat;
      ams[i].alarmDataArraySize = ams5.alarmDataArraySize;
      ams[i].severity = ams5.severity;
      ams[i].descriptor = ams5.descriptor;
      ams[i].alarmSystem = ams5.alarmSystem;
      if (ams[i].descriptor & almNEWALARM)
      {
        ams[i].starttime = ams5.timestamp;
      }
    }
  }
  else
  {
#   ifndef INTEL
    _StructSwapEx(AMSStructStruct,amsstream,(BYTE *)ams,RTOL,num,sizeof(AMS));
#   else
    if ((void *)amsstream != (void *)ams) memcpy(ams,amsstream,num*AMS_SIZE);
#   endif
  }
  return 0;  
}
void fixAlarmSystem(AMS *ams,int num,int alarmSystem)
{
  int i;
  for (i=0; i<num; i++)
  {
    if (ams[i].alarmSystem == 0) ams[i].alarmSystem = alarmSystem;
  }
}
#define ActionMask (almHEARTBEAT|almOSCILLATION|almTERMINATE)
#define CanTakeAction(d) (((d) & almTRANSIENT) || !((d) & ActionMask))
void ProcessAlarms(int id, int cc)
{
  int i=id-1,k,n;
  CAMS *cams = NULL;
  AMS *ams;
  CasActLstEntry *ca;
  CasDisableLstEntry *cd;
  CasTblEntry *cl = &casTbl[i];
  UINT32 starttime = 0;

  if (i<0 || i>= casTblSiz) return;
  cl->hasAcquiredInitialAlarmSet = TRUE;
  if (IsOffLine(cl)) return;
  if (tineDebug) dbglog("Process alarms for %.32s : %.32s",cl->devServer,erlst[cc]);
  if (cc) 
  {
    if (cc == address_unknown || cc == non_existent_elem || cc == non_existent)
    {
      cl->offline = CAS_STS_REMOVE;
      feclog("error processing alarms for /%.32s/%.32s : %.32s -> mark as off-line",
        cl->devContext,cl->devServer,cc2str(cc));
      return;
    }
    ccerr(cc);
  }
  n = cl->linkId;
  if (conTbl[n]->tineProtocol < 6) cl->legacy = TRUE;
  prepIncomingAMS(cl->almBuffer,(BYTE *)cl->almBuffer,cl->pending,cl->legacy);
  starttime = getStartFromAlmSet(cl->almBuffer,&cl->pending);
  fixAlarmSystem(cl->almBuffer,cl->pending,cl->AlarmSystem);
  if (cl->pending == 0) return;
  if (starttime <= 0) 
  {
    cl->pending = 0;
    feclog("process alarms for %.32s gave an invalid starttime %d",cl->devServer,starttime);
    return;
  }
  if (cl->legacy)
  {
    for (n=0; n<cl->pending; n++)
    { /* run thru the list to find the start time */
      if (cl->almBuffer[n].starttime != 0) continue; /* must have been a new one */
      if ((ams=findAlarmInList(cl->almList,cl->almBuffer[n].alarmCode,NULL)) != NULL)
      { /* found one */
        cl->almBuffer[n].starttime = ams->starttime;
      }
    }
  }
  if ((cams=(CAMS *)calloc(1,sizeof(CAMS))) == NULL) ccerr(out_of_server_memory);
  if (WaitForSystemMutex(hAmsTblMutex,-1) != 0) 
  {
    cl->pending = 0;
    feclog("processAlarms: AMS Table mutex unavailable !");
    return;
  }
  cams->starttime = starttime;
  cams->stoptime = cl->acqSnapshot.mrt;
  cams->nalms = cl->pending;
  cams->ams = cl->almBuffer;
  cams->nxt = cl->almList;
  cl->almList = cams;
  ReleaseMutex(hAmsTblMutex);
  for (k=0; k<nsysalias; k++)
  { /* any aliases to deal with ? */
    for (n=0; n<cams->nalms; n++)
    {
      if (cams->ams[n].alarmSystem == (UINT16)sysalias[k].system)
      {
        cams->ams[n].alarmSystem = (UINT16)sysalias[k].alias;
      }
    }
  }
  if (cl->disabledList != NULL)
  { /* are there incoming alarms to disable ? */
    CasDisableLstEntry *cdp, *cdn=NULL;
    int updateList = FALSE;
    for (n=0; n<cams->nalms; n++)
    { /* compare every alarm against the disabled list */
      cdp = cl->disabledList;
      for (cd=cdp; cd != NULL; cd=cdn)
      {
        cdn = cd->nxt;
        if (strlen(cd->device) > 0 && strncmp(cams->ams[n].device,cd->device,DEVICE_NAME_SIZE)) 
        { /* device names don't match */
          cdp = cd;
          continue;
        }
        if (cd->alarmCode == cams->ams[n].alarmCode)
        { /* a match ! */
          if (cams->ams[n].descriptor & almTERMINATE &&
              cd->disableType == ALM_DISABLE_CURRENT_ACTIVE)
          { /* current disabled alarm is no longer active ! */
            if (cdp != cl->disabledList) cdp->nxt = cd->nxt; else cdp = cl->disabledList = cd->nxt;
            free(cd); cd = cdp;
            if (cl->sizDisabledList > 0) cl->sizDisabledList--;
            updateList = TRUE;
          }
          else
          {
            cams->ams[n].descriptor |= (almDISABLED|almSUPPRESS);
          }
        }
        cdp = cd;
      }
    }
    if (updateList) writeCASDisabledList(i);
  }  
  if (cl->actionList != NULL)
  { /* are there action items related to the incoming alarms ? */
    for (n=0; n<cams->nalms; n++)
    {
      for (ca=cl->actionList; ca != NULL; ca=ca->nxt)
      {
        ca->retry = 3;
        if (ca->alarmCode == cams->ams[n].alarmCode)
        {
          if (CanTakeAction(cams->ams[n].descriptor))
          {
            ca->ams = &(cams->ams[n]);
	          TakeAction(ca);
          }
        }
      }
    }
  }
  /* are there non-target alarm systems to consider ? (nAlmSystemList */
# ifdef CAS2DOOCS
  for (n=0; n<cams->nalms; n++)
  { /* those registered servers will send to DOOCS */
    if (cams->ams[n].descriptor & almSUPPRESS) continue;
    SendAlarmToDoocs(&cams->ams[n],"run");
  }
# endif
err:
  cl->pending = 0; /* clear the pending here in case of retries ... */
  if (cc == 0) 
  { /* successful */
    cl->prevSnapshot = cl->acqSnapshot;
    cl->timestamp = cl->acqSnapshot.mrt;
    SystemScheduleProperty("CASEQM","NUMALARMS,NUMALARMS.RAW");
  }
  else
  {
    feclog("Process Alarm /%.32s/%.32s : %.32s",cl->devContext,cl->devServer,cc2str(cc));
  }
  return;
}
void AcquireAlarmInfo(int i)
{
  char lnkDevName[128];
  int lnkID, nalm;
  DTYPE dout,din;
  AMRB amrb;
  UINT32 buf[4];
  char str[128];

  casTbl[i].pending = 0;
  if (casTbl[i].thisSnapshot.nmrt == 0) 
  {
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  if (casTbl[i].thisSnapshot.nmrt > casTbl[i].thisSnapshot.nalm)
  {
    feclog("irregularity in number of recent alarms for server %.32s : %d > %d",
      casTbl[i].devServer,casTbl[i].thisSnapshot.nmrt,casTbl[i].thisSnapshot.nalm);
    casTbl[i].thisSnapshot.nmrt = casTbl[i].thisSnapshot.nalm;    
  }
  if (casTbl[i].thisSnapshot.nmrt > MAX_ALARMSNAPSHOT_SIZE)
  {
    feclog("truncating number of recent alarms for server %.32s from %d to %d",
      casTbl[i].devServer,casTbl[i].thisSnapshot.nmrt,MAX_ALARMSNAPSHOT_SIZE);
    casTbl[i].thisSnapshot.nmrt = MAX_ALARMSNAPSHOT_SIZE;    
  }
  nalm = casTbl[i].acqSnapshot.nalm; /* casTbl[i].acqSnapshot.nmrt; */
  casTbl[i].pending = nalm;
  if (casTbl[i].acqSnapshot.mrt == 0) 
  { /* if the device server has never set an alarm then acqSnapshot == 0 !*/
    casTbl[i].pending = 0;
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  if ((casTbl[i].almBuffer=(AMS *)calloc(nalm,AMS_SIZE)) == NULL)
  {
    feclog("insufficient memory for %d alarms for %.32s",nalm,casTbl[i].devServer);
    casTbl[i].pending = 0;
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  if (casTbl[i].legacy)
  { /* pre-release 4.0 */
    dout.dFormat = CF_STRUCT;
    dout.dArrayLength = AMS5_SIZE*nalm;
    dout.data.vptr = casTbl[i].almBuffer;
    memset(dout.dTag,0,TAG_NAME_SIZE);
  }
  else
  {
    dout.dFormat = CF_STRUCT;
    dout.dArrayLength = nalm;
    dout.data.vptr = casTbl[i].almBuffer;
    strncpy(dout.dTag,"AMSr4",TAG_NAME_SIZE);
  }
  sprintf(lnkDevName,"/%.32s/%.32s/*",casTbl[i].devContext,casTbl[i].devServer);

  amrb.almStartTime = casTbl[i].prevSnapshot.mrt + 1;
  amrb.almStopTime = casTbl[i].acqSnapshot.mrt;
  amrb.almMinSeverity = casTbl[i].SeverityLevel;
  amrb.reserved = 0;
  PrepOutgoingAMRB(&amrb,buf);

  din.dFormat = CF_STRUCT;
  din.dArrayLength = AMRB_SIZE;
  din.data.vptr = buf;
  din.dTag[0] = 0;

  lnkID = AttachLinkEx(lnkDevName,"ALARMSEXT",&dout,&din,CA_READ,1000,ProcessAlarms,CM_SINGLE,i+1);
  if (lnkID < 0)
  {
    casTbl[i].pending = 0;
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    sprintf(str,"Alarm Info Link to %.128s : %.32s",lnkDevName,cc2str(-lnkID));
    feclog(str);
    logproblem(str);
  }
}
AMS *findAlarmInList(CAMS *cams,UINT32 almCode,char *almTag)
{
  while (cams != NULL)
  {
    if (cams->ams->alarmCode == almCode) 
    { /* code matches */
      if (almTag == NULL) return cams->ams; /* tag unimportant */
      if (!strncmp(cams->ams->alarmTag,almTag,ALARM_TAG_SIZE)) return cams->ams;
    }
    /* not found, keep looking */
    cams = cams->nxt;
  }
  /* came up empty */
  return NULL;
}
void ClearServerDownAlarm(int i)
{
  CAMS *cams = NULL;
  AMS *ams = NULL;
  UINT32 st = time(NULL);
  int n;

  if (casTbl[i].markedAsDown != TRUE) return;
  feclog("clear server down alarm for %s",casTbl[i].devServer);
  casTbl[i].markedAsDown = FALSE;
  if (WaitForSystemMutex(hAmsTblMutex,-1) != 0) return;
  if ((ams=findAlarmInList(casTbl[i].almList,NOT_RESPONDING,TAG_NOT_RESPONDING)) != NULL)
  { /* fix the start time */
    st = ams->starttime;
  }
  for (cams=casTbl[i].almList; cams != NULL; cams=cams->nxt)
  {
    for (n=0; n<cams->nalms; n++) 
    { /* make sure any references are de-activated */
      if (cams->ams[n].alarmCode == NOT_RESPONDING)
        cams->ams[n].descriptor |= almDISABLED;
    }
  }
  AppendGeneratedAlarm(i,casTbl[i].AlarmSystem,ALMSYSTEM_INFO,almTERMINATE,st);
  AppendGeneratedAlarm(i,ALMSYSTEM_INFRASTRUCTURE,casTbl[i].AlarmLevel,almTERMINATE,st);
  ReleaseMutex(hAmsTblMutex);
}
int EmptyAlarmTable(int i,int reacquire)
{
  CAMS *cams = NULL;
  int n, actionTaken = FALSE;
  double ts = MakeDataTimeStamp();

  if (i < 0 || i >= casTblSiz) return invalid_index;
  casTbl[i].adsReacquire = reacquire; /* maybe something has changed ? */
  for (cams=casTbl[i].almList; cams != NULL; cams=cams->nxt)
  {
    for (n=0; n<cams->nalms; n++) 
    {
      if (cams->ams[n].alarmCode == NOT_RESPONDING) continue;
      if (isDuplicatedAlarm(&(cams->ams[n]),casTbl[i].almList)) continue;
      if (!(cams->ams[n].descriptor & almTERMINATE)) 
      {
        cams->ams[n].timestamp = (SINT32)ts;
        cams->ams[n].timestampUSec = 0;
        actionTaken = TRUE;
      }
      cams->ams[n].descriptor |= almTERMINATE;
    }
    if (actionTaken) 
    {
      cams->stoptime = (SINT32)ts;
      casTbl[i].timestamp = (time_t)ts;
    }
  }
# ifdef CAS2DOOCS
  SendStartupToDoocs(casTbl[i].devServer,casTbl[i].AlarmSystem);
# endif
  return 0;
}
void TerminateDanglingAlarms(int i)
{
  CAMS *cams = NULL, *pcams = NULL;
  AMS *ams = NULL;
  time_t t = time(NULL);
  int n, cc = 0;
  char *srv = "unknown";
  
  if (i < 0 || i >= casTblSiz) ccerr(invalid_index);
  for (cams=casTbl[i].almList; cams != NULL; cams=cams->nxt)
  {
    for (n=0; n<cams->nalms; n++)
    {
      if (!(cams->ams[n].descriptor & almTERMINATE))
      { /* not already terminated ! */
        if (t > (time_t)(cams->ams[n].timestamp + 2*DEFAULT_ALARM_HEARTBEAT))
        { /* no new heartbeat or anything for a long time (~40 minutes)! */
          if (isDuplicatedAlarm(&(cams->ams[n]),casTbl[i].almList)) continue;
          cams->ams[n].descriptor |= almDISABLED;
          goto generateTermination;
        }
      }
    }
  }
  ccerr(0);
generateTermination:
  if (cams == NULL) ccerr(code_failure);
  srv = casTbl[i].devServer;
  feclog("Generate alarm termination for %.32s",srv);
  if ((ams=(AMS *)calloc(1,sizeof(AMS))) == NULL) ccerr(out_of_local_memory);
  if ((pcams=(CAMS *)calloc(1,sizeof(CAMS))) == NULL) ccerr(out_of_local_memory);
  memcpy(ams,&(cams->ams[n]),sizeof(AMS));
  ams->timestamp = t;
  ams->descriptor = almTERMINATE;
  strncpy(ams->alarmTag,"dangling alarm",ALARM_TAG_SIZE);
  pcams->nalms = 1;
  pcams->stoptime = t;
  pcams->starttime = cams->stoptime-1;
  pcams->ams = ams;
  pcams->nxt = casTbl[i].almList;
  casTbl[i].almList = pcams;
  casTbl[i].timestamp = t;
err:
  if (cc) feclog("TerminateDanglingAlarms for %.32s : %.32s",srv,cc2str(cc));
  return;
}
void GenerateAndForwardAlarm(AMS *ams,int almsys)
{
  CAMS *cams = NULL;
  AMS *nams = NULL;
  int cc = 0;
  if (almsys == 0) return;

  if ((nams=(AMS *)calloc(1,sizeof(AMS))) == NULL) ccerr(out_of_local_memory);
  if ((cams=(CAMS *)calloc(1,sizeof(CAMS))) == NULL) ccerr(out_of_local_memory);
  memcpy(nams,ams,sizeof(AMS));
  nams->alarmSystem = almsys;
  cams->stoptime = ams->timestamp;
  cams->starttime = cams->stoptime-1;
  cams->ams = nams;
  cams->nxt = casTbl[i].almList;
  casTbl[i].almList = cams;
err:
  if (cc)
  {
    if (nams != NULL) free(nams);
    if (cams != NULL) free(cams);
  }
  return;
}
void AppendGeneratedAlarm(int i,int almsys,int almsev,int almdsc,UINT32 starttime)
{
  CAMS *cams = NULL;
  AMS *ams = NULL;
  time_t t = time(NULL);
  CasActLstEntry *ca;
  int cc = 0;
  char *srv = "unknown";
  
  if (i < 0 || i >= casTblSiz) ccerr(invalid_index);
  if ((ams=(AMS *)calloc(1,sizeof(AMS))) == NULL) ccerr(out_of_local_memory);
  if ((cams=(CAMS *)calloc(1,sizeof(CAMS))) == NULL) ccerr(out_of_local_memory);
  srv = casTbl[i].devServer;
  strncpy(ams->server,casTbl[i].devServer,EXPORT_NAME_SIZE);
  strncpy(ams->device,casTbl[i].devServer,EXPORT_NAME_SIZE);
  strncpy(ams->alarmTag,TAG_NOT_RESPONDING,ALARM_TAG_SIZE);
  ams->alarmCode = NOT_RESPONDING;
  ams->timestamp = t;
  ams->timestampUSec = 0;
  ams->alarmMask = getPrimaryRegionForServer(casTbl[i].devServer);
  ams->alarmDataFormat = BFMT(CF_NULL);
  ams->alarmDataArraySize = 0;
  ams->severity = almsev;
  ams->descriptor = (BYTE)almdsc;
  ams->alarmSystem = almsys;
  ams->starttime = starttime;
  ams->starttimeUSec = 0;

  cams->nalms = 1;
  cams->stoptime = t;
  cams->starttime = cams->stoptime-1;
  cams->ams = ams;
  cams->nxt = casTbl[i].almList;
  casTbl[i].almList = cams;
  casTbl[i].timestamp = t;

  for (ca=casTbl[i].actionList; ca != NULL; ca=ca->nxt)
  {
    ca->retry = 3;
    if (ca->alarmCode == NOT_RESPONDING)
    {
      ca->ams = ams;
	    TakeAction(ca);
    }
  }
err:
  if (cc) 
  {
    feclog("AppendGeneratedAlarm for %.32s : %.32s",srv,cc2str(cc));
    if (ams != NULL) free(ams);
    if (cams != NULL) free(cams);
  }
  return;
}
ADS adsSrvDown = 
{ 
  TAG_NOT_RESPONDING, 
  NOT_RESPONDING, 
  0x0, 
  ALMSYSTEM_INFRASTRUCTURE, 
  ALMSYSTEM_INFO, 
  BFMT(CF_NULL), 0, 
  0, 0,
  TEXT_NOT_RESPONDING,
  "Front end computer name which is not responding",
  "not applicable",
  "not applicable"
};
void GenerateServerDownAlarm(int i)
{
  AMS *ams = NULL;
  time_t t = time(NULL), mrt = 0;
  UINT32 st = (UINT32)t;
  int isInList = 0, needsHeartbeat = 0;
  BYTE descriptor = 0;
  
  if (WaitForSystemMutex(hAmsTblMutex,-1) != 0) return;
  if ((ams=findAlarmInList(casTbl[i].almList,NOT_RESPONDING,TAG_NOT_RESPONDING)) != NULL)
  {
    isInList = TRUE; /* found it */
    st = ams->starttime;
    if ((time_t)ams->timestamp > mrt) 
    {
      mrt = ams->timestamp;
      descriptor = ams->descriptor;
    }
  }
  /* check most recent time and descriptor at that time  */
  if ((descriptor & almTERMINATE) == almTERMINATE) 
  {
    isInList = FALSE;
    st = (UINT32)t;
  }
  needsHeartbeat = (t < mrt + DEFAULT_ALARM_HEARTBEAT) ? FALSE : TRUE;
  casTbl[i].markedAsDown = TRUE;
  if (isInList)
  {
    if (needsHeartbeat)
    {
      feclog("Generate server down heartbeat for %.32s",casTbl[i].devServer);
      AppendGeneratedAlarm(i,casTbl[i].AlarmSystem,ALMSYSTEM_INFO,almHEARTBEAT,st);
      AppendGeneratedAlarm(i,ALMSYSTEM_INFRASTRUCTURE,casTbl[i].AlarmLevel,almHEARTBEAT,st);
    }
    ReleaseMutex(hAmsTblMutex);
    return;
  }
  feclog("Generate server down alarm for %.32s",casTbl[i].devServer);
  AppendGeneratedAlarm(i,casTbl[i].AlarmSystem,ALMSYSTEM_INFO,almNEWALARM,st);
  AppendGeneratedAlarm(i,ALMSYSTEM_INFRASTRUCTURE,casTbl[i].AlarmLevel,almNEWALARM,st);
  ReleaseMutex(hAmsTblMutex);
  return;
}
void CheckAlarms(int id, int cc)
{
  int i=id-1, k;
  time_t mrt;
  DTYPE dout;
  char lnkDevName[128];
  CasTblEntry *cl = &casTbl[i];
  
  if (i < 0 || i >= casTblSiz) 
  {
    if (tineDebug || casdebug) dbglog("CheckAlarms: called with invalid index %d (list contains %d entries)",i,casTblSiz);
    feclog("CheckAlarms: called with invalid index %d (list contains %d entries)",i,casTblSiz);
    return;
  }
  if (IsOffLine(cl)) 
  {
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  if (tineDebug > 1) dbglog("CheckAlarms: /%.32s/%.32s/* : incoming",casTbl[i].devContext,casTbl[i].devServer);
  if (g_facilityNotRunning && casTbl[i].offline == CAS_STS_IGNORE_NOTRUNNING) 
  {
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  k = casTbl[i].linkId;
  if (cc != 0) 
  {
    if (cc == address_unknown || cc == non_existent_elem || cc == non_existent)
    {
      casTbl[i].offline = CAS_STS_REMOVE;
      casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
      feclog("error checking alarms for /%.32s/%.32s : %.32s -> mark as off-line and close connection",
        casTbl[i].devContext,casTbl[i].devServer,cc2str(cc));
      CloseLink(k);
      return;
    }
    if (cc == illegal_read_write && !casTbl[i].legacy)
    {
      CloseLink(casTbl[i].linkId);
      casTbl[i].legacy = TRUE;
      casTbl[i].linkId = i + 1;
      dout.dFormat = CF_LONG;  /* CAS_SNAPSHOT is 6 longs (modern) or 5 longs (legacy) */
      dout.dArrayLength = 5;
      dout.dTag[0] = 0;
      dout.data.lptr = (SINT32 *)&casTbl[i].thisSnapshot;
      sprintf(lnkDevName,"/%.32s/%.32s/*",casTbl[i].devContext,casTbl[i].devServer);
      casTbl[i].linkId = AttachLinkEx(lnkDevName,"NALARMS",&dout,NULL,CA_READ,1000,CheckAlarms,CM_REFRESH,i+1);
      if (tineDebug || casdebug) dbglog("CheckAlarms: change %.128s to legacy server",lnkDevName);
      return;
    }
    casTbl[i].linkStatus = cc;
    mrt = casTbl[i].prevSnapshot.mrt;
    memset(&casTbl[i].prevSnapshot,0,sizeof(ALM_SNAPSHOT));
    casTbl[i].prevSnapshot.mrt = mrt;
    casTbl[i].pending = 0;
    conTbl[k]->dtimestamp = 0; /* could be stuck in neutral if timestamp correction has gone back in time */
    if (casTbl[i].linkTimeoutCounter++ > TIMEOUT_THRESHOLD) 
    {
      GenerateServerDownAlarm(i);
      EmptyAlarmTable(i,FALSE);
      casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    }
    if (tineDebug || casdebug) dbglog("CheckAlarms: /%.32s/%.32s/* : %.32s",casTbl[i].devContext,casTbl[i].devServer,cc2str(cc));
    return;
  }
  if (conTbl[k]->tineProtocol < 6) casTbl[i].legacy = TRUE;
  if (casTbl[i].linkStatus != 0) 
  { /* server is back */
    feclog("Reconnected to server %.32s",casTbl[i].devServer);
    if (tineDebug || casdebug) dbglog("CheckAlarms: Reconnected to server %.32s",casTbl[i].devServer);
    ClearServerDownAlarm(i);
  }
  casTbl[i].linkStatus = cc;
  casTbl[i].linkTimeoutCounter = 0;
  if (casTbl[i].thisSnapshot.nalm == 0) casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
  if (casTbl[i].thisSnapshot.nalm > 0 && casTbl[i].thisSnapshot.hsv == 0) 
  {
    if (tineDebug || casdebug) dbglog("CheckAlarms: alarms from %.32s have severity = 0",casTbl[i].devServer);
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  if (!memcmp(&casTbl[i].thisSnapshot,&casTbl[i].prevSnapshot,sizeof(ALM_SNAPSHOT))) 
  {
    if (tineDebug || casdebug) dbglog("CheckAlarms: %.32s : current alarm set = previous alarm set",casTbl[i].devServer);
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  if (casTbl[i].pending) 
  { /* don't react if still acquiring previous set */
    feclog("acquire alarms for %.32s postponed due to %d pending",casTbl[i].devServer,casTbl[i].pending);
    if (tineDebug || casdebug) dbglog("CheckAlarms: acquire alarms for %.32s postponed due to %d pending",casTbl[i].devServer,casTbl[i].pending);
    return;
  }
  if (casTbl[i].thisSnapshot.hsv > 15)
  {
    if (tineDebug || casdebug) dbglog("CheckAlarms: %.32s : rejecting corrupt alarm severities",casTbl[i].devServer);
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
    return;
  }
  if (casTbl[i].thisSnapshot.mrt > casTbl[i].prevSnapshot.mrt) 
  {
    if (tineDebug || casdebug) dbglog("CheckAlarms: new alarm information for %.32s",casTbl[i].devServer);
    casTbl[i].acqSnapshot = casTbl[i].thisSnapshot;
    AcquireAlarmInfo(i);
  }
  else
  {
    if (tineDebug || casdebug) dbglog("CheckAlarms: %.32s : current alarm set not more recent than previous alarm set",casTbl[i].devServer);
    casTbl[i].hasAcquiredInitialAlarmSet = TRUE;
  }
  return;
}
int findHighestSeverity(CAMS *cams)
{
  int n,hsv=0;

  for (n=0; n<cams->nalms; n++)
  {
    if (cams->ams[n].severity > hsv) hsv = cams->ams[n].severity;
  }
  return hsv;
}
void caseqm_bkg(void)
{
  int i,k,n,hsv;
  time_t t=time(NULL);
  CAMS *cams, *pcams, *ncams;
  static time_t lastAcquired = 0;

  if (needToAcquireAlarmDefs && t > lastAcquired + 10)
  {
    n = 0;
    for (i=0; i<casTblSiz; i++)
    {
      if (!casTbl[i].adsListAcquired) getAlarmDefinitions(&casTbl[i]);
      else n++;
    }
    if (n == casTblSiz) needToAcquireAlarmDefs = FALSE; /* all done */
    lastAcquired = t;
  }

  /* could generate server down alarms this way: */
  /* ClearAlarm(CASEQM_TAG,-1); */
 
  /* is it time to archive alarms and/or free memory ? */
  if (WaitForSystemMutex(hAmsTblMutex,-1) != 0) 
  {
    feclog("caseqm_bkg: AMS Table mutex unavailable !");
    return;
  }
  for (i=0; i<casTblSiz; i++)
  {
    TerminateDanglingAlarms(i);
#   ifdef CAS2DOOCS
    SendHeartbeatToDoocs(casTbl[i].devServer,casTbl[i].AlarmSystem);
#   endif
    for (cams=casTbl[i].almList,pcams=cams; cams != NULL; )
    {
      for (k=0; k<cams->nalms; k++)
      {
        if ((t > cams->starttime + casTbl[i].Retention) &&
            (cams->ams[k].descriptor == almNEWALARM))
        {
           feclog("CAS: long term new alarm : %.32s %.64s %.32s",
             casTbl[i].devServer,cams->ams[k].device,cams->ams[k].alarmTag);
           feclog("CAS: time now %d > %d + %d",
             t,cams->starttime,casTbl[i].Retention);
           /* long-term NEW alarm with no heartbeat ! */
           cams->stoptime = cams->starttime; 
           /* generate terminated alarm ? */
        }
      }
      if (t > cams->stoptime + casTbl[i].Retention) 
      { /* free memory and archive (Retention typically 2 hours) */
        if ((hsv=findHighestSeverity(cams)) >= MIN_ARCHIVE_LEVEL)
        {
          ArchiveAMS(cams->nalms,cams->ams,casTbl[i].devServer,casTbl[i].arcExt,cams->stoptime);
        }
        free(cams->ams);
        /* join ... */
        if (cams == casTbl[i].almList) 
        { /* first in the list */
          casTbl[i].almList = cams->nxt;
        }
        else
        { /* somewhere down the chain ... */
          pcams->nxt = cams->nxt;
        }
        ncams = cams->nxt;
        /* free ... */
        free(cams);
        /* re-assign ... */
        cams = ncams;
        continue;
      }
      pcams = cams;
      cams = cams->nxt;
    }
  }
  ReleaseMutex(hAmsTblMutex);
  /* do outside of the the hAmsTblMutex as getAlarmDefinitionsFromServer might call callbacks ... */
  for (i=0; i<casTblSiz; i++)
  {
    if (casTbl[i].adsReacquire && casTbl[i].linkStatus == 0) 
    {
      feclog("signal to reacquire alarm definitions for %.32s",casTbl[i].devServer);
      getAlarmDefinitionsFromServer(&casTbl[i]);
      casTbl[i].adsReacquire = FALSE;
    }
  }
  if (gCasInitializedTs == 0) 
  {
    for (i=0; i<casTblSiz; i++) if (!casTbl[i].hasAcquiredInitialAlarmSet) break;
    if (i == casTblSiz) gCasInitializedTs = time(NULL);
  }
  return;
}
time_t getBaseTime(void)
{
  int i;
  time_t ts = gCasInitializedTs;
  for (i=0; i<casTblSiz; i++) 
    if (casTbl[i].timestampDisabled > ts) ts = casTbl[i].timestampDisabled;
  return ts;
}
int findAlarmSystemEntry(int start, int row, int column)
{
  int i;
  for (i=start; i<nAlarmSystems; i++)
  {
    if (AlarmSystems[i].row != row) continue;
    if (AlarmSystems[i].column != column) continue;
    return i;
  }
  return -1;
}
void sortAlarmSystemList(void)
{
  int i, k, r, c;
  int rmin = 1000, rmax = 0, cmin = 1000, cmax = 0;
  CasAlmSysTblEntry as;
  for (i=0; i<nAlarmSystems; i++)
  {
    if (AlarmSystems[i].row > rmax) rmax = AlarmSystems[i].row;
    if (AlarmSystems[i].row < rmin) rmin = AlarmSystems[i].row;
    if (AlarmSystems[i].column > cmax) cmax = AlarmSystems[i].column;
    if (AlarmSystems[i].column < cmin) cmin = AlarmSystems[i].column;
  }
  r = rmin; c = cmin; i = 0;
  while (i<nAlarmSystems)
  {
    if ((k=findAlarmSystemEntry(i,r,c)) < 0)
    { /* not found, move along */
      if (++r > rmax)
      {
        r = rmin;
        if (++c > cmax) break;
      }
      continue;
    }
    if (k != i)
    { /* swap them */
      as = AlarmSystems[i];
      AlarmSystems[i] = AlarmSystems[k];
      AlarmSystems[k] = as;
    }
    i++;
  }
}

int nRunningCriteria = 0;
RunningCriteria *runningCriteria = NULL;
int getRunningCriteria(char *fn)
{
  FILE *fp = NULL;
  int cc=0,done=0,i=0,fmt,siz;
  int key_col,fmt_col,siz_col,low_col,hi_col,tgt_col;
  char str[256],hdr[256],scratch[64];

  if ((nRunningCriteria=GetDataBaseSize(fn)) < 0) ccerr(-nRunningCriteria);
  if (nRunningCriteria == 0) ccerr(nRunningCriteria);
  if (runningCriteria != NULL) free(runningCriteria);
  if ((runningCriteria=(RunningCriteria *)calloc(nRunningCriteria,sizeof(RunningCriteria))) == NULL) ccerr(out_of_server_memory);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((key_col=findcol(hdr,"KEYWORD")) < 0) ccerr(no_such_column);
      fmt_col = findcol(hdr,"FORMAT");
      siz_col = findcol(hdr,"SIZE");
      low_col = findcol(hdr,"THRESHOLD_LOW");
      hi_col  = findcol(hdr,"THRESHOLD_HIGH");
      tgt_col = findcol(hdr,"TARGET");
      done = 1; /* read header info */
      continue;
    }
    fmt = CF_FLOAT; siz = 1; /* default values */
    strncpy(runningCriteria[i].key,colptr(key_col,str),DEVICE_NAME_SIZE);
    if (fmt_col >= 0) 
    {
      strncpy(scratch,colptr(fmt_col,str),16);
      fmt = GetFormatFromString(scratch);
      if (!isSimpleFormat((short)fmt)) continue;
    }
    runningCriteria[i].fmt = fmt;
    if (siz_col >= 0)
    {
      strncpy(scratch,colptr(siz_col,str),16);
      siz = atoi(scratch);
      if (siz < 1) continue;
    }
    runningCriteria[i].siz = siz;
    if (low_col >= 0)
    {
      strncpy(scratch,colptr(low_col,str),64);
      runningCriteria[i].chk = CHK_LOW;
    }
    else if (hi_col >= 0)
    {
      strncpy(scratch,colptr(hi_col,str),64);
      runningCriteria[i].chk = CHK_HIGH;
    }
    else if (tgt_col >= 0)
    {
      strncpy(scratch,colptr(tgt_col,str),64);
      runningCriteria[i].chk = CHK_TGT;
    }
    else
    { /* no threshold or target given ! */
      continue;
    }
    switch (fmt)
    {
      case CF_BYTE:
        runningCriteria[i].siz = 1;
        *(BYTE *)runningCriteria[i].cmp = (BYTE)atoi(scratch); break;
      case CF_INT16:
        runningCriteria[i].siz = 1;
        *(SINT16 *)runningCriteria[i].cmp = (SINT16)atoi(scratch); break;
      case CF_INT32:
        runningCriteria[i].siz = 1;
        *(SINT32 *)runningCriteria[i].cmp = (SINT32)atoi(scratch); break;
      case CF_FLOAT:
        runningCriteria[i].siz = 1;
        *(float *)runningCriteria[i].cmp = (float)atof(scratch); break;
      case CF_DOUBLE:
        runningCriteria[i].siz = 1;
        *(double *)runningCriteria[i].cmp = (double)atof(scratch); break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        runningCriteria[i].siz = 1;
        memcpy(runningCriteria[i].cmp,scratch,fmtsizeof(fmt)); break;
      case CF_TEXT:
        memcpy(runningCriteria[i].cmp,scratch,siz); break;
      default: continue;
    }
    /* safe to increment */
    i++;
  }
  nRunningCriteria = i;
err:
  if (fp) fclose(fp);
  feclog("running criteria file %s : %.32s",fn,cc ? cc2str(cc) : "found");
  return cc;
}

void cbIsRunning(int id,int cc)
{
  static int last_cc = 0;
  int i,running=TRUE;

  if (cc) return;
  for (i=0; i<nRunningCriteria && running; i++)
  {
    switch (runningCriteria[i].chk)
    {
      case CHK_LOW:
        switch (runningCriteria[i].fmt)
        {
          case CF_BYTE:
            if (*(BYTE *)runningCriteria[i].dat < *(BYTE *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_INT16:
            if (*(SINT16 *)runningCriteria[i].dat < *(SINT16 *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_INT32:
            if (*(SINT32 *)runningCriteria[i].dat < *(SINT32 *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_FLOAT:
            if (*(float *)runningCriteria[i].dat < *(float *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_DOUBLE:
            if (*(double *)runningCriteria[i].dat < *(double *)runningCriteria[i].cmp) running = FALSE; break;
        }
        break;
      case CHK_HIGH:
        switch (runningCriteria[i].fmt)
        {
          case CF_BYTE:
            if (*(BYTE *)runningCriteria[i].dat > *(BYTE *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_INT16:
            if (*(SINT16 *)runningCriteria[i].dat > *(SINT16 *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_INT32:
            if (*(SINT32 *)runningCriteria[i].dat > *(SINT32 *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_FLOAT:
            if (*(float *)runningCriteria[i].dat > *(float *)runningCriteria[i].cmp) running = FALSE; break;
          case CF_DOUBLE:
            if (*(double *)runningCriteria[i].dat > *(double *)runningCriteria[i].cmp) running = FALSE; break;
        }
        break;
      case CHK_TGT:
        if (memcmp(runningCriteria[i].dat,runningCriteria[i].cmp,runningCriteria[i].siz*fmtsizeof(runningCriteria[i].fmt)) != 0)
          running = FALSE;
    }
  }

  if (!running)
  {
    if (!g_facilityNotRunning)
    { /* just turned off */
      feclog("%.32s flagged as not running!",gDeviceContext);
      for (i=0; i<casTblSiz; i++)
      { /* remove relevant alarms */
        if (casTbl[i].offline == CAS_STS_IGNORE_NOTRUNNING) EmptyAlarmTable(i,FALSE);
      }
    }
    g_facilityNotRunning = TRUE;
  }
  else
  {
    if (g_facilityNotRunning)
    { /* just turned off: remove relevant alarms */
      feclog("%.32s flagged as running!",gDeviceContext);
    }
    g_facilityNotRunning = FALSE;
  }
}
void checkFacilityRunning(void)
{
  DTYPE dout;
  int i, k, ival;
  float fval;
  char dev[256],valStr[64],*condStr;

  getRunningCriteria(RUNNING_CRITERIA_FILE);

  memset(&dout,0,sizeof(DTYPE));

  for (i=0; i<nRunningCriteria; i++)
  {
    dout.dFormat = runningCriteria[i].fmt;
    dout.dArrayLength = runningCriteria[i].siz;
    dout.data.bptr = runningCriteria[i].dat;
    sprintf(dev,"/%.32s/%.64s",gDeviceContext,runningCriteria[i].key);
    runningCriteria[i].lnkId = recvNetGlobalEx(dev,&dout,cbIsRunning,UNASSIGNED_CALLBACKID);
    if (runningCriteria[i].lnkId < 0)
    {
      feclog("global keyword %.64s does not exist : not checking running criteria",runningCriteria[i].key);
      for (k=0; k<i; k++) CloseGlobalLink(runningCriteria[k].lnkId);
      break;
    }
    switch (runningCriteria[i].chk)
    {
      default:
      case CHK_LOW:
        condStr = "greater than"; break;
      case CHK_HIGH:
        condStr = "less than"; break;
      case CHK_TGT:
        condStr = "matches"; break;
    }
    switch (runningCriteria[i].fmt)
    {
      case CF_BYTE:
        ival = *(BYTE *)runningCriteria[i].cmp;
        sprintf(valStr,"%d",ival); break;
      case CF_INT16:
        ival = *(SINT16 *)runningCriteria[i].cmp;
        sprintf(valStr,"%d",ival); break;
      case CF_INT32:
        ival = *(SINT32 *)runningCriteria[i].cmp;
        sprintf(valStr,"%d",ival); break;
      case CF_FLOAT:
        fval = *(float *)runningCriteria[i].cmp;
        sprintf(valStr,"%g",fval); break;
      case CF_DOUBLE:
        fval = (float)(*(double *)runningCriteria[i].cmp);
        sprintf(valStr,"%g",fval); break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
      case CF_TEXT:
        strncpy(valStr,runningCriteria[i].cmp,64); break;
    }
    feclog("running criterion %.64s %s %.64s",runningCriteria[i].key,condStr,valStr);
  }  
  return;
}
void caseqm_ini(void)
{
  char lnkDevName[128],str[128],*c;
  int cc,i;
  DTYPE dout;
  time_t t = time(NULL);

  initRegionsDb();
  RegisterCaseqmProperties();
  remove("problems.log");

  registerCASStructs();
  getAdministrators(ADMINS_FILE);
  getSystemAliases(SYSALIAS_FILE);
  if ((cc=getAlarmSystemList(ALARM_SYSTEM_FILE)) != 0)
  {
    if (cc == no_such_file)
    { /* initial start with no pre-defined systems file ? */
      AlarmSystems = defAlarmSystems;
      nAlarmSystems = sizeof(defAlarmSystems)/sizeof(CasAlmSysTblEntry);
      putAlarmSystemsList(ALARM_SYSTEM_FILE,AlarmSystems,nAlarmSystems);
    }
    else
    {
      feclog("getAlarmSystemList: %.32s",cc2str(cc));
    }
  }
  sortAlarmSystemList();
  switch ((cc=getCasTbl(casListFile)))
  {
    case no_such_file:
      appendCasTbl(TRUE);
      break;
    case 0: 
      appendCasTbl(casTblSiz > 0 ? FALSE : TRUE);
      break;
    default:
      tineDebug = 1;
      feclog("CAS Initialization error: %.32s",cc2str(cc));
      exit(1);  /* stop right here */
  }
  /* use CAS device name = targetted device server */
  for (i=0; i<casTblSiz; i++)
  {
    casTbl[i].prevSnapshot.mrt = t - 3600;
    if (!strnicmp(gDeviceContext,casTbl[i].devContext,CONTEXT_NAME_SIZE))
      strncpy(str,casTbl[i].devServer,EXPORT_NAME_SIZE);
    else
      sprintf(str,"%.32s@%.32s",casTbl[i].devServer,casTbl[i].devContext);
    cc = RegisterDeviceName(CASEQM_TAG,str,i);
    if (cc != 0) feclog("cannot register %.32s as device name : %.32s",str,cc2str(cc));
  }
  /* or CAS device name = targetted sub system */
  for (i=0; i<nAlarmSubSystems; i++)
  {
    cc = RegisterDeviceName(CASEQM_TAG,AlarmSubSystems[i].name,casTblSiz+i);
    if (cc != 0) feclog("cannot register %.16s as device name : %.32s",
      AlarmSubSystems[i].name,cc2str(cc));
  }
  /* start the links to the targetted FECs */
  dout.dFormat = CF_LONG;  /* CAS_SNAPSHOT is 6 longs (modern) or 5 longs (legacy) */
  dout.dArrayLength = 6;
  for (i=0; i<casTblSiz; i++)
  {
    if (IsOffLine(&casTbl[i])) continue;
    dout.data.lptr = (SINT32 *)&casTbl[i].thisSnapshot;
    sprintf(lnkDevName,"/%.32s/%.32s/*",casTbl[i].devContext,casTbl[i].devServer);
    casTbl[i].linkId = AttachLinkEx(lnkDevName,"NALARMS",&dout,NULL,CA_READ,CAS_ACQ_RATE,CheckAlarms,CM_DATACHANGE,i+1);
    if (casTbl[i].linkId < 0)
    {
      cc = -casTbl[i].linkId;
      sprintf(str,"Alarm Snapshot Link to %.128s : %.32s",lnkDevName,cc2str(cc));
      feclog(str);
      logproblem(str);
    }
    else
    { /* everythings okay, so get the alarm definition information */
      if (getAlarmDefinitions(&casTbl[i]) != 0) needToAcquireAlarmDefs = TRUE;
    }
  }
  if ((c=GetExportedName(CASEQM_TAG)) != NULL)
    strncpy(casExportName,c,EXPORT_NAME_SIZE);
  else 
    strncpy(casExportName,"CAS",EXPORT_NAME_SIZE);
  printf("CAS services exported as %.32s\n",casExportName);

  if ((cc=InitMutex(&hSendMailMutex)) != 0) 
  {
    feclog("could not initialize send mail mutex : %.32s",cc2str(cc));
  }
  if ((cc=InitMutex(&hAmsTblMutex)) != 0) 
  {
    feclog("could not initialize AMS Table mutex : %.32s",cc2str(cc));
  }
  RegisterUserCommand("casdbg",NULL,&casdebug,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("aliases",dumpSysAlias,NULL,NULL,CA_READ);
  RegisterUserCommand("castable",dumpCasTbl,NULL,NULL,CA_READ);
  
  checkFacilityRunning();
# ifdef CAS2DOOCS
  initCas2Doocs();
  for (i=0; i<casTblSiz; i++)
  {
    SendStartupToDoocs(casTbl[i].devServer,casTbl[i].AlarmSystem);
  }
# endif
}
void caseqm_exi(void)
{
  int i;
  CasActLstEntry *ca=NULL, *canxt=NULL;
  CasDisableLstEntry *cd=NULL, *cdnxt=NULL;
  /* flush the archive queue, clear any 'down' alarms */
  for (i=0; i<casTblSiz; i++) 
  {
    ClearServerDownAlarm(i);
    casTbl[i].Retention = 0;
  }
  caseqm_bkg(); /* Retention = 0: will free the current alarms lists */
  for (i=0; i<casTblSiz; i++)
  {
    if (casTbl[i].linkId >= 0) CloseLink(casTbl[i].linkId);
    if (casTbl[i].adsList != NULL) free(casTbl[i].adsList);
    for (ca=casTbl[i].actionList; ca != NULL; ca=canxt)
    {
      canxt = ca->nxt;
      if (ca->din.data.vptr != NULL) free(ca->din.data.vptr);
      free(ca);
    }
    removeDisabledList(&casTbl[i]);
    if (casTbl[i].almSystemList != NULL) free(casTbl[i].almSystemList);
  }
}
/* return TRUE if ams is found previously in cams */
/* note: the most recent alarms are at the head of the linked list */
int isDuplicatedAlarm(AMS *ams, CAMS *pcams)
{
  CAMS *cams;
  int i;
  if (ams->descriptor & almINSTANT) return FALSE;
  for (cams=pcams; cams != NULL; cams=cams->nxt)
  {
    for (i=0; i<cams->nalms; i++)
    {
      if (ams == &(cams->ams[i])) return FALSE; /* obviously not duplicated */
      if (ams->descriptor & almDISABLED) return TRUE; /* disabled; */
      if (cams->ams[i].starttime != ams->starttime) continue;
      if (cams->ams[i].starttimeUSec != ams->starttimeUSec) continue;
      if (cams->ams[i].alarmCode != ams->alarmCode) continue;
      if (cams->ams[i].alarmSystem != ams->alarmSystem) continue;
      if (strncmp(cams->ams[i].server,ams->server,EXPORT_NAME_SIZE)) continue;
      if (cams->ams[i].device[0] == 0 && ams->device[0] != 0) continue;
      if (cams->ams[i].device[0] != 0 && ams->device[0] == 0) continue;
      if (strlen(ams->device) > 0 &&
          strncmp(cams->ams[i].device,ams->device,DEVICE_NAME_SIZE)) continue;
      ams->descriptor |= almDISABLED; /* disable it for future checks */
      return TRUE;
    }
  }
  return FALSE;
}
int numNonTerminatedAlarmsInList(int i,UINT32 start,UINT32 stop,int sev,int sys,int filterLast,time_t *ts)
{
  CAMS *cams;
  int j, n=0, showHidden=FALSE;
  int thisAlarmSystem,dftAlarmSystem = casTbl[i].AlarmSystem;
  time_t mrt = 0;

  if (sys & ALM_SYSTEM_HIDDEN)
  {
    showHidden = TRUE;
    dftAlarmSystem |= ALM_SYSTEM_HIDDEN;
  }
  for (cams=casTbl[i].almList; cams != NULL; cams=cams->nxt)
  {
    if ((UINT32)cams->stoptime >= start && (UINT32)cams->starttime <= stop)
    {
      for (j=0; j<cams->nalms; j++)
      {
        thisAlarmSystem = cams->ams[j].alarmSystem;
        if (thisAlarmSystem == 0) /* wasn't set ! */
          thisAlarmSystem = dftAlarmSystem;
        if (sys > -1 && thisAlarmSystem != sys) continue;
        if (cams->ams[j].descriptor & almTERMINATE) continue;
        if (filterLast && (cams->ams[j].descriptor & almDISABLED)) continue;
        if (filterLast && isDuplicatedAlarm(&(cams->ams[j]),casTbl[i].almList)) continue;
        if (cams->ams[j].severity >= sev) n++;
        if (cams->ams[j].timestamp > (UINT32)mrt) mrt = (time_t)cams->ams[j].timestamp;
      }
    }
  }
  if (ts != NULL) *ts = mrt;
  return n;
}
int numAlarmsInList(int i,UINT32 start,UINT32 stop,int sev,int sys,int filterLast,time_t *ts)
{
  CAMS *cams;
  int j, n=0, showHidden=FALSE;
  int thisAlarmSystem,dftAlarmSystem = casTbl[i].AlarmSystem;
  time_t mrt = 0;

  if (sys & ALM_SYSTEM_HIDDEN)
  {
    showHidden = TRUE;
    dftAlarmSystem |= ALM_SYSTEM_HIDDEN;
  }
  for (cams=casTbl[i].almList; cams != NULL; cams=cams->nxt)
  {
    if ((UINT32)cams->stoptime >= start && (UINT32)cams->starttime <= stop) 
    {
      if (sev == -1 && sys == -1) n += cams->nalms; /* sev == -1 effectively turns this off */
      else
      {
        for (j=0; j<cams->nalms; j++) 
        {
          thisAlarmSystem = cams->ams[j].alarmSystem;
          if (thisAlarmSystem == 0) /* wasn't set ! */
            thisAlarmSystem = dftAlarmSystem;
          if (sys > -1 && thisAlarmSystem != sys) continue;
          if (filterLast && (cams->ams[j].descriptor & almDISABLED)) continue;
          if (filterLast && (cams->ams[j].descriptor & almTERMINATE)) continue;
          if (cams->ams[j].descriptor & almSUPPRESS) continue;
          if (cams->ams[j].severity >= sev) n++;
          if (cams->ams[j].timestamp > (UINT32)mrt) mrt = (time_t)cams->ams[j].timestamp;
        }
      }
    }
  }
  if (ts != NULL) *ts = mrt;
  return n;
}
int getAlarmSystemNumberFromTag(char *tag)
{
  int i;
  for (i=0; i<nAlarmSystems; i++)
    if (!strncmp(tag,AlarmSystems[i].almsystem,ALARM_TAG_SIZE)) break;
  return i < nAlarmSystems ? AlarmSystems[i].number : -1;
}
char *getAlarmSystemTagFromNumber(int nr)
{
  int i;
  for (i=0; i<nAlarmSystems; i++) 
    if (AlarmSystems[i].number == nr) return AlarmSystems[i].almsystem;
  return "unknown";
}
ADS *getAlarmDefinition(CasTblEntry *caslst,UINT32 almCode)
{
  int k;
  for (k=0; k<caslst->nAdsList; k++)
  {
    if (almCode == caslst->adsList[k].alarmCode) return &caslst->adsList[k];
  }
  return NULL;
}
char *getAlarmText(CasTblEntry *caslst, AMS *ams)
{
  int k;
  for (k=0; k<caslst->nAdsList; k++)
  {
    if (ams->alarmCode == caslst->adsList[k].alarmCode)
    {
      if (strlen(caslst->adsList[k].alarmText) == 0) return NULL;
      return caslst->adsList[k].alarmText;
    }
  }
  return NULL;
}
int makeDisabledAMS(AMS *ams, CasTblEntry *caslst, char *dev, UINT32 almCode, char *usr, time_t ts)
{
  int k;
  ADS *ads;
  memset(ams,0,sizeof(AMS));
  ams->alarmCode = almCode;
  ams->alarmDataFormat = BFMT(CF_TEXT);
  ams->alarmDataArraySize = 16;
  strncpy(ams->server,caslst->devServer,EXPORT_NAME_SIZE);
  if (dev == NULL || strlen(dev) == 0) dev = "all devices";
  strncpy(ams->device,dev,DEVICE_NAME_SIZE);
  strncpy(ams->alarmData,usr,USERNAME_SIZE);
  ams->timestamp = (UINT32)ts;
  for (k=0; k<caslst->nAdsList; k++)
  {
    if (caslst->adsList[k].alarmCode == almCode)
    {
      ams->alarmSystem = caslst->adsList[k].alarmSystem;
      strncpy(ams->alarmTag,caslst->adsList[k].alarmTag,ALARM_TAG_SIZE);
      ams->severity = (BYTE)caslst->adsList[k].alarmSeverity;
      return 0;
    }
  }
  if (almCode == NOT_RESPONDING)
  {
    strncpy(ams->alarmTag,TAG_NOT_RESPONDING,ALARM_TAG_SIZE);
    ams->alarmSystem = ALMSYSTEM_INFRASTRUCTURE;
    ams->severity = caslst->AlarmLevel;
  }
  else if ((ads=getErrTblADS((short)almCode)) != NULL)
  {
    strncpy(ams->alarmTag,ads->alarmTag,ALARM_TAG_SIZE);
    ams->alarmSystem = ads->alarmSystem;
    ams->severity = (BYTE)ads->alarmSeverity;
  }
  else
  {
    ams->alarmSystem = caslst->AlarmSystem;
    sprintf(ams->alarmTag,"unknown alarm (%d)",almCode);
  }
  return 0;
}
void clearDisabledList(CasTblEntry *cl, char *dev, UINT32 almCode)
{
  int n;
  CAMS *cams;
  CasDisableLstEntry *cd=NULL;

  for (cams=cl->almList; cams != NULL; cams=cams->nxt)
  {
    for (n=0; n<cams->nalms; n++) 
    {
      if (dev != NULL && strlen(dev) > 0 && strncmp(dev,cams->ams[n].device,DEVICE_NAME_SIZE)) continue;
      if (cams->ams[n].alarmCode == almCode)
      {
        cams->ams[n].descriptor &= ~(almDISABLED|almSUPPRESS);
      }
    }
  }
}
void applyDisabledList(CasTblEntry *cl)
{
  int n;
  CAMS *cams;
  CasDisableLstEntry *cd=NULL;

  for (cd=cl->disabledList; cd != NULL; cd = cd->nxt)
  {
    for (cams=cl->almList; cams != NULL; cams=cams->nxt)
    {
      for (n=0; n<cams->nalms; n++) 
      {
        if (strlen(cd->device) > 0 && strncmp(cams->ams[n].device,cd->device,DEVICE_NAME_SIZE)) continue;
        if (cams->ams[n].alarmCode == cd->alarmCode)
        {
          cams->ams[n].descriptor |= (almDISABLED|almSUPPRESS);
        }
      }
    }
  }
  cl->timestampDisabled = time(NULL);
}
int appendToDisabledList(CasTblEntry *cl,char *dev,char *caller,int almCode,int type)
{
  CasDisableLstEntry *cd=NULL, *cdp, *cdn;
  int replaceWithALL = FALSE;
  if (dev != NULL && stricmp(dev,"ALL") == 0) dev = NULL;
  for (cd=cl->disabledList; cd != NULL; cd = cd->nxt)
  { /* already in list ? */
    if (cd->alarmCode != (UINT32)almCode) continue;
    if (strlen(cd->device) == 0) break; /* already disabled for ALL devices */
    if (dev == NULL || strlen(dev) == 0)
    { /* replace specific device disabling with ALL devices */
      replaceWithALL = TRUE;
      break;
    }
    if (!strncmp(dev,cd->device,DEVICE_NAME_SIZE)) break; /* exact match ! */
  }
  if (replaceWithALL)
  {
    for (cd=cl->disabledList, cdp=cd; cd != NULL; cd=cdn)
    { /* find and remove all current entries in list  */
      cdn = cd->nxt;
      if (cd->alarmCode == (UINT32)almCode)
      {
        if (cdp != cd) cdp->nxt = cd->nxt; else cdp = cl->disabledList = cd->nxt;
        free(cd); cd = cdp;
        if (cl->sizDisabledList > 0) cl->sizDisabledList--;
      }
      cdp = cd; /* mark the previous entry */
    }
    cd = NULL; /*add the ALL entry below */
  }
  if (cd == NULL)
  { /* no -> add to list ! */
    if ((cd=(CasDisableLstEntry *)calloc(1,sizeof(CasDisableLstEntry))) == NULL) return out_of_server_memory;
    cd->alarmCode = almCode;
    cd->timeDisabled = time(NULL);
    if (dev != NULL && strlen(dev) > 0) strncpy(cd->device,dev,DEVICE_NAME_SIZE);
    strncpy(cd->disabler,caller,USERNAME_SIZE);
    cd->disableType = type;
    cd->nxt = cl->disabledList;
    cl->disabledList = cd;
    cl->sizDisabledList++;
  }
  return 0;
}
int removeDisabledList(CasTblEntry *cl)
{
  CasDisableLstEntry *cd=cl->disabledList, *cdp;
  while (cd != NULL)
  { 
    clearDisabledList(cl,NULL,cd->alarmCode);
    cdp = cd->nxt;
    free(cd);
    cd = cdp;
  }
  cl->sizDisabledList = 0;
  cl->disabledList = NULL;
  cl->timestampDisabled = time(NULL);
  return 0;
}
int removeFromDisabledList(CasTblEntry *cl,char *dev,UINT32 almCode)
{
  CasDisableLstEntry *cd=NULL, *cdp;
  for (cd=cl->disabledList, cdp=cd; cd != NULL; cd=cd->nxt)
  { /* find it in list  */
    if (dev != NULL && strlen(dev) > 0 && strncmp(dev,cd->device,DEVICE_NAME_SIZE)) continue;
    if (cd->alarmCode == almCode)
    {
      clearDisabledList(cl,dev,almCode);
      if (cdp != cd) cdp->nxt = cd->nxt; else cl->disabledList = cd->nxt;
      free(cd);
      if (cl->sizDisabledList > 0) cl->sizDisabledList--;
      cl->timestampDisabled = time(NULL);
      return 0;
    }
    cdp = cd; /* mark the previous entry */
  }
  return 0;
}
int prepCASOutgoingAMS(CasTblEntry *caslst,BYTE *amsstream,AMS *amslst,char *stag,int length)
{
  int i,useAlarmText=FALSE;
  char *txt;
  XAMS *xams;
  XAMS5 *xams5;
  
  if (!strcmp(stag,"AMSr4") || !strcmp(stag,"AMS"))
  { /* just a re-hash */
    memcpy(amsstream,amslst,length*AMS_SIZE);
  }
  if (strlen(stag) == 0)
  { /* no tag! -> really old-fashioned ? */
    return prepOutgoingAMS(amsstream,amslst,stag,length);
  }
  if (!strcmp(stag,"XAMSr4"))
  { /* modern extended */
    xams = (XAMS *)amsstream;
    for (i=0; i<length; i++)
    {
      strncpy(xams[i].server,amslst[i].server,EXPORT_NAME_SIZE);
      strncpy(xams[i].device,amslst[i].device,DEVICE_NAME_SIZE);
      useAlarmText = FALSE;
      if ((txt=getAlarmText(caslst,&amslst[i])) != NULL)
      { /* replace almTag with almText */
        strncpy(xams[i].alarmText,txt,ALARM_TEXT_SIZE);
        useAlarmText = TRUE;
      }
      if (!useAlarmText) strncpy(xams[i].alarmText,amslst[i].alarmTag,ALARM_TAG_SIZE);
      xams[i].alarmCode = amslst[i].alarmCode;
      xams[i].timestamp = amslst[i].timestamp;
      xams[i].timestampUSec = amslst[i].timestampUSec;
      xams[i].starttime = amslst[i].starttime;
      xams[i].starttimeUSec = amslst[i].starttimeUSec;
      xams[i].alarmMask = amslst[i].alarmMask;
      xams[i].alarmSystem = amslst[i].alarmSystem;
      memcpy(xams[i].alarmData,amslst[i].alarmData,ALARM_DATA_SIZE);
      xams[i].alarmDataFormat = amslst[i].alarmDataFormat;
      xams[i].alarmDataArraySize = amslst[i].alarmDataArraySize;
      xams[i].severity = amslst[i].severity;
      xams[i].descriptor = amslst[i].descriptor;
    }
    return 0;
  }
  if (!strcmp(stag,"XAMS"))
  { /* legacy */
    xams5 = (XAMS5 *)amsstream;
    for (i=0; i<length; i++)
    {
      strncpy(xams5[i].server,amslst[i].server,EXPORT_NAME_SIZE);
      strncpy(xams5[i].device,amslst[i].device,DEVICE_NAME_SIZE);
      useAlarmText = FALSE;
      if ((txt=getAlarmText(caslst,&amslst[i])) != NULL)
      { /* replace almTag with almText */
        strncpy(xams5[i].alarmText,txt,40);
        useAlarmText = TRUE;
      }
      if (!useAlarmText) strncpy(xams5[i].alarmText,amslst[i].alarmTag,ALARM_TAG_SHORTSIZE);
      xams5[i].alarmCode = amslst[i].alarmCode;
      xams5[i].timestamp = amslst[i].timestamp;
      xams5[i].alarmMask = amslst[i].alarmMask;
      xams5[i].alarmSystem = amslst[i].alarmSystem;
      memcpy(xams5[i].alarmData,amslst[i].alarmData,8);
      xams5[i].alarmDataFormat = amslst[i].alarmDataFormat;
      xams5[i].alarmDataArraySize = amslst[i].alarmDataArraySize;
      xams5[i].severity = amslst[i].severity;
      xams5[i].descriptor = amslst[i].descriptor;
    }
    return 0;
  }
  return invalid_structure_tag;
}
int hasSubsystem(int i,int sys)
{
  int k;
# ifdef SCAN_ALL_ALARMS
  CAMS *cams;
  int dftAlarmSystem=0,thisAlarmSystem;
# endif

  if (i < 0 || i >= casTblSiz) return FALSE;
  if (sys < 0) return TRUE;
  if (sys == ALM_SYSTEM_SYSTEM || sys == ALM_SYSTEM_HARDWARE) return TRUE;
  if (casTbl[i].AlarmSystem == sys) return TRUE;
  for (k=0; k<casTbl[i].nAlmSystemList; k++)
  {
    if (casTbl[i].almSystemList[k] == sys) return TRUE;
  }
# ifdef SCAN_ALL_ALARMS
  dftAlarmSystem = casTbl[i].AlarmSystem;
  for (cams=casTbl[i].almList; cams != NULL; cams = cams->nxt)
  {
    for (k=0; k<cams->nalms; k++)
    {
      thisAlarmSystem = cams->ams[k].alarmSystem;
      if (thisAlarmSystem == 0) thisAlarmSystem = dftAlarmSystem;
      if (thisAlarmSystem == sys) return TRUE;
    }
  }
# endif
  return FALSE;
}
int getNonDecoratedServerAsDeviceNumber(char *devName)
{
  int i;
  for (i=0; i<casTblSiz; i++)
  {
    if (strnicmp(casTbl[i].devServer,devName,EXPORT_NAME_SIZE)) continue;
    return i;
  }
  return -1;
}
int caseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,almsysnr,i,k,m,n,cc = 0,len,nlen,ilen,siz,sev = 0,chkss;
  int addAlarmCode=0,addViewColumn=0,dftAlarmSystem=0,thisAlarmSystem,filterLast=0;
  int almcode=-1,isAllowed=FALSE,isstruct=FALSE,codeFound=TRUE,almsFound=FALSE,showHidden=FALSE;
  int amssize=AMS_SIZE;
  CasTblEntry *cl;
  CasDisableLstEntry *cd;
  CasDbItem *dbItm;
  CasAlmSysTblEntry *almSysItm;
  CasActionItem *actItm;
  CasActLstEntry *ca;
  time_t stoptime = time(NULL), starttime = gSystemStartupTime - 7200, mrt = 0, ts;
  short lcl_numalarms, lcl_nntalarms, lcl_disalarms, lcl_nalm[3];
  int lcl_ival[25], *da;
  CAMS *cams;
  BYTE *ptr, *daPtr;
  NAME64I *dnia;
  NAME64DBLDBL *dndd;
  char *c, caller[USERNAME_SIZE], devNameBuffer[256], *ctx, *srv, *dev=NULL;
  SINT32 *lptr;
  FecDataStruct fds;
  AMS ams;

  /* registered devices include the servers and their subsystems */
  /* additional usage of device name include the alarm system via number or name */
  /*    => begins with 'SYS#' or 'SYS:' */
  /* also (for disable comments) device name can be decorated with a device and alarm code */
  /*    => begins with server name followed by '/<device>:<code>' */
  if ((dev=strchr(devName,'/')) != NULL)
  { /* parse out the targeted device and code info */
    *dev++ = 0;
    if ((c=strchr(dev,':')) != NULL)
    {
      *c++ = 0;
      almcode = atoi(c);
    }
  }
  /* get device number from device name  */
  devnr = GetDeviceNumber(CASEQM_TAG,devName); /* -1 => ALL  */
  almsysnr = -1; /* ALL */
  if ((c=strstr(devName,"SYS#")) != NULL)
  {
    almsysnr = atoi(&c[4]);
    //if (almsysnr & ALM_SYSTEM_HIDDEN) showHidden = TRUE;
  }
  else if ((c=strstr(devName,"SYS:")) != NULL) 
  {
    almsysnr = getAlarmSystemNumberFromTag(&c[4]);
#   ifdef USE_THIS
    while (almsysnr == -1 && !showHidden)
    { /* only turn on HIDDEN if asking for SYS:ALL */
      if (c[4] == '.') 
      {
        almsysnr = getAlarmSystemNumberFromTag(&c[4]);
        showHidden = TRUE;
      }
      if (strnicmp(devName,"SYS:ALL",7) != 0) return non_existent;
      break;
    }
#   endif
  }
  almsysnr &= ~ALM_SYSTEM_HIDDEN;
  prpid = GetPropertyId(CASEQM_TAG,devProperty);

  switch (prpid)
  {
    case PRP_ISONLINE:
        if (devnr >= 0) 
        {
          if (dout->dFormat != CF_SHORT && dout->dFormat != CF_LONG) 
            return illegal_format;
          for (n=0,i=devnr; n<(int)dout->dArrayLength && i<casTblSiz; n++, i++)
          {
            k = (casTbl[i].offline || casTbl[i].markedAsDown) ? 0 : 1;
            if (dout->dFormat == CF_SHORT) 
              dout->data.sptr[n] = k;
            else 
              dout->data.lptr[n] = k;
          }
          dout->dArrayLength = n;
        }
        return 0;
    case PRP_REMOVEALARMS:
        if (access&CA_WRITE) 
        {
          strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
          if (GetFECAddr(caller,&fds) != 0) return unauthorised_action;
          return EmptyAlarmTable(devnr,TRUE);
        }
        return illegal_read_write;
    case PRP_ACKTIME:
        if (access&CA_WRITE) 
        {
          if (din->dArrayLength == 0) return illegal_read_write;
          if ((cc=GetValuesAsLong(din,(SINT32 *)&g_AckThreshold,1)) != 0) return cc;
        }
        if (dout->dArrayLength > 0)
        {
          return PutValuesFromLong(dout,(SINT32 *)&g_AckThreshold,1);
        }
        return 0;
    case PRP_ENABLE:
        if (devnr < 0 || devnr >= casTblSiz) return not_registered;
        cl = &casTbl[devnr];
        if (!(access&CA_WRITE)) return illegal_read_write;
#       ifdef NEED_ADMIN_TO_DISABLE
        strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
        if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
        /* if (!isAdministrator(caller)) return unauthorised_action; */
#       endif
        if (din->dArrayLength == 0) 
        { /* all disabled alarms for this server */
          removeDisabledList(cl);
          feclog("CAS: ALL alarms on server %.32s have been enabled by %.16s",cl->devServer,caller);
        }
        else
        { /* a specified list of alarm codes */
          if ((daPtr=(BYTE *)alloca(din->dArrayLength*sizeof(NAME64I))) == NULL) return out_of_server_memory;
          da = (SINT32 *)daPtr;
          dnia = (NAME64I *)daPtr;
          if (din->dFormat == CF_NAME64I)
          {
            memcpy(dnia,din->data.bptr,din->dArrayLength*sizeof(NAME64I));
            for (i=0; i<(int)din->dArrayLength; i++) 
            {
              if (dnia[i].ival == 0) continue;
              if ((cc=removeFromDisabledList(cl,dnia[i].name,dnia[i].ival)) != 0) return cc;
              feclog("CAS: alarms of type %.64s (code %d) on server %.32s have been enabled by %.16s for device %s",
                getAlarmMessageFromDefinition(devnr,dnia[i].ival),
                dnia[i].ival,cl->devServer,caller,
                (strlen(dnia[i].name)) ? dnia[i].name : "ALL devices");
            }
          }
          else
          {
            if ((cc=GetValuesAsLong(din,da,din->dArrayLength)) != 0) return cc;
            for (i=0; i<(int)din->dArrayLength; i++) 
            {
              if (da[i] == 0) continue;
              if ((cc=removeFromDisabledList(cl,dev,(UINT32)da[i])) != 0) return cc;
              feclog("CAS: alarms of type %.64s (code %d) on server %.32s have been enabled by %.16s for device %s",
                getAlarmMessageFromDefinition(devnr,da[i]),
                da[i],cl->devServer,caller,
                (dev != NULL && strlen(dev)) ? dev : "ALL devices");
            }
          }
        }
        writeCASDisabledList(devnr);
        return 0;
    case PRP_DISABLE:
        if ((n=MAX(din->dArrayLength,dout->dArrayLength)) == 0) return dimension_error;
        if ((daPtr=(BYTE *)alloca(n*sizeof(NAME64DBLDBL))) == NULL) return out_of_server_memory;
        da = (SINT32 *)daPtr;
        dnia = (NAME64I *)daPtr;
        if (access&CA_WRITE) 
        { /* setting disabled alarms ... */
          strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
#         ifdef NEED_ADMIN_TO_DISABLE
          if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
          /* if (!isAdministrator(caller)) return unauthorised_action; */
#         endif
          if (devnr < 0 || devnr >= casTblSiz) return not_registered;
          cl = &casTbl[devnr];
          if (din->dArrayLength == 0) return illegal_read_write;
          switch (din->dFormat)
          {
            case CF_NAME64DBLDBL:
              dndd = (NAME64DBLDBL *)daPtr;
              memcpy(dndd,din->data.bptr,din->dArrayLength*fmtsizeof(CF_NAME64DBLDBL));
              for (i=0; i<(int)din->dArrayLength; i++) 
              {
                if ((almcode = (int)dndd[i].d1val) == 0) continue;
                if ((cc=appendToDisabledList(cl,dndd[i].name,caller,almcode,(int)dndd[i].d2val)) != 0) return cc;
                feclog("CAS: current active alarms of type %.64s (code %d) on server %.32s have been disabled by %.16s for device %.64s",
                  getAlarmMessageFromDefinition(devnr,almcode),
                  almcode,cl->devServer,caller,
                  (strlen(dndd[i].name)) ? dndd[i].name : "ALL devices");
              }
              break;
            case CF_NAME64I:
              memcpy(dnia,din->data.bptr,din->dArrayLength*fmtsizeof(CF_NAME64I));
              for (i=0; i<(int)din->dArrayLength; i++) 
              {
                if ((almcode=dnia[i].ival) == 0) continue;
                if ((cc=appendToDisabledList(cl,dnia[i].name,caller,almcode,0)) != 0) return cc;
                feclog("CAS: alarms of type %.64s (code %d) on server %.32s have been disabled by %.16s for device %.64s",
                  getAlarmMessageFromDefinition(devnr,almcode),
                  almcode,cl->devServer,caller,
                  (strlen(dnia[i].name)) ? dnia[i].name : "ALL devices");
              }
              break;
            default:
              if ((cc=GetValuesAsLong(din,da,n)) != 0) return cc;
              for (i=0; i<(int)din->dArrayLength; i++) 
              {
                if ((almcode=da[i]) == 0) continue;
                if ((cc=appendToDisabledList(cl,dev,caller,almcode,0)) != 0) return cc;
                feclog("CAS: alarms of type %.64s (code %d) on server %.32s have been disabled by %.16s for device %64s",
                  getAlarmMessageFromDefinition(devnr,almcode),
                  almcode,cl->devServer,caller,
                  (dev != NULL && strlen(dev)) ? dev : "ALL devices");
              }
              break;
          }
          applyDisabledList(cl);
          writeCASDisabledList(devnr);
        }
        if (dout->dArrayLength > 0)
        { /* prepare the return list */
          if (!strcmp(dout->dTag,"AMSr4")) isstruct = TRUE;
          if (devnr < 0 || devnr >= casTblSiz)
          { /* either ALL or a specific subsystem */
            if (!isstruct) return illegal_format;
            for (i=0,n=0; i<casTblSiz && n<(int)dout->dArrayLength; i++) 
            {
              cl = &casTbl[i];
              if (almsysnr != -1 && !hasSubsystem(i,almsysnr)) continue;
              //if (almsysnr != cl->AlarmSystem && almsysnr != -1) continue;
              if (devnr > 0 && strncmp(devName,cl->subSystem,SUBSYSTEM_NAME_SIZE)) continue;
              for (cd=cl->disabledList; cd != NULL; cd=cd->nxt)
              {
                makeDisabledAMS(&ams,cl,cd->device,cd->alarmCode,cd->disabler,cd->timeDisabled);
                memcpy(dout->data.bptr+n*AMS_SIZE,&ams,AMS_SIZE);
                if (++n >= (int)dout->dArrayLength) break;
              }
            }
            dout->dArrayLength = n;
            return 0;
          }
          /* must be a device server ... */
          cl = &casTbl[devnr];
          for (cd=cl->disabledList,i=0; cd != NULL && i<(int)dout->dArrayLength; cd=cd->nxt, i++)
          {
            da[i] = cd->alarmCode;
            if (isstruct)
            {
              makeDisabledAMS(&ams,cl,cd->device,cd->alarmCode,cd->disabler,cd->timeDisabled);
              memcpy(dout->data.bptr+i*AMS_SIZE,&ams,AMS_SIZE);
            }
          }
          dout->dArrayLength = i;
          if (!isstruct && i > 0) cc = PutValuesFromLong(dout,da,n);
          return cc;
        }
        return 0;
    case PRP_DISABLECMT:
        if (devnr < 0 || devnr >= casTblSiz) return not_registered;
        cl = &casTbl[devnr];
        if (dout->dArrayLength > 0 && dout->dFormat != CF_TEXT) return illegal_format;
        if (access&CA_WRITE) 
        { /* setting disabled alarms ... */
#         ifdef NEED_ADMIN_TO_DISABLE
          strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
          if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
          /* if (!isAdministrator(caller)) return unauthorised_action; */
#         endif
          if (din->dFormat != CF_TEXT) return illegal_format;
          if ((cc=writeCASDisabledComment(devnr,dev,almcode,din->data.cptr,(int)din->dArrayLength)) != 0) return cc;
        }
        if (dout->dArrayLength == 0) return 0;
        return readCASDisabledComment(devnr,dev,almcode,dout->data.cptr,(int *)&dout->dArrayLength);
    case PRP_BASETIME:
        ts = getBaseTime();
        dout->dTimeStamp = (double)ts;
        if (isStringFormat(dout->dFormat))
        {
          char tsstr[80];
          strncpy(tsstr,ctime(&ts),24);
          return PutValuesFromString(dout,tsstr,CF_TEXT,24,0);
        }
        i = (int)ts;
        return PutValuesFromLong(dout,&i,1);
    case PRP_DISABLE_ISACTIVE:
        if (devnr < 0 || devnr >= casTblSiz) return not_registered;
        cl = &casTbl[devnr];
        if (dout->dArrayLength == 0) return dimension_error;
        if (din->dArrayLength > 0)
        {
          if ((cc=GetValuesAsLong(din,&almcode,1)) != 0) return cc;
        }
        if (cl->disabledList == NULL)
        {
          SetEqmCompletion(CASEQM_TAG,"alarm is not disabled");
          return 512;
        }
        n = 0;
        for (cams=casTbl[devnr].almList; cams != NULL; cams = cams->nxt)
        { /* most recent alarms are first in the linked list */
          for (k=0; k<cams->nalms; k++)
          {
            if (dev != NULL && strlen(dev) > 0 && strnicmp(dev,cams->ams[k].device,DEVICE_NAME_SIZE)) continue;
            if (cams->ams[k].alarmCode != (UINT32)almcode) continue;
            if (cams->ams[k].descriptor & almTERMINATE) goto listScanned; /* most recent alarm has been terminated */
            if (cams->ams[k].descriptor & almSUPPRESS)
            {
              n = 1;
              goto listScanned;
            }
          }
        }
listScanned:
        return PutValuesFromLong(dout,&n,1);
    case PRP_DISABLE_ACTIVE:
        if (devnr < 0 || devnr >= casTblSiz) return not_registered;
        cl = &casTbl[devnr];
        if (dout->dArrayLength == 0) return dimension_error;
        if (dout->dFormat != CF_NAME64I) return illegal_format;
        if (cl->disabledList == NULL)
        {
          dout->dArrayLength = 0;
          return 0;
        }
        n = 0;
        for (cams=casTbl[devnr].almList; cams != NULL; cams = cams->nxt)
        {
          for (k=0; k<cams->nalms; k++)
          {
            if (cams->ams[k].descriptor & almTERMINATE) continue;
            if (cams->ams[k].descriptor & almSUPPRESS)
            {
              for (i=0; i<n; i++)
              { /* already in return list ? */
                if (!strncmp(((NAME64I *)dout->data.vptr)[i].name,cams->ams[k].device,DEVICE_NAME_SIZE)) break;
              }
              if (i < n) continue;
              strncpy(((NAME64I *)dout->data.vptr)[n].name,cams->ams[k].device,DEVICE_NAME_SIZE);
              ((NAME64I *)dout->data.vptr)[n].ival = (int)cams->ams[k].alarmCode;
              if (++n == (int)dout->dArrayLength) return 0;
            }
          }
        }
        dout->dArrayLength = (UINT32)n;
        return 0;
    case PRP_SYSALARMS_ALL:
        //showHidden = TRUE;
    case PRP_SYSALARMS:
        filterLast = -1;
    case PRP_RAWSYSALARMS_ALL:
        //if (prpid != PRP_SYSALARMS) showHidden = TRUE;
    case PRP_RAWSYSALARMS:
        if (access&CA_WRITE) return illegal_read_write;
        if (din->dArrayLength > 0)
        {
          switch (din->dFormat)
          {
            case CF_SHORT:
              starttime = stoptime - din->data.sptr[0]*60;
              break;
            case CF_LONG:
              if ((din->dArrayLength > 2)) sev = din->data.lptr[2];
              if ((din->dArrayLength > 1) && (i=(int)din->data.lptr[1]) > 0)
              { /* accept as stoptime if > 0 */
                stoptime = din->data.lptr[1];
              }
              if ((i=(int)din->data.lptr[0]) >= 0) 
              { /* accept as starttime if > 0 */
                starttime = din->data.lptr[0];
              }
              else 
              { /* i is negative (given as a depth) */
                starttime = stoptime + i;
              }
              break;
            default:
              return illegal_format;
          }
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          ptr = dout->data.bptr;
          if (!strcmp(dout->dTag,"AMSr4"))
          {
            n = dout->dArrayLength;
          }
          else if (!strcmp(dout->dTag,"AMS"))
          {
            n = dout->dArrayLength;
            amssize = AMS5_SIZE;
          }
          else if (!strcmp(dout->dTag,"XAMSr4"))
          {
            n = dout->dArrayLength;
            amssize = XAMS_SIZE;
          }
          else if (!strcmp(dout->dTag,"XAMS"))
          {
            n = dout->dArrayLength;
            amssize = XAMS5_SIZE;
          }
          else
          {
            if (dout->dArrayLength%AMS5_SIZE) return illegal_read_write;
            if ((n=dout->dArrayLength/AMS5_SIZE) == 0) return buffer_too_small;
          }
          /* ptr fixed at the caller's read buffer, n is his buffer size */
          if (WaitForSystemMutex(hAmsTblMutex,-1) != 0) return semaphore_error;
          mrt = (time_t)starttime;
          /* mrt = gSystemStartupTime; */
          if (devnr < 0) 
          { /* all or specific alarm system (almsysnr) */
            m = 0;
            for (i=0; i<casTblSiz; i++)
            {
              almsFound = FALSE;
              if (m >= n) break;
              cl = &casTbl[i];
              dftAlarmSystem = casTbl[i].AlarmSystem;
              for (cams=casTbl[i].almList; cams != NULL; cams = cams->nxt)
              {
                if (m >= n) break;
                if ((time_t)cams->stoptime < starttime || 
                    (time_t)cams->starttime > stoptime) continue;
                if (sev == 0 && almsysnr == -1)
                {
                  almsFound = TRUE;
                  if (n < m + cams->nalms) break;
                  prepCASOutgoingAMS(cl,ptr,cams->ams,dout->dTag,cams->nalms);
                  ptr += cams->nalms*amssize;
                  m += cams->nalms;
                }
                else
                {
                  for (k=0; k<cams->nalms; k++)
                  {
                    if (m >= n) break;
                    thisAlarmSystem = cams->ams[k].alarmSystem;
                    //if (showHidden) thisAlarmSystem &= ~ALM_SYSTEM_HIDDEN;
                    if (thisAlarmSystem == 0) /* wasn't set ! */
                      thisAlarmSystem = dftAlarmSystem;
                    if (almsysnr > -1 && thisAlarmSystem != almsysnr) continue;
                    almsFound = TRUE;
                    if (filterLast && cams->ams[k].descriptor & almDISABLED) continue;
                    if (filterLast && (cams->ams[k].descriptor & almTERMINATE)) continue;
                    if (cams->ams[k].descriptor & almSUPPRESS) continue;
                    if (cams->ams[k].severity < sev) continue;
                    prepCASOutgoingAMS(cl,ptr,&cams->ams[k],dout->dTag,1);
                    ptr += amssize;
                    m++;
                  }
                }
              }
              if (!almsFound) continue;
              if (casTbl[i].timestamp == 0) casTbl[i].timestamp = time(NULL) - 7200;
              if (casTbl[i].timestamp > mrt) mrt = casTbl[i].timestamp;
            }
            if (strlen(dout->dTag) > 0) dout->dArrayLength = m;
          }
          else if (devnr >= casTblSiz) 
          { /* control subsystem */
            m = 0;
            for (i=0; i<casTblSiz; i++)
            {
              if (m >= n) break;
              if (strncmp(devName,casTbl[i].subSystem,SUBSYSTEM_NAME_SIZE)) continue;
              cl = &casTbl[i];
              for (cams=casTbl[i].almList; cams != NULL; cams = cams->nxt)
              {
                if (m >= n) break;
                if ((time_t)cams->stoptime < starttime || 
                    (time_t)cams->starttime > stoptime) continue;
                if (sev == 0)
                {
                  if (n < m + cams->nalms) break;
                  prepCASOutgoingAMS(cl,ptr,cams->ams,dout->dTag,cams->nalms);
                  ptr += cams->nalms*amssize;
                  m += cams->nalms;
                }
                else
                {
                  for (k=0; k<cams->nalms; k++)
                  {
                    if (m >= n) break;
                    if (filterLast && cams->ams[k].descriptor & almDISABLED) continue;
                    if (filterLast && (cams->ams[k].descriptor & almTERMINATE)) continue;
                    if (cams->ams[k].descriptor & almSUPPRESS) continue;
                    if (cams->ams[k].severity < sev) continue;
                    prepCASOutgoingAMS(cl,ptr,&cams->ams[k],dout->dTag,1);
                    ptr += amssize;
                    m++;
                  }
                }
              }
              if (casTbl[i].timestamp > mrt) mrt = casTbl[i].timestamp;
            }
            if (strlen(dout->dTag) > 0) dout->dArrayLength = m;
          }
          else 
          { /* targeted device server */
            m = 0;
            cl = &casTbl[devnr];
            for (cams=casTbl[devnr].almList; cams != NULL; cams = cams->nxt)
            {
              if (m >= n) break;
              if ((time_t)cams->stoptime < starttime || 
                  (time_t)cams->starttime > stoptime) continue;
              if (sev == 0)
              {
                if (n < m + cams->nalms) break;
                prepCASOutgoingAMS(cl,ptr,cams->ams,dout->dTag,cams->nalms);
                ptr += cams->nalms*amssize;
                m += cams->nalms;
              }
              else
              {
                for (k=0; k<cams->nalms; k++)
                {
                  if (m >= n) break;
                  if (filterLast && cams->ams[k].descriptor & almDISABLED) continue;
                  if (filterLast && (cams->ams[k].descriptor & almTERMINATE)) continue;
                  if (cams->ams[k].descriptor & almSUPPRESS) continue;
                  if (cams->ams[k].severity < sev) continue;
                  prepCASOutgoingAMS(cl,ptr,&cams->ams[k],dout->dTag,1);
                  ptr += amssize;
                  m++;
                }
              }
            }
            if (casTbl[devnr].timestamp > mrt) mrt = casTbl[devnr].timestamp;
            if (strlen(dout->dTag) > 0) dout->dArrayLength = m;
          }
        }
        ReleaseMutex(hAmsTblMutex);
        dout->dTimeStamp = (double)mrt;
        return 0;
    case PRP_NUMALARMS_ALL:
        //showHidden = TRUE;
    case PRP_NUMALARMS:
        filterLast = -1;
    case PRP_RAWNUMALARMS_ALL:
        //if (prpid != PRP_NUMALARMS) showHidden = TRUE;
    case PRP_RAWNUMALARMS:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength < 1 || dout->dArrayLength > 3) return dimension_error;
        if (din->dArrayLength > 0)
        {
          if ((cc=GetValuesAsLong(din,lcl_ival,3)) != 0) return cc;
          if ((din->dArrayLength > 2)) sev = lcl_ival[2];
          if ((din->dArrayLength > 1) && (i=lcl_ival[1]) > 0) stoptime = i;
          if ((i=lcl_ival[0]) > 0) starttime = lcl_ival[0];
          else starttime = stoptime + i;
        }
        if (WaitForSystemMutex(hAmsTblMutex,-1) != 0) return semaphore_error;
        lcl_numalarms = 0; lcl_nntalarms = 0; lcl_disalarms = 0;
        mrt = (time_t)starttime;
        if (devnr < 0) 
        { /* all or specific alarm system (almsysnr) */
          for (i=0; i<casTblSiz; i++) 
          {
            //if (showHidden) almsysnr &= ~ALM_SYSTEM_HIDDEN;
            if (!hasSubsystem(i,almsysnr)) continue;
            lcl_numalarms += numAlarmsInList(i,starttime,stoptime,sev,almsysnr,filterLast,&ts);
            //if (showHidden) 
            //  lcl_numalarms += numAlarmsInList(i,starttime,stoptime,sev,almsysnr|ALM_SYSTEM_HIDDEN,filterLast,&ts);
            if (ts > mrt) mrt = ts;
            lcl_nntalarms += numNonTerminatedAlarmsInList(i,starttime,stoptime,sev,almsysnr,filterLast,&ts);
            //if (showHidden) 
            //  lcl_nntalarms += numNonTerminatedAlarmsInList(i,starttime,stoptime,sev,almsysnr|ALM_SYSTEM_HIDDEN,filterLast,&ts);
            if (ts > mrt) mrt = ts;
            if (casTbl[i].timestamp == 0) casTbl[i].timestamp = time(NULL) - 7200;
            if (almsysnr == casTbl[i].AlarmSystem)
            {
              if (casTbl[i].timestamp > mrt) mrt = casTbl[i].timestamp;
//              if (casTbl[i].timestampDisabled > mrt) mrt = casTbl[i].timestampDisabled;
            }
            //if (almsysnr != casTbl[i].AlarmSystem && almsysnr != -1) continue; << to show disable alarms everywhere
            lcl_disalarms += casTbl[i].sizDisabledList;
          }
        }
        else if (devnr >= casTblSiz) 
        { /* control subsystem */
          for (i=0; i<casTblSiz; i++)
          {
            if (strncmp(devName,casTbl[i].subSystem,SUBSYSTEM_NAME_SIZE)) continue;
            lcl_numalarms += numAlarmsInList(i,starttime,stoptime,sev,almsysnr,filterLast,&ts);
            if (ts > mrt) mrt = ts;
            lcl_nntalarms += numNonTerminatedAlarmsInList(i,starttime,stoptime,sev,almsysnr,filterLast,&ts);
            if (ts > mrt) mrt = ts;
            if (casTbl[i].timestamp == 0) casTbl[i].timestamp = time(NULL) - 7200;
            if (almsysnr == casTbl[i].AlarmSystem)
            {
              if (casTbl[i].timestamp > mrt) mrt = casTbl[i].timestamp;
//              if (casTbl[i].timestampDisabled > mrt) mrt = casTbl[i].timestampDisabled;
            }
            lcl_disalarms += casTbl[i].sizDisabledList;
          }
        }
        else 
        { /* targeted device server */
          lcl_numalarms = numAlarmsInList(devnr,starttime,stoptime,sev,almsysnr,filterLast,&ts);
          lcl_nntalarms = numNonTerminatedAlarmsInList(devnr,starttime,stoptime,sev,almsysnr,filterLast,&ts);
          if (casTbl[devnr].timestamp == 0) casTbl[devnr].timestamp = time(NULL) - 7200;
          mrt = casTbl[devnr].timestamp;
//          if (casTbl[devnr].timestampDisabled > mrt) mrt = casTbl[devnr].timestampDisabled;
          lcl_disalarms = casTbl[devnr].sizDisabledList;
        }
        lcl_nalm[0] = lcl_numalarms;
        lcl_nalm[1] = lcl_nntalarms;
        lcl_nalm[2] = lcl_disalarms;
        ReleaseMutex(hAmsTblMutex);
        if ((cc=PutValuesFromShort(dout,lcl_nalm,3)) != 0) return cc;
        dout->dTimeStamp = (double)mrt;
        return 0;
    case PRP_SERVERINFO:
        if (devnr < 0) devnr = getNonDecoratedServerAsDeviceNumber(devName);
        if (devnr < 0 || devnr >= casTblSiz) return illegal_equipment_number;
        if (dout->dFormat != CF_STRUCT) return illegal_format;
        if (dout->dArrayLength != 1) return dimension_error;
        if (!strncmp(dout->dTag,"CASQS",TAG_NAME_SIZE))
        {
          strncpy(((CASQS *)(dout->data.vptr))->devContext,casTbl[devnr].devContext,CONTEXT_NAME_SHORTSIZE);
          strncpy(((CASQS *)(dout->data.vptr))->devServer,casTbl[devnr].devServer,EXPORT_NAME_SHORTSIZE);
          strncpy(((CASQS *)(dout->data.vptr))->subSystem,casTbl[devnr].subSystem,SUBSYSTEM_NAME_SIZE);
          strncpy(((CASQS *)(dout->data.vptr))->arcExt,casTbl[devnr].arcExt,XTN_NAME_SIZE);
          ((CASQS *)(dout->data.vptr))->SeverityLevel = casTbl[devnr].SeverityLevel;
          ((CASQS *)(dout->data.vptr))->ArchiveLevel = casTbl[devnr].ArchiveLevel;
          ((CASQS *)(dout->data.vptr))->Retention = casTbl[devnr].Retention;
          ((CASQS *)(dout->data.vptr))->AlarmLevel = casTbl[devnr].AlarmLevel;
          ((CASQS *)(dout->data.vptr))->AlarmSystem = casTbl[devnr].AlarmSystem;
          return 0;
        }
        if (!strncmp(dout->dTag,"CQS",TAG_NAME_SIZE))
        {
          strncpy(((CQS *)(dout->data.vptr))->devContext,casTbl[devnr].devContext,CONTEXT_NAME_SIZE);
          strncpy(((CQS *)(dout->data.vptr))->devServer,casTbl[devnr].devServer,EXPORT_NAME_SIZE);
          strncpy(((CQS *)(dout->data.vptr))->subSystem,casTbl[devnr].subSystem,SUBSYSTEM_NAME_SIZE);
          strncpy(((CQS *)(dout->data.vptr))->arcExt,casTbl[devnr].arcExt,XTN_NAME_SIZE);
          ((CQS *)(dout->data.vptr))->SeverityLevel = casTbl[devnr].SeverityLevel;
          ((CQS *)(dout->data.vptr))->ArchiveLevel = casTbl[devnr].ArchiveLevel;
          ((CQS *)(dout->data.vptr))->Retention = casTbl[devnr].Retention;
          ((CQS *)(dout->data.vptr))->AlarmLevel = casTbl[devnr].AlarmLevel;
          ((CQS *)(dout->data.vptr))->AlarmSystem = casTbl[devnr].AlarmSystem;
          return 0;
        }
        return invalid_structure_tag;
    case PRP_SUBSYSTEMS:
        if ((siz=dout->dArrayLength) > 0)
        {
          switch (dout->dFormat)
          {
            case CF_NAME8: 
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME48:
            case CF_NAME64: len = fmtsizeof(dout->dFormat); break;
            case CF_USTRING:
            case CF_INTINT: len = 0; break;
            case CF_TEXT: len = 16; siz = siz/16; break;
            default : return illegal_format;
          }
          for (i=0; i<siz && i<nAlarmSubSystems; i++)
          {
            if (len > 0)
              strncpy(dout->data.cptr+i*len,AlarmSubSystems[i].name,SUBSYSTEM_NAME_SIZE);
            if (dout->dFormat == CF_INTINT)
            {
              ((INTINT *)dout->data.vptr)[i].i1val = AlarmSubRanges[i].i1val;
              ((INTINT *)dout->data.vptr)[i].i2val = AlarmSubRanges[i].i2val;
            }
            if (dout->dFormat == CF_USTRING)
            {
              strncpy(dout->data.cptr+i*96+16,AlarmSubSystems[i].name,SUBSYSTEM_NAME_SIZE);
              ((DUSTRING *)dout->data.vptr)[i].f1val = (float)AlarmSubRanges[i].i1val;
              ((DUSTRING *)dout->data.vptr)[i].f2val = (float)AlarmSubRanges[i].i2val;
            }
          }
          dout->dArrayLength = dout->dFormat == CF_TEXT ? i*len : i;
        }
        return 0;
    case PRP_ALARMSYSTEMS_ALL:
        showHidden = TRUE; /* show the hidden subsystems */
    case PRP_ALARMSYSTEMS:
        if ((siz=dout->dArrayLength) > 0)
        {
          ilen = 0;
          switch (dout->dFormat)
          {
            case CF_NAME16II:
              addViewColumn = TRUE;
              ilen = 8;
            case CF_NAME16I:
            case CF_NAME32I:
            case CF_NAME48I:
            case CF_NAME64I: 
              if (ilen == 0) ilen = 4;
              addAlarmCode = TRUE;
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME48:
            case CF_NAME64: 
              len = fmtsizeof(dout->dFormat); 
              break;
            default : 
              return illegal_format;
          }
          nlen = len - ilen;
          for (i=0,n=0; i<nAlarmSystems && n<siz; i++)
          {
            if (!showHidden && AlarmSystems[i].almsystem[0] == '.') continue;
            strncpy(dout->data.cptr+n*len,AlarmSystems[i].almsystem,nlen);
            if (addAlarmCode) 
              *(UINT32 *)(dout->data.bptr+(n+1)*len - ilen) = AlarmSystems[i].number;
            if (addViewColumn) 
              *(UINT32 *)(dout->data.bptr+(n+1)*len - 4) = AlarmSystems[i].column;
            n++;
          }
          dout->dArrayLength = n;
        }
        return 0;
    case PRP_SERVERS:
        if ((siz=dout->dArrayLength) > 0)
        {
          switch (dout->dFormat)
          {
            case CF_NAME8: len = 8; break;
            case CF_NAME16: len = 16; break;
            case CF_NAME32: len = 32; break;
            case CF_NAME48: len = 48; break;
            case CF_NAME64: len = 64; break;
            case CF_TEXT: len = 16; siz = siz/16; break;
            default : return illegal_format;
          }
          devNameBuffer[EXPORT_NAME_SIZE] = 0;
          chkss = (devnr >= casTblSiz) ? -1 : 0;
          for (i=0,n=0; i<siz && i<casTblSiz; i++)
          {
            if (chkss && strncmp(devName,casTbl[i].subSystem,SUBSYSTEM_NAME_SIZE)) continue;
            if (strnicmp(gDeviceContext,casTbl[i].devContext,CONTEXT_NAME_SIZE))
              sprintf(devNameBuffer,"%.32s@%.32s",casTbl[i].devServer,casTbl[i].devContext);
            else
              strncpy(devNameBuffer,casTbl[i].devServer,EXPORT_NAME_SIZE);
            strncpy(dout->data.cptr+n*len,devNameBuffer,len);
            n++;
          }
          dout->dArrayLength = dout->dFormat == CF_TEXT ? n*len : n;
        }
        return 0;
    case PRP_NUMALMDEFS:
        if (devnr < 0) devnr = getNonDecoratedServerAsDeviceNumber(devName);
        if (devnr < 0 || devnr >= casTblSiz) return illegal_equipment_number;
        if (dout->dArrayLength == 0) return illegal_read_write;
        n = casTbl[devnr].nAdsList;
        return PutValuesFromLong(dout,&n,1);
    case PRP_SYSALMDEFS:
        if (devnr < 0) devnr = getNonDecoratedServerAsDeviceNumber(devName);
        if (devnr < 0 || devnr >= casTblSiz) return illegal_equipment_number;
        if (dout->dArrayLength == 0) return illegal_read_write;
        if (dout->dFormat != CF_STRUCT) return illegal_format;
        almcode = -1;
        if (casTbl[devnr].nAdsList == 0) codeFound = FALSE;
        if (din->dArrayLength > 0)
        {
          if ((cc=getValuesAsLong(din,(SINT32 *)&almcode,1)) != 0) return cc;
          codeFound = FALSE;
        }
        n = 0;
        ptr = dout->data.bptr;
        if (!strcmp(dout->dTag,"ADSr4"))
        { /* modern */
          if (almcode == NOT_RESPONDING)
          {
            memcpy(ptr,&adsSrvDown,ADS_SIZE);
            dout->dArrayLength = 1;
            codeFound = TRUE;
            return 0;
          }
          for (i=0; i<casTbl[devnr].nAdsList; i++)
          {
            if (almcode > 0 && 
                almcode != (int)casTbl[devnr].adsList[i].alarmCode) continue;
            codeFound = TRUE;
            if (n >= (int)dout->dArrayLength) break;
            memcpy(ptr,&casTbl[devnr].adsList[i],ADS_SIZE);
            ptr += ADS_SIZE;
            n++;
          }
          if (!codeFound)
          { /* looking for a specific code and came up empty (or no alarm definitions recorded) */
            needToAcquireAlarmDefs = TRUE;
            casTbl[devnr].adsListAcquired = FALSE;
          }
          dout->dArrayLength = (UINT32)n;
          return 0;
        }
        if (!strcmp(dout->dTag,"ADS"))
        { /* legacy */
          for (i=0; i<casTbl[devnr].nAdsList; i++)
          {
            if (almcode > 0 && 
                almcode != (int)casTbl[devnr].adsList[i].alarmCode) continue;
            codeFound = TRUE;
            if (n >= (int)dout->dArrayLength) break;
            ADStoADS5((ADS5 *)ptr,&casTbl[devnr].adsList[i]);
            ptr += ADS5_SIZE;
            n++;
          }
          if (!codeFound)
          { /* looking for a specific code and came up empty (or no alarm definitions recorded) */
            needToAcquireAlarmDefs = TRUE;
            casTbl[devnr].adsListAcquired = FALSE;
          }
          dout->dArrayLength = (UINT32)n;
          return 0;
        }
        /* dout->dArrayLength = n; */
        return illegal_format;
    case PRP_SYSALARMSEXT:
        return not_implemented;
    case PRP_SYSTEMLIST:
        if (devnr < 0) devnr = getNonDecoratedServerAsDeviceNumber(devName);
        if (devnr < 0 || devnr >= casTblSiz) return illegal_equipment_number;
        if (access & CA_WRITE)
        {
          strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
          if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
          /* if (!isAdministrator(caller)) return unauthorised_action; */
          if (din->dArrayLength > 0)
          {
            if ((cc=GetValuesAsLong(din,lcl_ival,25)) != 0) return cc;
            if (din->dArrayLength > 25) din->dArrayLength = 25;
            for (i=0; i<(int)din->dArrayLength; i++)
            {
              for (k=0; k<nAlarmSystems; k++) if (AlarmSystems[k].number == lcl_ival[i]) break;
              if (k == nAlarmSystems) return invalid_index;
            }
          }
          if ((cc=putServerAlarmSystemList(&casTbl[devnr],lcl_ival,din->dArrayLength)) != 0) return cc;
          rmvServerAlarmSystemList(&casTbl[devnr]);
          getServerAlarmSystemList(&casTbl[devnr]);
        }
        if ((siz=dout->dArrayLength) > 0)
        {
          n = 1 + casTbl[devnr].nAlmSystemList;
          if ((lptr=(SINT32 *)alloca(n*sizeof(SINT32))) == NULL) return out_of_server_memory;
          lptr[0] = casTbl[devnr].AlarmSystem;
          for (i=0; i<casTbl[devnr].nAlmSystemList; i++)
          {
            lptr[i+1] = casTbl[devnr].almSystemList[i];
          }
          if ((cc=putValuesFromLong(dout,lptr,(short)n)) != 0) return cc;
          if (siz > n) dout->dArrayLength = n;
        }
        return 0;
    case PRP_SERVERLIST:
        if (almsysnr < 0) return illegal_equipment_number;
        if ((siz=dout->dArrayLength) > 0)
        {
          switch (dout->dFormat)
          {
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME48: 
            case CF_NAME64:
              len = fmtsizeof(dout->dFormat); break;
            default : return illegal_format;
          }
          for (i=0,n=0; n<siz && i<casTblSiz; i++)
          {
            isAllowed = casTbl[i].AlarmSystem == almsysnr ? TRUE : FALSE;
            if (!isAllowed && casTbl[i].nAlmSystemList > 0)
            {
              for (k=0; k<casTbl[i].nAlmSystemList; k++)
              {
                if (casTbl[i].almSystemList[k] == almsysnr)
                {
                  isAllowed = TRUE;
                  break;
                }
              }
            }
            if (isAllowed)
            {
              devNameBuffer[EXPORT_NAME_SIZE] = 0;
              if (strnicmp(gDeviceContext,casTbl[i].devContext,CONTEXT_NAME_SIZE))
                sprintf(devNameBuffer,"%.32s@%.32s",casTbl[i].devServer,casTbl[i].devContext);
              else
                strncpy(devNameBuffer,casTbl[i].devServer,EXPORT_NAME_SIZE);
              strncpy(dout->data.cptr+n*len,devNameBuffer,len);
              n++;
            }
          }
          dout->dArrayLength = n;
        }
        return 0;
    case PRP_SERVERDB:
        if (access & CA_WRITE)
        {
          strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
          if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
          /* if (!isAdministrator(caller)) return unauthorised_action; */
          if (din->dArrayLength > 0)
          {
            if (din->dFormat != CF_STRUCT) return illegal_format;
            if (strncmp(din->dTag,CASDBITEM_TAG,TAG_NAME_SIZE)) return invalid_structure_tag;
            if ((cc=appendCASDb((CasDbItem *)din->data.vptr,din->dArrayLength)) != 0) return cc;
          }
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,CASDBITEM_TAG,TAG_NAME_SIZE)) return invalid_structure_tag;
          if (dout->dArrayLength > (UINT32)casTblSiz) dout->dArrayLength = (int)casTblSiz;
          dbItm = (CasDbItem *)dout->data.vptr;
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            strncpy(dbItm[i].devContext,casTbl[i].devContext,CONTEXT_NAME_SIZE);
            strncpy(dbItm[i].devServer,casTbl[i].devServer,EXPORT_NAME_SIZE);
            strncpy(dbItm[i].subSystem,casTbl[i].subSystem,SUBSYSTEM_NAME_SIZE);
            strncpy(dbItm[i].arcExt,casTbl[i].arcExt,XTN_NAME_SIZE);
            dbItm[i].SeverityLevel = casTbl[i].SeverityLevel;
            dbItm[i].ArchiveLevel = casTbl[i].ArchiveLevel;
            dbItm[i].Retention = casTbl[i].Retention;
            dbItm[i].AlarmLevel = casTbl[i].AlarmLevel;
            dbItm[i].AlarmSystem = casTbl[i].AlarmSystem;
            dbItm[i].offline = casTbl[i].offline;
          }
        }
        return 0;
    case PRP_ALMSYSDB:
        if (access & CA_WRITE)
        {
          strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
          if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
          /* if (!isAdministrator(caller)) return unauthorised_action; */
          if (din->dArrayLength > 0)
          {
            if (din->dFormat != CF_STRUCT) return illegal_format;
            if (strncmp(din->dTag,CASALMSYSITEM_TAG,TAG_NAME_SIZE)) return invalid_structure_tag;
            if ((cc=putAlarmSystemsList(ALARM_SYSTEM_FILE,(CasAlmSysTblEntry *)din->data.vptr,din->dArrayLength)) != 0)
              return cc;
          }
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,CASALMSYSITEM_TAG,TAG_NAME_SIZE)) return invalid_structure_tag;
          if (dout->dArrayLength > (UINT32)nAlarmSystems) dout->dArrayLength = (int)nAlarmSystems;
          almSysItm = (CasAlmSysTblEntry *)dout->data.vptr;
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            strncpy(almSysItm[i].subsystem,AlarmSystems[i].subsystem,SUBSYSTEM_NAME_SIZE);
            strncpy(almSysItm[i].almsystem,AlarmSystems[i].almsystem,ALARM_TAG_SIZE);
            almSysItm[i].column = AlarmSystems[i].column;
            almSysItm[i].number = AlarmSystems[i].number;
            almSysItm[i].row = AlarmSystems[i].row;
          }
        }
        return 0;
    case PRP_ACTIONDB:
        if (devnr < 0) devnr = getNonDecoratedServerAsDeviceNumber(devName);
        if (devnr < 0 || devnr >= casTblSiz) return illegal_equipment_number;
        if (access & CA_WRITE)
        {
          strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
          if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
          /* if (!isAdministrator(caller)) return unauthorised_action; */
          if (din->dArrayLength > 0)
          {
            if (din->dFormat != CF_STRUCT) return illegal_format;
            if (strncmp(din->dTag,CASACTIONITEM_TAG,TAG_NAME_SIZE)) return invalid_structure_tag;
          }
          if ((cc=writeAlmActionLst(devnr,(CasActionItem *)din->data.vptr,din->dArrayLength)) != 0) return cc;
          rmvCASActionList(devnr);
          casTbl[devnr].actionList = getCASActionList(devnr);
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_STRUCT) return illegal_format;
          if (strncmp(dout->dTag,CASACTIONITEM_TAG,TAG_NAME_SIZE)) return invalid_structure_tag;
          actItm = (CasActionItem *)dout->data.vptr;
          ca = casTbl[devnr].actionList;
          for (i=0; i<(int)dout->dArrayLength && ca != NULL; i++)
          {
            strncpy(actItm[i].email,ca->email,128);
            /* split ctx, srv, dev from devName */
            strncpy(devNameBuffer,ca->devName,128);
            if ((cc=tokenizeFullName(devNameBuffer,&ctx,&srv,&dev)) != 0) return cc;
            strncpy(actItm[i].ctx,ctx,CONTEXT_NAME_SIZE);
            strncpy(actItm[i].srv,srv,EXPORT_NAME_SIZE);
            strncpy(actItm[i].dev,dev,DEVICE_NAME_SIZE);
            strncpy(actItm[i].prp,ca->devProperty,PROPERTY_NAME_SIZE);
            strncpy(actItm[i].actionSignalString,ca->actionSignalString,128);
            actItm[i].actionSignalFormat = ca->actionSignalFormat;
            actItm[i].alarmCode = ca->alarmCode;
            ca = ca->nxt;
          }
          dout->dArrayLength = (UINT32)i;
        }
        return 0;
    case PRP_RESTART:
        if ((access & CA_WRITE) == 0) return illegal_read_write;
        strncpy(caller,GetCaller(CASEQM_TAG),USERNAME_SIZE);
        if (AssertIsAdministrator(caller) == FALSE) return unauthorised_action;
        /* if (!isAdministrator(caller)) return unauthorised_action; */
        needsRestart = TRUE;
        return 0;
    case PRP_NREGIONRULES:
        cc = getRegionRulesForServer(devName,NULL,&n);
        return cc ? cc : PutValuesFromLong(dout,&n,1);
    case PRP_REGIONRULES:
        if (dout->dFormat != CF_STRUCT) return illegal_format;
        if (strcmp(dout->dTag,"RgnRule")) return invalid_structure_tag;
        n = (int)dout->dArrayLength;
        cc = getRegionRulesForServer(devName,(DeviceRegionItem *)dout->data.vptr,&n);
        if (n < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)n;
        return cc;
    case PRP_REGIONSDB:
        if (dout->dFormat != CF_STRUCT) return illegal_format;
        if (strcmp(dout->dTag,DBTAG)) return invalid_structure_tag;
        if (dout->dArrayLength > (UINT32)nRegionDbItems) dout->dArrayLength = (UINT32)nRegionDbItems;
        memcpy(dout->data.vptr,regionDbItems,dout->dArrayLength*sizeof(RegionDbItem));
        return 0;
    case PRP_REGIONSXREF:
        if (dout->dFormat != CF_NAME16I) return illegal_format;
        if (dout->dArrayLength > (UINT32)nRegionsXRef) dout->dArrayLength = (UINT32)nRegionsXRef;
        memcpy(dout->data.vptr,regionsXRef,dout->dArrayLength*sizeof(NAME16I));
        return 0;
    default:
        return illegal_property;
  }
}
int logproblem(char *text,...)
{
  int len,cc=0;
  double ts = makeDataTimeStamp();
  char str[256];
  FILE *fp = NULL;
  va_list args;
  if (strlen(text) > 80) ccerr(string_too_long);
  va_start(args,text);
  sprintf(str,"%s : ",getDataTimeString(ts,0));
  vsprintf(&str[strlen(str)],text,args);
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  va_end(args);
  if ((fp=fopen("problems.log","a")) == NULL) ccerr(file_error);
  fwrite(str,len=strlen(str),1,fp);
err:
  if (fp) fclose(fp);
  return cc;
}

