/* ens.c - Equipment Name Server */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include "prolog.h"
#include "tine.h"
#include "srvip.h"

extern int NameServerLoaded;
extern TINE_EXPORT int gScanForNameLists;
int findcol(char *str,char *tag);
char *colptr(int col,char *str);
void HexToBytes(char *s,BYTE *b,int nibs);
extern int SrvTblSize,FecTblSize;
extern int CommandLevel;
char eqpfname[256],fecfname[256];
FecAddrStruct gTgtFec;
ExpDataStruct gTgtSrv;
int CtxTblSize = 0;
int debug = 0;
int checkExportFlag = 0;
int forceExportAddr = FALSE;
int gCanSendAddrInUse = TRUE;
int dumpFecFileFlag = 0;
int dumpEqpFileFlag = 0;
int scanFlag = 0;
int fixFecProtocol(int fecidx);
int isSystemContextDecoration(char *dec);
char *getShortSubSystemName(char *subLongName,int forceMatch);
void rmvWidowedFecs(void);
int dumpFwdQueues(int i1,int i2, int i3, int i4);
int dumpCorrections(int i1,int i2, int i3, int i4);
int dumpFixedAddresses(int i1,int i2, int i3, int i4);
void initENS(void);
void taskENS(void);
void stopENS(void);
int pollInit(void);
int enseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access);  
int getContexts(void);
NAME32 *contexts = NULL;
int ncontexts = 0;
INADR ens_iaddr;
int dbbackup(char *fn);
BYTE *getFwdIpAddr(char *tgt);
void checkFixedAddr(ExpDataStruct *srv,int *matchaddr, int *matchfec);
int removeDecoration(char *context);

char logbuf[212];

#define CHECK_VALID_CHARS(c) ((strlen(c) == 0 || !isalnum((c)[0]) || strpbrk(c,"/\\* ") != NULL) ? FALSE : TRUE)

#define FILESPACE 100
#define TMPSPACE  10 

#define CMDSFILE "cmds.log"

#define MAX_SEARCH_PARAMS 25
typedef struct SearchType
{
  char *lst[MAX_SEARCH_PARAMS];
  int len;
} SearchType;

SearchType ctxSearch = { NULL, 0};
SearchType srvSearch = { NULL, 0};
SearchType subSearch = { NULL, 0};
SearchType osSearch = { NULL, 0};
SearchType eqmSearch = { NULL, 0};
SearchType rspSearch = { NULL, 0};
SearchType impSearch = { NULL, 0};
SearchType ipSearch = { NULL, 0};
SearchType fecSearch = { NULL, 0};

int gUseSearchLists = FALSE;
void clearSearchLists(void)
{
  ctxSearch.len = subSearch.len = osSearch.len = 0;
  eqmSearch.len = rspSearch.len = impSearch.len = 0;
  srvSearch.len = ipSearch.len = fecSearch.len = 0;
  gUseSearchLists = FALSE;
  return;
}
int parseSearchString(char *str)
{
  char *c, *cv, *stok = "/";
  int i;
  clearSearchLists();
  if (strstr(str,stok) == NULL) return not_applicable;
  gUseSearchLists = TRUE;
  for (c=strtok(str,stok); c != NULL; c=strtok(NULL,stok))
  {
    if ((cv=strchr(c,'=')) == NULL) continue; // not well formed
    cv++; // move past the '='
    strtrm(cv); // trim the whitespace off the ends ...
    switch (tolower(*c))
    {
      case 'c':
        if ((i=ctxSearch.len) >= MAX_SEARCH_PARAMS) continue;
        ctxSearch.lst[i] = cv;
        ctxSearch.len++;
        break;
      case 's':
        if (strnicmp(c,"su",2) == 0)
        { /* subsystem */
          if ((i=subSearch.len) >= MAX_SEARCH_PARAMS) continue;
          subSearch.lst[i] = cv;
          subSearch.len++;
        }
        else
        { /* server */
          if ((i=srvSearch.len) >= MAX_SEARCH_PARAMS) continue;
          srvSearch.lst[i] = cv;
          srvSearch.len++;
        }
        break;
      case 'o':
        if ((i=osSearch.len) >= MAX_SEARCH_PARAMS) continue;
        osSearch.lst[i] = cv;
        osSearch.len++;
        break;
      case 'e':
        if ((i=eqmSearch.len) >= MAX_SEARCH_PARAMS) continue;
        eqmSearch.lst[i] = cv;
        eqmSearch.len++;
        break;
      case 'f':
        if ((i=fecSearch.len) >= MAX_SEARCH_PARAMS) continue;
        fecSearch.lst[i] = cv;
        fecSearch.len++;
        break;
      case 'r':
        if ((i=rspSearch.len) >= MAX_SEARCH_PARAMS) continue;
        rspSearch.lst[i] = cv;
        rspSearch.len++;
        break;
      case 'i':
        if (strnicmp(c,"ip",2) == 0)
        {
          if ((i=ipSearch.len) >= MAX_SEARCH_PARAMS) continue;
          ipSearch.lst[i] = cv;
          ipSearch.len++;
        }
        else
        {
          if ((i=impSearch.len) >= MAX_SEARCH_PARAMS) continue;
          impSearch.lst[i] = cv;
          impSearch.len++;
        }
        break;
    }
  }
  return 0;
}

typedef struct
{
  time_t lastPinged;
  time_t lastActive;
  int wasLogged;
  int pingPending;
  int rebootCounts;
  int disabled;
  int fecIndex;
} SrvInfo;

typedef struct
{
  short disable;
  short importance;
  char  ipv4[16];
  char  ipv6[40];
  BYTE  mac[6];
} FecInfoStructEx;

FecInfoStruct *FecInfoTbl = NULL,gTgtFecInfo;
FecInfoStructEx *FecInfoEx = NULL, gTgtFecInfoEx;
SrvInfo *SrvInfoTbl = NULL;

FecAddrStruct gFecSink;
char gFecSinkAddr[64] = "100.100.100.100";
ExpDataStruct gSrvSink;
FecInfoStruct gFecSinkInfo;

typedef struct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char fec[FEC_NAME_SIZE];
  char os[FEC_OS_SIZE];
  char sub[SUBSYSTEM_NAME_SIZE];
  char addr[ADDR_SIZE_IPV6];
  char ver[FEC_VERSION_SIZE];
  char resp[FEC_RESP_SIZE];
  int importance;
  int portoffset;
} SrvQueryInfo;
int registerSrvQueryInfoStruct(void)
{
  int cc;
  if ((cc=addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo,ctx),
    CONTEXT_NAME_SIZE, CF_TEXT, "context")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, srv),
    EXPORT_NAME_SIZE, CF_TEXT, "server")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, fec),
    FEC_NAME_SIZE, CF_TEXT, "fec")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, os),
    FEC_OS_SIZE, CF_TEXT, "os")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, sub),
    SUBSYSTEM_NAME_SIZE, CF_TEXT, "subsystem")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, addr),
    ADDR_SIZE_IPV6, CF_TEXT, "ip")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, ver),
    FEC_VERSION_SIZE, CF_TEXT, "version")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, resp),
    FEC_RESP_SIZE, CF_TEXT, "responsible")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, importance),
    1, CF_INT32, "importance")) != 0) return cc;
  if ((cc = addFieldToStruct("SRVQINF", (int)OFFSETIN(SrvQueryInfo, portoffset),
    1, CF_INT32, "port_offset")) != 0) return cc;

  if ((cc= sealTaggedStruct("SRVQINF", sizeof(SrvQueryInfo), 1000)) != 0) return cc;
  return 0;
}

UINT32 gMininumFreeDiskSpace = 200000;
#define FREESPACE_FILE "freespace.csv"
int getMinFreeDiskSpace(char *fname)
{
  char str[256],scratch[128];
  int done = 0,i,cc = 0;
  int col=0;
  FILE *fp = NULL;
 
  if ((fp=fopen(fname,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; csvGetLine(str,255,fp) != NULL; )   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((col=findcol(str,"DISK_SPACE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(col,str),64);
    gMininumFreeDiskSpace = strtol(scratch,NULL,0);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}

#define ONEHOUR 3600
#define ONEDAY (ONEHOUR * 24)
#define ONEMONTH (ONEDAY *30)
int gAllowedDeadTime = ONEMONTH * 3;
#ifdef SOLARIS
#define INADDR_NONE 0
#endif

#define LOCATIONS_FILE "locations.csv"
int hasLocationsFile = FALSE;
char *getCanonicalLocation(char *ip)
{
  FILE *fp=NULL;
  int ip_col=-1,loc_col=-1,done=0,cc=address_unknown;
  char str[256],hdr[256];
  static char location[64];

  if (ip == NULL || strlen(ip) == 0) return NULL;
  if ((fp=fopen(LOCATIONS_FILE,"r")) == NULL) 
  {
    hasLocationsFile = FALSE;
    return NULL;
  }
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((ip_col=findcol(hdr,"IP")) < 0) ccerr(no_such_line);
      if ((loc_col=findcol(hdr,"LOCATION")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    if (strncmp(ip,colptr(ip_col,str),16) != 0) continue;
    strncpy(location,colptr(loc_col,str),64);
    ccerr(0);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc ? NULL : location;
}
void checkCanonicalLocations(void)
{
  FILE *fp=NULL;
  int ip_col=-1,loc_col=-1,done=0,cc=0,i;
  char str[256],hdr[256];
  char ip[64], loc[64];

  if ((fp=fopen(LOCATIONS_FILE,"r")) == NULL) 
  {
    hasLocationsFile = FALSE;
    return;
  }
  while (csvGetLine(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((ip_col=findcol(hdr,"IP")) < 0) ccerr(no_such_line);
      if ((loc_col=findcol(hdr,"LOCATION")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(ip,colptr(ip_col,str),16);
    strncpy(loc,colptr(loc_col,str),64);
    for (i=0; i<FecTblSize; i++)
    {
      if (!strcmp(FecTbl[i].fecName,"FECSINK"))
      {
        strncpy(FecInfoTbl[i].loc,"nowhere",FEC_LOCATION_SIZE);
        continue;
      }
      if (!strnicmp(FecTbl[i].strAdr,ip,ADDR_SIZE_IPV6))
      {
        strncpy(FecInfoTbl[i].loc,loc,FEC_LOCATION_SIZE);
      }
    }  
  }
err:
  if (cc == 0) hasLocationsFile = TRUE;
  if (fp != NULL) fclose(fp);  
}
#if defined(WIN32) || defined(NEED_UNSETENV)
int unsetenv(char *env)
{
  char envstr[128];
  strncpy(envstr,env,120);
  strcat(envstr,"=");
# ifdef WIN32
  return _putenv(envstr);
# else
  return putenv(envstr);
# endif
}
#endif
int strcontains(const char *str1,const char *str2)
{
  char s1[64],s2[64];
  strncpy(s1,str1,64); strncpy(s2,str2,64);
  strupr(s1); strupr(s2);
  return strstr(s1,s2) == NULL ? FALSE : TRUE;
}
char thisLocation[FEC_LOCATION_SIZE] = "central location";
void PreSystemInit(void)
{
  char *c=NULL;
  FILE *fp=NULL;
  int keepEnvSettings=FALSE,n=0;
  if ((c=getenv("USE_TINE_HOME")) != NULL && strcmp(c,"TRUE") == 0)
     keepEnvSettings = TRUE;
  if (!keepEnvSettings)
  {
    /* take NO chances that this registers a name server ! */
    isEquipmentNameServer(TRUE);
    unsetenv("TINE_HOME");
    unsetenv("FEC_HOME");
    /* legacy env variable */
    unsetenv("CONTROLDB");
    unsetenv("FECDB");
  }
  SystemAssignBufferSpace(100000,32000);
  /* take a peek at the eqpdbase.csv file: */
  n = GetDataBaseSize(EQPDB_FILE);
  MaxSystemTransportSize = MAX(0x40000,n*100);
  SetAutoAdjustWorkAreaSize(FALSE);
  MaxNumContracts = 1000;
  MaxNumConnections = 1000;
  gRespondToServiceRequests = TRUE;
  gUseLoopback = TRUE;
  gStartupDebug = FALSE;
  gScanForNameLists = FALSE;
  FeclogDepth = 10000;
  useCycleTrigger = FALSE; /* don't care about this in the ENS */
  useGlobalSynchronization = FALSE; /* don't care about this in the ENS */
  gIgnoreFollowers = TRUE;

  if ((fp=fopen(FECID_FILE,"r")) == NULL)
  { /* no local file : register this way */
    RegisterFecInformation("ENS","SER","SITE","equipment name server",thisLocation,"none","control system admin",0);
  }
  else
  { /* local file found : close it and let the system pick it up later */
    fclose(fp);
  }
}
#define DEFAULT_DEV_TAG 0
#define FEC_DEV_TAG 1
#define SRV_DEV_TAG 2
#define INF_DEV_TAG 3
void PostSystemInit(void)
{
  DTYPE dout;
  
  DTYPEZERO(dout);

  RegisterEquipmentModule("ENS","ENSEQM",100,enseqm,initENS,taskENS,200,stopENS);

  RegisterProperty("ENSEQM","SIZES",20,CF_SHORT,CA_READ,"Database sizes");
  RegisterProperty("ENSEQM","EXPORT",310,CF_STRUCT,CA_WRITE,"Handle export information (add new FEC or Server)");
  RegisterProperty("ENSEQM","REMOVE",1,CF_NAME16,CA_WRITE,"remove FEC or Server entry");
  RegisterProperty("ENSEQM","REBOOTCOUNT",1,CF_LONG,CA_READ,"[0:1e11]running server reboot count");
  RegisterProperty("ENSEQM","REBOOTCOUNT.LIST",500,CF_NAME32I,CA_READ,"[0:1e11]list of server reboot counts");
  RegisterProperty("ENSEQM","LASTACTIVE",1,CF_LONG,CA_READ,"[0:1e11 sec]last time server marked active (utc)");
  RegisterProperty("ENSEQM","CHECKFLAG",1,CF_SHORT,CA_READ|CA_WRITE,"read/write export check flag");
  RegisterProperty("ENSEQM","TAGS",1000,CF_NAME32,CA_READ,"get exported servers list");
  RegisterProperty("ENSEQM","NTAGS",1,CF_SHORT,CA_READ,"get number of exported servers list");
  RegisterProperty("ENSEQM","SERVERS",1000,CF_NAME32,CA_READ,"get exported servers list");
  dout.dArrayLength = 1000;
  dout.dFormat = CF_STRUCT;
  strcpy(dout.dTag, "SRVQINF");
  RegisterPropertyInformation("ENSEQM","SERVERS", &dout, NULL, CA_READ, AT_SINGLE, 0, "get exported servers list", 0, NULL);
  RegisterProperty("ENSEQM","NSERVERS",1,CF_SHORT,CA_READ,"get number of exported servers list");
  RegisterProperty("ENSEQM","FECS",1000,CF_NAME16,CA_READ,"get fec list");
  RegisterProperty("ENSEQM","NFECS",1,CF_SHORT,CA_READ,"get number of fecs");
  RegisterProperty("ENSEQM","SuggestFecName",1,CF_NAME32,CA_READ,"suggested fec name based on input srv name");
  RegisterProperty("ENSEQM","CONTEXTS",200,CF_NAME32,CA_READ,"get context list");
  RegisterProperty("ENSEQM","NCONTEXTS",1,CF_SHORT,CA_READ,"get number of contexts");
  RegisterProperty("ENSEQM","SUBSYSTEM",1,CF_NAME16,CA_READ,"get assigned subsystem for targetted server");
  RegisterProperty("ENSEQM","SUBSYSTEMS",200,CF_NAME16,CA_READ,"get subsystem list");
  RegisterProperty("ENSEQM","SUBSYSTEMS.ALLOWED",200,CF_NAME16,CA_READ,"allowed subsystem list (empty => everything)");
  RegisterProperty("ENSEQM","SUBSYSTEMS.DESCRIPTIONS",200,CF_NAME64,CA_READ,"allowed subsystem list descriptions");
  RegisterProperty("ENSEQM","NSUBSYSTEMS",1,CF_SHORT,CA_READ,"get number of subsystems");
  RegisterProperty("ENSEQM","ALMTAGS",1000,CF_NAME16,CA_READ,"get exported servers list in alarm system");
  RegisterProperty("ENSEQM","NALMTAGS",1,CF_SHORT,CA_READ,"get number of exported servers list in alarm system");
  RegisterProperty("ENSEQM","WHOAMI",5,CF_NAME16,CA_READ,"requested server information");
  RegisterProperty("ENSEQM","IMPORTANCE",1,CF_NAME16,CA_READ|CA_WRITE,"importance level for given fec");
  RegisterProperty("ENSEQM","CONTEXTS.DECORATED",200,CF_NAME32,CA_READ,"decorated context list");
  RegisterProperty("ENSEQM","CONTEXT.ROOTS.ALLOWED",200,CF_NAME32,CA_READ,"allowed root contexts (empty list => everything)");
  RegisterProperty("ENSEQM","ADD.ALLOWED.ROOT",1,CF_NAME32,CA_WRITE,"add allowed root context to list");
  RegisterProperty("ENSEQM","DEL.ALLOWED.ROOT",1,CF_NAME32,CA_WRITE,"remove allowed root context from list");
  RegisterProperty("ENSEQM","ALLOWEDDEADTIME",1,CF_LONG,CA_READ|CA_WRITE,"[3600:1e6 sec]allowed dead time before auto-removal of server");
  RegisterProperty("ENSEQM","REMOVENOTRUNNING",0,CF_NULL,CA_WRITE,"start process to remove servers currently not running");
  RegisterProperty("ENSEQM","ADDRESSOF",sizeof(ExpDataStruct)+sizeof(FecDataStruct),CF_STRUCT,CA_READ,"address of targeted server");
  RegisterProperty("ENSEQM","ADDRESSOF",sizeof(ExpDataStruct)+sizeof(FecAddrStruct),CF_STRUCT,CA_READ,"address of targeted server");
  RegisterProperty("ENSEQM","ADDRESSOF",4,CF_NAME64,CA_READ,"main address parameters of targeted server as string");
  RegisterProperty("ENSEQM","EXPORT.ALIAS",2,CF_NAME64,CA_WRITE,"register export alias");
  RegisterProperty("ENSEQM","FECALIASLIST",32,CF_NAME16,CA_READ,"export alias list for target FEC");
  RegisterProperty("ENSEQM","EXPORTALIASLIST",32,CF_NAME64,CA_READ,"export alias list for target server");
  RegisterProperty("ENSEQM","FECALIASES",512,CF_NAME64,CA_READ,"list of all known FEC aliases");
  RegisterProperty("ENSEQM","EXPORTALIASES",1024,CF_NAME64,CA_READ,"list of all known server aliases");
  RegisterProperty("ENSEQM","DEVICE.SEQUENCE",1024,CF_NAME64,CA_READ,"provides a sequenced device list based on input device name");
  RegisterProperty("ENSEQM","DEVICE.ECHO",64,CF_NAME64,CA_READ,"echos the device list given");

  RegisterLocalAlias("ENSEQM","HERANS");
  RegisterDeviceName("ENSEQM","DEFAULT",DEFAULT_DEV_TAG);
  RegisterDeviceName("ENSEQM","FEC",FEC_DEV_TAG);
  RegisterDeviceName("ENSEQM","EQP",SRV_DEV_TAG);
  RegisterDeviceName("ENSEQM","FEC.EXT",INF_DEV_TAG);

  RegisterUserCommand("ensdebug",NULL,&debug,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("dumpfecs",NULL,&dumpFecFileFlag,NULL,CA_WRITE|CA_READ);
  RegisterUserCommand("dumpeqps",NULL,&dumpEqpFileFlag,NULL,CA_WRITE|CA_READ);

  RegisterUserCommand("forwards",dumpFwdQueues,NULL,NULL,CA_READ);
  RegisterUserCommand("corrections",dumpCorrections,NULL,NULL,CA_READ);
  RegisterUserCommand("pins",dumpFixedAddresses,NULL,NULL,CA_READ);

  getMinFreeDiskSpace(FREESPACE_FILE);
  if (gMininumFreeDiskSpace > 0) 
  {
    feclog("alarm condition with available disk space < %d bytes",gMininumFreeDiskSpace);
    SetFreeBlocksAlarmThreshold(".",gMininumFreeDiskSpace);
  }
  SetPutCommandsInCmdlog(FALSE);
}
#define ENSLOGDEPTH 10000
void nslog(char *text,...)
{
  static char fn[64] = "";
  static char bfn[64] = "";
  FILE *fp=NULL;
  time_t t = time(NULL);
  char str[512];
  long flen;
  int cc=0;
  va_list args;

  if (strlen(fn) == 0) 
  {
    sprintf(fn,"%.32s.log",gFecName);
    sprintf(bfn,"%.32s.bak",gFecName);
    strlwr(fn); strlwr(bfn);
  }
  sprintf(str,"%s[%.16s] ",getDataTimeString((double)t,0),gFecName);
# if defined(UNIX)
  va_start(args,text);
  vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# elif defined(WIN32)
  va_start(args,text);
  vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# else
  if (strlen(text) > 128) 
  {
    if (strchr(text,'%') == NULL)
    { /* then truncate */
      strncpy(&str[strlen(str)],text,256); str[256] = 0;
    }
    ccerr(string_too_long);
  }
  else
  {
    va_start(args,text);
    vsprintf(&str[strlen(str)],text,args);
    va_end(args);
  }
# endif
  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  fseek(fp, 0L, SEEK_END); flen = ftell(fp);
  if (flen > ENSLOGDEPTH*80)
  {
    fclose(fp); remove(bfn); rename(fn,bfn);
    if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  }
  fwrite(str,strlen(str),1,fp);
  if (debug) dbglog(str);
err:
  if (cc) dbglog("nslog: %.32s",erlst[cc]);
  if (fp != NULL) fclose(fp);
  return;
}
void pinglog(char *ctx,char *srv)
{
  static char fn[64] = "ping.log";
  static char bfn[64] = "ping.bak";
  FILE *fp=NULL;
  time_t t = time(NULL);
  char str[512];
  long flen;
  int cc=0;

  sprintf(str,"%s[%.16s] /%.32s/%.32s did not respond to ping\n",getDataTimeString((double)t,0),gFecName,ctx,srv);
  if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  fseek(fp, 0L, SEEK_END); flen = ftell(fp);
  if (flen > ENSLOGDEPTH*80)
  {
    fclose(fp); remove(bfn); rename(fn,bfn);
    if ((fp=fopen(fn,"a")) == NULL) ccerr(file_error);
  }
  fwrite(str,strlen(str),1,fp);
  if (debug) dbglog(str);
err:
  if (cc) dbglog("pinglog: %.32s",erlst[cc]);
  if (fp != NULL) fclose(fp);
  return;
}
int initSinkInfo(void)
{
  DWORD haddr = 0;
  char *ptr = getenv("ENS_FECSINK_ADDR");
  if (ptr != NULL) strncpy(gFecSinkAddr,ptr,16);
  memset(&gFecSink,0,sizeof(FecAddrStruct));
  strcpy(gFecSink.fecName,"FECSINK");
  strncpy(gFecSink.strAdr,gFecSinkAddr,16);
  haddr = inet_addr(gFecSinkAddr);
  memcpy(gFecSink.netAdr.ipv4,&haddr,4);

  gFecSink.inetProtocol = 0x02;
  gFecSink.tineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;

  memset(&gSrvSink,0,sizeof(ExpDataStruct));
  strcpy(gSrvSink.FecName,"FECSINK");
  strcpy(gSrvSink.SubSystem,"TEST"); /* so it won't appear in the FEC Remote Panel */
  strcpy(gSrvSink.EqmName,"SNKEQM");

  memset(&gFecSinkInfo,0,sizeof(FecInfoStruct));
  strcpy(gFecSinkInfo.os,"UNIX");
  strcpy(gFecSinkInfo.desc,"non-existent address sink");
  strcpy(gFecSinkInfo.loc,"nowhere");
  strcpy(gFecSinkInfo.ver,"4.0");
  strcpy(gFecSinkInfo.hdw,"none");
  strcpy(gFecSinkInfo.resp,"no one");

  nslog("using %.16s as FEC sink address",gFecSinkAddr);
  feclog("ENS: using %.16s as FEC sink address",gFecSinkAddr);
  return 0;
}
void sinkSrvAddr(char *ctx,char *srv)
{
  int i;
  char *ss, srvBuffer[EXPORT_NAME_SIZE+4], ctxBuffer[CONTEXT_NAME_SIZE+4];
  nslog("sinkSrvAddr: add sink for /%.32s/%.32s",ctx,srv);
  strncpy(gSrvSink.ExportName,srv,EXPORT_NAME_SIZE);
  if (strcmp(ctx,"DEFAULT") == 0) ctx = "TEST"; /* give it a real context */
  strncpy(gSrvSink.EqmContext,ctx,CONTEXT_NAME_SIZE);
  memcpy(&gTgtSrv,&gSrvSink,sizeof(ExpDataStruct));
  memcpy(&gTgtFec,&gFecSink,sizeof(FecAddrStruct));
  memcpy(&gTgtFecInfo,&gFecSinkInfo,sizeof(FecInfoStruct));
  for (i=0; i<SrvTblSize; i++)
  { /* do I already have this ? */
    if (strnicmp(SrvTbl[i].EqmContext,ctx,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(SrvTbl[i].ExportName,srv,EXPORT_NAME_SIZE)) continue;
    if (!strnicmp(SrvTbl[i].FecName,gTgtFec.fecName,FEC_NAME_SIZE)) 
    {
      nslog("sinkSrvAddr: sink addr for /%.32s/%.32s already entered!",ctx,srv);
      return;
    }
    break;
  }
  i = SrvTblSize;
  ss = gTgtSrv.SubSystem;
  strncpy(srvBuffer,gTgtSrv.ExportName,EXPORT_NAME_SIZE);
  srvBuffer[EXPORT_NAME_SIZE] = 0; strlwr(srvBuffer);
  strncpy(ctxBuffer,gTgtSrv.EqmContext,CONTEXT_NAME_SIZE);
  ctxBuffer[CONTEXT_NAME_SIZE] = 0; strlwr(ctxBuffer);
  if (strcontains(srvBuffer,"test") && !strcontains(ctxBuffer,"test"))
  { /* server name has a 'test' in it, but not in an approved 'test' context */
    if (stricmp(ss,"WDOG")) ss = "TEST";
  }
  strncpy(SrvTbl[i].ExportName,gTgtSrv.ExportName,EXPORT_NAME_SIZE);
  strncpy(SrvTbl[i].FecName,gTgtSrv.FecName,FEC_NAME_SIZE);
  strncpy(SrvTbl[i].EqmName,gTgtSrv.EqmName,EQM_NAME_SHORTSIZE);
  strncpy(SrvTbl[i].EqmContext,gTgtSrv.EqmContext,CONTEXT_NAME_SIZE);
  strncpy(SrvTbl[i].SubSystem,ss,SUBSYSTEM_NAME_SIZE);
  SrvInfoTbl[i].disabled = 0;
  nslog("sinkSrvAddr: new DATA entry - /%.32s/%.32s",gTgtSrv.EqmContext,gTgtSrv.ExportName);
  if (RegisterProperty("ENSEQM",SrvTbl[i].ExportName,
        sizeof(ExpDataStruct)+sizeof(FecAddrStruct),CF_STRUCT,
        0x01,"registered server") < 0)
  {
    nslog("sinkSrvAddr: couldn't register %.32s for export",SrvTbl[i].ExportName);
  }
  SrvTblSize++;
  dumpEqpFileFlag = 1;

  for (i=0; i<FecTblSize; i++)
  { /* do I already have a FEC sink ? */
    if (!strnicmp(FecTbl[i].fecName,gTgtFec.fecName,FEC_NAME_SIZE)) break;
  }
  if (i == FecTblSize) 
  { // TODO: ipv6
    strncpy(FecTbl[i].fecName,gTgtFec.fecName,FEC_NAME_SIZE);
    memcpy(&FecTbl[i].netAdr,&gTgtFec.netAdr,16);
    FecTbl[i].inetProtocol = gTgtFec.inetProtocol;
    strncpy(FecTbl[i].strAdr,gTgtFec.strAdr,ADDR_SIZE_IPV6);
    FecTbl[i].portOffset = gTgtFec.portOffset;
    memcpy(&FecInfoTbl[i],&gTgtFecInfo,sizeof(gTgtFecInfo));
    FecInfoEx[i].disable = 0;
    FecInfoEx[i].importance = 1;
    nslog("sinkSrvAddr: new FEC entry - %.16s",gTgtFec.fecName);
    if (RegisterProperty("ENSEQM",FecTbl[i].fecName,
                         sizeof(FecAddrStruct),CF_STRUCT,
                         0x01,"registered FEC") < 0)
    {
      sprintf(logbuf,"sinkSrvAddr: couldn't register %.16s for export",FecTbl[i].fecName);
      nslog(logbuf);
    }
    FecTblSize++;
    dumpFecFileFlag = 1;
  }
}
void resolveFecAddress(int i)
{ // TODO: ipv6
  DWORD haddr = 0;
  struct hostent *host;

  if (ipLoaded && (host=gethostbyname(FecTbl[i].fecName)) != NULL)
  {
    if (strlen(FecTbl[i].strAdr)) haddr = inet_addr(FecTbl[i].strAdr);
    if (haddr != *(DWORD *)host->h_addr)
    {
      printf("ENS: %.16s using new IP address\n>",FecTbl[i].fecName);
      haddr = *(DWORD *)host->h_addr;
      strncpy(FecTbl[i].strAdr,inet_ntoa(*(struct in_addr *)host->h_addr),ADDR_SIZE);
    }
    memcpy(FecTbl[i].netAdr.ipv4,&haddr,4);
  }
  return;
}
void nsSend(char *txt)
{
  DTYPE din;
  /* this is a bit of a kluge, but since I'm the name server */
  /* I know that NameServer is not used */
  memcpy(NameServer,&gTgtFec,sizeof(FecAddrStruct));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = (UINT32)strlen(txt);
  din.dFormat = CF_TEXT;
  din.data.cptr = txt;
  tExecLink("ENS",gTgtSrv.EqmName,"#0","MESSAGE",NULL,&din,CA_WRITE,100);
}
int verifyFecAddress(int n)
{
  int i,first=0,last=FecTblSize-1;
  if (n >= FecTblSize) return out_of_range;
  if (n >= 0) first = last = n;
  for (i=first; i<=last; i++) resolveFecAddress(i);
  return 0;
}
UINT32 getIPhaddr(char *ip)
{
  UINT32 haddr;
  char *c;
  if (ip == NULL || strlen(ip) == 0) return 0;
  for (c=ip; c != NULL; c++)
  { /* as read from the database, must be of the form a.b.c.d in decimal */
    if (isdigit(*c)) continue;
    if (*c == '.') continue;
    if (*c == 0) break;
    return 0;
  }
  if ((haddr=inet_addr(ip)) == INADDR_NONE) return 0;
  return haddr;
}

char *gImportanceLevel[] = {
  "ALL", "IMPORTANT", "ESSENTIAL", "CRITICAL" };
int nImportanceLevels = sizeof(gImportanceLevel)/sizeof(char *);

int getImportanctLevel(char *importanceAsString)
{
  int i, importance = 0;
  for (i=0; i<nImportanceLevels; i++)
    if (!stricmp(gImportanceLevel[i],importanceAsString)) break;
  if (i < nImportanceLevels) importance = i;
  return importance;
}
#define dbcmp(a,b,n,flg) {\
  if (memcmp(a,b,n)) { memcpy(a,b,n); flg = 1; }\
}
short numRemoteProps;
char *getCsvString(char *s,int length)
{
  char sbuf[128];
  char *c, *start;
  int len;

  if (s == NULL) return s; /* null */
  if (length > 126) length = 126; /* this won't happen */
  if ((c=strchr(s,'\n')) != NULL) *c = 0; /* no new-line char */
  if ((len=(int)strlen(s)) == 0) return s; /* empty */
  if (strchr(s,',') == NULL) return s; /* nothing to worry about */
  for (start=s; isspace(*start) && len > 0; start++, len--); 
  if (*start == '"') return s; /* already a csv string */
  if (len > length-2) len = length-2; /* hardly possible */
  sbuf[0] = '"'; start[len] = 0;
  strncpy(&sbuf[1],start,len+1);
  strcat(sbuf,"\"");
  strncpy(s,sbuf,length);
  return s;
}
int getFecAlias(FecAddrStruct *fec)
{
  int i;
  for (i=0; i<FecTblSize; i++)
  {
    if (!strncmp(FecTbl[i].fecName,fec->fecName,FEC_NAME_SIZE)) continue;
    if (strncmp(FecTbl[i].strAdr,fec->strAdr,ADDR_SIZE_IPV6)) continue;
    if (FecTbl[i].portOffset != fec->portOffset) continue;
    return i;
  }
  return -1;
}
int queryPending = -1,fecTargetId = -1;
void querycallback(int id,int cc) { id = id; queryPending = cc;}
enum { check_idle=0,
       check_tag,
       check_tagaddr,
       check_taginfo,
       check_fec,
       check_fecaddr,
       check_pending,
       check_fecinfo,
       check_finished };
int lastSrvTblIdx, lastFECindex;
void checkExport(void)
{ // TODO: ipv6
  static int i,idx;
  static int nextStep;
  static int srvTblIncremented = FALSE;
  static int matchaddr = FALSE, matchfec = FALSE;
  static time_t interim = 0;
  static char errstr[STATUS_SIZE];
  char fulldevname[128],prpname[PROPERTY_NAME_SIZE+4],scratch[64];
  DTYPE dout;
  int n;

  if (checkExportFlag == check_idle) 
  {
    srvTblIncremented = FALSE;
    return;
  }
  while(checkExportFlag != check_idle)
  {
   switch (checkExportFlag)
   {
    case check_idle: 
      srvTblIncremented = FALSE;
      return;
    case check_tag:
      checkFixedAddr(&gTgtSrv,&matchaddr,&matchfec);
      nslog("check_tag: check export tag /%.32s/%.32s",gTgtSrv.EqmContext,gTgtSrv.ExportName);
      if (matchfec) nslog("check_tag: fec name is pinned");
      if (matchaddr) nslog("check_tag: fec address is pinned");
      for (i=0; i<SrvTblSize; i++)
      {
        if (strlen(gTgtSrv.EqmContext) != 0 && 
            strlen(SrvTbl[i].EqmContext) != 0 &&
            strnicmp(gTgtSrv.EqmContext,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
        if (!strnicmp(SrvTbl[i].ExportName,gTgtSrv.ExportName,EXPORT_NAME_SIZE)) break;
      }
      lastSrvTblIdx = i;
      if (i == SrvTblSize) 
      { /* it's a new one */
        strncpy(SrvTbl[i].ExportName,gTgtSrv.ExportName,EXPORT_NAME_SIZE);
        strncpy(SrvTbl[i].FecName,gTgtSrv.FecName,FEC_NAME_SIZE);
        strncpy(SrvTbl[i].EqmName,gTgtSrv.EqmName,EQM_NAME_SHORTSIZE);
        strncpy(SrvTbl[i].EqmContext,gTgtSrv.EqmContext,CONTEXT_NAME_SIZE);
        if (strlen(gTgtSrv.SubSystem) == 0) 
        { /* as yet, no subsystem defined: any indication within the server name ? */
          strncpy(scratch,gTgtSrv.ExportName,EXPORT_NAME_SIZE); scratch[EXPORT_NAME_SIZE] = 0;
          strupr(scratch);
          if (strstr(scratch,"VAC.") != NULL || strstr(scratch,"VAC_") != NULL)
          {
            strcpy(gTgtSrv.SubSystem,"VAC");
          }
          else if (strstr(scratch,"RF.") != NULL || strstr(scratch,"RF_") != NULL)
          {
            strcpy(gTgtSrv.SubSystem,"RF");
          }
          else if (strstr(scratch,"MAG.") != NULL || strstr(scratch,"MAG_") != NULL)
          {
            strcpy(gTgtSrv.SubSystem,"MAG");
          }
          else if (strstr(scratch,"WATCH.") != NULL || strstr(scratch,".WATCH") != NULL)
          {
            strcpy(gTgtSrv.SubSystem,"WDOG");
          }
          else
          { /* punt */
            strcpy(gTgtSrv.SubSystem,"MISC");
          }
        }
        strncpy(SrvTbl[i].SubSystem,gTgtSrv.SubSystem,SUBSYSTEM_NAME_SIZE);
        SrvInfoTbl[i].disabled = 0;
        dumpEqpFileFlag = 1;
        nslog("check_tag: new DATA entry - /%.32s/%.32s",gTgtSrv.EqmContext,gTgtSrv.ExportName);
        strncpy(prpname,SrvTbl[i].ExportName,EXPORT_NAME_SIZE); prpname[EXPORT_NAME_SIZE] = 0;
        if (RegisterProperty("ENSEQM",prpname,
              sizeof(ExpDataStruct)+sizeof(FecAddrStruct),CF_STRUCT,
              0x01,"registered server") < 0)
        {
          nslog("check_tag: couldn't register %.32s for export",SrvTbl[i].ExportName);
        }
        SrvTblSize++;
        checkExportFlag = check_fec;
        srvTblIncremented = TRUE;
      }
      else                  
      { /* it's in the list */
        if (strlen(gTgtSrv.EqmContext) == 0) 
        { /* not given, copy it back */
          strncpy(gTgtSrv.EqmContext,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE);
          nslog("ENS: supply context %.32s to %.32s",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
        }
        checkExportFlag = check_tagaddr;
      }
      SrvInfoTbl[i].rebootCounts++;
      SrvInfoTbl[i].lastActive = SrvInfoTbl[i].lastPinged = time(NULL);
      if (!dumpEqpFileFlag &&
          strlen(SrvTbl[i].EqmContext) > 0 && 
          strnicmp(SrvTbl[i].EqmContext,"TEST",4) != 0) dumpEqpFileFlag = -1;
      break;
    case check_tagaddr:
      if (debug) 
      {
        dbglog("check_tagaddr: FEC %.16s",gTgtSrv.FecName);
      }
      checkExportFlag = check_taginfo;
      if (forceExportAddr) break;
      if (!strcmp(SrvTbl[i].FecName, "FECSINK")) break; /* will successfully timeout! */
      if (strnicmp(SrvTbl[i].FecName,gTgtSrv.FecName,FEC_NAME_SIZE))
      { /* same tag, new FEC !? */
        nslog("check_tagaddr: found %.16s vs %.16s",SrvTbl[i].FecName,gTgtSrv.FecName);
        if (!stricmp(SrvTbl[i].FecName,"GENS")) matchfec = TRUE;
        if (matchfec)
        { /* but the FEC names must match according to the fixed address rules ! */
          checkExportFlag = check_pending;
          nextStep = check_idle;
          queryPending = 0; /* signal address in use immediately ! */
          sprintf(errstr,"server name /%.32s/%.32s can only be used by FEC %.16s",
                  SrvTbl[i].EqmContext,SrvTbl[i].ExportName,SrvTbl[i].FecName);
          break;
        }
        if (strlen(SrvTbl[i].EqmContext) > 0)
          sprintf(fulldevname,"/%.32s/%.32s/#0",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
        else
          sprintf(fulldevname,"%.32s/#0",SrvTbl[i].ExportName);
        dout.dFormat = CF_SHORT;
        dout.dArrayLength = 1;
        dout.data.sptr = &numRemoteProps;
        fecTargetId = AttachLink(fulldevname,"NPROPS",&dout,NULL,CA_READ,1000,querycallback,CM_SINGLE);
        if (fecTargetId >= 0) 
        {
          idx = conTbl[fecTargetId]->fecIdx;
          if (debug) 
          {
            dbglog("check_tagaddr: check addr of %.16s vs %.16s",FecTbl[idx].fecName,gTgtFec.fecName);
          }
          if (!memcmp(&FecTbl[idx].netAdr,&gTgtFec.netAdr,16) &&
              gTgtFec.portOffset == FecTbl[idx].portOffset)
          { /* but same address ! */
            nslog("check_tagaddr: FEC name change only");
            break; 
          }
          interim = time(NULL);
          queryPending = -1;
          sprintf(errstr,"server name /%.32s/%.32s in use by FEC %.16s",
                  SrvTbl[i].EqmContext,SrvTbl[i].ExportName,SrvTbl[i].FecName);
          checkExportFlag = check_pending;
          nextStep = check_taginfo;
        }
      }
      break;
    case check_taginfo:
      if (debug) 
      {
        dbglog("check_taginfo: %.32s %.16s %.6s",
                SrvTbl[i].EqmContext,SrvTbl[i].FecName,SrvTbl[i].EqmName);
      }
      dbcmp(SrvTbl[i].FecName,gTgtSrv.FecName,sizeof(gTgtSrv.FecName),dumpEqpFileFlag);
      dbcmp(SrvTbl[i].EqmName,gTgtSrv.EqmName,sizeof(gTgtSrv.EqmName),dumpEqpFileFlag);
      dbcmp(SrvTbl[i].EqmContext,gTgtSrv.EqmContext,sizeof(gTgtSrv.EqmContext),dumpEqpFileFlag);
      if (strlen(gTgtSrv.SubSystem) > 0)
        dbcmp(SrvTbl[i].SubSystem,gTgtSrv.SubSystem,sizeof(gTgtSrv.SubSystem),dumpEqpFileFlag);
      checkExportFlag = check_fec;
      break;
    case check_fec:
      if (debug) dbglog("check_fec: %.16s",gTgtFec.fecName);
      for (i=0; i<FecTblSize; i++)
      {
        if (!strnicmp(FecTbl[i].fecName,gTgtFec.fecName,FEC_NAME_SIZE))
        { /* found the fec by name */
          if (!zerocmp(&gTgtFec.netAdr,16)) strcpy(gTgtFec.strAdr,"0.0.0.0");
          FecInfoEx[i].disable = 0;
          if (debug)
          {
            dbglog("check_fec: %.40s:%d vs %.40s:%d\n>",
                   FecTbl[i].strAdr,FecTbl[i].portOffset,
                   gTgtFec.strAdr,gTgtFec.portOffset);
          }
          break;
        }
      }
      if (lastSrvTblIdx >= 0)
      {
        SrvInfoTbl[lastSrvTblIdx].fecIndex = i;
        nslog("check_fec: server /%.32s/%.32s assigned to FEC %.16s", 
          SrvTbl[lastSrvTblIdx].EqmContext,
          SrvTbl[lastSrvTblIdx].ExportName,
          gTgtFec.fecName);
      }
      lastFECindex = i;
      if (i == FecTblSize) 
      { /* it's a new one */
        if ((n=getFecAlias(&gTgtFec)) >= 0)
        {
          if (gCanSendAddrInUse) 
          {
            sprintf(errstr,"FEC has an alias entry with name %.16s",FecTbl[n].fecName);
            nsSend(errstr);
          }
        }
        strncpy(FecTbl[i].fecName,gTgtFec.fecName,FEC_NAME_SIZE);
        memcpy(&FecTbl[i].netAdr,&gTgtFec.netAdr,16);
        strncpy(FecTbl[i].strAdr,gTgtFec.strAdr,ADDR_SIZE_IPV6);
        FecTbl[i].inetProtocol = getProtocolFromAddressString(FecTbl[i].strAdr);
        //memcpy(FecTbl[i].netAdr.ipv4,gTgtFec.netAdr.ipv4,4);
        FecTbl[i].portOffset = gTgtFec.portOffset;
        memcpy(FecTbl[i].lclAdr.mac,gTgtFec.lclAdr.mac,6);
        getCsvString(gTgtFecInfo.os,12);
        memcpy(FecInfoTbl[i].os,gTgtFecInfo.os,FEC_OS_SIZE);
        getCsvString(gTgtFecInfo.desc,FEC_DESC_SIZE);
        memcpy(FecInfoTbl[i].desc,gTgtFecInfo.desc,FEC_DESC_SIZE);
        getCsvString(gTgtFecInfo.loc,FEC_LOCATION_SIZE);
        memcpy(FecInfoTbl[i].loc,gTgtFecInfo.loc,FEC_LOCATION_SIZE);
        getCsvString(gTgtFecInfo.hdw,FEC_HDW_SIZE);
        memcpy(FecInfoTbl[i].hdw,gTgtFecInfo.hdw,FEC_HDW_SIZE);
        getCsvString(gTgtFecInfo.resp,FEC_RESP_SIZE);
        memcpy(FecInfoTbl[i].resp,gTgtFecInfo.resp,FEC_RESP_SIZE);
        getCsvString(gTgtFecInfo.ver,FEC_VERSION_SIZE);
        memcpy(FecInfoTbl[i].ver,gTgtFecInfo.ver,FEC_VERSION_SIZE);
        FecInfoEx[i].disable = 0;
        FecInfoEx[i].importance = 0;
        dumpFecFileFlag = 1;
        nslog("check_fec: new FEC entry - %.16s",gTgtFec.fecName);
        strncpy(prpname,FecTbl[i].fecName,FEC_NAME_SIZE); prpname[FEC_NAME_SIZE] = 0;
        if (RegisterProperty("ENSEQM",prpname,
                             sizeof(FecAddrStruct),CF_STRUCT,
                             0x01,"registered FEC") < 0)
        {
          nslog("check_fec: couldn't register %.16s for export",FecTbl[i].fecName);
        }
        FecTblSize++;
        matchaddr = matchfec = FALSE;
        checkExportFlag = check_idle; /* finished */
        forceExportAddr = FALSE;
        if (debug) dbglog("reload eqp tbl %d, reload fec tbl %d",dumpEqpFileFlag,dumpFecFileFlag);
      }
      else checkExportFlag = check_fecaddr;
      break;
    case check_fecaddr:
      if (debug) dbglog("check_fecaddr: %.16s %.40s",gTgtFec.fecName,gTgtFec.strAdr);
      checkExportFlag = check_fecinfo;
      if (forceExportAddr) break;
      if (memcmp(&FecTbl[i].netAdr,&gTgtFec.netAdr,16) ||
          FecTbl[i].portOffset != gTgtFec.portOffset)  
      { /* different address */
        if (isIpv4on6(FecTbl[i].strAdr) && !isIpv4on6(gTgtFec.strAdr))
        {
          nslog("check_fecaddr: move IPv4 mapped to IPv6 address on FEC %.16s",gTgtFec.fecName);
          break;
        }
        if (!isIpv4on6(FecTbl[i].strAdr) && isIpv4on6(gTgtFec.strAdr))
        {
          nslog("check_fecaddr: re-instate IPv4 mapped to IPv6 address on FEC %.16s",gTgtFec.fecName);
          break;
        }
        for (n=0; n<SrvTblSize; n++)
          if (!strnicmp(SrvTbl[n].FecName,gTgtFec.fecName,FEC_NAME_SIZE)) break;
        if (n < SrvTblSize)          
        { /* same name, different address and there are eqm pointers to FEC */
          nslog("check_fecaddr: found %.40s vs %.40s",FecTbl[i].strAdr,gTgtFec.strAdr);
          if (matchaddr)
          { /* but the FEC names must match according to the fixed address rules ! */
            checkExportFlag = check_pending;
            nextStep = check_idle;
            queryPending = 0;
            sprintf(errstr,"FEC name %.16s can only be used by address %.40s [%d]",
                    FecTbl[i].fecName,FecTbl[i].strAdr,FecTbl[i].portOffset);
            break;
          }
          if (strlen(SrvTbl[n].EqmContext) > 0)
            sprintf(fulldevname,"/%.32s/%.32s/#0",SrvTbl[n].EqmContext,SrvTbl[n].ExportName);
          else
            sprintf(fulldevname,"%.32s/#0",SrvTbl[n].ExportName);
          dout.dFormat = CF_SHORT;
          dout.dArrayLength = 1;
          dout.data.sptr = &numRemoteProps;
          fecTargetId = AttachLink(fulldevname,"NPROPS",&dout,NULL,CA_READ,1000,querycallback,CM_SINGLE);
          if (fecTargetId >= 0)
          {
            interim = time(NULL);
            queryPending = -1;
            sprintf(errstr,"FEC name %.16s in use by %.40s [%d]",
                    FecTbl[i].fecName,FecTbl[i].strAdr,FecTbl[i].portOffset);
            checkExportFlag = check_pending;
            nextStep = check_fecinfo;
          }
        }
      }
      break;
    case check_pending:
      if (debug) 
      {
        dbglog("check_pending: waiting for reply");
      }
      while (queryPending < 0 && time(NULL) < interim + 10) SystemDelay(500);
      if (queryPending != 0 ) 
      { /* this is okay (old address doesn't answer or non existent elem */
        checkExportFlag = nextStep;
        break;
      }
      nslog("check_pending: %.192s",errstr);
      if (gCanSendAddrInUse) nsSend(errstr);
      nslog(errstr);
      dumpFecFileFlag = 0; dumpEqpFileFlag = 0;
      matchaddr = matchfec = FALSE;
      checkExportFlag = check_idle; /* old address still in use */
      forceExportAddr = FALSE;
      if (srvTblIncremented)
      {
        SrvTblSize--;
        nslog("check_pending: discard new server entry /%.32s/%.32s",SrvTbl[SrvTblSize].EqmContext, SrvTbl[SrvTblSize].ExportName);
        srvTblIncremented = FALSE;
      }
      break;
    case check_fecinfo:
      if (debug) dbglog("check_fecinfo: final step");
      dbcmp(&FecInfoEx[i].ipv4,&gTgtFecInfoEx.ipv4,16,dumpFecFileFlag);
      dbcmp(&FecTbl[i].netAdr,&gTgtFec.netAdr,16,dumpFecFileFlag);
      dbcmp(FecTbl[i].strAdr,gTgtFec.strAdr,ADDR_SIZE_IPV6,dumpFecFileFlag);
      dbcmp(&FecTbl[i].portOffset,&gTgtFec.portOffset,4,dumpFecFileFlag);
      dbcmp(&FecTbl[i].lclAdr.mac,&gTgtFec.lclAdr.mac,6,dumpFecFileFlag);
      if (FecTbl[i].addrFlags != gTgtFec.addrFlags)
      {
        nslog("FEC %.16s was MCAST Rules %s, now MCAST Rules %s",FecTbl[i].fecName,
          ((FecTbl[i].addrFlags&ADDRFLAG_MCAST_MODERN)?"modern":"legacy"),
          ((gTgtFec.addrFlags&ADDRFLAG_MCAST_MODERN)?"modern":"legacy"));
      }
      dbcmp(&FecTbl[i].addrFlags,&gTgtFec.addrFlags,4,dumpFecFileFlag);
      dbcmp(FecInfoTbl[i].os,gTgtFecInfo.os,FEC_OS_SIZE,dumpFecFileFlag);
      if (gTgtFecInfo.desc[0] != '[' && FecInfoTbl[i].desc[0] == '[')
      {  /* keep assigned sub system and description ! */
        strncpy(gTgtFecInfo.desc,FecInfoTbl[i].desc,FEC_DESC_SIZE);
      }
      if (!strnicmp(FecInfoTbl[i].desc,"[XXX]",5))
      {  /* renegade FEC developer has mis-categorized his FEC ! */
        strncpy(gTgtFecInfo.desc,FecInfoTbl[i].desc,FEC_DESC_SIZE);
      }
      getCsvString(gTgtFecInfo.desc,FEC_DESC_SIZE);
      dbcmp(FecInfoTbl[i].desc,gTgtFecInfo.desc,FEC_DESC_SIZE,dumpFecFileFlag);
      if (gTgtFecInfo.loc[0] == 0) strncpy(gTgtFecInfo.loc,FecInfoTbl[i].loc,FEC_LOCATION_SIZE);
      getCsvString(gTgtFecInfo.loc,FEC_LOCATION_SIZE);
      dbcmp(FecInfoTbl[i].loc,gTgtFecInfo.loc,FEC_LOCATION_SIZE,dumpFecFileFlag);
      getCsvString(gTgtFecInfo.hdw,FEC_HDW_SIZE);
      dbcmp(FecInfoTbl[i].hdw,gTgtFecInfo.hdw,FEC_HDW_SIZE,dumpFecFileFlag);
      if (gTgtFecInfo.resp[0] == 0) strncpy(gTgtFecInfo.resp,FecInfoTbl[i].resp,FEC_RESP_SIZE);
      getCsvString(gTgtFecInfo.resp,FEC_RESP_SIZE);
      dbcmp(FecInfoTbl[i].resp,gTgtFecInfo.resp,FEC_RESP_SIZE,dumpFecFileFlag);
      getCsvString(gTgtFecInfo.ver,FEC_VERSION_SIZE);
      dbcmp(FecInfoTbl[i].ver,gTgtFecInfo.ver,FEC_VERSION_SIZE,dumpFecFileFlag);
      if (gTgtFecInfo.ver[0] == 0) strncpy(gTgtFecInfo.ver,FecInfoTbl[i].ver,FEC_VERSION_SIZE);
      matchaddr = matchfec = FALSE;
      checkExportFlag = check_idle; /* completely finished */
      forceExportAddr = FALSE;
      srvTblIncremented = FALSE;
      if (debug)
      {
        dbglog("reload eqp tbl %d, reload fec tbl %d",dumpEqpFileFlag,dumpFecFileFlag);
      }
      break;
   default:
     matchaddr = matchfec = FALSE;
     checkExportFlag = check_idle; /* turn off the check flag */
     forceExportAddr = FALSE;
     srvTblIncremented = FALSE;
  }
  SystemDelay(20);
 }
 matchaddr = matchfec = FALSE;
 if (debug) dbglog("check_finished: %.32s",gTgtSrv.ExportName);
 nslog("checkExport: finished checking %.32s",gTgtSrv.ExportName);
}
void fillinFecIndexTable(void)
{
  int i,j;
  for (i=0; i<SrvTblSize; i++)
  {
    for (j=0; j<FecTblSize; j++)
    {
      if (!strnicmp(SrvTbl[i].FecName,FecTbl[j].fecName,FEC_NAME_SIZE)) 
      {
        SrvInfoTbl[i].fecIndex = (short)j;
        /* if (strlen(SrvTbl[i].SubSystem) == 0) what was the idea here ? */
        break;
      }
    }
    if (j == FecTblSize)
    {
      SrvInfoTbl[i].disabled = TRUE;
      nslog("fillinFecIndexTable: orphaned server %.32s -> no fec entry %.16s found!",SrvTbl[i].ExportName,SrvTbl[i].FecName);
    }
  }
}
int getFecTblIndex(char *target)
{ /* internal: target never NULL */
  int i;
  for (i=0; i<FecTblSize; i++)
  { 
    if (FecInfoEx[i].disable) continue;
    if (!strnicmp(target,FecTbl[i].fecName,FEC_NAME_SIZE)) return i;
  }
  return -1;
}
int fecadrcmp(FecAddrStruct *fec1,FecAddrStruct *fec2)
{
  if (fec1 == fec2) return TRUE;
  if (!strnicmp(fec1->fecName,fec2->fecName,FEC_NAME_SIZE)) return TRUE;
  if (strncmp(fec1->strAdr,fec2->strAdr,40)) return FALSE;
  if (fec1->portOffset != fec2->portOffset) return FALSE;
  return TRUE;

}
extern int feccmp(int id1,int id2);
int getFecAliases(NAME64 *lst,int *lstSize)
{
  BYTE *skip;
  int i, k, n=0, na, siz;
  if (lst == NULL || lstSize == NULL) return invalid_parameter;
  if ((skip=alloca(FecTblSize)) != NULL) memset(skip,0,FecTblSize);
  siz = *lstSize;
  for (i=0; i<FecTblSize; i++)
  { /* scan thru for any aliases */
    if (FecInfoEx[i].disable) continue;
    if (skip != NULL && skip[i] != 0) continue;
    if (n > siz - 2) goto err;
    sprintf(lst[n].name,"%.40s:%d",FecTbl[i].strAdr,FecTbl[i].portOffset);
    strncpy(lst[n+1].name,FecTbl[i].fecName,FEC_NAME_SIZE);
    for (k=i+1,na=0; k<FecTblSize; k++)
    {
      if (FecInfoEx[k].disable) continue;
      if (fecadrcmp(&FecTbl[i],&FecTbl[k]))
      { /* found a match */
        if (na++ == 0) n += 2;
        if (n > siz - 1) goto err; 
        strncpy(lst[n++].name,FecTbl[k].fecName,FEC_NAME_SIZE);
        if (skip != NULL) skip[k] = 1;
      }
    }
  }
err:
  *lstSize = n;
  return 0;
}
int getFecAliasList(char *target,NAME16 *lst,int *lstSize)
{
  int i, n=0, siz;
  int tgtFecIdx;
  if (target == NULL || strlen(target) == 0) return invalid_parameter;
  if (lst == NULL || lstSize == NULL) return invalid_parameter;
  siz = *lstSize;
  if ((tgtFecIdx=getFecTblIndex(target)) < 0) return non_existent_fec;
  for (i=0; i<FecTblSize && n < siz; i++)
  { /* scan thru for the aliases (omit the target) */
    if (i == tgtFecIdx) continue;
    if (FecInfoEx[i].disable) continue;
    if (strncmp(FecTbl[tgtFecIdx].strAdr,FecTbl[i].strAdr,40)) continue;
    if (FecTbl[tgtFecIdx].portOffset != FecTbl[i].portOffset) continue;
    strncpy(lst[n++].name,FecTbl[i].fecName,16);
  }
  *lstSize = n;
  return 0;
}
int getSrvAliases(NAME64 *lst,int *lstSize)
{
  int i, k, fi, fk, n=0, na, siz;
  BYTE *skip;
  
  if (lst == NULL || lstSize == NULL) return invalid_parameter;
  if ((skip=alloca(SrvTblSize)) != NULL) memset(skip,0,SrvTblSize);
  siz = *lstSize;

  for (i=0; i<SrvTblSize; i++)
  {
    if (SrvInfoTbl[i].disabled) continue;
    if (skip != NULL && skip[i] != 0) continue;
    if (n > siz - 2) goto err;
    fi = SrvInfoTbl[i].fecIndex;
    sprintf(lst[n].name,"%.6s@%.40s:%d",SrvTbl[i].EqmName,FecTbl[fi].strAdr,FecTbl[fi].portOffset);
    sprintf(lst[n+1].name,"/%.20s/%.32s",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
    for (k=i+1,na=0; k<SrvTblSize; k++)
    {
      if (SrvInfoTbl[k].disabled) continue;
      fk = SrvInfoTbl[k].fecIndex;
      if (fk != fi && strnicmp(FecTbl[fi].strAdr,FecTbl[fk].strAdr,40)) continue;
      if (fk != fi && FecTbl[fi].portOffset != FecTbl[fk].portOffset) continue;
      if (!strnicmp(SrvTbl[i].EqmName,SrvTbl[k].EqmName,EQM_NAME_SHORTSIZE))
      {
        if (na++ == 0) n += 2;
        if (n > siz - 1) goto err; 
        sprintf(lst[n++].name,"/%.20s/%.32s",SrvTbl[k].EqmContext,SrvTbl[k].ExportName);
        if (skip != NULL) skip[k] = 1;
      }
    }
  }
err:
  *lstSize = n;
  return 0;
}
int getSrvAliasList(char *tgtCtx,char *tgtSrv,NAME64 *lst,int *lstSize)
{
  int i, k, n=0, siz;
  int tgtSrvIdx=-1, tgtFecIdx=-1;
  if (tgtSrv == NULL || strlen(tgtSrv) == 0) return invalid_parameter;
  if (tgtCtx == NULL || strlen(tgtCtx) == 0) return invalid_parameter;
  if (lst == NULL || lstSize == NULL) return invalid_parameter;
  siz = *lstSize;
  for (i=0; i<SrvTblSize; i++)
  { /* scan thru once to find the target */
    if (SrvInfoTbl[i].disabled) continue;
    if (strnicmp(tgtCtx,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(tgtSrv,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) continue;
    if ((tgtFecIdx=SrvInfoTbl[i].fecIndex) < 0 || tgtFecIdx > FecTblSize) return non_existent_fec;
    tgtSrvIdx = i;
    break;
  }
  if (tgtFecIdx == -1) return non_existent_fec;
  if (i == SrvTblSize) return non_existent_server;
  for (i=0; i<SrvTblSize && n < siz; i++)
  { /* scan thru for the aliases (omit the target) */
    if (i == tgtSrvIdx) continue;
    if (SrvInfoTbl[i].disabled) continue;
    if (strnicmp(SrvTbl[tgtSrvIdx].EqmName,SrvTbl[i].EqmName,EQM_NAME_SHORTSIZE)) continue;
    k = SrvInfoTbl[i].fecIndex;
    if (strncmp(FecTbl[tgtFecIdx].strAdr,FecTbl[k].strAdr,40)) continue;
    if (FecTbl[tgtFecIdx].portOffset != FecTbl[k].portOffset) continue;
    sprintf(lst[n++].name,"/%.30s/%.32s",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
  }
  *lstSize = n;
  return 0;
}
int getServerTableIndex(char *ctx,char *srv)
{
  int i;
  if (srv == NULL) return -1;
  if (ctx == NULL) ctx = "";
  for (i=0; i<SrvTblSize; i++)
  {
    if (SrvInfoTbl[i].disabled) continue;
    if (strlen(ctx) > 0 && 
        strlen(SrvTbl[i].EqmContext) != 0 &&
        strnicmp(ctx,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(SrvTbl[i].ExportName,srv,EXPORT_NAME_SIZE)) return i;
  }
  return -1;
}
int addExportAlias(char *target,char *alias)
{
  char ctxTgt[128], *expTgt, ctxAli[128], *expAli;
  int i, idxTgt;

  if (target == NULL || alias == NULL) return invalid_parameter;
  if (target[0] != '/') return illegal_context;
  strncpy(ctxTgt,&target[1],CONTEXT_NAME_SIZE);
  if ((expTgt=strchr(ctxTgt,'/')) == NULL) return invalid_name;
  *expTgt++ = 0;
  if (alias[0] != '/') 
  {
    strncpy(ctxAli,ctxTgt,CONTEXT_NAME_SIZE);
    ctxAli[CONTEXT_NAME_SIZE] = 0;
    expAli = alias;
  }
  else
  {
    strncpy(ctxAli,&alias[1],CONTEXT_NAME_SIZE);
    if ((expAli=strchr(ctxAli,'/')) == NULL) return invalid_name;
    *expAli++ = 0;
  }
  nslog("addExportAlias: add export alias /%.32s/%.32s for /%.32s/%.32s",ctxAli,expAli,ctxTgt,expTgt);
  for (i=0; i<SrvTblSize; i++)
  { /* find the target */
    if (strlen(SrvTbl[i].EqmContext) != 0 &&
        strnicmp(ctxTgt,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(SrvTbl[i].ExportName,expTgt,EXPORT_NAME_SIZE)) break;
  }
  if (i == SrvTblSize) 
  {
    nslog("addExportAlias: alias target %.32s is not a registered server",expTgt);
    return non_existent_server;
  }
  idxTgt = i;
  for (i=0; i<SrvTblSize; i++)
  { /* find the alias (if it's there) */
    if (strlen(SrvTbl[i].EqmContext) != 0 &&
        strnicmp(ctxAli,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(SrvTbl[i].ExportName,expAli,EXPORT_NAME_SIZE)) break;
  }
  if (i == SrvTblSize) 
  { /* it's a new one */
    strncpy(SrvTbl[i].ExportName,expAli,EXPORT_NAME_SIZE);
    strncpy(SrvTbl[i].FecName,SrvTbl[idxTgt].FecName,FEC_NAME_SIZE);
    strncpy(SrvTbl[i].EqmName,SrvTbl[idxTgt].EqmName,EQM_NAME_SHORTSIZE);
    strncpy(SrvTbl[i].EqmContext,ctxAli,CONTEXT_NAME_SIZE);
    strncpy(SrvTbl[i].SubSystem,SrvTbl[idxTgt].SubSystem,SUBSYSTEM_NAME_SIZE);
    SrvInfoTbl[i].disabled = 0;
    dumpEqpFileFlag = 1;
    nslog("addExportAlias: new alias DATA entry - /%.32s/%.32s",ctxAli,expAli);
    if (RegisterProperty("ENSEQM",expAli,
          sizeof(ExpDataStruct)+sizeof(FecAddrStruct),CF_STRUCT,
          0x01,"registered server") < 0)
    {
      nslog("addExportAlias: couldn't register %.32s for export",expAli);
    }
    SrvTblSize++;
  }
  else                  
  { /* it's in the list */
    if (i == idxTgt) 
    {
      nslog("addExportAlias: alias %.32s maps to itself %.32s",expAli,expTgt);
      return not_allowed;
    }
    if (SrvInfoTbl[i].rebootCounts > 0)
    {
      nslog("addExportAlias: alias %.32s refused (existing table entry is a registered server)",expAli);
      return not_allowed;
    }
    if (strnicmp(SrvTbl[i].FecName,SrvTbl[idxTgt].FecName,FEC_NAME_SIZE))
    {
      strncpy(SrvTbl[i].FecName,SrvTbl[idxTgt].FecName,FEC_NAME_SIZE);
      dumpEqpFileFlag = 1;
    }
    if (strnicmp(SrvTbl[i].EqmName,SrvTbl[idxTgt].EqmName,EQM_NAME_SHORTSIZE))
    {
      strncpy(SrvTbl[i].EqmName,SrvTbl[idxTgt].EqmName,EQM_NAME_SHORTSIZE);
      dumpEqpFileFlag = 1;
    }
    if (strncmp(SrvTbl[i].EqmContext,ctxAli,CONTEXT_NAME_SIZE))
    {
      strncpy(SrvTbl[i].EqmContext,ctxAli,CONTEXT_NAME_SIZE);
      dumpEqpFileFlag = 1;
    }
    if (strnicmp(SrvTbl[i].SubSystem,SrvTbl[idxTgt].SubSystem,SUBSYSTEM_NAME_SIZE))
    {
      strncpy(SrvTbl[i].SubSystem,SrvTbl[idxTgt].SubSystem,SUBSYSTEM_NAME_SIZE);
      dumpEqpFileFlag = 1;
    }
    SrvInfoTbl[i].disabled = 0;
  }
  return 0;
}

int nAllowedDecorations = 0;
NAME16 *gAllowedDecorations = NULL;
int nAllowedContexts = 0;
NAME32 *gAllowedContexts = NULL;
#define ROOTCTX_FNAME "root_contexts.csv"
#define MAX_ALLOWED_CONTEXTS 100

int getAllowedContexts(char *fn)
{
  FILE *fp = NULL;
  int i,n,cc=0,done=0;
  int ctx_col=-1;
  char str[256],hdr[256];

  if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
  if (n > MAX_ALLOWED_CONTEXTS) n = MAX_ALLOWED_CONTEXTS;
  if ((gAllowedContexts=(NAME32 *)calloc(MAX_ALLOWED_CONTEXTS,sizeof(NAME32))) == NULL)
    ccerr(out_of_local_memory);

  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<MAX_ALLOWED_CONTEXTS;)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((ctx_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(gAllowedContexts[i].name,colptr(ctx_col,str),32);
    i++;
  }
  nAllowedContexts = i;
err:
  if (fp) fclose(fp);
  return cc;
}
int writeAllowedContexts(char *fn)
{
  FILE *fp = NULL;
  int i, cc = 0;

  if ((fp=fopen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"CONTEXT\n");
  for (i=0; i<nAllowedContexts; i++)
  {
    fprintf(fp,"%.32s\n",gAllowedContexts[i].name);
  }
err:
  if (fp) fclose(fp);
  return cc;
}
int addAllowedContext(char *ctx)
{
  int i;

  if (gAllowedContexts == NULL &&
      (gAllowedContexts=(NAME32 *)calloc(MAX_ALLOWED_CONTEXTS,sizeof(NAME32))) == NULL)
  {
    return out_of_local_memory;
  }
  for (i=0; i<nAllowedContexts; i++)
  {
    if (!strnicmp(ctx,gAllowedContexts[i].name,32)) return already_assigned;
  }
  if (i < MAX_ALLOWED_CONTEXTS)
  {
    strncpy(gAllowedContexts[i].name,ctx,32);
    nAllowedContexts++;
    return 0;
  }
  return resources_exhausted;
}
int rmvAllowedContext(char *ctx)
{
  int i,rmv_index=-1;

  if (gAllowedContexts == NULL) return un_allocated;
  for (i=0; i<nAllowedContexts; i++)
  {
    if (!strnicmp(ctx,gAllowedContexts[i].name,32))
    {
      rmv_index = i;
      break;
    }
  }
  if (rmv_index < 0) return un_allocated;
  for (i=rmv_index+1; i<nAllowedContexts; i++)
  {
    strncpy(gAllowedContexts[i-1].name,gAllowedContexts[i].name,32);
  }
  nAllowedContexts--;
  return 0;
}
#define MAX_FIXED_ADDRESSES 256
#define FIXADR_FNAME "fixed.csv"
int nFixedAddr = 0;
typedef struct
{
  char eqm[EQM_NAME_SIZE];
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char fec[FEC_NAME_SIZE];
} FixedAddr;
FixedAddr *gFixedAddr = NULL;
int getFixedAddresses(char *fn)
{
  FILE *fp = NULL;
  int i,n,cc=0,done=0;
  int ctx_col=-1,srv_col=-1,eqm_col=-1,fec_col=-1;
  char str[256],hdr[256];

  if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
  if ((gFixedAddr=(FixedAddr *)calloc(n,sizeof(FixedAddr))) == NULL)
    ccerr(out_of_local_memory);

  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL  && i<MAX_FIXED_ADDRESSES;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      srv_col = findcol(hdr,"SERVER");
      ctx_col = findcol(hdr,"CONTEXT");
      eqm_col = findcol(hdr,"EQP_MODULE");
      fec_col = findcol(hdr,"FEC_NAME");
      done = 1;
      continue;
    }
    if (srv_col < 0 && ctx_col < 0 && eqm_col < 0 && fec_col < 0)
    { /* empty ! */
      continue;
    }
    if (srv_col >= 0) strncpy(gFixedAddr[i].srv,colptr(srv_col,str),EXPORT_NAME_SIZE);
    if (ctx_col >= 0) strncpy(gFixedAddr[i].ctx,colptr(ctx_col,str),CONTEXT_NAME_SIZE);
    if (eqm_col >= 0) strncpy(gFixedAddr[i].eqm,colptr(eqm_col,str),EQM_NAME_SIZE);
    if (fec_col >= 0) strncpy(gFixedAddr[i].fec,colptr(fec_col,str),FEC_NAME_SIZE);
    i++;
  }
  nFixedAddr = i;
err:
  if (fp) fclose(fp);
  return cc;
}
void checkFixedAddr(ExpDataStruct *srv,int *matchaddr, int *matchfec)
{ /* used locally -> don't need to check NULL pointers ... */
  int i;

  *matchaddr = FALSE; *matchfec = FALSE;
  for (i=0; i<nFixedAddr; i++)
  {
    if ((strlen(gFixedAddr[i].ctx) > 0) && strnicmp(gFixedAddr[i].ctx,srv->EqmContext,CONTEXT_NAME_SIZE))
    { /* context given and not equal -> don't have to match this one at all */
      continue;
    }
    if ((strlen(gFixedAddr[i].eqm) > 0) && !strncmp(gFixedAddr[i].eqm,srv->EqmName,EQM_NAME_SHORTSIZE))
    { /* eqm given and is equal -> must match everywhere */
      *matchaddr = TRUE;
      *matchfec = TRUE;
      return;
    }
    if ((strlen(gFixedAddr[i].srv) > 0) && !strnicmp(gFixedAddr[i].srv,srv->ExportName,EXPORT_NAME_SIZE))
    { /* server name given and is equal -> must match fec */
      *matchfec = TRUE;
    }
    if ((strlen(gFixedAddr[i].fec) > 0) && !strnicmp(gFixedAddr[i].fec,srv->FecName,EXPORT_NAME_SIZE))
    { /* FEC name given and is equal -> must match address */
      *matchaddr = TRUE;
    }
    if (*matchfec == TRUE && *matchaddr == TRUE) break; /* no need to keep looking */
  }
}
#define MAX_EXPORT_CORRECTIONS 256
#define EXPCOR_FNAME "corrections.csv"
int nExportCorrections = 0;
typedef struct
{
  char ctx[CONTEXT_NAME_SIZE];
  char srv[EXPORT_NAME_SIZE];
  char sub[SUBSYSTEM_NAME_SIZE];
  char rsp[FEC_RESP_SIZE];
} ExportCorrection;
ExportCorrection *gExportCorrections = NULL;
int getExportCorrections(char *fn)
{
  FILE *fp = NULL;
  int i,n,cc=0,done=0;
  int ctx_col=-1,srv_col=-1,sub_col=-1,rsp_col=-1;
  char str[256],hdr[256];

  if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
  if ((gExportCorrections=(ExportCorrection *)calloc(n,sizeof(ExportCorrection))) == NULL)
    ccerr(out_of_local_memory);

  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<MAX_EXPORT_CORRECTIONS;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((srv_col=findcol(hdr,"SERVER")) < 0) ccerr(no_such_line);
      ctx_col = findcol(hdr,"CONTEXT");
      sub_col = findcol(hdr,"SUBSYSTEM");
      rsp_col = findcol(hdr,"RESPONSIBLE");
      done = 1;
      continue;
    }
    strncpy(gExportCorrections[i].srv,colptr(srv_col,str),EXPORT_NAME_SIZE);
    if (ctx_col >= 0) strncpy(gExportCorrections[i].ctx,colptr(ctx_col,str),CONTEXT_NAME_SIZE);
    if (sub_col >= 0) strncpy(gExportCorrections[i].sub,colptr(sub_col,str),SUBSYSTEM_NAME_SIZE);
    if (rsp_col >= 0) strncpy(gExportCorrections[i].rsp,colptr(rsp_col,str),FEC_RESP_SIZE);
    if (strlen(gExportCorrections[i].sub) == 0 && 
        strlen(gExportCorrections[i].rsp) == 0) continue; /* nothing to correct */
    i++;
  }
  nExportCorrections = i;
err:
  if (fp) fclose(fp);
  return cc;
}
void applyExportCorrections(FecInfoStruct *inf, ExpDataStruct *srv)
{
  int i;
  
  for (i=0; i<nExportCorrections; i++)
  {
    if (strnicmp(gExportCorrections[i].srv,srv->ExportName,EXPORT_NAME_SIZE)) continue;
    if (strlen(gExportCorrections[i].ctx) > 0 && 
      strnicmp(gExportCorrections[i].ctx,srv->EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (strlen(gExportCorrections[i].sub) > 0)
    { /* pin the subystem to the correction provided (no matter what) */
      strncpy(srv->SubSystem,gExportCorrections[i].sub,SUBSYSTEM_NAME_SIZE);
    }
    if (strlen(gExportCorrections[i].rsp) > 0)
    { /* pin the responsible list to the correction provided (no matter what) */
      strncpy(inf->resp,gExportCorrections[i].rsp,FEC_RESP_SIZE);
    }
    return;
  }
}
int dumpFixedAddresses(int i1,int i2, int i3, int i4)
{
  int i;
  char *ctx, *srv, *eqm, *fec;
  i1 = i1; i2 = i2; i3 = i3; i4 = i4;
  if (nFixedAddr == 0)
  {
    ttyoutput("there are no pinned addresses configured");
    return 0;
  }
  ttyoutput("pinned servers with a configured address are checked according");
  ttyoutput("to the following address criteria:");
  for (i=0; i<nFixedAddr; i++)
  {
    ctx = gFixedAddr[i].ctx;
    if (strlen(ctx) == 0) ctx = "<any>";
    eqm = gFixedAddr[i].eqm;
    srv = gFixedAddr[i].srv;
    fec = gFixedAddr[i].fec;
    if (strlen(eqm) > 0)
    {
      ttyoutput("servers with EQM name %.6s (context %.32s): pinned to the reg. FEC name and addr",eqm,ctx);
      continue;
    }
    if (strlen(srv) > 0) ttyoutput("servers with export name %.32s (context %.32): pinned to FEC name",srv,ctx);
    if (strlen(fec) > 0) ttyoutput("FECs with FEC name %.16s: pinned to FEC addr",fec);
  }
  return 0;
}
int dumpCorrections(int i1,int i2, int i3, int i4)
{
  int i;
  char *ctx, *subs, *rsp;
  i1 = i1; i2 = i2; i3 = i3; i4 = i4;
  if (nExportCorrections == 0)
  {
    ttyoutput("there is no export correction information registered");
    return 0;
  }
  ttyoutput("export corrections:");
  for (i=0; i<nExportCorrections; i++)
  {
    ctx = gExportCorrections[i].ctx;
    if (strlen(ctx) == 0) ctx = "<all>";
    subs = gExportCorrections[i].sub;
    if (strlen(subs) == 0) subs = "<as registered>";
    rsp = gExportCorrections[i].rsp;
    if (strlen(rsp) == 0) rsp = "<as registered>";
    ttyoutput("server /%.32s/%.32s : subsystem -> %.16s, responsible -> %.16s",
      ctx,gExportCorrections[i].srv,subs,rsp);
  }
  return 0;
}
#define MAX_SUBSYSTEMS 64
int nSubSystems = 0;
NAME16 subSystems[MAX_SUBSYSTEMS];
typedef struct
{
  char name[16];
  char desc[64];
  char alias1[16];
  char alias2[16];
} SubSystemType;
SubSystemType *gAllowedSubsystems = NULL;
int nAllowedSubsystems = 0;
#define SUBSYS_FNAME "subsystems.csv"
int getAllowedSubsystems(char *fn)
{
  FILE *fp = NULL;
  int i,n,cc=0,done=0;
  int name_col=-1,desc_col=-1,cmp1_col=-1,cmp2_col=-1;
  char str[256],hdr[256];

  if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
  if ((gAllowedSubsystems=(SubSystemType *)calloc(n,sizeof(SubSystemType))) == NULL)
    ccerr(out_of_local_memory);

  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((name_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_line);
      cmp1_col = findcol(hdr,"ALIAS1");
      cmp2_col = findcol(hdr,"ALIAS2");
      desc_col = findcol(hdr,"DESCRIPTION");
      done = 1;
      continue;
    }
    strncpy(gAllowedSubsystems[i].name,colptr(name_col,str),16);
    if (cmp1_col >= 0) strncpy(gAllowedSubsystems[i].alias1,colptr(cmp1_col,str),16);
    if (cmp2_col >= 0) strncpy(gAllowedSubsystems[i].alias2,colptr(cmp2_col,str),16);
    if (desc_col >= 0) strncpy(gAllowedSubsystems[i].desc,colptr(desc_col,str),64);
    i++;
  }
  nAllowedSubsystems = i;
err:
  if (fp) fclose(fp);
  return cc;
}
int getAllowedSubsystemDescriptions(DTYPE *d,char *tgt)
{
  int i,n,siz,isnam=TRUE,hastgt=FALSE;
  DUSTRING *ustr = (DUSTRING *)d->data.vptr;
  if (tgt != NULL && strlen(tgt) > 0) 
  {
    hastgt = TRUE;
    d->dArrayLength = 1;
  }
  switch (d->dFormat)
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      siz = fmtsizeof(d->dFormat);
      break;
    case CF_USTRING:
      siz = 64;
      isnam = FALSE;
      break;
    default: 
      return illegal_format;
  }
  if (nAllowedSubsystems < (int)d->dArrayLength) d->dArrayLength = (UINT32)nAllowedSubsystems;
  for (i=0,n=0; n<(int)d->dArrayLength && i<nAllowedSubsystems; i++)
  {
    if (hastgt && stricmp(tgt,gAllowedSubsystems[i].name)) continue;
    if (isnam) 
    {
      strncpy(&d->data.cptr[n*siz],gAllowedSubsystems[i].desc,siz);
    }
    else
    { /* the USTRING request */
      sprintf(ustr[n].str,"%.12s: %.64s",gAllowedSubsystems[i].name,gAllowedSubsystems[i].desc);
    }
    n++;
  }
  return 0;
}
int getAllowedSubsystemNames(DTYPE *d)
{
  int i,siz,cpsiz;
  switch (d->dFormat)
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      siz = fmtsizeof(d->dFormat);
      break;
    default: 
      return illegal_format;
  }
  cpsiz = siz < 16 ? siz : 16;
  if (nAllowedSubsystems < (int)d->dArrayLength) d->dArrayLength = (UINT32)nAllowedSubsystems;
  for (i=0; i<(int)d->dArrayLength; i++)
  {
    strncpy(&d->data.cptr[i*siz],gAllowedSubsystems[i].name,cpsiz);
  }
  return 0;
}

void appendSubSystemList(char *substr)
{
  int i, is = -1, dir = 0;
  if (substr == NULL || strlen(substr) == 0) return;
  if (nSubSystems == MAX_SUBSYSTEMS) return;
  for (i=0; i<nSubSystems && strnicmp(substr,subSystems[i].name,SUBSYSTEM_NAME_SIZE); i++);
  for (i=0; i<nSubSystems; i++)
  {
    if ((dir=strnicmp(substr,subSystems[i].name,SUBSYSTEM_NAME_SIZE)) == 0) break;
    if (dir < 0 && is < 0)
    { /* make note of first case where string1 < string2 */
      is = i;
    }
  }
  if (i < nSubSystems) return;
  if (is < 0)
  { /* everything in proper order */
    strncpy(subSystems[i].name,substr,SUBSYSTEM_NAME_SIZE);
  }
  else
  { /* insert entry in proper place */
    for (i=nSubSystems; i>is; i--)
    { /* move these up */
      strncpy(subSystems[i].name,subSystems[i-1].name,SUBSYSTEM_NAME_SIZE);
    }
    strncpy(subSystems[is].name,substr,SUBSYSTEM_NAME_SIZE);
  }
  nSubSystems++; 
}
char *getShortSubSystemName(char *subLongName,int forceMatch)
{
  int i, slen;
  if (nAllowedSubsystems > 0)
  {
    for (i=0; i<nAllowedSubsystems; i++)
    {
      slen = (int)strlen(gAllowedSubsystems[i].name); /* straight up */
      if (!strnicmp(subLongName,gAllowedSubsystems[i].name,slen)) return gAllowedSubsystems[i].name;
      slen = (int)strlen(gAllowedSubsystems[i].alias1); /* 1st comparison string */
      if (slen > 0 && !strnicmp(subLongName,gAllowedSubsystems[i].alias1,slen)) return gAllowedSubsystems[i].name;
      slen = (int)strlen(gAllowedSubsystems[i].alias2); /* 2nd comparison string */
      if (slen > 0 && !strnicmp(subLongName,gAllowedSubsystems[i].alias2,slen)) return gAllowedSubsystems[i].name;
    }
    return forceMatch ? "MISC" : strupr(subLongName);
  }
  else
  {
    if (!strnicmp(subLongName,"magn",4)) return "MAG";
    if (!strnicmp(subLongName,"timin",5)) return "TIM";
    if (!strnicmp(subLongName,"time",4)) return "TIM";
    if (!strnicmp(subLongName,"injec",5)) return "INJ";
    if (!strnicmp(subLongName,"instr",5)) return "INSTR";
    if (!strnicmp(subLongName,"dia",3)) return "DIAG";
    if (!strnicmp(subLongName,"tun",3)) return "TUNE";
    if (!strnicmp(subLongName,"feedb",5)) return "FB";
    if (!strnicmp(subLongName,"fb",2)) return "FB";
    if (!strnicmp(subLongName,"serv",4)) return "SER";
    if (!strnicmp(subLongName,"subs",4)) return "INFRA";
    if (!strnicmp(subLongName,"infra",5)) return "INFRA";
    if (!strnicmp(subLongName,"util",4)) return "UTIL";
    if (!strnicmp(subLongName,"misc",4)) return "MISC";
    if (!strnicmp(subLongName,"vac",3)) return "VAC";
    if (!strnicmp(subLongName,"exp",3)) return "EXP";
    if (!strnicmp(subLongName,"cam",3)) return "VIDEO";
    if (!strnicmp(subLongName,"vid",3)) return "VIDEO";
    if (!strnicmp(subLongName,"rf",2)) return "RF";
    if (!strnicmp(subLongName,"quen",4)) return "QPS";
    if (!strnicmp(subLongName,"cryo",4)) return "CRYO";
    if (!strnicmp(subLongName,"kry",3)) return "CRYO";
    if (!strnicmp(subLongName,"hist",4)) return "HIST";
    if (!strnicmp(subLongName,"hst",3)) return "HIST";
    if (!strnicmp(subLongName,"arc",3)) return "HIST";
    if (!strnicmp(subLongName,"test",4)) return "TEST";
    if (!strnicmp(subLongName,"tst",3)) return "TEST";
    if (!strnicmp(subLongName,"aut",3)) return "AUTO";
    if (!strnicmp(subLongName,"calc",4)) return "CALC";
    if (!strnicmp(subLongName,"watch",5)) return "WDOG";
    if (!strnicmp(subLongName,"wdog",4)) return "WDOG";
  }
  return strupr(subLongName);
}
char *getSubsystemFromDescription(char *str)
{
  static char substr[SUBSYSTEM_NAME_SIZE];
  char *c,*c1,s[68];
  int len;

  substr[0] = 0;
  if (str == NULL) goto err;
  strncpy(s,str,64); s[64] = 0;
  if (s[0] != '[' || (c=strchr(s,']')) == NULL) goto err;
  *c++ = 0;  len = (int)strlen(c);
  strncpy(str,c,len); str[len] = 0;
   c = &s[1];
  while (*c == ' ') c++;
  if ((c1=strchr(c,' ')) != NULL) *c1 = 0;
  if (strlen(c) == 0) goto err;
  c = getShortSubSystemName(c,TRUE);
  strncpy(substr,c,SUBSYSTEM_NAME_SIZE);
err:
  return substr;
}
void checkLocationEntry(FecInfoStruct *fecinfo,FecAddrStruct *fec)
{
  char *loc;
  if (!hasLocationsFile) return;
  if ((loc=getCanonicalLocation(fec->strAdr))!= NULL) strncpy(fecinfo->loc,loc,FEC_LOCATION_SIZE);
  return;
}
void checkSubSytemEntry(FecInfoStruct *fecinfo,ExpDataStruct *srv)
{ /* this routine currently doesn't really do anything .... */
  char *c, *subs, subsBuffer[SUBSYSTEM_NAME_SIZE];
  int i, len;
  
  if (fecinfo == NULL) return;
  /* see if the subystem is entered in this old-fashioned way: */
  subs = getSubsystemFromDescription(fecinfo->desc);
  if (strlen(subs) == 0 && (c=strchr(srv->EqmContext,'.')) != NULL)
  { /* if not, see if the context might be carrying the subsystem */
    c++;
    strncpy(subsBuffer,c,SUBSYSTEM_NAME_SIZE);
    subs = subsBuffer;
  }
  if ((len=(int)strlen(srv->SubSystem)) > 0)
  { /* given normally */
    for (i=0; i<len && isalnum(srv->SubSystem[i]); i++);
    if (i == len) subs = srv->SubSystem;
  }
  /* find a match in the canonical list */
  subs = getShortSubSystemName(subs,TRUE);
  return;
}
int checkContextEntry(FecInfoStruct *inf, ExpDataStruct *srv)
{
  char *c, *subs;
  char ctxRoot[CONTEXT_NAME_SIZE], subsBuffer[SUBSYSTEM_NAME_SIZE];
  char srvBuffer[EXPORT_NAME_SIZE+4], ctxBuffer[CONTEXT_NAME_SIZE+4];
  int i;
 
  if (srv == NULL) return FALSE;
  if ((c=strchr(srv->EqmContext,'_')) != NULL)
  { /* could this be "_TEST" or "_SIM" ? */
    if (isSystemContextDecoration(&c[1]))
    { /* seems to be! replace with a '.' */
      *c = '.';
    }
  }
  strncpy(ctxRoot,srv->EqmContext,CONTEXT_NAME_SIZE);
  if ((c=strchr(ctxRoot,'.')) != NULL) *c = 0; /* establish root context */
  for (i=0; i<nAllowedContexts; i++)
  { /* have to pass this test ... */
    if (!strnicmp(gAllowedContexts[i].name,ctxRoot,CONTEXT_NAME_SIZE)) break;
  }
  if (nAllowedContexts > 0 && i == nAllowedContexts) return FALSE;
  if ((c=strchr(srv->EqmContext,'.')) == NULL) 
  { /* context not decorated */
    goto chkCtxCase;
  }
  if (isSystemContextDecoration(&c[1])) goto chkCtxCase;
  strncpy(subsBuffer,&c[1],SUBSYSTEM_NAME_SIZE);
  /* possible sub system decoration */
  subs = getShortSubSystemName(subsBuffer,TRUE);
  if (strlen(srv->SubSystem) > 0)
  { /* subsystem given (take it) */
    goto chkCtxCase;
  }
  /* allowed exceptions */
  for (i=0; i<nAllowedDecorations; i++)
  {
    if (strnicmp(subs,gAllowedDecorations[i].name,16) == 0) goto chkCtxCase;
  }
  *c = 0; /* context has been decorated with subsystem name: remove decoration */
  if ((c=strchr(srv->EqmContext,'_')) != NULL)
  { /* check "_TEST" or "_SIM" again (the famous doocs 'XFEL_SIM.DIAG' case) */
    if (isSystemContextDecoration(&c[1]))
    { /* seems to be! replace with a '.' */
      *c = '.';
    }
  }
  if (strlen(srv->SubSystem) == 0) strncpy(srv->SubSystem,subs,SUBSYSTEM_NAME_SIZE);
chkCtxCase:
  for (i=0; i<ncontexts; i++)
  { /* find a case-consistent match */
    if (strnicmp(srv->EqmContext,contexts[i].name,CONTEXT_NAME_SIZE)) continue;
    strncpy(srv->EqmContext,contexts[i].name,CONTEXT_NAME_SIZE);
    break;
  }
  /* now trap servers with a 'test' in the name */
  strncpy(srvBuffer,srv->ExportName,EXPORT_NAME_SIZE);
  srvBuffer[EXPORT_NAME_SIZE] = 0; strlwr(srvBuffer);
  strncpy(ctxBuffer,srv->EqmContext,CONTEXT_NAME_SIZE);
  ctxBuffer[CONTEXT_NAME_SIZE] = 0; strlwr(ctxBuffer);
  if (strcontains(srvBuffer,"test") && !strcontains(ctxBuffer,"test") && strcmp(srv->SubSystem,"WDOG"))
  { /* server name has a 'test' in it, but not in an approved 'test' context */
    strncpy(srv->SubSystem,"TEST",SUBSYSTEM_NAME_SIZE);
  }
  /* now trap the doocs watchdog */
  if (strcontains(srvBuffer,".watch"))
  { /* server name has a 'test' in it, but not in an approved 'test' context */
    strncpy(srv->SubSystem,"WDOG",SUBSYSTEM_NAME_SIZE);
  }
  /* apply any known export corrections */
  applyExportCorrections(inf,srv);
  return TRUE;
}

void strAdr2netAdr(FecAddrStruct *fte)
{
  int family = AF_INET;
  char *c, scratch[128];
  memset(&fte->netAdr,0,sizeof(FecNetAdr));  
  family = !isInvalidIpV4Addr(fte->strAdr) ? AF_INET : 
           !isInvalidIpV6Addr(fte->strAdr) ? AF_INET6 : 0;
  switch (family)
  {
    case AF_INET:
      inet_pton(family,fte->strAdr,&fte->netAdr);
      fte->inetProtocol = UDP;
      break;
    case AF_INET6:
      inet_pton(family,fte->strAdr,&fte->netAdr);
      fte->inetProtocol = UDP6;
      break;
    default:
      strncpy(scratch,fte->strAdr,40);
      if ((c=strchr(scratch,':')) != NULL) *c++ = 0;
      HexToBytes(scratch,fte->netAdr.ipx.IpxNet,8);
      if (c != NULL) HexToBytes(c,fte->netAdr.ipx.IpxNode,12);
      fte->inetProtocol = IPX;
      break;
  }
}
/* used internally: ver never NULL ! */
int getTineProtocolFromVersion(const char *ver)
{
  int v = atoi(ver);
  if (v > 2) v += 2;
  else if (v == 0) v = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
  return v;
}
int loadDatabaseFiles(void)
{
  FILE *fp = NULL;
  time_t ta=0,tnow=time(NULL);
  int i,family=AF_INET,po=0,cc=0,done=0;
  int name_col=-1,fecname_col=-1,eqpmodule_col=-1,
      context_col=-1,subsys_col=-1,reboot_col=-1;
  int ipx_col=-1,mac_col=-1,ip_col=-1,ipv4_col=-1,
      os_col=-1,description_col=-1,location_col=-1,hardware_col=-1,
      responsible_col=-1,portoffset_col=-1,import_col=-1,ver_col=-1,
      last_col=-1,rul_col=-1;
  char str[256],hdr[256],scratch[128],s[256];
  char prpname[36], *subs, *c;

  if (strlen(fecfname) == 0 || strlen(eqpfname) == 0) ccerr(un_allocated);

  if ((fp=fopen(eqpfname,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((name_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_line);
      if ((fecname_col=findcol(hdr,"FECNAME")) < 0) ccerr(no_such_line);
      if ((eqpmodule_col=findcol(hdr,"EQPMODULE")) < 0) ccerr(no_such_line);
      context_col = findcol(hdr,"CONTEXT");
      subsys_col = findcol(hdr,"SUBSYSTEM");
      reboot_col = findcol(hdr,"REBOOTCOUNT");
      last_col = findcol(hdr,"LASTACTIVE");
      done = 1;
      continue;
    }
    strncpy(SrvTbl[i].ExportName,colptr(name_col,str),EXPORT_NAME_SIZE);
    strncpy(SrvTbl[i].FecName,colptr(fecname_col,str),FEC_NAME_SIZE);
    strncpy(SrvTbl[i].EqmName,colptr(eqpmodule_col,str),EQM_NAME_SIZE);
    if (context_col >= 0)
    {
      strncpy(SrvTbl[i].EqmContext,colptr(context_col,str),CONTEXT_NAME_SIZE);
      if ((c=strchr(SrvTbl[i].EqmContext,'_')) != NULL)
      { /* could this be "_TEST" or "_SIM" ? */
        if (isSystemContextDecoration(&c[1]))
        { /* seems to be! replace with a '.' */
          *c = '.';
        }
      }
    }
    if (subsys_col >= 0)
    {
      strncpy(scratch,colptr(subsys_col,str),SUBSYSTEM_NAME_SIZE);
      subs = getShortSubSystemName(scratch,TRUE);
      strncpy(SrvTbl[i].SubSystem,subs,SUBSYSTEM_NAME_SIZE);
      if (strlen(SrvTbl[i].SubSystem) == 0) strcpy(SrvTbl[i].SubSystem,"TEST");
      appendSubSystemList(SrvTbl[i].SubSystem);
    }
    if (reboot_col >= 0)
    {
      strncpy(scratch,colptr(reboot_col,str),16);
      SrvInfoTbl[i].rebootCounts = atoi(scratch);
    }
    ta = 0;
    if (last_col >= 0)
    {
      strncpy(scratch,colptr(last_col,str),24); scratch[24] = 0;
      ta = parseCtimeStr(scratch);
    }
    if (ta == 0) ta = tnow; /* could be just starting ? */
    SrvInfoTbl[i].lastActive = SrvInfoTbl[i].lastPinged = ta;
    if (tineDebug) 
    { 
      dbglog("ENS: %.32s [FEC %.16s, EQP %.6s] added",
              SrvTbl[i].ExportName,SrvTbl[i].FecName,SrvTbl[i].EqmName);
    }

    if (GetPropertyId("ENSEQM",SrvTbl[i].ExportName) < 0)
    {
      strncpy(prpname,SrvTbl[i].ExportName,EXPORT_NAME_SIZE); prpname[EXPORT_NAME_SIZE] = 0;
      if (RegisterProperty("ENSEQM",prpname,
                           sizeof(ExpDataStruct)+sizeof(FecAddrStruct),CF_STRUCT,
                           CA_READ,"registered server") < 0)
      {
        nslog("loadDatabaseFiles: couldn't register %.32s for export",SrvTbl[i].ExportName);
      }
    }
    i++;
  }
  fclose(fp); fp = NULL;
  numSrvTblEntries = SrvTblSize = i;

  if ((fp=fopen(fecfname,"r")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((fecname_col=findcol(hdr,"FECNAME")) < 0) ccerr(no_such_line);
      if ((ipx_col=findcol(hdr,"IPX_ADDR")) < 0) ipx_col = findcol(hdr,"FECNETWORK");
      if ((mac_col=findcol(hdr,"MAC_ADDR")) < 0) mac_col = findcol(hdr,"FECNODE");
      if ((ip_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_line);
      ipv4_col = findcol(hdr,"IPV4_ADDR");
      if ((os_col=findcol(hdr,"OS")) < 0) ccerr(no_such_line);
      if ((description_col=findcol(hdr,"DESCRIPTION")) < 0) ccerr(no_such_line);
      if ((location_col=findcol(hdr,"LOCATION")) < 0) ccerr(no_such_line);
      if ((hardware_col=findcol(hdr,"HARDWARE")) < 0) ccerr(no_such_line);
      if ((responsible_col=findcol(hdr,"RESPONSIBLE")) < 0) ccerr(no_such_line);
      portoffset_col = findcol(hdr,"PORTOFFSET");
      import_col = findcol(hdr,"IMPORTANCE");
      ver_col=findcol(hdr,"VERSION");
      rul_col=findcol(hdr,"MCAST_RULES");
      done = 1;
      continue;
    }
    strncpy(FecTbl[i].fecName,colptr(fecname_col,str),FEC_NAME_SIZE);
    strncpy(FecTbl[i].strAdr,colptr(ip_col,str),ADDR_SIZE_IPV6);
    if (gAfInet == AF_INET6)
    { /* the canonical address familiy is IPv6 ! */
      strncpy(FecInfoEx[i].ipv6,FecTbl[i].strAdr,40);
      if (ipv4_col < 0 && isIpv4on6(FecTbl[i].strAdr)) toIpv4Addr(FecTbl[i].strAdr,FecInfoEx[i].ipv4);
    }
    else 
    {
      if (ipv4_col < 0) strncpy(FecInfoEx[i].ipv4,FecTbl[i].strAdr,16);
    }
    if (ipv4_col >= 0)
    {
      strncpy(FecInfoEx[i].ipv4,colptr(ipv4_col,str),16);
    }
    if (mac_col >= 0) 
    {
      strncpy(scratch,colptr(mac_col,str),16);
      HexToBytes(scratch,FecTbl[i].lclAdr.mac,6);
      memcpy(FecInfoEx[i].mac,FecTbl[i].lclAdr.mac,6);
    }
    if (strlen(FecTbl[i].strAdr) == 0)
    { /* still speaking IPX ? */
      if (ipx_col >= 0) strncpy(scratch,colptr(ipx_col,str),16); strcat(scratch,":");
      BytesToHex(FecTbl[i].lclAdr.mac,&scratch[strlen(scratch)],6);
    }
    strAdr2netAdr(&FecTbl[i]);
    if (portoffset_col >= 0)
    {
      strncpy(scratch,colptr(portoffset_col,str),16);
      po = atoi(scratch);
    }
    else
    {
      po = 0;
    }
    FecTbl[i].portOffset = po;
    /* NB: colptr() strips the leading and trailing '"' if there are any */
    /* so we have to put them back if they're necessary */
    strncpy(scratch,colptr(os_col,str),FEC_OS_SIZE); scratch[FEC_OS_SIZE] = 0;
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%.128s\"",scratch); else strncpy(s,scratch,128);
    strncpy(FecInfoTbl[i].os,s,FEC_OS_SIZE);

    strncpy(scratch,colptr(description_col,str),FEC_DESC_SIZE); scratch[FEC_DESC_SIZE] = 0;
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%.128s\"",scratch); else strncpy(s,scratch,128);
    strncpy(FecInfoTbl[i].desc,s,FEC_DESC_SIZE);

    strncpy(scratch,colptr(location_col,str),FEC_LOCATION_SIZE); scratch[FEC_LOCATION_SIZE] = 0;
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%.128s\"",scratch); else strncpy(s,scratch,128);
    strncpy(FecInfoTbl[i].loc,s,FEC_LOCATION_SIZE);

    strncpy(scratch,colptr(hardware_col,str),FEC_HDW_SIZE); scratch[FEC_HDW_SIZE] = 0;
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%.128s\"",scratch); else strncpy(s,scratch,128);
    strncpy(FecInfoTbl[i].hdw,s,FEC_HDW_SIZE);

    if (rul_col >= 0)
    {
      strncpy(scratch,colptr(rul_col,str),16);
      po = !stricmp(scratch,"MODERN") ? 1 : atoi(scratch);
      FecTbl[i].addrFlags = po;
    }
    if (ver_col >= 0)
    {
      strncpy(FecInfoTbl[i].ver,colptr(ver_col,str),FEC_VERSION_SIZE); scratch[FEC_VERSION_SIZE] = 0;
    }
    if (strlen(FecInfoTbl[i].ver) == 0)
    {
      strncpy(FecInfoTbl[i].ver,"3.31",FEC_VERSION_SIZE);
    }
    FecTbl[i].tineProtocol = getTineProtocolFromVersion(FecInfoTbl[i].ver);
    if (import_col >= 0)
    {
      strncpy(scratch,colptr(import_col,str),16);
      if (strlen(scratch) > 0) FecInfoEx[i].importance = (short)atoi(scratch);
    }

    strncpy(scratch,colptr(responsible_col,str),FEC_RESP_SIZE); scratch[FEC_RESP_SIZE] = 0;
    if (strchr(scratch,',') != NULL) sprintf(s,"\"%.128s\"",scratch); else strncpy(s,scratch,128);
    strncpy(FecInfoTbl[i].resp,s,FEC_RESP_SIZE);

    if (tineDebug) 
    { 
       dbglog("ENS: FEC %.16s [%.40s:%d] added",FecTbl[i].fecName,FecTbl[i].strAdr,FecTbl[i].portOffset);
    }

    if (GetPropertyId("ENSEQM",FecTbl[i].fecName) < 0)
    {
      strncpy(prpname,FecTbl[i].fecName,FEC_NAME_SIZE); prpname[FEC_NAME_SIZE] = 0;
      if (RegisterProperty("ENSEQM",prpname,
                           sizeof(FecAddrStruct),CF_STRUCT,
                           CA_READ,"registered FEC") < 0)
        feclog("ENS: couldn't register %.16s for export",FecTbl[i].fecName);
    }
    i++;
  }
  fclose(fp); fp = NULL;
  numFecTblEntries = FecTblSize = i;
  fillinFecIndexTable();
  return 0;

err: 
  if (fp) fclose(fp);
  return cc;
}
char gSrvDataBaseFullName[196], gFecDataBaseFullName[256];
int scanTables(void)
{
  BYTE *tmp;
  int cc;
  int srvTblSizeRead,fecTblSizeRead;
  static int srvTblSizeAllocated = 0,fecTblSizeAllocated = 0;

  feclog("ENS: Scanning database files");
  srvTblSizeRead = GetDataBaseSize(gSrvDataBaseFullName);
  fecTblSizeRead = GetDataBaseSize(gFecDataBaseFullName);
  if (srvTblSizeRead <= 0 || fecTblSizeRead <= 0)
  { /* no file found -> give it some space */
    srvTblSizeRead = TMPSPACE; fecTblSizeRead = TMPSPACE;  
  }
  else
  {
    feclog("read %d Server Records, %d FEC Records",srvTblSizeRead,fecTblSizeRead);
  }

  if (srvTblSizeAllocated == 0) srvTblSizeAllocated = srvTblSizeRead;
  if (fecTblSizeAllocated == 0) fecTblSizeAllocated = fecTblSizeRead;

  if (srvTblSizeRead >= srvTblSizeAllocated)
  {
    feclog("ENS: (re)allocating server table memory (%d entries needed vs. %d)", srvTblSizeRead, srvTblSizeAllocated);
    srvTblSizeAllocated = srvTblSizeRead + FILESPACE;
    if (!dbaseloaded)
    {
      if (SrvTbl != NULL) free(SrvTbl);
      if (FecTbl != NULL) free(FecTbl);
      SrvTbl = NULL;
      FecTbl = NULL;
    }
    if (SrvTbl == NULL)
    {
      if ((SrvTbl=(ExpDataStruct *)calloc(srvTblSizeAllocated+TMPSPACE,sizeof(ExpDataStruct))) == NULL)
        ccerr(out_of_server_memory);
      if ((SrvInfoTbl=(SrvInfo *)calloc(srvTblSizeAllocated+TMPSPACE,sizeof(SrvInfo))) == NULL)
        ccerr(out_of_server_memory);
    }
    else
    {
      if ((tmp=(BYTE *)realloc(SrvTbl,(srvTblSizeAllocated+TMPSPACE)*sizeof(ExpDataStruct))) == NULL)
        ccerr(out_of_server_memory);
      SrvTbl = (ExpDataStruct *)tmp;
      if ((tmp=(BYTE *)realloc(SrvInfoTbl,(srvTblSizeAllocated+TMPSPACE)*sizeof(SrvInfo))) == NULL)
        ccerr(out_of_server_memory);
      SrvInfoTbl = (SrvInfo *)tmp;
      memset(SrvInfoTbl,0,(srvTblSizeAllocated+TMPSPACE)*sizeof(SrvInfo));
    }
  }

  /* could be that these aren't allocated yet ! */
  if (SrvInfoTbl == NULL &&
      (SrvInfoTbl=(SrvInfo *)calloc(srvTblSizeAllocated+TMPSPACE,sizeof(SrvInfo))) == NULL)
    ccerr(out_of_server_memory);

  if (fecTblSizeRead >= fecTblSizeAllocated || FecTbl == NULL)
  {
    feclog("ENS: (re)allocating fec table memory (%d entries needed vs. %d)", fecTblSizeRead, fecTblSizeAllocated);
    fecTblSizeAllocated = fecTblSizeRead + FILESPACE;
    if (FecTbl == NULL)
    {
      if ((FecTbl=(FecAddrStruct *)calloc(fecTblSizeAllocated+TMPSPACE,sizeof(FecAddrStruct))) == NULL)
        ccerr(out_of_server_memory);
      if ((FecInfoTbl=(FecInfoStruct *)calloc(fecTblSizeAllocated+TMPSPACE,sizeof(FecInfoStruct))) == NULL)
        ccerr(out_of_server_memory);
      if ((FecInfoEx=(FecInfoStructEx *)calloc(fecTblSizeAllocated+TMPSPACE,sizeof(FecInfoStructEx))) == NULL)
        ccerr(out_of_server_memory);
    }
    else
    {
      if ((tmp=(BYTE *)realloc(FecTbl,(fecTblSizeAllocated+TMPSPACE)*sizeof(FecAddrStruct))) == NULL)
        ccerr(out_of_server_memory);
      FecTbl = (FecAddrStruct *)tmp;
      if ((tmp=(BYTE *)realloc(FecInfoTbl,(fecTblSizeAllocated+TMPSPACE)*sizeof(FecInfoStruct))) == NULL)
        ccerr(out_of_server_memory);
      FecInfoTbl = (FecInfoStruct *)tmp;
      if ((tmp=(BYTE *)realloc(FecInfoEx,(fecTblSizeAllocated+TMPSPACE)*sizeof(FecInfoStructEx))) == NULL)
        ccerr(out_of_server_memory);
      FecInfoEx = (FecInfoStructEx *)tmp;
      memset(FecInfoEx,0,(fecTblSizeAllocated+TMPSPACE)*sizeof(FecInfoStructEx));
    }
  }

  if (FecInfoTbl == NULL &&
      (FecInfoTbl=(FecInfoStruct *)calloc(fecTblSizeAllocated+TMPSPACE,sizeof(FecInfoStruct))) == NULL)
    ccerr(out_of_server_memory);
  if (FecInfoEx == NULL &&
      (FecInfoEx=(FecInfoStructEx *)calloc(fecTblSizeAllocated+TMPSPACE,sizeof(FecInfoStructEx))) == NULL)
    ccerr(out_of_server_memory);

  if (srvTblSizeAllocated == 0)
  {
    feclog("ENS: server table not allocated !");
    ccerr(un_allocated);
  }
  if (fecTblSizeAllocated == 0)
  {
    feclog("ENS: fec table not allocated !");
    ccerr(un_allocated);
  }
  /* Now we have enough room: */
  feclog("ENS: Loading database files");
  if ((cc=loadDatabaseFiles()) != 0) ccerr(database_not_loaded);
  dbaseloaded = 1;
  getContexts();
err:
  if (cc != 0)
  {
    nslog("scanTables: error loading database :: %.32s", cc2str(cc));
    feclog("ENS: (scanTables) error loading database :: %.32s", cc2str(cc));
  }
  return cc;
}
char fechdr[] = "FecName,IpxAddr,MacAddr,IpAddr,IpV4Addr,PortOffset,OS,Version,Description,Location,Hardware,Responsible,Importance,McastRules\n";
char eqphdr[] = "Name,FecName,EqpModule,Context,SubSystem,RebootCount,LastActive\n";

int dumpFECTable(void)
{
  FILE *fp = NULL;
  int i,rule,cc = 0;
  char tmpFecName[FEC_NAME_SIZE+4];
  char ipxaddr[32],macaddr[32],ipaddr[48],ipv4addr[16];

  dbbackup(fecfname);
  if ((fp=fopen(fecfname,"w+t")) == NULL) ccerr(file_error);
  fprintf(fp,"%s",fechdr);
  for (i = 0; i<FecTblSize; i++)
  {
    if (FecInfoEx[i].disable) continue;
    strncpy(tmpFecName,FecTbl[i].fecName,FEC_NAME_SIZE); tmpFecName[FEC_NAME_SIZE] = 0;
    if (FecTbl[i].inetProtocol == IPX)
    {
      IpxAdr ipx = FecTbl[i].netAdr.ipx;
      sprintf(ipxaddr,"%02x%02x%02x%02x",
        ipx.IpxNet[0],ipx.IpxNet[1],ipx.IpxNet[2],ipx.IpxNet[3]);
      sprintf(macaddr,"%02x%02x%02x%02x%02x%02x",
        ipx.IpxNode[0],ipx.IpxNode[1],ipx.IpxNode[2],ipx.IpxNode[3],ipx.IpxNode[4],ipx.IpxNode[5]);
      strcpy(ipaddr,"0");
      strcpy(ipv4addr,"0.0.0.0");
    }
    else
    { 
      strcpy(ipxaddr,"0");
      BytesToHex(FecTbl[i].lclAdr.mac,macaddr,6);
      strncpy(ipaddr,FecTbl[i].strAdr,ADDR_SIZE_IPV6);
      strncpy(ipv4addr,FecInfoEx[i].ipv4,16);
    }
    rule = (FecTbl[i].addrFlags & ADDRFLAG_MCAST_MODERN) ? 1 : 0;
    fprintf(fp,"%.16s,%s,%s,%.40s,%.16s,%d,%.16s,%.16s,%.64s,%.48s,%.32s,%.32s,%d,%d\n",
            tmpFecName,ipxaddr,macaddr,ipaddr,ipv4addr,
            FecTbl[i].portOffset,FecInfoTbl[i].os,FecInfoTbl[i].ver,
            FecInfoTbl[i].desc,FecInfoTbl[i].loc,
            FecInfoTbl[i].hdw,FecInfoTbl[i].resp,(int)FecInfoEx[i].importance,rule);
  }
  dumpFecFileFlag = 0;
err: if (fp) fclose(fp);
  if (cc) printf("ENS: write FEC Table - %.32s\n>",erlst[cc]);
  return cc;
}
int dumpEQPTable(void)
{
  FILE *fp = NULL;
  int i,cc = 0;

  dbbackup(eqpfname);
  if ((fp=fopen(eqpfname,"w+t")) == NULL) ccerr(file_error);
  fprintf(fp,"%s",eqphdr);
  for (i = 0; i<SrvTblSize; i++)
  {
    if (SrvInfoTbl[i].disabled) continue;
    fprintf(fp,"%.32s,%.16s,%.6s,%.32s,%.16s,%u,%.24s\n",
      SrvTbl[i].ExportName,SrvTbl[i].FecName,SrvTbl[i].EqmName,
      SrvTbl[i].EqmContext,SrvTbl[i].SubSystem,SrvInfoTbl[i].rebootCounts,ctime(&SrvInfoTbl[i].lastActive));
  }
  dumpEqpFileFlag = 0;
err: 
  if (fp) fclose(fp);
  if (cc) feclog("ENS: write EQP Table - %.32s",erlst[cc]);
  return cc;
}
#define ext_1day "1day"
#define ext_1week "1week"
int dbbackup(char *fn)
{
  struct stat fs, fsb;
  char fnb1[80],fnb2[80];
  int dayfile_exists = FALSE;
  time_t tnow = time(NULL);

  if (stat(fn,&fs) != 0) return no_such_file;
  if (fs.st_size > 0)
  { /* copy only if non-zero length */
    sprintf(fnb1,"%.72s.%s",fn,ext_1day);
    if (stat(fnb1,&fsb) == 0)
    { /* if there: needs to be at least 1 day old */
      dayfile_exists = TRUE;
      if (tnow < fsb.st_mtime + ONEDAY) return 0;
    }
    sprintf(fnb2,"%.72s.%s",fn,ext_1week);
    if (stat(fnb2,&fsb) == 0)
    { /* check this one too !*/
      if (tnow > fsb.st_mtime + ONEDAY * 7)
      {
        nslog("backup 1-week database for %s",fn);
        unlink(fnb2); rename(fnb1,fnb2);
      }
    }
    else if (dayfile_exists)
    { /* initial file */
      nslog("intial backup 1-week database for %s",fn);
      rename(fnb1,fnb2);
    }
    nslog("backup 1-day database for %s",fn);
    unlink(fnb1); rename(fn,fnb1);
  }
  return 0;
}

int relaxGensAddressCheck = FALSE;
void checkMyENSEntry(void)
{
  static char myaddr[64] = "";
  static int done = FALSE;
  static SCKADR sckadr;
  static BYTE badr[40];
  struct stat sbuf;
  int i;

  if (stat(MY_ADDRESS_FILE,&sbuf) == 0)
  { /* the ENS is forcing its IP address : some weird IP configuration is in play*/
    relaxGensAddressCheck = TRUE;
  }
  if (!done)
  {
    memcpy(badr,getmyipaddr(),gInAdrLen);
    if (!zerocmp(badr,gInAdrLen)) return;
    sckadr_in(&sckadr,badr,UDP_PORT,gAfInet);
    inet_ntop(gAfInet,badr,myaddr,40);
    done = TRUE;
  }
  if (isInAddrLoopback(&sckadr))
  {
    feclog("local IP address is the loopback address ! (VMWare ?)");
    feclog("Not updating Self entry in database !");
    return;
  }
  /* the ENS self entry should be one of the very first entries ... */
  for (i=0; i<FecTblSize; i++)
  {
    if (!strnicmp("ENS",FecTbl[i].fecName,FEC_NAME_SIZE)) break;
  }
  if (i == FecTblSize)
  {
    feclog("No ENS Self entry found in database !");
    return;
  }
  // TODO: ipv6
  memcpy(&ens_iaddr,badr,gInAdrLen);
  if (!memcmp(&FecTbl[i].netAdr,badr,gInAdrLen)) goto gens_check; /* this is fine */
  memcpy(&FecTbl[i].netAdr,badr,gInAdrLen);
  memcpy(FecTbl[i].strAdr,myaddr,ADDR_SIZE_IPV6);
  dumpFecFileFlag = 1;
  feclog("Correcting ENS self entry to %.40s",myaddr);
gens_check:
  /* now check the GENS, entry should also be one of the very first entries ... */
  for (i=0; i<FecTblSize; i++)
  {
    if (!strnicmp("GENS",FecTbl[i].fecName,FEC_NAME_SIZE)) break;
  }
  if (i == FecTblSize) return;
  if (!memcmp(&FecTbl[i].netAdr,badr,gInAdrLen)) return; /* this is fine */
  memcpy(&FecTbl[i].netAdr,badr,gInAdrLen);
  memcpy(FecTbl[i].strAdr,myaddr,ADDR_SIZE_IPV6);
  dumpFecFileFlag = 1;
  feclog("Correcting GENS self entry to %.16s",myaddr);
}

#ifdef WIN32
# define lstat(f,s) _stat(f,s)
#endif
int loadENS(void) /* read name information from file */
{
  char *ptr;
  static int done = 0, ENSCheck = TRUE;
# ifdef WIN32
  struct _stat eqpstat,fecstat;
# else
  struct stat eqpstat,fecstat;
# endif
  static time_t OLDeqptime,OLDfectime;
  int n, i, j, cc;

  if (!done)
  {
    tineHomePath[0] = 0;
    if ((ptr=getenv("TINE_HOME")) != NULL) strncpy(tineHomePath,ptr,72);
    sprintf(eqpfname,"%.128s%s",tineHomePath,EQPDB_FILE);
    sprintf(fecfname,"%.128s%s",tineHomePath,FECDB_FILE);
    sprintf(gSrvDataBaseFullName,"%.128s%s",tineHomePath,EQPDB_FILE);
    sprintf(gFecDataBaseFullName,"%.128s%s",tineHomePath,FECDB_FILE);
    if ((ptr=getenv("NO_ENS_CHECK")) != NULL)
    {
      if (stricmp(ptr,"YES") == 0 || stricmp(ptr,"TRUE") == 0)
        ENSCheck = FALSE;
      feclog("not checking ENS Address consistency due to environment");
    }
    done = 1;
  }
  if (lstat(eqpfname,&eqpstat) < 0)
  {
    nslog("loadENS: cannot find eqp database");
    return no_such_file;
  }
  if (lstat(fecfname,&fecstat) < 0)
  {
    nslog("loadENS: cannot find fec database");
    return no_such_file;
  }
  if (OLDeqptime != eqpstat.st_mtime || OLDfectime != fecstat.st_mtime)
  { /* new files found */
    if (SrvInfoTbl != NULL) memset(SrvInfoTbl,0,SrvTblSize*sizeof(SrvInfo)); 
    if (FecInfoEx != NULL) memset(FecInfoEx,0,FecTblSize*sizeof(FecInfoStructEx));
    if ((cc=scanTables()) != 0) feclog("ENS: %.32s",erlst[cc]);
    for (n=0; n<nConnectionTableEntries; n++)
    {
      for (i=0; i<SrvTblSize; i++)
      {
        if (!strnicmp(conTbl[n]->expName,SrvTbl[i].ExportName,EXPORT_NAME_SIZE))
        {
          for (j=0; j<FecTblSize; j++)
          {
            if (!strnicmp(SrvTbl[i].FecName,FecTbl[j].fecName,FEC_NAME_SIZE))
            {
              conTbl[n]->fecIdx = j;
              fixFecProtocol(j);
              break;
            }
          }
          if (j == FecTblSize)
          {
            nslog("loadENS: no FEC entry for %.16s link",SrvTbl[i].FecName);
          }
          conTbl[n]->srvIdx = i;
          break;
        }
      }
      if (i == SrvTblSize)
      {
        nslog("loadENS: no FEC entry for %.32s link",conTbl[n]->expName);
      }
    }
    OLDeqptime = eqpstat.st_mtime;
    OLDfectime = fecstat.st_mtime;
    if (ENSCheck) checkMyENSEntry();        /* check self entry */
    return -1;                              /* signal verification loop*/
  }
  return 0;                                 /* all's quiet */
}
typedef struct
{
  char pattern[CONTEXT_NAME_SIZE];
  char context[CONTEXT_NAME_SIZE];
} CONTEXTPATTERN;
CONTEXTPATTERN *contextPatterns = NULL;
#define CONTEXTS_FILE "contexts.csv"
int ncontextPatterns = 0;
int getContexts(void)
{
  int i,j,cc = 0;
  if (CtxTblSize == 0) CtxTblSize = 100; /* a nice size */
recheck:  
  if (ncontexts >= CtxTblSize)
  {
    if (CtxTblSize < SrvTblSize) CtxTblSize = SrvTblSize; /* worst case */
    if (contexts != NULL) free(contexts); 
    contexts = NULL;
  }
  if (contexts == NULL)
  {
    feclog("ENS: allocate context check table (%d entries)", CtxTblSize);
    if ((contexts=(NAME32 *)calloc(CtxTblSize,sizeof(NAME32))) == NULL)
      ccerr(out_of_server_memory);
    ncontexts = 0;
  }
  for (i=0; i<SrvTblSize; i++)
  {
    for (j=0; j<ncontexts; j++)
    {
      if (!strnicmp(contexts[j].name,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) break;
    }
    if (j == ncontexts && strlen(SrvTbl[i].EqmContext)) /* not found */
    {
      if (ncontexts == CtxTblSize) goto recheck;
      strncpy(contexts[j].name,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE);
      feclog("ENS: register context %.32s as device name", contexts[j].name);
      RegisterDeviceName("ENSEQM",contexts[j].name,4+j);
      ncontexts++;
    }
  }
  AssignNumModules("ENSEQM",ncontexts+4);
err:
  return cc;
}
#define isENS(nam) (!strncmp(nam,"ENS",3))
#define isGENS(nam) (!strncmp(nam,"GENS",4))
#define ENSFWD_FILE "forward.csv"
typedef struct FwdQ
{
  char dev[DEVICE_NAME_SIZE];
  char prp[PROPERTY_NAME_SIZE];
  DTYPE din;
  BYTE payload[512]; /* covers everything ... */
  int access;
  int ntries;
  int pending;
  int sts;
  struct FwdQ *nxt;
  struct FwdQ *prv;
} FWDQ;
typedef struct
{
  char ctx[64];
  char fwd[64];
  FWDQ *fwdq;
  INADR inadr;
  int qlen;
} FWD;
int nforwards = 0;
FWD *fwdLst = NULL;

int retryForwardOnError(int cc)
{
  switch (cc)
  {
    case link_timeout:
    case connection_timeout:
    case operation_busy:
      return TRUE;
    default:
      break;
  }
  return FALSE;
}
void chkForwardsList(int idx)
{
  DTYPE dout;
  FWDQ *q;
  int cc;
  char devsrv[132];

  memset(&dout,0,sizeof(DTYPE)); dout.dFormat = CF_NULL;
  while (fwdLst[idx].fwdq != NULL)
  {
    q = fwdLst[idx].fwdq;
    sprintf(devsrv,"%.64s/%.64s",fwdLst[idx].fwd,q->dev);
    cc = ExecLinkEx(devsrv,q->prp,&dout,&q->din,(short)q->access,200);
    if (retryForwardOnError(cc) && q->ntries++ < 60) break;
    if (cc != 0 && q->ntries >= 60)
    {
      feclog("abandon forward %.64s to %.64s : %.32s",q->prp,fwdLst[idx].fwd,cc2str(cc));
    }
    fwdLst[idx].fwdq = q->nxt;
    free(q);
    fwdLst[idx].qlen--;
  }
}
void addtoFwdQueue(int idx,char *dev,char *prp,DTYPE *din,int access)
{
  FWDQ *q, *qp, *qt;
  int sizeInBytes;
  if (din == NULL) return;
  sizeInBytes = din->dArrayLength * fmtsizeof(din->dFormat);
  if (din->dFormat == CF_STRUCT && din->dArrayLength == 1)
  {
    if (strlen(din->dTag) == 0) return;
    sizeInBytes *= GetStructSize(din->dTag);
  }
  if (sizeInBytes <= 0) return;
  if (sizeInBytes > 512)
  { /* tilt !!!  */
    feclog("unexpected payload size (%d) in forward property %.64s to %.64s",sizeInBytes,prp,fwdLst[idx].fwd);
    return;
  }
  if (fwdLst[idx].qlen > 25)
  {
    feclog("forward queue %.64s at capacity !",fwdLst[idx].fwd);
    return;
  }
  for (qt=fwdLst[idx].fwdq; qt != NULL; qt=qt->nxt)
  {
    if (strnicmp(qt->dev,dev,DEVICE_NAME_SIZE)) continue;
    if (strnicmp(qt->prp,prp,PROPERTY_NAME_SIZE)) continue;
    if (qt->din.dFormat != din->dFormat) continue;
    if (qt->din.dArrayLength != din->dArrayLength) continue;
    if (strncmp(qt->din.dTag,din->dTag,TAG_NAME_SIZE)) continue;
    if (memcmp(qt->payload,din->data.bptr,sizeInBytes)) continue;
    return; /* already in queue ! */
  }
  if ((q=(FWDQ *)calloc(1,sizeof(FWDQ))) == NULL) return;
  strncpy(q->dev,dev,DEVICE_NAME_SIZE);
  strncpy(q->prp,prp,PROPERTY_NAME_SIZE);
  q->access = access;
  q->din.dFormat = din->dFormat;
  q->din.dArrayLength = din->dArrayLength;
  strncpy(q->din.dTag,din->dTag,TAG_NAME_SIZE);
  q->din.data.bptr = q->payload;
  memcpy(q->din.data.bptr,din->data.bptr,sizeInBytes);
  /* find the tail (this needs to be a FIFO queue */
  qp = qt = fwdLst[idx].fwdq;
  while (qt != NULL) { qp = qt; qt = qt->nxt; }
  q->prv = qp;
  if (qp != NULL) qp->nxt = q; else fwdLst[idx].fwdq = q;
  fwdLst[idx].qlen++;
}
int getForwardsList(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int lst_col=-1,con_col=-1;
  char str[256],hdr[256];
  BYTE *badr;

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (fwdLst != NULL) 
  {
    free(fwdLst);
    fwdLst = NULL;
    nforwards = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((fwdLst=(FWD *)calloc(n,sizeof(FWD))) == NULL) ccerr(out_of_server_memory);
  nforwards = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<nforwards;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((lst_col=findcol(hdr,"ENS_LIST")) < 0) ccerr(no_such_line);
      con_col = findcol(hdr,"CONTEXT");
      done = 1;
      continue;
    }
    strncpy(fwdLst[i].fwd,colptr(lst_col,str),64);
    if (con_col >= 0) strncpy(fwdLst[i].ctx,colptr(con_col,str),64);
    if (fwdLst[i].ctx[0] == '*') fwdLst[i].ctx[0] = 0;
    if ((badr=getFwdIpAddr(fwdLst[i].fwd)) == NULL) continue;
    memcpy(&fwdLst[i].inadr,badr,SCKADR_LEN);
    i++;
  }
  nforwards = i;
err:
  if (fp != NULL) fclose(fp);
  return cc;
} 
void cbFwd(int id,int cc,void *ref)
{
  FWDQ *q = (FWDQ *)ref;
  char msg[128];
  if (id < 0 || id >= nforwards) return; /* shouldn't be ! */
  if (q == NULL || fwdLst[id].qlen == 0) return; /* no queue ? */
  q->pending = FALSE;
  q->sts = cc;
  if (retryForwardOnError(cc) && q->ntries++ < 10) return; /* keep trying */
  sprintf(msg,"%.32s",cc2str(cc));
  if (q->ntries >= 10) strcat(msg," (abandoned)");
  if (q->prv == NULL) 
  { /* then I'm at the head */
    fwdLst[id].fwdq = q->nxt; 
  }
  else 
  { /* splice */
    q->prv->nxt = q->nxt;
    if (q->nxt != NULL) q->nxt->prv = q->prv;
  }
  if (fwdLst[id].fwdq != NULL) fwdLst[id].fwdq->prv = NULL; /* ensure this */
  fwdLst[id].qlen--;
  feclog("forward %.64s to %.64s : %s -> queue length now %d",
    q->prp,fwdLst[id].fwd,msg,fwdLst[id].qlen);
  if (fwdLst[id].fwdq == NULL && fwdLst[id].qlen > 0)
  {
    fwdLst[id].qlen = 0;
    feclog("adjust forward %.64s queue length to 0!",fwdLst[id].fwd);
  }
  if (fwdLst[id].qlen == 0 && fwdLst[id].fwdq != NULL)
  {
    fwdLst[id].fwdq = NULL;
    feclog("force forward %.64s queue to NULL!",fwdLst[id].fwd);
  }
  free(q);
}
void chkForwardsQueue(int idx)
{
  DTYPE dout;
  FWDQ *q, *qnxt=NULL;
  int id,cc;
  char devsrv[132];

  DTYPEZERO(dout);
  for (q=fwdLst[idx].fwdq; q != NULL; q=qnxt)
  {
    qnxt = q->nxt; /* the callback below will free 'q' so do this 'just in case' ... */
    if (q->pending) continue; /* still waiting for an answer ... */
    q->pending = TRUE;
    sprintf(devsrv,"%.64s/%.64s",fwdLst[idx].fwd,q->dev);
    id = AttachLinkEx2(devsrv,q->prp,&dout,&q->din,(short)q->access,500,cbFwd,CM_SINGLE,idx,q);
    if (id < 0) 
    { /* unlikely */
      cc = -id;
      feclog("Forward %.64s to %.128s : %.32s",q->prp,devsrv,cc2str(cc));
      q->ntries = 100;
    }
  }
}
int asyncCmdForward(char *context,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int access,BYTE *badr)
{
  int i;
  dout = dout;
  for (i=0; i<nforwards; i++)
  {
    if (!memcmp(&fwdLst[i].inadr,badr,SCKADR_LEN)) continue;
    if (strlen(fwdLst[i].ctx) > 0 && strstr(fwdLst[i].ctx,context) == NULL) continue;
    addtoFwdQueue(i,devName,devProperty,din,access);
  }
  return 0;
}
int dumpFwdQueues(int i1,int i2, int i3, int i4)
{
  int i;
  FWDQ *q;
  i1 = i1; i2 = i2; i3 = i3; i4 = i4;
  if (nforwards == 0)
  {
    ttyoutput("there is no forwarding information registered");
    return 0;
  }
  ttyoutput("forward queue information:");
  for (i=0; i<nforwards; i++)
  {
    ttyoutput("%.64s : current queue length %d",fwdLst[i].fwd,fwdLst[i].qlen);
    for (q=fwdLst[i].fwdq; q != NULL; q=q->nxt)
    {
      ttyoutput("  %.64s[%.64s] : %.32s, %d tries so far (pending: %s)",
        q->dev,q->prp,cc2str(q->sts),q->ntries,(q->pending?"TRUE":"FALSE"));
    }
  }
  return 0;
}
#define ADMINS_FILE "admins.csv"
int nadmins = 0;
NAME16 *admins = NULL;
int isAdministrator(char *nam)
{
  int i;
  if (nadmins == 0) return FALSE; /* must be a registered admin ! */
  for (i=0; i<nadmins && strnicmp(nam,admins[i].name,USERNAME_SIZE); i++);
  return nadmins - i;
}
int getAdministrators(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int usr_col=-1;
  char str[256],hdr[256];
 
  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (admins != NULL)
  {
    free(admins);
    nadmins = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((admins=(NAME16 *)calloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
  nadmins = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<nadmins;)
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((usr_col=findcol(hdr,"USERNAME")) < 0 &&
          (usr_col=findcol(hdr,"USERS")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(admins[i].name,colptr(usr_col,str),USERNAME_SIZE);
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int getContextPatternInformation(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int pattern_col=-1,context_col=-1;
  char *c,str[256],hdr[256],scratch[64];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (contextPatterns != NULL)
  {
    free(contextPatterns);
    ncontextPatterns = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((contextPatterns=(CONTEXTPATTERN *)calloc(n,sizeof(CONTEXTPATTERN))) == NULL) ccerr(out_of_server_memory);
  ncontextPatterns = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<ncontextPatterns;)  /* scan file */
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((pattern_col=findcol(hdr,"PATTERN")) < 0) ccerr(no_such_line);
      if ((context_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(pattern_col,str),CONTEXT_NAME_SIZE);
    if ((c=strchr(scratch,'*')) != NULL) *c = 0;
    strncpy(contextPatterns[i].pattern,scratch,CONTEXT_NAME_SIZE);
    strncpy(contextPatterns[i].context,colptr(context_col,str),CONTEXT_NAME_SIZE);
    i++;
  }
  ncontextPatterns = i;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
void assignContext(char *expname,char *cntxt)
{
  char en[EXPORT_NAME_SIZE];
  int i,len;
  
  strncpy(en,expname,EXPORT_NAME_SIZE); strupr(en);
  for (i=0; i<ncontextPatterns; i++)
  {
    if ((len=(int)strlen(contextPatterns[i].pattern)) > CONTEXT_NAME_SIZE) len = CONTEXT_NAME_SIZE;
    if (strnicmp(en,contextPatterns[i].pattern,len)) continue;
    strncpy(cntxt,contextPatterns[i].context,CONTEXT_NAME_SIZE);
    return;
  }
  cntxt[0] = 0;
  return;
}
int serverNameExists(char *ctx,char *srv)
{
  int i = getServerTableIndex(ctx,srv);
  return i+1; /* TRUE or FALSE */
}
typedef struct
{
  char name[FEC_NAME_SIZE];
  char abbr[FEC_NAME_SIZE];
} FECABBR;
FECABBR *fecAbbr = NULL;
#define FECABBR_FILE "abbreviations.csv"
int nFecAbbr = 0;
int getFecAbbreviations(char *fn)
{
  FILE *fp = NULL;
  int n,i,cc=0,done=0;
  int abbr_col=-1,name_col=-1;
  char str[256],hdr[256],scratch[64];

  if (fn == NULL || strlen(fn) == 0) ccerr(un_allocated);
  if (fecAbbr != NULL)
  {
    free(fecAbbr);
    nFecAbbr = 0;
  }
  if ((n=GetDataBaseSize(fn)) <= 0) ccerr(-n);
  if ((fecAbbr=(FECABBR *)calloc(n,sizeof(FECABBR))) == NULL) ccerr(out_of_server_memory);
  nFecAbbr = n;
  if ((fp=fopen(fn,"rt")) == NULL) ccerr(no_such_file);
  for (done=0,i=0; csvGetLine(str,255,fp) != NULL && i<ncontextPatterns;)  /* scan file */
  {
    strtrm(str);
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((abbr_col=findcol(hdr,"ABBREVIATION")) < 0) ccerr(no_such_line);
      if ((name_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(abbr_col,str),FEC_NAME_SIZE);
    if (!CHECK_VALID_CHARS(scratch)) continue;
    strncpy(fecAbbr[i].abbr,scratch,FEC_NAME_SIZE);
    strncpy(scratch,colptr(name_col,str),FEC_NAME_SIZE);
    if (!CHECK_VALID_CHARS(scratch)) continue;
    strncpy(fecAbbr[i].name,scratch,FEC_NAME_SIZE);
    i++;
  }
  nFecAbbr = i;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
char *getAbbreviatedFecName(char *srv,int maxlen)
{
  static char fec[FEC_NAME_SIZE+8];
  int i,k,slen,nlen,alen,sflen=0;
  char *c, *suf=NULL, tmp[EXPORT_NAME_SIZE+8];
  strncpy(tmp,srv,EXPORT_NAME_SIZE); tmp[EXPORT_NAME_SIZE] = 0;
  slen = (int)strlen(tmp);
  for (i=slen-1; i > 0 && isdigit(tmp[i]); i--);
  suf = &tmp[i+1];
  if ((sflen= (int)strlen(suf)) > 5)
  { /* allow up to a 5-digit numerical name */
    suf = NULL; 
    sflen = 0;
  }
  slen -= sflen;
  for (i=0; i<nFecAbbr; i++)
  {
    if ((c=stristr(tmp,fecAbbr[i].name)) != NULL)
    {
      nlen = (int)strlen(fecAbbr[i].name);
      alen = (int)strlen(fecAbbr[i].abbr);
      strncpy(c,fecAbbr[i].abbr,alen);
      c += alen; 
      for (k=0; k<nlen-alen; k++) *c++ = '*';
    }
  }
  memset(fec,0,FEC_NAME_SIZE+8);
  for (i=0,k=0; i<slen && k<maxlen; i++)
  {
    if (tmp[i] == 0) break;
    if (tmp[i] == '*') continue;
    if (tmp[i] == '.') continue;
    if (tmp[i] == '_') continue;
    if (tmp[i] == '-') continue;
    fec[k++] = tmp[i];
  }
  if (sflen > 0)
  {
    if ((int)strlen(fec) > maxlen-sflen) fec[maxlen-sflen-1] = 0;
    strncat(fec,suf,sflen);
  }
  return fec;
}
char *suggestFecName(char *ctx,char *srv,int maxlen)
{
  static char fec[FEC_NAME_SIZE+8];
  char cx='X';
  int i,k,p,len;

  memset(fec,0,FEC_NAME_SIZE+8);
  if (srv == NULL) return fec;
  if (ctx != NULL)
  {
    char *c;
    removeDecoration(ctx);
    if ((c=strstr(ctx,".SIM")) != NULL) *c++ = 0;
    else if ((c=strstr(ctx,"_SIM")) != NULL) *c++ = 0;
    else if ((c=strstr(ctx,".TEST")) != NULL) *c++ = 0;
    else if ((c=strstr(ctx,"_TEST")) != NULL) *c++ = 0;
    for (i=0; i<ncontextPatterns && strnicmp(ctx,contextPatterns[i].context,CONTEXT_NAME_SIZE); i++);
    if (i < ncontextPatterns) 
    {
      strncpy(fec,contextPatterns[i].pattern,3);
      if (c != NULL) strncat(fec,c,1);
    }
  }
  if (maxlen > FEC_NAME_SIZE) maxlen = FEC_NAME_SIZE;
  len = maxlen - (int)strlen(fec);
  strncat(fec,getAbbreviatedFecName(srv,len),len);
  strupr(fec);
  for (i=0; i<FecTblSize; i++)
  {
    if (FecInfoEx[i].disable) continue;
    if (!strnicmp(fec,FecTbl[i].fecName,FEC_NAME_SIZE))
    { /* name collision ! */
      if (strstr(fec,"XXXX") != NULL) cx = 'Y';
      if (strstr(fec,"YYYY") != NULL) cx = 'Z';
      if (strstr(fec,"ZZZZ") != NULL) return "";
      if ((p=(int)strlen(fec)) == maxlen)
      { /* fec name already at 16 (or maxlen) chars -> backup ! */
        p--;
        for (k=0; k<3; k++) if (fec[p] == cx) p--;
      }
      fec[p] = cx;
    }
  }
  if ((int)strlen(fec) < (maxlen - 2) && strstr(fec,"FEC") == NULL) strcat(fec,"FEC");
  return fec;
}
void stopENS(void)
{
  dumpFECTable();
  dumpEQPTable();
}
BYTE *getFwdIpAddr(char *tgt)
{ // TODO: ipv6
  char fwdTgt[128],*ctx=NULL,*srv=NULL;
  int i,sid;
  strncpy(fwdTgt,tgt,64);
  if (fwdTgt[0] != '/') return NULL;
  ctx = &fwdTgt[1];
  if ((srv=strchr(ctx,'/')) == NULL) return NULL;
  *srv++ = 0;
  if ((sid=getServerTableIndex(ctx,srv)) < 0) return 0;
  for (i=0; i<FecTblSize; i++)
  {
    if (!strnicmp(FecTbl[i].fecName,SrvTbl[sid].FecName,FEC_NAME_SIZE))
    {
      return (BYTE *)&FecTbl[i].netAdr;
    }
  }
  return NULL;
}
void chkCmdsLog(void)
{
  FILE *fp1=NULL, *fp2=NULL;
  static int done = FALSE;
  if (done) return;
  if ((fp1=fopen("commands.log","r")) != NULL &&
      (fp2=fopen(CMDSFILE,"r")) == NULL)
  { // a sure sign that we need to rename 'commands.log' to 'cmds.log'
    fclose(fp1); fp1 = NULL;
    rename("commands.log",CMDSFILE);
  }
  if (fp1 != NULL) fclose(fp1);
  if (fp2 != NULL) fclose(fp2);
  done = TRUE;
  return;
}
void initENS(void)
{
  int cc;
  char *ptr = NULL;
  BYTE *ver = SystemVersion();
  static char myaddr[40] = "";
  static BYTE badr[40];

  debug = 1;
  dbglog("Initializing Equipment Name Server ...");
  registerSrvQueryInfoStruct();
  ens_iaddr = inadr_zero;
  SetAutoLinkErrorAlarms(FALSE);
  if (!NameServerLoaded)
  {
    dbglog("Allocating Name Server space ...");
    if ((NameServer=(FecAddrStruct *)calloc(1,sizeof(FecAddrStruct))) == NULL)
    {
      dbglog("Cannot allocate Name Server space !");
      exit(1);
    }
  }
  if ((ptr=getenv("ENS_ALLOWED_DEADTIME")) != NULL)
  {
    gAllowedDeadTime = atoi(ptr);
    if (gAllowedDeadTime < ONEHOUR) gAllowedDeadTime = ONEHOUR;
  }
  getAllowedContexts(ROOTCTX_FNAME);
  getContextPatternInformation(CONTEXTS_FILE);
  getFecAbbreviations(FECABBR_FILE);
  getAdministrators(ADMINS_FILE);
  getAllowedSubsystems(SUBSYS_FNAME);
  getExportCorrections(EXPCOR_FNAME);
  getFixedAddresses(FIXADR_FNAME);
  if ((cc=loadENS()) > 0)  
  { /* brand new startup -> no database files ? */
    if (cc == no_such_file)
    {
      if ((SrvTbl=(ExpDataStruct *)calloc(TMPSPACE,sizeof(ExpDataStruct))) == NULL) return;
      if ((SrvInfoTbl=(SrvInfo *)calloc(TMPSPACE,sizeof(SrvInfo))) == NULL) return;
      if ((FecTbl=(FecAddrStruct *)calloc(TMPSPACE,sizeof(FecAddrStruct))) == NULL) return ;
      if ((FecInfoTbl=(FecInfoStruct *)calloc(TMPSPACE,sizeof(FecInfoStruct))) == NULL) return;
      if ((FecInfoEx=(FecInfoStructEx *)calloc(TMPSPACE,sizeof(FecInfoStructEx))) == NULL) return;
      SrvTblSize = FecTblSize = 1;
      memcpy(badr,getmyipaddr(),gInAdrLen);
      if (zerocmp(badr,gInAdrLen))
      { /* good!  my address didn't come up empty */
        inet_ntop(gAfInet,badr,myaddr,40);
      }
      else
      {
        SCKADR sckadr;
        sckadr.ss_family = (short)gAfInet;
        setInAddrLoopback(&sckadr);
        inet_ntop(gAfInet,getInAddr(&sckadr),myaddr,40);
      }
      strcpy(FecTbl[0].fecName,"ENS");
      strncpy(FecTbl[0].strAdr,myaddr,ADDR_SIZE_IPV6);
      FecTbl[0].portOffset = 0;
      FecTbl[0].addrFlags |= (gMultiCastUseLegacyRules ? ADDRFLAG_MCAST_LEGACY : ADDRFLAG_MCAST_MODERN);
      strncpy(FecInfoTbl[0].os,OS_TAG,FEC_OS_SIZE);
      sprintf(FecInfoTbl[0].ver,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
      strcpy(FecInfoTbl[0].desc,"Equipment name server");
      strcpy(FecInfoTbl[0].hdw,"none");
      FecInfoEx[0].importance = 2;
      strcpy(SrvTbl[0].ExportName,"ENS");
      strcpy(SrvTbl[0].FecName,"ENS");
      strcpy(SrvTbl[0].EqmName,"ENSEQM");
      strcpy(SrvTbl[0].EqmContext,"SITE");
      strcpy(SrvTbl[0].SubSystem,"SER");
      sprintf(logbuf,"ENS: creating initial database"); 
      dumpFECTable();
      dumpEQPTable();
      scanTables();
    }
    else
    {
      sprintf(logbuf,"ENS: %.32s",erlst[cc]);
    }
    nslog(logbuf); 
  }
  checkCanonicalLocations();
  getForwardsList(ENSFWD_FILE);
  scanFlag = 0;
  initSinkInfo();
  chkCmdsLog();
  dbglog("Equipment Name Server Initialized");
  nslog("initENS: initialized");
  feclog("ENS: initialized");
  debug = 0;
  return;
}
void removeNotRunning(char *ctx)
{
  int i, chkCtx=FALSE;
  time_t tnow = time(NULL);
  if (ctx != NULL && strlen(ctx) > 0) chkCtx = TRUE;
  nslog("removal of servers in context %.32s started",chkCtx ? ctx : "(all contexts)");
  for (i=0; i<SrvTblSize; i++)
  {
    if (SrvInfoTbl[i].disabled) continue;
    if (chkCtx && strlen(ctx) > 0 && strnicmp(ctx,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE) != 0) continue;
    SrvInfoTbl[i].lastPinged = tnow - ONEHOUR; /* force immediate ping */
    SrvInfoTbl[i].lastActive = tnow - gAllowedDeadTime; /* ping must respond ! */
  }
}
void pingcallback(int id,int cc)
{
  SrvInfoTbl[id].pingPending = FALSE;
  if (cc == 0) 
  {
    SrvInfoTbl[id].lastActive = SrvInfoTbl[id].lastPinged;
  }
  else
  {
    pinglog(SrvTbl[id].EqmContext,SrvTbl[id].ExportName);
  }
}
void pingServers(void)
{
  DTYPE dout;
  int i, k, lid;
  char fulldevname[256];
  time_t tnow = time(NULL);

  for (i=0; i<SrvTblSize; i++)
  {
    if (SrvInfoTbl[i].disabled) continue;
    if (isENS(SrvTbl[i].ExportName)) continue;
    if (isGENS(SrvTbl[i].ExportName)) continue;
    if (!strcmp(SrvTbl[i].FecName,"GENS")) continue; /* let the GENS manage these */
    if (SrvInfoTbl[i].lastActive > SrvInfoTbl[i].lastPinged) 
    { /* how did this happen ? */
      SrvInfoTbl[i].lastPinged = SrvInfoTbl[i].lastActive;
    }
    if (SrvInfoTbl[i].lastPinged > tnow - ONEHOUR) 
    { /* recently pinged */
      if (!SrvInfoTbl[i].pingPending && SrvInfoTbl[i].lastActive < tnow - gAllowedDeadTime)
      { /* appears to be dead */
        if (SrvInfoTbl[i].wasLogged) continue;
        if ((k=SrvInfoTbl[i].fecIndex) >= 0 && FecInfoEx[k].importance == 0)
        { /* is NOT marked as IMPORTANT */
          SrvInfoTbl[i].disabled = TRUE;
          srvlog("activity","ping","disabling long-term inactive server /%.32s/%.32s",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
        }
        else
        {
          srvlog("activity","ping","important server /%.32s/%.32s is long-term inactive !",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
        }
        SrvInfoTbl[i].wasLogged = TRUE;
      }
      continue;
    }
    SrvInfoTbl[i].wasLogged = FALSE;
    SrvInfoTbl[i].pingPending = TRUE;
    SrvInfoTbl[i].lastPinged = tnow;
    if (strlen(SrvTbl[i].EqmContext) > 0)
      sprintf(fulldevname,"/%.32s/%.32s/#0",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
    else
      sprintf(fulldevname,"%.32s/#0",SrvTbl[i].ExportName);
    memset(&dout,0,sizeof(DTYPE));
    dout.dFormat = CF_SHORT;
    dout.dArrayLength = 1;
    dout.data.sptr = &numRemoteProps;
    lid = AttachLinkEx(fulldevname,"NPROPS",&dout,NULL,CA_READ,1000,pingcallback,CM_SINGLE,i);
    if (lid < 0) srvlog("activity","ping","cannot ping %.192s : <%d>",fulldevname,-lid);
    break; /* just do one at a time (help stagger the pings) */
  }
}
void taskENS(void)
{
  int i, cc;
  static int counter = 0;
  static time_t interim = 0;
  time_t t = time(NULL);
  if (checkExportFlag)
  {
    checkExport();
  }
  else
  {
    pingServers();
    if (dumpFecFileFlag) dumpFECTable();
    if (dumpEqpFileFlag > 0) dumpEQPTable();
  }
  for (i=0; i<nforwards; i++) chkForwardsQueue(i);
  if (checkExportFlag) return;
  if (t < interim + 60) return;
  if (dumpEqpFileFlag < 0) dumpEQPTable();
  interim = t; /* check for a new database each minute */
  counter++;   /* count the minutes                    */
  if ((cc=loadENS()) > 0) feclog("ENS: loadENS -> %.32s",erlst[cc]);
  if ((counter % 5) == 0) rmvWidowedFecs();
  if (counter >= 60) counter = 0;
  return;
}
typedef struct
{
  char *elem;
  int   len;
} CSVINFO; 

CSVINFO csvFecInfo[] = {
  { gTgtFecInfo.os,FEC_OS_SIZE },
  { gTgtFecInfo.desc,FEC_DESC_SIZE },
  { gTgtFecInfo.loc,FEC_LOCATION_SIZE },
  { gTgtFecInfo.hdw,FEC_HDW_SIZE },
  { gTgtFecInfo.resp,FEC_RESP_SIZE },
  { gTgtFecInfo.ver,FEC_VERSION_SIZE }
};
  
void checkCSVFormat(CSVINFO *inf,int num)
{
  char scratch[80];
  char *ptr,*loc;
  int i,j,slen,pos;

  for (i=0; i<num; i++)
  {
    loc = inf[i].elem; slen = inf[i].len;
    strncpy(scratch,loc,slen); scratch[slen] = 0;
    for (j=0; j<slen; j++) if (scratch[j] == '\n') scratch[j] = ' ';
    if ((ptr=strchr(scratch,'"')) != NULL)
    {
      if (!(ptr == scratch && scratch[strlen(scratch)-1] == '"'))
        loc[0] = 0;  /* don't know what's going on, and don't want to figure it out */
    }
    else if ((ptr=strchr(loc,',')) != NULL)
    {
      if ((pos=(int)strlen(loc)) >= slen-2) pos = slen-2;
      loc[0] = '"'; strncpy(&loc[1],scratch,slen-2);
      loc[pos+1] = '"';
    }
  }
}
int getSearchInfo(DTYPE *d,char *subs,int *imp)
{
  int len;
  int importance = 0, i;
  char scratch[64];
  clearSearchLists();
  if (d->dArrayLength > 0)
  {
    memset(subs,0,32);
    switch (d->dFormat)
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        len = fmtsizeof(d->dFormat);
        if (d->dArrayLength > 1)
        {
          strncpy(scratch,d->data.cptr+len,len);
          for (i=0; i<nImportanceLevels; i++)
            if (!stricmp(gImportanceLevel[i],scratch)) break;
          if (i < nImportanceLevels) importance = i;
        }
        break;
      case CF_TEXT:
        if (strchr(d->data.cptr,'/') != NULL)
        { /* volatile use of the input data : use only within EQM dispatch */
          return parseSearchString(d->data.cptr);
        }
        len = MIN(32,d->dArrayLength);
        break;
      default: 
        return illegal_format;
    }
    strncpy(subs,d->data.cptr,len); 
    if (imp != NULL) *imp = importance;
  }
  return 0;
}
int getRebootCount(char *ctx,char *srv)
{
  int i;
  if ((i=getServerTableIndex(ctx,srv)) < 0) return -1;
  return SrvInfoTbl[i].rebootCounts;
}
int getRebootCountAll(char *cntxt,NAME32I *ni,int len)
{
  int i,n=0;
  if (cntxt == NULL || ni == NULL) return -invalid_parameter;
  for (i=0; i<SrvTblSize && n<len; i++)
  {
    if (SrvInfoTbl[i].disabled) continue;
    if (strnicmp(cntxt,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    strncpy(ni[n].name,SrvTbl[i].ExportName,32);
    ni[n].ival = SrvInfoTbl[i].rebootCounts;
    n++;
  }
  return n;
}
time_t getServerLastActive(char *ctx,char *srv)
{
  int i;
  if ((i=getServerTableIndex(ctx,srv)) < 0) return 0;
  return SrvInfoTbl[i].lastActive;
}
int removeDecoration(char *context)
{
  char *c;
  if ((c=strchr(context,'.')) != NULL)
  {
    int i;
    char *udc, *ss;
    udc = c++;
    if (isSystemContextDecoration(c)) return invalid_data;
    *udc = 0; /* okay to remove decoration */
    ss = getShortSubSystemName(c,TRUE);
    /* MISC is returned as a best effort of getShortSubSystem() ! */
    if (!strcmp(ss,"MISC") && strnicmp(ss,c,4)) return invalid_data;
    if ((c=strchr(context,'_')) != NULL)
    { /* check "_TEST" or "_SIM" again (the famous doocs 'XFEL_SIM.DIAG' case) */
      if (isSystemContextDecoration(&c[1]))
      { /* seems to be! replace with a '.' */
        *c = '.';
      }
    }
    for (i=0; i<nSubSystems; i++)
    { /* first see if decoration is a valid subsystem */
      if (!strnicmp(ss,subSystems[i].name,SUBSYSTEM_NAME_SIZE)) return 0;
    }
  }
  return invalid_data;
}
int srvInSearchList(int id)
{
  int i, fid, retry_undecorated=FALSE;
  char *c;
  if (SrvInfoTbl[id].disabled) return FALSE;
  if (!gUseSearchLists) return TRUE;
retryContextSearch:
  for (i=0; i<ctxSearch.len; i++)
  {
    if (!strnicmp(SrvTbl[id].EqmContext,ctxSearch.lst[i],CONTEXT_NAME_SIZE)) break;
  }
  if (ctxSearch.len > 0 && ctxSearch.len == i) 
  {
    if (!retry_undecorated)
    {
      for (i=0; i<ctxSearch.len; i++) if (removeDecoration(ctxSearch.lst[i]) == 0) retry_undecorated = TRUE;
      if (retry_undecorated) goto retryContextSearch;
    }
    return FALSE;
  }
  for (i=0; i<srvSearch.len; i++)
  {
    if (!strnicmp(SrvTbl[id].ExportName,srvSearch.lst[i],SUBSYSTEM_NAME_SIZE)) break;
  }
  if (srvSearch.len > 0 && srvSearch.len == i) return FALSE;
  for (i=0; i<subSearch.len; i++)
  {
    if (!strnicmp(SrvTbl[id].SubSystem,subSearch.lst[i],SUBSYSTEM_NAME_SIZE)) break;
  }
  if (subSearch.len > 0 && subSearch.len == i) return FALSE;
  for (i=0; i<eqmSearch.len; i++)
  {
    if (!strnicmp(SrvTbl[id].EqmName,eqmSearch.lst[i],EQM_NAME_SIZE)) break;
  }
  if (eqmSearch.len > 0 && eqmSearch.len == i) return FALSE;
  for (i=0; i<fecSearch.len; i++)
  {
    if (!strnicmp(SrvTbl[id].FecName,fecSearch.lst[i],FEC_NAME_SIZE)) break;
  }
  if (fecSearch.len > 0 && fecSearch.len == i) return FALSE;

  fid = SrvInfoTbl[id].fecIndex;
  if (FecInfoEx[fid].disable) return FALSE;
  for (i=0; i<impSearch.len; i++)
  {
    if (FecInfoEx[fid].importance < getImportanctLevel(impSearch.lst[i])) continue;
  }
  if (impSearch.len > 0 && impSearch.len == i) return FALSE;

  for (i=0; i<ipSearch.len; i++)
  {
    if ((c=strchr(ipSearch.lst[i],'*')) != NULL) *c = 0;
    if (!strnicmp(FecTbl[fid].strAdr,ipSearch.lst[i],strlen(ipSearch.lst[i]))) break;
  }
  if (ipSearch.len > 0 && ipSearch.len == i) return FALSE;
  for (i=0; i<osSearch.len; i++)
  {
    if (!strnicmp(FecInfoTbl[fid].os,osSearch.lst[i],8)) break;
  }
  if (osSearch.len > 0 && osSearch.len == i) return FALSE;
  for (i=0; i<rspSearch.len; i++)
  {
    if (stristr(FecInfoTbl[fid].resp,rspSearch.lst[i]) != NULL) break;
  }
  if (rspSearch.len > 0 && rspSearch.len == i) return FALSE;

  return TRUE;
}
int getNumExportTags(char *subsys,char *cntxt,int importance)
{
  int i,j,k=-1,n=0;
  int subschk=0,cntxtchk=0,isFec=0,isEqm=0;
  char *c, *subs = subsys, eqm[16]; 

  if (!gUseSearchLists)
  { /* old style */
    if (cntxt != NULL &&  strlen(cntxt) > 0) cntxtchk = -1;
    if (!stricmp(cntxt,"DEFAULT")) cntxtchk = 0;
    if (subsys != NULL && !stricmp(subsys,"ALL")) subsys[0] = 0;
    if (subsys != NULL && strlen(subsys) > 0)
    {
      if ((i=GetPropertyId("ENSEQM",subsys)) >= 0)
      {
        for (k=0; k<FecTblSize && strnicmp(subsys,FecTbl[k].fecName,FEC_NAME_SIZE); k++);
        if (k < FecTblSize) isFec = -1;
      }
      if (subsys[0] == '(')
      { /* specific equipment module name */
        strncpy(eqm,&subsys[1],EQM_NAME_SIZE);
        eqm[EQM_NAME_SIZE] = 0;
        if ((c=strchr(eqm,')')) != NULL) *c = 0;
        isEqm = -1;
      }
      else if (!isFec)
      {
        if (!strnicmp(subsys,"UNKNOWN",SUBSYSTEM_NAME_SIZE)) subschk = -1; else subschk = 1;
        subs = getShortSubSystemName(subsys,FALSE);
      }
    }
    if (subschk == 0 && cntxtchk == 0 && isFec == 0 && isEqm == 0 && importance <= 0) 
      return SrvTblSize;
  }
  else
  { /* context search coming in extra ? */
    if (cntxt != NULL && strlen(cntxt) > 0)
    {
      for (i=0; i<ctxSearch.len; i++) if (!strnicmp(ctxSearch.lst[i],cntxt,CONTEXT_NAME_SIZE)) break;
      if (i == ctxSearch.len)
      { /* not found ! */
        if (ctxSearch.len < MAX_SEARCH_PARAMS)
        {
          ctxSearch.lst[i] = cntxt;
          ctxSearch.len++;
        }
      }
    }
  }
retry_undecorated:
  for (i=0; i<SrvTblSize; i++)
  {
    if (gUseSearchLists)
    {
      if (!srvInSearchList(i)) continue;
    }
    else
    {
      if (SrvInfoTbl[i].disabled) continue;
      if (cntxtchk && strnicmp(cntxt,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
      j = SrvInfoTbl[i].fecIndex;
      if (isFec && k != j) continue;
      if (!isFec && FecInfoEx[j].importance < importance) continue;
      if (isEqm && strnicmp(SrvTbl[i].EqmName,eqm,EQM_NAME_SIZE)) continue;
      if (subschk > 0 && stricmp(SrvTbl[i].SubSystem,subs)) continue;
      if (subschk < 0 && strlen(SrvTbl[i].SubSystem) > 0) continue;
    }
    n++;
  }
  if (n == 0 && cntxtchk && removeDecoration(cntxt) == 0)
   goto retry_undecorated;
  return n;
}
int getNumKnownFecs(char *contxt,char *subsys,int importance)
{
  int i,j,k,n=0;
  char *subs = subsys;
  int chksubs = (subsys == NULL || strlen(subsys) == 0) ? 0 : 1;
  int chkcon  = (contxt == NULL || strlen(contxt) == 0) ? 0 : 1;
  if (chkcon && !stricmp(contxt,"DEFAULT")) chkcon = 0;

  if (gUseSearchLists)
  { /* modern style */
    if (contxt != NULL && strlen(contxt) > 0)
    { /* context coming in extra */
      for (i=0; i<ctxSearch.len; i++) if (!strnicmp(ctxSearch.lst[i],contxt,CONTEXT_NAME_SIZE)) break;
      if (i == ctxSearch.len)
      { /* not found ! */
        if (ctxSearch.len < MAX_SEARCH_PARAMS)
        {
          ctxSearch.lst[i] = contxt;
          ctxSearch.len++;
        }
      }
    }
    if (ctxSearch.len) 
    {
      for (i=0; i<SrvTblSize; i++) if (srvInSearchList(i)) 
      { /* don't double count FECs ! */
        for (j=0; j<i; j++)
          if (!strnicmp(SrvTbl[j].FecName,SrvTbl[i].FecName,FEC_NAME_SIZE)) break;
        if (j < i) continue; 
        n++;
      }
      return n;
    }
    chksubs = 0;
  }

  if (chksubs && !stricmp(subsys,"ALL")) chksubs = 0;
  if (!chksubs)
  {
    if (!chkcon && importance <= 0) return FecTblSize;
  }
  else
  {
    if (!strnicmp(subsys,"UNKNOWN",32)) chksubs = -1;
    subs = getShortSubSystemName(subsys,FALSE);
  }

  if (chkcon)
  {
retry_undecorated:
    for (i=0; i<SrvTblSize; i++)
    {
      if (SrvInfoTbl[i].disabled) continue;
      if (strnicmp(SrvTbl[i].EqmContext,contxt,CONTEXT_NAME_SIZE)) continue;
      for (j=0; j<i; j++)
        if (!strnicmp(SrvTbl[j].FecName,SrvTbl[i].FecName,FEC_NAME_SIZE) &&
            !strnicmp(SrvTbl[j].EqmContext,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) break;
      if (j < i) continue; 
      k = SrvInfoTbl[i].fecIndex;
      if (k < 0 || k >= FecTblSize) continue;
      if (FecInfoEx[k].disable) continue;
      if (FecInfoEx[k].importance < importance) continue;
      if (chksubs)
      {
        if (chksubs > 0 && strnicmp(SrvTbl[k].SubSystem,subs,SUBSYSTEM_NAME_SIZE)) continue;
        if (chksubs < 0 && strlen(SrvTbl[k].SubSystem) > 0) continue;
      }
      n++;
    }
    if (n == 0 && removeDecoration(contxt) == 0)
      goto retry_undecorated;
  }
  else
  {
    for (i=0; i<FecTblSize; i++)
    {
      if (FecInfoEx[i].disable) continue;
      if (FecInfoEx[i].importance < importance) continue;
      n++;
    }
  }
  return n;
}
void rmvWidowedFecs(void)
{
  int i,k;
  for (i=0; i<FecTblSize; i++)
  {
    if (FecInfoEx[i].disable) continue;
    for (k=0; k<SrvTblSize; k++)
    { /* gotta find a matching entry */
      if (SrvInfoTbl[k].fecIndex == i) break;
    }
    if (k == SrvTblSize)
    { /* oops! */
      FecInfoEx[i].disable = TRUE;
      nslog("remove widowed FEC %.16s (%.40s)",FecTbl[i].fecName,FecTbl[i].strAdr);
      dumpFecFileFlag = 1;
    }
  }
}
int getKnownFecs(DTYPE *d,char *contxt,char *subsys,int importance)
{
  DUSTRING *ustr = NULL;
  char *subs = subsys;
  int i,j,k,m=0,n=-1,siz,dsiz,nlen,nleft;
  UINT32 haddr = 0;
  int chksubs = (subsys == NULL || strlen(subsys) == 0) ? 0 : 1;
  int chkcon  = (contxt == NULL || strlen(contxt) == 0) ? 0: 1;
  int chkip = FALSE;
  int allsubs=FALSE, isTag=FALSE, isMulti=FALSE;

  if (gUseSearchLists)
  { /* modern style */
    if (contxt != NULL && strlen(contxt) > 0)
    { /* context coming in extra */
      for (i=0; i<ctxSearch.len; i++) if (!strnicmp(ctxSearch.lst[i],contxt,CONTEXT_NAME_SIZE)) break;
      if (i == ctxSearch.len)
      { /* not found ! */
        if (ctxSearch.len < MAX_SEARCH_PARAMS)
        {
          ctxSearch.lst[i] = contxt;
          ctxSearch.len++;
        }
      }
    }
    if (ctxSearch.len) chkcon = TRUE;
    chksubs = 0;
    isMulti = TRUE;
  }
  if (chksubs)
  {
    if (!stricmp(subsys,"ALL")) { allsubs = TRUE; chksubs = 0; }
    if (!stricmp(subsys,"MULTI")) { isMulti = TRUE; chksubs = 0; }
  }
  dsiz = d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_NAME8: siz = 8; break;
    case CF_NAME16: siz = 16; break;
    case CF_NAME32: siz = 32; break;
    case CF_TEXT: siz = 16; dsiz /= 16; break;
    case CF_USTRING: 
      siz = 16; 
      ustr = (DUSTRING *)d->data.vptr;
      break;
    default: return illegal_format;
  }
  if (chksubs)
  {
    if ((i=GetPropertyId("ENSEQM",subsys)) >= 0)
    { /* kluge : passed an export name instead of subsystem ! */
      for (n=0; n<SrvTblSize && strnicmp(subsys,SrvTbl[n].ExportName,EXPORT_NAME_SIZE); n++);
      if (n < SrvTblSize) 
      {
        isTag = TRUE;
        chksubs = 0;
      }
    }
    else
    { /* maybe it's an ip query ? TODO: ipv6 */
      if ((haddr=getIPhaddr(subsys)) != 0) chkip = TRUE;
    }
    if (!isTag && !chkip)
    {
      if (!strnicmp(subsys,"UNKNOWN",32)) chksubs = -1;
      subs = getShortSubSystemName(subsys,FALSE);
    }
  }
  if (chkcon && !stricmp(contxt,"DEFAULT")) chkcon = 0;
  if (isTag || chkcon || chksubs || allsubs || isMulti)
  { /* filter parameters passed */
retry_undecorated:
    for (i=0,k=0; i<SrvTblSize && k<dsiz; i++)
    {
      if (gUseSearchLists)
      { /* modern search */
        if (!srvInSearchList(i)) continue;
      }
      else
      { /* old style */
        if (SrvInfoTbl[i].disabled) continue;
        if (chkcon && strnicmp(SrvTbl[i].EqmContext,contxt,CONTEXT_NAME_SIZE)) continue;
        if (isTag)
        { /* match the export name */
          if (strnicmp(SrvTbl[i].ExportName,subs,EXPORT_NAME_SIZE)) continue;
          if ((m=SrvInfoTbl[i].fecIndex) < 0 || m >= FecTblSize) continue;
        }
        else
        { /* don't double count the FECS in a subsystem */
          for (j=0; j<i; j++)
            if (!strnicmp(SrvTbl[j].FecName,SrvTbl[i].FecName,FEC_NAME_SIZE) &&
                !strnicmp(SrvTbl[j].EqmContext,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) break;
          if (!isMulti)
          { /* another server on same FEC ? */
            if (j < i) continue; 
          }
          else
          { /* looking for FECS with multiple servers ... */
            if (j == i) continue; 
          }
        }
        if ((m=SrvInfoTbl[i].fecIndex) < 0 || m >= FecTblSize) continue;
        if (FecInfoEx[m].disable) continue;
        if (FecInfoEx[m].importance < importance) continue;
        if (chkip)
        { /* match the IP address */ // TODO: ipv6
          if (memcmp(FecTbl[m].netAdr.ipv4,&haddr,4)) continue;
        }
        else if (chksubs)
        { /* match the subsystem */
          if (chksubs > 0 && strnicmp(SrvTbl[i].SubSystem,subs,SUBSYSTEM_NAME_SIZE)) continue;
          if (chksubs < 0 && strlen(SrvTbl[i].SubSystem) > 0) continue;
        }
      }
      if (ustr != NULL)
      { /* more than just the fec names ... */
        strncpy(ustr[k].str,SrvTbl[i].FecName,16);
        strncpy(&ustr[k].str[16],FecInfoTbl[m].os,8);
        strncpy(&ustr[k].str[24],SrvTbl[i].SubSystem,8);
        nlen = (int)MIN(strlen(SrvTbl[i].ExportName),32);
        strncpy(&ustr[k].str[32],SrvTbl[i].ExportName,nlen);
        ustr[k].str[32+nlen] = 0; nleft = 80 - 32 - nlen - 1 ;
        if (chkcon)
        {
          strncpy(&ustr[k].str[32+nlen+1],FecInfoTbl[m].desc,nleft);
        }
        else
        {
          strncpy(&ustr[k].str[32+nlen+1],SrvTbl[i].EqmContext,nleft);
        }
        memcpy(&ustr[k].ival,FecTbl[m].netAdr.ipv4,4);
        ustr[k].tm = FecInfoEx[m].importance;
      }
      else
      {
        if (isMulti)
        { /* don't double count the FEC names */
          for (j=0; j<k; j++)
            if (!strnicmp(&d->data.cptr[j*siz],SrvTbl[i].FecName,FEC_NAME_SIZE)) break;
          if (j < k) continue;
        }
        strncpy(&d->data.cptr[k*siz],SrvTbl[i].FecName,siz);
      }
      k++;
    }
    if (k == 0 && chkcon && removeDecoration(contxt) == 0)
      goto retry_undecorated;
  }
  else
  { /* no filter parameters passed => no srv table xref */
    for (i=0,k=0; i<FecTblSize && k<dsiz; i++)
    {
      if (FecInfoEx[i].disable) continue;
      if (FecInfoEx[i].importance < importance) continue;
      if (ustr != NULL)
      {
        strncpy(ustr[k].str,FecTbl[i].fecName,16);
        strncpy(&ustr[k].str[16],FecInfoTbl[i].os,8);
        strncpy(&ustr[k].str[48],FecInfoTbl[i].desc,32);
        memcpy(&ustr[k].ival,FecTbl[i].netAdr.ipv4,4);
        ustr[k].tm = FecInfoEx[i].importance;
      }
      else
      {
        strncpy(&d->data.cptr[k*siz],FecTbl[i].fecName,siz);
      }
      k++;
    }
  }
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}
int getSubsystemForServer(DTYPE *d,char *ctx,char *srv)
{
  int i = getServerTableIndex(ctx,srv);
  char *c = NULL;
  if (i < 0) return non_existent_server;
  return PutValuesFromString(d,SrvTbl[i].SubSystem,CF_TEXT,SUBSYSTEM_NAME_SIZE,0);
}
int getKnownSubsystems(DTYPE *d,char *contxt)
{
  char *ptr=NULL, *tptr, *nptr;
  int i,j,k,siz,asiz,dsiz;
  int dir = 0, is = -1;
  int chkcon  = (contxt == NULL || strlen(contxt) == 0) ? 0 : 1;

  asiz = d->dArrayLength;
  dsiz = fmtsizeof(d->dFormat);
  ptr = d->data.cptr;
  switch (d->dFormat)
  {
    case CF_NAME16: siz = 16; break;
    case CF_NAME32: siz = 32; break;
    case CF_TEXT: dsiz = 16; asiz /= 16; break;
    case CF_USTRING: 
      siz = 16; 
      ptr = ((DUSTRING *)d->data.vptr)->str;
      break;
    default: return illegal_format;
  }
  if (asiz < 1) return dimension_error;
  if (chkcon && !stricmp(contxt,"DEFAULT")) chkcon = 0;
  if (chkcon)
  {
retry_undecorated:
    for (i=0,k=0; i<SrvTblSize && k<asiz; i++)
    {
      if (SrvInfoTbl[i].disabled) continue;
      if (strnicmp(SrvTbl[i].EqmContext,contxt,CONTEXT_NAME_SIZE)) continue;
      nptr = SrvTbl[i].SubSystem; is = -1;
      for (j=0; j<k; j++) 
      {
        tptr = ptr + j*dsiz;
        if ((dir=strnicmp(nptr,tptr,SUBSYSTEM_NAME_SIZE)) == 0) break;
        if (dir < 0 && is < 0)
        { /* make note of first case where string1 < string2 */
          is = j;
        }
      }
      if (j < k) continue; /* already counted */
      if (strlen(SrvTbl[i].SubSystem) == 0) continue;
      if (is < 0)
      {
        strncpy(ptr+k*dsiz,nptr,SUBSYSTEM_NAME_SIZE);
      }
      else
      { /* insert entry in proper place */
        for (j=k; j>is; j--)
        { /* move these up */
          strncpy(ptr+j*dsiz,ptr+(j-1)*dsiz,SUBSYSTEM_NAME_SIZE);
        }
        strncpy(ptr+is*dsiz,nptr,SUBSYSTEM_NAME_SIZE);
      }
      k++;
    }
    if (k == 0 && removeDecoration(contxt) == 0)
      goto retry_undecorated;
  }
  else
  {
    for (i=0,k=0; i<nSubSystems && k<dsiz; i++)
    {
      if (strlen(subSystems[i].name) == 0) continue;
      strncpy(ptr+k*dsiz,subSystems[i].name,SUBSYSTEM_NAME_SIZE);
      k++;
    }
  }
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}

int getExportsForFec(int idx,BYTE *buf,int bufsiz,int explen)
{
  int i, n = 0, sizeleft=bufsiz;
  if (bufsiz < explen || buf == NULL) return 0;
  for (i=0; i<SrvTblSize && sizeleft >= explen; i++)
  {
    if (!strnicmp(SrvTbl[i].FecName,FecTbl[idx].fecName,FEC_NAME_SIZE))
    {
      memcpy(&buf[n*explen],SrvTbl[i].ExportName,explen);
      sizeleft -= explen; n++;
    }
  }
  return n;
}
int getAllowedRootContextNames(DTYPE *d)
{
  int i,siz,cpsiz;
  switch (d->dFormat)
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      siz = fmtsizeof(d->dFormat);
      break;
    default: 
      return illegal_format;
  }
  cpsiz = siz < 32 ? siz : 32;
  if (nAllowedContexts < (int)d->dArrayLength) d->dArrayLength = (UINT32)nAllowedContexts;
  for (i=0; i<(int)d->dArrayLength; i++)
  {
    strncpy(&d->data.cptr[i*siz],gAllowedContexts[i].name,cpsiz);
  }
  return 0;
}
int getContextNames(DTYPE *d,char *srv,int importance)
{
  int i,j,k,m,siz,cpsiz,dsiz;
  int chkSrv = FALSE;

  /* parseSearchString() happens in getSearchInfo() */
  if (gUseSearchLists)
  { /* modern style */
    chkSrv = TRUE;
  }
  else
  {
    if (srv != NULL && strlen(srv) > 0) chkSrv = TRUE;
  }
  
  dsiz = d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      siz = fmtsizeof(d->dFormat);
      break;
    case CF_TEXT: siz = 16; dsiz /= 16; break;
    default: return illegal_format;
  }
  cpsiz = siz < 32 ? siz : 32;
  for (i=0,k=0; i<ncontexts && k<dsiz; i++)
  {
    if (chkSrv)
    { /* server name filter given */
      for (m=0; m<SrvTblSize; m++)
      { 
        if (gUseSearchLists)
        {
          ctxSearch.len = 1;
          ctxSearch.lst[0] = contexts[i].name;
          if (!srvInSearchList(m)) continue;
        }
        else
        {
          if ((j=SrvInfoTbl[m].fecIndex) < 0 || j >= FecTblSize) continue;
          if (FecInfoEx[j].importance < importance) continue;
          if (SrvInfoTbl[m].disabled) continue;
          if (strnicmp(SrvTbl[m].EqmContext,contexts[i].name,CONTEXT_NAME_SIZE)) continue;
          if (strnicmp(SrvTbl[m].ExportName,srv,EXPORT_NAME_SIZE)) continue;
        }
        break;
      }
      if (m == SrvTblSize) continue; /* no such entry */
    }
    strncpy(&d->data.cptr[k*siz],contexts[i].name,cpsiz);
    k++;
  }
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}
int getDecoratedContextNames(DTYPE *d)
{
  int i,j,k,m,siz,dsiz;
  char decname[64];
  dsiz = d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      siz = fmtsizeof(d->dFormat);
      break;
    case CF_TEXT: siz = 16; dsiz /= 16; break;
    default: return illegal_format;
  }
  for (j=0,k=0; j<SrvTblSize && k<dsiz; j++)
  {
    if (SrvInfoTbl[j].disabled) continue;
    if (strlen(SrvTbl[j].EqmContext) == 0) continue;
    if ((m=SrvInfoTbl[j].fecIndex) < 0 || m >= FecTblSize) continue;
    if (FecInfoEx[m].disable) continue;
    if (strchr(SrvTbl[j].EqmContext,'.') != NULL) 
    {
      strncpy(decname,SrvTbl[j].EqmContext,CONTEXT_NAME_SIZE);
    }
    else
    {
      if (strlen(SrvTbl[j].SubSystem) == 0) continue;
      sprintf(decname,"%.32s.%.5s",SrvTbl[j].EqmContext,SrvTbl[j].SubSystem);
    }
    for (i=0; i<k; i++)
    {
      if (!strnicmp(decname,&d->data.cptr[i*siz],CONTEXT_NAME_SIZE)) break;
    }
    if (i == k)
    {
      strncpy(&d->data.cptr[k*siz],decname,siz);
      k++;
    }
  }
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}
int getExportTags(DTYPE *d,char *subsys,char *cntxt,int importance)
{
  DUSTRING *ustr = NULL;
  SrvQueryInfo *sqi = NULL;
  char exp[128], eqm[16], *c, *subs = subsys;
  int i,j,k,n=-1,nlen,nleft,siz,dsiz;
  int subschk=0,cntxtchk=0,ipchk=0,isFec=0,isEqm=0;
  UINT32 haddr=0;
  dsiz = d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      siz = fmtsizeof(d->dFormat);
      break;
    case CF_TEXT: siz = 16; dsiz /= 16; break;
    case CF_USTRING: 
      siz = 16; 
      ustr = (DUSTRING *)d->data.vptr;
      break;
    case CF_STRUCT:
      if (strcmp(d->dTag,"SRVQINF")) return invalid_structure_tag;
      sqi = (SrvQueryInfo *)d->data.vptr;
      break;
    default: 
      return illegal_format;
  }
  if (!gUseSearchLists)
  {
    if (cntxt != NULL && strlen(cntxt) > 0) cntxtchk = -1;
    if (!stricmp(cntxt,"DEFAULT")) cntxtchk = 0;
    if (subsys != NULL && !stricmp(subsys,"ALL")) subsys[0] = 0;
    if (subsys != NULL && strlen(subsys) > 0) 
    {
      if ((i=GetPropertyId("ENSEQM",subsys)) >= 0)
      { /* a registered property! -> either a FEC or a Server */
        for (n=0; n<FecTblSize && strnicmp(subsys,FecTbl[n].fecName,FEC_NAME_SIZE); n++);
        if (n < FecTblSize) isFec = -1;
      }
      else if ((haddr=getIPhaddr(subsys)) != 0)
      { /* an ip address query ? TODO: ipv6 */
        ipchk = -1;
      }
      else if (subsys[0] == '(')
      { /* specific equipment module name */
        strncpy(eqm,&subsys[1],EQM_NAME_SIZE);
        eqm[EQM_NAME_SIZE] = 0;
        if ((c=strchr(eqm,')')) != NULL) *c = 0;
        isEqm = -1;
      }
      if (!isFec && !ipchk && !isEqm)
      { /* maybe a subsystem query ? */
        if (!strnicmp(subsys,"UNKNOWN",32)) subschk = -1; else subschk = 1;
        subs = getShortSubSystemName(subsys,FALSE);
      }
    }
  }
  else
  { /* context search coming in extra ? */
    if (cntxt != NULL && strlen(cntxt) > 0)
    {
      for (i=0; i<ctxSearch.len; i++) if (!strnicmp(ctxSearch.lst[i],cntxt,CONTEXT_NAME_SIZE)) break;
      if (i == ctxSearch.len)
      { /* not found ! */
        if (ctxSearch.len < MAX_SEARCH_PARAMS)
        {
          ctxSearch.lst[i] = cntxt;
          ctxSearch.len++;
        }
      }
    }
  }
retry_undecorated:
  for (i=0,k=0; i<SrvTblSize && k<dsiz; i++)
  {
    if (gUseSearchLists)
    {
      if (!srvInSearchList(i)) continue;
      if ((j=SrvInfoTbl[i].fecIndex) < 0 || j >= FecTblSize) continue;
    }
    else
    { // TODO: ipv6
      if (SrvInfoTbl[i].disabled) continue;
      if ((j=SrvInfoTbl[i].fecIndex) < 0 || j >= FecTblSize) continue;
      if (cntxtchk && strnicmp(SrvTbl[i].EqmContext,cntxt,CONTEXT_NAME_SIZE)) continue;
      if (ipchk && memcmp(FecTbl[j].netAdr.ipv4,&haddr,4)) continue;
      if (isFec && n != j) continue;
      if (!isFec && FecInfoEx[j].importance < importance) continue;
      if (isEqm && strnicmp(SrvTbl[i].EqmName,eqm,EQM_NAME_SIZE)) continue;
      if (subschk)
      {
        if (subschk > 0 && strnicmp(SrvTbl[i].SubSystem,subs,SUBSYSTEM_NAME_SIZE)) continue;
        if (subschk < 0 && strlen(SrvTbl[i].SubSystem) > 0) continue;
      } 
    }
    if (ustr != NULL)
    {
      nlen = (int)MIN(strlen(SrvTbl[i].ExportName),32);
      if (nlen > 16) 
      {
        strncpy(exp,SrvTbl[i].ExportName,15);
        exp[15] = '&';
        nleft = nlen - 15;
        strncpy(&exp[16],&SrvTbl[i].ExportName[15],nleft);
      }
      else
      {
        strncpy(exp,SrvTbl[i].ExportName,16);
        nleft = 0;
      }
      strncpy(ustr[k].str,exp,16);
      strncpy(&ustr[k].str[16],FecInfoTbl[j].os,8);
      strncpy(&ustr[k].str[24],SrvTbl[i].SubSystem,8);
      strncpy(&ustr[k].str[32],SrvTbl[i].FecName,16);
      if (nleft > 0) strncpy(&ustr[k].str[48],&exp[16],nleft);
      memcpy(&ustr[k].ival,FecTbl[j].netAdr.ipv4,4);
      ustr[k].tm = FecInfoEx[j].importance;
      ustr[k].f1val = (float)FecTbl[j].portOffset;
    }
    else if (sqi != NULL)
    {
      strncpy(sqi[k].ctx, SrvTbl[i].EqmContext, CONTEXT_NAME_SIZE);
      strncpy(sqi[k].srv, SrvTbl[i].ExportName, EXPORT_NAME_SIZE);
      strncpy(sqi[k].fec, SrvTbl[i].FecName, FEC_NAME_SIZE);
      strncpy(sqi[k].sub, SrvTbl[i].SubSystem, SUBSYSTEM_NAME_SIZE);
      strncpy(sqi[k].os, FecInfoTbl[j].os, FEC_OS_SIZE);
      strncpy(sqi[k].resp, FecInfoTbl[j].resp, FEC_RESP_SIZE);
      strncpy(sqi[k].ver, FecInfoTbl[j].ver, FEC_VERSION_SIZE);
      strncpy(sqi[k].addr, FecTbl[j].strAdr, FEC_OS_SIZE);
      sqi[k].importance = FecInfoEx[j].importance;
      sqi[k].portoffset = FecTbl[j].portOffset;
    }
    else
    {
      if (isFec && !cntxtchk)
      {
        sprintf(exp,"/%.32s/%.32s",SrvTbl[i].EqmContext,SrvTbl[i].ExportName);
      }
      else
      {
        strncpy(exp,SrvTbl[i].ExportName,EXPORT_NAME_SIZE);
      }
      strncpy(&d->data.cptr[k*siz],exp,siz);
    }
    k++;
  }
  if (k == 0 && cntxtchk && removeDecoration(cntxt) == 0)
    goto retry_undecorated;
  if (k < dsiz) d->dArrayLength = k;
  return 0;
}
void trimname(char *nam,int len)
{
  int i, cplen=len;
  char tmp[64],*c = tmp;
  if (len > 62) len = 62;
  strncpy(tmp,nam,len); tmp[len] = 0;
  while (*c == ' ') { c++;  cplen--; }
  while (c[len-1] == ' ') { c[len-1] = 0; len--; cplen--; }
  strncpy(nam,c,cplen);
  /* trap this absurdity ... */
  for (i=0; i<(int)strlen(nam) && i<len; i++) if (nam[i] == '\n') nam[i] = ' ';
}
/* used to check certain names : */
int isallowedstr(const char *str, int len)
{
# define ALLOWEDCHARS "._-:"
  int i, slen = 0;
  if (str == NULL) return FALSE;
  slen = (int)strlen(str);
  if (slen == 0) return FALSE; /* no zero-length names allowed */
  if (!isalnum(str[0])) return FALSE; /* 1st char must be alpha-numeric */
  if (slen > len) slen = len; /* check up to the length of this string */
  for (i = 0; i < slen; i++)
  { /* all character must be alpha-numeric or from a speical allowed set */
    if (!isalnum(str[i]) && strchr(ALLOWEDCHARS,str[i]) == NULL) return FALSE;
  }
  return TRUE;
}
int checkNames(void)
{
# define CHECK_NAME(c,n) if (!isallowedstr(c,n)) return -1;
  CHECK_NAME(gTgtSrv.FecName,FEC_NAME_SIZE);
  strntrm(gTgtSrv.FecName,FEC_NAME_SIZE);
  CHECK_NAME(gTgtSrv.ExportName,EXPORT_NAME_SIZE);
  strntrm(gTgtSrv.ExportName,EXPORT_NAME_SIZE);
  CHECK_NAME(gTgtSrv.EqmContext,CONTEXT_NAME_SIZE);
  strntrm(gTgtSrv.EqmContext,CONTEXT_NAME_SIZE);
  CHECK_NAME(gTgtSrv.EqmName,EQM_NAME_SHORTSIZE);
  CHECK_NAME(gTgtFec.fecName,FEC_NAME_SIZE);
  strntrm(gTgtFec.fecName,FEC_NAME_SIZE);
  return 0;
}
void splitContextAndSubsystem(char *contxt,char *subsys)
{
  int i;
  char *c;
  if (contxt == NULL) return;
  if (subsys == NULL) return; /* no point in doing this .. */
  for (i=0; i<ncontexts && strnicmp(contexts[i].name,contxt,CONTEXT_NAME_SIZE); i++);
  if (i < ncontexts) return;
  if ((c=strchr(contxt,'.')) == NULL) return;
  *c++ = 0;
  strncpy(subsys,c,SUBSYSTEM_NAME_SIZE); subsys[SUBSYSTEM_NAME_SIZE] = 0; /* divvy the 32-char devname up */
}
#define PRP_FECINFO_BASESIZE sizeof(FecInfoStruct) /* 208 bytes */
#define LEGACY_PRP_FECINFO_BASESIZE sizeof(RPCFecExStruct) /* 188 bytes */
#define ADDRPKTSIZE6   (EXPDATASTRUCT_SIZE + FECADDRSTRUCT_SIZE + sizeof(FecInfoStruct))
#define ADDRPKTSIZE4  (EXPDATASTRUCT_SIZE + FECDATASTRUCT_SIZE + sizeof(FecInfoStruct))
#define ADDRPKTSIZE4L (RPCFECSTRUCT_SIZE + RPCDATASTRUCT_SIZE + sizeof(RPCFecExStruct))
int prepNewExportRequest(BYTE *reqBuffer,int datalength)
{
  RPCDataStruct lTgtSrv;
  RPCFecStruct lTgtFec;
  RPCFecExStruct lTgtFecInfo;
  FecDataStruct lTgtFec4;
  int po =0;
  switch (datalength)
  {
    case ADDRPKTSIZE4L:
    {
      UINT32 haddr = 0;
      memcpy(&lTgtSrv,reqBuffer,sizeof(RPCDataStruct));
      memcpy(&lTgtFec,&reqBuffer[sizeof(RPCDataStruct)],sizeof(RPCFecStruct));
      memcpy(&lTgtFecInfo,&reqBuffer[sizeof(RPCDataStruct)+sizeof(RPCFecStruct)],sizeof(RPCFecExStruct));
      memset(&gTgtFec,0,sizeof(RPCFecStruct));
      strncpy(gTgtFec.fecName,lTgtFec.fecName,FEC_NAME_SIZE);
      if ((haddr = *(UINT32 *)lTgtFec.IPh_addr) == 0)
      {
        memcpy(gTgtFec.netAdr.ipx.IpxNet,lTgtFec.Network,16);
      }
      else
      {
        strncpy(gTgtFec.strAdr,lTgtFec.IP,ADDR_SIZE);
        memcpy(gTgtFec.netAdr.ipv4,&haddr,4);
      }
      po = lTgtFec.Port[0] + ((int)lTgtFec.Port[1] >> 8);
      gTgtFec.portOffset = po > 0 ? po : lTgtFec.portOffset;
      gTgtFec.inetProtocol = lTgtFec.Prot;
      memset(&gTgtSrv,0,sizeof(ExpDataStruct));
      strncpy(gTgtSrv.FecName,lTgtSrv.FecName,FEC_NAME_SIZE);
      strncpy(gTgtSrv.ExportName,lTgtSrv.ExportName,EXPORT_NAME_SHORTSIZE);
      strncpy(gTgtSrv.EqmName,lTgtSrv.EqmName,EQM_NAME_SHORTSIZE+1);
      strncpy(gTgtSrv.EqmContext,lTgtSrv.EqmContext,CONTEXT_NAME_SHORTSIZE);
      memset(&gTgtFecInfo,0,sizeof(RPCFecExStruct));
      strncpy(gTgtFecInfo.os,lTgtFecInfo.os,12);
      strncpy(gTgtFecInfo.desc,lTgtFecInfo.desc,64);
      strncpy(gTgtFecInfo.loc,lTgtFecInfo.loc,32);
      strncpy(gTgtFecInfo.hdw,lTgtFecInfo.hdw,32);
      strncpy(gTgtFecInfo.resp,lTgtFecInfo.resp,32);
      strncpy(gTgtSrv.SubSystem,lTgtFecInfo.sub,SUBSYSTEM_NAME_SIZE);
      if (strlen(gTgtSrv.SubSystem) == 0)
      {
        strncpy(gTgtSrv.SubSystem,lTgtSrv.SubSystem,SUBSYSTEM_NAME_SHORTSIZE);
      }
      strcpy(gTgtFecInfo.ver,"3.31");
      break;
    }
    case ADDRPKTSIZE4:
      memcpy(&gTgtSrv,reqBuffer,sizeof(ExpDataStruct));
#     ifdef INTEL
      memcpy(&lTgtFec4,&reqBuffer[sizeof(ExpDataStruct)],sizeof(FecDataStruct));
#     else
      _StructSwap(FDSStructStruct,&reqBuffer[sizeof(ExpDataStruct)],0,NULL,(BYTE *)&lTgtFec4,RTOL);
#     endif
      toFecAddr(&gTgtFec,&lTgtFec4);
      memcpy(&gTgtFecInfo,&reqBuffer[sizeof(ExpDataStruct)+sizeof(FecDataStruct)],sizeof(FecInfoStruct));
      break;
    case ADDRPKTSIZE6:
      memcpy(&gTgtSrv,reqBuffer,sizeof(ExpDataStruct));
#     ifdef INTEL
      memcpy(&gTgtFec,&reqBuffer[sizeof(ExpDataStruct)],sizeof(FecAddrStruct));
#     else
      _StructSwap(FASStructStruct,&reqBuffer[sizeof(ExpDataStruct)],0,NULL,(BYTE *)&gTgtFec,RTOL);
#     endif
      memcpy(&gTgtFecInfo,&reqBuffer[sizeof(ExpDataStruct)+sizeof(FecAddrStruct)],sizeof(FecInfoStruct));
      break;
    default:
      return illegal_data_size;
  }
  return 0;
}
void prepFwdExportRequest(BYTE *reqBuffer,int datalength)
{ /* ready to forward the payload, but I might be a big-endian machine */
  RPCFecStruct lTgtFec4L;
  FecDataStruct lTgtFec4;

  switch (datalength)
  {
    case ADDRPKTSIZE4L:
      memcpy(lTgtFec4L.IP,gTgtFec.strAdr,16);
      memcpy(lTgtFec4L.IPh_addr,gTgtFec.netAdr.ipv4,4);
      memcpy(&reqBuffer[sizeof(RPCDataStruct)],&lTgtFec4L,sizeof(RPCFecStruct)); 
      break;
    case ADDRPKTSIZE4:
      toFecData(&lTgtFec4,&gTgtFec);
#     ifdef INTEL
      memcpy(&reqBuffer[sizeof(ExpDataStruct)],&lTgtFec4,sizeof(FecDataStruct)); 
#     else
      _StructSwap(FDSStructStruct,&reqBuffer[sizeof(ExpDataStruct)],0,NULL,(BYTE *)&lTgtFec4,LTOR);
#     endif
      break;
    case ADDRPKTSIZE6:
#     ifdef INTEL
      memcpy(&reqBuffer[sizeof(ExpDataStruct)],&gTgtFec,sizeof(FecAddrStruct)); 
#     else
      _StructSwap(FASStructStruct,&reqBuffer[sizeof(FecAddrStruct)],0,NULL,(BYTE *)&gTgtFec,LTOR);
#     endif
      break;
  }
}
int prepFecAddrRequest(BYTE *reqBuffer,FecAddrStruct *fec,int datalength)
{
  switch (datalength)
  {
    case ADDRPKTSIZE4L:
    {
      RPCFecStruct *lfec = (RPCFecStruct *)reqBuffer;
      memset(lfec,0,sizeof(RPCFecStruct));
      strncpy(lfec->fecName,fec->fecName,FEC_NAME_SIZE);
      if (fec->inetProtocol == IPX)
      {
        memcpy(lfec->Network,fec->netAdr.ipx.IpxNet,16);
      }
      else
      {
        strncpy(lfec->IP,fec->strAdr,ADDR_SIZE);
        memcpy(lfec->IPh_addr,fec->netAdr.ipv4,4);
      }
      lfec->Port[0] = (BYTE)fec->portOffset;
      lfec->Port[1] = (BYTE)(fec->portOffset >> 8);
      lfec->portOffset = (BYTE)fec->portOffset;
      lfec->Prot = (short)fec->inetProtocol;
      break;
    }
    case ADDRPKTSIZE4:
    {
      FecDataStruct *lfec = (FecDataStruct *)reqBuffer;
      toFecData(lfec,fec);
      break;
    }
    case ADDRPKTSIZE6:
      memcpy(reqBuffer,fec,sizeof(FecAddrStruct));
      break;
    default:
      return illegal_data_size;
  }
  return 0;
}

int prepSrvAddrRequest(BYTE *reqBuffer,ExpDataStruct *srv,FecAddrStruct *fec,int fmt,int datalength)
{
  NAME64 *n64 = (NAME64 *)reqBuffer;
  if (datalength < 8 && fmt != CF_NAME64) return illegal_format;
  switch (datalength)
  {
    case 4:
      strncpy(n64[3].name,srv->EqmName,EQM_NAME_SHORTSIZE);
    case 3:
      strncpy(n64[2].name,fec->fecName,FEC_NAME_SIZE);
    case 2:
      sprintf(n64[1].name,"%d",fec->portOffset);
    case 1:
      strncpy(n64[0].name,fec->strAdr,64);
      break;
    case ADDRPKTSIZE4L:
    {
      RPCFecStruct *lfec = (RPCFecStruct *)(&reqBuffer[sizeof(RPCDataStruct)]);
      RPCDataStruct *lsrv = (RPCDataStruct *)reqBuffer;
      memset(lfec,0,sizeof(RPCFecStruct));
      strncpy(lfec->fecName,fec->fecName,FEC_NAME_SIZE);
      if (fec->inetProtocol == IPX)
      {
        memcpy(lfec->Network,fec->netAdr.ipx.IpxNet,16);
      }
      else
      {
        strncpy(lfec->IP,fec->strAdr,ADDR_SIZE);
        memcpy(lfec->IPh_addr,fec->netAdr.ipv4,4);
      }
      lfec->Port[0] = (BYTE)fec->portOffset;
      lfec->Port[1] = (BYTE)(fec->portOffset >> 8);
      lfec->portOffset = (BYTE)fec->portOffset;
      lfec->Prot = (short)fec->inetProtocol;
      memset(lsrv,0,sizeof(RPCDataStruct));
      strncpy(lsrv->FecName,srv->FecName,FEC_NAME_SIZE);
      strncpy(lsrv->SubSystem,srv->SubSystem,SUBSYSTEM_NAME_SHORTSIZE);
      strncpy(lsrv->ExportName,srv->ExportName,EXPORT_NAME_SHORTSIZE);
      strncpy(lsrv->EqmName,srv->EqmName,EQM_NAME_SHORTSIZE+1);
      strncpy(lsrv->EqmContext,srv->EqmContext,CONTEXT_NAME_SHORTSIZE);
      break;
    }
    case ADDRPKTSIZE4:
    {
      FecDataStruct lfec;
      toFecData(&lfec,fec);
      memcpy(reqBuffer,srv,sizeof(ExpDataStruct));
#     ifdef INTEL
      memcpy(&reqBuffer[sizeof(ExpDataStruct)],&lfec,sizeof(FecDataStruct));
#     else
      _StructSwap(FDSStructStruct,(unsigned char *)&reqBuffer[sizeof(ExpDataStruct)],0,NULL,(unsigned char *)fec,LTOR);
#     endif
      break;
    }
    case ADDRPKTSIZE6:
      memcpy(reqBuffer,srv,sizeof(ExpDataStruct));
#     ifdef INTEL
      memcpy(&reqBuffer[sizeof(ExpDataStruct)],fec,sizeof(FecAddrStruct));
#     else
      _StructSwap(FASStructStruct,(unsigned char *)&reqBuffer[sizeof(ExpDataStruct)],0,NULL,(unsigned char *)fec,LTOR);
#     endif
      break;
    default:
      return illegal_data_size;
  }
  return 0;
}
int prepFecInfoRequest(BYTE *reqBuffer,int fecindex,FecInfoStruct *inf,int sizeInBytes,int pktsize)
{
  char *explst;
  int n = 0,lstSizeInBytes;

  switch (pktsize)
  {
    case ADDRPKTSIZE4L:
    {
      RPCFecExStruct *linf = (RPCFecExStruct *)reqBuffer;
      strncpy(linf->os,inf->os,12);
      strncpy(linf->desc,inf->desc,64);
      strncpy(linf->loc,inf->loc,32);
      strncpy(linf->hdw,inf->hdw,32);
      strncpy(linf->resp,inf->resp,32);
      strncpy(linf->sub,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
      if (sizeInBytes > LEGACY_PRP_FECINFO_BASESIZE) 
      { 
        explst = (char *)&reqBuffer[LEGACY_PRP_FECINFO_BASESIZE];
        lstSizeInBytes = sizeInBytes - LEGACY_PRP_FECINFO_BASESIZE;
        n = getExportsForFec(fecindex,(BYTE *)explst,lstSizeInBytes,EXPORT_NAME_SHORTSIZE);
      }
      break;
    }
    case ADDRPKTSIZE4:
    case ADDRPKTSIZE6:
      memcpy(reqBuffer,inf,PRP_FECINFO_BASESIZE);
      if (sizeInBytes > PRP_FECINFO_BASESIZE) 
      { 
        explst = (char *)&reqBuffer[PRP_FECINFO_BASESIZE];
        lstSizeInBytes = sizeInBytes - PRP_FECINFO_BASESIZE;
        n = getExportsForFec(fecindex,(BYTE *)explst,lstSizeInBytes,EXPORT_NAME_SIZE);
      }
      break;
  }
  return n;
}
int logAndReturn(char *usr,char *prp)
{
  char es[256];
  feclog("refused access to property %.64s for %.16s (not an administrator)",prp,usr);
  sprintf(es,"access denied: user %.16s is not an administrator",usr);
  SetEqmCompletion("ENSEQM",es);
  return access_denied;
}
char last_caller[32];
char last_ctxreq[CONTEXT_NAME_SIZE];
char last_srvreq[EXPORT_NAME_SIZE];
time_t last_reqts = 0;
int last_reqcnt = 0;
BYTE *getNetAdrFromSckAdr(SCKADR *sckadr)
{
  static BYTE bytebuf[16];
  int off;
  memset(bytebuf,0,16);
  if (sckadr->ss_family == AF_INET)
  {
    off = OFFSETIN(struct sockaddr_in,sin_addr);
    memcpy(bytebuf,(BYTE *)sckadr + off,4);
  }
  else
  {
    off = OFFSETIN(struct sockaddr_in6,sin6_addr);
    memcpy(bytebuf,(BYTE *)sckadr + off,16);
  }
  return bytebuf;
}
int enseqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int i,j=0,k,n;
  int cc,chkCtx=FALSE,devnr=0,imp=0,ncallers=1,isLegacyCall=FALSE,isFEC=TRUE,tgtFound,tgtFamily=gAfInet;
  UINT32 dsize,ssize;
  short fmt;
  BYTE *dbuf = dout->data.bptr;
  char *c;
  time_t ttmp;
  static char tgt[72],subs[64],cntxt[64];
  static char usr[USERNAME_SIZE+4],resp[FEC_RESP_SIZE],lvl[64];
  char tgtsrv[64],tgtctx[72];
  ClnInfo clnInfo;
  BYTE inaddr[16], *netAdr;
# define deliverAbsoluteIp4Address (tgtFamily != gAfInet && tgtFamily == AF_INET)

  if ((devnr=GetDeviceNumber("ENSEQM",devName)) >= 0 && devnr < 4) 
  { /* one of : "DEFAULT", "FEC", "EQP", "FEC.EXT" */
    strcpy(tgtctx,"DEFAULT");
    memset(tgtsrv,0,EXPORT_NAME_SIZE);
  }
  else if (strlen(devName) > 0 && devName[0] != '#')
  { /* context given as 'devName' */
    devnr = 0;
    chkCtx = TRUE;
    if ((c=strchr(devName,'_')) != NULL)
    { /* could this be "_TEST" or "_SIM" ? */
      if (isSystemContextDecoration(&c[1]))
      { /* seems to be! replace with a '.' */
        *c = '.';
      }
    }
    c = devName[0] == '/' ? &devName[1] : devName;
    strncpy(tgtctx,c,72);
    if ((c=strchr(tgtctx,'/')) != NULL)
    { /* compound device name : <context>/<server> */
      *c++ = 0;
      strncpy(tgtsrv,c,EXPORT_NAME_SIZE);
    }
  }
  else
  { /* empty device name or begins with a '#4' or higher */
    devnr = 0;
    memset(tgtctx,0,CONTEXT_NAME_SIZE);
    memset(tgtsrv,0,EXPORT_NAME_SIZE);
  }

  fmt = dout->dFormat;
  dsize = dout->dArrayLength;
  
  GetCallerInformation("ENSEQM",&clnInfo,&ncallers); 
  netAdr = getNetAdrFromSckAdr(&clnInfo.addr); // clnInfo.addr is a SCKADR -> includes port and family
  if (tgtFamily == AF_INET6)
  { /* check whether the caller wants to speak IPv4 and we're an IPv6 ENS */
    char adr[64];
    inetStrAdr(&clnInfo.addr,adr,FALSE);
    if (isIpv4on6(adr)) tgtFamily = AF_INET;
  }
  strncpy(usr,clnInfo.userName,16); strupr(usr);
  if (!stricmp(devProperty,"ADDRESSOF"))
  {
    char *pctx, *psrv, *pdev;
    devnr = SRV_DEV_TAG;
    if (devName[0] != '/') sprintf(tgt,"/%.70s",devName);
    else strncpy(tgt,devName,72);
    if ((cc=tokenizeFullName(tgt,&pctx,&psrv,&pdev)) != 0) return cc;
    strncpy(tgtctx,pctx,CONTEXT_NAME_SIZE);
    strncpy(tgtsrv,psrv,EXPORT_NAME_SIZE);
    goto handle_addr_query;
  }
  if (!stricmp(devProperty,"SIZES"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (dsize != 2) return out_of_range;
    switch (fmt)
    {
      case CF_SHORT:
        dout->data.sptr[0] = (short)SrvTblSize;
        dout->data.sptr[1] = (short)FecTblSize;
        break;
      case CF_LONG:
        dout->data.lptr[0] = SrvTblSize;
        dout->data.lptr[1] = FecTblSize;
        break;
      default: return illegal_format;
    }
    return 0;
  }
  if (!stricmp(devProperty,"WHOAMI"))
  {
    NAME16 inpt[2];
    int po = -1;
    tgt[0] = 0;
    if (din->dArrayLength > 0)
    {
      if (access & CA_WRITE) return illegal_read_write;
      if ((cc=GetValuesAsStringEx(din,inpt[0].name,CF_NAME16,2,0)) != 0) return cc;
      strncpy(tgt,inpt[0].name,EQM_NAME_SIZE);
      if (din->dArrayLength > 1) po = atoi(inpt[1].name);
    }
    for (i=0; i<FecTblSize; i++)
    { // TODO: ipv6
      if (memcmp(netAdr,&FecTbl[i].netAdr,SCKADR_LEN)) continue; // netAdr is 16 bytes (raw IP address: no port)
      if (po >= 0 && po != FecTbl[i].portOffset) continue;
      for (j=0; j<SrvTblSize; j++)
      {
        if (strnicmp(SrvTbl[j].FecName,FecTbl[i].fecName,FEC_NAME_SIZE)) continue;
        if (strlen(tgt) > 0 && strnicmp(tgt,SrvTbl[j].EqmName,EQM_NAME_SIZE)) continue;
        break;
      }
      if (j < SrvTblSize) break;
    }
    if (i == FecTblSize) return not_exported;
    if (j == SrvTblSize) return not_exported;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        ssize = fmtsizeof(dout->dFormat);
        break;
      default:
        return illegal_format;
    }
    switch (dout->dArrayLength)
    {
      case 5:
        sprintf(dout->data.cptr+4*ssize,"%d",FecTbl[i].portOffset);
      case 4:
        strncpy(dout->data.cptr+3*ssize,SrvTbl[j].EqmContext,ssize);
      case 3:
        strncpy(dout->data.cptr+2*ssize,SrvTbl[j].FecName,ssize);
      case 2:
        strncpy(dout->data.cptr+ssize,SrvTbl[j].EqmName,ssize);
      case 1:
        strncpy(dout->data.cptr,SrvTbl[j].ExportName,ssize);
        break;
      default:
        return dimension_error;
    }
    return 0;
  }

  if (!stricmp(devProperty,"CHECKFLAG"))
  {
    if (din->dArrayLength > 0)
    {
      if (!(access & CA_WRITE)) return illegal_read_write;
      if ((cc=GetValuesAsLong(din,(SINT32 *)&checkExportFlag,1)) != 0) return cc;
    }
    if (dout->dArrayLength > 0)
    {
      if ((cc=PutValuesFromLong(dout,(SINT32 *)&checkExportFlag,1)) != 0) return cc;
    }
    return 0;
  }
  if (!stricmp(devProperty,"IMPORTANCE"))
  {
    if (din == NULL || din->dArrayLength == 0) return argument_list_error;
    switch (din->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        imp = 0;
        ssize = fmtsizeof(din->dFormat);
        strncpy(tgt,din->data.cptr,ssize);
        if (din->dArrayLength > 1)
        {
          if (!(access & CA_WRITE)) return illegal_read_write;
          strncpy(lvl,din->data.cptr+ssize,ssize);
          imp = getImportanctLevel(lvl);
        }
        break;
      default:
        return illegal_format;
    }
    for (i=0; i<FecTblSize; i++)
    {
      if (FecInfoEx[i].disable) continue;
      if (!strnicmp(tgt,FecTbl[i].fecName,FEC_NAME_SIZE)) break;
    }
    if (i == FecTblSize) return no_such_computer;
    if (din->dArrayLength > 1)
    {
      FecInfoEx[i].importance = (short)imp;
      dumpFecFileFlag = 1;
    }
    if (dout->dArrayLength > 0)
    {
      imp = FecInfoEx[i].importance;
      switch (dout->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          ssize = fmtsizeof(dout->dFormat);
          strncpy(dout->data.cptr,gImportanceLevel[imp],ssize);
          return 0;
        case CF_TEXT:
          ssize = dout->dArrayLength;
          strncpy(dout->data.cptr,gImportanceLevel[imp],ssize);
          return 0;
       default:
          return putValuesFromLong(dout,(SINT32 *)&FecInfoEx[i].importance,1);
      }
    }
    return 0;
  }
  if (!stricmp(devProperty,"REBOOTCOUNT"))
  {
    if (din->dArrayLength == 0) return illegal_read_write;
    if ((cc=GetValuesAsStringEx(din,tgt,CF_TEXT,EXPORT_NAME_SIZE,0)) != 0) return cc;
    if (strlen(tgt) == 0) return illegal_read_write;
    if ((n=getRebootCount(tgtctx,tgt)) < 0) return device_not_connected; 
    return PutValuesFromLong(dout,(SINT32 *)&n,1);
  }
  if (!stricmp(devProperty,"REBOOTCOUNT.LIST"))
  {
    if (dout->dFormat != CF_NAME32I) return illegal_format;
    n = dout->dArrayLength;
    if ((n=getRebootCountAll(tgtctx,(NAME32I *)dout->data.vptr,n)) < 0) return -n; 
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"LASTACTIVE"))
  {
    char tstr[64];
    if (din->dArrayLength == 0) return illegal_read_write;
    if ((cc=GetValuesAsStringEx(din,tgt,CF_TEXT,EXPORT_NAME_SIZE,0)) != 0) return cc;
    if (strlen(tgt) == 0) return illegal_read_write;
    if ((ttmp=getServerLastActive(tgtctx,tgt)) < 0) return device_not_connected;
    switch (dout->dFormat)
    {
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
      case CF_TEXT:
        strncpy(tstr,ctime(&ttmp),24);
        return PutValuesFromString(dout,tstr,CF_TEXT,24,0);
      default:
        return PutValuesFromLong(dout,(SINT32 *)&n,1);
    }
  }
  if (!stricmp(devProperty,"ALLOWEDDEADTIME"))
  {
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr)) return logAndReturn(usr,devProperty);
      if ((cc=GetValuesAsLong(din,(SINT32 *)&i,1)) != 0) return cc;
      if (i < ONEHOUR) return out_of_range;
      gAllowedDeadTime = i;
      asyncCmdForward(tgtctx,devName,devProperty,dout,din,access,(BYTE *)netAdr);
    }
    if (dout->dArrayLength > 0)
    {
      if ((cc=PutValuesFromLong(dout,(SINT32 *)&gAllowedDeadTime,1)) != 0) return cc;
    }
    return 0;
  }
  if (!stricmp(devProperty,"REMOVENOTRUNNING"))
  {
    if (access & CA_WRITE)
    {
      if (!isAdministrator(usr)) return logAndReturn(usr,devProperty);
      srvlog(CMDSFILE,devProperty,"(context %.32s) called by %.16s",tgtctx,usr);
      removeNotRunning(tgtctx);
      asyncCmdForward(tgtctx,devName,devProperty,dout,din,access,(BYTE *)netAdr);
      return 0;
    }
    return illegal_read_write;
  }
  if (!stricmp(devProperty,"REMOVE"))
  {
    if (din == NULL || din->dArrayLength == 0) return illegal_read_write;
    if ((cc=GetValuesAsStringEx(din,tgt,CF_TEXT,EXPORT_NAME_SIZE,0)) != 0) return cc;
    switch (devnr)
    {
      case FEC_DEV_TAG: /* FEC Name */
        if (strnicmp(tgt,"GENS",4) == 0 || 
            strnicmp(tgt,"ENS",3) == 0 ||
            stricmp(tgt,"FECSINK") == 0)
        { 
          nslog("enseqm: remove fec %.16s by %.16s refused !",tgt,usr);
          return not_allowed;
        }
        srvlog(CMDSFILE,devProperty,"remove fec %.16s by %.16s",tgt,usr);
        sprintf(logbuf,"enseqm: remove fec %.16s by %.16s",tgt,usr);
        for (i=0; i<FecTblSize; i++)
        {
          if (FecInfoEx[i].disable) continue;
          if (!strnicmp(tgt,FecTbl[i].fecName,FEC_NAME_SIZE)) /* found it */
          {
            strncpy(resp,FecInfoTbl[i].resp,32); strupr(resp);
            if (!isAdministrator(usr) && !isENS(usr) && 
                 strlen(resp) && strstr(resp,usr) == NULL)
            {
              strcat(logbuf," -> access denied");
              nslog(logbuf);
              return logAndReturn(usr,devProperty);
            }
            FecInfoEx[i].disable = -1;
            for (n=0; n<SrvTblSize; n++) 
              if (SrvInfoTbl[n].fecIndex == i) SrvInfoTbl[n].disabled = TRUE;
            dumpFecFileFlag = 1;
            dumpEqpFileFlag = 1;
            nslog(logbuf);
            if (!isGENS(usr)) asyncCmdForward(tgtctx,devName,devProperty,dout,din,access,(BYTE *)netAdr);
            return 0;
          }
        }
        strcat(logbuf," -> no such computer");
        nslog(logbuf);
        return no_such_computer;
      case DEFAULT_DEV_TAG:
        if (strlen(tgtctx) == 0) break;
      case SRV_DEV_TAG: /* EQP Name */
        if (stricmp(tgt,"GENS") == 0 || stricmp(tgt,"ENS") == 0)
        { 
          nslog("enseqm: remove export %.32s by %.16s refused !",tgt,usr);
          return not_allowed;
        }
        if (strlen(tgtctx) == 0) strcpy(tgtctx,"DEFAULT");
        srvlog(CMDSFILE,devProperty,"remove server %.32s in context %.32s by %.16s",tgt,tgtctx,usr);
        sprintf(logbuf,"ENS: remove server %.32s in context %.32s by %.16s",tgt,tgtctx,usr);
        tgtFound = FALSE;
retry_remove_undecorated:
        for (i=0; i<SrvTblSize; i++)
        {
          if (SrvInfoTbl[i].disabled) continue;
          if (chkCtx && strnicmp(tgtctx,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
          if (!strnicmp(tgt,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) /* found it */
          {
            n = SrvInfoTbl[i].fecIndex;
            if (n < 0 || n >= FecTblSize) 
            { /* really sick ! -> allow anyone to remove this */
              strncpy(resp,usr,16);
            }
            else
            {
              strncpy(resp,FecInfoTbl[n].resp,FEC_RESP_SIZE); strupr(resp);
            }
            if (!isAdministrator(usr) && !isENS(usr) && 
                strlen(resp) && strstr(resp,usr) == NULL) 
            {
              strcat(logbuf," -> access denied");
              nslog(logbuf);
              return logAndReturn(usr,devProperty);
            }
            SrvInfoTbl[i].disabled = TRUE;
            dumpEqpFileFlag = 1;
            nslog(logbuf);
            if (!isGENS(usr)) asyncCmdForward(SrvTbl[i].EqmContext,devName,devProperty,dout,din,access,(BYTE *)netAdr);
            if (chkCtx) return 0;
            nslog("%s -> context %.32s found and removed",logbuf,SrvTbl[i].EqmContext);
            tgtFound = TRUE;
          }
        }
        if (!tgtFound) 
        { /* did not find a match */
          if (chkCtx && removeDecoration(tgtctx) == 0)
            goto retry_remove_undecorated;
        }
        if (tgtFound) return 0;
        strcat(logbuf," -> non existent server");
        nslog(logbuf);
        return non_existent_server;
      default: 
        return illegal_equipment_number;
    }
    return illegal_context;
  }
  if (!stricmp(devProperty,"EXPORT.ALIAS"))
  {
    char *aptr, *tptr;
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isAdministrator(usr) && !isENS(usr)) return logAndReturn(usr,devProperty);
    if (din->dArrayLength != 2) return dimension_error;
    switch (din->dFormat)
    {
      case CF_NAME32:
      case CF_NAME64:
        aptr = din->data.cptr;
        tptr = din->data.cptr + fmtsizeof(din->dFormat);
        break;
      case CF_STRING:
        aptr = din->data.strptr[0];
        tptr = din->data.strptr[1];
        break;
      default:
        return illegal_format;
    }
    srvlog(CMDSFILE,devProperty,"add alias %.32s for %.32s by %.16s",aptr,tptr,usr);
    return addExportAlias(tptr,aptr);
  }
  if (!stricmp(devProperty,"DEVICE.SEQUENCE"))
  {
    int nlen = 0;
    if ((access & CA_WRITE)) return illegal_read_write;
    if (strlen(devName) == 0) return illegal_device;
    switch (dout->dFormat)
    {
      case CF_NAME64:
      case CF_NAME32:
      case CF_NAME16:
        nlen = fmtsizeof(dout->dFormat);
        break;
      default:
        return illegal_format;
    }
    for (i=0; i<(int)dout->dArrayLength; i++)
    {
      snprintf(dout->data.cptr+i*nlen,nlen,"%s%d",devName,i+1);
    }
    return 0;
  }
  if (!stricmp(devProperty,"DEVICE.ECHO"))
  {
    if ((access & CA_WRITE)) return illegal_read_write;
    if (din->dArrayLength == 0) return dimension_error;
    if (!isSimpleStringFormat(din->dFormat)) return illegal_format;
    return PutValuesFromString(dout,din->data.cptr,din->dFormat,din->dArrayLength,0);
  }
  if (!stricmp(devProperty,"EXPORTALIASES"))
  {
    if ((access & CA_WRITE)) return illegal_read_write;
    if (dout->dFormat != CF_NAME64) return illegal_format;
    return getSrvAliases((NAME64 *)dout->data.vptr,(int *)&dout->dArrayLength);
  }
  if (!stricmp(devProperty,"EXPORTALIASLIST"))
  {
    char *cptr, *sptr;
    if ((access & CA_WRITE)) return illegal_read_write;
    if (din->dArrayLength == 2)
    {
      switch (din->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          cptr = din->data.cptr;
          sptr = din->data.cptr + fmtsizeof(din->dFormat);
          break;
        case CF_STRING:
          cptr = din->data.strptr[0];
          sptr = din->data.strptr[1];
          break;
        default:
          return illegal_format;
      }
    }
    else
    {
      cptr = devName;
      if (cptr[0] == '/') cptr++;
      if ((sptr=strchr(cptr,'/')) == NULL) return invalid_parameter;
      *sptr++ = 0;
    }
    if (dout->dFormat != CF_NAME64) return illegal_format;
    return getSrvAliasList(cptr,sptr,(NAME64 *)dout->data.vptr,(int *)&dout->dArrayLength);
  }
  if (!stricmp(devProperty,"FECALIASES"))
  {
    if ((access & CA_WRITE)) return illegal_read_write;
    if (dout->dFormat != CF_NAME64) return illegal_format;
    return getFecAliases((NAME64 *)dout->data.vptr,(int *)&dout->dArrayLength);
  }
  if (!stricmp(devProperty,"FECALIASLIST"))
  {
    char *tptr;
    if ((access & CA_WRITE)) return illegal_read_write;
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_CHAR:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          tptr = din->data.cptr;
          break;
        case CF_STRING:
          tptr = din->data.strptr[0];
          break;
        default:
          return illegal_format;
      }
    }
    else
    {
      tptr = devName;
    }
    if (dout->dFormat != CF_NAME16) return illegal_format;
    return getFecAliasList(tptr,(NAME16 *)dout->data.vptr,(int *)&dout->dArrayLength);
  }
  if (!stricmp(devProperty,"EXPORT"))
  {
    int isadm = isAdministrator(usr);
    if (debug) dbglog("ENS: incoming export ...");
    if (checkExportFlag) return operation_busy;
    if (din->dFormat != CF_STRUCT) return illegal_format;
    if (!(access & CA_WRITE)) return illegal_read_write;
    /* need to distinguish between pre- and post- release 4.0 */
    dsize = din->dArrayLength;
    if (dsize == 1 && strlen(din->dTag) > 0)
    {
      dsize *= GetStructSize(din->dTag);
    }
    if ((cc=prepNewExportRequest(din->data.bptr,dsize)) != 0) return cc;
    if (strstr(gTgtFecInfo.desc,FAILOVER_MASTER_DESC) != NULL) forceExportAddr = TRUE;
    if (checkNames()) 
    {
      if (!strcmp(gTgtSrv.FecName, "__LCLRPTRFEC__")) return 0; /* ignore a (legacy) local repeater */
      feclogEx(TINE_LOGLEVEL_ERROR,"ENS: invalid name in EXPORT /%.32s/%.32s = %.6s@%.16s (%.16s)",
        gTgtSrv.EqmContext,gTgtSrv.ExportName,gTgtSrv.EqmName,gTgtSrv.FecName,gTgtFec.fecName);
      return invalid_name;
    }
    checkSubSytemEntry(&gTgtFecInfo,&gTgtSrv);
    if (!checkContextEntry(&gTgtFecInfo,&gTgtSrv)) return illegal_context;
    checkCSVFormat(csvFecInfo,5);
    if (!strnicmp(gTgtSrv.FecName,"ENS",3)) return 0; /* don't check yourself! */
    bzero(&gTgtFecInfoEx,sizeof(FecInfoStructEx));
    if (!isENS(usr) && !isadm)
    { /* must be a FEC ! */
      if (strnicmp(gTgtSrv.FecName,usr,USERNAME_SIZE) != 0) return logAndReturn(usr,devProperty);
      inet_ntop(gAfInet,netAdr,gTgtFec.strAdr,40);
      if (tgtFamily == AF_INET6) strncpy(gTgtFecInfoEx.ipv6,gTgtFec.strAdr,40);
      memcpy(gTgtFecInfoEx.mac,gTgtFec.lclAdr.mac,6);
      memcpy(inaddr,gTgtFec.netAdr.ipv4,4);
      inet_ntop(AF_INET,inaddr,gTgtFecInfoEx.ipv4,16);
      memset(&gTgtFec.netAdr,0,16);
      memcpy(&gTgtFec.netAdr,netAdr,gInAdrLen);
      /* prepare data input for the forward list */
      prepFwdExportRequest(din->data.bptr,dsize);
      isFEC = TRUE;
    }
    else
    { /* either a forward or an addition via an admin */
      if (isadm) gTgtFec.addrFlags |= ADDRFLAG_MCAST_MODERN; /* bug in ENS admin ? */
      memcpy(gTgtFecInfoEx.mac,gTgtFec.lclAdr.mac,6);
      memcpy(inaddr,gTgtFec.netAdr.ipv4,4);
      inet_ntop(AF_INET,inaddr,gTgtFecInfoEx.ipv4,16);
      strAdr2netAdr(&gTgtFec);
      if (gTgtFec.inetProtocol == UDP6) strncpy(gTgtFecInfoEx.ipv6,gTgtFec.strAdr,40);
      isFEC = FALSE;
    }
    /* do this to circumvent alignment problems : */
    if (!zerocmp(&gTgtFec.netAdr,gInAdrLen))
    { /* don't let this get through ! */
      feclogEx(TINE_LOGLEVEL_ERROR,"ENS: invalid ip address from %.32s",gTgtSrv.ExportName);
      return invalid_data;
    }
    if (!strnicmp(gTgtSrv.FecName,"GENS",4))
    { /* this address must be consistent */
      char ip1[64],ip2[64];
      if (zerocmp(&ens_iaddr,gInAdrLen) && memcmp(&ens_iaddr,&gTgtFec.netAdr,gInAdrLen))
      { /* I'm checking my ENS/GENS entries and this doesn't match ! */
        if (relaxGensAddressCheck)
        { /* overcome some weird IP sleight-of-hand */
          memcpy(&gTgtFec.netAdr,&ens_iaddr,gInAdrLen);
          inet_ntop(gAfInet,&ens_iaddr,gTgtFec.strAdr,40);
        }
        else
        {
          inet_ntop(gAfInet,&gTgtFec.netAdr,ip1,40);
          inet_ntop(gAfInet,&ens_iaddr,ip2,40);
          feclogEx(TINE_LOGLEVEL_ERROR,"ENS: GENS EXPORT address mismatch %.40s should be %.40s",ip1,ip2);
          return invalid_data;
        }
      }
    }
    checkLocationEntry(&gTgtFecInfo,&gTgtFec);
    if (debug) dbglog("ENS: fec %.16s context %.32s export %.32s",gTgtFec.fecName,gTgtSrv.EqmContext,gTgtSrv.ExportName);
    if (strlen(gTgtSrv.EqmContext) == 0 && strlen(tgtctx) > 0)
    {
      strncpy(gTgtSrv.EqmContext,tgtctx,CONTEXT_NAME_SIZE);
      if (debug) dbglog("ENS: apply context from device name");
      feclogEx(TINE_LOGLEVEL_WARN,"ENS: apply context %.32s from device name",tgtctx);
    }
    if (strlen(gTgtSrv.ExportName) &&
        strlen(gTgtSrv.EqmName)  &&
        strlen(gTgtSrv.FecName) &&
        strlen(gTgtFec.fecName)) 
    {
      if (strlen(gTgtSrv.EqmContext) < 3 || !isalpha(gTgtSrv.EqmContext[0])) /* not allowed */
        assignContext(gTgtSrv.ExportName,gTgtSrv.EqmContext);
      if (!stricmp(gTgtSrv.EqmContext,"DEFAULT") || !stricmp(gTgtSrv.EqmContext,"CONTEXT")) 
      { /* somebody claims context DEFAULT or context CONTEXT! */
        gTgtSrv.EqmContext[0] = 0; /* remove it ! */
        feclogEx(TINE_LOGLEVEL_WARN,"ENS: erase explicit DEFAULT context");
      }
      if (!strnicmp(gTgtSrv.FecName,"ENS",3)) return 0; /* don't check yourself! */
      trimname(gTgtSrv.ExportName,EXPORT_NAME_SIZE);
      trimname(gTgtSrv.EqmName,EQM_NAME_SHORTSIZE);
      trimname(gTgtSrv.FecName,FEC_NAME_SIZE);
      trimname(gTgtFec.fecName,FEC_NAME_SIZE);
      checkExportFlag = check_tag;
      lastSrvTblIdx = lastFECindex = fecTargetId = -1; 
      gCanSendAddrInUse = isFEC ? TRUE : FALSE;
      if (forceExportAddr) gCanSendAddrInUse = FALSE;
    }
    if (!isGENS(usr)) asyncCmdForward(gTgtSrv.EqmContext,devName,devProperty,dout,din,access,(BYTE *)netAdr);
    return 0;
  }

  /* below are READ-ONLY operations ... */

  cntxt[0] = 0; subs[0] = 0;
  if (chkCtx) 
  { /* the next several queries allow a subsystem decorated context */
    strncpy(cntxt,tgtctx,CONTEXT_NAME_SIZE);
    splitContextAndSubsystem(cntxt,subs);
  }
  if (!stricmp(devProperty,"NFECS"))
  {
    if ((cc=getSearchInfo(din,subs,&imp)) != 0) return cc;
    n = getNumKnownFecs(cntxt,subs,imp);
    return putValuesFromLong(dout,(SINT32 *)&n,1);
  }
  if (!stricmp(devProperty,"FECS"))
  {
    if ((cc=getSearchInfo(din,subs,&imp)) != 0) return cc;
    return getKnownFecs(dout,cntxt,subs,imp);
  }
  if (!stricmp(devProperty,"SuggestFecName"))
  {
    char *sn;
    int fnl = 16;
    if (din->dArrayLength == 0) return argument_list_error;
    switch (din->dFormat)
    {
      case CF_NAME32I:
        sn = ((NAME32I *)din->data.vptr)[0].name;
        fnl = ((NAME32I *)din->data.vptr)[0].ival;
        break;
      case CF_NAME64I:
        sn = ((NAME64I *)din->data.vptr)[0].name;
        fnl = ((NAME64I *)din->data.vptr)[0].ival;
        break;
      default:
        if (!isSimpleStringFormat(din->dFormat)) return illegal_format;
        sn = din->data.cptr;
        break;
    }
    if (*sn == 0) return argument_list_error;
    /* does server name already exist ? */
    if ((i=serverNameExists(cntxt,sn)) > 0)
    {
      i--;
      if ((cc=PutValuesFromString(dout,SrvTbl[i].FecName,CF_TEXT,FEC_NAME_SIZE,0)) != 0)
        return cc;
      return name_already_exists + CE_SENDDATA;
    }
    /* suggest name */
    c = suggestFecName(cntxt,sn,fnl);
    if (c == NULL || strlen(c) == 0) return name_already_exists;
    return PutValuesFromString(dout,c,CF_TEXT,FEC_NAME_SIZE,0);
  }
  if (!stricmp(devProperty,"NCONTEXTS"))
  {
    return putValuesFromLong(dout,(SINT32 *)&ncontexts,1);
  }
  if (!stricmp(devProperty,"CONTEXTS"))
  {
    if ((cc=getSearchInfo(din,subs,&imp)) != 0) return cc;
    c = din->dArrayLength > 0 ? subs : NULL;
    return getContextNames(dout,c,imp);
  }
  if (!stricmp(devProperty,"CONTEXTS.DECORATED"))
  {
    return getDecoratedContextNames(dout);
  }
  if (!stricmp(devProperty,"CONTEXT.ROOTS.ALLOWED"))
  {
    return getAllowedRootContextNames(dout);
  }
  if (!stricmp(devProperty,"ADD.ALLOWED.ROOT"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isAdministrator(usr)) return logAndReturn(usr,devProperty);
    if (din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsStringEx(din,cntxt,CF_TEXT,32,0)) != 0) return cc;
    }
    else if (chkCtx)
    {
      if ((c=strchr(tgtctx,'.')) != NULL) *c = 0;
      strncpy(cntxt,tgtctx,CONTEXT_NAME_SIZE);
    }
    else
    {
      return invalid_data;
    }
    if ((cc=addAllowedContext(cntxt)) != 0) return cc;
    if ((cc=writeAllowedContexts(ROOTCTX_FNAME)) != 0) return cc;
    return 0;
  }
  if (!stricmp(devProperty,"DEL.ALLOWED.ROOT"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isAdministrator(usr)) return logAndReturn(usr,devProperty);
    if (din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsStringEx(din,cntxt,CF_TEXT,32,0)) != 0) return cc;
    }
    else if (chkCtx)
    {
      if ((c=strchr(tgtctx,'.')) != NULL) *c = 0;
      strncpy(cntxt,tgtctx,CONTEXT_NAME_SIZE);
    }
    else
    {
      return invalid_data;
    }
    if ((cc=rmvAllowedContext(cntxt)) != 0) return cc;
    if ((cc=writeAllowedContexts(ROOTCTX_FNAME)) != 0) return cc;
    return 0;
  }
  if (!stricmp(devProperty,"NSERVERS") || !stricmp(devProperty,"NTAGS"))
  {
    if ((cc=getSearchInfo(din,subs,&imp)) != 0) return cc;
    n = getNumExportTags(subs,cntxt,imp);
    return putValuesFromLong(dout,(SINT32 *)&n,1);
  }
  if (!stricmp(devProperty,"SERVERS") || !stricmp(devProperty,"TAGS"))
  {
    if ((cc=getSearchInfo(din,subs,&imp)) != 0) return cc;
    return getExportTags(dout,subs,cntxt,imp);
  }
  if (!stricmp(devProperty,"NSUBSYSTEMS"))
  {
    return putValuesFromLong(dout,(SINT32 *)&nSubSystems,1);
  }
  if (!stricmp(devProperty,"SUBSYSTEMS"))
  {
    return getKnownSubsystems(dout,cntxt);
  }
  if (!stricmp(devProperty,"SUBSYSTEMS.ALLOWED"))
  {
    return getAllowedSubsystemNames(dout);
  }
  if (!stricmp(devProperty,"SUBSYSTEMS.DESCRIPTIONS"))
  {
    char *ptgt = NULL;
    if (din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsStringEx(din,tgt,CF_TEXT,16,0)) != 0) return cc;
      ptgt = tgt;
    }
    return getAllowedSubsystemDescriptions(dout,ptgt);
  }
  if (!stricmp(devProperty,"SUBSYSTEM"))
  {
    if (din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsStringEx(din,tgtsrv,CF_TEXT,32,0)) != 0) return cc;
    }
    return getSubsystemForServer(dout,cntxt,tgtsrv);
  }

  /* below: address request queries (FECs and SRVs are properties ...) */
  strncpy(tgtsrv,devProperty,PROPERTY_NAME_SIZE);
handle_addr_query: // TODO: ipv6 : tgtFamily = AF_INET (do I have both types?) 
  switch (devnr)
  {
    case FEC_DEV_TAG: /* FEC Name query */
      if (tineDebug) dbglog("ENS: addr request for fec %.64s",tgtsrv);
      if (fmt != CF_STRUCT) return illegal_format;
      switch (dsize)
      {
        case sizeof(RPCFecStruct): dsize = ADDRPKTSIZE4L; break;
        case sizeof(FecDataStruct): dsize = ADDRPKTSIZE4; break;
        case sizeof(FecAddrStruct): dsize = ADDRPKTSIZE6; break;
        default : return illegal_data_size;
      }
      for (i=0; i<FecTblSize; i++)
      {
        if (FecInfoEx[i].disable) continue;
        if (!strnicmp(tgtsrv,FecTbl[i].fecName,FEC_NAME_SIZE)) 
        { /* found it */
          FecAddrStruct fec = FecTbl[i];
          if (deliverAbsoluteIp4Address)
          {
            BYTE b4adr[16];
            inet_pton(AF_INET,FecInfoEx[i].ipv4,b4adr);
            if (*(UINT32 *)b4adr != 0) 
            {
              memset(fec.strAdr,0,ADDR_SIZE_IPV6);
              strncpy(fec.strAdr,FecInfoEx[i].ipv4,16);
            }
            toIpv4Addr(fec.strAdr,fec.strAdr);
            strAdr2netAdr(&fec);
          }
          return prepFecAddrRequest(dbuf,&fec,dsize);
        }
      }
      return no_such_computer;
    case DEFAULT_DEV_TAG:
      if (strlen(tgtctx) == 0) 
      {
        if (tineDebug) dbglog("ENS: addr request %.64s %s not well formed",tgtsrv,devName);
        break;
      }
    case SRV_DEV_TAG: /* Export Server Name query */
      if (tineDebug) dbglog("ENS: addr request for srv %.64s in context %.32s",tgtsrv,tgtctx);
      switch (fmt)
      {
        case CF_STRUCT:
          switch (dsize)
          {
            case sizeof(RPCFecStruct)+sizeof(RPCDataStruct): dsize = ADDRPKTSIZE4L; break;
            case sizeof(ExpDataStruct)+sizeof(FecDataStruct): dsize = ADDRPKTSIZE4; break;
            case sizeof(ExpDataStruct)+sizeof(FecAddrStruct): dsize = ADDRPKTSIZE6; break;
            default : return illegal_data_size;
          }
          break;
        case CF_NAME64:
          dsize = dout->dArrayLength;
          break;
        default:
          return illegal_format;
      }
retry_undecorated:
      for (n=0; n<SrvTblSize; n++)
      {
        if (SrvInfoTbl[n].disabled) continue;
        if (chkCtx && strnicmp(tgtctx,SrvTbl[n].EqmContext,CONTEXT_NAME_SIZE)) continue;
        if (!strnicmp(tgtsrv,SrvTbl[n].ExportName,EXPORT_NAME_SIZE)) break;
      }
      if (n == SrvTblSize) 
      { /* did not find a match */
        if (chkCtx && removeDecoration(tgtctx) == 0)
          goto retry_undecorated;
        if (!chkCtx) strcpy(tgtctx,"DEFAULT");
        ttmp = time(NULL);
        if (!stricmp(tgtsrv,"CAS") || !stricmp(tgtsrv,"STATE") || !stricmp(tgtsrv,"CYCLER"))
        { /* these are allowed to return non_existent multiple times */
          return non_existent;
        }
        if (strnicmp(last_caller,usr,16) == 0 &&
            strnicmp(last_ctxreq,tgtctx,CONTEXT_NAME_SIZE) == 0 &&
            strnicmp(last_srvreq,tgtsrv,EXPORT_NAME_SIZE) == 0 &&
            ttmp == last_reqts)
        { /* same caller wanting the same non-existent address */
          if (++last_reqcnt > 3)
          { /* put a stop to this ! */
            char ipadr[64];
            inet_ntop(gAfInet,netAdr,ipadr,40);
            nslog("third request for non-existent /%.32s/%.32s by %.16s @ %.40s within 1 second!",
              tgtctx,tgtsrv,usr,ipadr);
            if (!checkExportFlag) sinkSrvAddr(tgtctx,tgtsrv);
          }
        }
        else
        {
          strncpy(last_caller,usr,16);
          strncpy(last_ctxreq,tgtctx,CONTEXT_NAME_SIZE);
          strncpy(last_srvreq,tgtsrv,EXPORT_NAME_SIZE);
          last_reqts = ttmp;
          last_reqcnt = 1;
        }
        return non_existent;
      }
      if (!chkCtx)
      { /* no context given, is it ambiguous ? */
        for (k=n+1; k<SrvTblSize; k++)
        { /* look for others from where we left off*/
          if (SrvInfoTbl[k].disabled) continue;
          if (!strnicmp(tgtsrv,SrvTbl[k].ExportName,EXPORT_NAME_SIZE)) 
          { /* same export names!  do they point to different FECs ? */
            if (strnicmp(SrvTbl[n].FecName,SrvTbl[k].FecName,FEC_NAME_SIZE)) break;
            if (strncmp(SrvTbl[n].EqmName,SrvTbl[k].EqmName,EQM_NAME_SHORTSIZE)) break;
          }
        }
#       ifndef ALLOW_AMBIGUOUS
        if (!stricmp(tgtsrv,"GLOBALS")) k = SrvTblSize; 
        if (k < SrvTblSize) return ambiguous;
#       endif
      }
      for (i=0; i<FecTblSize; i++)
      {
        if (FecInfoEx[i].disable) continue;
        if (!strnicmp(SrvTbl[n].FecName,FecTbl[i].fecName,FEC_NAME_SIZE))
        { /* found it */
          FecAddrStruct fec = FecTbl[i];
          if (tineDebug) dbglog("ENS: return info for srv %.64s",tgtsrv);
          if (deliverAbsoluteIp4Address)
          {
            BYTE b4adr[16];
            inet_pton(AF_INET,FecInfoEx[i].ipv4,b4adr);
            if (*(UINT32 *)b4adr != 0) 
            {
              memset(fec.strAdr,0,ADDR_SIZE_IPV6);
              strncpy(fec.strAdr,FecInfoEx[i].ipv4,16);
            }
            toIpv4Addr(fec.strAdr,fec.strAdr);
            strAdr2netAdr(&fec);
          }
          return prepSrvAddrRequest(dbuf,&SrvTbl[n],&fec,fmt,dsize);
        }
      }  
      return no_such_computer;
    case INF_DEV_TAG: /* FEC extended info query */
      if (tineDebug) dbglog("ENS: request extended info for fec %.64s",tgtsrv);
      for (i=0; i<FecTblSize; i++)
      {
        if (FecInfoEx[i].disable) continue;
        if (!strnicmp(tgtsrv,FecTbl[i].fecName,FEC_NAME_SIZE)) break;
      }
      if (i == FecTblSize) return no_such_computer;
      if (fmt == CF_TEXT)
      { /* extended info query */
        int pktsize = 0;
        if (dsize < LEGACY_PRP_FECINFO_BASESIZE) return illegal_data_size; /* true in general */
        if ((dsize-LEGACY_PRP_FECINFO_BASESIZE) % EXPORT_NAME_SHORTSIZE == 0) 
        {
          pktsize = ADDRPKTSIZE4L;
          isLegacyCall = TRUE;
        }
        else
        {
          if ((dsize-PRP_FECINFO_BASESIZE) % EXPORT_NAME_SIZE != 0) return illegal_data_size;
          pktsize = ADDRPKTSIZE4;
        }
        for (i=0; i<FecTblSize; i++)
        {
          if (FecInfoEx[i].disable) continue;
          if (!strnicmp(tgtsrv,FecTbl[i].fecName,FEC_NAME_SIZE)) 
          { /* found it */
            n = prepFecInfoRequest(dbuf,i,&FecInfoTbl[i],dsize,pktsize);
            dout->dArrayLength = isLegacyCall ? LEGACY_PRP_FECINFO_BASESIZE + n*EXPORT_NAME_SHORTSIZE :
                                                PRP_FECINFO_BASESIZE + n*EXPORT_NAME_SIZE;
            return 0;
          }
        }
      }
      /* return the fec importance */
      return putValuesFromShort(dout,&FecInfoEx[i].importance,1);
    default: break;
  }
  return illegal_property;
}
